package animatlab.kodamat.core.adds;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.io.DataLoadingPolicy;
import animatlab.kodamat.core.io.IToDot;
import animatlab.kodamat.core.io.IToXml;
import animatlab.kodamat.core.io.XmlUtils;
import animatlab.kodamat.core.io.DotUtils.ImageMap;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;

/**
 * 
 * Algebraic decision diagram
 * 
 * @author Thomas Degris, Tomas Simovic
 * @param <T>
 *          type of data contained in this ADD
 * 
 */
public class ADD<T> implements IToXml<T>, IFunction<T>, IToDot<T> {
	/**
	 * 
	 * Namespace of xml node name
	 * 
	 */
	public static final String NAMESPACE = "cud";

	/**
	 * 
	 * Name of the node in xml to represent an edge
	 * 
	 */
	public static final String NODE_NAME = NAMESPACE + ":" + "add";

	/**
	 * 
	 * Name of the node in xml to represent an edge
	 * 
	 */
	public static final String NODE_NAME_LEAF = NAMESPACE + ":" + "leaf";

	/**
	 * 
	 * Name of the node in xml to represent an edge
	 * 
	 */
	public static final String NODE_NAME_EDGE = NAMESPACE + ":" + "edge";

	/**
	 * 
	 * Name of the node in xml to represent an edge
	 * 
	 */
	public static final String NODE_NAME_DECISION = NAMESPACE + ":" + "decision";

	/**
	 * 
	 * Name of the node in xml to represent an edge
	 * 
	 */
	public static final String DOT_NAME_LEAF = "leaf";

	/**
	 * 
	 * Name of the node in xml to represent an edge
	 * 
	 */
	public static final String DOT_NAME_DNODE = "dnode";

	/**
	 * 
	 * Attribute used in a xml node to specify the variable to test
	 * 
	 */
	public static final String ATTRIBUTE_VARIABLE_TESTED = "variable";

	/**
	 * 
	 * Attribute used in a xml node to specify the identity of the ADD node
	 * 
	 */
	public static final String ATTRIBUTE_IDENTITY = "id";

	/**
	 * 
	 * Attribute used in a xml node to specify the origin of the edge
	 * 
	 */
	private static final String ATTRIBUTE_FROM = "from";

	/**
	 * 
	 * Attribute used in a xml node to specify the destination of the edge
	 * 
	 */
	private static final String ATTRIBUTE_TO = "to";

	/**
	 * 
	 * Attribute used in a xml node to specify the side of the edge
	 * 
	 */
	private static final String ATTRIBUTE_SIDE = "side";

	/**
	 * 
	 * Value of the attribute used in a xml node to specify the side of the edge
	 * 
	 */
	private static final String ATTRIBUTE_SIDE_LEFT = "left";

	/**
	 * 
	 * Value of the attribute used in a xml node to specify the side of the edge
	 * 
	 */
	private static final String ATTRIBUTE_SIDE_RIGHT = "right";

	/**
	 * 
	 * List of edge in the ADD
	 * 
	 */
	private final List<DNode<T>> dnodeList = new ArrayList<DNode<T>>();

	/**
	 * 
	 * List of leaves in the ADD
	 * 
	 */
	private final List<TerminalNode<T>> tnodeList = new ArrayList<TerminalNode<T>>();

	/**
	 * 
	 * The root of this tree
	 * 
	 */
	protected ADDNode<T> root = null;

	/**
	 * 
	 * Constructor from a xml tree
	 * 
	 * @param root
	 *          root of the xml tree
	 * @param dataLoadingPolicy
	 *          the policy to use to create the nodes
	 * 
	 */
	public ADD(Element root, DataLoadingPolicy<T> dataLoadingPolicy) {
		assert (NODE_NAME.equals(root.getNodeName()));
		boolean rootSet = false;
		NodeList nodeList = root.getChildNodes();
		Map<String, ADDNode<T>> loadingMap = new LinkedHashMap<String, ADDNode<T>>();
		for (int index = 0; index < nodeList.getLength(); index++) {
			Node childNode = nodeList.item(index);
			if (childNode.getNodeType() == Node.ELEMENT_NODE) {
				Element element = (Element) childNode;
				ADDNode<T> createdNode = null;
				if (childNode.getNodeName().equals(NODE_NAME_LEAF)) {
					createdNode = createTerminal(dataLoadingPolicy.xmlToData(element));
					if (rootSet == false) {
						this.root = createdNode; // ugly!!!
						rootSet = true;
					}
				}
				if (childNode.getNodeName().equals(NODE_NAME_DECISION)) {
					String variable = element.getAttribute(ATTRIBUTE_VARIABLE_TESTED);
					createdNode = createDecision(Variable.getVariable(variable
							.replaceFirst("var", "")));
					if (rootSet == false) {
						this.root = createdNode;
						rootSet = true;
					}
				}
				if (createdNode != null) {
					loadingMap.put(element.getAttribute(ATTRIBUTE_IDENTITY), createdNode);
				}
				if (childNode.getNodeName().equals(NODE_NAME_EDGE)) {
					String idFrom = element.getAttribute(ATTRIBUTE_FROM);
					String idTo = element.getAttribute(ATTRIBUTE_TO);
					DNode<T> fromNode = (DNode<T>) loadingMap.get(idFrom);
					ADDNode<T> toNode = loadingMap.get(idTo);
					if (element.getAttribute(ATTRIBUTE_SIDE).equals(ATTRIBUTE_SIDE_LEFT)) {
						fromNode.setLeft(toNode);
						toNode.setValue("0");
					} else {
						fromNode.setRight(toNode);
						toNode.setValue("1");
					}
				}
			}
		}
	}

	/**
	 * 
	 * Add a terminal node in the ADD
	 * 
	 * @param data
	 *          the data contained at the leaf
	 * @return the new terminal node
	 * 
	 */
	private TerminalNode<T> createTerminal(T data) {
		TerminalNode<T> result = new TerminalNode<T>(this, data);
		tnodeList.add(result);
		return result;
	}

	/**
	 * 
	 * Add a decision node in the ADD
	 * 
	 * @param variable
	 *          the variable tested in the new decision node
	 * @return the new decision node
	 * 
	 */
	private DNode<T> createDecision(Variable variable) {
		DNode<T> result = new DNode<T>(this, variable);
		dnodeList.add(result);
		return result;
	}

	/**
	 * @see animatlab.kodamat.core.io.IToXml#toXml(org.w3c.dom.Document,
	 *      animatlab.kodamat.core.io.IToXml.IToXmlPolicy)
	 */
	public Element toXml(Document document, IToXmlPolicy<T> policy) {
		XmlUtils.declareNS(document, NAMESPACE);
		Element resultNode = document.createElement(NODE_NAME);
		Map<ADDNode<T>, String> identityMap = new LinkedHashMap<ADDNode<T>, String>();
		for (int index = 0; index < dnodeList.size(); index++) {
			DNode<T> decisionNode = dnodeList.get(index);
			Element decisionElement = document.createElement(NODE_NAME_DECISION);
			String identity = String.valueOf(index);
			identityMap.put(decisionNode, identity);
			decisionElement.setAttribute(ATTRIBUTE_IDENTITY, identity);
			decisionElement.setAttribute(ATTRIBUTE_VARIABLE_TESTED, String
					.valueOf(decisionNode.getVariableTested()));
			resultNode.appendChild(decisionElement);
		}
		for (int index = 0; index < tnodeList.size(); index++) {
			TerminalNode<T> terminalNode = tnodeList.get(index);
			Element terminalElement = document.createElement(NODE_NAME_LEAF);
			String identity = String.valueOf(index + dnodeList.size());
			identityMap.put(terminalNode, identity);
			terminalElement.setAttribute(ATTRIBUTE_IDENTITY, identity);
			policy.toXml(terminalElement, terminalNode.get());
			resultNode.appendChild(terminalElement);
		}
		for (int index = 0; index < dnodeList.size(); index++) {
			DNode<T> decisionNode = dnodeList.get(index);

			Element leftEdgeElement = document.createElement(NODE_NAME_EDGE);
			leftEdgeElement.setAttribute(ATTRIBUTE_FROM, identityMap
					.get(decisionNode));
			leftEdgeElement.setAttribute(ATTRIBUTE_TO, identityMap.get(decisionNode
					.left()));
			leftEdgeElement.setAttribute(ATTRIBUTE_SIDE, ATTRIBUTE_SIDE_LEFT);
			resultNode.appendChild(leftEdgeElement);

			Element rightEdgeElement = document.createElement(NODE_NAME_EDGE);
			rightEdgeElement.setAttribute(ATTRIBUTE_FROM, identityMap
					.get(decisionNode));
			rightEdgeElement.setAttribute(ATTRIBUTE_TO, identityMap.get(decisionNode
					.right()));
			rightEdgeElement.setAttribute(ATTRIBUTE_SIDE, ATTRIBUTE_SIDE_RIGHT);
			resultNode.appendChild(rightEdgeElement);
		}
		return resultNode;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuilder result = new StringBuilder();
		for (DNode<T> node : dnodeList) {
			result.append(node.toString());
			result.append("\n");
		}
		return result.toString();
	}

	/**
	 * 
	 * Get the index of the decision node in the decision node list
	 * 
	 * @param decisionNode
	 *          the decision node to get in the decision node list
	 * 
	 * @return the index of the decision node to get in the decision node list
	 * 
	 */
	public int indexOf(DNode<T> decisionNode) {
		return dnodeList.indexOf(decisionNode);
	}

	/**
	 * @see animatlab.kodamat.core.functions.IFunction#classify(Context)
	 */
	public T classify(Context state) {
		return getTerminalNode(state).get();
	}

	/**
	 * 
	 * @return the number of nodes in this structure
	 * 
	 */
	public int numberOfNodes() {
		return tnodeList.size() + dnodeList.size();
	}

	/**
	 * 
	 * Get the leaf corresponding to a vector of values
	 * 
	 * @param values
	 *          the vector of values to classify in the add
	 * 
	 * @return the ADDNode corresponding to the class of the vector of values
	 * 
	 */
	public TerminalNode<T> getTerminalNode(Context values) {
		return getTerminalNode(getRoot(), values);
	}

	/**
	 * 
	 * Get the leaf corresponding to a vector of values
	 * 
	 * @param node
	 *          root of the subtree to look for
	 * @param state
	 *          the vector of values to classify in the decision tree
	 * 
	 * @return the ADDNode corresponding to the class of the vector of values
	 * 
	 */

	public TerminalNode<T> getTerminalNode(ADDNode<T> node, Context state) {
		ADDNode<T> currentNode = node;
		while (!currentNode.terminal) {
			DNode<T> decisionNode = (DNode<T>) currentNode;
			currentNode = decisionNode.getChild(state);
		}
		return (TerminalNode<T>) currentNode;
	}

	/**
	 * 
	 * @return Returns the root of this add.
	 * 
	 */
	public ADDNode<T> getRoot() {
		return root;
	}

	/**
	 * @see java.lang.Object#clone()
	 */
	@Override
	public ADD<T> clone() {
		assert (false);
		return null;
	}

	/**
	 * @see animatlab.kodamat.core.io.IToDot#toDot(java.lang.StringBuilder,
	 *      animatlab.kodamat.core.io.IToDot.IToDotPolicy, ImageMap)
	 */
	public void toDot(StringBuilder output, IToDotPolicy<T> policy,
			ImageMap<T> imageMap) {
		output.append("digraph G {\n");
		Map<ADDNode<T>, String> identityMap = new LinkedHashMap<ADDNode<T>, String>();
		for (int index = 0; index < dnodeList.size(); index++) {
			DNode<T> dnode = dnodeList.get(index);
			String identity = DOT_NAME_DNODE + String.valueOf(index);
			identityMap.put(dnode, identity);
			output.append("  ");
			output.append(identity);
			output.append("[texlbl=\"");
			output.append(dnode.getVariableTested().toString());
			output.append("\",label=\"");
			output.append(dnode.getVariableTested().toString());
			output.append("\"];\n");
		}
		for (int index = 0; index < tnodeList.size(); index++) {
			TerminalNode<T> tnode = tnodeList.get(index);
			String identity = DOT_NAME_LEAF + String.valueOf(index);
			identityMap.put(tnode, identity);
			output.append("  ");
			output.append(identity);
			String label = "null";
			if ((tnode.get() != null) && (policy == null)) {
				label = tnode.get().toString();
			} else {
				if (policy != null) {
					label = policy.toDotLabel(tnode.get());
				}
			}
			output.append("[texlbl=\"");
			output.append(label);
			output.append("\",label=\"");
			output.append(label);
			output.append("\"];\n");
		}
		for (int index = 0; index < dnodeList.size(); index++) {
			DNode<T> dnode = dnodeList.get(index);
			String dnodeId = identityMap.get(dnode);
			output.append("  ");
			output.append(dnodeId);
			output.append("->");
			output.append(identityMap.get(dnode.left()));
			output.append("[style=\"filled\"];\n  ");
			output.append(dnodeId);
			output.append("->");
			output.append(identityMap.get(dnode.right()));
			output.append("[style=\"dotted\"];\n");
		}
		output.append("}\n");
	}

	/**
	 * @see animatlab.kodamat.core.io.IToDot#dotName()
	 */
	public String dotName() {
		return "";
	}
}
