/**
 * 
 **/
package animatlab.kodamat.core.conditions;

import java.util.Random;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import animatlab.kodamat.core.io.XmlUtils;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;

/**
 * 
 * Condition on one variable
 * 
 * @author Thomas Degris
 * 
 */
public class VariableCondition implements IStateCondition {
	/**
	 * 
	 * Node for a condition on a variable
	 * 
	 */
	static public final String NODE_NAME = NAMESPACE + ":" + "node";

	/**
	 * 
	 * Attribute indicating the value required to be true
	 * 
	 */
	static public final String ATTRIBUTE_VALUE_NAME = "value";

	/**
	 * 
	 * Variable tested on this condition
	 * 
	 */
	private Variable variable;

	/**
	 * 
	 * Value required to make this condition true
	 * 
	 */
	private String valueRequired;

	/**
	 * Constructor : create the condition from a xml node
	 * 
	 * @param xmlNode
	 *          the xml node to build the condition from
	 * 
	 */
	VariableCondition(Node xmlNode) {
		this(Variable.load(xmlNode), xmlNode.getAttributes().getNamedItem(
				ATTRIBUTE_VALUE_NAME).getTextContent());
		assert (xmlNode.getNodeName().equals(NODE_NAME));
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param variable
	 *          variable to test
	 * @param value
	 *          value required for the variable
	 * 
	 */
	public VariableCondition(Variable variable, String value) {
		this.variable = variable;
		this.valueRequired = value;
	}

	/**
	 * @see animatlab.kodamat.core.conditions.IStateCondition#isTrue(Context)
	 */
	public boolean isTrue(Context state) {
		return valueRequired.equals(state.get(variable));
	}

	/**
	 * @see animatlab.kodamat.core.conditions.IStateCondition#setToTrue(java.util.Random,
	 *      Context)
	 */
	public void setToTrue(Random random, Context state) {
		state.put(variable, valueRequired);
	}

	/**
	 * Convert a xml node into a set of state conditions
	 * 
	 * @param xmlNode
	 *          the xml node to convert
	 * 
	 * @return the condition build from the xml structure
	 * 
	 */
	public static IStateCondition load(Node xmlNode) {
		if (xmlNode.getNodeName().equals(NODE_NAME)) {
			return new VariableCondition(xmlNode);
		}
		if (xmlNode.getNodeName().equals(OperatorCondition.NODE_NAME)) {
			String operatorName = xmlNode.getAttributes().getNamedItem(
					OperatorCondition.ATTRIBUTE_OPERATOR_NAME).getNodeValue();
			if (operatorName.equals(OrCondition.OPERATOR_OR_NAME)) {
				return new OrCondition(xmlNode);
			}
			if (operatorName.equals(AndCondition.OPERATOR_AND_NAME)) {
				return new AndCondition(xmlNode);
			}
		}

		assert (false);
		return null;
	}

	/**
	 * @see animatlab.kodamat.core.io.IToXml#toXml(org.w3c.dom.Document,
	 *      animatlab.kodamat.core.io.IToXml.IToXmlPolicy)
	 */
	public Element toXml(Document document, IToXmlPolicy<Object> policy) {
		XmlUtils.declareNS(document, NAMESPACE);
		Element variableNode = document.createElement(NODE_NAME);
		variable.setAttribute(variableNode);
		variableNode.setAttribute(ATTRIBUTE_VALUE_NAME, valueRequired);
		return variableNode;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuffer result = new StringBuffer("(");
		result.append(variable.toString());
		result.append("=");
		result.append(valueRequired);
		result.append(")");
		return result.toString();
	}

}
