package animatlab.kodamat.core.functions;

import java.util.ArrayList;
import java.util.List;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import animatlab.kodamat.core.io.IToDot;
import animatlab.kodamat.core.io.IToXml;
import animatlab.kodamat.core.io.DotUtils.ImageMap;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.VariableOrdering;

/**
 * 
 * Weighted sum of functions
 * 
 * @author Thomas Degris
 * 
 */
public class FunctionsSum implements IFunction<Double>, IToXml<Double>,
		IToDot<Double> {
	/**
	 * 
	 * SerialVersionUID
	 * 
	 */
	private static final long serialVersionUID = 2338058621609715764L;

	/**
	 * 
	 * Name of the node describing the function
	 * 
	 */
	public static final String NODE_NAME = "fst:weightedfunctionssum";

	/**
	 * 
	 * Attribute used to specify the weight of each function
	 * 
	 */
	private static final String ATTRIBUTE_WEIGHT = "weight";

	/**
	 * 
	 * List of function to use in the sum
	 * 
	 */
	private final ArrayList<IFunction<Double>> functions = new ArrayList<IFunction<Double>>();

	/**
	 * 
	 * Weight associated to each function
	 * 
	 */
	private final ArrayList<Double> weight = new ArrayList<Double>();

	/**
	 * 
	 * Constructor
	 * 
	 * 
	 */
	public FunctionsSum() {
		super();
	}

	/**
	 * 
	 * Constructor to load the sum from a xml node
	 * 
	 * @param element
	 *          root of the xml tree
	 * @param variableOrdering
	 *          legend associated the variables used in this tree
	 * @param loadingPolicy
	 *          policy to use to load the functions
	 * 
	 */
	public FunctionsSum(Element element, VariableOrdering variableOrdering,
			IPartitionFunction.LoadXmlPolicy<Double> loadingPolicy) {
		loadXml(element, loadingPolicy);
	}

	/**
	 * 
	 * Load this sum from a xml node
	 * 
	 * @param element
	 *          root of the xml tree
	 * @param loadingPolicy
	 *          policy to use to load the functions
	 * 
	 */
	public void loadXml(Element element,
			IPartitionFunction.LoadXmlPolicy<Double> loadingPolicy) {
		assert (NODE_NAME.equals(element.getNodeName()));
		NodeList nodeList = element.getChildNodes();
		for (int index = 0; index < nodeList.getLength(); index++) {
			Node childNode = nodeList.item(index);
			if (childNode.getNodeType() == Node.ELEMENT_NODE) {
				Element functionNode = (Element) childNode;
				IFunction<Double> function = loadingPolicy.loadXml(functionNode);
				double nodeWeight = 1.0;
				if (functionNode.getAttribute(ATTRIBUTE_WEIGHT).length() > 0) {
					nodeWeight = Double.parseDouble(functionNode
							.getAttribute(ATTRIBUTE_WEIGHT));
				}
				add(function, nodeWeight);
			}
		}
	}

	/**
	 * 
	 * Add a function to this sum
	 * 
	 * @param function
	 *          the function to add
	 * @param functionWeight
	 *          the weight of the function in the sum
	 * 
	 */
	public void add(IFunction<Double> function, double functionWeight) {
		functions.add(function);
		weight.add(functionWeight);
	}

	/**
	 * 
	 * Add a function to this sum (with a weight of 1.0)
	 * 
	 * @param function
	 *          the function to add
	 * 
	 */
	public void add(IFunction<Double> function) {
		add(function, 1.0);
	}

	/**
	 * @see animatlab.kodamat.core.functions.IFunction#classify(Context)
	 */
	public Double classify(Context state) {
		double result = 0;
		for (int index = 0; index < size(); index++) {
			Double value = functions.get(index).classify(state);
			if (value != null) {
				result = result + value * weight.get(index);
			}
		}
		return result;
	}

	/**
	 * @see animatlab.kodamat.core.functions.IFunction#numberOfNodes()
	 */
	public int numberOfNodes() {
		int result = 0;
		for (IFunction<Double> rules : functions) {
			result = result + rules.numberOfNodes();
		}
		return result;
	}

	/**
	 * @see animatlab.kodamat.core.io.IToXml#toXml(org.w3c.dom.Document,
	 *      animatlab.kodamat.core.io.IToXml.IToXmlPolicy)
	 */
	@SuppressWarnings("unchecked")
	public Element toXml(Document document, IToXmlPolicy<Double> policy) {
		Element result = document.createElement(NODE_NAME);
		for (int index = 0; index < size(); index++) {
			Element functionNode = ((IToXml<Double>) functions.get(index)).toXml(
					document, IToXml.doubleToXmlPolicy);
			result.appendChild(functionNode);
		}
		return result;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuilder result = new StringBuilder("{");
		for (int index = 0; index < size(); index++) {
			result.append(String.valueOf(weight.get(index)) + " * ");
			result.append(functions.get(index));
			if (index != (functions.size() - 1)) {
				result.append(" + ");
			}
		}
		result.append("}");
		return result.toString();
	}

	/**
	 * @see java.lang.Object#clone()
	 */
	@Override
	public FunctionsSum clone() {
		FunctionsSum result = new FunctionsSum();
		for (int index = 0; index < size(); index++) {
			result.add(functions.get(index).clone(), weight.get(index));
		}
		return result;
	}

	/**
	 * 
	 * @return number of function in the sum
	 * 
	 */
	public int size() {
		return functions.size();
	}

	/**
	 * 
	 * Set the weight of one of the function
	 * 
	 * @param index
	 *          the index of the function to set the weight for
	 * @param weight
	 *          the weight of the function
	 * 
	 */
	public void setWeight(int index, double weight) {
		this.weight.set(index, weight);
	}

	/**
	 * 
	 * @return the list of function in the sum
	 * 
	 */
	public List<IFunction<Double>> functions() {
		return functions;
	}

	/**
	 * 
	 * Remove all the functions of this sum
	 * 
	 */
	public void clear() {
		functions.clear();
		weight.clear();
	}

	/**
	 * 
	 * The weight for a given index
	 * 
	 * @param index
	 *          the index to get the weight from
	 * 
	 * @return the value of the weight
	 * 
	 */
	public double getWeight(int index) {
		return weight.get(index);
	}

	/**
	 * 
	 * Set a new function at a specific position
	 * 
	 * @param index
	 *          index of the function to set
	 * @param function
	 *          the new function to set
	 * 
	 */
	public void setFun(int index, IFunction<Double> function) {
		functions.set(index, function);
	}

	/**
	 * @see animatlab.kodamat.core.io.IToDot#dotName()
	 */
	public String dotName() {
		return "";
	}

	/**
	 * @see animatlab.kodamat.core.io.IToDot#toDot(java.lang.StringBuilder,
	 *      animatlab.kodamat.core.io.IToDot.IToDotPolicy, ImageMap)
	 */
	@SuppressWarnings("unchecked")
	public void toDot(StringBuilder output,
			animatlab.kodamat.core.io.IToDot.IToDotPolicy<Double> policy,
			ImageMap<Double> imageMap) {
		output.append("digraph G {\n");
		output.append("ordering=out\n");
		for (IFunction<Double> function : functions()) {
			Tree<Double> tree = (Tree<Double>) function;
			for (IPartition<Double> leaf : tree.partitions()) {
				((IToDot<Double>) leaf).toDot(output, policy, imageMap);
			}
			tree.getRoot().toDot(output, policy, imageMap);
		}
		output.append("}\n");
	}

	/**
	 * 
	 * Remove a function from the sum
	 * 
	 * @param function
	 *          the function to remove
	 * 
	 */
	public void remove(IFunction<Double> function) {
		int index = functions().indexOf(function);
		if (index == -1) {
			return;
		}
		functions.remove(index);
		weight.remove(index);
	}
}
