package animatlab.kodamat.core.functions;

import animatlab.kodamat.core.utils.Parameters;

/**
 * 
 * Policy to apply on rule in rules
 * 
 * @param <T>
 *          type of the rules
 * 
 */
public abstract class UnaryOperatorPolicy<T> {
	/**
	 * 
	 * Policy to multiply a set of rules by a constant
	 * 
	 */
	public static class MultiplyOperator extends UnaryOperatorPolicy<Double> {
		/**
		 * 
		 * Constant used to multiply the value of the rules
		 * 
		 */
		private double constant;

		/**
		 * 
		 * Constructor
		 * 
		 */
		public MultiplyOperator() {
			super();
		}

		/**
		 * 
		 * Constructor
		 * 
		 * @param constant
		 *          the constant to use when applying the operator
		 * 
		 */
		public MultiplyOperator(double constant) {
			this.constant = constant;
		}

		/**
		 * 
		 * Operator implementation
		 * 
		 * @param value
		 *          value of the rules
		 * @return the result of the multiplication with constant
		 * 
		 */
		@Override
		public Double apply(Double value) {
			assert (!Double.isNaN(value));
			return value * constant;
		}

		/**
		 * 
		 * Set constant
		 * 
		 * @param constant
		 *          The constant to set.
		 * 
		 */
		public void setConstant(double constant) {
			this.constant = constant;
		}
	}

	/**
	 * 
	 * Operator to apply GAMMA on a function
	 * 
	 */
	public static final UnaryOperatorPolicy.MultiplyOperator discountFactor = new UnaryOperatorPolicy.MultiplyOperator(
			Parameters.GAMMA);

	/**
	 * 
	 * Operator to apply GAMMA on a function
	 * 
	 */
	public static final UnaryOperatorPolicy.MultiplyOperator minusOneFactor = new UnaryOperatorPolicy.MultiplyOperator(
			-1.0);

	/**
	 * 
	 * Math.abs operator
	 * 
	 */
	public static final UnaryOperatorPolicy<Double> absOperator = new UnaryOperatorPolicy<Double>() {
		@Override
		protected Double apply(Double value) {
			return Math.abs(value);
		}
	};

	/**
	 * 
	 * Method applied on each rule
	 * 
	 * @param value
	 *          pointed value of the rule
	 * 
	 * @return the new value
	 * 
	 */
	abstract protected T apply(T value);

	/**
	 * 
	 * Apply a policy on the rules on all this set
	 * 
	 * @param function
	 *          function to apply this policy on
	 * 
	 */
	final public void applyPolicy(IPartitionFunction<T> function) {
		for (IPartition<T> partition : function.partitions()) {
			partition.set(apply(partition.get()));
		}
	}
}