package animatlab.kodamat.core.io;

import java.awt.Rectangle;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintStream;
import java.io.UnsupportedEncodingException;
import java.util.LinkedHashMap;
import java.util.Map;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import animatlab.kodamat.core.io.IToDot.IToDotPolicy;

/**
 * 
 * Some function to output dot file
 * 
 * @author Thomas Degris
 * 
 */
public class DotUtils {
	/**
	 * 
	 * Map containing position to object
	 * 
	 * @param <T>
	 *          type pointed by node in the image
	 * 
	 */
	static public class ImageMap<T> extends LinkedHashMap<Rectangle, T> {
		/**
		 * 
		 * Serial version UID
		 * 
		 */
		private static final long serialVersionUID = 1L;

		/**
		 * 
		 * Map from dot name to object
		 * 
		 */
		Map<String, T> dotNameToObject = null;

		/**
		 * Constructor
		 */
		public ImageMap() {
			super();
		}

		/**
		 * 
		 * Register a new dot name
		 * 
		 * @param dotName
		 *          the dot name to register
		 * @param object
		 *          the object corresponding to the dot name
		 * @return the string to add for the dot object
		 * 
		 */
		@SuppressWarnings("unchecked")
		public String registerDotName(String dotName, Object object) {
			if (dotNameToObject == null) {
				return "";
			}
			dotNameToObject.put(dotName, (T) object);
			return "URL=\"" + dotName + "\" ";
		}

		/**
		 * 
		 * Parse a map file to get coordinate for each objects
		 * 
		 * @param mapPath
		 *          path to the file map to load
		 * 
		 */
		public void parse(String mapPath) {
			Element rootNode = XmlUtils.getRootFromFile(mapPath);
			assert (rootNode.getNodeName().equals("map"));
			NodeList nodeList = rootNode.getChildNodes();
			for (int index = 0; index < nodeList.getLength(); index++) {
				Node childNode = nodeList.item(index);
				if (childNode.getNodeType() == Node.ELEMENT_NODE) {
					if (childNode.getNodeName().equals("area")) {
						Element areaNode = (Element) childNode;
						registerArea(areaNode.getAttribute("href"), areaNode
								.getAttribute("coords"));
					}
				}
			}
			assert (dotNameToObject.size() == size());
		}

		/**
		 * 
		 * Register an area
		 * 
		 * @param dotName
		 *          the dot name associated to the area
		 * @param coords
		 *          the coordinate of the area
		 * 
		 */
		private void registerArea(String dotName, String coords) {
			String[] splitedCoord = coords.split(",");
			int x = Integer.parseInt(splitedCoord[0]);
			int y = Integer.parseInt(splitedCoord[1]);
			int width = Integer.parseInt(splitedCoord[2]) - x;
			int height = Integer.parseInt(splitedCoord[3]) - y;
			put(new Rectangle(x, y, width, height), dotNameToObject.get(dotName));
		}

		/**
		 * 
		 * Get the object given a set x,y coordinates
		 * 
		 * @param x
		 *          the x coordinate
		 * @param y
		 *          the y coordinate
		 * @return the object corresponding the coordinates
		 * 
		 */
		public T get(int x, int y) {
			for (Map.Entry<Rectangle, T> entry : entrySet()) {
				if (entry.getKey().contains(x, y)) {
					return entry.getValue();
				}
			}
			return null;
		}
	}

	/**
	 * 
	 * Show a data structure into a dot viewer
	 * 
	 * @param itodot
	 *          the data structure to show
	 * @param policy
	 *          the policy to use to output the data structure into a dot file
	 * 
	 */
	static public void show(IToDot<?> itodot, IToDot.IToDotPolicy<?> policy) {
		File dotBuffer = null;
		try {
			dotBuffer = File.createTempFile("spiti", ".dot");
		} catch (IOException e) {
			e.printStackTrace();
			return;
		}
		System.out.println("Saving in " + dotBuffer.getAbsolutePath() + "...");
		DotUtils.save(itodot, policy, dotBuffer.getAbsolutePath(), null);
		String outputFile = dotBuffer.getAbsolutePath().replaceAll("dot", "ps");
		convert(dotBuffer.getAbsolutePath(), outputFile, "ps", null);
		Exec.exec("kghostview " + outputFile);
	}

	/**
	 * 
	 * Save a data structure into a dot file
	 * 
	 * @param itodot
	 *          the data structure to save
	 * @param policy
	 *          the policy to use to output the data structure
	 * @param filepath
	 *          the path of the file to use
	 * @param format
	 *          format used to export the dot file
	 * @param imageMap
	 *          map of the image to fill
	 * 
	 */
	static public void export(IToDot<?> itodot, IToDot.IToDotPolicy<?> policy,
			String filepath, String format, ImageMap<?> imageMap) {
		try {
			File dotFile = File.createTempFile("todot", ".dot");
			save(itodot, policy, dotFile.getAbsolutePath(), imageMap);
			convert(dotFile.getAbsolutePath(), filepath, format, imageMap);
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
		}
	}

	/**
	 * 
	 * Save a data structure into a dot file
	 * 
	 * @param itodot
	 *          the data structure to save
	 * @param policy
	 *          the policy to use to output the data structure
	 * @param filepath
	 *          the path of the file to use
	 * @param imageMap
	 *          map of the image to fill
	 * 
	 */
	@SuppressWarnings("unchecked")
	public static void save(IToDot itodot, IToDotPolicy policy, String filepath,
			ImageMap imageMap) {
		PrintStream output = null;
		try {
			output = new PrintStream(filepath, "UTF-8");
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			System.exit(1);
			return;
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
			System.exit(1);
			return;
		}
		StringBuilder buffer = new StringBuilder();
		if (policy == null) {
			policy = IToDot.toStringPolicy;
		}
		if (imageMap == null) {
			imageMap = new ImageMap();
		} else {
			imageMap.dotNameToObject = new LinkedHashMap<String, Object>();
		}
		itodot.toDot(buffer, policy, imageMap);
		output.print(buffer.toString());
		output.flush();
		output.close();
	}

	/**
	 * 
	 * Output a data structure into a html file using dot
	 * 
	 * @param itodot
	 *          the data structure to output
	 * @param policy
	 *          the policy used to output the data structure
	 * @param folder
	 *          folder where to outputs the figure
	 * @param filename
	 *          name of the file to use
	 * 
	 * @return the corresponding html source
	 * 
	 */
	public static String toHtmlDot(IToDot<?> itodot, IToDotPolicy<?> policy,
			String folder, String filename) {
		StringBuilder result = new StringBuilder();
		result.append("<center>\n" + "  <img align='center' src='" + filename
				+ ".png'/><br/>\n" + "  <p>\n");
		String dotFile = folder + "/" + filename + ".dot";
		save(itodot, policy, dotFile, null);
		String[] formats = { "png" }; // { "png", "svg", "ps" };
		result.append("    <a href='./" + filename + ".dot'>" + filename
				+ ".dot</a>\n");
		for (String format : formats) {
			convert(dotFile, folder + "/" + filename + "." + format, format, null);
			result.append("    <a href='./" + filename + "." + format + "'>"
					+ filename + "." + format + "</a>\n");
		}
		result.append("</p>\n" + "</center>\n");
		return result.toString();
	}

	/**
	 * 
	 * Convert a dot file into a figure
	 * 
	 * @param dotFile
	 *          the dot file to convert
	 * @param outputFile
	 *          the filename of the output file (with the extension)
	 * @param format
	 *          the format to convert dot file into
	 * @param imageMap
	 *          map from position to object in the generated image
	 * 
	 */
	public static void convert(String dotFile, String outputFile, String format,
			ImageMap<?> imageMap) {
		Exec.exec("dot -T" + format + " " + dotFile + " -o " + outputFile);
		if (imageMap != null) {
			File mapFile = null;
			try {
				mapFile = File.createTempFile("dot", ".map");
			} catch (IOException e) {
				e.printStackTrace();
				return;
			}
			Exec.exec("dot -Tcmapx " + dotFile + " -o " + mapFile.getAbsolutePath());
			imageMap.parse(mapFile.getAbsolutePath());
			mapFile.delete();
		}
	}

	/**
	 * 
	 * Convert a string into a normalized string (with no funny char in it)
	 * 
	 * @param str
	 *          the string to convert
	 * 
	 * @return the normalize string
	 * 
	 */
	public static String toNormalizeString(String str) {
		String result = str.replaceAll("\\ ", "");
		result = result.replaceAll("\\-", "");
		result = result.replaceAll("\\'", "");
		result = result.replaceAll("_", "");
		result = result.replaceAll("\\?", "");
		return result;
	}
}
