package animatlab.kodamat.core.io;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * 
 * Html page page to represent dot trees
 * 
 * @author Thomas Degris
 * 
 */
public class HtmlPage {
	/**
	 * 
	 * Folder containing the html page
	 * 
	 */
	final private String folder;

	/**
	 * 
	 * File used to write the html page
	 * 
	 */
	final private String htmlPageFilename;

	/**
	 * 
	 * Title of the html page
	 * 
	 */
	private String title;

	/**
	 * 
	 * Title of the set of generated html pages
	 * 
	 */
	private String htmlPageSetTitle;

	/**
	 * 
	 * Parent of this page
	 * 
	 */
	final private HtmlPage parent;

	/**
	 * 
	 * Body of this html page
	 * 
	 */
	private final StringBuilder body = new StringBuilder();

	/**
	 * 
	 * Children of this page
	 * 
	 */
	private final Set<HtmlPage> children = new LinkedHashSet<HtmlPage>();

	/**
	 * 
	 * Content of this set of html page
	 * 
	 */
	private StringBuilder content = new StringBuilder();

	/**
	 * 
	 * Constructor
	 * 
	 * @param folder
	 *          folder where the page has to be saved
	 * @param htmlPageFileName
	 *          name of the file of the html page
	 * 
	 */
	public HtmlPage(String folder, String htmlPageFileName) {
		this(null, folder, htmlPageFileName);
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param page
	 *          parent of this page
	 * @param folder
	 *          folder where the page has to be saved
	 * @param htmlPageFileName
	 *          name of the file of the html page
	 * 
	 */
	private HtmlPage(HtmlPage page, String folder, String htmlPageFileName) {
		this.folder = folder;
		this.htmlPageFilename = htmlPageFileName;
		File folderFile = new File(folder);
		folderFile.mkdirs();
		this.parent = page;
	}

	/**
	 * 
	 * Set the title of the html page
	 * 
	 * @param title
	 *          title of the html page
	 * 
	 */
	public void setTitle(String title) {
		this.title = title;
	}

	/**
	 * 
	 * Set the title of the html page
	 * 
	 * @param htmlPageSetTitle
	 *          title of the set of html pages
	 * 
	 */
	public void setHtmlPageSetTitle(String htmlPageSetTitle) {
		this.htmlPageSetTitle = htmlPageSetTitle;
	}

	/**
	 * 
	 * @return the absolute path of the html page
	 * 
	 */
	public String getAbsolutePath() {
		return getFile().getAbsolutePath();
	}

	/**
	 * 
	 * @return the file representing the html page
	 * 
	 */
	private File getFile() {
		return new File(folder + "/" + htmlPageFilename);
	}

	/**
	 * 
	 * Write the html page into a file
	 * 
	 */
	public void write() {
		writeChildren();
		try {
			PrintStream outputStream;
			outputStream = new PrintStream(getFile());
			outputStream.println("<html>" + "  <head>" + "    <title>"
					+ htmlPageSetTitle + ": " + title + "</title>" + "  </head>"
					+ "  <body>" + "    <center><h1>" + getTitle() + "</h1></center>");

			outputStream.println(getContent());
			outputStream.println("</br>");
			outputStream.println(body.toString());
			outputStream.println("  </body>" + "</html>");
			outputStream.flush();
			outputStream.close();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}

	/**
	 * 
	 * Write all the children of this page
	 * 
	 */
	private void writeChildren() {
		for (HtmlPage child : children) {
			child.write();
		}
	}

	/**
	 * 
	 * @return the file name of this page
	 * 
	 */
	private String getFileName() {
		return htmlPageFilename;
	}

	/**
	 * 
	 * @return the table of contents of this html page
	 * 
	 */
	private String getContent() {
		if (!children.isEmpty()) {
			return content.toString();
		}
		if (parent != null) {
			return parent.getContent();
		}
		return null;
	}

	/**
	 * 
	 * @return the chapter associated to this page
	 * 
	 */
	public String getTitle() {
		if ((title == null) && (parent == null)) {
			return "Table of contents";
		}
		if (title != null) {
			return title;
		}
		return "";
	}

	/**
	 * 
	 * @return the folder where the html page is writtent
	 * 
	 */
	public String getFolder() {
		return folder;
	}

	/**
	 * 
	 * Append a string to the body of the html page
	 * 
	 * @param htmlSource
	 *          the source to append to the body of the html page
	 * 
	 */
	public void append(String htmlSource) {
		body.append(htmlSource);
	}

	/**
	 * 
	 * Append a string to the body of the html page
	 * 
	 * @param htmlSource
	 *          the source to append to the body of the html page
	 * 
	 */
	public void appendToContent(String htmlSource) {
		content.append(htmlSource);
	}

	/**
	 * 
	 * Create a new page child of this page
	 * 
	 * @param title
	 *          title of the new html page
	 * 
	 * @return the new html page
	 * 
	 */
	public HtmlPage newChild(String title) {
		return newChild(title, title);
	}

	/**
	 * 
	 * Create a new page child of this page
	 * 
	 * @param title
	 *          title of the new html page
	 * @param linkName
	 *          name of the link to point to the child
	 * 
	 * @return the new html page
	 * 
	 */
	public HtmlPage newChild(String title, String linkName) {
		HtmlPage child = new HtmlPage(this, folder, DotUtils.toNormalizeString(
				title).toLowerCase()
				+ ".html");
		child.setHtmlPageSetTitle(htmlPageSetTitle);
		child.setTitle(title);
		content.append("<a href='./" + child.getFileName() + "'>" + linkName
				+ "</a> ");
		children.add(child);
		return child;
	}

}
