/*
 * 
 * Created on 23 dec. 2005
 *
 */
package animatlab.kodamat.core.io;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import animatlab.kodamat.core.io.IToXml.IToXmlPolicy;

/**
 * 
 * Some utils function to manipulate xml data
 * 
 * @author Thomas Degris
 * 
 */
public class XmlUtils {

	/**
	 * 
	 * Name of the root used temporary to store namespace at the root of the
	 * document
	 * 
	 */
	static private final String TEMPORARY_NAMESPACE_NODE_NAME = "temporary_node_name";

	/**
	 * 
	 * Load a xml document from a file path
	 * 
	 * @param filePath
	 *          the path of the file to load
	 * 
	 * @return a xml document representing the document
	 * 
	 */
	static public Document getXmlDocFromFile(String filePath) {
		try {
			return getXmlDoc(new InputSource(new FileInputStream(filePath)));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * 
	 * Parse a string as a xml document
	 * 
	 * @param xmlString
	 *          the string of xml data
	 * 
	 * @return a xml document
	 * 
	 */
	static public Document getXmlDoc(String xmlString) {
		return getXmlDoc(new InputSource(new StringReader(xmlString)));
	}

	/**
	 * 
	 * Parse an input source to create a xml document
	 * 
	 * @param source
	 *          the source to parse
	 * 
	 * @return the xml document
	 * 
	 */
	static public Document getXmlDoc(InputSource source) {
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder;
		Document xmlDoc = null;
		try {
			builder = factory.newDocumentBuilder();
			xmlDoc = builder.parse(source);
		} catch (ParserConfigurationException e) {
			e.printStackTrace();
		} catch (SAXException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return xmlDoc;
	}

	/**
	 * 
	 * Get the root element of a xml tree represented in a string
	 * 
	 * @param xmlString
	 *          the string to parse
	 * 
	 * @return the root of the xml document
	 * 
	 */
	static public Element getRoot(String xmlString) {
		return getXmlDoc(xmlString).getDocumentElement();
	}

	/**
	 * 
	 * Get the root element of a xml tree loaded from a file
	 * 
	 * @param filePath
	 *          the path of the file to load
	 * 
	 * @return the root of the xml document
	 * 
	 */
	static public Element getRootFromFile(String filePath) {
		return getXmlDocFromFile(filePath).getDocumentElement();
	}

	/**
	 * 
	 * @return an empty xml document
	 * 
	 */
	public static Document createXmlDocument() {
		DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory
				.newInstance();
		DocumentBuilder documentBuilder = null;
		try {
			documentBuilder = documentBuilderFactory.newDocumentBuilder();
		} catch (ParserConfigurationException e) {
			e.printStackTrace();
			return null;
		}
		Document document = documentBuilder.newDocument();
		// Create a root element to store namespace
		document.appendChild(document.createElement(TEMPORARY_NAMESPACE_NODE_NAME));
		return document;
	}

	/**
	 * 
	 * Convert an instance into a xml string using a policy
	 * 
	 * @param toxml
	 *          the instance to convert
	 * @param policy
	 *          the policy to use
	 * 
	 * @return the xml string
	 * 
	 */
	@SuppressWarnings("unchecked")
	public static String toXmlString(IToXml toxml, IToXmlPolicy policy) {
		Document document = createXmlDocument();
		setDocumentRoot(document, toxml.toXml(document, policy));
		return toXmlString(document);
	}

	/**
	 * 
	 * Build a string representation of a xml document
	 * 
	 * @param element
	 *          an element belonging to the document (may be any node)
	 * @return a string representation of the document
	 * 
	 */
	public static String toXmlString(Element element) {
		return toXmlString(element.getOwnerDocument());
	}

	/**
	 * 
	 * Convert a document into a xml string
	 * 
	 * @param document
	 *          the document to convert
	 * @return the xml string
	 * 
	 */
	private static String toXmlString(Document document) {
		StringWriter writer = new StringWriter();
		toXmlWriter(document, new StreamResult(writer));
		return writer.toString();
	}

	/**
	 * 
	 * Save a xml tree into a file
	 * 
	 * @param document
	 *          the xml document to save
	 * @param filePath
	 *          the path of the file to save
	 * 
	 */
	public static void toXmlFile(Document document, String filePath) {
		FileOutputStream writer = null;
		try {
			writer = new FileOutputStream(filePath);
		} catch (FileNotFoundException e1) {
			e1.printStackTrace();
			System.exit(1);
		}
		try {
			toXmlWriter(document, new StreamResult(new OutputStreamWriter(writer,
					"utf-8")));
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
			System.exit(1);
		}
	}

	/**
	 * 
	 * Convert a xml tree into a writer
	 * 
	 * @param document
	 *          the root of the xml document to save
	 * @param result
	 *          output
	 * 
	 */
	private static void toXmlWriter(Document document, StreamResult result) {
		TransformerFactory transformerFactory = TransformerFactory.newInstance();
		transformerFactory.setAttribute("indent-number", 2);
		Transformer transformer = null;
		try {
			transformer = transformerFactory.newTransformer();
		} catch (TransformerConfigurationException e) {
			e.printStackTrace();
			return;
		}
		transformer.setOutputProperty(OutputKeys.INDENT, "yes");
		try {
			transformer.transform(new DOMSource(document), result);
		} catch (TransformerException e) {
			e.printStackTrace();
			System.exit(1);
		}
	}

	/**
	 * 
	 * Save an instance in a xml file
	 * 
	 * @param toxml
	 *          the instance to save
	 * @param policy
	 *          policy to use to output the data
	 * @param filePath
	 *          the path of the file to save
	 * 
	 */
	@SuppressWarnings("unchecked")
	public static void save(IToXml toxml, IToXmlPolicy policy, String filePath) {
		Document document = createXmlDocument();
		setDocumentRoot(document, toxml.toXml(document, policy));
		toXmlFile(document, filePath);
	}

	/**
	 * 
	 * Return the first child of a xml node
	 * 
	 * @param xmlNode
	 *          the parent of the first child to get
	 * 
	 * @return the child if found, else null
	 * 
	 */
	public static Element getFirstChild(Node xmlNode) {
		NodeList nodeList = xmlNode.getChildNodes();
		for (int index = 0; index < nodeList.getLength(); index++) {
			Node childNode = nodeList.item(index);
			if (childNode.getNodeType() == Node.ELEMENT_NODE) {
				return (Element) childNode;
			}
		}
		return null;
	}

	/**
	 * 
	 * Declare a namespace at this node
	 * 
	 * @param document
	 *          the document using the namespace
	 * @param namespace
	 *          the namespace to declare
	 * 
	 */
	public static void declareNS(Document document, String namespace) {
		Element namespaceElement = document.getDocumentElement();
		namespaceElement.setAttribute("xmlns:" + namespace,
				"http://animatlab.lip6.fr/spiti");
	}

	/**
	 * 
	 * Set the document root
	 * 
	 * @param document
	 *          document to set the root element
	 * @param root
	 *          element to set as root
	 * 
	 */
	public static void setDocumentRoot(Document document, Element root) {
		Element namespaceElement = document.getDocumentElement();
		assert (namespaceElement.getNodeName()
				.equals(TEMPORARY_NAMESPACE_NODE_NAME));
		NamedNodeMap nodeMap = namespaceElement.getAttributes();
		for (int attrIndex = 0; attrIndex < nodeMap.getLength(); attrIndex++) {
			Attr nsAttr = (Attr) nodeMap.item(attrIndex);
			root.setAttribute(nsAttr.getName(), nsAttr.getValue());
		}
		document.replaceChild(root, namespaceElement);
	}

	/**
	 * 
	 * Build a string representing a xml node (debugging purpose only: this method
	 * should be called from some code)
	 * 
	 * @param element
	 *          the node to print into a string
	 * @return a string representing the xml node
	 * 
	 */
	public static String elementToString(Element element) {
		StringBuilder result = new StringBuilder("<");
		result.append(element.getNodeName());
		result.append(" ");
		NamedNodeMap namedNodeMap = element.getAttributes();
		for (int index = 0; index < namedNodeMap.getLength(); index++) {
			Node attr = namedNodeMap.item(index);
			result.append(attr.getNodeName());
			result.append("='");
			result.append(attr.getNodeValue());
			result.append("' ");
		}
		result.append(">");
		return result.toString();
	}
}
