package animatlab.kodamat.core.proba;

import java.util.Set;

/**
 * 
 * Some usefull functions to compute Chi2 values
 * 
 * @author Thomas Degris
 * @param <T>
 *          type of data to compute Chi2 value for
 * 
 */
public class Chi2Utils<T> {
	/**
	 * 
	 * Number of iteration maximum for Numerical Recipes in C algorithms
	 * 
	 */
	private static final int ITMAX = 100;

	/**
	 * 
	 * Epsilon used in Numerical Recipes in C algorithms
	 * 
	 */
	private static final double EPS = 3.0e-7;

	/**
	 * 
	 * Value from numerical recipies in C alforithms
	 * 
	 */
	private static final double FPMIN = 0;

	/**
	 * 
	 * Compute the Chi2 value for an observed samples and a theoric samples
	 * 
	 * @param obsStats
	 *          the observed statistics
	 * @param theoStats
	 *          the theoritical statistics
	 * @return the chi2 value computed
	 * 
	 */
	public double computeChi2(ILearningProbabilities<T> obsStats,
			IProbabilities<T> theoStats) {
		double result = 0.0;
		for (T value : theoStats.getValueSet()) {
			double theo = theoStats.getProbability(value)
					* obsStats.getNumberOfExamples();
			double obs = obsStats.getNumberOfExamples(value);
			double diff = obs - theo;
			result = result + (diff * diff) / theo;
		}
		return result;
	}

	/**
	 * 
	 * Compute the Chi2 value for a set observed samples and a theoric samples
	 * 
	 * @param obsStats
	 *          the set of observed statistics
	 * @param theoStats
	 *          the theoritical statistics
	 * @return the chi2 value computed
	 * 
	 */
	public double computeChi2(Set<ILearningProbabilities<T>> obsStats,
			IProbabilities<T> theoStats) {
		double result = 0.0;
		for (ILearningProbabilities<T> obsStat : obsStats) {
			result = result + computeChi2(obsStat, theoStats);
		}
		return result;
	}

	/**
	 * 
	 * Compute the Chi2 value for a set observed samples
	 * 
	 * @param obsStats
	 *          the set of observed statistics
	 * @return the chi2 value computed
	 * 
	 */
	public double computeChi2(Set<ILearningProbabilities<T>> obsStats) {
		ValueStatistics<T> theoStats = new ValueStatistics<T>();
		for (ILearningProbabilities<T> obsStat : obsStats) {
			theoStats.merge(obsStat);
		}
		return computeChi2(obsStats, theoStats);
	}

	/**
	 * 
	 * Compute the alpha value for a leaf
	 * 
	 * @param obsStats
	 *          the set of observed statistics
	 * @param theoStats
	 *          the theoritical statistics
	 * 
	 * @return the computed alpha value
	 * 
	 */
	public double computeAlphaValue(ILearningProbabilities<T> obsStats,
			IProbabilities<T> theoStats) {
		if ((obsStats == null)
				|| (theoStats.getValueSet().size() < obsStats.getValueSet().size())) {
			return 0.0;
		}
		if (theoStats.getValueSet().size() == 1) {
			T value = theoStats.getBest();
			if (obsStats.getProbability(value) == 1.0) {
				return 1.0;
			}
			return 0.0;
		}
		return chsone(obsStats, theoStats);
	}

	/**
	 * Given the distribution learningNode containing the observed frequency, and
	 * an array referenceNode containing the expected frequency, and given the
	 * number of constraints knstrn (normally one), this routine computes
	 * (trivially) the number of degrees of freedom df, and (nontrivially) the
	 * chi-square chsq and the signifiance prob. A small value of prob indicates a
	 * signifiant difference between the distributions referenceNode and
	 * learningNode. From Numerical Recipes in C Page 621
	 * 
	 * @param obsStats
	 *          the set of observed statistics
	 * @param theoStats
	 *          the theoritical statistics
	 * 
	 * @return the signifiance prob
	 * 
	 */
	private double chsone(ILearningProbabilities<T> obsStats,
			IProbabilities<T> theoStats) {
		final double knstrn = 1;
		double df = theoStats.getValueSet().size() - knstrn;
		double chsq = computeChi2(obsStats, theoStats);
		return gammq(0.5 * df, 0.5 * chsq);
	}

	/**
	 * 
	 * Compute the incomplete gamma function Q(a,x)=1-P(a,x) From Numerical
	 * Recipes in C Page 218
	 * 
	 * @param a
	 *          the degrees of freedom
	 * @param x
	 *          the chi square value
	 * 
	 * @return the incomplete gamma function Q(a,x)=1-P(a,x)
	 * 
	 */
	private double gammq(double a, double x) {
		assert (a > 0.0);
		assert (x >= 0.0);
		if (x < (a + 1.0)) {
			return 1.0 - gser(a, x);
		}
		return gcf(a, x);

	}

	/**
	 * 
	 * Compute the incomplete gamma function Q(a,x) evaluated by its continued
	 * fraction representation as gammcf From Numerical Recipes in C Page 219
	 * 
	 * @param a
	 *          the degrees of freedom
	 * @param x
	 *          the chi square value
	 * 
	 * @return the incomplete gamma function Q(a,x)
	 * 
	 */
	private double gcf(double a, double x) {
		int i;
		double an, b, c, d, del, h;
		double gIn = gammln(a);
		b = x + 1.0 - a;
		c = 1.0 / FPMIN;
		d = 1.0 / b;
		h = d;
		for (i = 1; i <= ITMAX; i++) {
			an = -i * (i - a);
			b += 2.0;
			d = an * d + b;
			if (Math.abs(d) < FPMIN) {
				d = FPMIN;
			}
			c = b + an / c;
			if (Math.abs(c) < FPMIN) {
				c = FPMIN;
			}
			d = 1.0 / d;
			del = d * c;
			h *= del;
			if (Math.abs(del - 1.0) < EPS) {
				break;
			}
		}
		assert (i <= ITMAX);
		return Math.exp(-x + a * Math.log(x) - gIn) * h;
	}

	/**
	 * 
	 * Compute the incomplete gamma function P(a,x) evaluated by its series
	 * representation as gamser From Numerical Recipes in C Page 218
	 * 
	 * @param a
	 *          the degrees of freedom
	 * @param x
	 *          the chi square value
	 * 
	 * @return the incomplete gamma function P(a,x)
	 * 
	 */
	private double gser(double a, double x) {
		int n;
		double sum, del, ap;
		double gamser;
		double gIn = gammln(a);
		assert (x >= 0.0);
		if (x == 0.0) {
			gamser = 0.0;
			return gamser;
		}
		ap = a;
		sum = 1.0 / a;
		del = sum;
		for (n = 1; n <= ITMAX; n++) {
			++ap;
			del *= x / ap;
			sum += del;
			if (Math.abs(del) < Math.abs(sum) * EPS) {
				gamser = sum * Math.exp(-x + a * Math.log(x) - gIn);
				return gamser;
			}
		}
		assert (false);
		return Double.NaN;
	}

	/**
	 * 
	 * Compute the value ln(Sigma(xx)) for xx > 0 From Numerical Recipes in C Page
	 * 218
	 * 
	 * @param xx
	 *          the degrees of freedom
	 * 
	 * @return the value ln(Sigma(xx))
	 * 
	 */
	private double gammln(double xx) {
		double x, y, tmp, ser;
		final double[] cof = { 76.18009172947146, -86.50532032941677,
				24.01409824083091, -1.231739572450155, 0.120865097386617e-2,
				-0.5395239384953e-5 };
		int j;

		x = xx;
		y = x;
		tmp = x + 5.5;
		tmp -= (x + 0.5) * Math.log(tmp);
		ser = 1.000000000190015;
		for (j = 0; j <= 5; j++) {
			ser += cof[j] / ++y;
		}
		return -tmp + Math.log(2.5066282746310005 * ser / x);
	}

	/**
	 * 
	 * Chi2 utils instance
	 * 
	 */
	static public final Chi2Utils<String> chi2Utils = new Chi2Utils<String>();

}
