package animatlab.kodamat.core.proba;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.utils.Parameters;

/**
 * 
 * Count the number of times a value occurs to be able to compute its
 * probability
 * 
 * @param <T>
 *          type of the data to count
 * 
 * @author Thomas Degris
 * 
 */
public class ValueStatistics<T> implements ILearningProbabilities<T> {
	/**
	 * 
	 * Number of examples added to this
	 * 
	 */
	private int sampleSize = 0;

	/**
	 * 
	 * Hash map used to count the number of times a value occurs
	 * 
	 */
	private Map<T, Integer> stats = new LinkedHashMap<T, Integer>();

	/**
	 * 
	 * Constructor
	 * 
	 * 
	 */
	public ValueStatistics() {
		super();
	}

	/**
	 * 
	 * Merge stats stored in a map
	 * 
	 * @param otherStats
	 *          the other stats to merge
	 * 
	 */
	public void merge(Map<T, Integer> otherStats) {
		for (Map.Entry<T, Integer> entry : otherStats.entrySet()) {
			int thisStat = getNumberOfExamples(entry.getKey());
			stats.put(entry.getKey(), thisStat + entry.getValue());
			sampleSize = sampleSize + entry.getValue();
		}
	}

	/**
	 * 
	 * Merge from other stats
	 * 
	 * @param otherStats
	 *          the other stats to merge
	 * 
	 */
	public void merge(ILearningProbabilities<T> otherStats) {
		for (T value : otherStats.getValueSet()) {
			int thisStat = getNumberOfExamples(value);
			int otherStat = otherStats.getNumberOfExamples(value);
			stats.put(value, thisStat + otherStat);
			sampleSize = sampleSize + otherStat;
		}
	}

	/**
	 * @see animatlab.kodamat.core.proba.ILearningProbabilities#getNumberOfExamples()
	 */
	final public int getNumberOfExamples() {
		return sampleSize;
	}

	/**
	 * @see animatlab.kodamat.core.proba.IProbabilities#getBest()
	 */
	final public T getBest() {
		int max = 0;
		T result = null;
		for (T value : getValueSet()) {
			int numberOfExamples = getNumberOfExamples(value);
			if (numberOfExamples > max) {
				result = value;
				max = numberOfExamples;
			}
		}
		return result;
	}

	/**
	 * 
	 * Get the probability associated with one value
	 * 
	 * @param value
	 *          the value to get the probability from
	 * 
	 * @return the probability of corresponding to the value
	 * 
	 */
	final public double getProbability(T value) {
		assert (getNumberOfExamples() > 0);
		return (double) getNumberOfExamples(value) / getNumberOfExamples();
	}

	/**
	 * @see animatlab.kodamat.core.proba.IProbabilities#getValueSet()
	 */
	final public Set<T> getValueSet() {
		return stats.keySet();
	}

	/**
	 * 
	 * Get the number of times a value occured
	 * 
	 * @param value
	 *          the value to look at
	 * @return the number of times the value occurs in the sample
	 * 
	 */
	final public int getNumberOfExamples(T value) {
		Integer result = stats.get(value);
		if (result == null) {
			return 0;
		}
		return result;
	}

	/**
	 * 
	 * Add a new example to the sample
	 * 
	 * @param value
	 *          the example to add
	 * 
	 */
	final public void add(T value) {
		int currentStat = getNumberOfExamples(value);
		stats.put(value, currentStat + 1);
		sampleSize++;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuilder result = new StringBuilder();
		for (Map.Entry<T, Integer> entry : stats.entrySet()) {
			result.append(entry.getKey() + ":"
					+ Parameters.format((double) entry.getValue() / sampleSize));
		}
		return result.toString();
	}

}
