package animatlab.kodamat.core.rules;

import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import animatlab.kodamat.core.functions.IPartition;
import animatlab.kodamat.core.functions.IPartitionFunction;
import animatlab.kodamat.core.io.DataLoadingPolicy;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.reinforcementlearning.planning.guestrin.linearprogramming.LPVariable;

/**
 * 
 * Set of rules defining a factored function
 * 
 * @param <T>
 *          type of the result of the function
 * 
 * @author Thomas Degris
 * 
 */
public class PlainRules<T> extends Rules<T> {
	/**
	 * 
	 * Policy to add set of rules together
	 * 
	 */
	final static public PlainRules.CombinaisonPolicy<Double> doubleMinusPolicy = new PlainRules.CombinaisonPolicy<Double>() {
		public Double operator(Double value01, Double value02) {
			return value01 - value02;
		}
	};

	/**
	 * 
	 * Policy to choose the max between two rules
	 * 
	 */
	final static public PlainRules.CombinaisonPolicy<Double> doubleMaxPolicy = new PlainRules.CombinaisonPolicy<Double>() {
		public Double operator(Double value01, Double value02) {
			return Math.max(value01, value02);
		}
	};

	/**
	 * 
	 * Policy to add set of rules together
	 * 
	 */
	final static public PlainRules.CombinaisonPolicy<Double> doubleMultiplyPolicy = new PlainRules.CombinaisonPolicy<Double>() {
		public Double operator(Double value01, Double value02) {
			return value01 * value02;
		}
	};

	/**
	 * 
	 * Policy to add set of rules together
	 * 
	 */
	final static public PlainRules.CombinaisonPolicy<Double> doubleSumPolicy = new PlainRules.CombinaisonPolicy<Double>() {
		public Double operator(Double value01, Double value02) {
			return value01 + value02;
		}
	};

	/**
	 * 
	 * Policy to load a rule
	 * 
	 */
	final static public IPartitionFunction.LoadXmlPolicy<Double> loadingPolicy = new IPartitionFunction.LoadXmlPolicy<Double>() {
		public IPartitionFunction<Double> loadXml(Element element) {
			return new PlainRules<Double>(element,
					DataLoadingPolicy.doubleLoadingPolicy);
		}
	};

	/**
	 * 
	 * SerialVersionUID
	 * 
	 */
	private static final long serialVersionUID = -918109180827666861L;

	/**
	 * 
	 * Set of rules contained by this rules
	 * 
	 */
	Set<IPartition<T>> rules = new LinkedHashSet<IPartition<T>>();

	/**
	 * 
	 * Constructor
	 * 
	 */
	public PlainRules() {
		super();
	}

	/**
	 * 
	 * Constructor of this set of rules from a tree representation
	 * 
	 * @param tree
	 *          the tree to build the set of leaves from
	 * 
	 */
	public PlainRules(Tree<T> tree) {
		for (IPartition<T> leaf : tree.partitions()) {
			PlainRule<T> rule = createRule(leaf.get());
			rule.setContext(leaf.getContext());
			add(rule);
		}
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param element
	 *          root of the xml tree
	 * @param dataLoadingPolicy
	 *          the policy to use to load the data
	 * 
	 */
	public PlainRules(Element element, DataLoadingPolicy<T> dataLoadingPolicy) {
		assert (NODE_NAME.equals(element.getNodeName()));
		NodeList nodeList = element.getChildNodes();
		for (int index = 0; index < nodeList.getLength(); index++) {
			Node childNode = nodeList.item(index);
			if (childNode.getNodeType() == Node.ELEMENT_NODE) {
				if (childNode.getNodeName().equals(Rule.NODE_NAME)) {
					PlainRule<T> rule = new PlainRule<T>((Element) childNode,
							dataLoadingPolicy);
					add(rule);
				}
			}
		}
	}

	/**
	 * 
	 * Create a rule in this set
	 * 
	 * @param value
	 *          value pointed by the rule
	 * 
	 * @return the rule created
	 * 
	 */
	public PlainRule<T> createRule(T value) {
		return new PlainRule<T>(value);
	}

	/**
	 * 
	 * Add a rule to this set. Once added, the rule is declared as unmutable
	 * 
	 * @param rule
	 *          the rule to add
	 * 
	 * @return true if the element has been added
	 * 
	 */
	public boolean add(PlainRule<T> rule) {
		return rules.add(rule);
	}

	/**
	 * 
	 * Combine two sets of rules together
	 * 
	 * @param rules
	 *          the other set of rules to combine to this one
	 * @param operator
	 *          operator to use to combine rules with the same context
	 * @param valuesSet
	 *          set of possible values for the variables
	 * 
	 */
	public void combine(IPartitionFunction<T> rules,
			CombinaisonPolicy<T> operator, Map<Variable, Set<String>> valuesSet) {

		PlainRules<T> workingRules = new PlainRules<T>();
		for (IPartition<T> rule : rules.partitions()) {
			PlainRule<T> clone = workingRules.createRule(rule.get());
			clone.setContext(rule.getContext());
			workingRules.add(clone);
		}

		// While there are two consistent rules n1 = [c1:p1] and n2 = [c2:p2]:
		// If c1 = c2, replace these two rules by [c1:p1p2];
		// Else replace these two rules by the set: Split(n1<c2)USplit(n2<c1)
		RulePair<T> rulePair = getConsistentRules(workingRules, null);
		while (rulePair != null) {
			Rule<T> p1 = rulePair.rule01;
			Rule<T> p2 = rulePair.rule02;
			remove(p1);
			workingRules.remove(p2);
			if (p1.hasSameContext(p2)) {
				// If c1 = c2, replace these two rules by [c1:p1p2];
				p1.set(operator.operator(p1.get(), p2.get()));
				if (p1.get() != null) {
					add(p1);
				}
			} else {
				// Else replace these two rules by the set: Split(n1<c2)USplit(n2<c1)
				p1.split(this, p2.getContext(), valuesSet);
				p2.split(workingRules, p1.getContext(), valuesSet);
			}
			rulePair = getConsistentRules(workingRules, null);
		}

		for (IPartition<T> rule : workingRules.partitions()) {
			PlainRule<T> clone = createRule(rule.get());
			clone.setContext(rule.getContext());
			add(clone);
		}
	}

	/**
	 * 
	 * Remove a rule from this set
	 * 
	 * @param rule
	 *          the rule to remove
	 * 
	 */
	@Override
	public void remove(Rule<T> rule) {
		rules.remove(rule);
	}

	/**
	 * @see animatlab.kodamat.core.io.IToXml#toXml(org.w3c.dom.Document,
	 *      animatlab.kodamat.core.io.IToXml.IToXmlPolicy)
	 */
	public Element toXml(Document document, IToXmlPolicy<T> policy) {
		Element result = document.createElement(NODE_NAME);
		for (Rule<T> rule : this) {
			result.appendChild(rule.toXml(document, policy));
		}
		return result;
	}

	/**
	 * 
	 * Get the first rule callable for a state
	 * 
	 * @param state
	 *          the state to get the rule from
	 * @return the rule associated with the state
	 * 
	 */
	public Rule<T> getRule(Context state) {
		for (Rule<T> current : this) {
			if (Context.satisfy(current.getContext(), state)) {
				return current;
			}
		}
		return null;
	}

	/**
	 * @see animatlab.kodamat.core.functions.IFunction#classify(Context)
	 */
	public T classify(Context state) {
		Rule<T> result = getRule(state);
		if (result == null) {
			return null;
		}
		return result.get();
	}

	/**
	 * @see animatlab.kodamat.core.functions.IFunction#numberOfNodes()
	 */
	public int numberOfNodes() {
		int result = 0;
		for (Rule<T> rule : this) {
			result = result + rule.getContext().size();
		}
		return result;
	}

	/**
	 * 
	 * Find rules with the same context
	 * 
	 * @param ignoredVariable
	 *          the variable to ignore
	 * 
	 * @return a set of rules with the same context (the set is empty if (no rule
	 *         was found)
	 * 
	 */
	public Set<Rule<T>> findSameContext(Variable ignoredVariable) {
		Set<Rule<T>> result = new LinkedHashSet<Rule<T>>();
		for (Rule<T> ruleLookedFor : this) {
			for (Rule<T> rule : this) {
				if ((rule != ruleLookedFor)
						&& (rule.hasSameContext(ruleLookedFor, ignoredVariable))) {
					result.add(rule);
				}
			}
			if (!result.isEmpty()) {
				result.add(ruleLookedFor);
				return result;
			}
		}
		return result;
	}

	/**
	 * @see animatlab.kodamat.core.rules.Rules#isEmpty()
	 */
	@Override
	public boolean isEmpty() {
		return rules.isEmpty();
	}

	/**
	 * @see animatlab.kodamat.core.rules.Rules#size()
	 */
	@Override
	public int size() {
		return rules.size();
	}

	/**
	 * 
	 * Clear all the rules of this set of rules
	 * 
	 */
	public void clear() {
		rules.clear();
	}

	/**
	 * @see java.lang.Iterable#iterator()
	 */
	public Iterator<Rule<T>> iterator() {
		return new Iterator<Rule<T>>() {
			private final Iterator<IPartition<T>> iterator = rules.iterator();

			public boolean hasNext() {
				return iterator.hasNext();
			}

			public Rule<T> next() {
				return (Rule<T>) iterator.next();
			}

			public void remove() {
				iterator.remove();
			}
		};
	}

	/**
	 * 
	 * Remove a set of rules from this rule
	 * 
	 * @param rulesSet
	 *          the set of rules to remove from this set
	 * 
	 */
	public void removeAll(Set<Rule<LPVariable>> rulesSet) {
		rules.removeAll(rulesSet);
	}

	/**
	 * @see animatlab.kodamat.core.rules.Rules#add(animatlab.kodamat.core.rules.Rule)
	 */
	@Override
	public void add(Rule<T> rule) {
		rules.add(rule);
	}

	/**
	 * @see animatlab.kodamat.core.functions.IPartitionFunction#partitions()
	 */
	public Set<IPartition<T>> partitions() {
		return rules;
	}
}
