package animatlab.kodamat.core.rules;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import animatlab.kodamat.core.functions.IPartition;
import animatlab.kodamat.core.io.IToXml;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;

/**
 * 
 * Rule data structure
 * 
 * @param <T>
 *          type of the value pointed by the rule
 * 
 * @author Thomas Degris
 * 
 */
public abstract class Rule<T> implements IToXml<T>, IPartition<T> {
	/**
	 * 
	 * Name of the xml node
	 * 
	 */
	static protected final String NODE_NAME = Rules.NAMESPACE + ":" + "rule";

	/**
	 * 
	 * Name of the xml node for a constraint
	 * 
	 */
	static protected final String NODE_NAME_CONSTRAINT = Rules.NAMESPACE + ":"
			+ "cond";

	/**
	 * 
	 * Attribute of a constraint to get the value of the variable
	 * 
	 */
	public static final String ATTRIBUTE_VALUE = "value";

	/**
	 * 
	 * Buffered hashcode for this rule
	 * 
	 */
	private int hashCode = 0;

	/**
	 * 
	 * Constructor
	 * 
	 */
	protected Rule() {
		super();
	}

	/**
	 * 
	 * Reset the hashcode of this instance
	 * 
	 */
	protected void resetHashCode() {
		hashCode = 0;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuilder result = new StringBuilder("(");
		result.append(Context.contextToString(getContext()));
		if (get() != null) {
			result.append(" :> ");
			result.append(get());
		}
		result.append(")");
		return result.toString();
	}

	/**
	 * @see animatlab.kodamat.core.io.IToXml#toXml(org.w3c.dom.Document,
	 *      animatlab.kodamat.core.io.IToXml.IToXmlPolicy)
	 */
	public Element toXml(Document document, IToXmlPolicy<T> policy) {
		Element result = document.createElement(NODE_NAME);
		policy.toXml(result, get());
		for (Map.Entry<Variable, String> entry : getContext().entrySet()) {
			Element constraintElement = document.createElement(NODE_NAME_CONSTRAINT);
			entry.getKey().setAttribute(constraintElement);
			constraintElement.setAttribute(ATTRIBUTE_VALUE, entry.getValue());
			result.appendChild(constraintElement);
		}
		return result;
	}

	/**
	 * 
	 * Test if the context part of this rule is equal to another rule
	 * 
	 * @param rule
	 *          the rule to test if the context are equals
	 * @param ignoredVariable
	 *          the variable to ignore during the test
	 * 
	 * @return true if the context are equals
	 * 
	 */
	protected boolean hasSameContext(Rule<T> rule, Variable ignoredVariable) {
		if (rule.getContext().size() != getContext().size()) {
			return false;
		}
		for (Map.Entry<Variable, String> entry : getContext().entrySet()) {
			if ((ignoredVariable == null)
					|| (!entry.getKey().equals(ignoredVariable))) {
				String thisConstraint = entry.getValue();
				String otherConstraint = rule.getContext().get(entry.getKey());
				if (!thisConstraint.equals(otherConstraint)) {
					return false;
				}
			}
		}
		return true;
	}

	/**
	 * 
	 * Test if the context part of this rule is equal to another rule
	 * 
	 * @param rule
	 *          the rule to test if the context are equals
	 * 
	 * @return true if the context are equals
	 * 
	 */
	public boolean hasSameContext(Rule<T> rule) {
		return hasSameContext(rule, null);
	}

	/**
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@SuppressWarnings("unchecked")
	@Override
	public boolean equals(Object object) {
		if (object == null) {
			return false;
		}
		Rule<T> other = (Rule<T>) object;
		if ((get() == null) && (other.get() != null)) {
			return false;
		}
		if ((get() != null) && (!get().equals(other.get()))) {
			return false;
		}
		return hasSameContext(other);
	}

	/**
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		if (hashCode == 0) {
			StringBuilder strHashCode = new StringBuilder();
			List<Variable> list = new ArrayList<Variable>(scope());
			Collections.sort(list);
			for (Variable variable : list) {
				strHashCode.append(variable.toString());
				strHashCode.append(getContext().get(variable));
			}
			hashCode = strHashCode.toString().hashCode();
		}
		return hashCode;
	}

	/**
	 * 
	 * Check this rule is consistent with another rule
	 * 
	 * @param rule
	 *          the rule to test against
	 * 
	 * @return true if it is consistant
	 * 
	 */
	public boolean isConsistent(Rule<?> rule) {
		return isConsistent(rule.getContext(), null);
	}

	/**
	 * 
	 * Check this rule is consistent with a context
	 * 
	 * @param otherContext
	 *          the context to check against
	 * @param ignoreVariable
	 *          the variable to ignore
	 * 
	 * @return true if it is consistant
	 * 
	 */
	public boolean isConsistent(Context otherContext, Variable ignoreVariable) {
		Set<Variable> domain = new LinkedHashSet<Variable>(getContext().keySet());
		domain.retainAll(otherContext.keySet());
		if (ignoreVariable != null) {
			domain.remove(ignoreVariable);
		}
		for (Variable variable : domain) {
			if (!getContext().get(variable).equals(otherContext.get(variable))) {
				return false;
			}
		}
		return true;
	}

	/**
	 * @see java.lang.Object#clone()
	 */
	@Override
	public PlainRule<T> clone() {
		PlainRule<T> result = new PlainRule<T>(get());
		result.setContext(getContext());
		return result;
	}

	/**
	 * @see animatlab.kodamat.core.functions.IPartition#scope()
	 */
	public Set<Variable> scope() {
		return getContext().keySet();
	}

	/**
	 * 
	 * Split operator on rule (Guestrin thesis, page 123, figure 7.2.4)
	 * 
	 * @param rules
	 *          set where to add the rules
	 * @param b
	 *          set of conditions to use in the split
	 * @param valuesSet
	 *          set of values of the variables
	 * 
	 */
	public void split(Rules<T> rules, Context b,
			Map<Variable, Set<String>> valuesSet) {
		// 1. {p} if c is not consistent with b
		if (!isConsistent(b, null)) {
			rules.add(clone());
			return;
		}
		// 2. {p} if Scope[B] include in Scope[C]
		Set<Variable> scopeB = b.keySet();
		Set<Variable> scopeC = scope();
		if (scopeC.containsAll(scopeB)) {
			rules.add(clone());
			return;
		}
		// 3. {Split(pi<b)|pi IN Split(p<Y),
		// for some variable Y IN Scope(B)-Scope(C)
		scopeB = new LinkedHashSet<Variable>(scopeB);
		scopeB.removeAll(scopeC);
		Variable varY = scopeB.iterator().next();
		PlainRules<T> splitOnY = new PlainRules<T>();
		split(splitOnY, varY, valuesSet);
		while (!splitOnY.isEmpty()) {
			Rule<T> splittedRuleOnY = splitOnY.iterator().next();
			splitOnY.remove(splittedRuleOnY);
			splittedRuleOnY.split(rules, b, valuesSet);
		}
	}

	/**
	 * 
	 * Split operator using one variable as context (Guestrin thesis, page 123,
	 * figure 7.2.3)
	 * 
	 * @param rules
	 *          set where to add the rules
	 * @param y
	 *          set of conditions to use in the split
	 * @param valuesSet
	 *          set of values of the variables
	 * 
	 */
	public void split(Rules<T> rules, Variable y,
			Map<Variable, Set<String>> valuesSet) {
		// If Y IN Scope(C), then Split(p<Y)={p}
		if (scope().contains(y)) {
			rules.add(clone());
			return;
		}
		// Otherwise: Split(p<Y)={[c&Y=yi:v]|yi IN Dom(Y)}
		for (String value : valuesSet.get(y)) {
			PlainRule<T> addedRule = clone();
			addedRule.addConstraint(y, value);
			rules.add(addedRule);
		}
	}
}
