package animatlab.kodamat.core.rules;

import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.functions.IPartitionFunction;
import animatlab.kodamat.core.io.IToXml;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;

/**
 * 
 * Abstract class on rules
 * 
 * @author Thomas Degris
 * @param <T>
 *          type pointed by the rules
 * 
 */
public abstract class Rules<T> implements IToXml<T>, Iterable<Rule<T>>,
		IPartitionFunction<T>, Cloneable {
	/**
	 * 
	 * Namespace of xml node name
	 * 
	 */
	public static final String NAMESPACE = "rdr";

	/**
	 * 
	 * Name of the node in xml to represent an decision node
	 * 
	 */
	public static final String NODE_NAME = NAMESPACE + ":" + "rules";

	/**
	 * 
	 * Policy to combine rule with the same context together
	 * 
	 * @param <T>
	 *          type used in the rules
	 * 
	 */
	public interface CombinaisonPolicy<T> {
		/**
		 * 
		 * Operator on two values of rules
		 * 
		 * @param value01
		 *          the first argument of the operator
		 * @param value02
		 *          the second argument of the operator
		 * 
		 * @return the result of the combinaison
		 * 
		 */
		T operator(T value01, T value02);
	}

	/**
	 * 
	 * Pair of rules
	 * 
	 * @param <U>
	 *          type of the rule to find
	 * 
	 */
	static public class RulePair<U> {
		/**
		 * 
		 * First rule of the pair
		 * 
		 */
		public Rule<U> rule01;

		/**
		 * 
		 * Second rule of the pair
		 * 
		 */
		public Rule<U> rule02;

		/**
		 * 
		 * Constructor
		 * 
		 * @param rule01
		 *          first rule of the pair
		 * @param rule02
		 *          second rule of the pair
		 * 
		 */
		public RulePair(Rule<U> rule01, Rule<U> rule02) {
			this.rule01 = rule01;
			this.rule02 = rule02;
		}

		/**
		 * @see java.lang.Object#toString()
		 */
		@Override
		public String toString() {
			return "[First: " + rule01 + " Second: " + rule02 + "]";
		}
	}

	/**
	 * 
	 * Constructor
	 * 
	 */
	public Rules() {
		super();
	}

	/**
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@SuppressWarnings("unchecked")
	@Override
	public boolean equals(Object object) {
		if (object == null) {
			return false;
		}
		Rules<T> other = (Rules<T>) object;
		if (size() != other.size()) {
			return false;
		}
		for (Rule<T> rule : this) {
			if (!other.contains(rule)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * 
	 * Check if this set of rules contain a rule
	 * 
	 * @param rule
	 *          the rule to check
	 * 
	 * @return true if the rule is contained
	 * 
	 */
	private boolean contains(Rule<T> rule) {
		for (Rule<T> current : this) {
			if (current.equals(rule)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * 
	 * @return the size of this set of rules
	 * 
	 */
	abstract public int size();

	/**
	 * 
	 * @return true if this set of rules is empty
	 * 
	 */
	abstract public boolean isEmpty();

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuilder result = new StringBuilder("[\n");
		String separator = "\n";
		if (!isEmpty()) {
			for (Rule<T> rule : this) {
				result.append(" ");
				result.append(rule.toString());
				result.append(separator);
			}
		}
		result.append("]");
		return result.toString();
	}

	/**
	 * 
	 * @return the scope of this set of rules
	 * 
	 */
	public Set<Variable> scope() {
		LinkedHashSet<Variable> result = new LinkedHashSet<Variable>();
		for (Rule<T> rule : this) {
			result.addAll(rule.scope());
		}
		return result;
	}

	/**
	 * @see java.lang.Object#clone()
	 */
	@Override
	public PlainRules<T> clone() {
		PlainRules<T> result = new PlainRules<T>();
		for (Rule<T> rule : this) {
			result.add(rule.clone());
		}
		return result;
	}

	/**
	 * 
	 * Find in this a rule that satisfy a particular context
	 * 
	 * @param context
	 *          the context to satisfy
	 * 
	 * @return a rule if found
	 * 
	 */
	public Rule<T> findConsistent(Context context) {
		for (Rule<T> rule : this) {
			if (rule.isConsistent(context, null)) {
				return rule;
			}
		}
		return null;
	}

	/**
	 * 
	 * Get the first pair of consistent rule from a rule set (ignoring a variable)
	 * 
	 * @param ignoredVariable
	 *          the variable ignored
	 * 
	 * @return the first pair found, null if not found
	 * 
	 */
	public RulePair<T> getConsistentRules(Variable ignoredVariable) {
		return getConsistentRules(this, ignoredVariable);
	}

	/**
	 * 
	 * Add a new rule to this rule set
	 * 
	 * @param rule
	 *          the rule to add
	 * 
	 */
	abstract public void add(Rule<T> rule);

	/**
	 * 
	 * Remove a rule from this set of rules
	 * 
	 * @param rule
	 *          the rule to remove
	 * 
	 */
	abstract public void remove(Rule<T> rule);

	/**
	 * 
	 * Get the first consistent rule in this rule (ignoring a variable)
	 * 
	 * @param rule
	 *          the rule to get check the consistency with
	 * @param ignoredVariable
	 *          the variable to ignore
	 * 
	 * @return the first rule consistent with rule (null if not found)
	 * 
	 */
	public Rule<T> getConsistentRule(Rule<T> rule, Variable ignoredVariable) {
		for (Rule<T> current : this) {
			if ((rule != current)
					&& current.isConsistent(rule.getContext(), ignoredVariable)) {
				return current;
			}
		}
		return null;
	}

	/**
	 * 
	 * Get the first consistent rule in this rule (ignoring a variable)
	 * 
	 * @param context
	 *          the context to get check the consistency with
	 * 
	 * @return the first rule consistent with rule (null if not found)
	 * 
	 */
	public Rule<T> getConsistentRule(Context context) {
		for (Rule<T> current : this) {
			if (current.isConsistent(context, null)) {
				return current;
			}
		}
		return null;
	}

	/**
	 * 
	 * Get the first pair of consistent rule from a rule set (ignoring a variable)
	 * 
	 * @param rules
	 *          set of rules where to look for
	 * @param ignoredVariable
	 *          the variable ignored
	 * 
	 * @return the first pair found, null if not found
	 * 
	 */
	public RulePair<T> getConsistentRules(Rules<T> rules, Variable ignoredVariable) {
		for (Rule<T> rule : this) {
			Rule<T> consistentRule = rules.getConsistentRule(rule, ignoredVariable);
			if (consistentRule != null) {
				return new RulePair<T>(rule, consistentRule);
			}
		}
		return null;
	}

	/**
	 * 
	 * Update a variable set of values given a particular context
	 * 
	 * @param context
	 *          the context used to update the set
	 * @param valueSets
	 *          the array of set of values to update
	 * 
	 */
	private void updateValueSet(Context context,
			Map<Variable, Set<String>> valueSets) {
		for (Map.Entry<Variable, String> entry : context.entrySet()) {
			Variable variable = entry.getKey();
			Set<String> valuesSet = valueSets.get(variable);
			if (valuesSet == null) {
				valuesSet = new LinkedHashSet<String>();
				valueSets.put(variable, valuesSet);
			}
			valuesSet.add(entry.getValue());
		}
	}

	/**
	 * 
	 * Merge rules together by applying an operator on them
	 * 
	 * @param operator
	 *          the operator to apply
	 * 
	 */
	public void mergeRules(CombinaisonPolicy<T> operator) {
		Map<Variable, Set<String>> valuesSet = new LinkedHashMap<Variable, Set<String>>();
		for (Rule<T> rule : this) {
			updateValueSet(rule.getContext(), valuesSet);
		}
		mergeRules(operator, valuesSet);
	}

	/**
	 * 
	 * Merge rules together by applying an operator on them
	 * 
	 * @param operator
	 *          the operator to apply
	 * @param valuesSet
	 *          set of possible values for the variables
	 * 
	 */
	public void mergeRules(CombinaisonPolicy<T> operator,
			Map<Variable, Set<String>> valuesSet) {
		// While there are two consistent rules n1 = [c1:p1] and n2 = [c2:p2]:
		// If c1 = c2, replace these two rules by [c1:p1p2];
		// Else replace these two rules by the set: Split(n1<c2)USplit(n2<c1)
		RulePair<T> rulePair = getConsistentRules(null);
		while (rulePair != null) {
			Rule<T> p1 = rulePair.rule01;
			Rule<T> p2 = rulePair.rule02;
			remove(p1);
			remove(p2);
			if (p1.hasSameContext(p2)) {
				// If c1 = c2, replace these two rules by [c1:p1p2];
				p1.set(operator.operator(p1.get(), p2.get()));
				if (p1.get() != null) {
					add(p1);
				}
			} else {
				// Else replace these two rules by the set: Split(n1<c2)USplit(n2<c1)
				p1.split(this, p2.getContext(), valuesSet);
				p2.split(this, p1.getContext(), valuesSet);
			}
			rulePair = getConsistentRules(null);
		}
	}

}
