package animatlab.kodamat.core.rules;

import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Set;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import animatlab.kodamat.core.functions.IPartition;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.LeafListener;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Context;

/**
 * 
 * Set of rules build from a tree
 * 
 * @author Thomas Degris
 * @param <T>
 *          type of data pointed by the rules
 * 
 */
public class RulesTree<T> extends Rules<T> implements LeafListener {
	/**
	 * 
	 * Tree monitored by this set of rules
	 * 
	 */
	private final Tree<T> tree;

	/**
	 * 
	 * Map between leaves in the tree and their corresponding rules
	 * 
	 */
	LinkedHashMap<IPartition<T>, LeafRule<T>> leafRuleMap = new LinkedHashMap<IPartition<T>, LeafRule<T>>();

	/**
	 * 
	 * Constructor
	 * 
	 * @param tree
	 *          the tree to monitored by this set of rules
	 * 
	 */
	public RulesTree(Tree<T> tree) {
		this.tree = tree;
		tree.setLeafListener(this);
		for (IPartition<T> leaf : tree.partitions()) {
			leafIsCreated((Leaf<T>) leaf);
		}
	}

	/**
	 * @see animatlab.kodamat.core.rules.Rules#size()
	 */
	@Override
	public int size() {
		return tree.numberOfLeaves();
	}

	/**
	 * @see java.lang.Iterable#iterator()
	 */
	public Iterator<Rule<T>> iterator() {
		return new Iterator<Rule<T>>() {
			Iterator<LeafRule<T>> iterator = leafRuleMap.values().iterator();

			public boolean hasNext() {
				return iterator.hasNext();
			}

			public Rule<T> next() {
				return iterator.next();
			}

			public void remove() {
				iterator.remove();
			}
		};
	}

	/**
	 * @see animatlab.kodamat.core.rules.Rules#isEmpty()
	 */
	@Override
	public boolean isEmpty() {
		return leafRuleMap.isEmpty();
	}

	/**
	 * @see animatlab.kodamat.core.trees.LeafListener#leafIsCreated(animatlab.kodamat.core.trees.Leaf)
	 */
	@SuppressWarnings("unchecked")
	public void leafIsCreated(Leaf<?> leaf) {
		final Leaf<T> typedLeaf = (Leaf<T>) leaf;
		leafRuleMap.put(typedLeaf, new LeafRule<T>(typedLeaf));
	}

	/**
	 * @see animatlab.kodamat.core.trees.LeafListener#leafIsRemoved(animatlab.kodamat.core.trees.Leaf)
	 */
	public void leafIsRemoved(Leaf<?> leaf) {
		leafRuleMap.remove(leaf);
	}

	/**
	 * @see animatlab.kodamat.core.functions.IFunction#classify(Context)
	 */
	public T classify(Context state) {
		return tree.classify(state);
	}

	/**
	 * @see animatlab.kodamat.core.functions.IFunction#numberOfNodes()
	 */
	public int numberOfNodes() {
		return tree.numberOfLeaves();
	}

	/**
	 * @see animatlab.kodamat.core.io.IToXml#toXml(org.w3c.dom.Document,
	 *      animatlab.kodamat.core.io.IToXml.IToXmlPolicy)
	 */
	public Element toXml(Document document, IToXmlPolicy<T> policy) {
		return tree.toXml(document, policy);
	}

	/**
	 * @see animatlab.kodamat.core.rules.Rules#add(animatlab.kodamat.core.rules.Rule)
	 */
	@Override
	public void add(Rule<T> rule) {
		assert (false);
	}

	/**
	 * @see animatlab.kodamat.core.rules.Rules#remove(animatlab.kodamat.core.rules.Rule)
	 */
	@Override
	public void remove(Rule<T> rule) {
		assert (false);
	}

	/**
	 * @see animatlab.kodamat.core.functions.IPartitionFunction#partitions()
	 */
	public Set<IPartition<T>> partitions() {
		return tree.partitions();
	}

}
