package animatlab.kodamat.core.rules.tests;

import static animatlab.kodamat.core.io.XmlUtils.getRoot;
import static animatlab.kodamat.core.utils.Context.getState;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

import junit.framework.TestCase;
import animatlab.kodamat.core.functions.FunctionsSum;
import animatlab.kodamat.core.io.DataLoadingPolicy;
import animatlab.kodamat.core.io.IToXml;
import animatlab.kodamat.core.io.XmlUtils;
import animatlab.kodamat.core.rules.PlainRule;
import animatlab.kodamat.core.rules.PlainRules;
import animatlab.kodamat.core.rules.Rule;
import animatlab.kodamat.core.rules.RulesTree;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.core.utils.VariableOrdering;

/**
 * 
 * Test the set of rules
 * 
 * @author Thomas Degris
 * 
 */
public class TestRulesSet extends TestCase {
	/**
	 * 
	 * Legend to use in the test
	 * 
	 */
	final VariableOrdering variableOrdering = new VariableOrdering(
			XmlUtils
					.getRoot("<spi:variable-legend>"
							+ "  <spi:variable legend='a' index='0'/>"
							+ "  <spi:variable legend='b' index='1'/>"
							+ "  <spi:variable legend='c' index='2'/>"
							+ "  <spi:variable legend='d' index='3'/>"
							+ "  <spi:variable legend='e' index='4'/>"
							+ "</spi:variable-legend>"));

	/**
	 * 
	 * Prototype used to create an empty rules set with a value set defined
	 * 
	 */
	final Map<Variable, Set<String>> valuesSet = Context.loadValueSet(
			new LinkedHashMap<Variable, Set<String>>(), XmlUtils
					.getRoot("  <dtr:variablevalueset>"
							+ "    <dtr:valueset variable='a'>"
							+ "      <dtr:value value='0'/>" + "      <dtr:value value='1'/>"
							+ "    </dtr:valueset>" + "    <dtr:valueset variable='b'>"
							+ "      <dtr:value value='0'/>" + "      <dtr:value value='1'/>"
							+ "    </dtr:valueset>" + "    <dtr:valueset variable='c'>"
							+ "      <dtr:value value='0'/>" + "      <dtr:value value='1'/>"
							+ "    </dtr:valueset>" + "    <dtr:valueset variable='d'>"
							+ "      <dtr:value value='0'/>" + "      <dtr:value value='1'/>"
							+ "    </dtr:valueset>" + "  </dtr:variablevalueset>"));

	/**
	 * 
	 * Build a rules set from a xml definition
	 * 
	 * @param xmlString
	 *          the xml definition of the set of rules
	 * 
	 * @return the new set of rules
	 * 
	 */
	private PlainRules<Double> getRules(String xmlString) {
		return new PlainRules<Double>(getRoot(xmlString),
				DataLoadingPolicy.doubleLoadingPolicy);
	}

	/**
	 * 
	 * Build a rule from a xml definition
	 * 
	 * @param xmlString
	 *          the xml definition of the rule
	 * 
	 * @return the new rule
	 * 
	 */
	private Rule<Double> getRule(String xmlString) {
		return new PlainRule<Double>(getRoot(xmlString),
				DataLoadingPolicy.doubleLoadingPolicy);
	}

	/**
	 * 
	 * Test expand against xml loading
	 * 
	 */
	public void testXmlLoadingSaving() {
		PlainRules<Double> rules = new PlainRules<Double>();
		PlainRule<Double> rule = rules.createRule(1.2);
		rule.addConstraint(Variable.getVariable("a"), "1");
		rule.addConstraint(Variable.getVariable("b"), "0");
		rules.add(rule);
		rule = rules.createRule(2.4);
		rule.addConstraint(Variable.getVariable("a"), "0");
		rules.add(rule);
		String xmlString = "<rdr:rules>" + "  <rdr:rule data='1.2'>"
				+ "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='b' value='0'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='2.4'>" + "    <rdr:cond variable='a' value='0'/>"
				+ "  </rdr:rule>" + "</rdr:rules>";
		assertEquals(getRules(xmlString).toString(), rules.toString());

		xmlString = XmlUtils.toXmlString(rules, IToXml.doubleToXmlPolicy);
		assertEquals(getRules(xmlString).toString(), rules.toString());
	}

	/**
	 * 
	 * Test equals operator on the contional part
	 * 
	 */
	public void testCondEqualsOperator() {
		PlainRules<Double> rules01 = getRules("<rdr:rules>"
				+ "  <rdr:rule data='1.2'>" + "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='b' value='0'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='2.4'>" + "    <rdr:cond variable='a' value='0'/>"
				+ "  </rdr:rule>" + "  <rdr:rule data='2.4'>"
				+ "    <rdr:cond variable='a' value='0'/>" + "  </rdr:rule>"
				+ "</rdr:rules>");
		assertEquals(2, rules01.size());
		Rule<Double> rule01;
		PlainRules<Double> rules02 = getRules("<rdr:rules>"
				+ "  <rdr:rule data='2.4'>" + "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='b' value='0'/>" + "  </rdr:rule>"
				+ "</rdr:rules>");
		Rule<Double> rule02;

		rule01 = rules01.getRule(getState("a=0,b=0", null));
		rule02 = rules01.getRule(getState("a=0", null));
		assertTrue(rule01 != null);
		assertEquals(rule01, rule02);

		rule01 = rules01.getRule(getState("a=1", null));
		assertEquals(null, rule01);

		rule01 = rules01.getRule(getState("a=1,b=0", null));
		rule02 = rules02.getRule(getState("a=1,b=0", null));
		assertEquals(true, rule01.hasSameContext(rule02));

		rule02 = rules01.getRule(getState("a=0", null));
		assertEquals(false, rule01.hasSameContext(rule02));
	}

	/**
	 * 
	 * Test the hash code returned by the rules
	 * 
	 */
	public void testRuleHashCode() {
		PlainRules<Double> rules01 = getRules("<rdr:rules>"
				+ "  <rdr:rule data='1.2'>" + "    <rdr:cond variable='a' value='0'/>"
				+ "    <rdr:cond variable='c' value='1'/>" + "  </rdr:rule>"
				+ "</rdr:rules>");
		PlainRules<Double> rules02 = getRules("<rdr:rules>"
				+ "  <rdr:rule data='1.2'>" + "    <rdr:cond variable='c' value='1'/>"
				+ "    <rdr:cond variable='a' value='0'/>" + "  </rdr:rule>"
				+ "</rdr:rules>");
		assertEquals(rules01.iterator().next().hashCode(), rules02.iterator()
				.next().hashCode());
	}

	/**
	 * 
	 * Test equals operator on rules
	 * 
	 */
	public void testEqualsOperator() {
		PlainRules<Double> rules01 = getRules("<rdr:rules>"
				+ "  <rdr:rule data='1.2'>" + "    <rdr:cond variable='a' value='1'/>"
				+ "  </rdr:rule>" + "  <rdr:rule data='2.4'>"
				+ "    <rdr:cond variable='a' value='0'/>" + "  </rdr:rule>"
				+ "</rdr:rules>");
		PlainRules<Double> rules02 = getRules("<rdr:rules>"
				+ "  <rdr:rule data='2.4'>" + "    <rdr:cond variable='a' value='0'/>"
				+ "  </rdr:rule>" + "  <rdr:rule data='1.2'>"
				+ "    <rdr:cond variable='a' value='1'/>" + "  </rdr:rule>"
				+ "</rdr:rules>");
		assertEquals(rules01, rules02);
	}

	/**
	 * 
	 * Test the consistent operator
	 * 
	 */
	public void testFindContext() {
		PlainRules<Double> rules = getRules("<rdr:rules>"
				+ "  <rdr:rule data='1.2'>" + "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='b' value='0'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='2.4'>" + "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='b' value='1'/>" + "  </rdr:rule>"
				+ "</rdr:rules>");
		assertEquals(0, rules.findSameContext(null).size());
		assertEquals(2, rules.findSameContext(Variable.getVariable("b")).size());
	}

	/**
	 * 
	 * Test the consistent operator
	 * 
	 */
	public void testConsistent() {
		PlainRules<Double> rules = getRules("<rdr:rules>"
				+ "  <rdr:rule data='2.4'>" + "    <rdr:cond variable='a' value='0'/>"
				+ "  </rdr:rule>" + "  <rdr:rule data='1.2'>"
				+ "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='b' value='0'/>" + "  </rdr:rule>"
				+ "</rdr:rules>");
		Rule<Double> expectedRule = getRule("  <rdr:rule data='1.2'>"
				+ "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='b' value='0'/>" + "  </rdr:rule>");

		Rule<Double> consistentRule = rules.getConsistentRule(
				getRule("  <rdr:rule>" + "    <rdr:cond variable='a' value='1'/>"
						+ "  </rdr:rule>"), null);
		assertEquals(expectedRule, consistentRule);
		consistentRule = rules.getConsistentRule(getRule("  <rdr:rule>"
				+ "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='c' value='1'/>" + "  </rdr:rule>"), null);
		assertEquals(expectedRule, consistentRule);
		consistentRule = rules.getConsistentRule(getRule("  <rdr:rule>"
				+ "    <rdr:cond variable='a' value='2'/>"
				+ "    <rdr:cond variable='b' value='0'/>" + "  </rdr:rule>"), null);
		assertEquals(null, consistentRule);
		consistentRule = rules.getConsistentRule(getRule("  <rdr:rule>"
				+ "    <rdr:cond variable='b' value='2'/>"
				+ "    <rdr:cond variable='a' value='0'/>" + "  </rdr:rule>"), Variable
				.getVariable("0"));
		expectedRule = getRule("  <rdr:rule data='2.4'>"
				+ "    <rdr:cond variable='a' value='0'/>" + "  </rdr:rule>");
		assertEquals(expectedRule, consistentRule);
	}

	/**
	 * 
	 * Test the split operator with a single variable as context
	 * 
	 */
	public void testVariableSplitOperator() {
		PlainRules<Double> expectedRules = getRules("<rdr:rules>"
				+ "  <rdr:rule data='1.2'>" + "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='b' value='0'/>"
				+ "    <rdr:cond variable='c' value='0'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='1.2'>" + "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='b' value='0'/>"
				+ "    <rdr:cond variable='c' value='1'/>" + "  </rdr:rule>"
				+ "</rdr:rules>");
		Rule<Double> splittedRule = getRule("  <rdr:rule data='1.2'>"
				+ "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='b' value='0'/>" + "  </rdr:rule>");
		PlainRules<Double> resultedRules = new PlainRules<Double>();
		splittedRule.split(resultedRules, Variable.getVariable("c"), valuesSet);
		assertEquals(expectedRules, resultedRules);
	}

	/**
	 * 
	 * Test the split operator
	 * 
	 */
	public void testRecursiveSplitOperator() {
		PlainRules<Double> resultedRules = new PlainRules<Double>();
		Rule<Double> splittedRule = getRule("  <rdr:rule>"
				+ "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='b' value='0'/>" + "  </rdr:rule>");
		// Testing with rule not consistent
		Rule<Double> contextRule = getRule("  <rdr:rule>"
				+ "    <rdr:cond variable='a' value='2'/>" + "  </rdr:rule>");
		splittedRule.split(resultedRules, contextRule.getContext(), valuesSet);
		PlainRules<Double> expectedRules = getRules("<rdr:rules>" + "  <rdr:rule>"
				+ "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='b' value='0'/>" + "  </rdr:rule>"
				+ "</rdr:rules>");
		assertEquals(expectedRules, resultedRules);
		// Testing when the scope of the context is included in the rule
		contextRule = getRule("  <rdr:rule>"
				+ "    <rdr:cond variable='a' value='1'/>" + "  </rdr:rule>");
		resultedRules.clear();
		splittedRule.split(resultedRules, contextRule.getContext(), valuesSet);
		assertEquals(expectedRules, resultedRules);
		// Testing splitting against a set of variables
		splittedRule = getRule("  <rdr:rule>"
				+ "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='b' value='1'/>" + "  </rdr:rule>");
		contextRule = getRule("  <rdr:rule>"
				+ "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='c' value='0'/>"
				+ "    <rdr:cond variable='d' value='1'/>" + "  </rdr:rule>");
		resultedRules.clear();
		splittedRule.split(resultedRules, contextRule.getContext(), valuesSet);
		expectedRules = getRules("<rdr:rules>" + "  <rdr:rule>"
				+ "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='b' value='1'/>"
				+ "    <rdr:cond variable='c' value='1'/>" + "  </rdr:rule>"
				+ "  <rdr:rule>" + "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='b' value='1'/>"
				+ "    <rdr:cond variable='c' value='0'/>"
				+ "    <rdr:cond variable='d' value='1'/>" + "  </rdr:rule>"
				+ "  <rdr:rule>" + "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='b' value='1'/>"
				+ "    <rdr:cond variable='c' value='0'/>"
				+ "    <rdr:cond variable='d' value='0'/>" + "  </rdr:rule>"
				+ "</rdr:rules>");
		assertEquals(expectedRules, resultedRules);
	}

	/**
	 * 
	 * Test the classify function in exclusive and not exclusive mode
	 * 
	 */
	public void testClassify() {
		PlainRules<Double> rules = getRules("<rdr:rules>"
				+ "  <rdr:rule data='1.2'>" + "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='b' value='1'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='2.4'>" + "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='b' value='1'/>" + "  </rdr:rule>"
				+ "</rdr:rules>");
		assertEquals(1.2, rules.classify(getState("a=1,b=1", null)));
	}

	/**
	 * 
	 * Test the classify function in exclusive and not exclusive mode
	 * 
	 */
	public void testClassifyWeightedsum() {
		FunctionsSum sum = new FunctionsSum(getRoot("<fst:weightedfunctionssum>"
				+ "  <rdr:rules weight='0.5'>" + "    <rdr:rule data='1.2'>"
				+ "      <rdr:cond variable='a' value='1'/>"
				+ "      <rdr:cond variable='b' value='1'/>" + "    </rdr:rule>"
				+ "  </rdr:rules>" + "  <rdr:rules weight='2'>"
				+ "    <rdr:rule data='2.4'>"
				+ "      <rdr:cond variable='a' value='1'/>"
				+ "      <rdr:cond variable='b' value='1'/>" + "    </rdr:rule>"
				+ "  </rdr:rules>" + "</fst:weightedfunctionssum>"), variableOrdering,
				PlainRules.loadingPolicy);
		assert (Math.abs(5.4 - sum.classify(getState("a=1,b=1", null))) < Parameters.EPSILON);
	}

	/**
	 * 
	 * Test the combination opertor using sum on doubles
	 * 
	 */
	public void testAddingRules() {
		PlainRules<Double> rules01 = getRules("<rdr:rules>"
				+ "  <rdr:rule data='5'/>" + "</rdr:rules>");
		PlainRules<Double> rules02 = getRules("<rdr:rules>"
				+ "  <rdr:rule data='3'>" + "    <rdr:cond variable='a' value='1'/>"
				+ "  </rdr:rule>" + "  <rdr:rule data='4'>"
				+ "    <rdr:cond variable='a' value='0'/>" + "  </rdr:rule>"
				+ "</rdr:rules>");
		rules01.combine(rules02, PlainRules.doubleSumPolicy, valuesSet);
		PlainRules<Double> expectedRules = getRules("<rdr:rules>"
				+ "  <rdr:rule data='8'>" + "    <rdr:cond variable='a' value='1'/>"
				+ "  </rdr:rule>" + "  <rdr:rule data='9'>"
				+ "    <rdr:cond variable='a' value='0'/>" + "  </rdr:rule>"
				+ "</rdr:rules>");
		assertEquals(expectedRules, rules01);

		rules01 = getRules("<rdr:rules>" + "  <rdr:rule data='3'>"
				+ "    <rdr:cond variable='a' value='1'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='4'>" + "    <rdr:cond variable='a' value='0'/>"
				+ "  </rdr:rule>" + "</rdr:rules>");
		rules02 = getRules("<rdr:rules>" + "  <rdr:rule data='5'/>"
				+ "</rdr:rules>");
		rules01.combine(rules02, PlainRules.doubleSumPolicy, valuesSet);
		assertEquals(expectedRules, rules01);

		rules01 = getRules("<rdr:rules>" + "  <rdr:rule data='5'>"
				+ "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='b' value='1'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='2'>" + "    <rdr:cond variable='a' value='0'/>"
				+ "    <rdr:cond variable='b' value='1'/>" + "  </rdr:rule>"
				+ "</rdr:rules>");
		rules02 = getRules("<rdr:rules>" + "  <rdr:rule data='3'>"
				+ "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='c' value='0'/>"
				+ "    <rdr:cond variable='d' value='1'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='4'>" + "    <rdr:cond variable='a' value='0'/>"
				+ "    <rdr:cond variable='b' value='0'/>" + "  </rdr:rule>"
				+ "</rdr:rules>");
		rules01.combine(rules02, PlainRules.doubleSumPolicy, valuesSet);
		expectedRules = getRules("<rdr:rules>" + "  <rdr:rule data='5'>"
				+ "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='b' value='1'/>"
				+ "    <rdr:cond variable='c' value='1'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='5'>" + "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='b' value='1'/>"
				+ "    <rdr:cond variable='c' value='0'/>"
				+ "    <rdr:cond variable='d' value='0'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='8'>" + "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='b' value='1'/>"
				+ "    <rdr:cond variable='c' value='0'/>"
				+ "    <rdr:cond variable='d' value='1'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='3'>" + "    <rdr:cond variable='a' value='1'/>"
				+ "    <rdr:cond variable='b' value='0'/>"
				+ "    <rdr:cond variable='c' value='0'/>"
				+ "    <rdr:cond variable='d' value='1'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='2'>" + "    <rdr:cond variable='a' value='0'/>"
				+ "    <rdr:cond variable='b' value='1'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='4'>" + "    <rdr:cond variable='a' value='0'/>"
				+ "    <rdr:cond variable='b' value='0'/>" + "  </rdr:rule>"
				+ "</rdr:rules>");
		assertEquals(expectedRules, rules01);
	}

	/**
	 * 
	 * Test the combination opertor using sum on doubles
	 * 
	 */
	public void testMinusOnRules() {
		PlainRules<Double> rules01 = getRules("<rdr:rules>"
				+ "  <rdr:rule data='0.9'/>" + "</rdr:rules>");
		PlainRules<Double> rules02 = getRules("<rdr:rules>"
				+ "  <rdr:rule data='1.0'>" + "    <rdr:cond variable='a' value='1'/>"
				+ "  </rdr:rule>" + "  <rdr:rule data='0.0'>"
				+ "    <rdr:cond variable='a' value='0'/>" + "  </rdr:rule>"
				+ "</rdr:rules>");
		rules01.combine(rules02, PlainRules.doubleMinusPolicy, valuesSet);
		PlainRules<Double> expectedRules = getRules("<rdr:rules>"
				+ "  <rdr:rule data='-0.1'>" + "    <rdr:cond variable='a' value='1'/>"
				+ "  </rdr:rule>" + "  <rdr:rule data='0.9'>"
				+ "    <rdr:cond variable='a' value='0'/>" + "  </rdr:rule>"
				+ "</rdr:rules>");
		assert ((rules01.classify(getState("a=1", null)) - expectedRules
				.classify(getState("a=1", null))) < Parameters.EPSILON);
		assert ((rules01.classify(getState("a=0", null)) - expectedRules
				.classify(getState("a=0", null))) < Parameters.EPSILON);
	}

	/**
	 * 
	 * Test the set of rules associated to a tree
	 * 
	 */
	public void testRulesOnTree() {
		Tree<String> tree = new Tree<String>();
		RulesTree<String> testedRules = new RulesTree<String>(tree);
		tree.load(XmlUtils.getRoot("<dtr:dnode variable='e'>"
				+ "  <dtr:leaf value='0' data='Hello'/>"
				+ "  <dtr:dnode value='1' variable='d'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:leaf value='1' data='and'/>"
				+ "    <dtr:leaf value='2' data='everybody'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>"), DataLoadingPolicy.stringLoadingPolicy);
		PlainRules<String> expectedRules = new PlainRules<String>(XmlUtils
				.getRoot("<rdr:rules>" + "  <rdr:rule data='Hello'>"
						+ "    <rdr:cond variable='e' value='0'/>" + "  </rdr:rule>"
						+ "  <rdr:rule data='world'>"
						+ "    <rdr:cond variable='e' value='1'/>"
						+ "    <rdr:cond variable='d' value='0'/>" + "  </rdr:rule>"
						+ "  <rdr:rule data='and'>"
						+ "    <rdr:cond variable='e' value='1'/>"
						+ "    <rdr:cond variable='d' value='1'/>" + "  </rdr:rule>"
						+ "  <rdr:rule data='everybody'>"
						+ "    <rdr:cond variable='e' value='1'/>"
						+ "    <rdr:cond variable='d' value='2'/>" + "  </rdr:rule>"
						+ "</rdr:rules>"), DataLoadingPolicy.stringLoadingPolicy);
		assertEquals(expectedRules, testedRules);

		testedRules = new RulesTree<String>(tree);
		assertEquals(expectedRules, testedRules);
	}

}
