/*
 * 
 * Created on Dec 29, 2005
 *
 */
package animatlab.kodamat.core.trees;

import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;

/**
 * 
 * Policy used to convert a tree of something into a tree of something else
 * 
 * @param <O>
 *          the type of something
 * @param <R>
 *          the type of something else
 * 
 * @author Thomas Degris
 * 
 */
abstract public class ConvertPolicy<O, R> {
	/**
	 * 
	 * Convert a data of something into something else
	 * 
	 * @param leaf
	 *          the data to convert
	 * 
	 * @return the data converted
	 * 
	 */
	abstract public R convert(Leaf<O> leaf);

	/**
	 * 
	 * Clone a tree of something into a tree of something else
	 * 
	 * @param source
	 *          the original tree of something
	 * @param target
	 *          the tree of something else (warning: all its content is removed)
	 * @return the tree converted
	 * 
	 */
	public Tree<R> convert(Tree<O> source, Tree<R> target) {
		createVertex(null, source.getRoot(), target, null);
		return target;
	}

	/**
	 * 
	 * Clone a tree of something into a tree of something else
	 * 
	 * @param valuesSet
	 *          set of values to guarantee that for a tested variables, it exists
	 *          a child for each value (may be null)
	 * @param source
	 *          the original tree of something
	 * @param target
	 *          the tree of something else (warning: all its content is removed)
	 * @return the tree converted
	 * 
	 */
	public Tree<R> convert(Map<Variable, Set<String>> valuesSet, Tree<O> source,
			Tree<R> target) {
		if (source == null) {
			Leaf<R> leaf = (Leaf<R>) target.getRoot();
			leaf.set(convert(null));
			return target;
		}
		createVertex(valuesSet, source.getRoot(), target, null);
		return target;
	}

	/**
	 * 
	 * Clone a tree of something into a tree of something else
	 * 
	 * @param source
	 *          the original tree of something
	 * @param target
	 *          the tree of something else (warning: all its content is removed)
	 * @param testsList
	 *          the set of constraint of this tree
	 * @return the tree converted
	 */
	public Tree<R> convert(Tree<O> source, Tree<R> target, Context testsList) {
		createVertex(source.getRoot(), target, null, null, testsList);
		return target;
	}

	/**
	 * 
	 * Create a new vertex in this tree from another tree
	 * 
	 * @param valuesSet
	 *          set of values to guarantee that for a tested variables, it exists
	 *          a child for each value (may be null)
	 * @param source
	 *          vertex to convert from the other tree
	 * @param tree
	 *          the result tree
	 * @param parent
	 *          the parent of this vertex
	 * 
	 * @return the newly created vertex
	 * 
	 */
	protected Vertex<R> createVertex(Map<Variable, Set<String>> valuesSet,
			Vertex<O> source, Tree<R> tree, DNode<R> parent) {
		if (source.isLeaf) {
			return convertLeaf(source, tree, parent, getTestedValue(source));
		}

		DNode<O> sourceDNode = (DNode<O>) source;
		DNode<R> dnode = tree.createDNode(parent, getTestedValue(sourceDNode),
				sourceDNode.getTestedVariable());
		for (Vertex<O> child : sourceDNode.children()) {
			createVertex(valuesSet, child, tree, dnode);
		}
		Set<String> valueSet = null;
		if (valuesSet != null) {
			valueSet = valuesSet.get(sourceDNode.getTestedVariable());
		}
		if ((valueSet != null) && (dnode.children().size() < valueSet.size())) {
			valueSet = new LinkedHashSet<String>(valueSet);
			valueSet.removeAll(dnode.getChildrenValues());
			for (String value : valueSet) {
				convertLeaf(null, tree, dnode, value);
			}
		}
		return dnode;
	}

	/**
	 * 
	 * Convert a leaf
	 * 
	 * @param source
	 *          vertex to convert from the other tree
	 * @param tree
	 *          the result tree
	 * @param parent
	 *          the parent of this vertex
	 * @param value
	 *          the value to reach the leaf
	 * @return the new leaf
	 * 
	 */
	private Leaf<R> convertLeaf(Vertex<O> source, Tree<R> tree, DNode<R> parent,
			String value) {
		Leaf<O> sourceLeaf = (Leaf<O>) source;
		Leaf<R> leaf = tree.createLeaf(parent, value);
		if (sourceLeaf != null) {
			leaf.set(convert(sourceLeaf));
		} else {
			leaf.set(convert(null));
		}
		return leaf;
	}

	/**
	 * 
	 * Create a new vertex in this tree from another tree
	 * 
	 * @param source
	 *          vertex to convert from the other tree
	 * @param tree
	 *          the result tree
	 * @param parent
	 *          the parent of this vertex
	 * @param testsList
	 *          the set of constraint of this tree
	 * @param value
	 *          value to reach the converted node from the parent
	 * 
	 * @return the newly created vertex
	 * 
	 */
	private Vertex<R> createVertex(Vertex<O> source, Tree<R> tree,
			DNode<R> parent, String value, Context testsList) {
		Vertex<O> currentVertex = source;
		Vertex<O> selectedVertex;
		do {
			selectedVertex = null;
			if (!currentVertex.isLeaf) {
				DNode<O> sourceDNode = (DNode<O>) currentVertex;
				String selectedValue = null;
				if (testsList != null) {
					selectedValue = testsList.get(sourceDNode.testedVariable);
				}
				if (selectedValue != null) {
					selectedVertex = sourceDNode.getChild(selectedValue);
				}
				if ((selectedVertex == null) && (sourceDNode.children().size() == 1)) {
					selectedVertex = sourceDNode.children().iterator().next();
				}
				if (selectedVertex != null) {
					currentVertex = selectedVertex;
				}
			}
		} while (selectedVertex != null);

		Vertex<R> result;
		if (currentVertex.isLeaf) {
			result = convertLeaf(currentVertex, tree, parent, value);
		} else {
			DNode<O> sourceDNode = (DNode<O>) currentVertex;
			DNode<R> dnode = tree.createDNode(parent, value, sourceDNode
					.getTestedVariable());
			for (Vertex<O> child : sourceDNode.children()) {
				createVertex(child, tree, dnode, getTestedValue(child), testsList);
			}
			result = dnode;
		}
		return result;
	}

	/**
	 * 
	 * Get the value instead of the wildcard
	 * 
	 * @param source
	 *          the child node concerned by the transformation
	 * 
	 * @return the value replacing the wildcard
	 * 
	 */
	private String getTestedValue(Vertex<O> source) {
		if (source.parent == null) {
			return null;
		}
		// FIXME: binary value hack
		// Hack to avoid the conversion of non binary tree
		// using wildcards
		String result = source.getTestedValue();
		if (!result.equals(DNode.VALUE_WILDCARD)) {
			return result;
		}
		assert (source.parent.children().size() <= 2);
		for (Vertex<O> child : source.parent.children()) {
			if (child != source) {
				if (child.value.equals("0")) {
					return "1";
				} else if (child.value.equals("1")) {
					return "0";
				}
				if (child.value.equals("true")) {
					return "false";
				}
				if (child.value.equals("false")) {
					return "true";
				}
				assert (false);
				return null;
			}
		}
		assert (false);
		return null;
	}

}
