/*
 * 
 * Created on 22 dec. 2005
 *
 */
package animatlab.kodamat.core.trees;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import animatlab.kodamat.core.io.DotUtils;
import animatlab.kodamat.core.io.IContainerToString;
import animatlab.kodamat.core.io.DotUtils.ImageMap;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;

/**
 * 
 * An decision node (containing a test) in the tree
 * 
 * @author Thomas Degris
 * @param <T>
 *          the type of instances contained in the tree
 * 
 */
public class DNode<T> extends Vertex<T> {
	/**
	 * 
	 * Name of the node in xml to represent an decision node
	 * 
	 */
	public static final String NODE_NAME = Tree.NAMESPACE + ":" + "dnode";

	/**
	 * 
	 * Wildcard that can be used for values
	 * 
	 */
	static public final String VALUE_WILDCARD = "*";

	/**
	 * 
	 * Variable tested at this decision node
	 * 
	 */
	protected Variable testedVariable;

	/**
	 * 
	 * Map of value to children of this decision node
	 * 
	 */
	protected final Map<String, Vertex<T>> children = new LinkedHashMap<String, Vertex<T>>();

	/**
	 * 
	 * Constructor
	 * 
	 * @param tree
	 *          the tree owning this decision node
	 * @param value
	 *          value of the tested variable
	 * @param parent
	 *          Parent in the tree of this node
	 * @param testedVariable
	 *          tested variable at this decision node
	 * 
	 */
	public DNode(Tree<T> tree, DNode<T> parent, String value,
			Variable testedVariable) {
		super(tree, parent, value, false);
		this.testedVariable = testedVariable;
	}

	/**
	 * @see animatlab.kodamat.core.trees.Vertex#replace(animatlab.kodamat.core.utils.Variable)
	 */
	@Override
	public DNode<T> replace(Variable testedVariable) {
		makeChildrenOrphan();
		tree.remove(this);
		this.testedVariable = testedVariable;
		tree.add(this);
		return this;
	}

	/**
	 * @see animatlab.kodamat.core.trees.Vertex#remove()
	 */
	@Override
	public void remove() {
		makeChildrenOrphan();
		tree.remove(this);
		super.remove();
	}

	/**
	 * 
	 * Make all the children of this node (and their children) orphan
	 * 
	 */
	private void makeChildrenOrphan() {
		while (!children.values().isEmpty()) {
			children.values().iterator().next().remove();
		}
		assert (children.isEmpty());
	}

	/**
	 * @see animatlab.kodamat.core.trees.Vertex#toString(int)
	 */
	@Override
	public String toString(int indentation) {
		StringBuilder result = new StringBuilder(super.toString(indentation));
		result.append("(");
		result.append(testedVariable.toString());
		result.append("=  ");
		result.append(dataToString());
		result.append(localInfoToString());
		result.append("\n");
		List<String> sortedValues = new ArrayList<String>(children.keySet());
		Collections.sort(sortedValues);
		for (String value : sortedValues) {
			Vertex<T> child = getChild(value);
			result.append(child.toString(indentation + 1));
		}
		return result.toString();
	}

	/**
	 * 
	 * @return a string representation of the local info of this decision node
	 * 
	 */
	protected String localInfoToString() {
		return "";
	}

	/**
	 * 
	 * @return private data from type inheriting of this
	 * 
	 */
	protected String dataToString() {
		return "";
	}

	/**
	 * 
	 * Register a vertex as child of this decision node
	 * 
	 * @param value
	 *          the value to reach this child
	 * @param vertex
	 *          the vertex to register as child
	 * 
	 */
	final protected void registerAsChild(String value, Vertex<T> vertex) {
		Vertex<T> oldChild = children.get(value);
		assert (oldChild != vertex);
		if (oldChild != null) {
			oldChild.remove();
		}
		children.put(value, vertex);
	}

	/**
	 * 
	 * Unregister a vertex as child of this decision node
	 * 
	 * @param value
	 *          the value to reach this child
	 * 
	 */
	final protected void unregisterAsChild(String value) {
		children.remove(value);
	}

	/**
	 * 
	 * Add a new leaf to this decision node
	 * 
	 * @param value
	 *          the value to reach this node
	 * @param data
	 *          the data associated with the new leaf (may be null)
	 * 
	 * @return the newly created leaf
	 * 
	 */
	public Leaf<T> addLeaf(String value, T data) {
		Leaf<T> leaf = addLeaf(value);
		leaf.set(data);
		return leaf;
	}

	/**
	 * 
	 * Add a new leaf to this decision node
	 * 
	 * @param value
	 *          the value to reach this node
	 * 
	 * @return the newly created leaf
	 * 
	 */
	public Leaf<T> addLeaf(String value) {
		Leaf<T> leaf = tree.createLeaf(this, value);
		return leaf;
	}

	/**
	 * 
	 * Add a new decision node to this decision node
	 * 
	 * @param value
	 *          the value to reach this node
	 * @param testedVariable
	 *          the variable tested at the new child decision node
	 * 
	 * @return the newly created decision node
	 * 
	 */
	public DNode<T> addDNode(String value, Variable testedVariable) {
		return tree.createDNode(this, value, testedVariable);
	}

	/**
	 * 
	 * Remove a child from the children list
	 * 
	 * @param value
	 *          the value that was used to reach the child to remove
	 * 
	 */
	final protected void removeChild(String value) {
		children.remove(value);
	}

	/**
	 * @see animatlab.kodamat.core.trees.Vertex#clone(animatlab.kodamat.core.trees.ClonePolicy,
	 *      animatlab.kodamat.core.trees.Tree, animatlab.kodamat.core.trees.DNode,
	 *      java.lang.String, animatlab.kodamat.core.utils.Context)
	 */
	@Override
	protected Vertex<T> clone(ClonePolicy<T> clonePolicy, Tree<T> result,
			DNode<T> parent, String value, Context testsList) {
		String constraint = null;
		if (testsList != null) {
			constraint = testsList.get(testedVariable);
		}
		if (constraint != null) {
			Vertex<T> child = getChild(constraint);
			if (child == null) {
				Leaf<T> newLeaf = result.createLeaf(parent, value);
				if (clonePolicy != null) {
					newLeaf.data = clonePolicy.cloneData(null);
				}
				return newLeaf;
			}
			return getChild(constraint).clone(clonePolicy, result, parent, value,
					testsList);
		}
		DNode<T> dnode = result.createDNode(parent, value, testedVariable);
		for (Map.Entry<String, Vertex<T>> entry : children.entrySet()) {
			entry.getValue().clone(clonePolicy, result, dnode, entry.getKey(),
					testsList);
		}
		return dnode;
	}

	/**
	 * 
	 * @return a collection of the children of this decision node
	 * 
	 */
	public Collection<Vertex<T>> children() {
		return children.values();
	}

	/**
	 * 
	 * Create a new set and put the children in
	 * 
	 * @return a set of all the children independant of the set used by this
	 *         decision node
	 * 
	 */
	public Collection<Vertex<T>> getNewChildrenSet() {
		return new LinkedHashSet<Vertex<T>>(children.values());
	}

	/**
	 * 
	 * Get a child from the value of the variable
	 * 
	 * @param string
	 *          the value of the variable
	 * 
	 * @return the child
	 * 
	 */
	public Vertex<T> getChild(String string) {
		Vertex<T> result = children.get(string);
		if (result == null) {
			return children.get(VALUE_WILDCARD);
		}
		return result;
	}

	/**
	 * 
	 * Get a child from an array of value
	 * 
	 * @param values
	 *          the set of values for variable
	 * 
	 * @return the child
	 * 
	 */
	final public Vertex<T> getChild(Context values) {
		return getChild(values.get(testedVariable));
	}

	/**
	 * 
	 * @return get the tested variable at this decision node
	 * 
	 */
	final public Variable getTestedVariable() {
		return testedVariable;
	}

	/**
	 * @see animatlab.kodamat.core.trees.Vertex#changeTree(animatlab.kodamat.core.trees.Tree)
	 */
	@Override
	protected void changeTree(Tree<T> tree) {
		super.changeTree(tree);
		for (Vertex<T> child : children()) {
			child.changeTree(tree);
		}
	}

	/**
	 * @see animatlab.kodamat.core.trees.Vertex#clearTestsList()
	 */
	@Override
	protected void clearTestsList() {
		if (testsList != null) {
			super.clearTestsList();
			for (Vertex<T> child : children()) {
				child.clearTestsList();
			}
		}
	}

	/**
	 * @see animatlab.kodamat.core.trees.Vertex#cut(animatlab.kodamat.core.utils.Context)
	 */
	@Override
	public void cut(Context testsList) {
		if (children.size() == 0) {
			return;
		}
		Vertex<T> selectedChild = getSelectedChild(testsList);
		if (selectedChild == null) {
			for (Vertex<T> child : children.values()) {
				child.cut(testsList);
			}
			return;
		}
		selectedChild.cut(testsList);
		replaceWithChild(getSelectedChild(testsList));
	}

	/**
	 * 
	 * Get the selected child in a child list when the decision node is simplified
	 * with a constraint list
	 * 
	 * @param testsList
	 *          the constraint list
	 * 
	 * @return the selected child replacing this node
	 * 
	 */
	private Vertex<T> getSelectedChild(Context testsList) {
		Vertex<T> selectedChild = null;
		String selectedValue = testsList.get(testedVariable);
		if (selectedValue != null) {
			selectedChild = getChild(selectedValue);
		}
		if (children.size() == 1) {
			selectedChild = children.values().iterator().next();
		}
		return selectedChild;
	}

	/**
	 * @see animatlab.kodamat.core.trees.Vertex#simplify(EqualPolicy,
	 *      MergeLeavesPolicy)
	 */
	@Override
	public void simplify(EqualPolicy<T> equalPolicy,
			MergeLeavesPolicy<T> mergePolicy) {
		if (children.size() == 0) {
			return;
		}
		for (Vertex<T> child : children.values()) {
			child.simplify(equalPolicy, mergePolicy);
		}

		Vertex<T> last = null;
		for (Vertex<T> child : children.values()) {
			if ((last != null) && !last.equals(child, equalPolicy)) {
				return;
			}
			last = child;
		}
		Vertex<T> selected = mergePolicy.merge(children.values());
		if (selected != null) {
			replaceWithChild(selected);
		}
	}

	/**
	 * 
	 * Replace this decision node by a selected child
	 * 
	 * @param selected
	 *          the child selected to replace this node
	 * 
	 */
	protected void replaceWithChild(Vertex<T> selected) {
		Collection<Vertex<T>> childSet = getNewChildrenSet();
		for (Vertex<T> child : childSet) {
			if (child != selected) {
				child.remove();
			}
		}

		selected.clearTestsList();
		selected.parent = parent;
		selected.value = value;
		tree.remove(this);
		if (parent != null) {
			parent.children.put(value, selected);
		} else {
			tree.root = selected;
		}
	}

	/**
	 * @see animatlab.kodamat.core.trees.Vertex#equals(animatlab.kodamat.core.trees.Vertex,
	 *      EqualPolicy)
	 */
	@SuppressWarnings("unchecked")
	@Override
	public boolean equals(Vertex<T> vertex, EqualPolicy<T> equalPolicy) {
		if (!super.equals(vertex, equalPolicy)) {
			return false;
		}
		DNode<T> dnode = (DNode<T>) vertex;
		if (!testedVariable.equals(dnode.testedVariable)) {
			return false;
		}
		if (children.size() != dnode.children.size()) {
			return false;
		}
		for (Map.Entry<String, Vertex<T>> entry : children.entrySet()) {
			if (!entry.getValue().equals(dnode.children.get(entry.getKey()),
					equalPolicy)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * @see animatlab.kodamat.core.io.IToXml#toXml(org.w3c.dom.Document,
	 *      animatlab.kodamat.core.io.IToXml.IToXmlPolicy)
	 */
	public Element toXml(Document document, IToXmlPolicy<T> policy) {
		Element dnodeXml = document.createElement(NODE_NAME);
		if (parent != null) {
			dnodeXml.setAttribute(ATTRIBUTE_VALUE, value);
		}
		testedVariable.setAttribute(dnodeXml);
		for (Vertex<T> child : children.values()) {
			dnodeXml.appendChild(child.toXml(document, policy));
		}
		return dnodeXml;
	}

	/**
	 * 
	 * @return the set of values to reach the children
	 * 
	 */
	public Set<String> getChildrenValues() {
		return children.keySet();
	}

	/**
	 * @see animatlab.kodamat.core.io.IToDot#toDot(java.lang.StringBuilder,
	 *      animatlab.kodamat.core.io.IToDot.IToDotPolicy, ImageMap)
	 */
	@SuppressWarnings("unchecked")
	public void toDot(StringBuilder output, IToDotPolicy<T> policy,
			ImageMap<T> imageMap) {
		String label = DotUtils.toNormalizeString(getTestedVariable().toString())
				.toLowerCase();
		if (policy instanceof IContainerToString) {
			label = ((IContainerToString<Vertex<T>>) policy).containerToString(this,
					label);
		}
		output.append(dotName() + " [" + "label=\"" + label + "\""
				+ imageMap.registerDotName(dotName(), this) + "];\n");
		ArrayList<String> values = new ArrayList<String>(children.keySet());
		Collections.sort(values);
		Collections.reverse(values);
		for (String value : values) {
			Vertex<T> child = getChild(value);
			output.append(dotName() + " -> " + child.dotName() + " [label=\"" + value
					+ "\"];\n");
			if (!child.isLeaf) {
				child.toDot(output, policy, imageMap);
			}
		}
	}

	/**
	 * 
	 * Replace this decision with a leaf
	 * 
	 * @return the leaf replacing this node
	 * 
	 */
	public Leaf<T> replaceWithLeaf() {
		String testedValue = getTestedValue();
		DNode<T> parent = getParent();
		Tree<T> owningTree = tree;
		remove();
		return owningTree.createLeaf(parent, testedValue);
	}

}
