/*
 * 
 * Created on 22 dec. 2005
 *
 */
package animatlab.kodamat.core.trees;

import java.util.Set;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import animatlab.kodamat.core.functions.IPartition;
import animatlab.kodamat.core.io.IContainerToString;
import animatlab.kodamat.core.io.DotUtils.ImageMap;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;

/**
 * 
 * An leaf (containing data) in the tree
 * 
 * @param <T>
 *          the type of instances contained in the tree
 * 
 * @author Thomas Degris
 * 
 */
public class Leaf<T> extends Vertex<T> implements IPartition<T> {
	/**
	 * 
	 * Attribute used in a xml node to specify the variable to test
	 * 
	 */
	static public final String NODE_NAME = Tree.NAMESPACE + ":" + "leaf";

	/**
	 * 
	 * The data associated with this node
	 * 
	 */
	protected T data = null;

	/**
	 * 
	 * Constructor
	 * 
	 * @param tree
	 *          the tree owning this dnode
	 * @param value
	 *          value of the tested variable
	 * @param parent
	 *          Parent in the tree of this node
	 * 
	 */
	public Leaf(Tree<T> tree, DNode<T> parent, String value) {
		super(tree, parent, value, true);
	}

	/**
	 * @param data
	 *          the data to set
	 * @see animatlab.kodamat.core.io.DataContainer#set(Object)
	 */
	public void set(T data) {
		this.data = data;
	}

	/**
	 * @see animatlab.kodamat.core.io.DataContainer#get()
	 */
	public T get() {
		return data;
	}

	/**
	 * @see animatlab.kodamat.core.trees.Vertex#replace(animatlab.kodamat.core.utils.Variable)
	 */
	@Override
	public DNode<T> replace(Variable testedVariable) {
		return tree.createDNode(parent, value, testedVariable);
	}

	/**
	 * @see animatlab.kodamat.core.trees.Vertex#remove()
	 */
	@Override
	public void remove() {
		tree.remove(this);
		super.remove();
	}

	/**
	 * 
	 * @return convert the data at the leaf into a string
	 * 
	 */
	protected String dataToString() {
		return get().toString();
	}

	/**
	 * @see animatlab.kodamat.core.trees.Vertex#toString(int)
	 */
	@Override
	public String toString(int indentation) {
		StringBuilder result = new StringBuilder(super.toString(indentation));
		result.append(": ");
		if (get() == null) {
			result.append("null");
		} else {
			result.append(dataToString());
		}
		result.append("\n");
		return result.toString();
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return toString(0);
	}

	/**
	 * @see animatlab.kodamat.core.trees.Vertex#clone(animatlab.kodamat.core.trees.ClonePolicy,
	 *      animatlab.kodamat.core.trees.Tree, animatlab.kodamat.core.trees.DNode,
	 *      java.lang.String, animatlab.kodamat.core.utils.Context)
	 */
	@Override
	protected Leaf<T> clone(ClonePolicy<T> clonePolicy, Tree<T> result,
			DNode<T> parent, String value, Context testsList) {
		Leaf<T> leaf = result.createLeaf(parent, value);
		if (data != null) {
			if (clonePolicy != null) {
				leaf.data = clonePolicy.cloneData(data);
			} else {
				leaf.data = data;
			}
		}
		return leaf;
	}

	/**
	 * @see animatlab.kodamat.core.trees.Vertex#simplify(EqualPolicy,
	 *      MergeLeavesPolicy)
	 */
	@Override
	public void simplify(EqualPolicy<T> equalPolicy,
			MergeLeavesPolicy<T> mergePolicy) {
		return;
	}

	/**
	 * @see animatlab.kodamat.core.trees.Vertex#cut(animatlab.kodamat.core.utils.Context)
	 */
	@Override
	public void cut(Context testsList) {
		return;
	}

	/**
	 * @see animatlab.kodamat.core.trees.Vertex#equals(animatlab.kodamat.core.trees.Vertex,
	 *      EqualPolicy)
	 */
	@SuppressWarnings("unchecked")
	@Override
	public boolean equals(Vertex<T> vertex, EqualPolicy<T> equalPolicy) {
		if (!super.equals(vertex, equalPolicy)) {
			return false;
		}
		Leaf<T> leaf = (Leaf<T>) vertex;
		if (data == leaf.data) {
			return true;
		}
		if (data != null) {
			if (equalPolicy != null) {
				return equalPolicy.equals(this, leaf);
			}
			return data.equals(leaf.data);
		}
		return false;
	}

	/**
	 * 
	 * Append a tree at this leaf
	 * 
	 * @param subtree
	 *          the subtree to append to this leaf
	 * 
	 */
	final public void replace(Tree<T> subtree) {
		tree.replace(this, subtree);
	}

	/**
	 * @see animatlab.kodamat.core.io.IToXml#toXml(org.w3c.dom.Document,
	 *      animatlab.kodamat.core.io.IToXml.IToXmlPolicy)
	 */
	public Element toXml(Document document, IToXmlPolicy<T> policy) {
		Element leafXml = document.createElement(NODE_NAME);
		if (parent != null) {
			leafXml.setAttribute(ATTRIBUTE_VALUE, value);
		}
		if (data != null) {
			policy.toXml(leafXml, data);
		}
		return leafXml;
	}

	/**
	 * @see animatlab.kodamat.core.functions.IPartition#scope()
	 */
	public Set<Variable> scope() {
		return getContext().keySet();
	}

	/**
	 * 
	 * Get the label of this leaf
	 * 
	 * @param policy
	 *          the policy to use to get the label of the data
	 * 
	 * @return the dot label of this leaf
	 * 
	 */
	protected String getDotLabel(IToDotPolicy<T> policy) {
		String leafData = "";
		if (get() != null) {
			leafData = policy.toDotLabel(get());
		}
		return leafData;
	}

	/**
	 * @see animatlab.kodamat.core.io.IToDot#toDot(StringBuilder,
	 *      animatlab.kodamat.core.io.IToDot.IToDotPolicy, ImageMap)
	 */
	@SuppressWarnings("unchecked")
	public void toDot(StringBuilder output, IToDotPolicy<T> policy,
			ImageMap<T> imageMap) {
		String label = getDotLabel(policy);
		if (policy instanceof IContainerToString) {
			label = ((IContainerToString<Vertex<T>>) policy).containerToString(this,
					label);
		}
		output.append(dotName() + "[label=\"" + label + "\" ");
		output.append("shape=box ");
		output.append(imageMap.registerDotName(dotName(), this));
		output.append("];\n");
	}

	/**
	 * 
	 * Method called when this leaf is used as prototype
	 * 
	 * @param tree
	 *          the tree owning this dnode
	 * @param value
	 *          value of the tested variable
	 * @param parent
	 *          Parent in the tree of this node
	 * 
	 * @return the new leaf
	 * 
	 */
	public Leaf<T> newInstance(Tree<T> tree, DNode<T> parent, String value) {
		return new Leaf<T>(tree, parent, value);
	}
}
