/*
 * 
 * Created on 22 dec. 2005
 *
 */
package animatlab.kodamat.core.trees;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import animatlab.kodamat.core.functions.IPartition;
import animatlab.kodamat.core.functions.IPartitionFunction;
import animatlab.kodamat.core.io.DataLoadingPolicy;
import animatlab.kodamat.core.io.IToDot;
import animatlab.kodamat.core.io.IToXml;
import animatlab.kodamat.core.io.XmlUtils;
import animatlab.kodamat.core.io.DotUtils.ImageMap;
import animatlab.kodamat.core.utils.Prototype;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;

/**
 * 
 * Decision tree in Spiti
 * 
 * @param <T>
 *          The type of instances contained in the tree
 * 
 * @author Thomas Degris
 * 
 */
public class Tree<T> implements Cloneable, IPartitionFunction<T>, IToXml<T>,
		IToDot<T>, Prototype {
	/**
	 * 
	 * Namespace of xml node name
	 * 
	 */
	public static final String NAMESPACE = "dtr";

	/**
	 * 
	 * Policy to load a rule
	 * 
	 */
	final static public IPartitionFunction.LoadXmlPolicy<Double> loadingPolicy = new IPartitionFunction.LoadXmlPolicy<Double>() {
		public IPartitionFunction<Double> loadXml(Element element) {
			return new Tree<Double>(element, DataLoadingPolicy.doubleLoadingPolicy);
		}
	};

	/**
	 * 
	 * A description of this tree
	 * 
	 */
	private String description = null;

	/**
	 * 
	 * The root of this tree
	 * 
	 */
	protected Vertex<T> root = null;

	/**
	 * 
	 * Number of nodes in the tree
	 * 
	 */
	protected int numberOfNodes = 0;

	/**
	 * 
	 * Set of leaves in this tree (maintained each time a leaf is removed or
	 * added)
	 * 
	 */
	final Set<IPartition<T>> leafSet = new LinkedHashSet<IPartition<T>>();

	/**
	 * 
	 * Leaf listener to call every time a leaf is created or destroyed
	 * 
	 */
	private LeafListener leafListener = null;

	/**
	 * 
	 * Map used to cache scope info
	 * 
	 */
	private Map<Variable, Integer> scopeMap = new LinkedHashMap<Variable, Integer>();

	/**
	 * 
	 * Default constructor
	 * 
	 */
	public Tree() {
		super();
	}

	/**
	 * 
	 * Constructor from a xml tree
	 * 
	 * @param element
	 *          root of the xml tree
	 * @param dataLoadingPolicy
	 *          the policy to use to create the nodes
	 * 
	 */
	public Tree(Element element, DataLoadingPolicy<T> dataLoadingPolicy) {
		load(element, dataLoadingPolicy);
	}

	/**
	 * 
	 * Set this tree from a xml tree
	 * 
	 * @param element
	 *          root of the xml tree
	 * @param dataLoadingPolicy
	 *          the policy to use to create the nodes
	 * 
	 */
	public void load(Element element, DataLoadingPolicy<T> dataLoadingPolicy) {
		setRoot(createVertex(element, null, dataLoadingPolicy));
		description = element.getAttribute("description");
	}

	/**
	 * 
	 * Set the root of the tree
	 * 
	 * @param root
	 *          the new root of the tree (may be null if the tree is empty)
	 * 
	 */
	final protected void setRoot(Vertex<T> root) {
		this.root = root;
	}

	/**
	 * 
	 * Create a new node in this tree from a xml element
	 * 
	 * @param element
	 *          xml element to create the node from
	 * @param parent
	 *          the parent of this new node
	 * @param dataLoadingPolicy
	 *          the policy to use to create the nodes
	 * 
	 * @return the new node (composed of its children)
	 * 
	 */
	private Vertex<T> createVertex(Element element, DNode<T> parent,
			DataLoadingPolicy<T> dataLoadingPolicy) {
		Vertex<T> result;
		String value = element.getAttribute(Vertex.ATTRIBUTE_VALUE);
		if (value.length() == 0) {
			value = null;
		}
		if (element.getNodeName().equals(DNode.NODE_NAME)) {
			Variable testedVariable = Variable.load(element);
			if (parent == null) {
				result = createDNode(parent, value, testedVariable);
			} else {
				result = parent.addDNode(value, testedVariable);
			}
		} else {
			Leaf<T> leaf;
			if (parent != null) {
				leaf = parent.addLeaf(value);
			} else {
				leaf = createLeaf(null, null);
			}
			if (dataLoadingPolicy != null) {
				leaf.set(dataLoadingPolicy.xmlToData(element));
			}
			result = leaf;
		}
		if (element.getNodeName().equals(DNode.NODE_NAME)) {
			NodeList nodeList = element.getChildNodes();
			for (int index = 0; index < nodeList.getLength(); index++) {
				Node childNode = nodeList.item(index);
				if (childNode.getNodeType() == Node.ELEMENT_NODE) {
					createVertex((Element) childNode, (DNode<T>) result,
							dataLoadingPolicy);
				}
			}
		}
		return result;
	}

	/**
	 * 
	 * @return Returns the root of this tree.
	 * 
	 */
	final public Vertex<T> getRoot() {
		if (root == null) {
			createLeaf(null, null);
		}
		return root;
	}

	/**
	 * 
	 * @return the number of leaves in the tree
	 * 
	 */
	final public int numberOfLeaves() {
		return leafSet.size();
	}

	/**
	 * @see animatlab.kodamat.core.functions.IFunction#numberOfNodes()
	 */
	final public int numberOfNodes() {
		return numberOfNodes;
	}

	/**
	 * 
	 * Remove a vertex in this tree
	 * 
	 * @param vertex
	 *          the vertex to remove
	 * 
	 */
	protected void remove(Vertex<T> vertex) {
		numberOfNodes--;
		if (vertex.isLeaf) {
			if (leafListener != null) {
				leafListener.leafIsRemoved((Leaf<T>) vertex);
			}
			leafSet.remove(vertex);
		} else {
			Variable testedVariable = ((DNode<T>) vertex).getTestedVariable();
			int variableCount = scopeMap.get(testedVariable);
			if (variableCount == 1) {
				scopeMap.remove(testedVariable);
			} else {
				scopeMap.put(testedVariable, variableCount - 1);
			}
		}
		if (root == vertex) {
			setRoot(null);
		}
	}

	/**
	 * 
	 * @return check the scope of the tree (used only for debugging)
	 * 
	 */
	protected boolean checkScopeDebug() {
		Set<Variable> scope = new LinkedHashSet<Variable>();
		for (IPartition<?> partition : partitions()) {
			scope.addAll(partition.getContext().keySet());
		}
		return scope.equals(scope());
	}

	/**
	 * 
	 * Add a vertex in this tree
	 * 
	 * @param vertex
	 *          the vertex to add
	 * 
	 */
	protected void add(Vertex<T> vertex) {
		numberOfNodes++;
		if (vertex.isLeaf) {
			leafSet.add((Leaf<T>) vertex);
		} else {
			Variable testedVariable = ((DNode<T>) vertex).getTestedVariable();
			Integer variableCount = scopeMap.get(testedVariable);
			if (variableCount == null) {
				scopeMap.put(testedVariable, 1);
			} else {
				scopeMap.put(testedVariable, variableCount + 1);
			}
		}
		if (vertex.getParent() == null) {
			setRoot(vertex);
		}
	}

	/**
	 * @see animatlab.kodamat.core.utils.Prototype#newInstance()
	 */
	public Tree<T> newInstance() {
		return new Tree<T>();
	}

	/**
	 * 
	 * Create a new leaf in this tree
	 * 
	 * @param parent
	 *          the parent of this leaf
	 * @param value
	 *          the value of the tested variable to reach this node
	 * 
	 * @return the new leaf
	 * 
	 */
	protected Leaf<T> newLeaf(DNode<T> parent, String value) {
		return new Leaf<T>(this, parent, value);
	}

	/**
	 * 
	 * Create a new decision node in this tree
	 * 
	 * @param parent
	 *          the parent of this decision node
	 * @param value
	 *          the value of the tested variable to reach this node
	 * @param testedVariable
	 *          the variable tested at the decision node
	 * 
	 * @return the new decision node
	 * 
	 */
	protected DNode<T> newDNode(DNode<T> parent, String value,
			Variable testedVariable) {
		return new DNode<T>(this, parent, value, testedVariable);
	}

	/**
	 * 
	 * Create a new decision node in this tree
	 * 
	 * @param parent
	 *          the parent of this decision node
	 * @param value
	 *          the value of the tested variable to reach this node
	 * @param testedVariable
	 *          the variable tested at the decision node
	 * 
	 * @return the new decision node
	 * 
	 */
	final protected DNode<T> createDNode(DNode<T> parent, String value,
			Variable testedVariable) {
		DNode<T> dnode = newDNode(parent, value, testedVariable);
		if (parent == null) {
			if (root != null) {
				root.remove();
			}
		}
		add(dnode);
		return dnode;
	}

	/**
	 * 
	 * Create a new leaf in this tree
	 * 
	 * @param parent
	 *          the parent of this leaf
	 * @param value
	 *          the value of the tested variable to reach this node
	 * 
	 * @return the new leaf
	 * 
	 */
	public final Leaf<T> createLeaf(DNode<T> parent, String value) {
		Leaf<T> leaf = newLeaf(parent, value);
		add(leaf);
		if (leafListener != null) {
			leafListener.leafIsCreated(leaf);
		}
		if (parent == null) {
			setRoot(leaf);
		}
		return leaf;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		if (root == null) {
			return "null";
		}
		StringBuilder result = new StringBuilder();
		result.append("(");
		result.append(numberOfLeaves());
		result.append("/");
		result.append(numberOfNodes());
		result.append(")\n");
		result.append(root.toString());
		return result.toString();
	}

	/**
	 * 
	 * Clone this tree executing a clone policy for the data associated at each
	 * leaf
	 * 
	 * @param testsList
	 *          the constraint to apply on the clone of this tree
	 * @param clonePolicy
	 *          the clone policy to execute at each leaf
	 * 
	 * @return the new cloned tree
	 * 
	 */
	public Tree<T> clone(ClonePolicy<T> clonePolicy, Context testsList) {
		Tree<T> result = newClonedTree();
		if (root != null) {
			result.setRoot(root.clone(clonePolicy, result, null, null, testsList));
		}
		return result;
	}

	/**
	 * 
	 * @return a new tree used to clone this
	 * 
	 */
	protected Tree<T> newClonedTree() {
		return new Tree<T>();
	}

	/**
	 * @see java.lang.Object#clone()
	 */
	@Override
	public Tree<T> clone() {
		return clone(null);
	}

	/**
	 * 
	 * Clone a part of this tree
	 * 
	 * @param testsList
	 *          the constraint to apply on the clone of this tree
	 * 
	 * @return a new tree
	 * 
	 */
	public Tree<T> clone(Context testsList) {
		return clone(null, testsList);
	}

	/**
	 * @see animatlab.kodamat.core.functions.IFunction#classify(Context)
	 */
	public T classify(Context state) {
		Leaf<T> leaf = getLeaf(state);
		if (leaf == null) {
			return null;
		}
		return leaf.get();
	}

	/**
	 * 
	 * Get the leaf corresponding to a vector of values
	 * 
	 * @param values
	 *          the vector of values to classify in the decision tree
	 * 
	 * @return the vertex corresponding to the class of the vector of values
	 * 
	 */
	final public Leaf<T> getLeaf(Context values) {
		Vertex<T> vertex = getRoot();
		while (!vertex.isLeaf) {
			Vertex<T> child = ((DNode<T>) vertex).getChild(values);
			if (child == null) {
				return null;
			}
			vertex = child;
		}
		return (Leaf<T>) vertex;
	}

	/**
	 * 
	 * Replace a vertex by a tree
	 * 
	 * @param vertex
	 *          the vertex to replace
	 * @param subTree
	 *          the tree used to replace this node
	 * 
	 */
	final public void replace(Vertex<T> vertex, Tree<T> subTree) {
		Vertex<T> subTreeRoot = subTree.getRoot();
		subTreeRoot.parent = vertex.parent;
		subTreeRoot.value = vertex.value;
		vertex.remove();
		subTreeRoot.changeTree(this);
		if (subTreeRoot.parent != null) {
			subTreeRoot.parent.registerAsChild(subTreeRoot.value, subTreeRoot);
		} else {
			setRoot(subTreeRoot);
		}
	}

	/**
	 * 
	 * Simplify this tree by removing vertex that does not correspond to a set of
	 * constraint
	 * 
	 * @param testsList
	 *          the set of constraint of this tree
	 * 
	 */
	final public void cut(Context testsList) {
		getRoot().cut(testsList);
	}

	/**
	 * 
	 * Simplify this tree using a policy to compare the data at the leaves
	 * 
	 * @param equalPolicy
	 *          the policy to compare two data at the leaves
	 * 
	 */
	public void simplify(EqualPolicy<T> equalPolicy) {
		getRoot().simplify(equalPolicy, new MergeLeavesPolicy<T>() {
			public Vertex<T> merge(Collection<Vertex<T>> vertexSet) {
				return vertexSet.iterator().next();
			}
		});
	}

	/**
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@SuppressWarnings("unchecked")
	@Override
	public boolean equals(Object object) {
		if (object == null) {
			return false;
		}
		return equals((Tree<T>) object);
	}

	/**
	 * 
	 * Test the equality between this tree and another tree
	 * 
	 * @param tree
	 *          the other tree to compare
	 * 
	 * @return true if the two trees are equal
	 * 
	 */
	public boolean equals(Tree<T> tree) {
		return equals(tree, null);
	}

	/**
	 * 
	 * Test the equality between this tree and another tree using a policy
	 * 
	 * @param tree
	 *          the other tree to compare
	 * @param equalPolicy
	 *          the policy to compare two data at the leaves
	 * 
	 * @return true if the two trees are equal
	 * 
	 */
	public boolean equals(Tree<T> tree, EqualPolicy<T> equalPolicy) {
		if (tree == null) {
			return false;
		}
		if (tree == this) {
			return true;
		}
		return getRoot().equals(tree.getRoot(), equalPolicy);
	}

	/**
	 * 
	 * Compare this tree to another tree
	 * 
	 * @param idtTree
	 *          the tree to compare with this
	 * @param policy
	 *          the policy to used to compare at each node
	 * 
	 */
	public void compare(Tree<T> idtTree, ComparePolicy<T> policy) {
		for (IPartition<T> idtNode : leafSet) {
			Tree<T> idtSubTree = idtTree.clone(idtNode.getContext());
			for (IPartition<T> idtNodeSubTree : idtSubTree.leafSet) {
				policy.compare((Leaf<T>) idtNode, (Leaf<T>) idtNodeSubTree);
			}
		}
	}

	/**
	 * @see animatlab.kodamat.core.functions.IPartitionFunction#partitions()
	 */
	final public Set<IPartition<T>> partitions() {
		if (leafSet.isEmpty()) {
			getRoot();
		}
		return leafSet;
	}

	/**
	 * @see animatlab.kodamat.core.io.IToXml#toXml(org.w3c.dom.Document,
	 *      animatlab.kodamat.core.io.IToXml.IToXmlPolicy)
	 */
	public Element toXml(Document document, IToXmlPolicy<T> policy) {
		XmlUtils.declareNS(document, NAMESPACE);
		Element result = getRoot().toXml(document, policy);
		result.setAttribute("nbleaf", String.valueOf(numberOfLeaves()));
		result.setAttribute("nbnode", String.valueOf(numberOfNodes()));
		return result;
	}

	/**
	 * 
	 * Set leafListener
	 * 
	 * @param leafListener
	 *          The leafListener to set.
	 * 
	 */
	public final void setLeafListener(LeafListener leafListener) {
		this.leafListener = leafListener;
	}

	/**
	 * 
	 * @return a clone of the set of leaves in the tree
	 * 
	 */
	public Set<Leaf<T>> getClonedLeafSet() {
		Set<Leaf<T>> result = new LinkedHashSet<Leaf<T>>();
		for (IPartition<T> partition : partitions()) {
			result.add((Leaf<T>) partition);
		}
		return result;
	}

	/**
	 * 
	 * Append a tree to this tree
	 * 
	 * @param addedTree
	 *          the tree to append
	 * @param appendPolicy
	 *          the policy to use to append the other tree
	 * @param clonePolicy
	 *          policy to use to clone part of the tree
	 * 
	 */
	final public void append(Tree<T> addedTree, AppendPolicy<T> appendPolicy,
			ClonePolicy<T> clonePolicy) {
		for (Leaf<T> leaf : getClonedLeafSet()) {
			append(leaf, addedTree, appendPolicy, clonePolicy);
		}
	}

	/**
	 * 
	 * Append a tree to a leaf of this tree
	 * 
	 * @param leaf
	 *          the leaf where to append the tree
	 * @param addedTree
	 *          the tree to append to the leaf
	 * @param appendPolicy
	 *          the policy to use to append the other tree
	 * @param clonePolicy
	 *          policy to use to clone part of the tree
	 * 
	 */
	public void append(Leaf<T> leaf, Tree<T> addedTree,
			AppendPolicy<T> appendPolicy, ClonePolicy<T> clonePolicy) {
		assert (leaf.getTree() == this);
		Tree<T> clonedTree = addedTree.clone(clonePolicy, leaf.getContext());
		if (appendPolicy != null) {
			for (IPartition<T> appendedLeaf : clonedTree.partitions()) {
				appendPolicy.append(leaf, (Leaf<T>) appendedLeaf);
			}
		}
		replace(leaf, clonedTree);
	}

	/**
	 * @see animatlab.kodamat.core.io.IToDot#toDot(StringBuilder,
	 *      animatlab.kodamat.core.io.IToDot.IToDotPolicy, ImageMap)
	 */
	@SuppressWarnings("unchecked")
	public void toDot(StringBuilder output, IToDotPolicy<T> policy,
			ImageMap<T> imageMap) {
		output.append("digraph G {\n");
		output.append("ordering=out\n");
		for (IPartition<T> leaf : leafSet) {
			((IToDot<T>) leaf).toDot(output, policy, imageMap);
		}
		getRoot().toDot(output, policy, imageMap);
		output.append("}\n");
	}

	/**
	 * @see animatlab.kodamat.core.io.IToDot#dotName()
	 */
	public String dotName() {
		return "";
	}

	/**
	 * 
	 * @return the scope of this tree
	 * 
	 */
	public Set<Variable> scope() {
		return scopeMap.keySet();
	}

	/**
	 * 
	 * Get the vertex associated to a context
	 * 
	 * @param context
	 *          the context to get the vertex for
	 * @return the vertex associated to this context
	 * 
	 */
	public Vertex<T> getVertex(Context context) {
		Vertex<T> currentVertex = getRoot();
		while (!currentVertex.isLeaf) {
			DNode<T> dnode = (DNode<T>) currentVertex;
			currentVertex = dnode.getChild(context.get(dnode.getTestedVariable()));
			if (currentVertex == null) {
				return dnode;
			}
		}
		return currentVertex;
	}

	/**
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		return toString().hashCode();
	}

	/**
	 * 
	 * @return returns the description associated to this tree
	 * 
	 */
	public String getDescription() {
		return description;
	}
}
