/*
 * 
 * Created on 22 dec. 2005
 *
 */
package animatlab.kodamat.core.trees;

import java.util.Map;

import animatlab.kodamat.core.io.IToDot;
import animatlab.kodamat.core.io.IToXml;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;

/**
 * 
 * A node (decision node or leaf) in the decision tree
 * 
 * @author Thomas Degris
 * @param <T>
 *          The type of instances contained in the tree
 * 
 */
abstract public class Vertex<T> implements IToXml<T>, IToDot<T> {
	/**
	 * 
	 * Attribute in a xml node representing the value to reach this node
	 * 
	 */
	public static final String ATTRIBUTE_VALUE = "value";

	/**
	 * 
	 * The tree owning this node
	 * 
	 */
	protected Tree<T> tree;

	/**
	 * 
	 * Parent of this node
	 * 
	 */
	protected DNode<T> parent;

	/**
	 * 
	 * Value of the tested variable (of the parent node) to reach this node
	 * 
	 */
	protected String value;

	/**
	 * 
	 * Tests list to reach this node from the root of the tree
	 * 
	 */
	protected Context testsList = null;

	/**
	 * 
	 * True if this vertex is a leaf
	 * 
	 */
	final public boolean isLeaf;

	/**
	 * 
	 * Constructor
	 * 
	 * @param tree
	 *          The tree owning this node
	 * @param parent
	 *          Parent in the tree of this node
	 * @param value
	 *          value of the tested variable
	 * @param isLeaf
	 *          true if this vertex is a leaf
	 * 
	 */
	public Vertex(Tree<T> tree, DNode<T> parent, String value, boolean isLeaf) {
		this.tree = tree;
		this.parent = parent;
		this.value = value;
		this.isLeaf = isLeaf;
		if (parent != null) {
			parent.registerAsChild(value, this);
		}
	}

	/**
	 * 
	 * Remove this node from the tree
	 * 
	 */
	public void remove() {
		if (parent != null) {
			parent.removeChild(value);
		}
		tree = null;
		parent = null;
		value = null;
	}

	/**
	 * 
	 * Replace this node in the tree by an decision node
	 * 
	 * @param testedVariable
	 *          the variable tested in the decision node
	 * @return the newly created decision node
	 * 
	 */
	abstract public DNode<T> replace(Variable testedVariable);

	/**
	 * 
	 * Build a string representing this node and its children
	 * 
	 * @param indentation
	 *          recursive parameter indicating the indentation for this node in
	 *          the string
	 * @return a string representation of this node and its children
	 * 
	 */
	public String toString(int indentation) {
		StringBuffer result = new StringBuffer();
		for (int indentationIndex = 0; indentationIndex < indentation; indentationIndex++) {
			result.append(" ");
		}
		if (parent != null) {
			result.append("|=\"");
			result.append(value);
			result.append("\")");
		}
		return result.toString();
		// return result.toString() + getTestsList().toString();
	}

	/**
	 * 
	 * Test if this node corresponds to a particular state (used mainly for
	 * debugging)
	 * 
	 * @param constraint
	 *          the constraint to test
	 * @return true if this node corresponds
	 * 
	 */
	final public boolean satisfy(Context constraint) {
		Context tests = getContext();
		for (Map.Entry<Variable, String> entry : tests.entrySet()) {
			String value = constraint.get(entry.getKey());
			if ((value != null) && (!value.equals(entry.getValue()))) {
				return false;
			}
		}
		return true;
	}

	/**
	 * 
	 * @return the tested value to reach this node
	 * 
	 */
	public String getTestedValue() {
		return value;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return toString(0);
	}

	/**
	 * 
	 * Clone this vertex
	 * 
	 * @param clonePolicy
	 *          policy to use to clone data at the leaves
	 * @param result
	 *          the tree containing the new vertex
	 * @param parent
	 *          the parent of the clone
	 * @param value
	 *          value to reach the cloned node
	 * @param testsList
	 *          the tests list when only a part of the tree is extracted (may be
	 *          null when the full tree is cloned)
	 * @return a new clone of this vertex
	 * 
	 */
	protected abstract Vertex<T> clone(ClonePolicy<T> clonePolicy,
			Tree<T> result, DNode<T> parent, String value, Context testsList);

	/**
	 * 
	 * @return a list of test to reach this node
	 * 
	 */
	final public Context getContext() {
		if (testsList == null) {
			testsList = new Context();
		}
		if ((parent != null) && (testsList.size() == 0)) {
			testsList.putAll(parent.getContext());
			testsList.put(parent.testedVariable, value);
		}
		return testsList;
	}

	/**
	 * 
	 * Change the tree belonging this node
	 * 
	 * @param tree
	 *          the new tree owning this vertex
	 * 
	 */
	protected void changeTree(Tree<T> tree) {
		if (testsList != null) {
			testsList.clear();
		}
		this.tree.remove(this);
		this.tree = tree;
		tree.add(this);
	}

	/**
	 * 
	 * Clear the tests list of this vertex and its children
	 * 
	 */
	protected void clearTestsList() {
		if (testsList != null) {
			testsList.clear();
		}
	}

	/**
	 * 
	 * Extract from this tree a subtree taken this node as the root of the new
	 * subtree
	 * 
	 * @return a new subtree
	 * 
	 */
	final public Tree<T> extract() {
		Tree<T> result = tree.newInstance();
		assert (result.root == null);
		result.root = this;
		changeTree(result);
		parent.unregisterAsChild(value);
		parent = null;
		value = null;
		return result;
	}

	/**
	 * 
	 * Simplify this tree by removing vertex that does not correspond to a set of
	 * constraint
	 * 
	 * @param testsList
	 *          the set of constraint of this tree
	 * 
	 */
	abstract public void cut(Context testsList);

	/**
	 * 
	 * Simplify this vertex and its children
	 * 
	 * @param equalPolicy
	 *          the policy to compare two data at the leaves
	 * @param mergePolicy
	 *          policy to use to merge nodes
	 * 
	 */
	abstract public void simplify(EqualPolicy<T> equalPolicy,
			MergeLeavesPolicy<T> mergePolicy);

	/**
	 * 
	 * Test if this vertex is equal to another vertex
	 * 
	 * @param vertex
	 *          the vrtex to compare with
	 * @param equalPolicy
	 *          the policy to compare two data at the leaves
	 * 
	 * @return true if they are equal
	 * 
	 */
	public boolean equals(Vertex<T> vertex, EqualPolicy<T> equalPolicy) {
		if (vertex == null) {
			return false;
		}
		if (vertex == this) {
			return true;
		}
		if (isLeaf != vertex.isLeaf) {
			return false;
		}
		return true;
	}

	/**
	 * 
	 * @return the tree owning this vertex
	 * 
	 */
	public Tree<T> getTree() {
		return tree;
	}

	/**
	 * @see animatlab.kodamat.core.io.IToDot#dotName()
	 */
	public String dotName() {
		return getClass().getSimpleName() + super.hashCode();
	}

	/**
	 * 
	 * @return the parent of this vertex
	 * 
	 */
	public DNode<T> getParent() {
		return parent;
	}

}
