/*
 * 
 * Created on 2 janv. 2006
 *
 */
package animatlab.kodamat.core.utils;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import animatlab.kodamat.core.IModel;
import animatlab.kodamat.core.proba.IProbabilities;

/**
 * 
 * Some utils function to manipulate states
 * 
 * @author Thomas Degris
 * 
 */
public class Context extends LinkedHashMap<Variable, String> {
	/**
	 * 
	 * Serial version UID
	 * 
	 */
	private static final long serialVersionUID = -7484090589995722427L;

	/**
	 * 
	 * Name of the node describing the set of values for all variables
	 * 
	 */
	public static final String NODE_NAME_VALUESET = "dtr:variablevalueset";

	/**
	 * 
	 * Value set used when generating a complete set of set
	 * 
	 */
	public static enum ValueSet {
		/**
		 * 
		 * Binary set
		 * 
		 */
		BINARY,

		/**
		 * 
		 * Octal set
		 * 
		 */
		OCTAL
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param state
	 *          the state to create a clone for
	 * 
	 */
	public Context(Context state) {
		super(state);
	}

	/**
	 * 
	 * Constructor
	 * 
	 * 
	 */
	public Context() {
		super();
	}

	/**
	 * 
	 * Create a set of all possible states for a state size
	 * 
	 * @param stateSize
	 *          the size of the state to create
	 * @param valueSet
	 *          the value set of the variable
	 * @return an array of all the possible states
	 * 
	 */
	static public Context[] createStateSet(int stateSize, ValueSet valueSet) {
		int numberOfStates = 0;
		switch (valueSet) {
		case BINARY:
			numberOfStates = (int) Math.pow(2, stateSize);
			break;
		case OCTAL:
			numberOfStates = (int) Math.pow(8, stateSize);
			break;
		}
		assert (numberOfStates != 0);
		Context[] states = new Context[numberOfStates];
		for (int index = 0; index < numberOfStates; index++) {
			String state = null;
			switch (valueSet) {
			case BINARY:
				state = Integer.toBinaryString(index);
				break;
			case OCTAL:
				state = Integer.toOctalString(index);
				break;
			default:
				return null;
			}
			StringBuilder buffer = new StringBuilder();
			for (int currentLength = state.length(); currentLength < stateSize; currentLength++) {
				buffer.append("0");
			}
			buffer.append(state);
			for (int charIndex = buffer.length() - 1; charIndex != 0; charIndex--) {
				buffer.insert(charIndex, ",");
			}
			states[index] = getState(buffer.toString(), null);
		}
		return states;

	}

	/**
	 * 
	 * Create a reward list from a string
	 * 
	 * @param state
	 *          the string describing the reward list
	 * @return a list of reward
	 * 
	 */
	public static List<Double> getRewards(String state) {
		ArrayList<Double> result = new ArrayList<Double>();
		String[] stringArrayRewards = state.split("\\,");
		for (String rewardStr : stringArrayRewards) {
			result.add(Double.parseDouble(rewardStr));
		}
		return result;
	}

	/**
	 * 
	 * Create a state from a string
	 * 
	 * @param state
	 *          the string describing the state
	 * @param variableOrdering
	 *          variable ordering to use to parse the string
	 * @return the new state
	 * 
	 */
	public static Context getState(String state, VariableOrdering variableOrdering) {
		Context result = new Context();
		String[] stringArrayState = state.split("\\,");
		for (int index = 0; index < stringArrayState.length; index++) {
			String value = stringArrayState[index].replaceAll(" ", "");
			int equalIndex = value.indexOf("=");
			Variable variable;
			if (equalIndex != -1) {
				variable = Variable.getVariable(value.substring(0, equalIndex));
				value = value.substring(equalIndex + 1, value.length());
			} else {
				if (variableOrdering == null) {
					variable = Variable.getVariable(String.valueOf(index));
				} else {
					variable = variableOrdering.get(index);
				}
			}
			result.put(variable, value);
		}
		return result;
	}

	/**
	 * 
	 * Load the value set from a xml tree
	 * 
	 * @param variableValueSet
	 *          where to put the result
	 * @param root
	 *          the root of the xml tree to load
	 * 
	 * @return the map where the result has been put
	 * 
	 */
	static public Map<Variable, Set<String>> loadValueSet(
			Map<Variable, Set<String>> variableValueSet, Element root) {
		assert (root.getNodeName().equals(NODE_NAME_VALUESET));
		NodeList nodeList = root.getChildNodes();
		for (int index = 0; index < nodeList.getLength(); index++) {
			Node childNode = nodeList.item(index);
			if (childNode.getNodeType() == Node.ELEMENT_NODE) {
				if (childNode.getNodeName().equals("dtr:valueset")) {
					Element valueSetElement = (Element) childNode;
					Variable variable = Variable.load(valueSetElement);
					variableValueSet.put(variable, loadValueSet(valueSetElement));
				}
			}
		}
		return variableValueSet;
	}

	/**
	 * 
	 * Create a set of value from a xml tree
	 * 
	 * @param root
	 *          the root of the xml tree to load
	 * 
	 * @return the new created set
	 * 
	 */
	private static Set<String> loadValueSet(Element root) {
		Set<String> result = new LinkedHashSet<String>();
		NodeList nodeList = root.getChildNodes();
		for (int index = 0; index < nodeList.getLength(); index++) {
			Node childNode = nodeList.item(index);
			if (childNode.getNodeType() == Node.ELEMENT_NODE) {
				if (childNode.getNodeName().equals("dtr:value")) {
					result.add(((Element) childNode).getAttribute("value"));
				}
			}
		}
		return result;
	}

	/**
	 * Pretty print a context into a string
	 * 
	 * @param context
	 *          the context to print
	 * @return the string representing the context of this rule
	 * 
	 */
	public static String contextToString(Context context) {
		StringBuilder result = new StringBuilder();
		final String operator = " & ";
		List<Variable> variables = new ArrayList<Variable>(context.keySet());
		Collections.sort(variables);
		for (Variable variable : variables) {
			result.append(variable.toString());
			result.append("=");
			result.append(context.get(variable));
			result.append(operator);
		}
		if (context.size() > 0) {
			result.delete(result.length() - operator.length(), result.length());
		}
		return result.toString();
	}

	/**
	 * 
	 * Create an array of all the possible given a particular model
	 * 
	 * @param model
	 *          the model to build all the possible states for
	 * 
	 * @return an array of all the possible states
	 * 
	 */
	public static Context[] createStateSet(IModel model) {
		return createStateSet(new ArrayList<Variable>(model.getVariableSet()),
				model.getValueSets());
	}

	/**
	 * Create recursively an array of all the possible given a particular model
	 * 
	 * @param variables
	 *          the set of variables to which build states for
	 * @param variablesValues
	 *          values for each variable
	 * 
	 * @return an array of all the possible states
	 * 
	 */
	private static Context[] createStateSet(List<Variable> variables,
			Map<Variable, Set<String>> variablesValues) {
		if (variables.size() == 0) {
			Context[] result = new Context[1];
			result[0] = new Context();
			return result;
		}

		Variable variable = variables.remove(variables.size() - 1);
		Context[] recursion = createStateSet(variables, variablesValues);

		Context[] result = new Context[variablesValues.get(variable).size()
				* recursion.length];
		int arrayIndex = 0;
		for (Context recursionState : recursion) {
			for (String value : variablesValues.get(variable)) {
				result[arrayIndex] = new Context(recursionState);
				result[arrayIndex].put(variable, value);
				arrayIndex++;
			}
		}
		return result;
	}

	/**
	 * 
	 * Get all the possible context for a set of variables
	 * 
	 * @param variables
	 *          the set of variables to get the context for
	 * @param valueSets
	 *          the set of values for each variables
	 * @return the set of all possible context
	 * 
	 */
	public static Set<Context> getAllContext(Set<Variable> variables,
			Map<Variable, Set<String>> valueSets) {
		Set<Context> setK = new LinkedHashSet<Context>();
		setK.add(new Context());
		for (Variable variable : variables) {
			Set<Context> setKp1 = new LinkedHashSet<Context>();
			for (Context context : setK) {
				for (String values : valueSets.get(variable)) {
					Context contextP1 = new Context(context);
					contextP1.put(variable, values);
					setKp1.add(contextP1);
				}
			}
			setK = setKp1;
		}
		return setK;
	}

	/**
	 * 
	 * Build a string representing a set of probilities of a SpiSviInfo node
	 * 
	 * @param info
	 *          the node containing the probabilities to represent
	 * @return a string representing the probabilities
	 * 
	 */
	public static String probabilitiesToString(
			Map<Variable, IProbabilities<String>> info) {
		StringBuilder result = new StringBuilder("{");
		final String operator = ", ";
		List<Variable> scope = new ArrayList<Variable>(info.keySet());
		Collections.sort(scope);
		for (Variable variable : scope) {
			IProbabilities<String> value = info.get(variable);
			result.append(variable.toString());
			result.append("=");
			result.append(value.toString());
			result.append(operator);
		}
		if (info.size() > 0) {
			result.delete(result.length() - operator.length(), result.length());
		}

		result.append("}");
		return result.toString();
	}

	/**
	 * 
	 * Extract a list from a string passed from Kodabot
	 * 
	 * @param listAsString
	 *          the string passed from Kodabot
	 * 
	 * @return the list build from the string
	 * 
	 */
	public static List<String> extractList(String listAsString) {
		if (listAsString == null) {
			return null;
		}
		if (listAsString.length() == 0) {
			return null;
		}
		String[] elementArray = listAsString.split(",\\s*");
		List<String> result = new ArrayList<String>();
		for (String element : elementArray) {
			result.add(element);
		}
		return result;
	}

	/**
	 * 
	 * Extract a list from a string passed from Kodabot
	 * 
	 * @param listAsString
	 *          the string passed from Kodabot
	 * 
	 * @return the list build from the string
	 * 
	 */
	public static List<Double> extractRewardList(String listAsString) {
		if (listAsString == null) {
			return null;
		}
		if (listAsString.length() == 0) {
			return null;
		}
		String[] actionArray = listAsString.split(",");
		List<Double> result = new ArrayList<Double>();
		for (String action : actionArray) {
			result.add(Double.parseDouble(action));
		}
		return result;
	}

	/**
	 * 
	 * Sum a list of double
	 * 
	 * @param rewards
	 *          the list of double to sum
	 * 
	 * @return the result of the sum
	 * 
	 */
	public static double getSum(List<Double> rewards) {
		if (rewards == null) {
			return Double.NaN;
		}
		assert (rewards.size() > 0);
		double result = 0;
		for (Double reward : rewards) {
			if (reward == null) {
				return Double.NaN;
			}
			result = result + reward;
		}
		return result;
	}

	/**
	 * 
	 * Convert a single reward into a list of rewards
	 * 
	 * @param reward
	 *          the reward to convert
	 * 
	 * @return the list containing only one element
	 * 
	 */
	public static List<Double> rewardToList(double reward) {
		if (Double.isNaN(reward)) {
			return null;
		}
		List<Double> rewards = new ArrayList<Double>();
		rewards.add(reward);
		return rewards;
	}

	/**
	 * 
	 * Output the reward as a string
	 * 
	 * @param rewardList
	 *          the reward to represent
	 * 
	 * @return the string representing the reward
	 * 
	 */
	public static String rewardToString(List<Double> rewardList) {
		if (rewardList == null) {
			return "NaN";
		}
		StringBuilder result = new StringBuilder();
		if (rewardList.size() > 1) {
			for (Double currentReward : rewardList) {
				result.append(currentReward + "+");
			}
			result.delete(result.length() - 1, result.length());
			result.append("=");
		}
		result.append(Context.getSum(rewardList));
		return result.toString();
	}

	/**
	 * 
	 * Convert a context into a list of value
	 * 
	 * @param variableOrdering
	 *          legend to get variable ordering
	 * @param context
	 *          the context to convert
	 * @return a new list with the values from the context
	 * 
	 */
	public static List<String> contextToList(VariableOrdering variableOrdering,
			Context context) {
		List<String> result = new ArrayList<String>();
		for (Map.Entry<Variable, String> entry : context.entrySet()) {
			int variableIndex = variableOrdering.indexOf(entry.getKey());
			while (result.size() <= variableIndex) {
				result.add(null);
			}
			result.set(variableIndex, entry.getValue());
		}
		return result;
	}

	/**
	 * 
	 * Check if a state satisfy a context
	 * 
	 * @param variableOrdering
	 *          legend to get variable ordering
	 * @param context
	 *          the context to check with
	 * @param example
	 *          the state to check
	 * 
	 * @return true if the context is satisfied
	 * 
	 */
	static public boolean satisfy(VariableOrdering variableOrdering,
			Context context, List<String> example) {
		for (Map.Entry<Variable, String> entry : context.entrySet()) {
			int variableIndex = variableOrdering.indexOf(entry.getKey());
			if ((example.size() > variableIndex)
					&& (!entry.getValue().equals(example.get(variableIndex)))) {
				return false;
			}
		}
		return true;
	}

	/**
	 * 
	 * Test if a context satisfy another context
	 * 
	 * @param referenceContext
	 *          context used as reference
	 * @param testedContext
	 *          the context used to test with
	 * 
	 * @return true if the context is satisfied
	 * 
	 */
	static public boolean satisfy(Context referenceContext, Context testedContext) {
		for (Map.Entry<Variable, String> entry : referenceContext.entrySet()) {
			String requiredValue = entry.getValue();
			String value = testedContext.get(entry.getKey());
			if (!requiredValue.equals(value)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * 
	 * Get the number of state in a restricted sub-space of the state space of a
	 * forest
	 * 
	 * @param variableValueSets
	 *          set of values for each variables
	 * @param assignedVariableSet
	 *          set of variables assigned delimiting the subspace
	 * 
	 * @return the number of states in the subspace restricted by assigned
	 *         variables
	 * 
	 */
	static public long getRestrictedStateSize(
			Map<Variable, Set<String>> variableValueSets,
			Set<Variable> assignedVariableSet) {
		long restrictedStateSize = 1;
		for (Map.Entry<Variable, Set<String>> entry : variableValueSets.entrySet()) {
			Variable variable = entry.getKey();
			if ((assignedVariableSet == null)
					|| (!assignedVariableSet.contains(variable))) {
				restrictedStateSize = restrictedStateSize * entry.getValue().size();
			}
		}
		return restrictedStateSize;
	}

	/**
	 * 
	 * Build a context from a set of proba (take the best proba for each variable)
	 * 
	 * @param probas
	 *          the proba from which to build the context
	 * 
	 * @return the context build from the set of proba
	 * 
	 */
	public static Context getContext(Map<Variable, IProbabilities<String>> probas) {
		Context result = new Context();
		for (Map.Entry<Variable, IProbabilities<String>> entry : probas.entrySet()) {
			result.put(entry.getKey(), entry.getValue().getBest());
		}
		return result;
	}
}
