/*
 * 
 * Created on 8 janv. 2006
 *
 */
package animatlab.kodamat.core.utils;

import java.text.NumberFormat;
import java.util.LinkedHashMap;
import java.util.Map;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import animatlab.kodamat.core.io.IToXml;
import animatlab.kodamat.core.io.XmlUtils;

/**
 * 
 * Class defining parameters
 * 
 * @author Thomas Degris
 * 
 */
public class Parameters implements IToXml<Object> {
	/**
	 * 
	 * Name of the key in the parameters to set the treshold used for the
	 * convergence
	 * 
	 */
	public static final String PARAM_KEY_THRESHOLD_CONVERGENCE = "epsilonThreshold";

	/**
	 * 
	 * Name of the key in the parameters to set the treshold used for ordering the
	 * value function
	 * 
	 */
	public static final String PARAM_KEY_ORDERING_THRESHOLD = "orderingThreshold";

	/**
	 * 
	 * Name of the key in the parameters to enable/disable variable ordering
	 * 
	 */
	public static final String PARAM_KEY_ORDERING = "ordering";

	/**
	 * 
	 * Name of the key in the parameters to enable/disable variable ordering
	 * 
	 */
	public static final String PARAM_KEY_AUTO_AGENT = "autoAgent";

	/**
	 * 
	 * Name of the key in the parameters to set the gamma value
	 * 
	 */
	public static final String PARAM_KEY_GAMMA = "gamma";

	/**
	 * 
	 * Name of the key in the parameters to enable/disable variable ordering
	 * 
	 */
	public static final String PARAM_KEY_EPISODE_TIME = "episodeTime";

	/**
	 * 
	 * Name of the key in the parameters to enable/disable the reset of the value
	 * function when the reward changes
	 * 
	 */
	public static final String PARAM_KEY_RESET_REWARD = "resetReward";

	/**
	 * 
	 * Name of the key in the parameters to enable/disable the reset of the value
	 * function when the reward changes
	 * 
	 */
	public static final String PARAM_KEY_TRANSITION_LEARNING_THRESHOLD = "transitionLearningThreshold";

	/**
	 * 
	 * Name of the xml node to represent a legend for decision trees
	 * 
	 */
	public static final String NAMESPACE = "dtr";

	/**
	 * 
	 * Name of the xml node to represent a set of keys
	 * 
	 */
	public static final String NODE_NAME_KEYS = NAMESPACE + ":" + "keys";

	/**
	 * 
	 * Name of the xml node to represent a key
	 * 
	 */
	public static final String NODE_NAME_KEY = NAMESPACE + ":" + "key";

	/**
	 * 
	 * Name of the attribute to named a key
	 * 
	 */
	static private final String ATTRIBUTE_NAME_ID = "id";

	/**
	 * 
	 * Name of the attribute to named a value of a key
	 * 
	 */
	static private final String ATTRIBUTE_NAME_VALUE = "value";

	/**
	 * 
	 * Epsilon used to compare two values
	 * 
	 */
	public static final double EPSILON = 1E-8;

	/**
	 * 
	 * Bonus to give to RMax when the reward maximum is not known
	 * 
	 */
	public static final double REWARD_MAX_BONUS = 0.1;

	/**
	 * 
	 * Threshold to use to compare two value functions and test if they have
	 * converged
	 * 
	 */
	static public double EPSILON_THRESHOLD_CONVERGENCE = 0.1;

	/**
	 * 
	 * Discount factor in the algorithm
	 * 
	 */
	static final public double GAMMA = 0.99;

	/**
	 * 
	 * Discount factor in the algorithm
	 * 
	 */
	static final public int DEFAULT_EPISODE_LENGTH = 15;

	/**
	 * 
	 * Output folder of the tests
	 * 
	 */
	static final public String OUTPUT_JUNIT_FOLDER = "./.reports";

	/**
	 * The map containing the pairs key/value
	 */
	final private Map<String, String> keys = new LinkedHashMap<String, String>();

	/**
	 * 
	 * Name of the planning method key
	 * 
	 */
	static public final String PARAM_KEY_PLANNING_METHOD = "planningMethod";

	/**
	 * 
	 * Name of the planning method key
	 * 
	 */
	static public final String PARAM_KEY_RUNNING_AGENT = "preferedRunningAgent";

	/**
	 * Instance used to format value number
	 */
	static final private NumberFormat numberFormat = NumberFormat.getInstance();

	/**
	 * 
	 * Name of the "decompose basis function" key
	 * 
	 */
	public static final String PARAM_KEY_DECOMPOSE_BASISFUNCTION = "decbf";

	/**
	 * 
	 * Key to set the scope of the basis function when they are generated
	 * 
	 */
	public static final String PARAM_KEY_BASIS_FUNCTION_SCOPE_SIZE = "bfscopesize";

	/**
	 * 
	 * Key to define the viewer of a FMDP
	 * 
	 */
	public static final String PARAM_KEY_VIEWER = "viewer";

	static {
		Parameters.numberFormat.setMinimumFractionDigits(2);
		Parameters.numberFormat.setMaximumFractionDigits(2);
	}

	/**
	 * Get a VMax value from a reward
	 * 
	 * @param reward
	 *          the reward to get the VMax value from (usually it is the best
	 *          reward received)
	 * 
	 * @return the corresponding VMax value to the reward
	 * 
	 */
	static public double getVMaxFromReward(double reward) {
		return (reward * (1.0 / (1.0 - GAMMA)));
	}

	/**
	 * Get a VMax value from a reward
	 * 
	 * @param reward
	 *          the reward to get the VMax value from (usually it is the best
	 *          reward received)
	 * 
	 * @return the corresponding VMax value to the reward
	 * 
	 */
	static public double getRMaxFromReward(double reward) {
		return (reward + REWARD_MAX_BONUS);
	}

	/**
	 * 
	 * Format a number
	 * 
	 * @param number
	 *          the number to represent
	 * 
	 * @return a string representing the number
	 * 
	 */
	static public String format(Double number) {
		if (number == null) {
			return "null";
		}
		if (Double.isNaN(number)) {
			return "nan";
		}
		return numberFormat.format(Math.round(number * 100.0) / 100.0);
	}

	/**
	 * 
	 * Constructor
	 * 
	 */
	public Parameters() {
		super();
	}

	/**
	 * 
	 * Load the parameters from a xml tree
	 * 
	 * @param xmlNode
	 *          the root of a set of key
	 * 
	 */
	public void load(Element xmlNode) {
		assert (xmlNode.getNodeName().equals(NODE_NAME_KEYS));
		NodeList nodeList = xmlNode.getChildNodes();
		for (int index = 0; index < nodeList.getLength(); index++) {
			Node variableNode = nodeList.item(index);
			if (variableNode.getNodeType() == Node.ELEMENT_NODE) {
				Element keyElement = (Element) variableNode;
				assert (keyElement.getNodeName().equals(NODE_NAME_KEY));
				String id = keyElement.getAttribute(ATTRIBUTE_NAME_ID);
				String value = keyElement.getAttribute(ATTRIBUTE_NAME_VALUE);
				setKeyValue(id, value);
			}
		}
	}

	/**
	 * 
	 * Get the value of a key
	 * 
	 * @param id
	 *          the id of the key
	 * @return the value of the key with the id 'id'
	 * 
	 */
	public int getKeyI(String id) {
		return Integer.parseInt(getKeyS(id));
	}

	/**
	 * 
	 * Get the value of a key
	 * 
	 * @param id
	 *          the id of the key
	 * @return the value of the key with the id 'id'
	 * 
	 */
	public boolean getKeyB(String id) {
		return Boolean.parseBoolean(getKeyS(id));
	}

	/**
	 * 
	 * Get the value of a key
	 * 
	 * @param id
	 *          the id of the key
	 * @return the value of the key with the id 'id'
	 * 
	 */
	public String getKeyS(String id) {
		return keys.get(id);
	}

	/**
	 * 
	 * Get the value of a key
	 * 
	 * @param id
	 *          the id of the key
	 * @return the value of the key with the id 'id'
	 * 
	 */
	public Double getKeyD(String id) {
		String value = getKeyS(id);
		if ("null".equals(value)) {
			return null;
		}
		return Double.parseDouble(value);
	}

	/**
	 * 
	 * Check if a key has a value
	 * 
	 * @param id
	 *          the id of the key
	 * @return true if the key has a value
	 * 
	 */
	public boolean hasKey(String id) {
		return (getKeyS(id) != null);
	}

	/**
	 * 
	 * Set a value for a key
	 * 
	 * @param id
	 *          the id of the key to set
	 * @param value
	 *          the value associated to the key
	 * 
	 */
	public void setKeyValue(String id, String value) {
		keys.put(id, value);
	}

	/**
	 * @see animatlab.kodamat.core.io.IToXml#toXml(org.w3c.dom.Document,
	 *      animatlab.kodamat.core.io.IToXml.IToXmlPolicy)
	 */
	public Element toXml(Document document, IToXmlPolicy<Object> policy) {
		Element legendElement = document.createElement(NODE_NAME_KEYS);
		XmlUtils.declareNS(document, NAMESPACE);
		for (Map.Entry<String, String> entry : keys.entrySet()) {
			Element keyElement = document.createElement(NODE_NAME_KEY);
			keyElement.setAttribute(ATTRIBUTE_NAME_ID, entry.getKey());
			keyElement.setAttribute(ATTRIBUTE_NAME_VALUE, entry.getValue());
			legendElement.appendChild(keyElement);
		}
		return legendElement;
	}

	/**
	 * 
	 * Set all the parameters from another parameters set
	 * 
	 * @param parameters
	 *          the parameters to set
	 * 
	 */
	public void setAll(Parameters parameters) {
		keys.putAll(parameters.keys);
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return keys.toString();
	}

	/**
	 * 
	 * Clear all the parameters
	 * 
	 */
	public void clear() {
		keys.clear();
	}
}
