package animatlab.kodamat.core.utils;

import java.util.LinkedHashMap;
import java.util.Map;

import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * 
 * Variable used in a rule
 * 
 * @author Thomas Degris
 * 
 */
public class Variable implements Comparable<Variable> {
	/**
	 * 
	 * Attribute indicating the variable to test
	 * 
	 */
	static public final String ATTRIBUTE_VARIABLE_NAME = "variable";

	/**
	 * 
	 * Attribute indicating the variable to test
	 * 
	 */
	static public final String ATTRIBUTE_TIME_NAME = "time";

	/**
	 * 
	 * Variable used to test on an action
	 * 
	 */
	static public final Variable ACTION = new Variable("Action", (byte) 0);

	/**
	 * 
	 * Time associated to this variable i.e time = -1 for t-1 i.e time = 0 for t
	 * i.e time = 1 for t+1
	 * 
	 */
	final public byte time;

	/**
	 * 
	 * Variable represented by this
	 * 
	 */
	final public String variable;

	/**
	 * 
	 * Hash code for this variable
	 * 
	 */
	final private int hashCode;

	/**
	 * 
	 * Buffer used to avoid the creation of too many variable
	 * 
	 */
	final static private Map<String, Variable> presentVariableBuffer = new LinkedHashMap<String, Variable>();

	/**
	 * 
	 * Buffer used to avoid the creation of too many variable
	 * 
	 */
	final static private Map<String, Variable> futurVariableBuffer = new LinkedHashMap<String, Variable>();

	/**
	 * 
	 * Constructor of a variable to a specific time
	 * 
	 * @param variable
	 *          variable represented by this
	 * @param time
	 *          time associated to this variable
	 * 
	 */
	private Variable(String variable, byte time) {
		this.time = time;
		this.variable = variable;
		hashCode = toString().hashCode();
	}

	/**
	 * 
	 * @return the time suffix for this variable
	 * 
	 */
	private String getSuffix() {
		String suffix;
		switch (time) {
		case 0:
			suffix = "";
			break;
		case 1:
			suffix = "'";
			break;
		default:
			String timeSuffix;
			if (time > 0) {
				timeSuffix = "+" + String.valueOf(time);
			} else {
				timeSuffix = "-" + String.valueOf(time);
			}
			suffix = "{t" + timeSuffix + "}";
		}
		return suffix;
	}

	/**
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object object) {
		if (object == this) {
			return true;
		}
		if (object == null) {
			return false;
		}
		Variable other = (Variable) object;
		if (variable != other.variable) {
			return false;
		}
		if (time != other.time) {
			return false;
		}
		return true;
	}

	/**
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		return hashCode;
	}

	/**
	 * 
	 * Get a variable associated with a variable
	 * 
	 * @param variable
	 *          the index of the variable to get
	 * @return the variable associated to the index
	 * 
	 */
	static public Variable getVariable(String variable) {
		Variable result = presentVariableBuffer.get(variable);
		if (result == null) {
			result = new Variable(variable, (byte) 0);
			presentVariableBuffer.put(variable, result);
		}
		return presentVariableBuffer.get(variable);
	}

	/**
	 * 
	 * Get a variable associated with a variable
	 * 
	 * @param variable
	 *          the index of the variable to get
	 * @param time
	 *          time associated to the variable
	 * @return the variable associated to the index
	 * 
	 */
	static private Variable getVariable(String variable, int time) {
		if (time == 0) {
			return getVariable(variable);
		}
		assert (time == 1);
		Variable result = futurVariableBuffer.get(variable);
		if (result == null) {
			result = new Variable(variable, (byte) 1);
			futurVariableBuffer.put(variable, result);
		}
		return futurVariableBuffer.get(variable);
	}

	/**
	 * 
	 * Get a variable associated with a variable
	 * 
	 * @param variable
	 *          the index of the variable to get
	 * @param time
	 *          time associated to the variable
	 * @return the variable associated to the index
	 * 
	 */
	static public Variable getVariable(Variable variable, int time) {
		return getVariable(variable.variable, time);
	}

	/**
	 * 
	 * Set the attributes of a xml node for this variable
	 * 
	 * @param variableNode
	 *          the xml node to set the attributes
	 * 
	 */
	public void setAttribute(Element variableNode) {
		variableNode.setAttribute(ATTRIBUTE_VARIABLE_NAME, variable);
		variableNode.setAttribute(ATTRIBUTE_TIME_NAME, String.valueOf(time));
	}

	/**
	 * 
	 * Load a variable from a xml node
	 * 
	 * @param xmlNode
	 *          the xml node describing the variable
	 * @return the variable corresponding to the node
	 * 
	 */
	public static Variable load(Node xmlNode) {
		String variable = xmlNode.getAttributes().getNamedItem(
				ATTRIBUTE_VARIABLE_NAME).getNodeValue();
		Node timeNode = xmlNode.getAttributes().getNamedItem(ATTRIBUTE_TIME_NAME);
		if (timeNode == null) {
			return getVariable(variable);
		}
		return getVariable(variable, Byte.parseByte(timeNode.getNodeValue()));
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return variable + getSuffix();

	}

	/**
	 * 
	 * Implementation of the Comparable interface
	 * 
	 * @param other
	 *          other object to compare to
	 * @return a negative integer, zero, or a positive integer as this object is
	 *         less than, equal to, or greater than the specified object.
	 * 
	 */
	public int compareTo(Variable other) {
		if (other.time > time) {
			return -1;
		}
		if (other.time < time) {
			return 1;
		}
		return other.variable.compareTo(variable);
	}

	/**
	 * 
	 * Test if a variable is an action
	 * 
	 * @param variable
	 *          the variable to test
	 * @return true when the variable is an action
	 * 
	 */
	final public static boolean isAction(Variable variable) {
		return ACTION.variable.equals(variable.variable);
	}
}
