/*
 * 
 * Created on 24 mars 2005
 *
 */
package animatlab.kodamat.core.utils;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import animatlab.kodamat.core.io.IToXml;
import animatlab.kodamat.core.io.XmlUtils;

/**
 * 
 * Provides a legend for variables
 * 
 * @author Thomas Degris
 * 
 */
public class VariableOrdering implements IToXml<Object> {
	/**
	 * 
	 * SerialVersionUID
	 * 
	 */
	private static final long serialVersionUID = -1442174098817688901L;

	/**
	 * 
	 * Name of the xml node to represent a legend for decision trees
	 * 
	 */
	public static final String NAMESPACE = "dtr";

	/**
	 * 
	 * Name of the xml node to represent a legend for decision trees
	 * 
	 */
	public static final String NODE_NAME_LEGEND = NAMESPACE + ":" + "legend";

	/**
	 * 
	 * Name of the xml node to represent a legend for decision trees
	 * 
	 */
	public static final String NODE_NAME_VARIABLE = NAMESPACE + ":" + "variable";

	/**
	 * 
	 * Name of the attribute to named a variable
	 * 
	 */
	static private final String ATTRIBUTE_NAME_LEGEND = "legend";

	/**
	 * 
	 * List of variables indicating the variable order
	 * 
	 */
	private final List<Variable> list = new ArrayList<Variable>();

	/**
	 * 
	 * Map to get quickly the index of a variable
	 * 
	 */
	private final Map<Variable, Integer> variableIndex = new LinkedHashMap<Variable, Integer>();

	/**
	 * 
	 * Constructor
	 * 
	 */
	protected VariableOrdering() {
		// Nothing to do by default
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param root
	 *          root of the xml tree to load
	 * 
	 */
	public VariableOrdering(Element root) {
		loadVariableOrdering(root);
	}

	/**
	 * 
	 * Load this legend from a xml tree
	 * 
	 * @param xmlNode
	 *          root node of the xml tree
	 * 
	 */
	private void loadVariableOrdering(Element xmlNode) {
		NodeList nodeList = xmlNode.getChildNodes();
		for (int index = 0; index < nodeList.getLength(); index++) {
			Node variableNode = nodeList.item(index);
			if (variableNode.getNodeType() == Node.ELEMENT_NODE) {
				Element variableElement = (Element) variableNode;
				add(Variable.getVariable(variableElement
						.getAttribute(ATTRIBUTE_NAME_LEGEND)));
			}
		}
	}

	/**
	 * @see animatlab.kodamat.core.io.IToXml#toXml(org.w3c.dom.Document,
	 *      animatlab.kodamat.core.io.IToXml.IToXmlPolicy)
	 */
	public Element toXml(Document document, IToXmlPolicy<Object> policy) {
		Element legendElement = document.createElement(NODE_NAME_LEGEND);
		XmlUtils.declareNS(document, NAMESPACE);
		for (Variable variable : list) {
			if (!Variable.isAction(variable)) {
				Element variableElement = document.createElement(NODE_NAME_VARIABLE);
				variableElement.setAttribute(ATTRIBUTE_NAME_LEGEND, variable.variable);
				legendElement.appendChild(variableElement);
			}
		}
		return legendElement;
	}

	/**
	 * 
	 * Get the variable corresponding to an index
	 * 
	 * @param index
	 *          the index to get the variable from
	 * @return the corresponding variable
	 * 
	 */
	final public Variable get(int index) {
		if (index == variableIndex.size()) {
			return Variable.ACTION;
		}
		return list.get(index);
	}

	/**
	 * 
	 * Get the index of a given variable
	 * 
	 * @param variable
	 *          the variable to get the index from
	 * @return the index of the variable in the variable order
	 * 
	 */
	public int indexOf(Variable variable) {
		if (Variable.isAction(variable)) {
			return variableIndex.size();
		}
		return variableIndex.get(variable);
	}

	/**
	 * 
	 * Add a new variable in the variable order
	 * 
	 * @param variable
	 *          the variable to add
	 * 
	 */
	public void add(Variable variable) {
		list.add(variable);
		variableIndex.put(variable, list.size() - 1);
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return list.toString();
	}

	/**
	 * 
	 * @return the number of variables in the variable order
	 * 
	 */
	public int size() {
		return list.size();
	}

	/**
	 * 
	 * Build a legend for numeric variables
	 * 
	 * @param stateSize
	 *          the size of the state to build the legend for
	 * @return a new legend with variables in numeric ordering
	 * 
	 */
	public static VariableOrdering buildNumericVariableOrdering(int stateSize) {
		VariableOrdering variableOrdering = new VariableOrdering();
		for (int variableIndex = 0; variableIndex < stateSize; variableIndex++) {
			variableOrdering.add(Variable.getVariable(String.valueOf(variableIndex)));
		}
		return variableOrdering;
	}

	/**
	 * 
	 * Build a legend for numeric variables
	 * 
	 * @param legendString
	 *          the string describing the legend
	 * @return a new legend with variables in numeric ordering
	 * 
	 */
	public static VariableOrdering buildVariableOrderingFromString(
			String legendString) {
		VariableOrdering variableOrdering = new VariableOrdering();
		for (String variable : Context.extractList(legendString)) {
			variableOrdering.add(Variable.getVariable(variable));
		}
		return variableOrdering;
	}

	/**
	 * 
	 * Build a list representing a state
	 * 
	 * @param state
	 *          the state to represent
	 * @return the list representing the state
	 * 
	 */
	final public List<String> stateToList(Context state) {
		List<String> result = new ArrayList<String>();
		for (int index = 0; index < variableIndex.size(); index++) {
			result.add(state.get(get(index)));
		}
		return result;
	}

	/**
	 * 
	 * Convert a list into a state
	 * 
	 * @param values
	 *          the values for each variable
	 * @return the corresponding to the list
	 * 
	 */
	final public Context listToState(List<String> values) {
		if (values == null) {
			return null;
		}
		Context result = new Context();
		for (int index = 0; index < values.size(); index++) {
			result.put(get(index), values.get(index));
		}
		return result;
	}

	/**
	 * 
	 * Build an order from a set of variable
	 * 
	 * @param variables
	 *          the set of variables
	 * @return a variable order
	 * 
	 */
	public static VariableOrdering buildOrdering(Set<Variable> variables) {
		VariableOrdering variableOrdering = new VariableOrdering();
		for (Variable variable : variables) {
			variableOrdering.add(variable);
		}
		return variableOrdering;
	}
}
