package animatlab.kodamat.experiments.protocol;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import animatlab.kodamat.core.io.XmlUtils;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.problems.ProblemRunner;
import animatlab.kodamat.reinforcementlearning.agent.IAgent;
import animatlab.kodamat.reinforcementlearning.agent.IAgentFactory;
import animatlab.kodamat.reinforcementlearning.agent.ModelBasedAgent;
import animatlab.kodamat.reinforcementlearning.agent.ModelBasedAgentFactory;
import animatlab.kodamat.reinforcementlearning.core.acting.RandomActing;
import animatlab.kodamat.reinforcementlearning.core.stats.Measure;
import animatlab.kodamat.reinforcementlearning.core.stats.MeasureWriter;
import animatlab.kodamat.reinforcementlearning.core.trajectory.Trajectory;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractForest;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractLearningForest;
import animatlab.kodamat.reinforcementlearning.model.forest.StaticForest;
import animatlab.kodamat.reinforcementlearning.model.forest.VariableActionLearningForest;
import animatlab.kodamat.reinforcementlearning.model.forest.VariableLearningForest;
import animatlab.kodamat.reinforcementlearning.planning.IPlanning;
import animatlab.kodamat.reinforcementlearning.planning.PlanningUtils;
import animatlab.kodamat.supervisedlearning.entropy.Chi2Measure;
import animatlab.kodamat.supervisedlearning.entropy.VariableMeasure;
import animatlab.kodamat.supervisedlearning.learningtree.ID3Tree;
import animatlab.kodamat.supervisedlearning.learningtree.LearningTree;
import animatlab.kodamat.supervisedlearning.regression.LeastSquareMeasure;

/**
 * 
 * Experiment used to run an agent on a problem over time
 * 
 * @author Thomas Degris
 * 
 */
public class AgentOverChi2Exp extends Experiment {
	/**
	 * 
	 * List of algorithm to run
	 * 
	 */
	private List<ModelBasedAgentFactory> learningAgentsFactory = new ArrayList<ModelBasedAgentFactory>();

	/**
	 * 
	 * Add new algorithm to run during the experiment
	 * 
	 * @param agentFactory
	 *          the algorithm to run for each pool
	 * 
	 */
	public void addLearningAgentFactory(ModelBasedAgentFactory agentFactory) {
		learningAgentsFactory.add(agentFactory);
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param problemList
	 *          list of problem to run
	 * @param numberOfTimeStep
	 *          number of time step to run the experiment with
	 * 
	 */
	public AgentOverChi2Exp(String[] problemList, int numberOfTimeStep) {
		super("chiexp", problemList);
		this.numberOfTimeStep = numberOfTimeStep;
	}

	/**
	 * @see animatlab.kodamat.experiments.protocol.Experiment#runAgent(java.util.Random,
	 *      IAgentFactory)
	 */
	@Override
	protected void runAgent(Random random, IAgentFactory agentFactory) {
		IAgent agent = agentFactory.createAgent(random, worldModel, loadedPlanning);
		Trajectory trajectory = new Trajectory(worldModel.getVariableOrdering(),
				worldModel.getActionList());
		ProblemRunner problemRunner = new ProblemRunner(random, worldModel, agent,
				log);
		problemRunner.registerListener(trajectory);
		problemRunner.runProblem(numberOfTimeStep);
		for (ModelBasedAgentFactory learningAgentFactory : learningAgentsFactory) {
			MeasureWriter measureWriter = new MeasureWriter(measures,
					getNetworkFolder() + "/" + getAgentFilePrefix(learningAgentFactory));
			for (double chiValue : measurePoints) {
				LearningTree<String, IProbabilities<String>> prototype = ((AbstractLearningForest) learningAgentFactory
						.getModel()).getInductionPrototype();
				VariableMeasure<String> variableMeasure = prototype.getMeasure();
				variableMeasure.setThreshold(chiValue);

				ModelBasedAgent learningAgent = learningAgentFactory.createAgent(
						random, worldModel, loadedPlanning);
				AbstractLearningForest learningModel = (AbstractLearningForest) learningAgent
						.getModel();
				learningModel.learnFromTrajectory(trajectory, null);
				if (loadAgentFlag) {
					String planningMethod = ((AbstractForest) worldModel).getParameters()
							.getKeyS(Parameters.PARAM_KEY_PLANNING_METHOD);
					IPlanning planning = PlanningUtils.getPlanningMethod(learningModel,
							planningMethod);
					planning.solve();
					learningAgent.setPlanning(planning);
				}
				for (Measure measure : measures) {
					measure.setMeasureData(worldModel, loadedPlanning, learningAgent,
							getNetworkFolder() + "/" + getAgentFilePrefix(agentFactory));
					measure.reset();
					String agentPath = getLocalFolder() + "/"
							+ getAgentFilePrefix(learningAgentFactory) + ".chi" + chiValue;
					save(learningAgent, agentPath);
					// learningAgent.dump(agentPath);
				}
				measureWriter.computeMeasures("chi", chiValue);
			}
			measureWriter.terminate();
		}
	}

	/**
	 * 
	 * Save the state of an agent xml
	 * 
	 * @param agent
	 *          the agent to save
	 * @param filename
	 *          name of the file
	 * 
	 */
	private void save(IAgent agent, String filename) {
		if (!(agent instanceof ModelBasedAgent)) {
			return;
		}

		ModelBasedAgent modelBasedAgent = (ModelBasedAgent) agent;
		String extension = StaticForest.PROBLEM_EXTENSION;
		if (modelBasedAgent.getPlanning() != null) {
			extension = ".agent";
		}
		XmlUtils.save(modelBasedAgent, null, filename + extension);
	}

	/**
	 * @see animatlab.kodamat.experiments.protocol.Experiment#getProblemStampFileName()
	 */
	@Override
	protected String getProblemStampFileName() {
		return null;
	}

	/**
	 * @see animatlab.kodamat.experiments.protocol.Experiment#getExperimentStampFileName()
	 */
	@Override
	public String getExperimentStampFileName() {
		return null;
	}

	/**
	 * @see animatlab.kodamat.experiments.protocol.Experiment#getAgentFilePrefix(animatlab.kodamat.reinforcementlearning.agent.IAgentFactory)
	 */
	@Override
	protected String getAgentFilePrefix(IAgentFactory agentFactory) {
		return (getExperimentName() + "." + currentProblemName + "."
				+ numberOfTimeStep + "." + agentFactory.getAgentName());
	}

	/**
	 * 
	 * Create a batch learning experiment to test on different chi2 variable
	 * 
	 * @param problem
	 *          the problem used to run the experiment
	 * @param numberOfTimeStep
	 *          number of time step for the sample
	 * @param measurePoints
	 *          period where to compute the measure
	 * @param measures
	 *          the measure to run for each size
	 * 
	 * @return the new experiment
	 * 
	 */
	static public Experiment addBatchChi2Exp(String problem,
			int numberOfTimeStep, double[] measurePoints, Measure... measures) {
		AgentOverChi2Exp agentOverChi2Exp = new AgentOverChi2Exp(problem
				.split(",\\s*"), numberOfTimeStep);
		agentOverChi2Exp.addAgentFactory(new RandomActing());
		ID3Tree.Prob<String> inductionTree = new ID3Tree.Prob<String>(
				new Chi2Measure(), null);
		ID3Tree.Plain<Double> regressionTree = new ID3Tree.Plain<Double>(
				new LeastSquareMeasure(null), null);
		agentOverChi2Exp.addLearningAgentFactory(new ModelBasedAgentFactory(null,
				new VariableActionLearningForest(null, null, inductionTree,
						regressionTree), null));
		agentOverChi2Exp.addLearningAgentFactory(new ModelBasedAgentFactory(null,
				new VariableLearningForest(null, null, inductionTree, regressionTree),
				null));
		for (Measure measure : measures) {
			agentOverChi2Exp.addMeasure(measure);
		}
		agentOverChi2Exp.setMeasurePoints(measurePoints);
		return agentOverChi2Exp;
	}

}
