package animatlab.kodamat.experiments.protocol;

import java.util.List;
import java.util.Random;

import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.problems.ProblemRunner;
import animatlab.kodamat.problems.generated.ForestGeneration;
import animatlab.kodamat.reinforcementlearning.agent.IAgentFactory;
import animatlab.kodamat.reinforcementlearning.agent.ModelBasedAgentFactory;
import animatlab.kodamat.reinforcementlearning.core.IStepListener;
import animatlab.kodamat.reinforcementlearning.core.acting.EpsilonGreedy;
import animatlab.kodamat.reinforcementlearning.core.acting.RandomActing;
import animatlab.kodamat.reinforcementlearning.core.stats.Measure;
import animatlab.kodamat.reinforcementlearning.core.stats.MeasureWriter;
import animatlab.kodamat.reinforcementlearning.model.forest.VariableActionLearningForest;
import animatlab.kodamat.reinforcementlearning.model.forest.VariableLearningForest;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.LearningSpiSvi;
import animatlab.kodamat.reinforcementlearning.planning.stats.BatchLearningMeasure;
import animatlab.kodamat.supervisedlearning.entropy.Chi2Measure;
import animatlab.kodamat.supervisedlearning.learningtree.ID3Tree;
import animatlab.kodamat.supervisedlearning.learningtree.SID4Tree;
import animatlab.kodamat.supervisedlearning.regression.LeastSquareMeasure;

/**
 * 
 * Experiment used to run an agent on a problem over time
 * 
 * @author Thomas Degris
 * 
 */
public class AgentOverSizeExp extends Experiment {

	/**
	 * 
	 * Measure wrapper to avoid that the terminate method is called at this of a
	 * ProblemRunner
	 * 
	 */
	private class MeasureWrapper implements IStepListener {
		/**
		 * 
		 * The measure to wrap
		 * 
		 */
		private Measure measure;

		/**
		 * 
		 * Constructor
		 * 
		 * @param measure
		 *          the measure to wrap
		 * 
		 */
		MeasureWrapper(Measure measure) {
			this.measure = measure;
		}

		/**
		 * @see animatlab.kodamat.reinforcementlearning.core.IStepListener#listenStep(int,
		 *      Context, java.lang.String, Context, java.util.List)
		 */
		public void listenStep(int time, Context currentState,
				String actionChoosen, Context nextState, List<Double> lastReward) {
			measure.listenStep(time, currentState, actionChoosen, nextState,
					lastReward);
		}

		/**
		 * @see animatlab.kodamat.reinforcementlearning.core.IStepListener#terminate()
		 */
		public void terminate() {
			// We do nothing : the call is intercepted
		}
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param numberOfTimeStep
	 *          number of time step for the agent in the problem
	 * @param problemList
	 *          list of problem to run
	 * 
	 */
	public AgentOverSizeExp(int numberOfTimeStep, String[] problemList) {
		super("sizeexp", problemList);
		this.numberOfTimeStep = numberOfTimeStep;
	}

	/**
	 * @see animatlab.kodamat.experiments.protocol.Experiment#getListener(animatlab.kodamat.reinforcementlearning.core.stats.Measure)
	 */
	@Override
	protected IStepListener getListener(Measure measure) {
		return new MeasureWrapper(measure);
	}

	/**
	 * @see animatlab.kodamat.experiments.protocol.Experiment#runAgent(java.util.Random,
	 *      IAgentFactory)
	 */
	@Override
	protected void runAgent(Random random, IAgentFactory agentFactory) {
		MeasureWriter measureWriter = new MeasureWriter(measures,
				getNetworkFolder() + "/" + getAgentFilePrefix(agentFactory));
		for (double size : measurePoints) {
			if (size >= 2) {
				ForestGeneration forestGeneration = ForestGeneration
						.createProblemGeneration(currentProblemName);
				forestGeneration.setProblemData((int) size);
				loadProblem(forestGeneration.getProblemDescription());
				ProblemRunner problemRunner = createProblemRunner(random, agentFactory);
				problemRunner.runProblem(numberOfTimeStep);
				measureWriter.computeMeasures("size", size);
			}
		}
		measureWriter.terminate();
	}

	/**
	 * @see animatlab.kodamat.experiments.protocol.Experiment#runProblem(java.lang.String)
	 */
	@Override
	protected void runProblem(String problemName) {
		for (IAgentFactory factory : agentsFactory) {
			for (Measure measure : measures) {
				measure.setMeasureData(null, null, null, getNetworkFolder() + "/"
						+ getAgentFilePrefix(factory));
				measure.reset();
			}
			runAgentFactory(factory);
			for (Measure measure : measures) {
				measure.terminate();
			}
		}
	}

	/**
	 * @see animatlab.kodamat.experiments.protocol.Experiment#getExperimentName()
	 */
	@Override
	protected String getExperimentName() {
		return super.getExperimentName() + numberOfTimeStep;
	}

	/**
	 * @see animatlab.kodamat.experiments.protocol.Experiment#getExperimentStampFileName()
	 */
	@Override
	public String getExperimentStampFileName() {
		return null;
	}

	/**
	 * @see animatlab.kodamat.experiments.protocol.Experiment#getAgentStampFileName(animatlab.kodamat.reinforcementlearning.agent.IAgentFactory)
	 */
	@Override
	protected String getAgentStampFileName(IAgentFactory agentFactory) {
		return null;
	}

	/**
	 * 
	 * Create a batch learning experiment on a variable size problem
	 * 
	 * @param problem
	 *          the problem used to run the experiment
	 * @param minSize
	 *          mininal size to start the experiment with
	 * @param maxSize
	 *          maximal size to start the experiment with
	 * @param numberOfTimeStep
	 *          number of time step to run the experiment
	 * @param measures
	 *          the measure to run for each size
	 * @return the new experiment
	 * 
	 */
	static public Experiment addBatchLearningExp(String problem, int minSize,
			int maxSize, int numberOfTimeStep, Measure... measures) {
		BatchLearningMeasure batchMeasure = new BatchLearningMeasure();
		for (Measure measure : measures) {
			batchMeasure.addMeasure(measure);
		}
		AgentOverSizeExp agentOverSizeExp = new AgentOverSizeExp(numberOfTimeStep,
				problem.split(",\\s*"));
		ID3Tree.Prob<String> inductionTree = new ID3Tree.Prob<String>(
				new Chi2Measure(), null);
		ID3Tree.Plain<Double> regressionTree = new ID3Tree.Plain<Double>(
				new LeastSquareMeasure(null), null);
		batchMeasure.addAgentFactory(new ModelBasedAgentFactory(null,
				new VariableActionLearningForest(null, null, inductionTree,
						regressionTree), null));
		batchMeasure.addAgentFactory(new ModelBasedAgentFactory(null,
				new VariableLearningForest(null, null, inductionTree, regressionTree),
				null));
		agentOverSizeExp.addMeasure(batchMeasure);
		agentOverSizeExp.addAgentFactory(new RandomActing());
		agentOverSizeExp.setMeasurePoints(Measure
				.measurePoints(minSize, maxSize, 1));
		return agentOverSizeExp;
	}

	/**
	 * 
	 * Create an incremental learning experiment on a variable size problem
	 * 
	 * @param problem
	 *          the problem used to run the experiment
	 * @param minSize
	 *          mininal size to start the experiment with
	 * @param maxSize
	 *          maximal size to start the experiment with
	 * @param numberOfTimeStep
	 *          number of time step to run the experiment
	 * @param measures
	 *          the measure to run for each size
	 * @return the new experiment
	 * 
	 */
	static public Experiment addIncLearningExp(String problem, int minSize,
			int maxSize, int numberOfTimeStep, Measure... measures) {
		AgentOverSizeExp agentOverSizeExp = new AgentOverSizeExp(numberOfTimeStep,
				problem.split(",\\s*"));
		agentOverSizeExp.addAgentFactory(AgentOverTimeExp.createFIEAgentFactory(1,
				0.0, 0.0));
		agentOverSizeExp.addAgentFactory(AgentOverTimeExp.createFIEAgentFactory(0,
				100.0, 1));
		agentOverSizeExp.addAgentFactory(new ModelBasedAgentFactory(
				new LearningSpiSvi(), new VariableLearningForest(null, null,
						new SID4Tree.Prob<String>(new Chi2Measure(), null),
						new SID4Tree.Plain<Double>(new LeastSquareMeasure(null), null)),
				new EpsilonGreedy()));
		for (Measure measure : measures) {
			agentOverSizeExp.addMeasure(measure);
		}
		agentOverSizeExp.setMeasurePoints(Measure
				.measurePoints(minSize, maxSize, 1));
		agentOverSizeExp.setExperimentSuffix("mbieeb");
		return agentOverSizeExp;
	}

	/**
	 * Last minute additional results Create an incremental learning experiment on
	 * a variable size problem
	 * 
	 * @param problem
	 *          the problem used to run the experiment
	 * @param minSize
	 *          mininal size to start the experiment with
	 * @param maxSize
	 *          maximal size to start the experiment with
	 * @param numberOfTimeStep
	 *          number of time step to run the experiment
	 * @param measures
	 *          the measure to run for each size
	 * @return the new experiment
	 * 
	 */
	static public Experiment addIncLearningExpParamExplo(String problem,
			int minSize, int maxSize, int numberOfTimeStep, Measure... measures) {
		AgentOverSizeExp agentOverSizeExp = new AgentOverSizeExp(numberOfTimeStep,
				problem.split(",\\s*"));
		agentOverSizeExp.addAgentFactory(AgentOverTimeExp.createFIEAgentFactory(1,
				0.0, 0.0));
		for (Measure measure : measures) {
			agentOverSizeExp.addMeasure(measure);
		}
		agentOverSizeExp.setMeasurePoints(Measure
				.measurePoints(minSize, maxSize, 1));
		agentOverSizeExp.setExperimentSuffix("mbieeblm");
		return agentOverSizeExp;
	}

}
