package animatlab.kodamat.experiments.protocol;

import java.util.Random;

import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.problems.ProblemRunner;
import animatlab.kodamat.reinforcementlearning.agent.IAgentFactory;
import animatlab.kodamat.reinforcementlearning.agent.ModelBasedAgentFactory;
import animatlab.kodamat.reinforcementlearning.core.acting.EpsilonGreedy;
import animatlab.kodamat.reinforcementlearning.core.acting.Greedy;
import animatlab.kodamat.reinforcementlearning.core.acting.RandomActing;
import animatlab.kodamat.reinforcementlearning.core.stats.Measure;
import animatlab.kodamat.reinforcementlearning.core.stats.MeasureWriter;
import animatlab.kodamat.reinforcementlearning.core.stats.RewardMeasure;
import animatlab.kodamat.reinforcementlearning.core.stats.StateActionMeasure;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractLearningForest;
import animatlab.kodamat.reinforcementlearning.model.forest.VariableActionLearningForest;
import animatlab.kodamat.reinforcementlearning.model.forest.VariableLearningForest;
import animatlab.kodamat.reinforcementlearning.model.tabular.TabularLearningModel;
import animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning;
import animatlab.kodamat.reinforcementlearning.planning.dyna.DynaQ;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.LearningFactoredIE;
import animatlab.kodamat.reinforcementlearning.planning.stats.BatchPlanningMeasure;
import animatlab.kodamat.reinforcementlearning.planning.stats.Chi2ModelMeasure;
import animatlab.kodamat.reinforcementlearning.planning.stats.ComputationTimeMeasure;
import animatlab.kodamat.reinforcementlearning.planning.stats.FunctionSizeMeasure;
import animatlab.kodamat.supervisedlearning.entropy.Chi2Measure;
import animatlab.kodamat.supervisedlearning.learningtree.ID3Tree;
import animatlab.kodamat.supervisedlearning.learningtree.SID4Tree;
import animatlab.kodamat.supervisedlearning.regression.LeastSquareMeasure;
import animatlab.kodamat.supervisedlearning.regression.StructureLearningMeasure;

/**
 * 
 * Experiment used to run an agent on a problem over time
 * 
 * @author Thomas Degris
 * 
 */
public class AgentOverTimeExp extends Experiment {
	/**
	 * 
	 * Constructor
	 * 
	 * @param problemList
	 *          list of problem to run
	 * 
	 */
	public AgentOverTimeExp(String[] problemList) {
		super("spiexp", problemList);
	}

	/**
	 * @see animatlab.kodamat.experiments.protocol.Experiment#runAgent(java.util.Random,
	 *      IAgentFactory)
	 */
	@Override
	protected void runAgent(Random random, IAgentFactory agentFactory) {
		ProblemRunner problemRunner = createProblemRunner(random, agentFactory);
		MeasureWriter measureWriter = new MeasureWriter(measures,
				getNetworkFolder() + "/" + getAgentFilePrefix(agentFactory));
		problemRunner.registerListener(measureWriter);
		measureWriter.setMeasurePoints(measurePoints);
		problemRunner.runProblem(numberOfTimeStep + 1);
	}

	/**
	 * @see animatlab.kodamat.experiments.protocol.Experiment#getProblemStampFileName()
	 */
	@Override
	protected String getProblemStampFileName() {
		return null;
	}

	/**
	 * @see animatlab.kodamat.experiments.protocol.Experiment#getExperimentStampFileName()
	 */
	@Override
	public String getExperimentStampFileName() {
		return null;
	}

	/**
	 * @see animatlab.kodamat.experiments.protocol.Experiment#setMeasurePoints(double[])
	 */
	public @Override
	void setMeasurePoints(double[] points) {
		super.setMeasurePoints(points);
		numberOfTimeStep = (int) (measurePoints[measurePoints.length - 1]) + 1;
	}

	/**
	 * 
	 * Set the number of time step to run
	 * 
	 * @param numberOfTimeStep
	 *          the number of time step to run
	 * 
	 */
	public void setNumberOfTimeStep(int numberOfTimeStep) {
		this.numberOfTimeStep = numberOfTimeStep;
	}

	/**
	 * 
	 * Create a MBIEEB agent factory
	 * 
	 * @param betaK
	 *          beta value used for known state
	 * @param betaU
	 *          beta value used for unknown state
	 * @param nU
	 *          weight of unknown states
	 * @return the RMax agent factory
	 * 
	 */
	public static ModelBasedAgentFactory createFIEAgentFactory(int betaK,
			double betaU, double nU) {
		final Chi2Measure chi2Measure;
		if (betaK == 0) {
			chi2Measure = new Chi2Measure(null);
		} else {
			chi2Measure = new Chi2Measure();
		}
		AbstractLearningForest learningForest = new VariableLearningForest(null,
				null, new SID4Tree.Prob<String>(chi2Measure, null),
				new SID4Tree.Plain<Double>(
						new StructureLearningMeasure(betaU, nU, 1E-4), null));
		learningForest.setIncludeActionInRewardLearningFlag(true);
		if (betaK == 0) {
			learningForest.getParameters().setKeyValue(
					Parameters.PARAM_KEY_RESET_REWARD, "false");
		}
		learningForest.getParameters().setKeyValue(Parameters.PARAM_KEY_ORDERING,
				"true");
		// if (betaK > 0) {
		// learningForest.getParameters().setKeyValue(Parameters.PARAM_KEY_ORDERING_THRESHOLD,
		// "0.01");
		// }

		ModelBasedAgentFactory modelBasedAgentFactory = new ModelBasedAgentFactory(
				new LearningFactoredIE(betaK), learningForest, new Greedy());
		return modelBasedAgentFactory;
	}

	/**
	 * 
	 * Planning/EGreedy agent compared to DynaQ
	 * 
	 * @param problem
	 *          problem to solve
	 * @param planning
	 *          planning method to use to solve the problem
	 * @param numberOfTimeStep
	 *          number of time step to run the experiment
	 * @return the new experiment
	 * 
	 */
	public static AgentOverTimeExp getEGreedyOverTime(String problem,
			IIncrementalPlanning planning, int numberOfTimeStep) {
		AgentOverTimeExp spiExperiment = new AgentOverTimeExp(problem
				.split(",\\s*"));
		// Agent factory
		SID4Tree.Prob<String> inductionTree = new SID4Tree.Prob<String>(
				new Chi2Measure(), null);
		SID4Tree.Plain<Double> regressionTree = new SID4Tree.Plain<Double>(
				new LeastSquareMeasure(null), null);
		spiExperiment.addAgentFactory(new RandomActing());
		spiExperiment.addAgentFactory(new Greedy());
		spiExperiment.addAgentFactory(new ModelBasedAgentFactory(planning,
				new VariableActionLearningForest(null, null, inductionTree,
						regressionTree), new EpsilonGreedy()));
		spiExperiment.addAgentFactory(new ModelBasedAgentFactory(planning,
				new VariableLearningForest(null, null, inductionTree, regressionTree),
				new EpsilonGreedy()));
		spiExperiment.addAgentFactory(new ModelBasedAgentFactory(new DynaQ(),
				new TabularLearningModel(), new EpsilonGreedy()));
		// Measures
		addIncrementalMeasure(spiExperiment);

		spiExperiment.setNumberOfTimeStep(numberOfTimeStep);
		return spiExperiment;
	}

	/**
	 * 
	 * Add some incremental measures
	 * 
	 * @param spiExperiment
	 *          the experiment where to add the measure
	 * 
	 */
	private static void addIncrementalMeasure(AgentOverTimeExp spiExperiment) {
		spiExperiment.addMeasure(new RewardMeasure());
		spiExperiment.addMeasure(new FunctionSizeMeasure());
		spiExperiment.addMeasure(new ComputationTimeMeasure());
		spiExperiment.addMeasure(new StateActionMeasure());
	}

	/**
	 * 
	 * Create a batch planning experiment on a given problem
	 * 
	 * @param problem
	 *          the problem used to run the experiment
	 * @param endExp
	 *          end of the experiment
	 * 
	 * @return the new experiment
	 * 
	 */
	static public Experiment addBatchPlanningExp(String problem, double endExp) {
		ID3Tree.Prob<String> inductionTree = new ID3Tree.Prob<String>(
				new Chi2Measure(), null);
		ID3Tree.Plain<Double> regressionTree = new ID3Tree.Plain<Double>(
				new LeastSquareMeasure(null), null);
		BatchPlanningMeasure batchLearningMeasure = new BatchPlanningMeasure();
		batchLearningMeasure.addAgentFactory(new RandomActing());
		batchLearningMeasure.addAgentFactory(new Greedy());
		batchLearningMeasure.addAgentFactory(new ModelBasedAgentFactory(null,
				new VariableActionLearningForest(null, null, inductionTree,
						regressionTree), null));
		batchLearningMeasure.addAgentFactory(new ModelBasedAgentFactory(null,
				new VariableLearningForest(null, null, inductionTree, regressionTree),
				null));
		batchLearningMeasure.addMeasure(new RewardMeasure());
		batchLearningMeasure.addMeasure(new FunctionSizeMeasure());
		batchLearningMeasure.addMeasure(new Chi2ModelMeasure());
		AgentOverTimeExp spiExperiment = new AgentOverTimeExp(problem
				.split(",\\s*"));
		spiExperiment.addAgentFactory(new RandomActing());
		spiExperiment.addMeasure(batchLearningMeasure);
		spiExperiment.setPrefix("batch");
		spiExperiment.setMeasurePoints(Measure.measurePoints(endExp * 0.02, endExp,
				endExp / 10.0));
		return spiExperiment;
	}

	// /**
	// *
	// * Planning RMax agent
	// *
	// * @param problem problem to solve
	// * @param numberOfTry number of try before a state is known
	// * @param rmaxValue rmax value used for unknown states
	// * @param numberOfTimeStep number of time step to run the experiment
	// * @return the new experiment
	// *
	// **/
	// public static Experiment getRMaxOverTime(String problem, int numberOfTry,
	// double rmaxValue, int numberOfTimeStep) {
	// AgentOverTimeExp spiExperiment = new
	// AgentOverTimeExp(problem.split(",\\s*"));
	// spiExperiment.addAgentFactory(createMBIEEBAgentFactory(numberOfTry,
	// rmaxValue));
	// addIncrementalMeasure(spiExperiment);
	// spiExperiment.setNumberOfTimeStep(numberOfTimeStep);
	// return spiExperiment;
	//
	// }
	//	
	/**
	 * 
	 * Planning RMax agent
	 * 
	 * @param problem
	 *          problem to solve
	 * @param numberOfTimeStep
	 *          number of time step to run the experiment
	 * @param betaK
	 *          beta value used for known state
	 * @param betaU
	 *          beta value used for unknown state
	 * @param nU
	 *          weight of unknown states
	 * @return the new experiment
	 * 
	 */
	public static Experiment getFIEOverTime(String problem, int numberOfTimeStep,
			int betaK, double betaU, double nU) {
		AgentOverTimeExp spiExperiment = new AgentOverTimeExp(problem
				.split(",\\s*"));
		spiExperiment.addAgentFactory(createFIEAgentFactory(betaK, betaU, nU));
		addIncrementalMeasure(spiExperiment);
		spiExperiment.setNumberOfTimeStep(numberOfTimeStep);
		return spiExperiment;

	}

}
