package animatlab.kodamat.experiments.protocol;

import static animatlab.kodamat.core.io.StampFile.exists;
import static animatlab.kodamat.core.io.StampFile.touchFinished;
import static animatlab.kodamat.core.io.StampFile.touchWorking;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import animatlab.kodamat.core.IProblemDefinition;
import animatlab.kodamat.core.io.ILog;
import animatlab.kodamat.core.io.Log;
import animatlab.kodamat.core.io.XmlUtils;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.problems.ProblemRunner;
import animatlab.kodamat.problems.generated.ForestGeneration;
import animatlab.kodamat.reinforcementlearning.agent.IAgent;
import animatlab.kodamat.reinforcementlearning.agent.IAgentFactory;
import animatlab.kodamat.reinforcementlearning.core.IStepListener;
import animatlab.kodamat.reinforcementlearning.core.stats.Measure;
import animatlab.kodamat.reinforcementlearning.model.forest.StaticForest;
import animatlab.kodamat.reinforcementlearning.planning.IPlanning;
import animatlab.kodamat.reinforcementlearning.planning.guestrin.StaticGuestrin;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.StaticSpiSvi;
import animatlab.kodamat.reinforcementlearning.planning.spudd.StaticSpudd;

/**
 * 
 * Abstract class for experiment
 * 
 * @author Thomas Degris
 * 
 */
public abstract class Experiment implements ILog {
	/**
	 * 
	 * Loader of customized problem
	 * 
	 */
	protected interface CustomizedProblemLoader {
		/**
		 * 
		 * Load a customize problem
		 * 
		 * @param problemName
		 *          the name of the problem to load
		 * @return the problem once loaded
		 * 
		 */
		IProblemDefinition loadCustomizedProblem(String problemName);
	}

	/**
	 * 
	 * Number of time step to run the agent
	 * 
	 */
	protected int numberOfTimeStep;

	/**
	 * 
	 * Name of the problem to test
	 * 
	 */
	final private String[] problemList;

	/**
	 * 
	 * Folder where all the results and stamps are written
	 * 
	 */
	private String networkFolder;

	/**
	 * 
	 * Folder where all the logs are written
	 * 
	 */
	private String localFolder;

	/**
	 * 
	 * Name of the experiment
	 * 
	 */
	final private String experimentName;

	/**
	 * 
	 * World model used in the experiment
	 * 
	 */
	protected IProblemDefinition worldModel = null;

	/**
	 * 
	 * World model used in the experiment
	 * 
	 */
	protected IPlanning loadedPlanning = null;

	/**
	 * 
	 * Name of the current problem
	 * 
	 */
	protected String currentProblemName = null;

	/**
	 * 
	 * Flag to decide weither or not an agent is loaded or the forest only
	 * 
	 */
	protected boolean loadAgentFlag = false;

	/**
	 * 
	 * Index of the current pool
	 * 
	 */
	private int poolIndex = -1;

	/**
	 * 
	 * Output log
	 * 
	 */
	protected Log log;

	/**
	 * 
	 * Loader of customized problem
	 * 
	 */
	private CustomizedProblemLoader customizedProblemLoader = null;

	/**
	 * 
	 * Points where to compute measures
	 * 
	 */
	protected double[] measurePoints = null;

	/**
	 * 
	 * List of algorithm to run
	 * 
	 */
	protected List<IAgentFactory> agentsFactory = new ArrayList<IAgentFactory>();

	/**
	 * 
	 * The measures associated to the experiment
	 * 
	 */
	protected List<Measure> measures = new ArrayList<Measure>();

	/**
	 * 
	 * Prefix to use in file name
	 * 
	 */
	private String prefix = null;

	/**
	 * 
	 * Suffix to add to the experiment name
	 * 
	 */
	private String experimentSuffix = "";

	/**
	 * 
	 * Constructor
	 * 
	 * @param experimentName
	 *          name of the experiment to run
	 * @param problemList
	 *          list of problem to run
	 * 
	 */
	protected Experiment(String experimentName, String[] problemList) {
		this.experimentName = experimentName;
		this.problemList = problemList;
	}

	/**
	 * 
	 * Add new algorithm to run during the experiment
	 * 
	 * @param agentFactory
	 *          the algorithm to run for each pool
	 * 
	 */
	public void addAgentFactory(IAgentFactory agentFactory) {
		agentsFactory.add(agentFactory);
		loadAgentFlag = loadAgentFlag || agentFactory.requireSolution();
	}

	/**
	 * 
	 * Add a new measure to the experiment
	 * 
	 * @param measure
	 *          the measure to add
	 * 
	 */
	public void addMeasure(Measure measure) {
		measures.add(measure);
		loadAgentFlag = loadAgentFlag || measure.requireSolution();
	}

	/**
	 * 
	 * Load the problem
	 * 
	 * @param problemName
	 *          the description of the problem to load
	 * 
	 */
	protected void loadProblem(String problemName) {
		loadedPlanning = null;
		if (loadAgentFlag) {
			loadedPlanning = loadAgent(problemName);
		}
		if (loadedPlanning == null) {
			worldModel = loadForest(problemName);
		} else {
			worldModel = (IProblemDefinition) loadedPlanning.getModel();
		}
		if ((worldModel == null) && (customizedProblemLoader != null)) {
			worldModel = customizedProblemLoader.loadCustomizedProblem(problemName);
		}
		assert (worldModel != null);
	}

	/**
	 * 
	 * Load a forest from a problem name
	 * 
	 * @param problemName
	 *          the name of the problem to load
	 * 
	 * @return the model loaded
	 * 
	 */
	public static StaticForest loadForest(String problemName) {
		return loadForest(problemName, "");
	}

	/**
	 * 
	 * Load a forest from a problem name
	 * 
	 * @param problemName
	 *          the name of the problem to load
	 * @param folderPrefix
	 *          prefix of the folder to access to model file
	 * 
	 * @return the model loaded
	 * 
	 */
	public static StaticForest loadForest(String problemName, String folderPrefix) {
		StaticForest result = ForestGeneration.toForest(problemName);
		if (result != null) {
			return result;
		}
		String forestFolderPrefix = folderPrefix + "data/forests/";
		String forestFileName = forestFolderPrefix + problemName
				+ StaticForest.PROBLEM_EXTENSION;
		if (new File(forestFileName).canRead()) {
			return new StaticForest(XmlUtils.getRootFromFile(forestFileName));
		}
		int noisyIndex = problemName.lastIndexOf(".noisy");
		if (noisyIndex == -1) {
			return null;
		}
		String originalProblemName = problemName.substring(0, noisyIndex);
		forestFileName = forestFolderPrefix + originalProblemName
				+ StaticForest.PROBLEM_EXTENSION;
		result = new StaticForest(XmlUtils.getRootFromFile(forestFileName));
		result.toNoisy(0.1);
		return result;
	}

	/**
	 * 
	 * Load the agent used as reference
	 * 
	 * @param problemName
	 *          the problem to load the agent for
	 * 
	 * @return the agent loaded
	 * 
	 */
	private IPlanning loadAgent(String problemName) {
		String forestFolder = "data/forests";
		String forestFilename = forestFolder + "/"
				+ ForestGeneration.getFileName(problemName);
		String truncFilename = forestFilename.substring(0, forestFilename
				.lastIndexOf("."));
		if (new File(truncFilename + StaticSpiSvi.AGENT_EXTENSION).canRead()) {
			return new StaticSpiSvi(XmlUtils.getRootFromFile(truncFilename
					+ StaticSpiSvi.AGENT_EXTENSION));
		}
		if (new File(truncFilename + StaticSpudd.AGENT_EXTENSION).canRead()) {
			return new StaticSpudd(XmlUtils.getRootFromFile(truncFilename
					+ StaticSpudd.AGENT_EXTENSION));
		}
		if (new File(truncFilename + StaticGuestrin.AGENT_EXTENSION).canRead()) {
			return new StaticGuestrin(XmlUtils.getRootFromFile(truncFilename
					+ StaticGuestrin.AGENT_EXTENSION));
		}
		return forestToAgent(problemName, forestFolder);
	}

	/**
	 * 
	 * Convert a forest (a problem without solution) into an agent (a problem with
	 * a near optimal solution)
	 * 
	 * @param problemName
	 *          path of the file representing the problem to solve
	 * @param agentFolder
	 *          folder where to save the agent
	 * 
	 * @return the planning method used to convert the forest into an agent
	 * 
	 */
	static public IPlanning forestToAgent(String problemName, String agentFolder) {
		String forestFilename = "data/forests/"
				+ ForestGeneration.getFileName(problemName);
		String extension = forestFilename
				.substring(forestFilename.lastIndexOf("."));
		assert (extension.equals(StaticForest.PROBLEM_EXTENSION));
		String agentFilename = agentFolder + "/" + problemName;
		StaticForest forest = loadForest(problemName);
		if (forest == null) {
			return null;
		}
		if ("false".equals(forest.getParameters().getKeyS(
				Parameters.PARAM_KEY_AUTO_AGENT))) {
			return null;
		}
		String planningMethod = forest.getParameters().getKeyS(
				Parameters.PARAM_KEY_PLANNING_METHOD);
		if (StaticGuestrin.PLANNING_METHOD_NAME.equals(planningMethod)) {
			return StaticGuestrin.convertProblemToAgent(forest, agentFilename);
		} else if (StaticSpudd.PLANNING_METHOD_NAME.equals(planningMethod)) {
			return StaticSpudd.convertProblemToAgent(forest, agentFilename);
		}
		return StaticSpiSvi.convertProblemToAgent(forest, agentFilename);
	}

	/**
	 * 
	 * Get the full path of a file for a pool
	 * 
	 * @return the full path (null if the filename is null)
	 * 
	 */
	public String getNetworkFolder() {
		return networkFolder;
	}

	/**
	 * 
	 * Get the full path of a file for a pool
	 * 
	 * @return the full path (null if the filename is null)
	 * 
	 */
	public String getLocalFolder() {
		return localFolder;
	}

	/**
	 * 
	 * Get the file name in the pool associated with the next running agent
	 * 
	 * @param agentFactory
	 *          the agent to get the filename from
	 * 
	 * @return the file name associated with the next running agent used to stamp
	 *         the experiment
	 * 
	 */
	protected String getAgentStampFileName(IAgentFactory agentFactory) {
		return getNetworkFolder() + "/" + getAgentFilePrefix(agentFactory);
	}

	/**
	 * 
	 * Get the file name in the pool associated with the current problem
	 * 
	 * @return the file name associated with the next running problem used to
	 *         stamp the experiment
	 * 
	 */
	protected String getProblemStampFileName() {
		return getNetworkFolder() + "/" + getExperimentName() + "."
				+ currentProblemName;
	}

	/**
	 * 
	 * Get the file name in the pool associated with the current experiment
	 * 
	 * @return the file name associated with the next running problem used to
	 *         stamp the experiment
	 * 
	 */
	public String getExperimentStampFileName() {
		return getNetworkFolder() + "/" + getExperimentName();
	}

	/**
	 * 
	 * @return the name of the experiment
	 * 
	 */
	protected String getExperimentName() {
		return experimentName + experimentSuffix;
	}

	/**
	 * 
	 * Build a file path prefix (with folders) for the current agent
	 * 
	 * @param agentFactory
	 *          the agent factory to get the file prefix for
	 * 
	 * @return the file path prefix
	 * 
	 */
	protected String getAgentFilePrefix(IAgentFactory agentFactory) {
		StringBuilder result = new StringBuilder(getExperimentName());
		if (prefix != null) {
			result.append(prefix);
		}
		result.append(".");
		return result.toString() + currentProblemName + "."
				+ agentFactory.getAgentName();
	}

	/**
	 * 
	 * Run a pool of experiment (with different policies)
	 * 
	 * @param problemName
	 *          name of the problem to use in the experiment
	 * 
	 */
	protected void runProblem(String problemName) {
		loadProblem(problemName);
		for (IAgentFactory factory : agentsFactory) {
			runAgentFactory(factory);
		}
	}

	/**
	 * 
	 * Run an agent factory
	 * 
	 * @param agentFactory
	 *          the agent factory to run
	 * 
	 */
	protected void runAgentFactory(IAgentFactory agentFactory) {
		Random random = new Random(poolIndex);
		String agentStampPath = getAgentStampFileName(agentFactory);
		if (exists(agentStampPath)) {
			log.out().println("Skipping " + agentFactory.getAgentName());
			return;
		}
		touchWorking(agentStampPath);
		log.out().println("Starting to log " + getAgentFilePrefix(agentFactory));
		agentFactory.setLog(log);
		runAgent(random, agentFactory);
		touchFinished(agentStampPath);
		return;
	}

	/**
	 * 
	 * Run the experiment for an agent
	 * 
	 * @param agentFactory
	 *          the agent to run
	 * @param random
	 *          random instance to use in the pool
	 * 
	 */
	abstract void runAgent(Random random, IAgentFactory agentFactory);

	/**
	 * 
	 * Create the problem runner (register all the measure and create the agent)
	 * 
	 * @param random
	 *          random instance used by the agent
	 * @param agentFactory
	 *          the factory used to create the agent
	 * @return the problem runner
	 * 
	 */
	protected ProblemRunner createProblemRunner(Random random,
			IAgentFactory agentFactory) {
		IAgent agent = agentFactory.createAgent(random, worldModel, loadedPlanning);
		ProblemRunner problemRunner = new ProblemRunner(random, worldModel, agent,
				log);
		for (Measure measure : measures) {
			measure.setMeasureData(worldModel, loadedPlanning, agent,
					getNetworkFolder() + "/" + getAgentFilePrefix(agentFactory));
			measure.setLog(log);
			problemRunner.registerListener(getListener(measure));
		}
		return problemRunner;
	}

	/**
	 * 
	 * Get the listener associated with a measure (by default, it is the measure
	 * itself)
	 * 
	 * @param measure
	 *          the measure to create the listener for
	 * @return a step listener that will be called at each time problem of a
	 *         problem runner
	 * 
	 */
	protected IStepListener getListener(Measure measure) {
		return measure;
	}

	/**
	 * 
	 * Run a problem in a pool
	 * 
	 * @param poolIndex
	 *          index of the pool to run
	 * 
	 */
	final protected void runExperiment(int poolIndex) {
		this.poolIndex = poolIndex;
		for (String problemName : problemList) {
			currentProblemName = problemName;
			String problemStampPath = getProblemStampFileName();
			if (exists(problemStampPath)) {
				log.out().println("Skipping " + currentProblemName);
			} else {
				touchWorking(problemStampPath);
				runProblem(problemName);
				touchFinished(problemStampPath);
			}
		}
	}

	/**
	 * 
	 * Set loadAgentFlag
	 * 
	 * @param loadAgentFlag
	 *          The loadAgentFlag to set.
	 * 
	 */
	public void setLoadAgentFlag(boolean loadAgentFlag) {
		this.loadAgentFlag = loadAgentFlag;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuilder result = new StringBuilder(getClass().getSimpleName() + ": ");
		for (String problemName : problemList) {
			result.append(problemName);
			result.append(", ");
		}
		return result.substring(0, result.length() - 2);
	}

	/**
	 * 
	 * Set the output folder for the next experiment
	 * 
	 * @param networkFolder
	 *          network folder used for the results and stamps
	 * @param localFolder
	 *          local folder used for logs
	 * 
	 */
	public void setOutputFolders(String networkFolder, String localFolder) {
		this.networkFolder = networkFolder;
		this.localFolder = localFolder;
	}

	/**
	 * @see animatlab.kodamat.core.io.ILog#setLog(animatlab.kodamat.core.io.Log)
	 */
	public void setLog(Log log) {
		this.log = log;
	}

	/**
	 * 
	 * Set measurePoints
	 * 
	 * @param measurePointsString
	 *          the measurePoints to set (as a string)
	 * 
	 */
	public void setMeasurePoints(String measurePointsString) {
		String[] pointsArray = measurePointsString.split(",");
		double[] points = new double[pointsArray.length];
		for (int index = 0; index < pointsArray.length; index++) {
			points[index] = Double.parseDouble(pointsArray[index]);
		}
		setMeasurePoints(points);
	}

	/**
	 * 
	 * Set the measure points for this experiment
	 * 
	 * @param points
	 *          the set of points to measure
	 * 
	 */
	public void setMeasurePoints(double[] points) {
		this.measurePoints = points;
	}

	/**
	 * 
	 * Set the prefix in the file name
	 * 
	 * @param prefix
	 *          the prefix to use in the filename
	 * 
	 */
	public void setPrefix(String prefix) {
		this.prefix = prefix;
	}

	/**
	 * 
	 * Set customizedProblemLoader
	 * 
	 * @param customizedProblemLoader
	 *          The customizedProblemLoader to set.
	 * 
	 */
	public void setCustomizedProblemLoader(
			CustomizedProblemLoader customizedProblemLoader) {
		this.customizedProblemLoader = customizedProblemLoader;
	}

	/**
	 * 
	 * Set the suffix to add to the experiement name
	 * 
	 * @param experimentSuffix
	 *          The experimentSuffix to set.
	 * 
	 */
	public void setExperimentSuffix(String experimentSuffix) {
		this.experimentSuffix = experimentSuffix;
	}

}
