package animatlab.kodamat.experiments.protocol;

import static animatlab.kodamat.core.io.StampFile.exists;
import static animatlab.kodamat.core.io.StampFile.touchFinished;
import static animatlab.kodamat.core.io.StampFile.touchWorking;

import java.io.File;
import java.util.LinkedList;
import java.util.List;

import animatlab.kodamat.core.IProblemDefinition;
import animatlab.kodamat.core.io.Log;
import animatlab.kodamat.experiments.protocol.Experiment.CustomizedProblemLoader;

/**
 * 
 * Class to decide what to run for a pool
 * 
 * @author Thomas Degris
 * 
 */
public abstract class PoolRunner implements CustomizedProblemLoader {
	/**
	 * 
	 * Prefix added before pool index in the folder name
	 * 
	 */
	private static final String POOL_PREFIX = "data";

	/**
	 * 
	 * Name of the
	 * 
	 */
	private final String studyName;

	/**
	 * 
	 * Number of experiment to do
	 * 
	 */
	protected int numberOfExperiment = 30;

	/**
	 * 
	 * Output folder of all the results
	 * 
	 */
	protected String networkFolder;

	/**
	 * 
	 * Output folder of all the results
	 * 
	 */
	protected String localFolder;

	/**
	 * 
	 * Constructor
	 * 
	 * @param studyName
	 *          name of the study regrouping a number of pool of experiments
	 * 
	 * 
	 */
	public PoolRunner(String studyName) {
		this.studyName = studyName;
	}

	/**
	 * 
	 * Get the string for the output folder for a specific pool
	 * 
	 * @param outputFolder
	 *          output folder for the pool
	 * @param poolIndex
	 *          index of the pool
	 * 
	 * @return a string representing the path of the output folder of the pool
	 * 
	 */
	private String getPoolFolder(String outputFolder, int poolIndex) {
		return outputFolder + "/" + POOL_PREFIX + String.valueOf(poolIndex);
	}

	/**
	 * @see animatlab.kodamat.experiments.protocol.Experiment.CustomizedProblemLoader#loadCustomizedProblem(java.lang.String)
	 */
	public IProblemDefinition loadCustomizedProblem(String problemName) {
		return null;
	}

	/**
	 * 
	 * Run all the pools
	 * 
	 * @param args
	 *          args passed to the study
	 * 
	 */
	public void runPools(String[] args) {
		if ((args.length != 2) && (args.length != 3)) {
			System.err
					.println("Usage: java animatlab.kodamat.experiments.main.<Study> network_folder local_folder");
			System.exit(1);
		}
		networkFolder = args[0] + "/results/" + studyName;
		localFolder = args[1] + "/results/" + studyName;
		System.out.println("Network path: " + networkFolder);
		System.out.println("Local path: " + localFolder);
		for (int poolIndex = 0; poolIndex < numberOfExperiment; poolIndex++) {
			List<Experiment> experiments = new LinkedList<Experiment>();
			buildExperimentList(experiments);
			checkFolder(getPoolFolder(networkFolder, poolIndex));
			checkFolder(getPoolFolder(localFolder, poolIndex));
			while (!experiments.isEmpty()) {
				Experiment experiment = experiments.remove(0);
				Log log = new Log(getPoolFolder(localFolder, poolIndex) + "/pool.log");
				experiment.setLog(log);
				experiment.setCustomizedProblemLoader(this);
				experiment.setOutputFolders(getPoolFolder(networkFolder, poolIndex),
						getPoolFolder(localFolder, poolIndex));
				String problemStampPath = null;
				if (experiment.getExperimentStampFileName() != null) {
					problemStampPath = getPoolFolder(networkFolder, poolIndex) + "/"
							+ experiment.getExperimentStampFileName();
				}
				if (exists(problemStampPath)) {
					log.out().println("Skipping " + experiment.getExperimentName());
				} else {
					log.out().println("Running " + experiment.getExperimentName());
					touchWorking(problemStampPath);
					experiment.runExperiment(poolIndex);
					touchFinished(problemStampPath);
				}
			}
		}
	}

	/**
	 * 
	 * Check if a folder exist, if not, create it
	 * 
	 * @param folderPath
	 *          the folder to check
	 * 
	 */
	private void checkFolder(String folderPath) {
		File folderFile = new File(folderPath);
		if (!folderFile.canRead()) {
			folderFile.mkdirs();
		}
	}

	/**
	 * 
	 * Build a list of experiments to run during a pool
	 * 
	 * @param experiments
	 *          list of experiments to run during a pool
	 * 
	 */
	abstract protected void buildExperimentList(List<Experiment> experiments);
}
