/*
 * 
 * Created on Jan 11, 2006
 *
 */
package animatlab.kodamat.problems;

import java.util.LinkedList;
import java.util.List;
import java.util.Random;

import animatlab.kodamat.core.IProblemDefinition;
import animatlab.kodamat.core.io.Log;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.reinforcementlearning.agent.IAgent;
import animatlab.kodamat.reinforcementlearning.core.IStepListener;

/**
 * 
 * Class used to run steps of an agent in a problem
 * 
 * @author Thomas Degris
 * 
 */
public class ProblemRunner {
	/**
	 * 
	 * The problem to run
	 * 
	 */
	final public IProblemDefinition problem;

	/**
	 * 
	 * The agent to run
	 * 
	 */
	final public IAgent agent;

	/**
	 * 
	 * List of the listener to run at each time step
	 * 
	 */
	final public List<IStepListener> stepListeners = new LinkedList<IStepListener>();

	/**
	 * 
	 * Current state of the agent
	 * 
	 */
	private Context currentState;

	/**
	 * 
	 * Reward received by the agent
	 * 
	 */
	private List<Double> reward;

	/**
	 * 
	 * Time in the problem
	 * 
	 */
	private int time;

	/**
	 * 
	 * Maximum time of an episode
	 * 
	 */
	private Integer maxEpisodeTime = null;

	/**
	 * 
	 * Last time an episode was started
	 * 
	 */
	private int lastEpisodeTime;

	/**
	 * 
	 * Log of the experiment
	 * 
	 */
	protected Log log;

	/**
	 * 
	 * Random instance to use to run the agent in the problem
	 * 
	 */
	final private Random random;

	/**
	 * 
	 * Constructor
	 * 
	 * @param random
	 *          random instance to use in the problem
	 * @param problem
	 *          the problem to run
	 * @param agent
	 *          the agent to run
	 * @param log
	 *          where the log the experiment
	 * 
	 * 
	 */
	public ProblemRunner(Random random, IProblemDefinition problem, IAgent agent,
			Log log) {
		this.random = random;
		this.problem = problem;
		this.agent = agent;
		this.log = log;
		agent.setLog(log);
		initProblem();
	}

	/**
	 * 
	 * Register a new step listener
	 * 
	 * @param listener
	 *          the listener to register
	 * 
	 */
	public void registerListener(IStepListener listener) {
		stepListeners.add(listener);
	}

	/**
	 * 
	 * Initialize the agent data to start a new problem
	 * 
	 */
	private void initProblem() {
		currentState = problem.getInitialState(random);
		reward = null;
		time = 0;
		lastEpisodeTime = 0;
		if (problem.getParameters().hasKey(Parameters.PARAM_KEY_EPISODE_TIME)) {
			maxEpisodeTime = problem.getParameters().getKeyI(
					Parameters.PARAM_KEY_EPISODE_TIME);
		}
	}

	/**
	 * 
	 * Run the coffee robot problem
	 * 
	 * @param numberOfSteps
	 *          time of the experiment
	 * 
	 */
	public void runProblem(int numberOfSteps) {
		initProblem();
		for (int stepIndex = 0; stepIndex < numberOfSteps; stepIndex++) {
			oneStep();
		}
		for (IStepListener listener : stepListeners) {
			listener.terminate();
		}
	}

	/**
	 * 
	 * Run the problem for one step
	 * 
	 */
	public void oneStep() {
		if (log != null) {
			log.out().print(time + ": " + String.valueOf(currentState) + " ");
		}
		String action = agent.getAction(time, currentState, reward);
		if (currentState == null) {
			assert (action == null);
		}
		Context nextState = new Context();
		reward = problem.getNextState(random, currentState, nextState, action);
		if ((nextState.size() == 0)
				|| ((maxEpisodeTime != null) && ((time - lastEpisodeTime) > maxEpisodeTime))) {
			nextState = null;
			lastEpisodeTime = time;
		}
		for (IStepListener listener : stepListeners) {
			listener.listenStep(time, currentState, action, nextState, reward);
		}
		if (log != null) {
			log.out().print(
					action + " " + String.valueOf(nextState) + " "
							+ Context.rewardToString(reward));
			log.out().println();
		}
		currentState = nextState;
		time++;
	}

	/**
	 * 
	 * @return the current time
	 * 
	 */
	public int getTime() {
		return time;
	}

}
