/*
 * 
 * Created on 13 sept. 2005
 *
 */
package animatlab.kodamat.problems.generated.syntheticmdps.tests;

import static animatlab.kodamat.problems.generated.syntheticmdps.ExponProblem.WORSTCASE_PROBLEM_NAME;
import static animatlab.kodamat.problems.generated.syntheticmdps.LinearProblem.BESTCASE_PROBLEM_NAME;
import static animatlab.kodamat.problems.generated.syntheticmdps.NoisyLinearProblem.NOISYCASE_PROBLEM_NAME;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import junit.framework.TestCase;
import animatlab.kodamat.core.IProblemDefinition;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.problems.generated.ForestGeneration;
import animatlab.kodamat.problems.generated.syntheticmdps.ExponProblem;
import animatlab.kodamat.problems.generated.syntheticmdps.LinearProblem;
import animatlab.kodamat.problems.generated.syntheticmdps.SyntheticMDPsProblem;
import animatlab.kodamat.reinforcementlearning.model.forest.StaticForest;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.SpiSvi;

/**
 * 
 * Best case problem defined in Stochastic Dynamic Programming with Factored
 * Representations (Boutilier, Dearden, Goldszmidt)
 * 
 * @author Thomas Degris
 * 
 */
public class TestSyntheticMdps extends TestCase {

	/**
	 * 
	 * Size of the problem used for the testing suite
	 * 
	 */
	final static int PROBLEM_SIZE = 11;

	/**
	 * 
	 * Variable ordering to use
	 * 
	 */
	final static VariableOrdering variableOrdering = VariableOrdering
			.buildNumericVariableOrdering(PROBLEM_SIZE);

	/**
	 * 
	 * Level of noise in stochastics problem
	 * 
	 */
	final static double NOISE_LEVEL = 0.2;

	/**
	 * 
	 * Number of random state to try
	 * 
	 */
	final static int NUMBER_OF_TRY = 500;

	/**
	 * 
	 * Randomizer used for testing purpose
	 * 
	 */
	final static Random random = new Random(0);

	/**
	 * 
	 * Best case problem directly coded in Java to compare with the XMl
	 * representation
	 * 
	 * @author Thomas Degris
	 * 
	 */
	abstract class TheoricCaseProblem implements IProblemDefinition {

		/**
		 * 
		 * Size (number of variables and number of states) of this best case problem
		 * 
		 */
		final int problemSize;

		/**
		 * 
		 * Actions available for this agent for this problem
		 * 
		 */
		final List<String> actionList = new ArrayList<String>();

		/**
		 * 
		 * Constructor
		 * 
		 * @param problemSize
		 *          number of variables/actions for this problem
		 * 
		 */
		public TheoricCaseProblem(int problemSize) {
			super();
			this.problemSize = problemSize;
			for (int actionIndex = 0; actionIndex < problemSize; actionIndex++) {
				actionList.add(String.valueOf(actionIndex));
			}
		}

		/**
		 * @see animatlab.kodamat.core.IProblemDefinition#getInitialState(Random)
		 */
		public Context getInitialState(Random random) {
			Context initialState;
			do {
				initialState = new Context();
				for (int variableIndex = 0; variableIndex < problemSize; variableIndex++) {
					initialState.put(variableOrdering.get(variableIndex), String
							.valueOf(random.nextBoolean()));
				}
			} while (isTerminal(initialState));
			return initialState;
		}

		/**
		 * @see animatlab.kodamat.core.IProblemDefinition#getNextState(Random,
		 *      Context, Context, java.lang.String)
		 */
		public List<Double> getNextState(Random random, Context currentState,
				Context nextState, String action) {
			nextState.clear();
			if (currentState == null) {
				nextState.putAll(getInitialState(random));
				return null;
			}
			if (isTerminal(currentState)) {
				return Context.rewardToList(SyntheticMDPsProblem.REWARD_VALUE);
			}

			int actionIndex = Integer.parseInt(action);
			getNextState(currentState, nextState, actionIndex);
			return Context.rewardToList(0.0);
		}

		/**
		 * 
		 * Compute the next state from the current state and an action
		 * 
		 * @param currentState
		 *          the current state of the agent
		 * @param nextState
		 *          the next state filled by the method
		 * @param action
		 *          the action executed by the agent
		 * 
		 */
		abstract protected void getNextState(Context currentState,
				Context nextState, int action);

		/**
		 * @see animatlab.kodamat.core.IModel#getActionList()
		 */
		public List<String> getActionList() {
			return actionList;
		}

		/**
		 * 
		 * Check if a subset of state is terminal
		 * 
		 * @param state
		 *          the state to check
		 * @param end
		 *          ending index defining the subset
		 * @return true if the state is terminal
		 * 
		 */
		boolean isTerminal(Context state, int end) {
			Context buffer = new Context();
			for (int variableIndex = 0; variableIndex < end; variableIndex++) {
				Variable variable = Variable.getVariable(String.valueOf(variableIndex));
				buffer.put(variable, state.get(variable));
			}
			return isTerminal(buffer);
		}

		/**
		 * @see animatlab.kodamat.core.IProblemDefinition#isTerminal(Context)
		 */
		public boolean isTerminal(Context state) {
			boolean result = true;
			for (String variableValue : state.values()) {
				result = result && variableValue.equals("true");
			}
			return result;
		}

		/**
		 * 
		 * @return the initial state of this problem
		 * 
		 */
		public ArrayList<String> getInitialState() {
			ArrayList<String> terminalState = new ArrayList<String>(problemSize);
			for (int variableIndex = 0; variableIndex < problemSize; variableIndex++) {
				terminalState.add("false");
			}
			return terminalState;
		}

		/**
		 * 
		 * @return a state to get a reward
		 * 
		 */
		public Context getRewardedState() {
			Context terminalState = new Context();
			for (int variableIndex = 0; variableIndex < problemSize; variableIndex++) {
				terminalState.put(variableOrdering.get(variableIndex), "true");
			}
			return terminalState;
		}

		/**
		 * @see animatlab.kodamat.core.IModel#getTransitionsModelSize()
		 */
		public long getTransitionsModelSize() {
			return getNumberOfStates() * getActionList().size();
		}

		/**
		 * @see animatlab.kodamat.core.IProblemDefinition#getNumberOfStates()
		 */
		public long getNumberOfStates() {
			return (long) Math.pow(2, problemSize);
		}

		/**
		 * @see animatlab.kodamat.core.IProblemDefinition#getStateSize()
		 */
		public int getStateSize() {
			return problemSize;
		}

		/**
		 * @see animatlab.kodamat.core.IModel#getRewardModelSize()
		 */
		public int getRewardModelSize() {
			assert (false);
			return 0;
		}

		/**
		 * @see animatlab.kodamat.core.IProblemDefinition#getVariableOrdering()
		 */
		public VariableOrdering getVariableOrdering() {
			return VariableOrdering.buildNumericVariableOrdering(getStateSize());
		}

		/**
		 * 
		 * @return the name of the problem
		 * 
		 */
		abstract protected String getProblemName();

		/**
		 * @see animatlab.kodamat.core.IModel#getName()
		 */
		public String getName() {
			return getProblemName() + getStateSize();
		}

		/**
		 * @see animatlab.kodamat.core.IProblemDefinition#getParameters()
		 */
		public Parameters getParameters() {
			return null;
		}
	}

	/**
	 * 
	 * Best case problem
	 * 
	 */
	class BestCaseProblem extends TheoricCaseProblem {
		/**
		 * 
		 * Constructor
		 * 
		 * @param problemSize
		 *          size of the problem
		 * 
		 */
		public BestCaseProblem(int problemSize) {
			super(problemSize);
		}

		/**
		 * @see animatlab.kodamat.problems.generated.syntheticmdps.tests.TestSyntheticMdps.TheoricCaseProblem#getNextState(Context,
		 *      Context, int)
		 */
		@Override
		protected void getNextState(Context currentState, Context nextState,
				int action) {
			nextState.putAll(currentState);
			nextState.put(Variable.getVariable(String.valueOf(action)), String
					.valueOf(isTerminal(currentState, action)));
			for (int variableIndex = action + 1; variableIndex < problemSize; variableIndex++) {
				nextState.put(variableOrdering.get(variableIndex), "false");
			}
		}

		/**
		 * @see animatlab.kodamat.problems.generated.syntheticmdps.tests.TestSyntheticMdps.TheoricCaseProblem#getProblemName()
		 */
		@Override
		protected String getProblemName() {
			return BESTCASE_PROBLEM_NAME;
		}

		/**
		 * @see animatlab.kodamat.core.IModel#getRewardMax()
		 */
		public double getRewardMax() {
			assert (false);
			return 0;
		}

		/**
		 * @see animatlab.kodamat.core.IModel#getValueSet(animatlab.kodamat.core.utils.Variable)
		 */
		public Set<String> getValueSet(Variable variable) {
			assert (false);
			return null;
		}

		/**
		 * @see animatlab.kodamat.core.IModel#getVariableSet()
		 */
		public Set<Variable> getVariableSet() {
			assert (false);
			return null;
		}

		/**
		 * @see animatlab.kodamat.core.IModel#getValueSets()
		 */
		public Map<Variable, Set<String>> getValueSets() {
			assert (false);
			return null;
		}

		/**
		 * @see animatlab.kodamat.core.IProblemDefinition#isStochastic()
		 */
		public boolean isStochastic() {
			assert (false);
			return false;
		}
	}

	/**
	 * 
	 * Worst case problem
	 * 
	 */
	class WorstCaseProblem extends TheoricCaseProblem {
		/**
		 * 
		 * Constructor
		 * 
		 * @param problemSize
		 *          size of the problem
		 * 
		 */
		public WorstCaseProblem(int problemSize) {
			super(problemSize);
		}

		/**
		 * @see animatlab.kodamat.problems.generated.syntheticmdps.tests.TestSyntheticMdps.TheoricCaseProblem#getNextState(Context,
		 *      Context, int)
		 */
		@Override
		protected void getNextState(Context currentState, Context nextState,
				int action) {
			nextState.putAll(currentState);
			nextState.put(Variable.getVariable(String.valueOf(action)), String
					.valueOf(isTerminal(currentState, action)));
			for (int variableIndex = 0; variableIndex < action; variableIndex++) {
				nextState.put(variableOrdering.get(variableIndex), "false");
			}
		}

		/**
		 * @see animatlab.kodamat.problems.generated.syntheticmdps.tests.TestSyntheticMdps.TheoricCaseProblem#getProblemName()
		 */
		@Override
		protected String getProblemName() {
			return WORSTCASE_PROBLEM_NAME;
		}

		/**
		 * @see animatlab.kodamat.core.IModel#getRewardMax()
		 */
		public double getRewardMax() {
			assert (false);
			return 0;
		}

		/**
		 * @see animatlab.kodamat.core.IModel#getValueSet(animatlab.kodamat.core.utils.Variable)
		 */
		public Set<String> getValueSet(Variable variable) {
			assert (false);
			return null;
		}

		/**
		 * @see animatlab.kodamat.core.IModel#getVariableSet()
		 */
		public Set<Variable> getVariableSet() {
			assert (false);
			return null;
		}

		/**
		 * @see animatlab.kodamat.core.IModel#getValueSets()
		 */
		public Map<Variable, Set<String>> getValueSets() {
			assert (false);
			return null;
		}

		/**
		 * @see animatlab.kodamat.core.IProblemDefinition#isStochastic()
		 */
		public boolean isStochastic() {
			assert (false);
			return false;
		}
	}

	/**
	 * 
	 * Test problem description parameters (size and noise) parsing
	 * 
	 */
	public void testProblemDescriptionParsing() {
		ForestGeneration forestGeneration = ForestGeneration
				.createProblemGeneration(BESTCASE_PROBLEM_NAME);
		assertEquals(0, forestGeneration.getProblemSize());
		assertEquals(0.0, forestGeneration.getNoise());

		forestGeneration = ForestGeneration
				.createProblemGeneration(BESTCASE_PROBLEM_NAME + "noise20");
		assertEquals(0, forestGeneration.getProblemSize());
		assertEquals(0.2, forestGeneration.getNoise());

		forestGeneration = ForestGeneration
				.createProblemGeneration(BESTCASE_PROBLEM_NAME + "20");
		assertEquals(20, forestGeneration.getProblemSize());
		assertEquals(0.0, forestGeneration.getNoise());

		forestGeneration = ForestGeneration
				.createProblemGeneration(BESTCASE_PROBLEM_NAME + "20noise20");
		assertEquals(20, forestGeneration.getProblemSize());
		assertEquals(0.2, forestGeneration.getNoise());
	}

	/**
	 * 
	 * Test a determinist problem
	 * 
	 * @param problem
	 *          the class generating the model of the problem in a xml file
	 * @param problemReference
	 *          instance used as reference to test the problem generated in xml
	 * 
	 */
	private void testDeterministTheoricProblem(ForestGeneration problem,
			IProblemDefinition problemReference) {
		problem.setProblemData(PROBLEM_SIZE);
		StaticForest spiForest = problem.toForest();
		testActionSet(problemReference, spiForest);
		testReward(problemReference, spiForest);
		testStateTransitionAndReward(problemReference, spiForest);
		// testInitialState(problemReference, spiForest);
	}

	/**
	 * 
	 * Test the initial state
	 * 
	 * @param problem
	 *          the problem used as reference
	 * @param spiForest
	 *          the forest to test
	 * 
	 */
	protected void testInitialState(IProblemDefinition problem,
			IProblemDefinition spiForest) {
		assertEquals(((TheoricCaseProblem) problem).getInitialState(), spiForest
				.getInitialState(new Random(0)));
	}

	/**
	 * 
	 * Test only the reward of the spi forest
	 * 
	 * @param problem
	 *          the problem used as reference
	 * @param spiForest
	 *          the forest to test
	 * 
	 */
	private void testReward(IProblemDefinition problem,
			IProblemDefinition spiForest) {
		TheoricCaseProblem bestCaseProblem = (TheoricCaseProblem) problem;
		Context currentState = bestCaseProblem.getRewardedState();
		Context nextStateProblem = new Context();
		Context nextStateSpiForest = new Context();
		String firstAction = problem.getActionList().iterator().next();
		assertEquals(problem.getNextState(random, currentState, nextStateProblem,
				firstAction), spiForest.getNextState(random, currentState,
				nextStateSpiForest, firstAction));
		for (int trialIndex = 0; trialIndex < NUMBER_OF_TRY; trialIndex++) {
			currentState = problem.getInitialState(random);
			assertEquals(problem.getNextState(random, currentState, nextStateProblem,
					firstAction), spiForest.getNextState(random, currentState,
					nextStateSpiForest, firstAction));
		}
	}

	/**
	 * 
	 * Test the get actions method of the problem
	 * 
	 * @param problem
	 *          the problem used to validate the spiForest
	 * @param spiForest
	 *          the problem to test
	 * 
	 */
	private void testActionSet(IProblemDefinition problem,
			IProblemDefinition spiForest) {
		assertEquals(problem.getActionList(), spiForest.getActionList());
	}

	/**
	 * 
	 * Test the state transition and the reward
	 * 
	 * @param problem
	 *          the problem used as reference
	 * @param spiForest
	 *          the forest tested
	 * 
	 */
	private void testStateTransitionAndReward(IProblemDefinition problem,
			IProblemDefinition spiForest) {
		List<String> actionList = new ArrayList<String>(problem.getActionList());
		for (int trialIndex = 0; trialIndex < NUMBER_OF_TRY; trialIndex++) {
			Context currentState = null;
			for (int actionTrialIndex = 0; actionTrialIndex < NUMBER_OF_TRY; actionTrialIndex++) {
				if ((currentState == null) || (currentState.size() == 0)) {
					currentState = problem.getInitialState(random);
				}
				Context nextStateProblem = new Context();
				Context nextStateSpiForest = new Context();
				String action = actionList.get(random.nextInt(actionList.size()));
				assertEquals(problem.getNextState(random, currentState,
						nextStateProblem, action), spiForest.getNextState(random,
						currentState, nextStateSpiForest, action));
				assertEquals(nextStateProblem, nextStateSpiForest);
				currentState = nextStateProblem;
			}
		}
	}

	/**
	 * 
	 * Test the worst case problem with some noise in it
	 * 
	 */
	public void testStochasticWorstCase() {
		StaticForest spiForest = ForestGeneration.toForest(WORSTCASE_PROBLEM_NAME
				+ PROBLEM_SIZE + ForestGeneration.NOISE_STRING + NOISE_LEVEL);
		for (int actionIndex = 0; actionIndex < PROBLEM_SIZE; actionIndex++) {
			for (int variableIndex = 0; variableIndex <= actionIndex; variableIndex++) {
				Tree<IProbabilities<String>> cptTree = spiForest.getTransitionFunction(
						ForestGeneration.getVariable(variableIndex), String
								.valueOf(actionIndex));

				Context constraints = new Context();
				for (int valueIndex = 0; valueIndex < actionIndex; valueIndex++) {
					constraints.put(variableOrdering.get(valueIndex), "true");
				}
				IProbabilities<String> probabilityNode = cptTree.getLeaf(constraints)
						.get();
				if (variableIndex < actionIndex) {
					assertEquals(1.0, probabilityNode.getProbability("false"));
				} else {
					assertEquals(1.0 - NOISE_LEVEL, probabilityNode
							.getProbability("true"));
					assertEquals(NOISE_LEVEL, probabilityNode.getProbability("false"));
				}
			}
		}
	}

	/**
	 * 
	 * Test the worst case problem
	 * 
	 */
	public void testWorstCase() {
		testDeterministTheoricProblem(new ExponProblem(), new WorstCaseProblem(
				PROBLEM_SIZE));
	}

	/**
	 * 
	 * Test the best case problem
	 * 
	 */
	public void testBestCase() {
		testDeterministTheoricProblem(new LinearProblem(), new BestCaseProblem(
				PROBLEM_SIZE));
	}

	/**
	 * 
	 * Test the best case problem with some noise in it
	 * 
	 */
	public void testStochasticBestCase() {
		StaticForest spiForest = ForestGeneration.toForest(BESTCASE_PROBLEM_NAME
				+ PROBLEM_SIZE + ForestGeneration.NOISE_STRING + NOISE_LEVEL);
		for (int sizeIndex = 0; sizeIndex < PROBLEM_SIZE; sizeIndex++) {
			for (int variableIndex = 0; variableIndex < sizeIndex; variableIndex++) {
				final Variable variable = ForestGeneration.getVariable(variableIndex);
				Tree<IProbabilities<String>> cptTree = spiForest.getTransitionFunction(
						variable, String.valueOf(sizeIndex));

				Context constraints = new Context();
				IProbabilities<String> probabilityNode;

				constraints.put(variable, "false");
				probabilityNode = cptTree.getLeaf(constraints).get();
				assertEquals(1.0, probabilityNode.getProbability("false"));

				constraints.put(variable, "true");
				probabilityNode = cptTree.getLeaf(constraints).get();
				if (variableIndex == (sizeIndex - 1)) {
					assertEquals(1.0 - NOISE_LEVEL, probabilityNode
							.getProbability("true"));
					assertEquals(NOISE_LEVEL, probabilityNode.getProbability("false"));
				} else {
					assertEquals(1.0, probabilityNode.getProbability("true"));
				}
			}
		}
	}

	/**
	 * 
	 * Test the noisy case problem with some noise in it
	 * 
	 */
	public void testStochasticNoisyCase() {
		StaticForest spiForest = ForestGeneration.toForest(NOISYCASE_PROBLEM_NAME
				+ PROBLEM_SIZE + ForestGeneration.NOISE_STRING + NOISE_LEVEL);
		for (int sizeIndex = 0; sizeIndex < PROBLEM_SIZE; sizeIndex++) {
			for (int variableIndex = 0; variableIndex < sizeIndex; variableIndex++) {
				final Variable variable = ForestGeneration.getVariable(variableIndex);
				Tree<IProbabilities<String>> cptTree = spiForest.getTransitionFunction(
						variable, String.valueOf(sizeIndex));
				Context constraints = new Context();
				IProbabilities<String> probabilityNode;
				constraints.put(variable, "false");
				probabilityNode = cptTree.getLeaf(constraints).get();
				assertEquals(1.0 - NOISE_LEVEL, probabilityNode.getProbability("false"));
				assertEquals(NOISE_LEVEL, probabilityNode.getProbability("true"));

				constraints.put(variable, "true");
				probabilityNode = cptTree.getLeaf(constraints).get();
				assertEquals(1.0 - NOISE_LEVEL, probabilityNode.getProbability("true"));
				assertEquals(NOISE_LEVEL, probabilityNode.getProbability("false"));
			}
		}
	}

	/**
	 * 
	 * Test the best case problem
	 * 
	 */
	public void testBestCaseValueFunction() {
		StaticForest spiForest = ForestGeneration.toForest(BESTCASE_PROBLEM_NAME
				+ PROBLEM_SIZE);
		SpiSvi spisvi = new SpiSvi(spiForest);
		spisvi.getSVI();
		Tree<Double> optimalValueFunction = spisvi.getOptimalValueFunction();
		assertEquals(PROBLEM_SIZE + 1, optimalValueFunction.numberOfLeaves());
	}

	/**
	 * 
	 * Test the best case problem
	 * 
	 */
	public void testWorstCaseValueFunction() {
		final int problemSize = 9;
		StaticForest spiForest = ForestGeneration.toForest(WORSTCASE_PROBLEM_NAME
				+ problemSize + "noise20");
		SpiSvi spisvi = new SpiSvi(spiForest);
		spisvi.solve();
		Tree<Double> optimalValueFunction = spisvi.getOptimalValueFunction();
		assertEquals((int) Math.pow(2, problemSize), optimalValueFunction
				.numberOfLeaves());
	}

	/**
	 * 
	 * Test if the stochasticity of a problem is correctly computed
	 * 
	 */
	public void testStochasticity() {
		StaticForest spiForest = ForestGeneration.toForest(BESTCASE_PROBLEM_NAME
				+ PROBLEM_SIZE);
		assertEquals(false, spiForest.isStochastic());
		spiForest = ForestGeneration.toForest(BESTCASE_PROBLEM_NAME + PROBLEM_SIZE
				+ ForestGeneration.NOISE_STRING + NOISE_LEVEL);
		assertEquals(true, spiForest.isStochastic());
	}
}
