package animatlab.kodamat.problems.generated.sysadmin;

import java.util.List;
import java.util.Map;

import animatlab.kodamat.core.functions.FunctionsSum;
import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.functions.UnaryOperatorPolicy;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.DNode;
import animatlab.kodamat.core.trees.EqualPolicy;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.MergePolicy;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.problems.generated.ForestGeneration;
import animatlab.kodamat.reinforcementlearning.model.StaticCPD;
import animatlab.kodamat.reinforcementlearning.model.forest.StaticForest;
import animatlab.kodamat.reinforcementlearning.planning.guestrin.BasisFunctionUtils;
import animatlab.kodamat.reinforcementlearning.planning.guestrin.StaticGuestrin;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.SpiSvi;

/**
 * 
 * Abstract class to generate Sys admin problem
 * 
 * @author Thomas Degris
 * 
 */
public abstract class SysAdminProblem extends ForestGeneration {
	/**
	 * 
	 * Decompose basis function suffix
	 * 
	 */
	private static final String DECBF_SUFFIX = "decbf";

	/**
	 * 
	 * Size of the scope of a basis function
	 * 
	 */
	public static final int BASIS_FUNCTION_SCOPE_SIZE = 2;

	/**
	 * 
	 * Decompose the set of basis function
	 * 
	 */
	private boolean decomposeBasisFunction = false;

	/**
	 * 
	 * Constructor
	 * 
	 * @param problemName
	 *          name of the problem
	 * 
	 */
	protected SysAdminProblem(String problemName) {
		super(problemName);
	}

	/**
	 * @see animatlab.kodamat.problems.generated.ForestGeneration#buildParameters(animatlab.kodamat.reinforcementlearning.model.forest.StaticForest)
	 */
	@Override
	protected void buildParameters(StaticForest forest) {
		forest.getParameters().setKeyValue(Parameters.PARAM_KEY_PLANNING_METHOD,
				StaticGuestrin.PLANNING_METHOD_NAME);
		forest.getParameters().setKeyValue(
				Parameters.PARAM_KEY_DECOMPOSE_BASISFUNCTION,
				String.valueOf(decomposeBasisFunction));
		forest.getParameters().setKeyValue(
				Parameters.PARAM_KEY_BASIS_FUNCTION_SCOPE_SIZE,
				String.valueOf(BASIS_FUNCTION_SCOPE_SIZE));
	}

	/**
	 * @see animatlab.kodamat.problems.generated.ForestGeneration#getBasisFunction(animatlab.kodamat.reinforcementlearning.model.forest.StaticForest)
	 */
	@Override
	protected FunctionsSum getBasisFunction(StaticForest forest) {
		if (!decomposeBasisFunction) {
			return computeTreeBasisFunction(forest);
		}
		return BasisFunctionUtils.generateBasisFunction(forest,
				BasisFunctionUtils.iterationPolicy01, BASIS_FUNCTION_SCOPE_SIZE - 1);
	}

	/**
	 * @see animatlab.kodamat.problems.generated.ForestGeneration#buildRewardFunctions(animatlab.kodamat.reinforcementlearning.model.forest.StaticForest)
	 */
	@Override
	protected void buildRewardFunctions(StaticForest forest) {
		for (int index = 0; index < problemSize; index++) {
			Tree<Double> tree = new Tree<Double>();
			DNode<Double> dnode = tree.getRoot().replace(getVariable(index));
			double rewardValue = 1.0;
			if (index == problemSize - 1) {
				rewardValue = 2.0;
			}
			dnode.addLeaf("W", rewardValue);
			dnode.addLeaf("D", 0.0);
			forest.addReward(tree);
		}
	}

	/**
	 * @see animatlab.kodamat.problems.generated.ForestGeneration#extractActionName(int)
	 */
	@Override
	public String extractActionName(int actionIndex) {
		return "Reboot" + super.extractActionName(actionIndex);
	}

	/**
	 * @see animatlab.kodamat.problems.generated.ForestGeneration#getActionList()
	 */
	@Override
	protected List<String> getActionList() {
		List<String> result = super.getActionList();
		result.add("DoNothing");
		return result;
	}

	/**
	 * 
	 * Add a new rule to the rule set
	 * 
	 * @param tree
	 *          tree where to set the proba
	 * @param stringContext
	 *          the context pointing to the leaf that needs to be set
	 * @param workingProbability
	 *          probability that this machine will work at next state
	 * @param variables
	 *          variables defining the context
	 * 
	 */
	protected void setProba(Tree<IProbabilities<String>> tree,
			String stringContext, double workingProbability, Variable... variables) {
		Context context = new Context();
		String[] values = stringContext.split(",");
		assert (values.length == variables.length);
		for (int index = 0; index < values.length; index++) {
			context.put(variables[index], values[index]);
		}
		StaticCPD proba = new StaticCPD();
		if (workingProbability > 0) {
			proba.setProba("W", workingProbability);
		}
		if ((1 - workingProbability) > 0) {
			proba.setProba("D", 1 - workingProbability);
		}
		Leaf<IProbabilities<String>> leaf = tree.getLeaf(context);
		leaf.set(proba);
	}

	/**
	 * 
	 * Compute a set of tree to use as basis functions
	 * 
	 * @param forest
	 *          the forest to use as basis function
	 * 
	 * @return the sum of basis function
	 * 
	 */
	protected FunctionsSum computeTreeBasisFunction(StaticForest forest) {
		FunctionsSum basisFunctionsSet = new FunctionsSum();
		Tree<Double> constantBasisFunction = new Tree<Double>();
		((Leaf<Double>) constantBasisFunction.getRoot()).set(1.0);
		basisFunctionsSet.add(constantBasisFunction);
		SpiSvi spisvi = new SpiSvi(forest);
		for (IFunction<Double> reward : forest.getRewardFunction(null).functions()) {
			basisFunctionsSet.add(computeValueFunction(spisvi, (Tree<Double>) reward
					.clone(), BASIS_FUNCTION_SCOPE_SIZE));
		}
		return basisFunctionsSet;
	}

	/**
	 * 
	 * Compute a new basis function
	 * 
	 * @param spisvi
	 *          spisvi instance to compute the regress operator
	 * @param treeH
	 *          the basis function to compute
	 * @return a new basis function
	 * 
	 */
	@SuppressWarnings("null")
	private Tree<Double> basisFunctionIteration(SpiSvi spisvi, Tree<Double> treeH) {
		Tree<Double> treeVh = null;
		for (String action : getActionList()) {
			Tree<Map<Variable, IProbabilities<String>>> treePQaH = spisvi
					.getPRegress(treeH.getRoot(), action, null);
			Tree<Double> treeQaH = spisvi.constructFVTree(action, treePQaH, treeH);
			UnaryOperatorPolicy.discountFactor.applyPolicy(treeQaH);
			treeVh = MergePolicy.minPolicy.merge(treeQaH, treeVh);
		}
		treeVh.simplify(EqualPolicy.valueLeafEqualPolicy);
		return treeVh;
	}

	/**
	 * 
	 * Compute a value function until a specified scope size (or until the scope
	 * size does not move)
	 * 
	 * @param spisvi
	 *          the spisvi instance used to develop the value function
	 * @param valueFunction
	 *          value function to start to iterate with
	 * @param maxScopeSize
	 *          maximum size of the scope
	 * @return the value function computed
	 * 
	 */
	private Tree<Double> computeValueFunction(SpiSvi spisvi,
			Tree<Double> valueFunction, int maxScopeSize) {
		Tree<Double> result = valueFunction;
		int currentScopeSize = valueFunction.scope().size();
		if (currentScopeSize >= maxScopeSize) {
			return result;
		}
		int lastScopeSize = -1;
		while ((currentScopeSize != lastScopeSize)
				&& (currentScopeSize < maxScopeSize)) {
			lastScopeSize = currentScopeSize;
			result = basisFunctionIteration(spisvi, result);
			currentScopeSize = result.scope().size();
		}
		return result;
	}

	/**
	 * @see animatlab.kodamat.problems.generated.ForestGeneration#setOptions(java.lang.String)
	 */
	@Override
	protected void setOptions(String problemDescription) {
		super.setOptions(problemDescription);
		if (problemDescription.endsWith(DECBF_SUFFIX)) {
			decomposeBasisFunction = true;
		}
	}

	/**
	 * @see animatlab.kodamat.problems.generated.ForestGeneration#getProblemFileName()
	 */
	@Override
	public String getProblemFileName() {
		String result = getProblemDescription();
		if (decomposeBasisFunction) {
			result = result + DECBF_SUFFIX;
		}
		return result + StaticForest.PROBLEM_EXTENSION;
	}

}
