package animatlab.kodamat.problems.generated.sysadmin.tests;

import java.util.Random;

import junit.framework.TestCase;
import animatlab.kodamat.core.functions.FunctionsSum;
import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.functions.IPartition;
import animatlab.kodamat.core.functions.IPartitionFunction;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.problems.generated.ForestGeneration;
import animatlab.kodamat.problems.generated.sysadmin.Disconnected;
import animatlab.kodamat.problems.generated.sysadmin.FullyConnected;
import animatlab.kodamat.problems.generated.sysadmin.SysAdminProblem;
import animatlab.kodamat.problems.generated.sysadmin.UnidirectionalRing;
import animatlab.kodamat.reinforcementlearning.model.forest.StaticForest;
import animatlab.kodamat.reinforcementlearning.planning.guestrin.StaticGuestrin;

/**
 * 
 * Test case to test sys admin generated problems
 * 
 * @author Thomas Degris
 * 
 */
public class TestSysAdminGeneratedProblem extends TestCase {
	/**
	 * 
	 * Size of the problem used for the testing suite
	 * 
	 */
	final static int PROBLEM_SIZE = 5;

	/**
	 * 
	 * Number of random state to try
	 * 
	 */
	final static int NUMBER_OF_TRY = 500;

	/**
	 * 
	 * Random series
	 * 
	 */
	final Random random = new Random(0);

	/**
	 * 
	 * Get a rules model from a generated problem
	 * 
	 * @param problem
	 *          the problem to generate
	 * 
	 * @return the new model
	 * 
	 */
	private StaticForest generateProblem(ForestGeneration problem) {
		problem.setProblemData(PROBLEM_SIZE);
		return problem.toForest();
	}

	/**
	 * 
	 * Test the unidirectional topology
	 * 
	 */
	public void testUnidirectionalRing() {
		StaticForest model = generateProblem(new UnidirectionalRing());
		assertEquals(model.getParameters().getKeyS(
				Parameters.PARAM_KEY_PLANNING_METHOD),
				StaticGuestrin.PLANNING_METHOD_NAME);
		testRewardFunction(model);
		testActionTransitions(model);
		for (IFunction<Double> basisFunction : model.getBasisFunctions()
				.functions()) {
			IPartitionFunction<Double> fun = (IPartitionFunction<Double>) basisFunction;
			if (fun.scope().size() > 0) {
				assertEquals(SysAdminProblem.BASIS_FUNCTION_SCOPE_SIZE, fun.scope()
						.size());
			}
		}
	}

	/**
	 * 
	 * Test the disconnected topology
	 * 
	 */
	public void testDisconnected() {
		StaticForest model = generateProblem(new Disconnected());
		assertEquals(model.getParameters().getKeyS(
				Parameters.PARAM_KEY_PLANNING_METHOD),
				StaticGuestrin.PLANNING_METHOD_NAME);
		testRewardFunction(model);
		testActionTransitions(model);
		testBasisFunction(model);
	}

	/**
	 * 
	 * Test general properties of transition functions
	 * 
	 * @param model
	 *          the model of rules to check
	 * 
	 */
	private void testActionTransitions(StaticForest model) {
		assertEquals(PROBLEM_SIZE + 1, model.getActionList().size());

		Context state = model.getInitialState(random);
		int workingMachine = PROBLEM_SIZE + 1;
		for (int trial = 0; trial < PROBLEM_SIZE * 2; trial++) {
			Context nextState = new Context();
			model.getNextState(random, state, nextState, "DoNothing");
			state = nextState;
			workingMachine = Math.min(workingMachine,
					getNumberOfWorkingMachine(state));
		}
		assert (workingMachine <= (PROBLEM_SIZE - (PROBLEM_SIZE * 0.95)));

		for (int trial = 0; trial < NUMBER_OF_TRY; trial++) {
			int machine = random.nextInt(PROBLEM_SIZE);
			Context nextState = new Context();
			model.getNextState(random, state, nextState, "Reboot"
					+ String.valueOf(machine));
			state = nextState;
			assertEquals("W", state.get(model.getVariableOrdering().get(machine)));
		}

	}

	/**
	 * 
	 * Check the reward function of a sys admin problem
	 * 
	 * @param model
	 * 
	 */
	private void testRewardFunction(StaticForest model) {
		for (int trial = 0; trial < NUMBER_OF_TRY; trial++) {
			Context state = model.getInitialState(random);
			double workingCount = getNumberOfWorkingMachine(state);
			if (state.get(model.getVariableOrdering().get(PROBLEM_SIZE - 1)).equals(
					"W")) {
				workingCount++;
			}
			assertEquals(workingCount, model.getRewardFunction(null).classify(state));
		}
	}

	/**
	 * 
	 * Check the set of basis function loaded
	 * 
	 * @param model
	 *          the model to test
	 * 
	 */
	private void testBasisFunction(StaticForest model) {
		FunctionsSum basisFunctions = model.getBasisFunctions();
		assertEquals(model.getStateSize() + 1, basisFunctions.size());
	}

	/**
	 * 
	 * Count the number of working machine in a state
	 * 
	 * @param state
	 *          the state where to count the number of working machine
	 * 
	 * @return the number of working machine in the state
	 * 
	 */
	private int getNumberOfWorkingMachine(Context state) {
		int workingCount = 0;
		for (String value : state.values()) {
			if (value.equals("W")) {
				workingCount++;
			}
		}
		return workingCount;
	}

	/**
	 * 
	 * Test the fully connected topology
	 * 
	 */
	public void testFullyConnected() {
		FullyConnected problem = new FullyConnected();
		StaticForest model = generateProblem(problem);
		assertEquals(model.getParameters().getKeyS(
				Parameters.PARAM_KEY_PLANNING_METHOD),
				StaticGuestrin.PLANNING_METHOD_NAME);
		testRewardFunction(model);
		for (Variable variable : model.getVariableSet()) {
			for (String action : model.getActionList()) {
				Tree<IProbabilities<String>> tree = model.getTransitionFunction(
						variable, action);
				if (action.equals(problem.extractActionName(variable.variable))) {
					assertEquals(0, tree.scope().size());
					assertEquals(1.0, ((Leaf<IProbabilities<String>>) tree.getRoot())
							.get().getProbability("W"));
				} else {
					assertEquals(PROBLEM_SIZE, tree.scope().size());
					for (IPartition<IProbabilities<String>> partition : tree.partitions()) {
						boolean working = true;
						for (String value : partition.getContext().values()) {
							working = working && value.equals("W");
						}
						if (!working) {
							assertEquals(0.9, partition.get().getProbability("D"));
						} else {
							assertEquals(0.9, partition.get().getProbability("W"));
						}
					}
				}
			}
		}
	}

}
