/*
 * 
 * Created on 11 mars 2005
 *
 */
package animatlab.kodamat.problems.mazespace;

import java.awt.Dimension;
import java.awt.Point;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import animatlab.kodamat.core.IProblemDefinition;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.core.utils.VariableOrdering;

/**
 * 
 * Model for maze
 * 
 * @author Thomas Degris
 * 
 */
public class MazeModel implements IProblemDefinition {
	/**
	 * 
	 * Maze defined by Sutton
	 * 
	 */
	static public final byte SUTTON_MAZE[][] = {
			{ 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
			{ 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
			{ 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 2, 1, 1 },
			{ 1, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 1 },
			{ 1, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 1 },
			{ 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1 },
			{ 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1 },
			{ 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1 },
			{ 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
			{ 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 } };

	/**
	 * 
	 * Local perception of the agent
	 * 
	 */
	static final public byte WOODS04[][] = { { -1, 0 }, { 0, +1 }, { +1, 0 },
			{ 0, -1 } };

	/**
	 * 
	 * Local perception of the agent with Sutton simple
	 * 
	 */
	static public final byte SUTTON_SIMPLE[][] = { { -1, -2 }, { -1, +1 },
			{ +1, 0 }, { +1, +1 }, { +2, +2 }, { 0, -2 }, { 0, +2 }, { +2, -2 },
			{ 0, -1 }, { -2, -2 }, { -2, 0 }, { -1, 0 } };

	/**
	 * 
	 * Local perception of the agent with Sutton simple
	 * 
	 */
	static final public byte SUTTON_WOODS25[][] = { { -2, -2 }, { -2, -1 },
			{ -2, 0 }, { -2, +1 }, { -2, +2 }, { -1, -2 }, { -1, -1 }, { -1, 0 },
			{ -1, +1 }, { -1, +2 }, { 0, -2 }, { 0, -1 }, { 0, 0 }, { 0, +1 },
			{ 0, +2 }, { +1, -2 }, { +1, -1 }, { +1, 0 }, { +1, +1 }, { +1, +2 },
			{ +2, -2 }, { +2, -1 }, { +2, 0 }, { +2, +1 }, { +2, +2 } };

	/**
	 * 
	 * Action available for the agent
	 * 
	 */
	static final public byte ACTIONS[][] = { { -1, 0 }, { 0, +1 }, { +1, 0 },
			{ 0, -1 } };

	/**
	 * 
	 * Value of the reward in the maze
	 * 
	 */
	static public final double REWARD = 1.0;

	/**
	 * 
	 * Value of a cell when it is empty
	 * 
	 */
	public static final byte EMPTY_VALUE = 0;

	/**
	 * 
	 * Value of the reward in the maze
	 * 
	 */
	public static final byte WALL_VALUE = 1;

	/**
	 * 
	 * Value of the reward in the maze
	 * 
	 */
	public static final byte REWARD_VALUE = 2;

	/**
	 * 
	 * Value of the reward in the maze
	 * 
	 */
	public static final byte AGENT_VALUE = 3;

	/**
	 * 
	 * Maze where the agent navigates
	 * 
	 */
	final private byte[][] maze;

	/**
	 * 
	 * Perception of the agent
	 * 
	 */
	final private byte[][] perception;

	/**
	 * 
	 * The dimension of the maze
	 * 
	 */
	final private Dimension mazeDimension;

	/**
	 * 
	 * Available actions for the agent
	 * 
	 */
	static final protected String[] actions = { "0", "1", "2", "3" };

	/**
	 * 
	 * View of the maze to display debugging data
	 * 
	 */
	MazeView mazeView = null;

	/**
	 * 
	 * Starting position of the agent
	 * 
	 */
	final Point startingPosition = new Point(2, 4);

	/**
	 * 
	 * Map from state of the agent to a position in the maze
	 * 
	 */
	final Map<Context, Point> stateToPosition = new LinkedHashMap<Context, Point>();

	/**
	 * 
	 * Variable ordering for this problem
	 * 
	 */
	private VariableOrdering variableOrdering;

	/**
	 * 
	 * Constructor
	 * 
	 * @param maze
	 *          maze where the agent navigates
	 * @param perception
	 *          perception of the agent
	 * 
	 */
	public MazeModel(byte[][] maze, byte[][] perception) {
		this.maze = maze;
		this.perception = perception;
		if (perception != null) {
			variableOrdering = VariableOrdering
					.buildNumericVariableOrdering(perception.length);
		}
		mazeDimension = new Dimension(maze[0].length, maze.length);
	}

	/**
	 * 
	 * Get the state corresponding at a position
	 * 
	 * @param position
	 *          the position to get the state from
	 * @param state
	 *          the state to iniatialize
	 * 
	 */
	public void setState(Point position, Context state) {
		state.clear();
		for (int index = 0; index < perception.length; index++) {
			byte value = maze[position.y + perception[index][1]][position.x
					+ perception[index][0]];
			if (value == REWARD_VALUE) {
				value = 0;
			}
			state.put(variableOrdering.get(index), String.valueOf(value));
		}
	}

	/**
	 * 
	 * Initialize the map from state to position in the maze
	 * 
	 */
	private void initializeStateToPosition() {
		for (int x = 0; x < mazeDimension.width; x++) {
			for (int y = 0; y < mazeDimension.height; y++) {
				if (maze[y][x] != WALL_VALUE) {
					Point position = new Point(x, y);
					Context state = new Context();
					setState(position, state);
					assert (!stateToPosition.containsKey(state));
					stateToPosition.put(state, position);
				}
			}
		}
	}

	/**
	 * 
	 * Get the reward at a position
	 * 
	 * @param position
	 *          the position where to get the reward from
	 * @return the reward
	 * 
	 */
	double getCurrentReward(Point position) {
		if (maze[position.y][position.x] == 2) {
			return REWARD;
		}
		return 0.0;
	}

	/**
	 * 
	 * Update the position of an agent for an action
	 * 
	 * @param currentPosition
	 *          the position of the agent
	 * @param action
	 *          the action done
	 * 
	 * @return the new position of the agent
	 * 
	 */
	Point updatePosition(Point currentPosition, String action) {
		Point nextPosition = new Point(currentPosition);
		if (maze[nextPosition.y][nextPosition.x] == REWARD_VALUE) {
			return startingPosition;
		}
		byte[] move = ACTIONS[Integer.parseInt(action)];
		if (maze[nextPosition.y + move[1]][nextPosition.x + move[0]] == WALL_VALUE) {
			return nextPosition;
		}
		nextPosition.x += move[0];
		nextPosition.y += move[1];
		return nextPosition;
	}

	/**
	 * 
	 * @return the dimension of the maze
	 * 
	 */
	public Dimension getDimension() {
		return mazeDimension;
	}

	/**
	 * 
	 * Get the value at a position in the maze
	 * 
	 * @param position
	 *          the position to get the value from
	 * @return the value at the position
	 * 
	 */
	public int getValueAt(Point position) {
		return maze[position.y][position.x];
	}

	/**
	 * @see animatlab.kodamat.core.IProblemDefinition#getNumberOfStates()
	 */
	public long getNumberOfStates() {
		int numberOfStates = 0;
		for (int x = 0; x < mazeDimension.width; x++) {
			for (int y = 0; y < mazeDimension.height; y++) {
				if (maze[y][x] != 1) {
					numberOfStates++;
				}
			}
		}
		return numberOfStates;
	}

	/**
	 * @see animatlab.kodamat.core.IProblemDefinition#getInitialState(Random)
	 */
	public Context getInitialState(Random random) {
		Context state = new Context();
		setState(startingPosition, state);
		return state;
	}

	/**
	 * @see animatlab.kodamat.core.IProblemDefinition#getNextState(Random,
	 *      Context, Context, java.lang.String)
	 */
	public List<Double> getNextState(Random random, Context state,
			Context nextState, String action) {
		if (stateToPosition.size() == 0) {
			initializeStateToPosition();
		}
		if (state == null) {
			nextState.clear();
			nextState.putAll(getInitialState(random));
			return null;
		}
		Point currentPosition = stateToPosition.get(state);
		assert (getValueAt(currentPosition) != WALL_VALUE);
		if (isTerminal(state)) {
			nextState.clear();
			return Context.rewardToList(getCurrentReward(currentPosition));
		}
		nextState.clear();
		setState(updatePosition(currentPosition, action), nextState);
		return Context.rewardToList(getCurrentReward(currentPosition));
	}

	/**
	 * @see animatlab.kodamat.core.IProblemDefinition#getVariableOrdering()
	 */
	public VariableOrdering getVariableOrdering() {
		return variableOrdering;
	}

	/**
	 * @see animatlab.kodamat.core.IProblemDefinition#getActionList()
	 */
	public List<String> getActionList() {
		List<String> result = new ArrayList<String>();
		for (String action : actions) {
			result.add(action);
		}
		return result;
	}

	/**
	 * @see animatlab.kodamat.core.IProblemDefinition#getStateSize()
	 */
	public int getStateSize() {
		return perception.length;
	}

	/**
	 * @see animatlab.kodamat.core.IProblemDefinition#isTerminal(Context)
	 */
	public boolean isTerminal(Context state) {
		// return (getValueAt(stateToPosition.get(state)) == REWARD_VALUE);
		return false;
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getName()
	 */
	public String getName() {
		return getClass().getSimpleName();
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getTransitionsModelSize()
	 */
	public long getTransitionsModelSize() {
		return getNumberOfStates() * getActionList().size();
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getRewardModelSize()
	 */
	public int getRewardModelSize() {
		int numberOfReward = 0;
		for (int x = 0; x < mazeDimension.width; x++) {
			for (int y = 0; y < mazeDimension.height; y++) {
				if (maze[y][x] != REWARD_VALUE) {
					numberOfReward++;
				}
			}
		}
		return numberOfReward;
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getRewardMax()
	 */
	public double getRewardMax() {
		return REWARD;
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getValueSet(animatlab.kodamat.core.utils.Variable)
	 */
	public Set<String> getValueSet(Variable variable) {
		assert (false);
		return null;
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getVariableSet()
	 */
	public Set<Variable> getVariableSet() {
		assert (false);
		return null;
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getValueSets()
	 */
	public Map<Variable, Set<String>> getValueSets() {
		assert (false);
		return null;
	}

	/**
	 * @see animatlab.kodamat.core.IProblemDefinition#isStochastic()
	 */
	public boolean isStochastic() {
		return false;
	}

	/**
	 * @see animatlab.kodamat.core.IProblemDefinition#getParameters()
	 */
	public Parameters getParameters() {
		Parameters parameters = new Parameters();
		parameters.setKeyValue(Parameters.PARAM_KEY_RESET_REWARD, "false");
		return parameters;
	}
}
