/*
 * 
 * Created on 11 mars 2005
 *
 */
package animatlab.kodamat.problems.mazespace;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Point;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import javax.imageio.ImageIO;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.Timer;
import javax.swing.ToolTipManager;

import animatlab.kodamat.core.io.Log;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.problems.ProblemRunner;
import animatlab.kodamat.reinforcementlearning.agent.ModelBasedAgent;
import animatlab.kodamat.reinforcementlearning.core.acting.EpsilonGreedy;
import animatlab.kodamat.reinforcementlearning.core.acting.Greedy;
import animatlab.kodamat.reinforcementlearning.core.acting.IActing;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractForest;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractLearningForest;
import animatlab.kodamat.reinforcementlearning.model.forest.VariableActionLearningForest;
import animatlab.kodamat.reinforcementlearning.model.forest.VariableLearningForest;
import animatlab.kodamat.reinforcementlearning.model.tabular.TabularLearningModel;
import animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning;
import animatlab.kodamat.reinforcementlearning.planning.dyna.DynaQ;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.LearningSpiSvi;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.LearningTreeBasedPlanning;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.StaticSpiSvi;
import animatlab.kodamat.supervisedlearning.entropy.Chi2Measure;
import animatlab.kodamat.supervisedlearning.entropy.KSMeasure;
import animatlab.kodamat.supervisedlearning.learningtree.ItiTree;
import animatlab.kodamat.supervisedlearning.learningtree.SID4Tree;
import animatlab.kodamat.supervisedlearning.regression.LeastSquareMeasure;
import animatlab.kodamat.supervisedlearning.regression.StructureLearningMeasure;
import animatlab.kodamat.supervisedlearning.rmax.RMaxTree;

/**
 * 
 * View of a maze
 * 
 * @author Thomas Degris
 * 
 */
public class MazeView extends JFrame implements ActionListener {

	/**
	 * 
	 */
	private static final long serialVersionUID = -4350561885194864763L;

	/**
	 * 
	 * Preferred case view size
	 * 
	 */
	final static protected int CASE_SIZE = 40;

	/**
	 * 
	 * MazeAgent running in the maze
	 * 
	 */
	MazeAgent mazeAgent;

	/**
	 * 
	 * Policy for drawing things in the maze
	 * 
	 */
	public interface DrawingPolicy {
		/**
		 * 
		 * Draw a position in the maze
		 * 
		 * @param position
		 *          the position of the maze to draw
		 * @param graphics
		 *          the panel where to draw the maze
		 * 
		 */
		void draw(Point position, Graphics graphics);
	}

	/**
	 * 
	 * Panel used to draw thing for the cases
	 * 
	 */
	private class ElementView {

		/**
		 * 
		 * Position of the element in the grid
		 * 
		 */
		final protected Point position;

		/**
		 * 
		 * Constructor
		 * 
		 * @param position
		 *          position of the element in the grid
		 * 
		 */
		ElementView(Point position) {
			this.position = position;
		}

		/**
		 * Set the background of this view
		 * 
		 * @param graphics
		 *          graphics to draw the image
		 * @param color
		 *          color of the background
		 * 
		 */
		void setBackground(Graphics graphics, Color color) {
			graphics.setColor(color);
			graphics.fillRect(position.x * CASE_SIZE, position.y * CASE_SIZE,
					(position.x + 1) * CASE_SIZE, (position.y + 1) * CASE_SIZE);
		}

		/**
		 * Paint this view in the panel
		 * 
		 * @param graphics
		 *          graphics to paint on
		 */
		public void paint(Graphics graphics) {
			Graphics2D g2 = (Graphics2D) graphics;
			g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
					RenderingHints.VALUE_ANTIALIAS_ON);
			if (mazeModel.getValueAt(position) == MazeModel.WALL_VALUE) {
				setBackground(graphics, Color.BLACK);
				return;
			}
			setBackground(graphics, Color.WHITE);
			if (drawingPolicyRegistered.size() == 0) {
				setBackground(graphics, Color.WHITE);
				return;
			}
			graphics.setColor(Color.BLACK);
			graphics.setFont(new Font("Serif", Font.PLAIN, 12));
			for (DrawingPolicy drawingPolicy : drawingPolicyRegistered) {
				drawingPolicy.draw(position, graphics);
			}
		}
	}

	/**
	 * 
	 * Panel to draw the maze
	 * 
	 */
	private class MazePanel extends JPanel {

		/**
		 * 
		 */
		private static final long serialVersionUID = -5717207060895100822L;

		/**
		 * 
		 * Constructor
		 * 
		 */
		MazePanel() {
			addMouseMotionListener(new MouseMotionAdapter() {
				private Point currentMousePosition = new Point();

				private Context displayedState = new Context();

				@Override
				public void mouseMoved(MouseEvent mouseEvent) {
					currentMousePosition.x = mouseEvent.getX() / CASE_SIZE;
					currentMousePosition.y = mouseEvent.getY() / CASE_SIZE;
					if (mazeModel.getValueAt(currentMousePosition) != MazeModel.WALL_VALUE) {
						mazeModel.setState(currentMousePosition, displayedState);
						setToolTipText(displayedState.toString());
					} else {
						setToolTipText("");
					}
				}
			});
		}

		/**
		 * @see java.awt.Container#paint(java.awt.Graphics)
		 */
		@Override
		public void paint(Graphics graphics) {
			for (int x = 0; x < mazeModel.getDimension().width; x++) {
				for (int y = 0; y < mazeModel.getDimension().height; y++) {
					elementViewGrid[y][x].paint(graphics);
				}
			}
		}
	}

	/**
	 * 
	 * Exit button of the application
	 * 
	 */
	private JButton exitButton;

	/**
	 * 
	 * Panel used to see the maze
	 * 
	 */
	private MazePanel mazeView;

	/**
	 * 
	 * Grid of panel
	 * 
	 */
	ElementView[][] elementViewGrid;

	/**
	 * 
	 * Model of the maze drawed
	 * 
	 */
	MazeModel mazeModel;

	/**
	 * 
	 * Current drawing policy called by the paint method of the ElementView
	 * 
	 */
	List<DrawingPolicy> drawingPolicyRegistered = new ArrayList<DrawingPolicy>();

	/**
	 * 
	 * The agent to run in the maze
	 * 
	 */
	private ProblemRunner problemRunner;

	/**
	 * 
	 * Constructor
	 * 
	 * @param mazeModel
	 *          the maze to visualize
	 * @param title
	 *          policy to draw the maze
	 * 
	 */
	MazeView(MazeModel mazeModel, String title) {
		super(title);
		this.mazeModel = mazeModel;
		initComponents(mazeModel.getDimension());
		initFrame();
		ToolTipManager toolTipManager = ToolTipManager.sharedInstance();
		toolTipManager.setDismissDelay(Integer.MAX_VALUE);
		toolTipManager.setInitialDelay(0);
	}

	/**
	 * 
	 * Draw a maze on the grid
	 * 
	 * @param drawingPolicy
	 *          policy to draw the maze
	 * 
	 */
	public void registerPolicy(DrawingPolicy drawingPolicy) {
		drawingPolicyRegistered.add(drawingPolicy);
	}

	/**
	 * 
	 * Remove all the drawing policies registered for this window
	 * 
	 */
	public void removePolicies() {
		drawingPolicyRegistered.clear();
	}

	/**
	 * 
	 * Initialize the component of this frame
	 * 
	 * @param mazeDimension
	 *          dimension of the maze to display
	 * 
	 */
	protected void initComponents(Dimension mazeDimension) {
		mazeView = new MazePanel();
		mazeView.setPreferredSize(new Dimension(mazeDimension.width * CASE_SIZE,
				mazeDimension.height * CASE_SIZE));
		elementViewGrid = new ElementView[mazeDimension.height][mazeDimension.width];
		for (int y = 0; y < mazeDimension.height; y++) {
			for (int x = 0; x < mazeDimension.width; x++) {
				elementViewGrid[y][x] = new ElementView(new Point(x, y));
			}
		}
		exitButton = new JButton("Exit");
		exitButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				System.exit(0);
			}
		});
		exitButton.setMnemonic(KeyEvent.VK_X);
	}

	/**
	 * 
	 * Initialise the window dudule
	 * 
	 */
	private void initFrame() {
		GridBagConstraints gridBagConstraints;
		getContentPane().setLayout(new GridBagLayout());

		gridBagConstraints = new java.awt.GridBagConstraints();
		gridBagConstraints.gridx = 0;
		gridBagConstraints.gridy = 0;
		gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
		getContentPane().add(mazeView, gridBagConstraints);

		gridBagConstraints = new GridBagConstraints();
		gridBagConstraints.gridx = 0;
		gridBagConstraints.gridy = 1;
		getContentPane().add(exitButton, gridBagConstraints);

		pack();

		addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent event) {
				System.exit(0);
			}
		});
	}

	/**
	 * 
	 * Save this view of the maze into an image
	 * 
	 * @param filename
	 *          name of the file to save
	 * @param format
	 *          format used to save the image
	 * 
	 */
	public void save(String filename, String format) {
		BufferedImage bufferedImage = new BufferedImage(mazeView.getWidth(),
				mazeView.getHeight(), BufferedImage.TYPE_INT_RGB);
		mazeView.paintAll(bufferedImage.createGraphics());
		try {
			ImageIO.write(bufferedImage, format, new FileOutputStream(filename + "."
					+ format));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * 
	 * Run an agent in the maze until the maze is known by the agent
	 * 
	 * @param agenttoRun
	 *          the agent to run
	 * 
	 */
	public void run(MazeAgent agenttoRun) {
		mazeAgent = agenttoRun;
		problemRunner = new ProblemRunner(null, mazeModel, agenttoRun, Log.sysout);
		setVisible(true);
		Timer timer = new Timer(0, this);
		timer.start();
	}

	/**
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	public void actionPerformed(ActionEvent e) {
		repaint();
		problemRunner.oneStep();
		if (problemRunner.getTime() > Double.MAX_VALUE) {
			AbstractLearningForest forest = (AbstractLearningForest) mazeAgent.agent
					.getModel();
			forest.setRewardAdjustmentPolicy(null);
			StaticSpiSvi.convertProblemToAgent(forest, "data/forests/suttonmaze");
			System.exit(0);
		}
	}

	/**
	 * 
	 * Main method
	 * 
	 * @param args
	 *          list of arguments passed to the program
	 * 
	 */
	public static void main(String[] args) {
		MazeModel mazeModel = new MazeModel(MazeModel.SUTTON_MAZE,
				MazeModel.SUTTON_SIMPLE);
		MazeView mazeView = new MazeView(mazeModel, "Maze");

		ModelBasedAgent agent = createSpiSviRMaxAgent(mazeModel);

		mazeView.run(new MazeAgent(mazeView, agent));
	}

	/**
	 * 
	 * Create a Spi/Svi agent
	 * 
	 * @param mazeModel
	 *          the model to run the agent
	 * 
	 * @return the nex agent created
	 * 
	 */
	protected static ModelBasedAgent createLTBPEGreedyAgent(MazeModel mazeModel) {
		ItiTree learningTree = new ItiTree(new KSMeasure(), mazeModel
				.getVariableOrdering());
		AbstractForest learningForest = new VariableActionLearningForest(mazeModel
				.getActionList(), mazeModel.getVariableOrdering(), learningTree, null);
		LearningTreeBasedPlanning planning = new LearningTreeBasedPlanning(
				learningForest);
		planning.setOrdering(new LeastSquareMeasure(null));
		IActing acting = new EpsilonGreedy(new Random(0), mazeModel.getActionList());
		ModelBasedAgent agent = new ModelBasedAgent(planning, acting);
		return agent;
	}

	/**
	 * 
	 * Create a Spi/Svi agent
	 * 
	 * @param mazeModel
	 *          the model to run the agent
	 * 
	 * @return the nex agent created
	 * 
	 */
	protected static ModelBasedAgent createSpiSviRMaxAgent(MazeModel mazeModel) {
		AbstractLearningForest learningForest = new VariableLearningForest(
				mazeModel.getActionList(), mazeModel.getVariableOrdering(),
				new SID4Tree.Prob<String>(new Chi2Measure(null), mazeModel
						.getVariableOrdering()), new SID4Tree.Plain<Double>(
						new StructureLearningMeasure(1, 1, null), mazeModel
								.getVariableOrdering()));
		learningForest.setIncludeActionInRewardLearningFlag(true);
		learningForest.getParameters().setAll(mazeModel.getParameters());
		learningForest.setIncludeActionInRewardLearningFlag(true);
		LearningSpiSvi planning = new LearningSpiSvi(learningForest);
		planning.setOrdering(new LeastSquareMeasure(null));
		ModelBasedAgent agent = new ModelBasedAgent(planning, new Greedy(
				new Random(0), mazeModel.getActionList()));
		return agent;
	}

	/**
	 * 
	 * Create a Spi/Svi agent
	 * 
	 * @param mazeModel
	 *          the model to run the agent
	 * 
	 * @return the nex agent created
	 * 
	 */
	protected static ModelBasedAgent createSpiSviEGreedyAgent(MazeModel mazeModel) {
		ItiTree learningTree = new ItiTree(new KSMeasure(), mazeModel
				.getVariableOrdering());
		AbstractForest learningForest = new VariableActionLearningForest(mazeModel
				.getActionList(), mazeModel.getVariableOrdering(), learningTree, null);
		LearningSpiSvi planning = new LearningSpiSvi(learningForest);
		planning.setOrdering(new LeastSquareMeasure(null));
		ModelBasedAgent agent = new ModelBasedAgent(planning, new EpsilonGreedy(
				new Random(0), mazeModel.getActionList()));
		return agent;
	}

	/**
	 * 
	 * Create a Spi/Svi agent
	 * 
	 * @param mazeModel
	 *          the model to run the agent
	 * 
	 * @return the nex agent created
	 * 
	 */
	protected static ModelBasedAgent createActionAggregatedSpiSviAgent(
			MazeModel mazeModel) {
		ItiTree learningTree = new ItiTree(new Chi2Measure(0.0), mazeModel
				.getVariableOrdering());
		RMaxTree rewardTree = new RMaxTree(1, mazeModel.getVariableOrdering());
		AbstractLearningForest learningForest = new VariableLearningForest(
				mazeModel.getActionList(), mazeModel.getVariableOrdering(),
				learningTree, rewardTree);
		LearningSpiSvi planning = new LearningSpiSvi(learningForest);
		planning.setOrdering(new LeastSquareMeasure(null));
		ModelBasedAgent agent = new ModelBasedAgent(planning, new Greedy(
				new Random(0), mazeModel.getActionList()));
		return agent;
	}

	/**
	 * 
	 * Create a Dyna agent
	 * 
	 * @param mazeModel
	 *          the model to run the agent
	 * 
	 * @return the new agent created
	 * 
	 */
	protected static ModelBasedAgent createDynaAgent(MazeModel mazeModel) {
		TabularLearningModel model = new TabularLearningModel(mazeModel
				.getVariableOrdering(), mazeModel.getActionList());
		IIncrementalPlanning planning = new DynaQ(model);
		ModelBasedAgent agent = new ModelBasedAgent(planning, new EpsilonGreedy(
				new Random(0), mazeModel.getActionList()));
		return agent;
	}
}
