package animatlab.kodamat.reinforcementlearning.agent;

import java.util.List;
import java.util.Map;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.io.HtmlPage;
import animatlab.kodamat.core.io.IToXml;
import animatlab.kodamat.core.io.Log;
import animatlab.kodamat.core.utils.Chrono;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.reinforcementlearning.core.acting.IActing;
import animatlab.kodamat.reinforcementlearning.model.ILearningModel;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractForest;
import animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning;
import animatlab.kodamat.reinforcementlearning.planning.IPlanning;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.SpiSvi;

/**
 * 
 * Model based agent
 * 
 * @author Thomas Degris
 * 
 */
public class ModelBasedAgent implements IAgent, IToXml<Object> {
	/**
	 * 
	 * Name of the time points used for learning
	 * 
	 */
	public static final String TIME_POINTS_LEARNING = "TIME_POINTS_LEARNING";

	/**
	 * 
	 * Name of the time points used for planning
	 * 
	 */
	public static final String TIME_POINTS_PLANNING = "TIME_POINTS_PLANNING";

	/**
	 * 
	 * Algorithm to learn the model
	 * 
	 */
	final private ILearningModel model;

	/**
	 * 
	 * Algorithm to compute the planning policy in the model
	 * 
	 */
	private IPlanning planning;

	/**
	 * 
	 * Algorithm to choose the action from a map of quality function
	 * 
	 */
	final private IActing acting;

	/**
	 * 
	 * Last state of this agent
	 * 
	 */
	private Context lastTimeState = null;

	/**
	 * 
	 * Last action executed by this agent
	 * 
	 */
	private String lastTimeAction = null;

	/**
	 * 
	 * Chrono used to measure learning and planning time
	 * 
	 */
	private Chrono chrono = new Chrono();

	/**
	 * 
	 * Constructor
	 * 
	 * @param planning
	 *          the incremental planning method of the agent
	 * @param acting
	 *          the policy to decide what to do
	 * 
	 */
	public ModelBasedAgent(IPlanning planning, IActing acting) {
		this.model = (ILearningModel) planning.getModel();
		this.planning = planning;
		this.acting = acting;
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param model
	 *          model used to learn the world transitions
	 * @param acting
	 *          the policy to decide what to do
	 * 
	 */
	public ModelBasedAgent(ILearningModel model, IActing acting) {
		this.model = model;
		this.planning = null;
		this.acting = acting;
	}

	/**
	 * 
	 * @return get the learning model used by the agent
	 * 
	 */
	public ILearningModel getModel() {
		return model;
	}

	/**
	 * 
	 * @return get the planning method used by the agent
	 * 
	 */
	public IPlanning getPlanning() {
		return planning;
	}

	/**
	 * 
	 * @return get the planning method used by the agent
	 * 
	 */
	private IIncrementalPlanning getIncPlanning() {
		return (IIncrementalPlanning) planning;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.agent.IAgent#getAction(int,
	 *      Context, java.util.List)
	 */
	public String getAction(int time, Context state, List<Double> reward) {
		chrono.clearTimePoints();
		chrono.start();
		getModel().learn(lastTimeState, lastTimeAction, state, reward);
		chrono.addTimePoint(TIME_POINTS_LEARNING);
		Map<String, IFunction<Double>> qualityFunctions = null;
		if ((planning != null) && (getModel().getActionTestedSet().size() > 1)) {
			chrono.start();
			qualityFunctions = getIncPlanning().getIncrementalPlanning(lastTimeState,
					lastTimeAction, state, reward);
			chrono.addTimePoint(TIME_POINTS_PLANNING);
		}
		String action = null;
		if (state != null) {
			action = acting.getAction(qualityFunctions, state);
		}
		lastTimeState = state;
		lastTimeAction = action;
		return action;
	}

	/**
	 * @see animatlab.kodamat.core.io.ILog#setLog(animatlab.kodamat.core.io.Log)
	 */
	public void setLog(Log log) {
		if (planning != null) {
			planning.setLog(log);
		}
		if (acting != null) {
			acting.setLog(log);
		}
	}

	/**
	 * @see animatlab.kodamat.core.io.IToXml#toXml(org.w3c.dom.Document,
	 *      animatlab.kodamat.core.io.IToXml.IToXmlPolicy)
	 */
	public Element toXml(Document document, IToXmlPolicy<Object> policy) {
		if ((planning != null) && (planning instanceof IToXml)) {
			return ((IToXml<?>) planning).toXml(document, null);
		}
		return ((IToXml<?>) model).toXml(document, null);
	}

	/**
	 * 
	 * Dump this agent into a html page
	 * 
	 * @param folderPath
	 *          path where to save the html page
	 * 
	 * 
	 */
	public void dump(String folderPath) {
		if ((planning != null) && (planning instanceof SpiSvi)) {
			((SpiSvi) planning).dump(folderPath, "ModelBasedAgent");
			return;
		}
		if (!(getModel() instanceof AbstractForest)) {
			return;
		}
		HtmlPage htmlDotPage = new HtmlPage(folderPath, "index.html");
		htmlDotPage.setHtmlPageSetTitle("ModelBasedAgent");
		htmlDotPage.setTitle("Index");
		((AbstractForest) getModel()).dump(htmlDotPage);
		htmlDotPage.write();
	}

	/**
	 * 
	 * @return acting instance of this agent
	 * 
	 */
	public IActing getActing() {
		return acting;
	}

	/**
	 * 
	 * Set the planning method of this agent
	 * 
	 * @param planning
	 *          the planning method to set
	 * 
	 */
	public void setPlanning(IPlanning planning) {
		this.planning = planning;
	}

	/**
	 * 
	 * @return Returns the chrono.
	 * 
	 */
	public Chrono getChrono() {
		return chrono;
	}
}
