/*
 * 
 * Created on Jan 16, 2006
 *
 */
package animatlab.kodamat.reinforcementlearning.agent;

import java.util.Random;

import animatlab.kodamat.core.IProblemDefinition;
import animatlab.kodamat.core.io.Log;
import animatlab.kodamat.reinforcementlearning.core.acting.IActing;
import animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel;
import animatlab.kodamat.reinforcementlearning.model.ILearningModel;
import animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning;
import animatlab.kodamat.reinforcementlearning.planning.IPlanning;

/**
 * 
 * Factory to create model based agents
 * 
 * @author Thomas Degris
 * 
 */
public class ModelBasedAgentFactory implements IAgentFactory {
	/**
	 * 
	 * Planning prototype
	 * 
	 */
	final private IIncrementalPlanning planningPrototype;

	/**
	 * 
	 * Model learning prototype
	 * 
	 */
	final private ILearningModel learningModelPrototype;

	/**
	 * 
	 * Acting prototype
	 * 
	 */
	final private IActing actingPrototype;

	/**
	 * 
	 * Log to set to new agents
	 * 
	 */
	private Log log;

	/**
	 * 
	 * Constructor
	 * 
	 * @param planningPrototype
	 *          planning prototype
	 * @param learningModelPrototype
	 *          model learning prototype
	 * @param actingPrototype
	 *          acting prototype
	 * 
	 */
	public ModelBasedAgentFactory(IIncrementalPlanning planningPrototype,
			ILearningModel learningModelPrototype, IActing actingPrototype) {
		this.planningPrototype = planningPrototype;
		this.learningModelPrototype = learningModelPrototype;
		this.actingPrototype = actingPrototype;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.agent.IAgentFactory#createAgent(java.util.Random,
	 *      IProblemDefinition,
	 *      animatlab.kodamat.reinforcementlearning.planning.IPlanning)
	 */
	public ModelBasedAgent createAgent(Random random, IProblemDefinition problem,
			IPlanning solution) {

		ILearningModel learningModel = learningModelPrototype.newInstance(problem);
		if ((learningModel instanceof AbstractFactoredModel)
				&& (problem instanceof AbstractFactoredModel)) {
			final AbstractFactoredModel factoredLearning = (AbstractFactoredModel) learningModel;
			final AbstractFactoredModel factoredProblem = (AbstractFactoredModel) problem;
			factoredLearning.setBasisFunctions(factoredProblem.getBasisFunctions()
					.clone());
		}

		IIncrementalPlanning planning = null;
		if (planningPrototype != null) {
			planning = planningPrototype.newInstance(learningModel);
		}

		IActing acting = null;
		if (actingPrototype != null) {
			acting = actingPrototype.newInstance(random, problem.getActionList(),
					planning);
		}

		ModelBasedAgent result;
		if (planning == null) {
			result = new ModelBasedAgent(learningModel, acting);
		} else {
			result = new ModelBasedAgent(planning, acting);
		}
		result.setLog(log);
		return result;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.agent.IAgentFactory#getAgentName()
	 */
	public String getAgentName() {
		StringBuilder name = new StringBuilder();
		name.append(learningModelPrototype.getName());
		if (planningPrototype != null) {
			name.append(".");
			name.append(planningPrototype.getName());
		}
		if (actingPrototype != null) {
			name.append(".");
			name.append(actingPrototype.getName());
		}
		return name.toString().toLowerCase();
	}

	/**
	 * @see animatlab.kodamat.core.io.ILog#setLog(animatlab.kodamat.core.io.Log)
	 */
	public void setLog(Log log) {
		this.log = log;
	}

	/**
	 * 
	 * @return the model used to create model based agent
	 * 
	 */
	public ILearningModel getModel() {
		return learningModelPrototype;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.agent.IAgentFactory#requireSolution()
	 */
	public boolean requireSolution() {
		return false;
	}
}