/*
 * 
 * Created on Jan 11, 2006
 *
 */
package animatlab.kodamat.reinforcementlearning.core.acting;

import static animatlab.kodamat.core.utils.Parameters.EPSILON;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import animatlab.kodamat.core.IProblemDefinition;
import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.reinforcementlearning.agent.IAgent;
import animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning;
import animatlab.kodamat.reinforcementlearning.planning.IPlanning;

/**
 * 
 * Greedy policy
 * 
 * @author Thomas Degris
 * 
 */
public class Greedy extends RandomActing implements IFunction<Set<String>> {
	/**
	 * 
	 * List of possible actions
	 * 
	 */
	final private List<String> greedyActionList = new ArrayList<String>();

	/**
	 * 
	 * Default action to choose when no random instance is available and an action
	 * between many must be choosen
	 * 
	 */
	private String defaultAction = null;

	/**
	 * 
	 * The set of quality functions to use when using this as a policy
	 * 
	 */
	final private Map<String, IFunction<Double>> actionValueFunctions;

	/**
	 * 
	 * When this is used as an acting instance, then we need a policy to execute
	 * 
	 */
	final private IFunction<Set<String>> optimalPolicy;

	/**
	 * 
	 * Constructor
	 * 
	 * @param random
	 *          random instance
	 * @param actionList
	 *          set of available actions
	 * 
	 */
	public Greedy(Random random, List<String> actionList) {
		super(random, actionList);
		actionValueFunctions = null;
		optimalPolicy = null;
	}

	/**
	 * 
	 * Constructor to use a set of quality function as a policy without explicitly
	 * building it
	 * 
	 * @param qualityFunctions
	 *          the quality functions to composing the policy
	 * 
	 */
	public Greedy(Map<String, IFunction<Double>> qualityFunctions) {
		super(null, new ArrayList<String>(qualityFunctions.keySet()));
		this.actionValueFunctions = qualityFunctions;
		optimalPolicy = null;
	}

	/**
	 * 
	 * Constructor when using this as agent factory
	 * 
	 */
	public Greedy() {
		super();
		actionValueFunctions = null;
		optimalPolicy = null;
	}

	/**
	 * 
	 * Constructor when using this as an acting instance
	 * 
	 * @param random
	 *          random instance
	 * @param planning
	 *          the instance to get the optimal policy to execute
	 * 
	 */
	public Greedy(Random random, IPlanning planning) {
		super(random, planning.getModel().getActionList());
		actionValueFunctions = null;
		optimalPolicy = planning.getOptimalPolicy();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.acting.IActing#getAction(java.util.Map,
	 *      Context)
	 */
	@Override
	public String getAction(Map<String, IFunction<Double>> qualityFunctions,
			Context state) {
		if ((qualityFunctions == null) && (optimalPolicy == null)) {
			if (random == null) {
				return actionList.get(0);
			}
			return randomAction();
		}
		buildGreedyActionList(greedyActionList, qualityFunctions, state);
		if (random == null) {
			if ((defaultAction == null)
					|| (!greedyActionList.contains(defaultAction))) {
				defaultAction = greedyActionList.get(0);
			}
			return defaultAction;
		}
		return greedyActionList.get(random.nextInt(greedyActionList.size()));
	}

	/**
	 * 
	 * Build a list of greedy actions
	 * 
	 * @param greedyList
	 *          the list to fill of greedy actions
	 * @param qualityFunctions
	 *          the map of quality functions
	 * @param state
	 *          the state to build the greedy action for
	 * 
	 */
	public void buildGreedyActionList(Collection<String> greedyList,
			Map<String, IFunction<Double>> qualityFunctions, Context state) {
		greedyList.clear();
		if (optimalPolicy != null) {
			greedyList.addAll(optimalPolicy.classify(state));
			assert (greedyList.size() > 0);
			return;
		}
		double bestQuality = -Double.MAX_VALUE;
		for (Map.Entry<String, IFunction<Double>> entry : qualityFunctions
				.entrySet()) {
			Double quality = entry.getValue().classify(state);
			if (quality == null) {
				quality = Double.MAX_VALUE;
			}
			double diff = quality - bestQuality;
			if (diff > EPSILON) {
				greedyList.clear();
				greedyList.add(entry.getKey());
				bestQuality = quality;
			} else if (Math.abs(diff) < EPSILON) {
				greedyList.add(entry.getKey());
			}
		}
	}

	/**
	 * @see animatlab.kodamat.core.functions.IFunction#classify(Context)
	 */
	public Set<String> classify(Context state) {
		buildGreedyActionList(greedyActionList, actionValueFunctions, state);
		return new LinkedHashSet<String>(greedyActionList);
	}

	/**
	 * @see animatlab.kodamat.core.functions.IFunction#numberOfNodes()
	 */
	public int numberOfNodes() {
		int result = 0;
		for (IFunction<Double> actionValueFunction : actionValueFunctions.values()) {
			result = result + actionValueFunction.numberOfNodes();
		}
		return result;
	}

	/**
	 * @see animatlab.kodamat.core.functions.IFunction#clone()
	 */
	@Override
	public IFunction<Set<String>> clone() {
		return new Greedy(actionValueFunctions);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.agent.IAgentFactory#createAgent(Random,
	 *      IProblemDefinition,
	 *      animatlab.kodamat.reinforcementlearning.planning.IPlanning)
	 */
	@Override
	public IAgent createAgent(Random random, IProblemDefinition problem,
			IPlanning solution) {
		return new Greedy(random, solution);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.agent.IAgent#getAction(int,
	 *      Context, java.util.List)
	 */
	@Override
	public String getAction(int time, Context state, List<Double> reward) {
		if (state == null) {
			return null;
		}
		greedyActionList.clear();
		greedyActionList.addAll(optimalPolicy.classify(state));
		return greedyActionList.get(random.nextInt(greedyActionList.size()));
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.agent.IAgentFactory#requireSolution()
	 */
	@Override
	public boolean requireSolution() {
		return true;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.acting.IActing#newInstance(java.util.Random,
	 *      java.util.List, IIncrementalPlanning)
	 */
	@Override
	public IActing newInstance(Random random, List<String> actionList,
			IIncrementalPlanning learningPlanning) {
		return new Greedy(random, actionList);
	}
}
