/*
 * 
 * Created on 14 nov. 2005
 *
 */
package animatlab.kodamat.reinforcementlearning.core.stats;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import animatlab.kodamat.core.IProblemDefinition;
import animatlab.kodamat.core.io.ILog;
import animatlab.kodamat.core.io.Log;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.reinforcementlearning.agent.IAgent;
import animatlab.kodamat.reinforcementlearning.core.IStepListener;
import animatlab.kodamat.reinforcementlearning.planning.IPlanning;

/**
 * 
 * Measure of an agent in a problem
 * 
 * @author Thomas Degris
 * 
 */
public abstract class Measure implements IStepListener, ILog {
	/**
	 * 
	 * The model of the world where the agent learn
	 * 
	 */
	protected IProblemDefinition problem = null;

	/**
	 * 
	 * Solution of the problem used as reference
	 * 
	 */
	protected IPlanning solution = null;

	/**
	 * 
	 * The agent to measure
	 * 
	 */
	protected IAgent agent;

	/**
	 * 
	 * The folder where to print out the results
	 * 
	 */
	protected String resultsFolder;

	/**
	 * 
	 * Pool log file
	 * 
	 */
	protected Log log = Log.sysout;

	/**
	 * 
	 * Constructor
	 * 
	 */
	public Measure() {
		super();
	}

	/**
	 * 
	 * Create an new agent
	 * 
	 * @param problem
	 *          the problem where the agent will be
	 * @param solution
	 *          the solution of the problem (may be null)
	 * @param agent
	 *          the agent to measure
	 * @param resultsFolder
	 *          the folder where the results will be printed out
	 * 
	 */
	public void setMeasureData(IProblemDefinition problem, IPlanning solution,
			IAgent agent, String resultsFolder) {
		this.problem = problem;
		this.solution = solution;
		this.agent = agent;
		this.resultsFolder = resultsFolder;
	}

	/**
	 * 
	 * Reset all the data of this
	 * 
	 */
	public abstract void reset();

	/**
	 * 
	 * Method to receive an event at each step of the agent
	 * 
	 * @param data
	 *          the map where to set the results computed
	 * 
	 */
	abstract public void computeMeasures(Map<String, Double> data);

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.IStepListener#listenStep(int,
	 *      Context, java.lang.String, Context, java.util.List)
	 */
	public void listenStep(int time, Context currentState, String actionChoosen,
			Context nextState, List<Double> lastReward) {
		// Nothing to do for this measure
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.IStepListener#terminate()
	 */
	public void terminate() {
		// Nothing to do for this measure
	}

	/**
	 * @see animatlab.kodamat.core.io.ILog#setLog(animatlab.kodamat.core.io.Log)
	 */
	public void setLog(Log log) {
		this.log = log;
	}

	/**
	 * 
	 * @return flag to indicate if this measure require the solution or not
	 * 
	 */
	public boolean requireSolution() {
		return false;
	}

	/**
	 * 
	 * Create a set of measure points from a period and a maximum time
	 * 
	 * @param start
	 *          value used to start the experiment
	 * @param end
	 *          end of the experiment
	 * @param period
	 *          period used to compute the measures
	 * 
	 * @return the set of measure points
	 * 
	 */
	static public double[] measurePoints(double start, double end, double period) {
		double[] points = new double[((int) ((end - start) / period)) + 1];
		for (int index = 0; index < points.length; index++) {
			points[index] = start + period * index;
		}
		return points;
	}

	/**
	 * 
	 * Add some measure points to a set of points
	 * 
	 * @param measurePoints
	 *          the set of points where to add the new measure points
	 * @param newPoints
	 *          the new points to add
	 * 
	 * @return the new ordered set of measure points
	 * 
	 */
	public static double[] addMesurePoints(double[] measurePoints,
			double... newPoints) {
		List<Double> list = new ArrayList<Double>();
		for (double measure : measurePoints) {
			list.add(measure);
		}
		for (double measure : newPoints) {
			list.add(measure);
		}
		Collections.sort(list);
		double[] result = new double[list.size()];
		for (int index = 0; index < list.size(); index++) {
			result[index] = list.get(index);
		}
		return result;
	}

}
