package animatlab.kodamat.reinforcementlearning.core.stats;

import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.reinforcementlearning.core.IStepListener;

/**
 * 
 * Writer to print out measures in xml file
 * 
 * @author Thomas Degris
 * 
 */
public class MeasureWriter implements IStepListener {
	/**
	 * 
	 * Output for logs
	 * 
	 */
	private PrintStream out = null;

	/**
	 * 
	 * Results to print out
	 * 
	 */
	final private Map<String, Double> results = new LinkedHashMap<String, Double>();

	/**
	 * 
	 * The list of measures to write
	 * 
	 */
	final private List<Measure> measures;

	/**
	 * 
	 * Points where to compute measures
	 * 
	 */
	private double[] measurePoints = null;

	/**
	 * 
	 * Current index of the next point to measure
	 * 
	 */
	private int currentPointIndex = 0;

	/**
	 * 
	 * Name of the xml root node name
	 * 
	 */
	static final private String rootNodeName = "results";

	/**
	 * 
	 * Name of the xml root node name
	 * 
	 */
	static final private String stepNodeName = "step";

	/**
	 * 
	 * Constructor
	 * 
	 * @param measures
	 *          the list of measures to write
	 * @param filepath
	 *          the file where to write
	 * 
	 */
	public MeasureWriter(List<Measure> measures, String filepath) {
		if (filepath != null) {
			try {
				out = new PrintStream(filepath + ".results.xml");
			} catch (FileNotFoundException e) {
				e.printStackTrace();
				System.exit(1);
			}
			out.println("<" + rootNodeName + ">");
		}
		this.measures = measures;
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param measures
	 *          list of measure to write
	 * 
	 */
	public MeasureWriter(Measure... measures) {
		this.measures = new ArrayList<Measure>();
		for (Measure measure : measures) {
			this.measures.add(measure);
		}
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.IStepListener#listenStep(int,
	 *      Context, java.lang.String, Context, java.util.List)
	 */
	public void listenStep(int time, Context currentState, String actionChoosen,
			Context nextState, List<Double> lastReward) {
		computeMeasures("time", time);
	}

	/**
	 * 
	 * Compute the set of measures given a map containing the primary key
	 * 
	 * @param data
	 *          the set of data to add to the results
	 * 
	 */
	public void computeMeasures(Map<String, Double> data) {
		results.putAll(data);
		computePrintResults();
	}

	/**
	 * 
	 * Compute the set of measures (if necessary) given a primary key
	 * 
	 * @param primKeyLabel
	 *          the label of the primary key
	 * @param primKeyValue
	 *          the value of the primary key
	 * 
	 */
	public void computeMeasures(String primKeyLabel, double primKeyValue) {
		if ((measurePoints != null)
				&& (currentPointIndex >= measurePoints.length || (primKeyValue < measurePoints[currentPointIndex]))) {
			return;
		}
		currentPointIndex++;
		results.put(primKeyLabel, primKeyValue);
		computePrintResults();
	}

	/**
	 * 
	 * Compute first the measures and then print them into the xml file
	 * 
	 */
	private void computePrintResults() {
		for (Measure measure : measures) {
			measure.computeMeasures(results);
		}
		if (out != null) {
			out.print("  <" + stepNodeName);
			for (Map.Entry<String, Double> entry : results.entrySet()) {
				out.print(" " + entry.getKey() + "='" + entry.getValue() + "'");
			}
			out.println("/>");
			out.flush();
		}
		results.clear();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.IStepListener#terminate()
	 */
	public void terminate() {
		if (out != null) {
			out.println("</" + rootNodeName + ">");
			out.flush();
			out.close();
			out = null;
		}
	}

	/**
	 * 
	 * Set the set points where to compute the measures
	 * 
	 * @param measurePoints
	 *          the set of (ordered) points where to compute the measures
	 * 
	 */
	public void setMeasurePoints(double[] measurePoints) {
		this.measurePoints = measurePoints;
	}

	/**
	 * 
	 * Add a data to the next time a result will be written
	 * 
	 * @param key
	 *          key of the result to print
	 * @param value
	 *          value of the result to print
	 * 
	 */
	public void addToNextStep(String key, double value) {
		results.put(key, value);
	}

	/**
	 * 
	 * Reset all the measures
	 * 
	 */
	public void reset() {
		for (Measure measure : measures) {
			measure.reset();
		}
	}
}
