package animatlab.kodamat.reinforcementlearning.model;

import static animatlab.kodamat.core.io.XmlUtils.getFirstChild;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import animatlab.kodamat.core.IModel;
import animatlab.kodamat.core.IProblemDefinition;
import animatlab.kodamat.core.functions.FunctionsSum;
import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.functions.IPartition;
import animatlab.kodamat.core.functions.IPartitionFunction;
import animatlab.kodamat.core.io.DataLoadingPolicy;
import animatlab.kodamat.core.io.IToXml;
import animatlab.kodamat.core.io.XmlUtils;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.rules.PlainRules;
import animatlab.kodamat.core.rules.Rules;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.core.utils.VariableOrdering;

/**
 * 
 * Utils function using factored model
 * 
 * @author Thomas Degris
 * 
 */
public abstract class AbstractFactoredModel implements IModel, IToXml<Object> {
	/**
	 * 
	 * Prefix for the forest in a xml file
	 * 
	 */
	static public final String NAMESPACE = "fst";

	/**
	 * 
	 * Name of the node of the reward function
	 * 
	 */
	static public final String NODE_NAME_REWARD = AbstractFactoredModel.NAMESPACE
			+ ":" + "reward";

	/**
	 * 
	 * Name of the node of the set of cpt-trees for all actions
	 * 
	 */
	static public final String NODE_NAME_TRANSITIONS = AbstractFactoredModel.NAMESPACE
			+ ":" + "transitions";

	/**
	 * 
	 * Name of the node of the set of CPT trees for an action
	 * 
	 */
	static public final String NODE_NAME_ACTION = AbstractFactoredModel.NAMESPACE
			+ ":" + "action";

	/**
	 * 
	 * Name of the node of the set of CPT trees for a variable
	 * 
	 */
	static public final String NODE_NAME_VARIABLE = AbstractFactoredModel.NAMESPACE
			+ ":" + "variable";

	/**
	 * 
	 * Name of the attribute to named an action
	 * 
	 */
	static public final String ATTRIBUTE_NAME_ACTION = "name";

	/**
	 * 
	 * Name of the attribute to named a variable
	 * 
	 */
	public static final String ATTRIBUTE_QUALITY = "data";

	/**
	 * 
	 * Name of the node defining the basis functions to use
	 * 
	 */
	public static final String NODE_NAME_BASIS_FUNCTIONS = AbstractFactoredModel.NAMESPACE
			+ ":" + "basisfunctions";

	/**
	 * 
	 * Policy used to load StaticCPD from a xml node
	 * 
	 */
	private class RewardLoadingPolicy implements DataLoadingPolicy<Double> {
		/**
		 * 
		 * Constructor
		 * 
		 */
		RewardLoadingPolicy() {
			super();
		}

		/**
		 * @see animatlab.kodamat.core.io.DataLoadingPolicy#xmlToData(org.w3c.dom.Element)
		 */
		public Double xmlToData(Element element) {
			double reward = DataLoadingPolicy.doubleLoadingPolicy
					.xmlToData(element);
			rewardMax = Math.max(getRewardMax(), reward);
			return reward;
		}
	}

	/**
	 * 
	 * Parameters of the model
	 * 
	 */
	final private Parameters parameters = new Parameters();

	/**
	 * 
	 * Set of availables action
	 * 
	 */
	final protected List<String> actionList;

	/**
	 * 
	 * Set of basis functions
	 * 
	 */
	protected FunctionsSum basisFunctionSet = null;

	/**
	 * 
	 * Reward function of the problem
	 * 
	 */
	protected FunctionsSum rewardFunction = new FunctionsSum();

	/**
	 * 
	 * Reward maximum learn from the beginning of the problem
	 * 
	 */
	protected double rewardMax = -Double.MAX_VALUE;

	/**
	 * 
	 * Constructor
	 * 
	 * @param actionList
	 *            action set used in this forest
	 * 
	 */
	public AbstractFactoredModel(List<String> actionList) {
		this.actionList = actionList;
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getActionList()
	 */
	public List<String> getActionList() {
		return actionList;
	}

	/**
	 * 
	 * @return the parameters of this forest
	 * 
	 */
	public Parameters getParameters() {
		return parameters;
	}

	/**
	 * 
	 * Compute the next state from the current state and an action
	 * 
	 * @param random
	 *            random instance
	 * @param problemDefinition
	 *            model to use
	 * @param state
	 *            the current state of the agent
	 * @param nextState
	 *            the next state filled by the method
	 * @param action
	 *            the action executed by the agent
	 * @return the current reward of the agent
	 * 
	 */
	static public List<Double> computeNextStateWithTerminal(Random random,
			IProblemDefinition problemDefinition, Context state,
			Context nextState, String action) {
		AbstractFactoredModel model = (AbstractFactoredModel) problemDefinition;
		if (state == null) {
			nextState.clear();
			nextState.putAll(problemDefinition.getInitialState(random));
			return null;
		}
		if (problemDefinition.isTerminal(state)) {
			nextState.clear();
			return model.getReward(state, action);
		}
		return computeNextState(random, model, state, nextState, action);
	}

	/**
	 * 
	 * Compute the next state from the current state and an action
	 * 
	 * @param randomModel
	 *            random instance
	 * @param factoredModel
	 *            model to use
	 * @param state
	 *            the current state of the agent
	 * @param nextState
	 *            the next state filled by the method
	 * @param action
	 *            the action executed by the agent
	 * @return the current reward of the agent
	 * 
	 */
	static public List<Double> computeNextState(Random randomModel,
			AbstractFactoredModel factoredModel, Context state,
			Context nextState, String action) {
		nextState.clear();
		for (Variable variable : factoredModel.getVariableSet()) {
			IFunction<IProbabilities<String>> variableModel = factoredModel
					.getTransitionFunction(variable, action);
			if (variableModel == null) {
				nextState.clear();
				return null;
			}
			IProbabilities<String> node = variableModel.classify(state);
			Set<String> valueSet = node.getValueSet();
			double randomValue = 0;
			if (valueSet.size() > 1) {
				randomValue = randomModel.nextDouble();
			}
			double probabilitySum = 0;
			String value = null;
			Iterator<String> iterator = valueSet.iterator();
			while (value == null) {
				String currentValue = iterator.next();
				probabilitySum = probabilitySum
						+ node.getProbability(currentValue);
				if (randomValue < probabilitySum) {
					value = currentValue;
				}
			}
			nextState.put(variable, value);
		}
		return factoredModel.getReward(state, action);
	}

	/**
	 * 
	 * Initialize randomly a state
	 * 
	 * @param variableOrdering
	 *            variable ordering
	 * @param randomModel
	 *            random instance to use
	 * @param state
	 *            the state to initialize
	 * @param variableValueSet
	 *            set of values for each variable
	 * @param stateSize
	 *            size of the state to initialize
	 * 
	 */
	static public void initializeState(VariableOrdering variableOrdering,
			Random randomModel, Context state,
			Map<Variable, Set<String>> variableValueSet, int stateSize) {
		for (int index = 0; index < stateSize; index++) {
			Variable variable = variableOrdering.get(index);
			Set<String> valueSet = variableValueSet.get(variable);
			int valueIndex = randomModel.nextInt(valueSet.size());
			Iterator<String> valueIterator = valueSet.iterator();
			String value;
			do {
				value = valueIterator.next();
				valueIndex--;
			} while (valueIndex >= 0);
			state.put(variable, value);
		}
	}

	/**
	 * 
	 * Compute the number of possible state of a problem
	 * 
	 * @param problem
	 *            the problem where to compute the number of possible state
	 * @return the number of possible state in the problem
	 * 
	 */
	static public long getNumberOfStates(IModel problem) {
		long result = 1;
		for (Variable variable : problem.getVariableSet()) {
			result *= problem.getValueSet(variable).size();
		}
		return result;
	}

	/**
	 * 
	 * Get a CPT tree from the forest
	 * 
	 * @param variable
	 *            variable to get the corresponding CPD from
	 * @param action
	 *            action to get the corresponding CPD from
	 * @return the CPD corresponding to the couple variable/action
	 * 
	 */
	abstract public IFunction<IProbabilities<String>> getTransitionFunction(
			Variable variable, String action);

	/**
	 * 
	 * Return a reward tree depending on a particular action
	 * 
	 * @param state
	 *            state of the agent
	 * @param action
	 *            the action for which the reward must be returned
	 * 
	 * @return the reward tree corresponding to the action
	 * 
	 */
	public List<Double> getReward(Context state, String action) {
		if ((state == null) || (action == null)) {
			return null;
		}
		List<Double> rewards = new ArrayList<Double>();
		Context modifiedState = new Context(state);
		modifiedState.put(Variable.ACTION, action);
		for (IFunction<Double> localizedReward : getRewardFunction(action)
				.functions()) {
			rewards.add(localizedReward.classify(modifiedState));
		}
		return rewards;
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getRewardModelSize()
	 */
	public int getRewardModelSize() {
		return rewardFunction.numberOfNodes();
	}

	/**
	 * 
	 * Get a reward function definition
	 * 
	 * @param action
	 *            the action to get the reward from
	 * @return the reward function
	 * 
	 */
	public FunctionsSum getRewardFunction(String action) {
		return rewardFunction;
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getName()
	 */
	public String getName() {
		return getClass().getSimpleName();
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuilder result = new StringBuilder();
		for (String action : getActionList()) {
			result.append("------------------------------------------------\n");
			for (Variable variable : getVariableSet()) {
				result.append("Action: " + action);
				result.append(" Variable: " + variable.toString());
				result.append("\n");
				result.append(getTransitionFunction(variable, action)
						.toString());
				result.append("\n");
			}
			FunctionsSum rewardFunction = getRewardFunction(action);
			for (int rewardIndex = 0; rewardIndex < rewardFunction.functions()
					.size(); rewardIndex++) {
				result.append("Action: " + action + " Reward " + rewardIndex
						+ "\n");
				result.append(rewardFunction.functions().get(rewardIndex)
						.toString());
				result.append("\n");
			}
		}
		return result.toString();
	}

	/**
	 * 
	 * Get the number of partition, given a defined scope
	 * 
	 * @param model
	 *            the model to split
	 * @param scope
	 *            the defined scope
	 * @return the number of partitions in the state space
	 * 
	 */
	public static long getNumberOfPartition(AbstractFactoredModel model,
			Set<Variable> scope) {
		assert (scope != null);
		long result = 1;
		for (Variable variable : scope) {
			result = result * model.getValueSet(variable).size();
		}
		return result;
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getVariableSet()
	 */
	public Set<Variable> getVariableSet() {
		return getValueSets().keySet();
	}

	/**
	 * 
	 * @return true when the problem is stochastic
	 * 
	 */
	public boolean isStochastic() {
		for (Variable variable : getVariableSet()) {
			for (String action : getActionList()) {
				IFunction<IProbabilities<String>> transition = getTransitionFunction(
						variable, action);
				for (IPartition<IProbabilities<String>> partition : ((IPartitionFunction<IProbabilities<String>>) transition)
						.partitions()) {
					if (partition.get().getValueSet().size() > 1) {
						return true;
					}
				}
			}
		}
		return false;
	}

	/**
	 * 
	 * @return the name of the xml root node
	 * 
	 */
	abstract protected String getXmlRootNodeName();

	/**
	 * 
	 * The xml node representing the reward tree
	 * 
	 * @param document
	 *            the xml document used to create xml elements
	 * @return the node representing the reward node
	 * 
	 */
	@SuppressWarnings("unchecked")
	private Element modelToXml(Document document) {
		Element cptForestNode = document
				.createElement(AbstractFactoredModel.NODE_NAME_TRANSITIONS);
		for (String action : actionList) {
			Element actionNode = document
					.createElement(AbstractFactoredModel.NODE_NAME_ACTION);
			actionNode.setAttribute(
					AbstractFactoredModel.ATTRIBUTE_NAME_ACTION, action);
			for (Variable variable : getVariableSet()) {
				Element variableNode = document
						.createElement(AbstractFactoredModel.NODE_NAME_VARIABLE);
				variable.setAttribute(variableNode);
				IFunction<IProbabilities<String>> transition = getTransitionFunction(
						variable, action);
				if (transition == null) {
					transition = new Tree<IProbabilities<String>>();
				}
				IToXml<IProbabilities<String>> toXml = (IToXml<IProbabilities<String>>) transition;
				variableNode.appendChild(toXml.toXml(document,
						StaticCPD.PROBABILITIES_POLICY));
				actionNode.appendChild(variableNode);
			}
			cptForestNode.appendChild(actionNode);
		}
		return cptForestNode;
	}

	/**
	 * 
	 * Comvert the reward function into a xml tree
	 * 
	 * @param document
	 *            the document owning the xml tree
	 * @param rewardNode
	 *            the xml node where to add the rewards
	 * 
	 */
	@SuppressWarnings("unchecked")
	final protected void rewardToXml(Document document, Element rewardNode) {
		for (IFunction<Double> reward : getRewardFunction(null).functions()) {
			rewardNode.appendChild(((IToXml<Double>) reward).toXml(document,
					IToXml.doubleToXmlPolicy));
		}
	}

	/**
	 * @see animatlab.kodamat.core.io.IToXml#toXml(org.w3c.dom.Document,
	 *      animatlab.kodamat.core.io.IToXml.IToXmlPolicy)
	 */
	public Element toXml(Document document, IToXmlPolicy<Object> policy) {
		Element spiProblemNode = document.createElement(getXmlRootNodeName());
		XmlUtils.declareNS(document, AbstractFactoredModel.NAMESPACE);
		XmlUtils.declareNS(document, IProbabilities.NAMESPACE);
		spiProblemNode.appendChild(modelToXml(document));
		basisFunctionToXml(document, spiProblemNode);
		spiProblemNode.appendChild(this.parameters.toXml(document, null));
		Element rewardNode = document
				.createElement(AbstractFactoredModel.NODE_NAME_REWARD);
		rewardToXml(document, rewardNode);
		spiProblemNode.appendChild(rewardNode);
		return spiProblemNode;
	}

	/**
	 * 
	 * Convert the basis function set into a xml tree
	 * 
	 * @param document
	 *            the document owning the xml tree
	 * @param parentNode
	 *            the parent node of the basis function node
	 * 
	 */
	private void basisFunctionToXml(Document document, Element parentNode) {
		if (basisFunctionSet != null) {
			Element root = document.createElement(NODE_NAME_BASIS_FUNCTIONS);
			root.appendChild(basisFunctionSet.toXml(document, null));
			parentNode.appendChild(root);
		}
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getTransitionsModelSize()
	 */
	public long getTransitionsModelSize() {
		int size = 0;
		for (Variable variable : getVariableSet()) {
			for (String action : actionList) {
				IFunction<IProbabilities<String>> transition = getTransitionFunction(
						variable, action);
				if (transition != null) {
					size += transition.numberOfNodes();
				}
			}
		}
		return size;
	}

	/**
	 * 
	 * @return the basis functions set
	 * 
	 */
	public FunctionsSum getBasisFunctions() {
		if (basisFunctionSet == null) {
			return new FunctionsSum();
		}
		return basisFunctionSet;
	}

	/**
	 * 
	 * Add a basis function to the set of basis functions
	 * 
	 * @param basisFunctions
	 * 
	 */
	public void addBasisFunction(IPartitionFunction<Double> basisFunctions) {
		basisFunctionSet.add(basisFunctions);
	}

	/**
	 * 
	 * Add the constant basis function to the basis function set
	 * 
	 * 
	 */
	public void addConstantBasisFunction() {
		Tree<Double> h0 = new Tree<Double>();
		((Leaf<Double>) h0.getRoot()).set(1.0);
		basisFunctionSet.add(h0);
	}

	/**
	 * 
	 * A new reward function to the list of function
	 * 
	 * @param reward
	 *            the reward function to add
	 * 
	 */
	public void addReward(IPartitionFunction<Double> reward) {
		rewardFunction.add(reward);
	}

	/**
	 * 
	 * Load a sum of function
	 * 
	 * @param childNode
	 *            the root of the tree to load
	 * @return the loaded sum function
	 * 
	 */
	static public FunctionsSum loadFunctionSum(Element childNode) {
		FunctionsSum basisFunctionSet = new FunctionsSum();
		Element firstChild = getFirstChild(childNode);
		if (firstChild.getNodeName().equals(Rules.NODE_NAME)) {
			basisFunctionSet.loadXml(childNode, PlainRules.loadingPolicy);
		} else {
			basisFunctionSet.loadXml(childNode, Tree.loadingPolicy);
		}
		return basisFunctionSet;
	}

	/**
	 * 
	 * Load the reward tree from a xml node
	 * 
	 * @param xmlNode
	 *            the xml node to load the reward tree from
	 * 
	 */
	protected void loadRewardFunctions(Node xmlNode) {
		NodeList nodeList = xmlNode.getChildNodes();
		for (int index = 0; index < nodeList.getLength(); index++) {
			if (nodeList.item(index).getNodeType() == Node.ELEMENT_NODE) {
				Element rewardNode = (Element) nodeList.item(index);
				addReward(new Tree<Double>(rewardNode,
						new RewardLoadingPolicy()));
			}
		}
	}

	/**
	 * 
	 * Load a set of parameters
	 * 
	 * @param element
	 *            the root of the xml tree
	 * 
	 */
	final protected void loadParameters(Element element) {
		parameters.load(element);
	}

	/**
	 * 
	 * Set the set of basis function to use with this model
	 * 
	 * @param basisFunctionSet
	 *            the set of basis function to use
	 * 
	 */
	public void setBasisFunctions(FunctionsSum basisFunctionSet) {
		this.basisFunctionSet = basisFunctionSet;
	}
}
