/*
 * 
 * Created on Dec 29, 2005
 *
 */
package animatlab.kodamat.reinforcementlearning.model;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

import org.w3c.dom.Element;
import org.w3c.dom.Node;

import animatlab.kodamat.core.io.DataLoadingPolicy;
import animatlab.kodamat.core.io.IToXml.IToXmlPolicy;
import animatlab.kodamat.core.proba.IProbabilities;

/**
 * 
 * Static probability distribution
 * 
 * @author Thomas Degris
 * 
 */
public class StaticCPD implements IProbabilities<String> {
	/**
	 * 
	 * Attribute used in a xml node to specify the value to test
	 * 
	 */
	static public final String ATTRIBUTE_VALUE_NAME = "value";

	/**
	 * 
	 * Name of the attribute the to get the value associated with the probability
	 * 
	 */
	final public static String NODE_PROBABILITY_NAME = NAMESPACE + ":"
			+ "probability";

	/**
	 * 
	 * Name of the attribute the to get the probability for a value
	 * 
	 */
	final public static String ATTRIBUTE_PROBABILITY_NAME = "probability";

	/**
	 * 
	 * Probability distributions
	 * 
	 */
	final private Map<String, Double> variableValueDistribution = new LinkedHashMap<String, Double>();

	/**
	 * 
	 * Policy used to load StaticCPD from a xml node
	 * 
	 */
	final public static DataLoadingPolicy<IProbabilities<String>> LOADING_POLICY = new DataLoadingPolicy<IProbabilities<String>>() {
		/**
		 * @see animatlab.kodamat.core.io.DataLoadingPolicy#xmlToData(org.w3c.dom.Element)
		 */
		public IProbabilities<String> xmlToData(Element element) {
			return new StaticCPD(element);
		}
	};

	/**
	 * 
	 * Policy used to transform a set of probabilities into a xml node
	 * 
	 */
	final public static IToXmlPolicy<IProbabilities<String>> PROBABILITIES_POLICY = new IToXmlPolicy<IProbabilities<String>>() {
		public void toXml(Element parent, IProbabilities<String> data) {
			for (String value : data.getValueSet()) {
				Element valueNode = parent.getOwnerDocument().createElement(
						NODE_PROBABILITY_NAME);
				valueNode.setAttribute(ATTRIBUTE_VALUE_NAME, value);
				valueNode.setAttribute(ATTRIBUTE_PROBABILITY_NAME, String.valueOf(data
						.getProbability(value)));
				parent.appendChild(valueNode);
			}
		}
	};

	/**
	 * 
	 * Constructor
	 * 
	 */
	public StaticCPD() {
		super();
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param xmlNode
	 *          the xml node to load the distribution from
	 * 
	 */
	public StaticCPD(Element xmlNode) {
		for (int index = 0; index < xmlNode.getChildNodes().getLength(); index++) {
			Node childNode = xmlNode.getChildNodes().item(index);
			if ((childNode.getNodeType() == Node.ELEMENT_NODE)
					&& (childNode.getNodeName().equals(NODE_PROBABILITY_NAME))) {
				// Load the probability
				String value = childNode.getAttributes().getNamedItem(
						ATTRIBUTE_VALUE_NAME).getNodeValue();
				double probability = Double.parseDouble(childNode.getAttributes()
						.getNamedItem(ATTRIBUTE_PROBABILITY_NAME).getNodeValue());
				setProba(value, probability);
			}
		}
	}

	/**
	 * 
	 * Set a probability to this distribution
	 * 
	 * @param value
	 *          the value associated to the added probability
	 * @param probability
	 *          the probability associated to the value
	 * 
	 */
	public void setProba(String value, double probability) {
		variableValueDistribution.put(value, probability);
	}

	/**
	 * @see animatlab.kodamat.core.proba.IProbabilities#getBest()
	 */
	@SuppressWarnings("null")
	public String getBest() {
		Map.Entry<String, Double> bestProb = null;
		for (Map.Entry<String, Double> entry : variableValueDistribution.entrySet()) {
			if ((bestProb == null) || (entry.getValue() > bestProb.getValue())) {
				bestProb = entry;
			}
		}
		return bestProb.getKey();
	}

	/**
	 * 
	 * Get the probability associated with one value
	 * 
	 * @param value
	 *          the value to get the probability from
	 * 
	 * @return the probability of corresponding to the value
	 * 
	 */
	public double getProbability(String value) {
		Double probability = variableValueDistribution.get(value);
		if (probability == null) {
			return 0;
		}
		return probability;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuilder result = new StringBuilder();
		result.append("{");
		for (Map.Entry<String, Double> entry : variableValueDistribution.entrySet()) {
			result.append(entry.getKey());
			result.append("=");
			result.append(entry.getValue());
			result.append(",");
		}
		result.deleteCharAt(result.length() - 1);
		result.append("}");
		return result.toString();
	}

	/**
	 * @see animatlab.kodamat.core.proba.IProbabilities#getValueSet()
	 */
	public Set<String> getValueSet() {
		return variableValueDistribution.keySet();
	}

}
