package animatlab.kodamat.reinforcementlearning.model.forest;

import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.IModel;
import animatlab.kodamat.core.functions.FunctionsSum;
import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.functions.IPartition;
import animatlab.kodamat.core.functions.IPartitionFunction;
import animatlab.kodamat.core.io.HtmlPage;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.ClonePolicy;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.LeafListener;
import animatlab.kodamat.core.trees.MergePolicy;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Chrono;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.reinforcementlearning.core.IStepListener;
import animatlab.kodamat.reinforcementlearning.core.trajectory.Observation;
import animatlab.kodamat.reinforcementlearning.core.trajectory.Trajectory;
import animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel;
import animatlab.kodamat.reinforcementlearning.model.ILearningModel;
import animatlab.kodamat.supervisedlearning.core.Example;
import animatlab.kodamat.supervisedlearning.entropy.VariableMeasure;
import animatlab.kodamat.supervisedlearning.learningtree.ID4Tree;
import animatlab.kodamat.supervisedlearning.learningtree.LearningTree;
import animatlab.kodamat.supervisedlearning.regression.LeastSquareMeasure;
import animatlab.kodamat.supervisedlearning.regression.RegressionMeasure;
import animatlab.kodamat.supervisedlearning.regression.StructureLearningMeasure;
import animatlab.kodamat.supervisedlearning.utils.LearningToDotUtils;

/**
 * 
 * Abstrat learning forest
 * 
 * @author Thomas Degris
 * 
 */
public abstract class AbstractLearningForest extends AbstractForest implements
		ILearningModel {
	/**
	 * 
	 * Number of examples learned since the beginning
	 * 
	 */
	private int numberOfExamples = 0;

	/**
	 * 
	 * Flag used to indicate when the reward function has changed during learning
	 * 
	 */
	protected boolean rewardFunctionModifFlag;

	/**
	 * 
	 * Flag used to indicate when the structure of the model has changed during
	 * learning
	 * 
	 */
	protected boolean modelModifFlag;

	/**
	 * 
	 * Chrono used to measure real time learning
	 * 
	 */
	private Chrono chrono = new Chrono();

	/**
	 * 
	 * Policy to apply when the reward function is extracted from the forest
	 * 
	 */
	private RewardAdjustmentPolicy rewardAdjustmentPolicy = null;

	/**
	 * 
	 * Prototype used to populate the forest
	 * 
	 */
	final protected LearningTree<String, IProbabilities<String>> inductionTreePrototype;

	/**
	 * 
	 * Prototype used to populate the forest
	 * 
	 */
	final protected LearningTree<Double, Double> regressionTreePrototype;

	/**
	 * 
	 * Set used to know which action has been already tested at least one time
	 * 
	 */
	protected Set<String> actionTestedSet = new LinkedHashSet<String>();

	/**
	 * 
	 * Include the action to learn the reward
	 * 
	 */
	protected boolean includeActionInRewardLearningFlag = false;

	/**
	 * 
	 * Policy used to sum rewards function
	 * 
	 */
	protected MergePolicy<Double> rewardSumPolicy = MergePolicy.sumPolicy;

	/**
	 * 
	 * Ordering used to build example
	 * 
	 */
	protected VariableOrdering variableOrdering;

	/**
	 * 
	 * Constructor
	 * 
	 * @param actionList
	 *          the set of actions to learn the model
	 * @param variableOrdering
	 *          legend associated to the variables
	 * @param learningTree
	 *          learning tree used to learn the world model
	 * @param regressionTree
	 *          reward tree to used in the forest (may be null)
	 * 
	 */
	public AbstractLearningForest(List<String> actionList,
			VariableOrdering variableOrdering,
			LearningTree<String, IProbabilities<String>> learningTree,
			LearningTree<Double, Double> regressionTree) {
		super(actionList);
		inductionTreePrototype = learningTree;
		if (regressionTree == null) {
			LeastSquareMeasure leastSquareError = new LeastSquareMeasure(null);
			regressionTreePrototype = new ID4Tree.Plain<Double>(leastSquareError,
					variableOrdering);
		} else {
			regressionTreePrototype = regressionTree;
		}
		setVariableOrdering(variableOrdering);
	}

	/**
	 * 
	 * Populate this forest by incremental learning tree
	 * 
	 * @param stateSize
	 *          size of the state in the problem
	 * 
	 */
	protected void populateForest(int stateSize) {
		if (getParameters().hasKey(
				Parameters.PARAM_KEY_TRANSITION_LEARNING_THRESHOLD)) {
			inductionTreePrototype.getMeasure().setThreshold(
					getParameters().getKeyD(
							Parameters.PARAM_KEY_TRANSITION_LEARNING_THRESHOLD));
		}
		forest = new LinkedHashMap<Variable, Map<String, Tree<IProbabilities<String>>>>();
		for (int index = 0; index < stateSize; index++) {
			Variable variable = variableOrdering.get(index);
			variableValueSet.put(variable, new LinkedHashSet<String>());
		}
		resetModifiedFlag();
	}

	/**
	 * 
	 * @return a new learning tree to learn the model
	 * 
	 */
	protected LearningTree<String, IProbabilities<String>> createLearningTree() {
		LearningTree<String, IProbabilities<String>> tree = inductionTreePrototype
				.newInstance();
		tree.setLeafListener(new LeafListener() {
			public void leafIsCreated(Leaf<?> leaf) {
				modelModifFlag = true;
			}

			public void leafIsRemoved(Leaf<?> leaf) {
				modelModifFlag = true;
			}
		});
		tree.setVariableOrdering(variableOrdering);
		return tree;
	}

	/**
	 * 
	 * Populate the list of rewards
	 * 
	 * @param rewardSize
	 *          the size of the list of reward to initialize
	 * 
	 */
	private void populateRewards(int rewardSize) {
		assert (rewardFunction.size() == 0);
		for (int rewardIndex = 0; rewardIndex < rewardSize; rewardIndex++) {
			LearningTree<Double, Double> rewardTree = regressionTreePrototype
					.newInstance();
			rewardTree.setLeafListener(new LeafListener() {
				public void leafIsCreated(Leaf<?> leaf) {
					rewardFunctionModifFlag = true;
				}

				public void leafIsRemoved(Leaf<?> leaf) {
					rewardFunctionModifFlag = true;
				}
			});
			rewardTree.setVariableOrdering(variableOrdering);
			addReward(rewardTree);
		}
	}

	/**
	 * 
	 * Get the reward regression function
	 * 
	 * @param rewardIndex
	 *          the index of the reward to get
	 * @return the regression function of the reward
	 * 
	 */
	@SuppressWarnings("unchecked")
	public LearningTree<Double, Double> getReward(int rewardIndex) {
		return (LearningTree<Double, Double>) rewardFunction.functions().get(
				rewardIndex);
	}

	/**
	 * 
	 * Learn the reward function of the problem
	 * 
	 * @param lastTimeState
	 *          last state the agent was
	 * @param lastTimeAction
	 *          last action the agent did
	 * @param reward
	 *          reward received by the agent
	 * 
	 */
	private void learnReward(Context lastTimeState, String lastTimeAction,
			List<Double> reward) {
		List<String> values = variableOrdering.stateToList(lastTimeState);
		if (includeActionInRewardLearningFlag) {
			values.add(lastTimeAction);
		}
		double rewardSum = 0;
		for (int rewardIndex = 0; rewardIndex < reward.size(); rewardIndex++) {
			double currentReward = reward.get(rewardIndex);
			assert (!Double.isNaN(currentReward));
			final LearningTree<Double, Double> rewardTree = getReward(rewardIndex);
			rewardTree.learn(new Example<Double>(currentReward, values));
			rewardSum = rewardSum + currentReward;
		}
		rewardMax = Math.max(rewardMax, rewardSum);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.ILearningModel#learn(Context,
	 *      java.lang.String, Context, List)
	 */
	public synchronized void learn(Context lastTimeState, String lastTimeAction,
			Context currentState, List<Double> reward) {
		registerValues(currentState);
		if (forest == null) {
			if (currentState == null) {
				return;
			}
			populateForest(currentState.size());
		}
		if ((rewardFunction.size() == 0) && (reward != null)) {
			populateRewards(reward.size());
		}

		if (lastTimeState == null) {
			return;
		}

		learnReward(lastTimeState, lastTimeAction, reward);
		if (currentState != null) {
			numberOfExamples++;
			actionTestedSet.add(lastTimeAction);
			learnModel(lastTimeState, lastTimeAction, currentState);
		}
	}

	/**
	 * 
	 * Reset flags indicating if the functions have been modified during learning
	 * 
	 */
	public void resetModifiedFlag() {
		rewardFunctionModifFlag = false;
		modelModifFlag = false;
	}

	/**
	 * 
	 * Register the values of the variables from a state
	 * 
	 * @param state
	 *          the state to register the values from
	 * 
	 */
	private void registerValues(Context state) {
		if (state == null) {
			return;
		}
		for (Variable variable : getVariableSet()) {
			Set<String> valueSet = variableValueSet.get(variable);
			if (valueSet == null) {
				valueSet = new LinkedHashSet<String>();
				variableValueSet.put(variable, valueSet);
			}
			valueSet.add(state.get(variable));
		}
		multiValuedFlag = checkIfMultiValued();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel#getTransitionFunction(animatlab.kodamat.core.utils.Variable,
	 *      java.lang.String)
	 */
	@SuppressWarnings("unchecked")
	@Override
	public LearningTree<String, IProbabilities<String>> getTransitionFunction(
			Variable variable, String action) {
		return (LearningTree<String, IProbabilities<String>>) forest.get(variable)
				.get(action);
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getStateSize()
	 */
	public int getStateSize() {
		return forest.size();
	}

	/**
	 * 
	 * Learn the model for all the variable
	 * 
	 * @param lastTimeState
	 *          last state the agent was
	 * @param lastTimeAction
	 *          last action the agent did
	 * @param currentState
	 *          current state of the agent
	 * 
	 */
	abstract protected void learnModel(Context lastTimeState,
			String lastTimeAction, Context currentState);

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.forest.AbstractForest#getRewardTree(java.lang.String)
	 */
	@Override
	public Tree<Double> getRewardTree(String action) {
		Tree<Double> actionRewardTree = null;
		for (IFunction<Double> reward : getRewardFunction(action).functions()) {
			Tree<Double> currentReward = extractRewardFunction((Tree<Double>) reward,
					action);
			if (actionRewardTree == null) {
				actionRewardTree = currentReward;
			} else {
				actionRewardTree = rewardSumPolicy.merge(currentReward,
						actionRewardTree);
			}
		}

		if ((rewardAdjustmentPolicy == null) || (action == null)) {
			return actionRewardTree;
		}
		return rewardAdjustmentPolicy.adjust(action, actionRewardTree);
	}

	/**
	 * 
	 * Extract a reward function given a particular action
	 * 
	 * @param globalReward
	 *          the reward from which the action reward has to be extracted
	 * @param action
	 *          the action to extract the reward for
	 * @return the action reward tree
	 * 
	 */
	@Override
	public Tree<Double> extractRewardFunction(Tree<Double> globalReward,
			String action) {
		Context actionConstraint = null;
		if (action != null) {
			actionConstraint = new Context();
			actionConstraint.put(Variable.ACTION, action);
		}
		Tree<Double> result = globalReward.clone(new ClonePolicy<Double>() {
			public Double cloneData(Double data) {
				if (data == null) {
					return getRewardMax();
				}
				return data;
			}
		}, actionConstraint);
		// result.simplify(EqualPolicy.valueLeafEqualPolicy);
		return result;
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getName()
	 */
	@Override
	public String getName() {
		StringBuilder name = new StringBuilder();
		name.append(getTypeName() + ".");
		String prototypeName = getName(inductionTreePrototype);
		name.append(prototypeName);
		String rewardAlgoName = getName(regressionTreePrototype);
		if (!prototypeName.toString().equals(rewardAlgoName)) {
			name.append("." + rewardAlgoName);
		}
		return name.toString().toLowerCase();
	}

	/**
	 * 
	 * Extract the name of a prototype given an instance of it
	 * 
	 * @param prototype
	 *          the prototype to extract the name from
	 * @return the name of the prototype
	 * 
	 */
	@SuppressWarnings("unchecked")
	private String getName(LearningTree<?, ?> prototype) {
		Class protoClass = prototype.getClass();
		if (protoClass.getDeclaringClass() != null) {
			protoClass = protoClass.getDeclaringClass();
		}
		return protoClass.getSimpleName() + prototype.getMeasure().getName();
	}

	/**
	 * 
	 * @return the name of the type of the forest
	 * 
	 */
	abstract protected String getTypeName();

	/**
	 * @see animatlab.kodamat.core.IModel#getRewardMax()
	 */
	@SuppressWarnings("unchecked")
	public double getRewardMax() {
		return Math.max(rewardMax,
				((RegressionMeasure) ((LearningTree<Double, Double>) rewardFunction
						.functions().get(0)).getMeasure()).getMax());
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.ILearningModel#getActionTestedSet()
	 */
	public Set<String> getActionTestedSet() {
		return actionTestedSet;
	}

	/**
	 * 
	 * Build this forest by learning from a trajectory
	 * 
	 * @param trajectory
	 *          the trajectory to learn from
	 * @param stepListener
	 *          listener to call at each time step
	 * 
	 */
	public void learnFromTrajectory(Trajectory trajectory,
			IStepListener stepListener) {
		chrono = trajectory.getChrono();
		for (int currentTime = 0; currentTime < trajectory.getTime(); currentTime++) {
			Observation experiment = trajectory.getExperimentAtTime(currentTime);
			learn(experiment.state, experiment.action, experiment.nextState,
					experiment.rewardList);
			if (stepListener != null) {
				stepListener.listenStep(currentTime, experiment.state,
						experiment.action, experiment.nextState, experiment.rewardList);
			}
		}
		if (stepListener != null) {
			stepListener.terminate();
		}
		flush();
	}

	/**
	 * 
	 * Flush all remaining examples
	 * 
	 */
	@SuppressWarnings("unchecked")
	protected void flush() {
		for (IFunction<Double> learningReward : rewardFunction.functions()) {
			((LearningTree<Double, Double>) learningReward).flush();
		}
	}

	/**
	 * 
	 * @return check if all the leaves of the model contains something
	 * 
	 */
	protected boolean checkLearning() {
		for (String action : getActionList()) {
			for (Variable variable : getVariableSet()) {
				Tree<IProbabilities<String>> transition = getTransitionFunction(
						variable, action);
				for (IPartition<IProbabilities<String>> partition : transition
						.partitions()) {
					if ((partition.get() == null)
							|| (partition.get().getValueSet().size() == 0)) {
						return false;
					}
				}
			}
			for (IFunction<Double> reward : getRewardFunction(action).functions()) {
				for (IPartition<Double> partition : ((IPartitionFunction<Double>) reward)
						.partitions()) {
					if ((partition.get() == null) || Double.isNaN(partition.get())) {
						return false;
					}
				}
			}
		}
		return true;
	}

	/**
	 * 
	 * Set the policy to adjust the reward when extracted from the forest
	 * 
	 * @param rewardAdjustmentPolicy
	 *          the policy to use
	 * 
	 */
	public void setRewardAdjustmentPolicy(
			RewardAdjustmentPolicy rewardAdjustmentPolicy) {
		this.rewardAdjustmentPolicy = rewardAdjustmentPolicy;
	}

	/**
	 * 
	 * @return the current reward adjustement policy
	 * 
	 */
	public RewardAdjustmentPolicy getRewardAdjustmentPolicy() {
		return rewardAdjustmentPolicy;
	}

	/**
	 * 
	 * Dump this as a html page
	 * 
	 * @param folderPath
	 *          the folder where to output the html file and figures
	 * @param title
	 *          title of the results
	 * 
	 */
	public void dump(String folderPath, String title) {
		HtmlPage htmlDotPage = new HtmlPage(folderPath, "index.html");
		htmlDotPage.setHtmlPageSetTitle(title);
		htmlDotPage.setTitle("Index");
		dump(htmlDotPage);
		htmlDotPage.write();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.forest.AbstractForest#dump(animatlab.kodamat.core.io.HtmlPage)
	 */
	@Override
	public void dump(HtmlPage htmlPage) {
		htmlPage.append("Learning since " + chrono.toString() + ".<br/>");
		LearningToDotUtils<Double> learningToDotUtils = new LearningToDotUtils<Double>();
		for (int rewardIndex = 0; rewardIndex < getRewardFunction(null).size(); rewardIndex++) {
			LearningTree<Double, Double> reward = getReward(rewardIndex);
			learningToDotUtils.dumpTree(reward, htmlPage.newChild("Reward "
					+ rewardIndex));
		}
	}

	/**
	 * 
	 * Set the set of basis functions to use in this problem
	 * 
	 * @param basisFunctions
	 *          the set of basis function to use in this problem
	 * 
	 */
	public void setBasisFunction(FunctionsSum basisFunctions) {
		this.basisFunctionSet = basisFunctions.clone();
	}

	/**
	 * 
	 * @return the number of learned examples
	 * 
	 */
	public int getNumberOfLearnedExamples() {
		return numberOfExamples;
	}

	/**
	 * 
	 * Set the flag to decide weither or not the action is included in reward
	 * learning
	 * 
	 * @param includeActionInRewardLearningFlag
	 *          the new value of the flag
	 * 
	 */
	public void setIncludeActionInRewardLearningFlag(
			boolean includeActionInRewardLearningFlag) {
		this.includeActionInRewardLearningFlag = includeActionInRewardLearningFlag;
	}

	/**
	 * 
	 * @return the prototype used to create the tree used in the transition
	 *         function learning
	 * 
	 */
	public LearningTree<String, IProbabilities<String>> getInductionPrototype() {
		return inductionTreePrototype;
	}

	/**
	 * 
	 * @return returns true when the reward has changed during last learning
	 * 
	 */
	public boolean isRewardFunctionModifFlag() {
		return rewardFunctionModifFlag;
	}

	/**
	 * 
	 * @return returns true when the reward has changed during last learning
	 * 
	 */
	public boolean isModelModifFlag() {
		return modelModifFlag;
	}

	/**
	 * 
	 * @return the induction measure used to learn the model
	 * 
	 */
	public VariableMeasure<String> getInductionMeasure() {
		return inductionTreePrototype.getMeasure();
	}

	/**
	 * 
	 * Set the parameters of this from the definition of a problem and a forest
	 * used as prototype
	 * 
	 * @param forest
	 *          the forest used as prototype
	 * @param problem
	 *          the problem definition
	 * 
	 */
	protected void adjustParameters(AbstractLearningForest forest, IModel problem) {
		setIncludeActionInRewardLearningFlag(forest.includeActionInRewardLearningFlag);
		VariableMeasure<Double> regressionMeasure = regressionTreePrototype
				.getMeasure();
		if (regressionMeasure instanceof StructureLearningMeasure) {
			StructureLearningMeasure mbieebMeasure = (StructureLearningMeasure) regressionMeasure;
			if (Double.isNaN(mbieebMeasure.getBetaU())) {
				mbieebMeasure.setBetaU(problem.getRewardMax());
			}
		}
		regressionTreePrototype.setVariableOrdering(variableOrdering);
		inductionTreePrototype.setVariableOrdering(variableOrdering);
		if (problem instanceof AbstractFactoredModel) {
			getParameters().setAll(((AbstractFactoredModel) problem).getParameters());
		}
		getParameters().setAll(forest.getParameters());
	}

	/**
	 * 
	 * @return the prototype used to learn the reward functions
	 * 
	 */
	public LearningTree<Double, Double> getRegressionPrototype() {
		return regressionTreePrototype;
	}

	/**
	 * 
	 * Get the number of examples estimating the current probability distribution
	 * for an action, a variable and a given context
	 * 
	 * @param action
	 *          the action to get the number of example for
	 * @param variable
	 *          the variable to get the number of example for
	 * @param context
	 *          the context that the probability distribution has to satisfy
	 * @return the number of examples found
	 * 
	 */
	abstract public int getNumberOfExamples(String action, Variable variable,
			Context context);

	/**
	 * 
	 * Set the policy used to sum rewards
	 * 
	 * @param rewardSumPolicy
	 *          the new policy used to sum rewards function
	 * 
	 */
	public void setRewardSumPolicy(MergePolicy<Double> rewardSumPolicy) {
		this.rewardSumPolicy = rewardSumPolicy;
	}

	/**
	 * 
	 * Set the legend for this FMDP
	 * 
	 * @param variableOrdering
	 *          the legend to set
	 * 
	 */
	public void setVariableOrdering(VariableOrdering variableOrdering) {
		if (variableOrdering != null) {
			assert (variableOrdering.size() > 0);
		}
		this.variableOrdering = variableOrdering;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.ILearningModel#getVariableOrdering()
	 */
	public VariableOrdering getVariableOrdering() {
		return variableOrdering;
	}
}
