/*
 * 
 * Created on Dec 29, 2005
 *
 */
package animatlab.kodamat.reinforcementlearning.model.forest;

import static animatlab.kodamat.core.io.XmlUtils.getFirstChild;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import animatlab.kodamat.core.IProblemDefinition;
import animatlab.kodamat.core.conditions.IStateCondition;
import animatlab.kodamat.core.conditions.VariableCondition;
import animatlab.kodamat.core.functions.FunctionsSum;
import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.functions.IPartition;
import animatlab.kodamat.core.functions.IPartitionFunction;
import animatlab.kodamat.core.io.DotUtils;
import animatlab.kodamat.core.io.HtmlPage;
import animatlab.kodamat.core.io.IToDot;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.rules.PlainRule;
import animatlab.kodamat.core.rules.PlainRules;
import animatlab.kodamat.core.trees.MergePolicy;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel;
import animatlab.kodamat.reinforcementlearning.model.StaticCPD;

/**
 * 
 * Load a static forest from a xml file
 * 
 * @author Thomas Degris
 * 
 */
public class StaticForest extends AbstractForest implements IProblemDefinition {
	/**
	 * 
	 * File to load to run the test
	 * 
	 */
	final static public String PROBLEM_EXTENSION = ".forest";

	/**
	 * 
	 * Name of the node of the set of cpt-trees for the terminal state
	 * 
	 */
	static public final String NODE_NAME_TERMINALSTATE = AbstractFactoredModel.NAMESPACE
			+ ":" + "terminal-state";

	/**
	 * 
	 * Name of the node of the set of cpt-trees for the initial state
	 * 
	 */
	static private final String NODE_NAME_INITIALSTATE = AbstractFactoredModel.NAMESPACE
			+ ":" + "initial-state";

	/**
	 * 
	 * Terminal state indicating the end of an experiment
	 * 
	 */
	private IStateCondition terminalStateCondition = null;

	/**
	 * 
	 * Conditions for a state to start an experiment
	 * 
	 */
	private IStateCondition initialStateCondition = null;

	/**
	 * 
	 * Default variable ordering
	 * 
	 */
	private VariableOrdering variableOrdering;

	/**
	 * 
	 * Constructor
	 * 
	 * @param variableOrdering
	 *          legend used in this forest
	 * @param actionList
	 *          action set used in this forest
	 * 
	 */
	public StaticForest(VariableOrdering variableOrdering, List<String> actionList) {
		super(actionList);
		forest = new LinkedHashMap<Variable, Map<String, Tree<IProbabilities<String>>>>();
		this.variableOrdering = variableOrdering;
	}

	/**
	 * 
	 * Constructor from a xml tree
	 * 
	 * @param element
	 *          root of the xml tree
	 * 
	 */
	public StaticForest(Element element) {
		super(new ArrayList<String>());
		assert (element.getNodeName().equals(AbstractForest.NODE_NAME_FOREST));
		forest = new LinkedHashMap<Variable, Map<String, Tree<IProbabilities<String>>>>();
		NodeList nodeList = element.getChildNodes();
		for (int index = 0; index < nodeList.getLength(); index++) {
			Node childNode = nodeList.item(index);
			if (childNode.getNodeType() == Node.ELEMENT_NODE) {
				if (childNode.getNodeName().equals(
						AbstractFactoredModel.NODE_NAME_REWARD)) {
					loadRewardFunctions(childNode);
				}
				if (childNode.getNodeName().equals(
						AbstractFactoredModel.NODE_NAME_TRANSITIONS)) {
					loadCpdTree(childNode);
				}
				if (childNode.getNodeName().equals(VariableOrdering.NODE_NAME_LEGEND)) {
					variableOrdering = new VariableOrdering((Element) childNode);
				}
				if (childNode.getNodeName().equals(Parameters.NODE_NAME_KEYS)) {
					loadParameters((Element) childNode);
				}
				if (childNode.getNodeName().equals(NODE_NAME_TERMINALSTATE)) {
					terminalStateCondition = VariableCondition
							.load(getFirstChild(childNode));
				}
				if (childNode.getNodeName().equals(NODE_NAME_INITIALSTATE)) {
					initialStateCondition = VariableCondition
							.load(getFirstChild(childNode));
				}
				if (childNode.getNodeName().equals(NODE_NAME_BASIS_FUNCTIONS)) {
					basisFunctionSet = loadFunctionSum(getFirstChild(childNode));
				}
			}
		}
		multiValuedFlag = checkIfMultiValued();
	}

	/**
	 * 
	 * @return flag weither to use initial/terminal conditions
	 * 
	 */
	private boolean episodeTimeoutFlag() {
		return getParameters().hasKey(Parameters.PARAM_KEY_EPISODE_TIME);
	}

	/**
	 * 
	 * Load the CPD trees from a NODE_NAME_CPDTREE node
	 * 
	 * @param xmlNode
	 *          the xml node to load the CPD trees from
	 * 
	 */
	private void loadCpdTree(Node xmlNode) {
		NodeList nodeList = xmlNode.getChildNodes();
		for (int actionIndex = 0; actionIndex < nodeList.getLength(); actionIndex++) {
			Node actionNode = nodeList.item(actionIndex);
			if ((actionNode.getNodeType() == Node.ELEMENT_NODE)
					&& (actionNode.getNodeName()
							.equals(AbstractFactoredModel.NODE_NAME_ACTION))) {
				String action = actionNode.getAttributes().getNamedItem(
						AbstractFactoredModel.ATTRIBUTE_NAME_ACTION).getNodeValue();
				actionList.add(action);
				loadActionTree(action, actionNode);
			}
		}
	}

	/**
	 * 
	 * Load a tree for an action a variable
	 * 
	 * @param action
	 *          action represented by actionNode
	 * @param actionNode
	 *          the node to load
	 * 
	 */
	private void loadActionTree(String action, Node actionNode) {
		NodeList variableNodeList = actionNode.getChildNodes();
		for (int variableIndex = 0; variableIndex < variableNodeList.getLength(); variableIndex++) {
			Node variableNode = variableNodeList.item(variableIndex);
			if (variableNode.getNodeType() == Node.ELEMENT_NODE) {
				Variable variable = Variable.load(variableNode);
				NodeList cpdNodeList = variableNode.getChildNodes();
				for (int elementIndex = 0; elementIndex < cpdNodeList.getLength(); elementIndex++) {
					Node cpdNode = cpdNodeList.item(elementIndex);
					if (cpdNode.getNodeType() == Node.ELEMENT_NODE) {
						setTransitionModel(action, variable,
								new Tree<IProbabilities<String>>((Element) cpdNode,
										StaticCPD.LOADING_POLICY));
					}
				}
			}
		}
	}

	/**
	 * 
	 * Set a probabilities distribution of the transition model
	 * 
	 * @param action
	 *          the action concerned by the probabilities distribution
	 * @param variable
	 *          the variable concerned by the probabilities distribution
	 * @param tree
	 *          the probabilities distribution
	 * 
	 */
	public void setTransitionModel(String action, Variable variable,
			Tree<IProbabilities<String>> tree) {
		Map<String, Tree<IProbabilities<String>>> variableModel = forest
				.get(variable);
		if (variableModel == null) {
			variableModel = new LinkedHashMap<String, Tree<IProbabilities<String>>>();
			forest.put(variable, variableModel);
		}
		variableModel.put(action, tree);

		Set<String> valueSet = variableValueSet.get(variable);
		if (valueSet == null) {
			valueSet = new LinkedHashSet<String>();
			variableValueSet.put(variable, valueSet);
		}
		for (IPartition<IProbabilities<String>> partition : tree.partitions()) {
			valueSet.addAll(partition.get().getValueSet());
		}
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel#getTransitionFunction(animatlab.kodamat.core.utils.Variable,
	 *      java.lang.String)
	 */
	@Override
	public Tree<IProbabilities<String>> getTransitionFunction(Variable variable,
			String action) {
		return forest.get(variable).get(action);
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getStateSize()
	 */
	public int getStateSize() {
		return forest.size();
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getRewardMax()
	 */
	public double getRewardMax() {
		return rewardMax;
	}

	/**
	 * @see animatlab.kodamat.core.IProblemDefinition#getNextState(Random,
	 *      Context, Context, java.lang.String)
	 */
	public List<Double> getNextState(Random random, Context state,
			Context nextState, String action) {
		return AbstractFactoredModel.computeNextStateWithTerminal(random, this,
				state, nextState, action);
	}

	/**
	 * @see animatlab.kodamat.core.IProblemDefinition#getInitialState(Random)
	 */
	public Context getInitialState(Random random) {
		Context initialState;
		do {
			initialState = new Context();
			AbstractFactoredModel.initializeState(variableOrdering, random,
					initialState, getValueSets(), getStateSize());
			if (initialStateCondition != null) {
				initialStateCondition.setToTrue(random, initialState);
			}
		} while (isTerminal(initialState));
		return initialState;
	}

	/**
	 * @see animatlab.kodamat.core.IProblemDefinition#isTerminal(Context)
	 */
	public boolean isTerminal(Context state) {
		if (terminalStateCondition == null) {
			return false;
		}
		if (episodeTimeoutFlag()) {
			return false;
		}
		return terminalStateCondition.isTrue(state);
	}

	/**
	 * @see animatlab.kodamat.core.io.IToXml#toXml(org.w3c.dom.Document,
	 *      animatlab.kodamat.core.io.IToXml.IToXmlPolicy)
	 */
	@Override
	public Element toXml(Document document, IToXmlPolicy<Object> policy) {
		Element forestNode = super.toXml(document, policy);
		if (variableOrdering != null) {
			forestNode.appendChild(variableOrdering.toXml(document, null));
		}
		if (terminalStateCondition != null) {
			Element terminalNode = document.createElement(NODE_NAME_TERMINALSTATE);
			terminalNode.appendChild(terminalStateCondition.toXml(document, null));
			forestNode.appendChild(terminalNode);
		}
		if (initialStateCondition != null) {
			Element initialNode = document.createElement(NODE_NAME_INITIALSTATE);
			initialNode.appendChild(initialStateCondition.toXml(document, null));
			forestNode.appendChild(initialNode);
		}
		return forestNode;
	}

	/**
	 * @see animatlab.kodamat.core.IProblemDefinition#getNumberOfStates()
	 */
	public long getNumberOfStates() {
		return getNumberOfStates(this);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.forest.AbstractForest#getRewardTree(java.lang.String)
	 */
	@Override
	public Tree<Double> getRewardTree(String action) {
		Tree<Double> result = null;
		for (IFunction<Double> reward : getRewardFunction(action).functions()) {
			result = MergePolicy.sumPolicy.merge(extractRewardFunction(
					(Tree<Double>) reward, action), result);
		}
		return result;
	}

	/**
	 * 
	 * Set the terminal condition of this forest
	 * 
	 * @param terminalCondition
	 *          the condition to set
	 * 
	 */
	public void setTerminalCondition(IStateCondition terminalCondition) {
		this.terminalStateCondition = terminalCondition;
	}

	/**
	 * 
	 * Set the initial condition of this forest
	 * 
	 * @param initialStateCondition
	 *          the condition to set
	 * 
	 */
	public void setInitialCondition(IStateCondition initialStateCondition) {
		this.initialStateCondition = initialStateCondition;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.forest.AbstractForest#dump(animatlab.kodamat.core.io.HtmlPage)
	 */
	@Override
	public void dump(HtmlPage htmlPage) {
		for (String action : getActionList()) {
			HtmlPage actionPage = htmlPage.newChild("Action " + action);
			for (Variable variable : getVariableSet()) {
				String name = "model"
						+ DotUtils.toNormalizeString(action + variable.toString());
				actionPage.append("Variable " + variable.toString() + " : <br/>\n");
				actionPage.append(DotUtils.toHtmlDot(getTransitionFunction(variable,
						action), null, actionPage.getFolder(), name.toLowerCase()));
			}
			FunctionsSum rewardFunction = getRewardFunction(action);
			for (int rewardIndex = 0; rewardIndex < rewardFunction.size(); rewardIndex++) {
				actionPage.append("Reward " + rewardIndex + ": <br/>\n");
				actionPage.append(DotUtils.toHtmlDot((IToDot<?>) rewardFunction
						.functions().get(rewardIndex), null, actionPage.getFolder(),
						"reward" + DotUtils.toNormalizeString(action)));
			}
		}
	}

	/**
	 * 
	 * Make all the determinist method noisy
	 * 
	 * @param noise
	 *          noise to set at the transitions
	 * 
	 */
	public void toNoisy(double noise) {
		for (String action : getActionList()) {
			for (Variable variable : getVariableSet()) {
				Set<String> valueSet = getValueSet(variable);
				for (IPartition<IProbabilities<String>> partition : getTransitionFunction(
						variable, action).partitions()) {
					if (partition.get().getValueSet().size() == 1) {
						StaticCPD cpd = (StaticCPD) partition.get();
						String bestValue = cpd.getBest();
						for (String value : valueSet) {
							if (value.equals(bestValue)) {
								cpd.setProba(value, 1 - noise);
							} else {
								cpd.setProba(value, noise / (valueSet.size() - 1));
							}
						}
					}
				}
			}
		}
		this.initialStateCondition = null;
		this.terminalStateCondition = null;
	}

	/**
	 * 
	 * Decompose a set of basis functions composed as partitions functions as a
	 * set of rules, each set containing only one rule
	 * 
	 * @param basisFunctions
	 *          the set of basis functions to decompose
	 * 
	 * @return a new decomposed set of rules
	 * 
	 */
	public FunctionsSum decomposeBasisFunction(FunctionsSum basisFunctions) {
		FunctionsSum result = new FunctionsSum();
		for (IFunction<Double> h : basisFunctions.functions()) {
			for (IPartition<Double> partition : ((IPartitionFunction<Double>) h)
					.partitions()) {
				if (Math.abs(partition.get()) > Parameters.EPSILON) {
					PlainRules<Double> plainRules = new PlainRules<Double>();
					PlainRule<Double> plainRule = new PlainRule<Double>(partition);
					plainRules.add(plainRule);
					result.add(plainRules);
				}
			}
		}
		return result;
	}

	/**
	 * @see animatlab.kodamat.core.IProblemDefinition#getVariableOrdering()
	 */
	public VariableOrdering getVariableOrdering() {
		return variableOrdering;
	}

}
