/*
 * 
 * Created on 28 avr. 2005
 *
 */
package animatlab.kodamat.reinforcementlearning.model.forest;

import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import animatlab.kodamat.core.IProblemDefinition;
import animatlab.kodamat.core.io.HtmlPage;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.reinforcementlearning.model.ILearningModel;
import animatlab.kodamat.supervisedlearning.core.Example;
import animatlab.kodamat.supervisedlearning.learningtree.ILearningVertex;
import animatlab.kodamat.supervisedlearning.learningtree.LearningTree;
import animatlab.kodamat.supervisedlearning.utils.LearningToDotUtils;

/**
 * 
 * Spi forest composed of incremental learning tree
 * 
 * @author Thomas Degris
 * 
 */
public class VariableActionLearningForest extends AbstractLearningForest {
	/**
	 * 
	 * Constructor
	 * 
	 * @param actionList
	 *          the set of actions to learn the model
	 * @param variableOrdering
	 *          legend associated to the variables
	 * @param learningTree
	 *          learning tree used to learn the world model
	 * @param regressionTree
	 *          reward tree to used in the forest (may be null)
	 * 
	 */
	public VariableActionLearningForest(List<String> actionList,
			VariableOrdering variableOrdering,
			LearningTree<String, IProbabilities<String>> learningTree,
			LearningTree<Double, Double> regressionTree) {
		super(actionList, variableOrdering, learningTree, regressionTree);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.forest.AbstractLearningForest#populateForest(int)
	 */
	@Override
	protected void populateForest(int stateSize) {
		super.populateForest(stateSize);
		for (int variableIndex = 0; variableIndex < stateSize; variableIndex++) {
			Map<String, Tree<IProbabilities<String>>> variableModel = new LinkedHashMap<String, Tree<IProbabilities<String>>>();
			for (String action : getActionList()) {
				LearningTree<String, IProbabilities<String>> tree = createLearningTree();
				variableModel.put(action, tree);
			}
			forest.put(variableOrdering.get(variableIndex), variableModel);
		}
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.forest.AbstractLearningForest#learnModel(Context,
	 *      java.lang.String, Context)
	 */
	@SuppressWarnings("unchecked")
	@Override
	protected void learnModel(Context lastTimeState, String lastTimeAction,
			Context currentState) {
		List<String> lastState = variableOrdering.stateToList(lastTimeState);
		for (Variable variable : getVariableSet()) {
			Map<String, Tree<IProbabilities<String>>> variableModel = forest
					.get(variable);
			LearningTree<String, IProbabilities<String>> learningTree = (LearningTree<String, IProbabilities<String>>) variableModel
					.get(lastTimeAction);
			learningTree.learn(new Example<String>(currentState.get(variable),
					lastState));
		}
	}

	// /**
	// * @see
	// animatlab.kodamat.core.IModel#getValueSet(animatlab.kodamat.core.utils.Variable)
	// **/
	// @Override
	// public Set<String> getValueSet(Variable variable) {
	// Set<String> result = new LinkedHashSet<String>();
	// for (Tree<IProbabilities<String>> actionModel :
	// forest.get(variable.variable).values()) {
	// LearningTree<String, IProbabilities<String>> learningTree =
	// (LearningTree<String, IProbabilities<String>>) actionModel;
	// result.addAll(learningTree.getStats().getValueSet());
	// }
	// return result;
	// }
	//
	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.forest.AbstractLearningForest#getTypeName()
	 */
	@Override
	protected String getTypeName() {
		return "variableaction";
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel#getTransitionFunction(animatlab.kodamat.core.utils.Variable,
	 *      java.lang.String)
	 */
	@SuppressWarnings("unchecked")
	@Override
	public LearningTree<String, IProbabilities<String>> getTransitionFunction(
			Variable variable, String action) {
		LearningTree<String, IProbabilities<String>> learningTree = (LearningTree<String, IProbabilities<String>>) forest
				.get(variable).get(action);
		if (learningTree.getStats().getNumberOfExamples() == 0) {
			return null;
		}
		return learningTree;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.ILearningModel#newInstance(IProblemDefinition)
	 */
	public ILearningModel newInstance(IProblemDefinition problem) {
		VariableActionLearningForest result = new VariableActionLearningForest(
				problem.getActionList(), problem.getVariableOrdering(),
				inductionTreePrototype.newInstance(), regressionTreePrototype
						.newInstance());
		result.adjustParameters(this, problem);
		return result;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.forest.AbstractLearningForest#flush()
	 */
	@SuppressWarnings("unchecked")
	@Override
	protected void flush() {
		super.flush();
		for (Map<String, Tree<IProbabilities<String>>> actionModel : forest
				.values()) {
			for (Tree<IProbabilities<String>> learningTree : actionModel.values()) {
				((LearningTree<String, IProbabilities<String>>) learningTree).flush();
			}
		}
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.forest.AbstractForest#dump(animatlab.kodamat.core.io.HtmlPage)
	 */
	@Override
	public void dump(HtmlPage htmlPage) {
		LearningToDotUtils<IProbabilities<String>> learningToDotUtils = new LearningToDotUtils<IProbabilities<String>>();
		for (String action : getActionList()) {
			htmlPage.appendToContent("Action " + action + " :");
			for (Variable variable : getVariableSet()) {
				learningToDotUtils.dumpTree(getTransitionFunction(variable, action),
						htmlPage
								.newChild(action + variable.toString(), variable.toString()));
			}
			htmlPage.appendToContent("</br>");
		}
		super.dump(htmlPage);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.forest.AbstractLearningForest#getNumberOfExamples(java.lang.String,
	 *      animatlab.kodamat.core.utils.Variable,
	 *      animatlab.kodamat.core.utils.Context)
	 */
	@SuppressWarnings("unchecked")
	@Override
	public int getNumberOfExamples(String action, Variable variable,
			Context context) {
		Tree<IProbabilities<String>> tree = getTransitionFunction(variable, action);
		if (tree == null) {
			return 0;
		}
		ILearningVertex<String> vertex = (ILearningVertex<String>) tree
				.getVertex(context);
		return vertex.getStats().getNumberOfExamples();
	}
}
