/*
 * 
 * Created on 28 avr. 2005
 *
 */
package animatlab.kodamat.reinforcementlearning.model.forest;

import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import animatlab.kodamat.core.IProblemDefinition;
import animatlab.kodamat.core.io.DotUtils;
import animatlab.kodamat.core.io.HtmlPage;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.reinforcementlearning.model.ILearningModel;
import animatlab.kodamat.supervisedlearning.core.Example;
import animatlab.kodamat.supervisedlearning.learningtree.ILearningVertex;
import animatlab.kodamat.supervisedlearning.learningtree.LearningTree;
import animatlab.kodamat.supervisedlearning.utils.LearningToDotUtils;

/**
 * 
 * Spi forest composed of incremental learning tree
 * 
 * @author Thomas Degris
 * 
 */
public class VariableLearningForest extends AbstractLearningForest {
	/**
	 * 
	 * Forest of CPD tree
	 * 
	 */
	private Map<Variable, LearningTree<String, IProbabilities<String>>> variablesForest;

	/**
	 * 
	 * Constructor
	 * 
	 * @param actionList
	 *          the set of actions to learn the model
	 * @param variableOrdering
	 *          legend associated to the variables
	 * @param learningTree
	 *          learning tree used to learn the world model
	 * @param regressionTree
	 *          reward tree to used in the forest (may be null)
	 * 
	 */
	public VariableLearningForest(List<String> actionList,
			VariableOrdering variableOrdering,
			LearningTree<String, IProbabilities<String>> learningTree,
			LearningTree<Double, Double> regressionTree) {
		super(actionList, variableOrdering, learningTree, regressionTree);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.forest.AbstractLearningForest#populateForest(int)
	 */
	@Override
	protected void populateForest(int stateSize) {
		super.populateForest(stateSize);
		variablesForest = new LinkedHashMap<Variable, LearningTree<String, IProbabilities<String>>>();
		for (int variableIndex = 0; variableIndex < stateSize; variableIndex++) {
			Variable variable = variableOrdering.get(variableIndex);
			variablesForest.put(variable, createLearningTree());
			forest.put(variable,
					new LinkedHashMap<String, Tree<IProbabilities<String>>>());
		}
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.forest.AbstractLearningForest#learnModel(Context,
	 *      java.lang.String, Context)
	 */
	@Override
	protected void learnModel(Context lastTimeState, String lastTimeAction,
			Context currentState) {
		for (Variable variable : getVariableSet()) {
			LearningTree<String, IProbabilities<String>> variableModel = variablesForest
					.get(variable);
			List<String> attributes = variableOrdering.stateToList(lastTimeState);
			attributes.add(lastTimeAction);
			Example<String> example = new Example<String>(currentState.get(variable),
					attributes);
			variableModel.learn(example);
			forest.get(variable).clear();
		}
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel#getTransitionFunction(animatlab.kodamat.core.utils.Variable,
	 *      java.lang.String)
	 */
	@SuppressWarnings("unchecked")
	@Override
	public LearningTree<String, IProbabilities<String>> getTransitionFunction(
			Variable variable, String action) {
		LearningTree<String, IProbabilities<String>> result = super
				.getTransitionFunction(variable, action);
		if (result == null) {
			actionConstraint.put(Variable.ACTION, action);
			result = (LearningTree<String, IProbabilities<String>>) variablesForest
					.get(variable).clone(actionConstraint);
			forest.get(variable).put(action, result);
		}
		return result;
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getTransitionsModelSize()
	 */
	@Override
	public long getTransitionsModelSize() {
		if (variablesForest == null) {
			return 0;
		}
		int size = 0;
		for (LearningTree<String, IProbabilities<String>> variableModel : variablesForest
				.values()) {
			size += variableModel.numberOfNodes();
		}
		return size;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.forest.AbstractLearningForest#getTypeName()
	 */
	@Override
	protected String getTypeName() {
		return "variable";
	}

	/**
	 * 
	 * Get the transition function for a variable
	 * 
	 * @param variable
	 *          the variable to get the transition from
	 * 
	 * @return the function representing the transition of the variable
	 * 
	 */
	public LearningTree<String, IProbabilities<String>> getTransitionFunction(
			Variable variable) {
		return variablesForest.get(variable);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.forest.AbstractForest#dump(animatlab.kodamat.core.io.HtmlPage)
	 */
	@Override
	public void dump(HtmlPage htmlPage) {
		LearningToDotUtils<IProbabilities<String>> learningToDotUtils = new LearningToDotUtils<IProbabilities<String>>();
		for (Variable variable : getVariableSet()) {
			learningToDotUtils.dumpTree(getTransitionFunction(variable), htmlPage
					.newChild("Variable " + variable.toString()));
		}
		super.dump(htmlPage);
		for (String action : getActionList()) {
			HtmlPage actionPage = htmlPage.newChild("Action " + action);
			for (Variable variable : getVariableSet()) {
				String name = "model"
						+ DotUtils.toNormalizeString(action + variable.toString());
				actionPage.append("Variable " + variable.toString() + " : <br/>\n");
				actionPage.append(DotUtils.toHtmlDot(getTransitionFunction(variable,
						action), null, actionPage.getFolder(), name.toLowerCase()));
			}
		}
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.ILearningModel#newInstance(IProblemDefinition)
	 */
	public ILearningModel newInstance(IProblemDefinition problem) {
		VariableLearningForest result = new VariableLearningForest(problem
				.getActionList(), problem.getVariableOrdering(), inductionTreePrototype
				.newInstance(), regressionTreePrototype.newInstance());
		result.adjustParameters(this, problem);
		return result;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.forest.AbstractLearningForest#flush()
	 */
	@SuppressWarnings("unchecked")
	@Override
	public void flush() {
		super.flush();
		for (Tree<IProbabilities<String>> learningTree : variablesForest.values()) {
			((LearningTree<String, IProbabilities<String>>) learningTree).flush();
		}
		assert (checkLearning());
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.forest.AbstractLearningForest#getNumberOfExamples(java.lang.String,
	 *      animatlab.kodamat.core.utils.Variable,
	 *      animatlab.kodamat.core.utils.Context)
	 */
	@SuppressWarnings("unchecked")
	@Override
	public int getNumberOfExamples(String action, Variable variable,
			Context context) {
		Context actionContext = new Context(context);
		actionContext.put(Variable.ACTION, action);
		Tree<IProbabilities<String>> tree = variablesForest.get(variable.variable);
		ILearningVertex<String> vertex = (ILearningVertex<String>) tree
				.getVertex(actionContext);
		return vertex.getStats().getNumberOfExamples();
	}
}
