/*
 * 
 * Created on 4 janv. 2006
 *
 */
package animatlab.kodamat.reinforcementlearning.model.forest.test;

import java.util.List;
import java.util.Random;

import junit.framework.TestCase;
import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.io.Log;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.problems.ProblemRunner;
import animatlab.kodamat.problems.generated.sysadmin.SysAdminProblem;
import animatlab.kodamat.problems.generated.sysadmin.UnidirectionalRing;
import animatlab.kodamat.problems.mazespace.MazeModel;
import animatlab.kodamat.reinforcementlearning.agent.ModelBasedAgent;
import animatlab.kodamat.reinforcementlearning.core.IStepListener;
import animatlab.kodamat.reinforcementlearning.core.acting.RandomActing;
import animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel;
import animatlab.kodamat.reinforcementlearning.model.ILearningModel;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractLearningForest;
import animatlab.kodamat.reinforcementlearning.model.forest.StaticForest;
import animatlab.kodamat.reinforcementlearning.model.forest.VariableActionLearningForest;
import animatlab.kodamat.reinforcementlearning.model.forest.VariableLearningForest;
import animatlab.kodamat.reinforcementlearning.model.tabular.TabularLearningModel;
import animatlab.kodamat.supervisedlearning.entropy.Chi2Measure;
import animatlab.kodamat.supervisedlearning.learningtree.ItiTree;
import animatlab.kodamat.supervisedlearning.learningtree.SID4Tree;
import animatlab.kodamat.supervisedlearning.rmax.RMaxTree;

/**
 * 
 * Test the learning method of learning forests in deterministic problem
 * 
 * @author Thomas Degris
 * 
 */
public class TestDeterministicLearningForest extends TestCase {

	/**
	 * 
	 * Random instance to use when initializing instances
	 * 
	 */
	final Random random = new Random(0);

	/**
	 * 
	 * Test a learning forest on a multiple reward problem
	 * 
	 */
	public void testLearningForestOnMultipleReward() {
		final int problemSize = 3;
		SysAdminProblem problem = new UnidirectionalRing();
		problem.setProblemData(problemSize);
		StaticForest sysAdminForest = problem.toForest();
		AbstractLearningForest forestTested = new VariableActionLearningForest(
				sysAdminForest.getActionList(), sysAdminForest.getVariableOrdering(),
				new SID4Tree.Prob<String>(new Chi2Measure(), sysAdminForest
						.getVariableOrdering()), null);
		ModelBasedAgent agent = new ModelBasedAgent(forestTested, new RandomActing(
				random, sysAdminForest.getActionList()));
		ProblemRunner problemRunner = new ProblemRunner(random, sysAdminForest,
				agent, Log.sysout);
		problemRunner.runProblem(1000);
		assertEquals(problemSize, forestTested.getRewardFunction(null).size());
		for (IFunction<Double> reward : forestTested.getRewardFunction(null)
				.functions()) {
			assertEquals(2, ((Tree<Double>) reward).numberOfLeaves());
		}
	}

	/**
	 * 
	 * Step listener called at each time step of the experiment to register the
	 * reward and the model
	 * 
	 */
	class StepListener implements IStepListener {
		/**
		 * 
		 * Current time
		 * 
		 */
		private int currentTime = 0;

		/**
		 * 
		 * Time where the test has to stop
		 * 
		 */
		private int stopTime = Integer.MAX_VALUE;

		/**
		 * 
		 * Transition registered from the experiment
		 * 
		 */
		final TabularLearningModel tabularModel;

		/**
		 * 
		 * Mazespace used in the experiment
		 * 
		 */
		final private MazeModel mazeSpace;

		/**
		 * 
		 * Constructor
		 * 
		 * @param mazeSpace
		 *          mazespace running the agent
		 * 
		 */
		StepListener(MazeModel mazeSpace) {
			this.mazeSpace = mazeSpace;
			tabularModel = new TabularLearningModel(mazeSpace.getVariableOrdering(),
					mazeSpace.getActionList());
		}

		/**
		 * @see animatlab.kodamat.reinforcementlearning.core.IStepListener#listenStep(int,
		 *      Context, java.lang.String, Context, List)
		 */
		public void listenStep(int time, Context currentState, String action,
				Context nextState, List<Double> reward) {
			if ((reward != null) && (Context.getSum(reward) != 0.0)) {
				stopTime = time + 100;
			}
			if (action != null) {
				tabularModel.learn(currentState, action, nextState, reward);
				if (nextState != null) {
					Context nextStateFromMaze = new Context();
					mazeSpace.getNextState(random, currentState, nextStateFromMaze,
							action);
					assertEquals(nextStateFromMaze, nextState);
					assertEquals(nextState, tabularModel.getTransition(currentState,
							action).getBest());
				}
			}
			currentTime = time;
		}

		/**
		 * 
		 * @return true if the test has to continue
		 * 
		 */
		boolean cont() {
			return (currentTime < stopTime);
		}

		/**
		 * @see animatlab.kodamat.reinforcementlearning.core.IStepListener#terminate()
		 */
		public void terminate() {
			return;
		}
	}

	/**
	 * 
	 * Test the learning forest on Mazespace
	 * 
	 * @param forestTested
	 *          the forest to test
	 * 
	 */
	private void testLearningForestWithMazeSpace(ILearningModel forestTested) {
		MazeModel mazeSpace = new MazeModel(MazeModel.SUTTON_MAZE,
				MazeModel.SUTTON_SIMPLE);
		StepListener stepListener = new StepListener(mazeSpace);
		ModelBasedAgent agent = new ModelBasedAgent(forestTested, new RandomActing(
				new Random(0), mazeSpace.getActionList()));
		ProblemRunner problemRunner = new ProblemRunner(random, mazeSpace, agent,
				Log.sysout);
		problemRunner.registerListener(stepListener);
		while (stepListener.cont()) {
			problemRunner.oneStep();
		}

		for (String action : stepListener.tabularModel.getActionTestedSet()) {
			for (Context state : stepListener.tabularModel.getStateSet(action)) {
				Context nextStateFromMaze = new Context();
				mazeSpace.getNextState(random, state, nextStateFromMaze, action);
				if (!nextStateFromMaze.isEmpty()) {
					Context nextStateFromTabular = stepListener.tabularModel
							.getTransition(state, action).getBest();
					assertEquals(nextStateFromMaze, nextStateFromTabular);
					Context nextStateFromLearningModel = new Context();
					List<Double> learnReward = AbstractFactoredModel.computeNextState(
							null, (AbstractFactoredModel) forestTested, state,
							nextStateFromLearningModel, action);
					assertEquals(nextStateFromTabular, nextStateFromLearningModel);
					assertEquals(stepListener.tabularModel.getReward(state, action),
							Context.getSum(learnReward));
				}
			}
		}
	}

	/**
	 * 
	 * Test the learning forest on Mazespace
	 * 
	 */
	public void testVariableActionLearningForestWithMazeSpace() {
		MazeModel mazeSpace = new MazeModel(MazeModel.SUTTON_MAZE,
				MazeModel.SUTTON_SIMPLE);
		Chi2Measure measure = new Chi2Measure(0.0);
		ItiTree itiTree = new ItiTree(measure, mazeSpace.getVariableOrdering());
		RMaxTree rewardTree = new RMaxTree(1, mazeSpace.getVariableOrdering());
		ILearningModel forestTested = new VariableActionLearningForest(mazeSpace
				.getActionList(), mazeSpace.getVariableOrdering(), itiTree, rewardTree);
		testLearningForestWithMazeSpace(forestTested);
	}

	/**
	 * 
	 * Test the learning forest on Mazespace
	 * 
	 */
	public void testVariableLearningForestWithMazeSpace() {
		MazeModel mazeSpace = new MazeModel(MazeModel.SUTTON_MAZE,
				MazeModel.SUTTON_SIMPLE);
		Chi2Measure measure = new Chi2Measure(0.0);
		ItiTree itiTree = new ItiTree(measure, mazeSpace.getVariableOrdering());
		RMaxTree rewardTree = new RMaxTree(1, mazeSpace.getVariableOrdering());
		ILearningModel forestTested = new VariableLearningForest(mazeSpace
				.getActionList(), mazeSpace.getVariableOrdering(), itiTree, rewardTree);
		testLearningForestWithMazeSpace(forestTested);
	}

}
