package animatlab.kodamat.reinforcementlearning.model.rules;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.functions.IPartitionFunction;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.rules.Rules;
import animatlab.kodamat.core.rules.RulesTree;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel;
import animatlab.kodamat.reinforcementlearning.model.forest.StaticForest;

/**
 * 
 * Rules model encapsulating a forest
 * 
 * @author Thomas Degris
 * 
 */
public class ForestRulesModel extends AbstractFactoredModel {
	/**
	 * 
	 * Reward max in this problem
	 * 
	 */
	final private double rewardMax;

	/**
	 * 
	 * Transition functions
	 * 
	 */
	final private Map<String, Map<Variable, RulesTree<IProbabilities<String>>>> transitions = new LinkedHashMap<String, Map<Variable, RulesTree<IProbabilities<String>>>>();

	/**
	 * 
	 * Value set for all the variables
	 * 
	 */
	final private Map<Variable, Set<String>> valueSet;

	/**
	 * 
	 * Constructor
	 * 
	 * @param forest
	 *          path of the static forest encapsulated by this model
	 * 
	 */
	public ForestRulesModel(StaticForest forest) {
		super(forest.getActionList());
		for (String action : forest.getActionList()) {
			for (Variable variable : forest.getVariableSet()) {
				registerTransitionTree(action, variable, forest.getTransitionFunction(
						variable, action));
			}
		}
		for (IFunction<Double> reward : forest.getRewardFunction(null).functions()) {
			addReward((IPartitionFunction<Double>) reward);
		}
		valueSet = forest.getValueSets();
		rewardMax = forest.getRewardMax();
		basisFunctionSet = forest.getBasisFunctions();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel#getTransitionFunction(animatlab.kodamat.core.utils.Variable,
	 *      java.lang.String)
	 */
	@Override
	public Rules<IProbabilities<String>> getTransitionFunction(Variable variable,
			String action) {
		return transitions.get(action).get(variable.variable);
	}

	/**
	 * 
	 * Register a new transition tree
	 * 
	 * @param action
	 *          the action to register
	 * @param variable
	 *          the variable to register
	 * @param tree
	 *          the transition function to register
	 * 
	 */
	public void registerTransitionTree(String action, Variable variable,
			Tree<IProbabilities<String>> tree) {
		Map<Variable, RulesTree<IProbabilities<String>>> variableTransition = transitions
				.get(action);
		if (variableTransition == null) {
			variableTransition = new LinkedHashMap<Variable, RulesTree<IProbabilities<String>>>();
			transitions.put(action, variableTransition);
		}
		variableTransition.put(variable,
				new RulesTree<IProbabilities<String>>(tree));
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel#getXmlRootNodeName()
	 */
	@Override
	protected String getXmlRootNodeName() {
		assert (false);
		return null;
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getValueSets()
	 */
	public Map<Variable, Set<String>> getValueSets() {
		return valueSet;
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getRewardMax()
	 */
	public double getRewardMax() {
		return rewardMax;
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getValueSet(animatlab.kodamat.core.utils.Variable)
	 */
	public Set<String> getValueSet(Variable variable) {
		return valueSet.get(variable);
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getStateSize()
	 */
	public int getStateSize() {
		return valueSet.keySet().size();
	}
}
