package animatlab.kodamat.reinforcementlearning.model.rules;

import static animatlab.kodamat.core.io.XmlUtils.getFirstChild;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import animatlab.kodamat.core.IProblemDefinition;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.rules.PlainRules;
import animatlab.kodamat.core.rules.Rule;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel;
import animatlab.kodamat.reinforcementlearning.model.StaticCPD;

/**
 * 
 * Static model (loaded from Xml) based on rules
 * 
 * @author Thomas Degris
 * 
 */
public class StaticRulesModel extends AbstractFactoredModel implements
		IProblemDefinition {
	/**
	 * 
	 * Name of the node of the reward function
	 * 
	 */
	static public final String NODE_NAME_RULES = AbstractFactoredModel.NAMESPACE
			+ ":" + "rules";

	/**
	 * 
	 * File to load to run the test
	 * 
	 */
	final static public String PROBLEM_EXTENSION = ".rules";

	/**
	 * 
	 * Random field used in the getNextState function and the model is stochastic
	 * 
	 */
	private final Random randomModel;

	/**
	 * 
	 * Set of possible value for each variables
	 * 
	 */
	private final Map<Variable, Set<String>> variableValueSet = new LinkedHashMap<Variable, Set<String>>();

	/**
	 * 
	 * Transition function of the problem
	 * 
	 */
	private final Map<Variable, Map<String, PlainRules<IProbabilities<String>>>> transitionFunction = new LinkedHashMap<Variable, Map<String, PlainRules<IProbabilities<String>>>>();

	/**
	 * 
	 * Variable ordering to use by default
	 * 
	 */
	private VariableOrdering variableOrdering;

	/**
	 * 
	 * Constructor
	 * 
	 * @param element
	 *          root of the xml tree
	 * @param random
	 *          random to use to build next states
	 * 
	 */
	public StaticRulesModel(Element element, Random random) {
		super(new ArrayList<String>());
		this.randomModel = random;
		assert (element.getNodeName().equals(NODE_NAME_RULES));
		NodeList nodeList = element.getChildNodes();
		for (int index = 0; index < nodeList.getLength(); index++) {
			Node childNode = nodeList.item(index);
			if (childNode.getNodeType() == Node.ELEMENT_NODE) {
				if (childNode.getNodeName().equals(
						AbstractFactoredModel.NODE_NAME_REWARD)) {
					loadRewardFunctions(getFirstChild(childNode));
				}
				if (childNode.getNodeName().equals(
						AbstractFactoredModel.NODE_NAME_TRANSITIONS)) {
					loadTransitions(childNode);
				}
				if (childNode.getNodeName().equals(VariableOrdering.NODE_NAME_LEGEND)) {
					variableOrdering = new VariableOrdering((Element) childNode);
				}
				if (childNode.getNodeName().equals(Context.NODE_NAME_VALUESET)) {
					Context.loadValueSet(getValueSets(), (Element) childNode);
				}
				if (childNode.getNodeName().equals(NODE_NAME_BASIS_FUNCTIONS)) {
					basisFunctionSet = loadFunctionSum(getFirstChild(childNode));
				}
			}
		}
	}

	/**
	 * 
	 * Load the transition model
	 * 
	 * @param childNode
	 *          root of the xml tree
	 * 
	 */
	private void loadTransitions(Node childNode) {
		NodeList nodeList = childNode.getChildNodes();
		for (int actionIndex = 0; actionIndex < nodeList.getLength(); actionIndex++) {
			Node actionNode = nodeList.item(actionIndex);
			if ((actionNode.getNodeType() == Node.ELEMENT_NODE)
					&& (actionNode.getNodeName()
							.equals(AbstractFactoredModel.NODE_NAME_ACTION))) {
				loadTransitionVariables((Element) actionNode);
			}
		}
	}

	/**
	 * 
	 * Load the transitions for a variables
	 * 
	 * @param actionElement
	 *          the element corresponding to the action
	 * 
	 */
	private void loadTransitionVariables(Element actionElement) {
		String actionName = actionElement
				.getAttribute(AbstractFactoredModel.ATTRIBUTE_NAME_ACTION);
		actionList.add(actionName);
		NodeList variableNodeList = actionElement.getChildNodes();
		for (int variableNodeIndex = 0; variableNodeIndex < variableNodeList
				.getLength(); variableNodeIndex++) {
			Node variableNode = variableNodeList.item(variableNodeIndex);
			if (variableNode.getNodeType() == Node.ELEMENT_NODE) {
				Element variableElement = (Element) variableNode;
				PlainRules<IProbabilities<String>> rules = new PlainRules<IProbabilities<String>>(
						getFirstChild(variableElement), StaticCPD.LOADING_POLICY);
				Variable variable = Variable.load(variableElement);
				Map<String, PlainRules<IProbabilities<String>>> variableModel = transitionFunction
						.get(variable);
				if (variableModel == null) {
					variableModel = new LinkedHashMap<String, PlainRules<IProbabilities<String>>>();
					transitionFunction.put(variable, variableModel);
				}
				variableModel.put(actionName, rules);
				registerValues(rules);
			}
		}
	}

	/**
	 * 
	 * Register the values used in each variables
	 * 
	 * @param rules
	 *          the rules where to get the information
	 * 
	 */
	private void registerValues(PlainRules<IProbabilities<String>> rules) {
		for (Variable variable : rules.scope()) {
			Set<String> valueSet = getValueSet(variable);
			if (valueSet == null) {
				valueSet = new LinkedHashSet<String>();
				getValueSets().put(variable, valueSet);
			}
			for (Rule<IProbabilities<String>> rule : rules) {
				String value = rule.getContext().get(variable);
				if (value != null) {
					valueSet.add(value);
				}
			}
		}
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel#getTransitionFunction(animatlab.kodamat.core.utils.Variable,
	 *      java.lang.String)
	 */
	@Override
	public PlainRules<IProbabilities<String>> getTransitionFunction(
			Variable variable, String action) {
		return transitionFunction.get(variable).get(action);
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getRewardMax()
	 */
	public double getRewardMax() {
		assert (false);
		return 0;
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getValueSet(animatlab.kodamat.core.utils.Variable)
	 */
	public Set<String> getValueSet(Variable variable) {
		return getValueSets().get(variable);
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getStateSize()
	 */
	public int getStateSize() {
		return transitionFunction.size();
	}

	/**
	 * @see animatlab.kodamat.core.IProblemDefinition#getNextState(Random,
	 *      Context, Context, java.lang.String)
	 */
	public List<Double> getNextState(Random random, Context state,
			Context nextState, String action) {
		return AbstractFactoredModel.computeNextStateWithTerminal(randomModel,
				this, state, nextState, action);
	}

	/**
	 * @see animatlab.kodamat.core.IProblemDefinition#getInitialState(Random)
	 */
	public Context getInitialState(Random random) {
		Context initialState = new Context();
		AbstractFactoredModel.initializeState(getVariableOrdering(), randomModel,
				initialState, getValueSets(), getStateSize());
		return initialState;
	}

	/**
	 * @see animatlab.kodamat.core.IProblemDefinition#isTerminal(Context)
	 */
	public boolean isTerminal(Context state) {
		return false;
	}

	/**
	 * @see animatlab.kodamat.core.IProblemDefinition#getNumberOfStates()
	 */
	public long getNumberOfStates() {
		return getNumberOfStates(this);
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getValueSets()
	 */
	public Map<Variable, Set<String>> getValueSets() {
		return variableValueSet;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel#getXmlRootNodeName()
	 */
	@Override
	protected String getXmlRootNodeName() {
		return AbstractFactoredModel.NAMESPACE + ":" + "rules";
	}

	/**
	 * @see animatlab.kodamat.core.IProblemDefinition#getVariableOrdering()
	 */
	public VariableOrdering getVariableOrdering() {
		return variableOrdering;
	}
}
