/*
 * 
 * Created on Jan 17, 2006
 *
 */
package animatlab.kodamat.reinforcementlearning.model.tabular;

import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.IProblemDefinition;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.reinforcementlearning.model.ILearningModel;
import animatlab.kodamat.supervisedlearning.core.Example;
import animatlab.kodamat.supervisedlearning.tabular.TabularLearning;

/**
 * 
 * Learning model using tabular representation
 * 
 * @author Thomas Degris
 * 
 */
public class TabularLearningModel implements ILearningModel {
	/**
	 * 
	 * Legend used to interprete examples
	 * 
	 */
	private VariableOrdering variableOrdering;

	/**
	 * 
	 * Map of action/state/state at time t+1/number of time it occured
	 * 
	 */
	private final Map<String, TabularLearning<Context>> model = new LinkedHashMap<String, TabularLearning<Context>>();

	/**
	 * 
	 * Model of the reward function Map of action/state/reward/number of time it
	 * occured
	 * 
	 */
	private final TabularLearning<Double> rewardModel;

	/**
	 * 
	 * Set of action available for the agent
	 * 
	 */
	private List<String> actionList;

	/**
	 * 
	 * Map to keep track of tested state with actions
	 * 
	 */
	private Map<String, Set<Context>> actionStateTested = new LinkedHashMap<String, Set<Context>>();

	/**
	 * 
	 * Reward maximum received by the agent
	 * 
	 */
	private double rewardMax = -Double.NaN;

	/**
	 * 
	 * Constructor
	 * 
	 * @param variableOrdering
	 *          legend to interprete examples
	 * @param actionList
	 *          set of actions available for the agent
	 * 
	 */
	public TabularLearningModel(VariableOrdering variableOrdering,
			List<String> actionList) {
		this.variableOrdering = variableOrdering;
		rewardModel = new TabularLearning<Double>(variableOrdering);
		this.actionList = actionList;
		for (String action : actionList) {
			model.put(action, new TabularLearning<Context>(variableOrdering));
			actionStateTested.put(action, new LinkedHashSet<Context>());
		}
	}

	/**
	 * 
	 * Constructor
	 * 
	 * 
	 */
	public TabularLearningModel() {
		variableOrdering = null;
		rewardModel = null;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.ILearningModel#learn(Context,
	 *      java.lang.String, Context, List)
	 */
	public void learn(Context lastTimeState, String lastTimeAction,
			Context currentState, List<Double> reward) {
		if (lastTimeState == null) {
			return;
		}
		actionStateTested.get(lastTimeAction).add(lastTimeState);
		learnReward(lastTimeState, lastTimeAction, reward);
		if (currentState != null) {
			learnModel(lastTimeState, lastTimeAction, currentState);
		}
	}

	/**
	 * 
	 * Learn the reward function of the problem
	 * 
	 * @param lastTimeState
	 *          last state the agent was
	 * @param lastTimeAction
	 *          last action the agent did
	 * @param rewards
	 *          reward received by the agent
	 * 
	 */
	private void learnReward(Context lastTimeState, String lastTimeAction,
			List<Double> rewards) {
		assert (rewards != null);
		double reward = Context.getSum(rewards);
		if (Double.isNaN(rewardMax)) {
			rewardMax = reward;
		} else {
			rewardMax = Math.max(reward, rewardMax);
		}
		rewardModel.learn(new Example<Double>(reward, variableOrdering
				.stateToList(lastTimeState)));
	}

	/**
	 * 
	 * Learn the model for all the variable
	 * 
	 * @param lastTimeState
	 *          last state the agent was
	 * @param lastTimeAction
	 *          last action the agent did
	 * @param currentState
	 *          current state of the agent
	 * 
	 */
	private void learnModel(Context lastTimeState, String lastTimeAction,
			Context currentState) {
		model.get(lastTimeAction).learn(
				new Example<Context>(currentState, variableOrdering
						.stateToList(lastTimeState)));
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.ILearningModel#getActionTestedSet()
	 */
	public Set<String> getActionTestedSet() {
		return model.keySet();
	}

	/**
	 * 
	 * The possible transitions (and their probabilities) from a state and an
	 * action
	 * 
	 * @param state
	 *          state at time t
	 * @param action
	 *          action executed at time t
	 * 
	 * @return the probabilities for state at time t+1
	 * 
	 */
	public IProbabilities<Context> getTransition(Context state, String action) {
		return model.get(action).classify(state);
	}

	/**
	 * 
	 * The reward from a state and a reward
	 * 
	 * @param state
	 *          state at time t
	 * @param action
	 *          action executed at time t
	 * 
	 * @return the reward received at time t+1
	 * 
	 */
	public Double getReward(Context state, String action) {
		return rewardModel.classify(state).getBest();
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getName()
	 */
	public String getName() {
		return getClass().getSimpleName();
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getTransitionsModelSize()
	 */
	public long getTransitionsModelSize() {
		int result = 0;
		for (TabularLearning<Context> actionModel : model.values()) {
			result += actionModel.numberOfNodes();
		}
		return result;
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getRewardModelSize()
	 */
	public int getRewardModelSize() {
		return rewardModel.numberOfNodes();
	}

	/**
	 * 
	 * Get the set of state where an action has been executed
	 * 
	 * @param action
	 *          the action to get the set of state
	 * 
	 * @return the set of state where this action has been executed
	 * 
	 */
	public Set<Context> getStateSet(String action) {
		return actionStateTested.get(action);
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getActionList()
	 */
	public List<String> getActionList() {
		return actionList;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "Action: \n" + model.toString() + "\n" + "Reward: \n"
				+ rewardModel.toString();
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getRewardMax()
	 */
	public double getRewardMax() {
		if (Double.isNaN(rewardMax)) {
			return 0;
		}
		return rewardMax;
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getValueSet(animatlab.kodamat.core.utils.Variable)
	 */
	public Set<String> getValueSet(Variable variable) {
		assert (false);
		return null;
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getStateSize()
	 */
	public int getStateSize() {
		// Not implemented
		assert (false);
		return 0;

	}

	/**
	 * @see animatlab.kodamat.core.IModel#getVariableSet()
	 */
	public Set<Variable> getVariableSet() {
		assert (false);
		return null;
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getValueSets()
	 */
	public Map<Variable, Set<String>> getValueSets() {
		assert (false);
		return null;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.ILearningModel#newInstance(IProblemDefinition)
	 */
	public ILearningModel newInstance(IProblemDefinition problem) {
		return new TabularLearningModel(problem.getVariableOrdering(), problem
				.getActionList());
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.ILearningModel#getVariableOrdering()
	 */
	public VariableOrdering getVariableOrdering() {
		return variableOrdering;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.ILearningModel#setVariableOrdering(animatlab.kodamat.core.utils.VariableOrdering)
	 */
	public void setVariableOrdering(VariableOrdering variableOrdering) {
		this.variableOrdering = variableOrdering;
	}

}
