/*
 * 
 * Created on Jan 17, 2006
 *
 */
package animatlab.kodamat.reinforcementlearning.planning.dyna;

import static animatlab.kodamat.core.utils.Parameters.EPSILON;

import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.IModel;
import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.io.Log;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.reinforcementlearning.model.ILearningModel;
import animatlab.kodamat.reinforcementlearning.model.tabular.TabularLearningModel;
import animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning;

/**
 * 
 * The algorithm DynaQ
 * 
 * @author Thomas Degris
 * 
 */
public class DynaQ implements IIncrementalPlanning {
	/**
	 * 
	 * Wrapper to expose the IFunction interface with a Java map
	 * 
	 */
	private class QualityMap extends LinkedHashMap<Context, Double> implements
			IFunction<Double> {
		/**
		 * SerialVersionUID
		 */
		private static final long serialVersionUID = -696679343066722739L;

		/**
		 * 
		 * Constructor
		 * 
		 */
		QualityMap() {
			super();
		}

		/**
		 * @see animatlab.kodamat.core.functions.IFunction#classify(Context)
		 */
		public Double classify(Context state) {
			Double result = get(state);
			if (result == null) {
				result = 0.0;
			}
			return result;
		}

		/**
		 * @see animatlab.kodamat.core.functions.IFunction#numberOfNodes()
		 */
		public int numberOfNodes() {
			return size();
		}

		/**
		 * @see java.lang.Object#clone()
		 */
		@Override
		public IFunction<Double> clone() {
			assert (false);
			return null;
		}
	}

	/**
	 * 
	 * Alpha constant used in DYNA-Q
	 * 
	 */
	private static final double ALPHA = 1.0;

	/**
	 * 
	 * Gamma value used in planning computation
	 * 
	 */
	public double gamma = Parameters.GAMMA;

	/**
	 * 
	 * Quality functions updated with the planning phase
	 * 
	 */
	final Map<String, QualityMap> qualityFunctions = new LinkedHashMap<String, QualityMap>();

	/**
	 * 
	 * Buffer listing all the possible states to avoid to compute something to
	 * know the size of the value function
	 * 
	 */
	final Set<Context> statesSet = new LinkedHashSet<Context>();

	/**
	 * 
	 * Tabular model
	 * 
	 */
	final TabularLearningModel model;

	/**
	 * 
	 * Optimal value function
	 * 
	 */
	final IFunction<Double> optimalValueFunction = new IFunction<Double>() {
		public Double classify(Context state) {
			if (model.getActionTestedSet().isEmpty()) {
				return null;
			}
			double maxValue = -Double.MAX_VALUE;
			for (String action : model.getActionTestedSet()) {
				Double value = qualityFunctions.get(action).get(state);
				if (value == null) {
					value = -Double.MAX_VALUE;
				}
				maxValue = Math.max(maxValue, value);
			}
			if (maxValue == -Double.MAX_VALUE) {
				return null;
			}
			return maxValue;
		}

		public int numberOfNodes() {
			return statesSet.size();
		}

		@Override
		public IFunction<Double> clone() {
			assert (false);
			return null;
		}
	};

	/**
	 * 
	 * Greedy policy
	 * 
	 */
	final IFunction<Set<String>> greedyPolicy = new IFunction<Set<String>>() {
		public Set<String> classify(Context state) {
			Set<String> result = new LinkedHashSet<String>();
			double maxValue = -Double.MAX_VALUE;
			for (String action : model.getActionTestedSet()) {
				Double value = qualityFunctions.get(action).get(state);
				if (value != null) {
					if (value > maxValue + EPSILON) {
						result.clear();
						result.add(action);
						maxValue = value;
					} else if (Math.abs(value - maxValue) < EPSILON) {
						result.add(action);
					}
				}
			}
			return result;
		}

		public int numberOfNodes() {
			return getActionValueSize();
		}

		@Override
		public IFunction<Set<String>> clone() {
			assert (false);
			return null;
		}
	};

	/**
	 * 
	 * Wrapper to present quality functions
	 * 
	 */
	final Map<String, IFunction<Double>> qualityFunctionsWrapper = new LinkedHashMap<String, IFunction<Double>>();

	/**
	 * 
	 * Constructor
	 * 
	 * @param model
	 *          model of the MDP
	 * 
	 */
	public DynaQ(TabularLearningModel model) {
		this.model = model;
		for (String action : model.getActionList()) {
			QualityMap actionQuality = new QualityMap();
			qualityFunctions.put(action, actionQuality);
			qualityFunctionsWrapper.put(action, actionQuality);
		}
	}

	/**
	 * 
	 * Constructor
	 * 
	 * 
	 */
	public DynaQ() {
		this.model = null;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning#getIncrementalPlanning(Context,
	 *      java.lang.String, Context, List)
	 */
	public Map<String, IFunction<Double>> getIncrementalPlanning(
			Context lastState, String action, Context currentState,
			List<Double> reward) {
		if (currentState != null) {
			statesSet.add(currentState);
		}
		if (lastState != null) {
			IProbabilities<Context> probabilities = model.getTransition(lastState,
					action);
			Double probability = 0.0;
			if (probabilities != null) {
				probability = probabilities.getProbability(currentState);
			} else {
				probability = 1.0;
			}
			updateQuality(lastState, action, currentState, Context.getSum(reward),
					probability);
		}
		computePlanning();
		return qualityFunctionsWrapper;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning#getActionValueFunctions()
	 */
	public Map<String, IFunction<Double>> getActionValueFunctions() {
		return qualityFunctionsWrapper;
	}

	/**
	 * 
	 * @return the size of the current action value used in DynaQ
	 * 
	 */
	int getActionValueSize() {
		if (model.getActionTestedSet().isEmpty()) {
			return 0;
		}
		int size = 0;
		for (String action : model.getActionTestedSet()) {
			size += qualityFunctions.get(action).size();
		}
		return size;
	}

	/**
	 * 
	 * Compute the planning phase of Dyna-Q
	 * 
	 */
	private void computePlanning() {
		for (String action : model.getActionTestedSet()) {
			for (Context state : model.getStateSet(action)) {
				updateQuality(state, action);
			}
		}
	}

	/**
	 * 
	 * Update the quality function using the model
	 * 
	 * @param state
	 *          state at time t
	 * @param action
	 *          action executed at time t
	 * 
	 */
	private void updateQuality(Context state, String action) {
		IProbabilities<Context> nextStateProbabilities = model.getTransition(state,
				action);
		double reward = model.getReward(state, action);
		double probabilitySum = 0.0;
		if (nextStateProbabilities != null) {
			for (Context statePrime : nextStateProbabilities.getValueSet()) {
				double probability = nextStateProbabilities.getProbability(statePrime);
				updateQuality(state, action, statePrime, reward, probability);
				probabilitySum += probability;
			}
			assert (probabilitySum < 1.0 + EPSILON);
			assert (probabilitySum > 1.0 - EPSILON);
		} else {
			updateQuality(state, action, null, reward, 1.0);
		}
	}

	/**
	 * 
	 * Update a quality in the quality functino
	 * 
	 * @param state
	 *          the state at time t
	 * @param action
	 *          the action at time t
	 * @param statePrime
	 *          the state at time t+1
	 * @param reward
	 *          the reward received by executing state/action
	 * @param probability
	 *          the probability of statePrime
	 * 
	 */
	private void updateQuality(Context state, String action, Context statePrime,
			double reward, double probability) {
		QualityMap qualityFunction = qualityFunctions.get(action);
		Double initialQuality = qualityFunction.classify(state);
		assert (initialQuality != null);
		Double statePrimeValue = 0.0;
		if (statePrime != null) {
			statePrimeValue = optimalValueFunction.classify(statePrime);
		} else {
			statePrimeValue = initialQuality;
		}
		if (statePrimeValue == null) {
			statePrimeValue = 0.0;
		}
		double updatedQuality = initialQuality + ALPHA * probability
				* (reward + gamma * statePrimeValue - initialQuality);
		qualityFunction.put(state, updatedQuality);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getOptimalPolicy()
	 */
	public IFunction<Set<String>> getOptimalPolicy() {
		return greedyPolicy;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getOptimalValueFunction()
	 */
	public IFunction<Double> getOptimalValueFunction() {
		return optimalValueFunction;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getName()
	 */
	public String getName() {
		return getClass().getSimpleName();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getModel()
	 */
	public IModel getModel() {
		return model;
	}

	/**
	 * @see animatlab.kodamat.core.io.ILog#setLog(animatlab.kodamat.core.io.Log)
	 */
	public void setLog(Log log) {
		// We do not use log
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#solve()
	 */
	public void solve() {
		assert (false);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning#newInstance(animatlab.kodamat.reinforcementlearning.model.ILearningModel)
	 */
	public IIncrementalPlanning newInstance(ILearningModel learningModel) {
		return new DynaQ(new TabularLearningModel(learningModel
				.getVariableOrdering(), learningModel.getActionList()));
	}
}
