package animatlab.kodamat.reinforcementlearning.planning.guestrin;

import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.functions.FunctionsSum;
import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.functions.IPartition;
import animatlab.kodamat.core.functions.IPartitionFunction;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.ConvertPolicy;
import animatlab.kodamat.core.trees.DNode;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.MergePolicy;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.trees.Vertex;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractForest;

/**
 * 
 * Some utils functions to manipulate basis function
 * 
 * @author Thomas Degris
 * 
 */
public class BasisFunctionUtils {
	/**
	 * 
	 * Policy to use to iterate through the basis function
	 * 
	 */
	public static interface IterationPolicy {
		/**
		 * 
		 * Iterate over a set of basis function
		 * 
		 * @param forest
		 *          forest for which to generate basis function
		 * @param basisFunctions
		 *          the set of basis function to iterate with
		 * @return the new set of basis function
		 * 
		 */
		public FunctionsSum iterate(AbstractForest forest,
				FunctionsSum basisFunctions);
	}

	/**
	 * 
	 * First attempt to generate basis function
	 * 
	 */
	static final public IterationPolicy iterationPolicy01 = new IterationPolicy() {
		public FunctionsSum iterate(AbstractForest forest,
				FunctionsSum basisFunctions) {
			Set<Tree<Double>> basisFunctionSet = new LinkedHashSet<Tree<Double>>();
			for (IFunction<Double> basisFunction : basisFunctions.functions()) {
				Tree<Double> newBasisFunction = new Tree<Double>();
				for (String action : forest.getActionList()) {
					dummyMergePolicy.merge(pregress(forest, action,
							((Tree<Double>) basisFunction).getRoot()), newBasisFunction);
				}
				basisFunctionSet.add(newBasisFunction);
			}
			FunctionsSum result = new FunctionsSum();
			for (Tree<Double> basisFunction : basisFunctionSet) {
				result.add(basisFunction);
			}
			return result;
		}
	};

	/**
	 * 
	 * First attempt to generate basis function
	 * 
	 */
	static final public IterationPolicy iterationPolicy02 = new IterationPolicy() {
		public FunctionsSum iterate(AbstractForest forest,
				FunctionsSum basisFunctions) {
			Set<Tree<Double>> basisFunctionSet = new LinkedHashSet<Tree<Double>>();
			for (IFunction<Double> basisFunction : basisFunctions.functions()) {
				Tree<Double> newBasisFunction = new Tree<Double>();
				for (String action : forest.getActionList()) {
					dummyMergePolicy.merge(pregress(forest, action,
							((Tree<Double>) basisFunction).getRoot()), newBasisFunction);
				}
				FunctionsSum rewards = forest.getRewardFunction(null);
				for (IFunction<Double> reward : rewards.functions()) {
					Tree<Double> workingBasisFunction = newBasisFunction.clone();
					workingBasisFunction = dummyMergePolicy.merge((Tree<Double>) reward,
							workingBasisFunction);
					basisFunctionSet.add(workingBasisFunction);
				}
			}
			FunctionsSum result = new FunctionsSum();
			for (Tree<Double> basisFunction : basisFunctionSet) {
				result.add(basisFunction);
			}
			return result;
		}
	};

	/**
	 * 
	 * Policy used to convert CPD trees
	 * 
	 */
	static private final ConvertPolicy<IProbabilities<String>, Double> convertPolicy = new ConvertPolicy<IProbabilities<String>, Double>() {
		@Override
		public Double convert(Leaf<IProbabilities<String>> leaf) {
			return null;
		}
	};

	/**
	 * 
	 * Policy to merge PTree together
	 * 
	 */
	static final MergePolicy<Double> dummyMergePolicy = new MergePolicy<Double>(
			null, null);

	/**
	 * 
	 * Generate a default set of basis function as a set of trees
	 * 
	 * @param forest
	 *          forest for which to generate basis function
	 * @param iterationPolicy
	 *          policy to use to iterate over the basis functions
	 * @param numberOfIterations
	 *          number of iterations to execute
	 * 
	 * @return the new set of basis function
	 * 
	 */
	static public FunctionsSum generateBasisFunction(AbstractForest forest,
			IterationPolicy iterationPolicy, int numberOfIterations) {
		assert (numberOfIterations >= 0);
		FunctionsSum result = new FunctionsSum();
		result.add(newConstantBasisFunction(forest));
		initializeBasisFunctionOrigin(forest, result);
		for (int currentIteration = 0; currentIteration < numberOfIterations; currentIteration++) {
			result = iterationPolicy.iterate(forest, result);
		}
		result = decomposeBasisFunction(forest, result);
		return result;
	}

	/**
	 * 
	 * Decompose a set of of basis function
	 * 
	 * @param forest
	 *          forest for which to generate basis function
	 * @param basisFunctions
	 *          the set of basis function to decompose
	 * @return the new set of basis function
	 * 
	 */
	public static FunctionsSum decomposeBasisFunction(AbstractForest forest,
			FunctionsSum basisFunctions) {
		Set<Tree<Double>> basisFunctionSet = new LinkedHashSet<Tree<Double>>();
		for (IFunction<Double> basisFunction : basisFunctions.functions()) {
			for (IPartition<Double> partition : ((IPartitionFunction<Double>) basisFunction)
					.partitions()) {
				Tree<Double> decomposedBasisFunction = new Tree<Double>();
				Leaf<Double> currentLeaf = (Leaf<Double>) decomposedBasisFunction
						.getRoot();
				for (Map.Entry<Variable, String> entry : partition.getContext()
						.entrySet()) {
					DNode<Double> dnode = currentLeaf.replace(entry.getKey());
					for (String value : forest.getValueSet(entry.getKey())) {
						dnode.addLeaf(value, 0.0);
					}
					currentLeaf = (Leaf<Double>) dnode.getChild(entry.getValue());
				}
				currentLeaf.set(1.0);
				basisFunctionSet.add(decomposedBasisFunction);
			}
		}
		FunctionsSum result = new FunctionsSum();
		for (Tree<Double> basisFunction : basisFunctionSet) {
			result.add(basisFunction);
		}
		return result;
	}

	/**
	 * 
	 * Compute a new iteration for a given basis function
	 * 
	 * @param action
	 *          action to iterate with
	 * @param forest
	 *          forest for which to generate basis function
	 * @param vertex
	 *          the basis function to iterate with
	 * @return the new basis function
	 * 
	 */
	static Tree<Double> pregress(AbstractForest forest, String action,
			Vertex<Double> vertex) {
		Tree<Double> result = new Tree<Double>();
		if (vertex.isLeaf) {
			return result;
		}
		final DNode<Double> dnode = (DNode<Double>) vertex;
		Map<String, Tree<Double>> testedVariableIterateSet = new LinkedHashMap<String, Tree<Double>>();
		for (Vertex<Double> child : dnode.children()) {
			testedVariableIterateSet.put(child.getTestedValue(), pregress(forest,
					action, child));
		}
		final Tree<IProbabilities<String>> testedVariableCPD = forest
				.getTransitionFunction(dnode.getTestedVariable(), action);
		convertPolicy.convert(testedVariableCPD, result, null);
		for (Leaf<Double> leaf : result.getClonedLeafSet()) {
			IProbabilities<String> probabilityDistributions = testedVariableCPD
					.getLeaf(leaf.getContext()).get();
			Tree<Double> treeAtLeaf = new Tree<Double>();
			for (String value : probabilityDistributions.getValueSet()) {
				dummyMergePolicy.merge(testedVariableIterateSet.get(value).clone(
						leaf.getContext()), treeAtLeaf);
			}
			result.append(leaf, treeAtLeaf, null, null);
		}
		return result;
	}

	/**
	 * 
	 * Initialize a first base of basis function
	 * 
	 * @param forest
	 *          forest for which the basis function are generated
	 * @param basisFunctionSet
	 *          the set of basis function to fill
	 * 
	 */
	static private void initializeBasisFunctionOrigin(AbstractForest forest,
			FunctionsSum basisFunctionSet) {
		for (Variable variable : forest.getVariableSet()) {
			Tree<Double> varTree = new Tree<Double>();
			DNode<Double> varDNode = ((Leaf<Double>) varTree.getRoot())
					.replace(variable);
			for (String value : forest.getValueSet(variable)) {
				varDNode.addLeaf(value);
			}
			basisFunctionSet.add(varTree);
		}
	}

	/**
	 * 
	 * Create a new constant basis function
	 * 
	 * @param forest
	 *          forest for which to generate basis function
	 * 
	 * @return the constant basis function
	 * 
	 */
	public static Tree<Double> newConstantBasisFunction(AbstractForest forest) {
		Tree<Double> result = new Tree<Double>();
		((Leaf<Double>) result.getRoot()).set(1.0);
		return result;
	}

}
