package animatlab.kodamat.reinforcementlearning.planning.guestrin;

import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Parameters;

/**
 * 
 * Policy represented as a decision list
 * 
 * @author Thomas Degris
 * 
 */
public class DecisionListPolicy implements IFunction<Set<String>> {
	/**
	 * 
	 * Infos required for an action to build a decision list defining a policy
	 * 
	 */
	protected class ActionInfos implements Comparable<ActionInfos> {
		/**
		 * 
		 * Context required to activate this action
		 * 
		 */
		public Context context;

		/**
		 * 
		 * Action defined by this info
		 * 
		 */
		public Set<String> actions = new LinkedHashSet<String>();

		/**
		 * 
		 * Bonus for taking this action compared to the default action
		 * 
		 */
		public double bonus;

		/**
		 * 
		 * Constructor
		 * 
		 * @param context
		 *          context required to activate this action
		 * @param action
		 *          action defined by this info
		 * @param bonus
		 *          bonus for taking this action compared to the default
		 * 
		 */
		public ActionInfos(Context context, String action, double bonus) {
			if (context != null) {
				this.context = context;
			} else {
				this.context = new Context();
			}
			actions.add(action);
			this.bonus = bonus;
		}

		/**
		 * 
		 * Compare with another element
		 * 
		 * @param other
		 *          the other element to compare with
		 * @return true if this executing this action is better than this other one
		 * 
		 */
		public int compareTo(ActionInfos other) {
			if (Math.abs(bonus - other.bonus) < Parameters.EPSILON) {
				return 0;
			}
			if (bonus < other.bonus) {
				return 1;
			}
			return -1;
		}

		/**
		 * @see java.lang.Object#toString()
		 */
		@Override
		public String toString() {
			StringBuilder result = new StringBuilder("({");
			result.append(Context.contextToString(context));
			result.append("},");
			result.append(actions.toString());
			result.append(",");
			result.append(Parameters.format(bonus));
			result.append(")");
			return result.toString();
		}
	}

	/**
	 * 
	 * Sorted list of action infos
	 * 
	 */
	final private List<ActionInfos> decisionList = new LinkedList<ActionInfos>();

	/**
	 * 
	 * Constructor
	 * 
	 */
	public DecisionListPolicy() {
		super();
	}

	/**
	 * @see animatlab.kodamat.core.functions.IFunction#classify(Context)
	 */
	public Set<String> classify(Context state) {
		for (ActionInfos actionInfo : decisionList) {
			if (Context.satisfy(actionInfo.context, state)) {
				return actionInfo.actions;
			}
		}
		// We should not be here
		assert (false);
		return null;
	}

	/**
	 * @see animatlab.kodamat.core.functions.IFunction#numberOfNodes()
	 */
	public int numberOfNodes() {
		return decisionList.size();
	}

	/**
	 * @see java.lang.Object#clone()
	 */
	@Override
	public DecisionListPolicy clone() {
		assert (false);
		return null;
	}

	/**
	 * 
	 * Add a new action info to this list
	 * 
	 * @param context
	 *          context required to activate this action
	 * @param action
	 *          action defined by this info
	 * @param bonus
	 *          bonus for taking this action compared to the default
	 * 
	 */
	public void add(Context context, String action, Double bonus) {
		decisionList.add(new ActionInfos(context, action, bonus));
	}

	/**
	 * 
	 * Sort this list in decreasing order
	 * 
	 */
	public void sort() {
		Collections.sort(decisionList);
		int index = 0;
		ActionInfos reference = null;
		do {
			ActionInfos actionInfo = decisionList.get(index);
			if ((reference == null)
					|| (Math.abs(actionInfo.bonus - reference.bonus) > Parameters.EPSILON)
					|| !actionInfo.context.equals(reference.context)) {
				reference = actionInfo;
				index++;
			} else {
				reference.actions.addAll(actionInfo.actions);
				decisionList.remove(actionInfo);
			}
		} while (index < decisionList.size());
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuilder result = new StringBuilder("[size: ");
		result.append(numberOfNodes());
		result.append("\n");
		final String separator = ",\n";
		for (ActionInfos actionInfo : decisionList) {
			result.append(actionInfo.toString());
			result.append(separator);
		}
		String subString = result
				.substring(0, result.length() - separator.length());
		return subString + "]";
	}

	/**
	 * 
	 * Remove rules that can be called
	 * 
	 */
	public void removeUselessRules() {
		for (int indexReference = 0; indexReference < decisionList.size(); indexReference++) {
			ActionInfos reference = decisionList.get(indexReference);
			int indexTest = indexReference + 1;
			while (indexTest < decisionList.size()) {
				ActionInfos tested = decisionList.get(indexTest);
				if ((Context.satisfy(reference.context, tested.context))
						&& (tested.context.keySet().containsAll(reference.context.keySet()))) {
					decisionList.remove(indexTest);
				} else {
					indexTest++;
				}
			}
		}
	}
}
