package animatlab.kodamat.reinforcementlearning.planning.guestrin;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import animatlab.kodamat.core.functions.FunctionsSum;
import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.functions.IPartition;
import animatlab.kodamat.core.functions.IPartitionFunction;
import animatlab.kodamat.core.functions.UnaryOperatorPolicy;
import animatlab.kodamat.core.io.IToXml;
import animatlab.kodamat.core.io.Log;
import animatlab.kodamat.core.rules.PlainRule;
import animatlab.kodamat.core.rules.PlainRules;
import animatlab.kodamat.core.rules.Rule;
import animatlab.kodamat.core.rules.Rules.RulePair;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.reinforcementlearning.core.acting.Greedy;
import animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractForest;
import animatlab.kodamat.reinforcementlearning.planning.IPlanning;
import animatlab.kodamat.reinforcementlearning.planning.guestrin.linearprogramming.LPComment;
import animatlab.kodamat.reinforcementlearning.planning.guestrin.linearprogramming.LPNamedVariable;
import animatlab.kodamat.reinforcementlearning.planning.guestrin.linearprogramming.LPProblem;
import animatlab.kodamat.reinforcementlearning.planning.guestrin.linearprogramming.LPVariable;

/**
 * 
 * Planning method using the Rule-based factor LP algorithm from Guestrin
 * (thesis chapter 7)
 * 
 * @author Thomas Degris
 * 
 */
public class GuestrinPlanning implements IPlanning, IToXml<Object> {
	/**
	 * 
	 * Prefix for the forest in a xml file
	 * 
	 */
	protected static final String PREFIX = "lp:";

	/**
	 * 
	 * Name of the node in xml to represent an decision node
	 * 
	 */
	static public final String NODE_NAME_AGENT = PREFIX + "agent";

	/**
	 * 
	 * Name of the node in xml to represent an decision node
	 * 
	 */
	protected static final String NODE_NAME_VALUEFUNCTION = PREFIX
			+ "valuefunction";

	/**
	 * 
	 * Name of the node in xml to represent an decision node
	 * 
	 */
	protected static final String NODE_NAME_POLICY = PREFIX + "policy";

	/**
	 * 
	 * Last value function computed
	 * 
	 */
	protected FunctionsSum optimalValueFunction = null;

	/**
	 * 
	 * Set of computed quality function
	 * 
	 */
	final protected Map<String, IFunction<Double>> lastComputedQualityFunction = new LinkedHashMap<String, IFunction<Double>>();

	/**
	 * 
	 * Name of the phi variable
	 * 
	 */
	private static final String LPVARIABLE_PHI = "phi";

	/**
	 * 
	 * Forest used to compute a policy
	 * 
	 */
	protected AbstractFactoredModel model;

	/**
	 * 
	 * LP Problem used to define and solve the LP program
	 * 
	 */
	final protected LPProblem omega = createLPProblem();

	/**
	 * 
	 * Wrapper to compute an action-value function
	 * 
	 */
	private class QualityWrapper implements IFunction<Double> {
		/**
		 * 
		 * Action associated with the quality
		 * 
		 */
		private final String action;

		/**
		 * 
		 * Constructor
		 * 
		 * @param action
		 *          action associated with this quality wrapper
		 * 
		 */
		QualityWrapper(String action) {
			this.action = action;
		}

		/**
		 * @see animatlab.kodamat.core.functions.IFunction#classify(Context)
		 */
		public Double classify(Context state) {
			double result = 0;
			for (PlainRules<Double> rules : basisFunctionsBackProjections.get(action)) {
				Double value = rules.classify(state);
				if (value != null) {
					result = result + value;
				}
			}
			result = result * gamma;
			List<Double> rewards = model.getReward(state, action);
			for (Double reward : rewards) {
				result = result + reward;
			}
			return result;
		}

		/**
		 * @see animatlab.kodamat.core.functions.IFunction#numberOfNodes()
		 */
		public int numberOfNodes() {
			return getOptimalValueFunction().numberOfNodes();
		}

		/**
		 * @see animatlab.kodamat.core.functions.IFunction#clone()
		 */
		@Override
		public IFunction<Double> clone() {
			assert (false);
			return null;
		}
	}

	/**
	 * 
	 * Gamma value used in planning computation
	 * 
	 */
	public double gamma = Parameters.GAMMA;

	/**
	 * 
	 * Projections of the basis functions for each action
	 * 
	 */
	Map<String, List<PlainRules<Double>>> basisFunctionsBackProjections;

	/**
	 * 
	 * Log instance
	 * 
	 */
	private Log log = Log.sysout;

	/**
	 * 
	 * Policy to use to compute backprojections
	 * 
	 */
	private GuestrinPolicy guestrinPolicy = null;

	/**
	 * 
	 * Constructor
	 * 
	 * @param model
	 * 
	 */
	public GuestrinPlanning(AbstractFactoredModel model) {
		setModel(model);
	}

	// /**
	// *
	// * Factored linear programming-based approximation algorithm
	// * (Guestrin thesis, page 64, eq 5.1)
	// *
	// * @return the basis function weights w to use in the linear approximation
	// * of the value function
	// *
	// **/
	// private List<LPNamedVariable> factoredNoMaxLPA() {
	// // Cache the backprojections of the basis functions.
	// Map<String, List<IFunction<Double>>> g = new LinkedHashMap<String,
	// List<IFunction<Double>>>();
	// // For each basis function hi IN H
	// for (IFunction<Double> hi : model.getBasisFunctions().functions()) {
	// // For each action a
	// for (String a : model.getActionList()) {
	// // Let gai = Backproja(hi)
	// IFunction<Double> gai = rulesBackProj(a, hi);
	// List<IFunction<Double>> ga = g.get(a);
	// if (ga == null) {
	// ga = new ArrayList<IFunction<Double>>();
	// g.put(a, ga);
	// }
	// ga.add(gai);
	// }
	// }
	//		
	// // Compute factored state relevance weights.
	// List<Double> alpha = new ArrayList<Double>();
	// // For each basis function hi, compute the factored state relevance
	// // weights alpha_i as in equation (5.2)
	// computeFactoredStateRelevance(alpha);
	// // Generate linear programming-based approximation constraints
	// // Let Omega={}
	// LPProblem omega = new LPProblem();
	// ArrayList<String>[] stateSpace = States.createStateSet(model);
	// // For each action a:
	// for (String action : model.getActionList()) {
	// omega.addConstraint(new LPComment("Action: " + action));
	// List<IFunction<Double>> ga = g.get(action);
	// for (List<String> state : stateSpace) {
	// omega.addConstraint(new LPComment("State: " + state));
	// Set<LPNamedVariable> namedVariables = new LinkedHashSet<LPNamedVariable>();
	// Set<LPNamedVariable> definedVariables = new
	// LinkedHashSet<LPNamedVariable>();
	// for (int i = 0; i < ga.size(); i++) {
	// double weight = Parameters.GAMMA * ga.get(i).classify(state);
	// weight = weight -
	// model.getBasisFunctions().functions().get(i).classify(state);
	// LPWeightedVariable variable = new LPWeightedVariable(
	// Math.abs(weight), getOmegaVariable(omega, i));
	// if (weight > 0) {
	// namedVariables.add(variable);
	// } else {
	// definedVariables.add(variable);
	// }
	// }
	//				
	// double rightArg = -model.getRewardFunction(action).classify(state);
	// omega.addConstraint(
	// new LPConstraintDefinedVariablesList(definedVariables, "<=", rightArg,
	// namedVariables));
	// }
	// }
	//		
	// // We can now obtain the solution weights by solving an LP.
	// // Let w be the solution of the linear program: minimize Sum_i alpha_i w_i
	// subject to the
	// // constraints Omega
	// List<LPNamedVariable> solutions =new ArrayList<LPNamedVariable>();
	// for (int index = 0; index < model.getBasisFunctions().size(); index++) {
	// solutions.add(getOmegaVariable(omega, index));
	// }
	// omega.setObjectiveFunction(LPProblem.Direction.MINIMIZE, alpha, solutions);
	// omega.solve();
	// return solutions;
	// }

	// /**
	// *
	// * Generate a set of constraint from a set of functions
	// *
	// * @param basisFunction the set of functions
	// * @param targetFunction the target functions
	// * @param omega set of constraint to fill
	// *
	// **/
	// private void factoredNoMaxLP(
	// List<PlainRules<Double>> basisFunction,
	// List<IFunction<Double>> targetFunction,
	// LPProblem omega) {
	// ArrayList<String>[] stateSpace = States.createStateSet(model);
	// for (List<String> state : stateSpace) {
	// omega.addConstraint(new LPComment("State: " + state));
	// Set<LPNamedVariable> namedVariables = new LinkedHashSet<LPNamedVariable>();
	// Set<LPNamedVariable> definedVariables = new
	// LinkedHashSet<LPNamedVariable>();
	// for (int i = 0; i < basisFunction.size(); i++) {
	// PlainRules<Double> bi = basisFunction.get(i);
	// double weight = bi.classify(state);
	// LPWeightedVariable variable = new LPWeightedVariable(
	// Math.abs(weight), getOmegaVariable(omega, i));
	// if (weight > 0) {
	// namedVariables.add(variable);
	// } else {
	// definedVariables.add(variable);
	// }
	// }
	// double rightArg = 0;
	// for (IFunction<Double> ti : targetFunction) {
	// rightArg = rightArg + ti.classify(state);
	// }
	// omega.addConstraint(
	// new LPConstraintDefinedVariablesList(definedVariables, "<=", -rightArg,
	// namedVariables));
	// }
	// }

	/**
	 * 
	 * Get the LP variable corresponding to a weight in the value function
	 * 
	 * @param omega
	 *          the lp problem owning the variable
	 * @param index
	 *          the index of the weight in the value function approximation
	 * @return the corresponding LP variable
	 * 
	 */
	private LPNamedVariable getOmegaVariable(LPProblem omega, int index) {
		return omega.getNamedVariable("w" + String.valueOf(index));
	}

	/**
	 * 
	 * Compute factored state relevance weights.
	 * 
	 * @param alpha
	 *          the vector of state relevance to compute
	 * 
	 */
	private void computeFactoredStateRelevance(List<Double> alpha) {
		for (IFunction<Double> hi : model.getBasisFunctions().functions()) {
			double alphaValue = 0.0;
			for (IPartition<Double> rule : ((IPartitionFunction<Double>) hi)
					.partitions()) {
				alphaValue = alphaValue
						+ (1.0 / AbstractFactoredModel.getNumberOfPartition(model, rule
								.scope())) * rule.get();
			}
			alpha.add(alphaValue);
		}
	}

	/**
	 * 
	 * Factored linear programming-based approximation algorithm (Guestrin thesis,
	 * page 64, eq 5.1)
	 * 
	 * @return the basis function weights w to use in the linear approximation of
	 *         the value function
	 * 
	 */
	public List<LPNamedVariable> factoredLPA() {
		// Cache the backprojections of the basis functions.
		Map<String, List<PlainRules<Double>>> g = computeBackProjections();

		// Compute factored state relevance weights.
		List<Double> alpha = new ArrayList<Double>();
		// For each basis function hi, compute the factored state relevance
		// weights alpha_i as in equation (5.2)
		computeFactoredStateRelevance(alpha);

		// Generate linear programming-based approximation constraints
		// Let Omega={}
		omega.reset();
		omega.setLog(log);
		// For each action a:
		for (String a : model.getActionList()) {
			omega.addConstraint(new LPComment("Action: " + a));
			List<PlainRules<Double>> c = new ArrayList<PlainRules<Double>>();
			for (int i = 0; i < model.getBasisFunctions().size(); i++) {
				c.add(computecaiFunction(g, a, i));
			}
			// Let Omega = Omega U FactoredLP({gamma * ga1-h1, ..., gamma * gak-hk},
			// Ra, O)
			factoredLP(c, model.getRewardFunction(a).functions(), omega);
		}
		// So far, our constraints guarantee that
		// phi>= R(x,a)+gamma Sum_x' P(x'|x,a)Sum_i w_ih_i(x')- Sum_i wi hi(x);
		// to satisfy the linear programming-approximation solution in (5.1), we
		// must add
		// a final constraint.
		omega.getNamedVariable(LPVARIABLE_PHI).solve(0.0);

		// We can now obtain the solution weights by solving an LP.
		// Let w be the solution of the linear program: minimize Sum_i alpha_i w_i
		// subject to the
		// constraints Omega
		List<LPNamedVariable> solutions = new ArrayList<LPNamedVariable>();
		for (int index = 0; index < model.getBasisFunctions().size(); index++) {
			solutions.add(getOmegaVariable(omega, index));
		}
		omega.setObjectiveFunction(LPProblem.Direction.MINIMIZE, alpha, solutions);
		omega.solve();
		return solutions;
	}

	/**
	 * 
	 * @return a new LP problem
	 * 
	 */
	protected LPProblem createLPProblem() {
		return new LPProblem();
	}

	/**
	 * 
	 * Compute the back projections for each actions for each basis functions
	 * 
	 * @return the set of back projections computed
	 * 
	 */
	protected Map<String, List<PlainRules<Double>>> computeBackProjections() {
		return computeBackProjections(model.getBasisFunctions());
	}

	/**
	 * 
	 * Compute the back projections for each actions for each basis functions
	 * 
	 * @param valueFunctions
	 *          the set of value functions to use in the backprojection
	 * @return the set of back projections computed
	 * 
	 */
	protected Map<String, List<PlainRules<Double>>> computeBackProjections(
			FunctionsSum valueFunctions) {
		basisFunctionsBackProjections = new LinkedHashMap<String, List<PlainRules<Double>>>();
		// For each action a
		for (String a : model.getActionList()) {
			List<PlainRules<Double>> ga = basisFunctionsBackProjections.get(a);
			if (ga == null) {
				ga = new ArrayList<PlainRules<Double>>();
				basisFunctionsBackProjections.put(a, ga);
			}
			// For each basis function hi IN H
			for (IFunction<Double> hi : valueFunctions.functions()) {
				// Let gai = Backproja(hi)
				PlainRules<Double> gai = guestrinPolicy.rulesBackProj(a,
						(IPartitionFunction<Double>) hi);
				if (gai.size() == 0) {
					gai.add(new PlainRule<Double>(0.0));
				}
				ga.add(gai);
			}
		}
		return basisFunctionsBackProjections;
	}

	/**
	 * 
	 * Compute the c^a_i functions (Guestrin thesis, page 68)
	 * 
	 * @param g
	 *          the set of back projection
	 * @param a
	 *          the action of the back projection
	 * @param i
	 *          index of the basis function
	 * @return the cai function
	 * 
	 */
	private PlainRules<Double> computecaiFunction(
			Map<String, List<PlainRules<Double>>> g, String a, int i) {
		IFunction<Double> hi = model.getBasisFunctions().functions().get(i);
		PlainRules<Double> gai = g.get(a).get(i).clone();
		UnaryOperatorPolicy.discountFactor.applyPolicy(gai);
		gai.combine((IPartitionFunction<Double>) hi, PlainRules.doubleMinusPolicy,
				model.getValueSets());
		return gai;
	}

	/**
	 * 
	 * Maximize out variable b from rule function f (Guestrin thesis, page 129,
	 * figure 7.3)
	 * 
	 * @param f
	 *          the function to remove the variable from
	 * @param b
	 *          the variable to do the maximization on
	 * @param problem
	 *          set of constraints to fill
	 * @return result of the maximisation
	 * 
	 */
	public PlainRules<LPVariable> ruleMaxOut(LPProblem problem,
			PlainRules<LPVariable> f, Variable b) {
		// Let g = {}
		PlainRules<LPVariable> g = new PlainRules<LPVariable>();
		// Add completing rules to f : [B = b_i : 0], omegaIndex=1,...,k
		for (String value : model.getValueSet(b)) {
			PlainRule<LPVariable> completingRule = f.createRule(problem
					.createConstant(0.0));
			completingRule.addConstraint(b, value);
			f.add(completingRule);
		}
		// Summing consistent rules:
		// While there are two consistent rules p1=[c1:v1] and p2=[c2:v2]:
		RulePair<LPVariable> rulePair = f.getConsistentRules(null);
		while (rulePair != null) {
			Rule<LPVariable> p1 = rulePair.rule01;
			Rule<LPVariable> p2 = rulePair.rule02;
			f.remove(p1);
			f.remove(p2);
			if (p1.hasSameContext(p2)) {
				// If c1 = c2, then replace these two rules by [c1:v1+v2]
				p1.set(problem.sum(p1.get(), p2.get()));
				f.add(p1);
			} else {
				// Else replace these two rules by the set: split(p1 < c2) U split(p2 <
				// c1)
				p1.split(f, p2.getContext(), model.getValueSets());
				p2.split(f, p1.getContext(), model.getValueSets());
			}
			rulePair = f.getConsistentRules(null);
		}
		// Maximizing out variable B:
		// Repeat until f is empty:
		while (!f.isEmpty()) {
			Set<Rule<LPVariable>> rules = f.findSameContext(b);
			// If there are rules [c & B=bi : vi], Forall bi IN Dom(B):
			if (!rules.isEmpty()) {
				// Remove these rules from f
				f.removeAll(rules);
				PlainRule<LPVariable> firstRule = (PlainRule<LPVariable>) rules
						.iterator().next();
				// and add rule [c:max_i v_i] to g
				firstRule.set(maximize(problem, rules));
				firstRule.getContext().remove(b);
				g.add(firstRule);
			} else {
				// Select two rules: pi=[ci&B=bi:vi] and pj=[cj&B=bj:vj] such
				// that ci is consistent with cj, but not identical, and replace
				// them with split(pi<cj)Usplit(pj<ci)
				rulePair = f.getConsistentRules(b);
				Rule<LPVariable> pi = rulePair.rule01;
				Rule<LPVariable> pj = rulePair.rule02;
				f.remove(pi);
				f.remove(pj);
				Context ci = new Context(pi.getContext());
				ci.remove(b);
				Context cj = new Context(pj.getContext());
				cj.remove(b);
				pi.split(f, cj, model.getValueSets());
				pj.split(f, ci, model.getValueSets());
			}
		}
		// Return g
		return g;
	}

	/**
	 * 
	 * Maximize a set of rules
	 * 
	 * @param constraints
	 *          the set of constraint to fill
	 * @param rules
	 *          the rules to maximize
	 * @return the new variable representing the maximization
	 * 
	 */
	private LPVariable maximize(LPProblem constraints, Set<Rule<LPVariable>> rules) {
		assert (rules.size() > 1);
		LPVariable[] variables = new LPVariable[rules.size()];
		int index = 0;
		for (Rule<LPVariable> rule : rules) {
			variables[index] = rule.get();
			index++;
		}
		return constraints.max(variables);
	}

	/**
	 * 
	 * Generate a set of constraint from a set of functions
	 * 
	 * @param basisFunction
	 *          the set of functions
	 * @param targetFunction
	 *          the target functions
	 * @param omega
	 *          set of constraint to fill
	 * 
	 */
	private void factoredLP(List<PlainRules<Double>> basisFunction,
			List<IFunction<Double>> targetFunction, LPProblem omega) {
		// Data structure for the intermediate functions generated in variable
		// elimination
		Set<PlainRules<LPVariable>> f = new LinkedHashSet<PlainRules<LPVariable>>();

		// Generate equality constraint to abstract away basis functions
		omega.addConstraint(new LPComment(
				"Generate equality constraint to abstract away basis functions"));
		// For each ci IN C:
		for (int i = 0; i < basisFunction.size(); i++) {
			PlainRules<Double> ci = basisFunction.get(i);
			omega.addConstraint(new LPComment("ci: " + ci.toString()));
			PlainRules<LPVariable> fi = new PlainRules<LPVariable>();
			// Let Z=Scope[ci]
			// For each assignment instantiation z IN Z
			for (Rule<Double> z : ci) {
				// create a new LP variable u_z^fi and add a constraint to Omega:
				PlainRule<LPVariable> rule = fi.createRule(omega
						.createWeightedVariable(z.get(), getOmegaVariable(omega, i)));
				rule.setContext(z.getContext());
				fi.add(rule);
			}
			// Store new function fi to use in variable elimination step F=FU{fi}
			f.add(fi);
		}

		// Generate equality constraint to abstract away target functions.
		omega.addConstraint(new LPComment(
				"Generate equality constraint to abstract away target functions"));
		// For each bj IN b:
		for (int j = 0; j < targetFunction.size(); j++) {
			IPartitionFunction<Double> bj = (IPartitionFunction<Double>) targetFunction
					.get(j);
			PlainRules<LPVariable> fj = new PlainRules<LPVariable>();
			// Let z=Scope[bj]
			// For each assignment instantiation z IN Z
			for (IPartition<Double> z : bj.partitions()) {
				// create a new LP variable u_z^fj and add a constraint to Omega:
				PlainRule<LPVariable> rule = fj.createRule(omega
						.createConstant(z.get()));
				rule.setContext(z.getContext());
				fj.add(rule);
			}
			// Store new function fj to use in variable elimination step F:FU{fj}
			f.add(fj);
		}

		// Now F contains all of the functions involved in the LP, our constraints
		// become:
		// Phi >= Sum_{ei IN F}(ei(x)), for all x, which we represent compactly
		// using a
		// variable elimination procedure.
		// For i=1 to Number of variables
		for (Variable i : model.getVariableSet()) {
			// Select the next variable to be eliminated
			// Let l = O(i)
			Variable l = i;
			omega.addConstraint(new LPComment("Eliminating " + i.toString()));
			// Select the relevant functions
			// Let e1, ..., eL be the function in F whose scope contains Xl, and let
			// Zj = Scope[ej]
			Set<PlainRules<LPVariable>> selected = select(f, l);
			// Define a new function e with scope Z = U^{L}_{j=1}Zj-{Xl} to represent
			// max_{xl}Sum_{j=1}^{L}ej.
			// Add constraints to Omega to enforce the maximum: for each assignment z
			// IN Z:
			// u^{e}_{z}>= Sum^{L}_{j=1}u^{ej}_{z,xl}[z_j] for all xl
			PlainRules<LPVariable> ue = sum(omega, selected);
			PlainRules<LPVariable> e = ruleMaxOut(omega, ue, l);
			// Update the set of functions
			// Update the set of functions F=F U {e} \ {e1, ..., eL}.
			f.removeAll(selected);
			f.add(e);
		}
		// Now, all the variables have been eliminated and all functions have empty
		// scope
		// Add last constraint to Omega: phi >= Sum_{ei IN F}ei
		LPVariable[] e = new LPVariable[f.size()];
		int i = 0;
		for (PlainRules<LPVariable> rules : f) {
			assert (rules.size() == 1);
			LPVariable ei = rules.iterator().next().get();
			e[i] = ei;
			i++;
		}
		omega.max(omega.getNamedVariable(LPVARIABLE_PHI), e);
	}

	/**
	 * 
	 * Sum a set of rules together
	 * 
	 * @param problem
	 *          the problem where to add the constraint
	 * @param rulesSet
	 *          the set of rules to sum
	 * @return the result of the sum
	 * 
	 */
	private PlainRules<LPVariable> sum(final LPProblem problem,
			Set<PlainRules<LPVariable>> rulesSet) {
		PlainRules<LPVariable> result = new PlainRules<LPVariable>();
		for (PlainRules<LPVariable> rules : rulesSet) {
			result.combine(rules, new PlainRules.CombinaisonPolicy<LPVariable>() {
				public LPVariable operator(LPVariable value01, LPVariable value02) {
					return problem.sum(value01, value02);
				}
			}, model.getValueSets());
		}
		return result;
	}

	/**
	 * 
	 * @return the last computed approximated value function
	 * 
	 */
	public IFunction<Double> getApproximatedValueFunction() {
		return optimalValueFunction;
	}

	/**
	 * 
	 * Select the functions having a scope including a selected variable
	 * 
	 * @param rulesSet
	 *          the set of functions to look in
	 * @param variable
	 *          the variable to look for in scope
	 * @return the set of selected function
	 * 
	 */
	private Set<PlainRules<LPVariable>> select(
			Set<PlainRules<LPVariable>> rulesSet, Variable variable) {
		Set<PlainRules<LPVariable>> result = new LinkedHashSet<PlainRules<LPVariable>>();
		for (PlainRules<LPVariable> rules : rulesSet) {
			if (rules.scope().contains(variable)) {
				result.add(rules);
			}
		}
		return result;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#solve()
	 */
	public void solve() {
		List<LPNamedVariable> omegaVariable = factoredLPA();
		FunctionsSum newValueFunction = new FunctionsSum();
		for (int index = 0; index < model.getBasisFunctions().size(); index++) {
			double weigth = omegaVariable.get(index).getSolution();
			if (!Double.isNaN(weigth) && (weigth != 0)) {
				IPartitionFunction<Double> basisFunction = (IPartitionFunction<Double>) model
						.getBasisFunctions().functions().get(index).clone();
				for (IPartition<Double> partition : basisFunction.partitions()) {
					partition.set(partition.get() * weigth);
				}
				newValueFunction.add(basisFunction);
			}
		}
		setLastComputedValueFunction(newValueFunction);
		computeBackProjections(newValueFunction);
	}

	/**
	 * 
	 * Set a new value value function for this instance
	 * 
	 * @param newValueFunction
	 *          the new value function to set
	 * 
	 */
	protected void setLastComputedValueFunction(FunctionsSum newValueFunction) {
		optimalValueFunction = newValueFunction;
	}

	/**
	 * 
	 * @return the optimal policy as a decision list
	 * 
	 */
	public DecisionListPolicy getOptimalDecisionListPolicy() {
		Map<String, FunctionsSum> qualityFunctions = new LinkedHashMap<String, FunctionsSum>();
		for (String action : model.getActionList()) {
			FunctionsSum quality = computeQuality(action, getOptimalValueFunction());
			qualityFunctions.put(action, quality);
		}
		return decisionListPolicy(qualityFunctions);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getOptimalPolicy()
	 */
	public IFunction<Set<String>> getOptimalPolicy() {
		Map<String, IFunction<Double>> qualityFunctions = new LinkedHashMap<String, IFunction<Double>>();
		for (String action : model.getActionList()) {
			FunctionsSum quality = computeQuality(action, getOptimalValueFunction());
			qualityFunctions.put(action, quality);
		}
		return new Greedy(qualityFunctions);
	}

	/**
	 * 
	 * Compute the quality function, given an action and a value function
	 * 
	 * @param action
	 *          the action for which we want to get the quality function
	 * @param valueFunction
	 *          the value function to compute from
	 * @return the quality function given the action and the value function
	 * 
	 */
	private FunctionsSum computeQuality(String action, FunctionsSum valueFunction) {
		FunctionsSum result = new FunctionsSum();

		// Build the sum of rewards
		result.add(getModel().getRewardFunction(action));

		// Build the sum of weighted backprojection
		FunctionsSum gaiSum = new FunctionsSum();
		List<PlainRules<Double>> ga = basisFunctionsBackProjections.get(action);
		for (int gaIndex = 0; gaIndex < ga.size(); gaIndex++) {
			gaiSum.add(ga.get(gaIndex), valueFunction.getWeight(gaIndex));
		}
		result.add(gaiSum);

		return result;
	}

	/**
	 * 
	 * Compute a factored policy given a set of quality functions
	 * 
	 * @param q
	 *          the set of quality functions to compute the policy from
	 * @return a policy
	 * 
	 */
	@SuppressWarnings("null")
	private DecisionListPolicy decisionListPolicy(Map<String, FunctionsSum> q) {
		// Let Delta = {}
		DecisionListPolicy delta = new DecisionListPolicy();
		// Compute the bonus functions
		String d = getModel().getActionList().iterator().next();
		FunctionsSum qd = q.get(d);
		assert (qd.size() == 2);
		FunctionsSum gd = (FunctionsSum) qd.functions().get(1);
		// For each action a other than the default action d:
		for (String action : getModel().getActionList()) {
			// Compute the bonus for taking action a,
			// delta_a(x) = Q_a(x) - Q_d(x);
			// as in equation (5.6). Note that delta_a has scope restricted
			// to T_a, as in Equation (5.7)
			// WARNING: we suppose R_a(x)=0
			if (action != d) {
				FunctionsSum qa = q.get(action);
				assert (qa.size() == 2);
				FunctionsSum ga = (FunctionsSum) qa.functions().get(1);
				assert (gd.size() == ga.size());
				List<PlainRules<Double>> deltaAISet = new ArrayList<PlainRules<Double>>();
				for (int gaIndex = 0; gaIndex < ga.size(); gaIndex++) {
					PlainRules<Double> gaiMinusGdi = (PlainRules<Double>) ga.functions()
							.get(gaIndex).clone();
					gaiMinusGdi.combine((IPartitionFunction<Double>) gd.functions().get(
							gaIndex), new PlainRules.CombinaisonPolicy<Double>() {
						public Double operator(Double value01, Double value02) {
							Double result = value01 - value02;
							if (Math.abs(result) < Parameters.EPSILON) {
								return null;
							}
							return result;
						}
					}, model.getValueSets());
					new UnaryOperatorPolicy.MultiplyOperator(gd.getWeight(gaIndex))
							.applyPolicy(gaiMinusGdi);
					deltaAISet.add(gaiMinusGdi);
				}
				PlainRules<Double> deltaA = null;
				for (PlainRules<Double> deltaAI : deltaAISet) {
					if (deltaA == null) {
						deltaA = deltaAI;
					} else {
						deltaA.combine(deltaAI, PlainRules.doubleSumPolicy, model
								.getValueSets());
					}
				}
				// Add states with positives bonus to the (unsorted) decision list
				// For each assignment T \in T_a :
				for (IPartition<Double> rule : deltaA.partitions()) {
					// If delta_a(t) > 0, add branch to decision list:
					// Delta = Delta U { (t,a,\delta_a(t))}
					if (rule.get() > 0) {
						delta.add(rule.getContext(), action, rule.get());
					}
				}
			}
		}
		// Add the default action to the (unsorted) decision list
		// Let Delta = Delta U { (EmptySet, d, 0) }.
		delta.add(null, d, 0.0);
		// Sort decision list to obtain final policy
		// Sort the decision list Delta in decreasing order on
		// the delta element of (t,a,delta)
		delta.sort();
		return delta;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getOptimalValueFunction()
	 */
	public FunctionsSum getOptimalValueFunction() {
		return optimalValueFunction;
	}

	/**
	 * 
	 * @return a quality function wrapper on the last computed value function
	 * 
	 */
	public Map<String, IFunction<Double>> getQualityFunctionWrapper() {
		if (basisFunctionsBackProjections == null) {
			return null;
		}
		return lastComputedQualityFunction;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getModel()
	 */
	public AbstractFactoredModel getModel() {
		return model;
	}

	/**
	 * @see animatlab.kodamat.core.io.ILog#setLog(animatlab.kodamat.core.io.Log)
	 */
	public void setLog(Log log) {
		this.log = log;
		omega.setLog(log);
	}

	/**
	 * 
	 * Set the model to use to compute the policy
	 * 
	 * @param model
	 *          the new model to use to compute the policy
	 * 
	 */
	public void setModel(AbstractFactoredModel model) {
		this.model = model;
		guestrinPolicy = null;
		if (model == null) {
			return;
		}
		if (model.getParameters().hasKey(Parameters.PARAM_KEY_GAMMA)) {
			gamma = model.getParameters().getKeyD(Parameters.PARAM_KEY_GAMMA);
		}
		for (String action : model.getActionList()) {
			lastComputedQualityFunction.put(action, new QualityWrapper(action));
		}
		FunctionsSum basisFunctions = model.getBasisFunctions();
		if ((basisFunctions != null) && (basisFunctions.size() > 0)) {
			IFunction<Double> function = basisFunctions.functions().get(0);
			if (function instanceof Tree) {
				setPolicy(new TreeBasedBackproj((AbstractForest) model, gamma));
			} else {
				setPolicy(new RuleBasedBackproj(model));
			}
		}
	}

	/**
	 * 
	 * Set the policy to use to compute backprojections
	 * 
	 * @param policy
	 *          the policy to use
	 * 
	 */
	public void setPolicy(GuestrinPolicy policy) {
		this.guestrinPolicy = policy;
	}

	/**
	 * @see animatlab.kodamat.core.io.IToXml#toXml(org.w3c.dom.Document,
	 *      animatlab.kodamat.core.io.IToXml.IToXmlPolicy)
	 */
	public Element toXml(Document document, IToXmlPolicy<Object> policy) {
		Element agentNode = document.createElement(NODE_NAME_AGENT);
		agentNode.appendChild(model.toXml(document, policy));

		if (optimalValueFunction != null) {
			Element valueFunctionNode = document
					.createElement(NODE_NAME_VALUEFUNCTION);
			valueFunctionNode.appendChild(optimalValueFunction.toXml(document, null));
			agentNode.appendChild(valueFunctionNode);
		}

		return agentNode;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getName()
	 */
	public String getName() {
		return guestrinPolicy.getName();
	}

	/**
	 * 
	 * Compute the back projection for a basis function (Guestrin thesis, page
	 * 127, eq 7.1)
	 * 
	 * @param a
	 *          the action to compute the back projection for
	 * @param basisFunction
	 *          the basis function to project
	 * 
	 * @return the result of the back projection
	 * 
	 */
	public PlainRules<Double> rulesBackProj(String a,
			IPartitionFunction<Double> basisFunction) {
		return guestrinPolicy.rulesBackProj(a, basisFunction);
	}

	/**
	 * 
	 * @return the policy used to compute backprojections
	 * 
	 */
	public GuestrinPolicy getPolicy() {
		return guestrinPolicy;
	}
}
