package animatlab.kodamat.reinforcementlearning.planning.guestrin;

import static animatlab.kodamat.core.utils.Parameters.PARAM_KEY_BASIS_FUNCTION_SCOPE_SIZE;
import static animatlab.kodamat.core.utils.Parameters.PARAM_KEY_DECOMPOSE_BASISFUNCTION;

import java.util.List;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.reinforcementlearning.model.ILearningModel;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractLearningForest;
import animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning;
import animatlab.kodamat.reinforcementlearning.planning.guestrin.linearprogramming.IncrementalLPProblem;
import animatlab.kodamat.reinforcementlearning.planning.guestrin.linearprogramming.LPProblem;

/**
 * 
 * Incremental planning techniques based on LP
 * 
 * @author Thomas Degris
 * 
 */
public class LearningGuestrinPlanning extends GuestrinPlanning implements
		IIncrementalPlanning {
	/**
	 * 
	 * Period for which the planning is not computed after the model has been
	 * modified (so, the agent act with the last planning result)
	 * 
	 */
	static private int MODEL_NOT_MODIFIED_BUFFER_PERIOD = 100;

	/**
	 * 
	 * Maximum period for which the planning is not computed
	 * 
	 */
	static private int NO_COMPUTATION_MAX_PERIOD = 1500;

	/**
	 * 
	 * Minimum period for which the planning is not computed
	 * 
	 */
	static private int NO_COMPUTATION_MIN_PERIOD = 50;

	/**
	 * 
	 * Last time the model has been modified
	 * 
	 */
	private int currentTime = 0;

	/**
	 * 
	 * Last time the model has been modified
	 * 
	 */
	private int lastTimeMDPModified = 0;

	/**
	 * 
	 * Last time the model has been modified
	 * 
	 */
	private int lastPlanningPeriod = -NO_COMPUTATION_MAX_PERIOD + 500;

	/**
	 * 
	 * Flag used to decide weither or not basis function must be generated
	 * 
	 */
	private boolean generateBasisFunction = true;

	/**
	 * 
	 * Constructor of the prototype
	 * 
	 */
	public LearningGuestrinPlanning() {
		this(null);
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param model
	 *          model to use to define the problem
	 * 
	 */
	public LearningGuestrinPlanning(AbstractLearningForest model) {
		super(model);
		if (model == null) {
			setPolicy(new TreeBasedBackproj(model, gamma));
		} else {
			model.getInductionMeasure().noEmptyLeaf = true;
		}
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.guestrin.GuestrinPlanning#createLPProblem()
	 */
	@Override
	protected LPProblem createLPProblem() {
		final IncrementalLPProblem lpSolver = new IncrementalLPProblem();
		lpSolver.timeout = 2;
		return lpSolver;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning#getIncrementalPlanning(Context,
	 *      java.lang.String, Context, List)
	 */
	public Map<String, IFunction<Double>> getIncrementalPlanning(
			Context lastState, String action, Context currentState,
			List<Double> reward) {
		currentTime++;
		if (getModel().getActionTestedSet().size() < getModel().getActionList()
				.size()) {
			return null;
		}
		if (getModel().getRewardFunction(null).numberOfNodes() <= getModel()
				.getRewardFunction(null).size()) {
			return null;
		}
		for (Map.Entry<Variable, Set<String>> entry : getModel().getValueSets()
				.entrySet()) {
			if (entry.getValue().size() <= 1) {
				return null;
			}
		}
		if (getModel().isRewardFunctionModifFlag() || getModel().isModelModifFlag()) {
			((IncrementalLPProblem) omega).resetSolution();
			lastTimeMDPModified = currentTime;
			getModel().resetModifiedFlag();
			generateBasisFunction = true;
		}
		if (((currentTime - lastTimeMDPModified > MODEL_NOT_MODIFIED_BUFFER_PERIOD) || (currentTime
				- lastPlanningPeriod > NO_COMPUTATION_MAX_PERIOD))
				&& (currentTime - lastPlanningPeriod > NO_COMPUTATION_MIN_PERIOD)) {
			lastPlanningPeriod = currentTime;
			if ((generateBasisFunction)
					&& (getModel().getParameters()
							.getKeyB(PARAM_KEY_DECOMPOSE_BASISFUNCTION))) {
				getModel().setBasisFunction(
						BasisFunctionUtils.generateBasisFunction(getModel(),
								BasisFunctionUtils.iterationPolicy01, getModel()
										.getParameters().getKeyI(
												PARAM_KEY_BASIS_FUNCTION_SCOPE_SIZE) - 1));
				generateBasisFunction = false;
			}
			solve();
		}
		return getQualityFunctionWrapper();
	}

	// /**
	// * @see
	// animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning#getIncrementalPlanning(Context,
	// * java.lang.String, Context, List)
	// */
	// public Map<String, IFunction<Double>> getIncrementalPlanning(Context
	// lastState, String action,
	// Context currentState, List<Double> reward) {
	// currentTime++;
	// if (getModel().getActionTestedSet().size() <
	// getModel().getActionList().size()) {
	// return null;
	// }
	// if (getModel().getRewardFunction(null).numberOfNodes() <=
	// getModel().getRewardFunction(null)
	// .size()) {
	// return null;
	// }
	// for (Map.Entry<Variable, Set<String>> entry :
	// getModel().getValueSets().entrySet()) {
	// if (entry.getValue().size() <= 1) {
	// return null;
	// }
	// }
	// if (getModel().isRewardFunctionModifFlag() ||
	// getModel().isModelModifFlag()) {
	// ((IncrementalLPProblem) omega).resetSolution();
	// lastTimeMDPModified = currentTime;
	// if (getModel().isModelModifFlag()) {
	// generateBasisFunction = true;
	// }
	// getModel().resetModifiedFlag();
	// }
	// if ((generateBasisFunction)
	// && (getModel().getParameters().getKeyB(PARAM_KEY_DECOMPOSE_BASISFUNCTION)))
	// {
	// getModel().setBasisFunction(
	// BasisFunctionUtils.generateBasisFunction(getModel(),
	// BasisFunctionUtils.iterationPolicy01, getModel().getParameters().getKeyI(
	// PARAM_KEY_BASIS_FUNCTION_SCOPE_SIZE) - 1));
	// generateBasisFunction = false;
	// }
	// solve();
	// return getQualityFunctionWrapper();
	// }

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning#getActionValueFunctions()
	 */
	public Map<String, IFunction<Double>> getActionValueFunctions() {
		return getQualityFunctionWrapper();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning#newInstance(animatlab.kodamat.reinforcementlearning.model.ILearningModel)
	 */
	public IIncrementalPlanning newInstance(ILearningModel learningModel) {
		return new LearningGuestrinPlanning((AbstractLearningForest) learningModel);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getName()
	 */
	@Override
	public String getName() {
		return "learning" + super.getName();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getModel()
	 */
	@Override
	public AbstractLearningForest getModel() {
		return (AbstractLearningForest) model;
	}
}
