package animatlab.kodamat.reinforcementlearning.planning.guestrin;

import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.functions.IPartition;
import animatlab.kodamat.core.functions.IPartitionFunction;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.rules.PlainRule;
import animatlab.kodamat.core.rules.PlainRules;
import animatlab.kodamat.core.rules.Rule;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel;

/**
 * 
 * Planning method using the Rule-based factor LP algorithm from Guestrin
 * (thesis chapter 7)
 * 
 * @author Thomas Degris
 * 
 */
public class RuleBasedBackproj implements GuestrinPolicy {
	/**
	 * 
	 * Model to solve
	 * 
	 */
	AbstractFactoredModel model;

	/**
	 * 
	 * Constructor
	 * 
	 * @param model
	 *          model to use to define the problem
	 * 
	 */
	public RuleBasedBackproj(AbstractFactoredModel model) {
		this.model = model;
	}

	/**
	 * 
	 * Rule-based backproject (Guestrin thesis, page 126, figure 7.2)
	 * 
	 * @param a
	 *          action used to perform the backprojection
	 * @param p
	 *          rule to use in the backprojection
	 * 
	 * @return a new set of rules, result of the backprojection
	 * 
	 */
	public PlainRules<Double> ruleBackProj(String a, IPartition<Double> p) {
		// Let g = {}
		PlainRules<Double> g = new PlainRules<Double>();
		// If this is a constant function, then return the same function
		if (p.scope().isEmpty()) {
			g.add(new PlainRule<Double>(p));
			return g;
		}
		// If the rule is equal to 0, we return an empty rule
		if (p.get() == 0) {
			return g;
		}
		// Select the set P of relevant probability rules:
		// P={nj IN P(X'omegaIndex|Parents(X'omegaIndex))|X'omegaIndex IN C and c is
		// consistent with cj}
		PlainRules<Double> setP = new PlainRules<Double>();
		for (Map.Entry<Variable, String> entry : p.getContext().entrySet()) {
			Variable variable = entry.getKey();
			String value = entry.getValue();
			IPartitionFunction<IProbabilities<String>> rules = (IPartitionFunction<IProbabilities<String>>) model
					.getTransitionFunction(variable, a);
			for (IPartition<IProbabilities<String>> nj : rules.partitions()) {
				double probability = nj.get().getProbability(value);
				PlainRule<Double> newRule = setP.createRule(probability);
				newRule.setContext(nj.getContext());
				newRule.addConstraint(Variable.getVariable(variable, 1), value);
				setP.add(newRule);
			}
		}

		// Remove the X' assignments from the context of all rules in P
		for (Rule<Double> rule : setP) {
			Set<Variable> scope = new LinkedHashSet<Variable>(rule.scope());
			for (Variable variable : scope) {
				if (variable.time > 0) {
					rule.getContext().remove(variable);
				}
			}
		}

		// Multiply consistent rules:
		// While there are two consistent rules n1 = [c1:p1] and n2 = [c2:p2]:
		// If c1 = c2, replace these two rules by [c1:p1p2];
		// Else replace these two rules by the set: Split(n1<c2)USplit(n2<c1)
		setP.mergeRules(PlainRules.doubleMultiplyPolicy);

		// Generate value rules:
		// For each rule ni IN P:
		while (!setP.isEmpty()) {
			PlainRule<Double> probRule = (PlainRule<Double>) setP.iterator().next();
			setP.remove(probRule);
			if (probRule.get() != 0) {
				// Update the backprojection g=gU{[ci:piv]}
				probRule.set(probRule.get() * p.get());
				g.add(probRule);
			}
		}
		return g;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.guestrin.GuestrinPolicy#rulesBackProj(java.lang.String,
	 *      animatlab.kodamat.core.functions.IPartitionFunction)
	 */
	public PlainRules<Double> rulesBackProj(String a,
			IPartitionFunction<Double> hj) {
		PlainRules<Double> gaj = null;
		for (IPartition<Double> pi : hj.partitions()) {
			PlainRules<Double> ruleBackProja = ruleBackProj(a, pi);
			if (gaj == null) {
				gaj = ruleBackProja;
			} else {
				gaj.combine(ruleBackProja, PlainRules.doubleSumPolicy, model
						.getValueSets());
				// Add consistent rules:
				// While there are two consistent rules n1 = [c1:p1] and n2 = [c2:p2]:
				// If c1 = c2, replace these two rules by [c1:p1p2];
				// Else replace these two rules by the set: Split(n1<c2)USplit(n2<c1)
				gaj.mergeRules(PlainRules.doubleSumPolicy, model.getValueSets());
			}
		}
		return gaj;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.guestrin.GuestrinPolicy#getName()
	 */
	public String getName() {
		return "guestrinrule";
	}
}
