package animatlab.kodamat.reinforcementlearning.planning.guestrin;

import static animatlab.kodamat.core.io.XmlUtils.getFirstChild;

import java.util.Set;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import animatlab.kodamat.core.functions.FunctionsSum;
import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.io.XmlUtils;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractForest;
import animatlab.kodamat.reinforcementlearning.model.forest.StaticForest;
import animatlab.kodamat.reinforcementlearning.planning.IPlanning;

/**
 * 
 * SpiSvi instance loaded from a xml file to get a pre-computed optimal value
 * function and optimal policy
 * 
 * @author Thomas Degris
 * 
 */
public class StaticGuestrin extends GuestrinPlanning {
	/**
	 * 
	 * File to load to run the test
	 * 
	 */
	static final public String AGENT_EXTENSION = ".guestrin.agent";

	/**
	 * 
	 * Name of this planning method
	 * 
	 */
	public static final String PLANNING_METHOD_NAME = "guestrin";

	/**
	 * 
	 * Constructor
	 * 
	 * @param element
	 *          xml tree to load the agent from
	 * 
	 */
	public StaticGuestrin(Element element) {
		super(null);
		assert (element.getNodeName().equals(NODE_NAME_AGENT) || element
				.getNodeName().equals(AbstractForest.NODE_NAME_FOREST));
		if (element.getNodeName().equals(AbstractForest.NODE_NAME_FOREST)) {
			setModel(new StaticForest(element));
		}

		NodeList nodeList = element.getChildNodes();
		for (int index = 0; index < nodeList.getLength(); index++) {
			Node childNode = nodeList.item(index);
			if (childNode.getNodeType() == Node.ELEMENT_NODE) {
				if (childNode.getNodeName().equals(AbstractForest.NODE_NAME_FOREST)) {
					setModel(new StaticForest((Element) childNode));
				}
				if (childNode.getNodeName().equals(NODE_NAME_VALUEFUNCTION)) {
					setLastComputedValueFunction(loadValueFunction(getFirstChild(childNode)));
					computeBackProjections((FunctionsSum) getApproximatedValueFunction());
				}
				if (childNode.getNodeName().equals(NODE_NAME_POLICY)) {
					setLastComputedPolicyFunction(loadPolicy(getFirstChild(childNode)));
				}
			}
		}
	}

	/**
	 * 
	 * Set a SpiSviTree as the last computed policy function
	 * 
	 * @param policy
	 *          the last computed policy tree
	 * 
	 */
	private void setLastComputedPolicyFunction(IFunction<Set<String>> policy) {
		// Nothing to do at the moment
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param model
	 *          the model used by the algorithm
	 * 
	 */
	public StaticGuestrin(AbstractForest model) {
		super(model);
	}

	/**
	 * 
	 * Load the optimal policy from a xml tree
	 * 
	 * @param xmlNode
	 *          the root the xml tree
	 * 
	 * @return the tree loaded
	 * 
	 */
	public IFunction<Set<String>> loadPolicy(Element xmlNode) {
		return null;
	}

	/**
	 * 
	 * Load the optimal value function from a xml tree
	 * 
	 * @param xmlNode
	 *          the root the xml tree
	 * 
	 * @return the value function loaded
	 * 
	 */
	public FunctionsSum loadValueFunction(Element xmlNode) {
		return AbstractFactoredModel.loadFunctionSum(xmlNode);
	}

	/**
	 * 
	 * Convert a forest xml file into an agent xml file
	 * 
	 * @param forest
	 *          the forest to convert
	 * @param outputFile
	 *          file where to save the forest
	 * @return the planning used to compute the policy of the agent
	 * 
	 */
	static public IPlanning convertProblemToAgent(StaticForest forest,
			String outputFile) {
		if (forest.getBasisFunctions() == null) {
			FunctionsSum basisFunctionSet = forest.getRewardFunction(null).clone();
			Tree<Double> constantBasisFunction = new Tree<Double>();
			((Leaf<Double>) constantBasisFunction.getRoot()).set(1.0);
			basisFunctionSet.add(constantBasisFunction);
			forest.setBasisFunctions(basisFunctionSet);
		}
		StaticGuestrin mdpLPSolver = new StaticGuestrin(forest);
		mdpLPSolver.solve();
		XmlUtils.save(mdpLPSolver, null, outputFile + AGENT_EXTENSION);
		return mdpLPSolver;
	}
}
