package animatlab.kodamat.reinforcementlearning.planning.guestrin;

import java.util.Map;

import animatlab.kodamat.core.functions.IPartition;
import animatlab.kodamat.core.functions.IPartitionFunction;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.rules.PlainRule;
import animatlab.kodamat.core.rules.PlainRules;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractForest;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.SpiSvi;

/**
 * 
 * Planning method using the Rule-based factor LP algorithm from Guestrin
 * (thesis chapter 7)
 * 
 * @author Thomas Degris
 * 
 */
public class TreeBasedBackproj implements GuestrinPolicy {
	/**
	 * 
	 * Spisvi used to compute the backprojection
	 * 
	 */
	private final SpiSvi spisvi;

	/**
	 * 
	 * Constructor
	 * 
	 * @param model
	 *          model to use to define the problem
	 * @param gamma
	 *          gamma value to use
	 * 
	 */
	public TreeBasedBackproj(AbstractForest model, double gamma) {
		spisvi = new SpiSvi(model);
		spisvi.CHECK_QUALITY = false;
		spisvi.gamma = gamma;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.guestrin.GuestrinPolicy#rulesBackProj(java.lang.String,
	 *      animatlab.kodamat.core.functions.IPartitionFunction)
	 */
	public PlainRules<Double> rulesBackProj(String a,
			IPartitionFunction<Double> basisFunction) {
		Tree<Double> backProjection = treeBackProj(a, (Tree<Double>) basisFunction);
		PlainRules<Double> result = new PlainRules<Double>();
		for (IPartition<Double> leaf : backProjection.partitions()) {
			if (Math.abs(leaf.get()) > Parameters.EPSILON) {
				PlainRule<Double> rule = result.createRule(leaf.get());
				rule.setContext(leaf.getContext());
				result.add(rule);
			}
		}
		return result;
	}

	/**
	 * 
	 * Compute the back projection for a value function
	 * 
	 * @param action
	 *          the action to compute the back projection for
	 * @param valueFunction
	 *          the value function to project
	 * 
	 * @return the result of the back projection
	 * 
	 */
	public Tree<Double> treeBackProj(String action, Tree<Double> valueFunction) {
		if (valueFunction.numberOfNodes() == 1) {
			return valueFunction.clone();
		}
		Tree<Map<Variable, IProbabilities<String>>> backProjection = spisvi
				.getPRegress(valueFunction.getRoot(), action, null);
		return spisvi.constructFVTree(action, backProjection, valueFunction);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.guestrin.GuestrinPolicy#getName()
	 */
	public String getName() {
		return "guestrintree";
	}
}
