package animatlab.kodamat.reinforcementlearning.planning.guestrin.linearprogramming;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StringWriter;
import java.io.Writer;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import org.w3c.dom.Element;

import animatlab.kodamat.core.io.DataLoadingPolicy;
import animatlab.kodamat.core.io.Exec;
import animatlab.kodamat.core.io.ILog;
import animatlab.kodamat.core.io.Log;

/**
 * 
 * The LP definition
 * 
 * @author Thomas Degris
 * 
 */
public class LPProblem implements ILog {
	/**
	 * 
	 * Binary to use to execute glpsol
	 * 
	 */
	static protected final String GLPSOL = "glpsol";

	/**
	 * 
	 * Define the optimization direction
	 * 
	 */
	public enum Direction {
		/**
		 * 
		 * The objective function must be minimized
		 * 
		 */
		MINIMIZE,

		/**
		 * 
		 * The objective function must be maximized
		 * 
		 */
		MAXIMIZE
	}

	/**
	 * 
	 * Index of the current LP variable
	 * 
	 */
	int lpVariableIndex = 1;

	/**
	 * 
	 * Set of variables in this LP problem
	 * 
	 */
	final private Set<LPNamedVariable> variables = new LinkedHashSet<LPNamedVariable>();

	/**
	 * 
	 * Set of constraints of the problem
	 * 
	 */
	final private Set<LPConstraintLine> constraints = new LinkedHashSet<LPConstraintLine>();

	/**
	 * 
	 * Direction optimization direction
	 * 
	 */
	private Direction direction;

	/**
	 * 
	 * Weights the weights associated to each variable
	 * 
	 */
	private List<Double> weights;

	/**
	 * 
	 * The set of variables to solve
	 * 
	 */
	private List<LPNamedVariable> weightVariables;

	/**
	 * 
	 * Timeout used for the solver
	 * 
	 */
	public int timeout = -1;

	/**
	 * 
	 * Policy to load the sum from a xml tree
	 * 
	 */
	final public DataLoadingPolicy<LPVariable> loadingPolicy = new DataLoadingPolicy<LPVariable>() {
		public LPVariable xmlToData(Element element) {
			return createNewVariable();
		}
	};

	/**
	 * 
	 * Log instance
	 * 
	 */
	protected Log log = Log.sysout;

	/**
	 * 
	 * Constructor
	 * 
	 * 
	 */
	public LPProblem() {
		reset();
	}

	/**
	 * 
	 * Reset this definition
	 * 
	 */
	public void reset() {
		lpVariableIndex = 1;
		variables.clear();
		constraints.clear();
		weights = null;
		weightVariables = null;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringWriter result = new StringWriter();
		try {
			result.append(objectiveToString());
			constraintsToCPLEXLP(result);
			variablesToCPLEXLP(result);
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
		}
		return result.toString();
	}

	/**
	 * 
	 * Convert the set of variables to a CPLEXLP string
	 * 
	 * @param buffer
	 *          the string to build
	 * @throws IOException
	 * 
	 */
	private void variablesToCPLEXLP(Writer buffer) throws IOException {
		buffer.append("Bounds\n");
		for (LPNamedVariable variable : variables) {
			buffer.append(variable.toBoundString());
			buffer.append("\n");
		}
	}

	/**
	 * 
	 * Convert the set of constraint to a CPLEXLP string
	 * 
	 * @param buffer
	 *          the string to build
	 * @throws IOException
	 * 
	 */
	private void constraintsToCPLEXLP(Writer buffer) throws IOException {
		buffer.append("Subject To\n");
		int constraintIndex = 0;
		for (LPConstraintLine constraint : constraints) {
			if (constraint.isConstraint) {
				buffer.append("constraint");
				buffer.append(String.valueOf(constraintIndex));
				constraintIndex++;
				buffer.append(": ");
			}
			buffer.append(constraint.toLPString());
			buffer.append("\n");
		}
	}

	/**
	 * 
	 * Get a variable with specific name
	 * 
	 * @param name
	 *          the name of the variable
	 * @return the LP variable
	 * 
	 */
	public LPNamedVariable getNamedVariable(String name) {
		for (LPNamedVariable variable : variables) {
			if (variable.name.equals(name)) {
				return variable;
			}
		}
		return createVariable(name);
	}

	/**
	 * 
	 * Create a new variable in this problem
	 * 
	 * @param name
	 *          the name of the new variable
	 * 
	 * @return the created LP variable
	 * 
	 */
	private LPNamedVariable createVariable(String name) {
		LPNamedVariable result = new LPNamedVariable(name);
		variables.add(result);
		return result;
	}

	/**
	 * 
	 * @return a new name for a LP variable
	 * 
	 */
	public LPNamedVariable createNewVariable() {
		LPNamedVariable result = createVariable("u" + lpVariableIndex);
		lpVariableIndex++;
		return result;
	}

	/**
	 * 
	 * @return the number of constraint in the problem
	 * 
	 */
	public int getNumberOfConstraint() {
		return constraints.size();
	}

	/**
	 * 
	 * Create a new constant in the LP problem
	 * 
	 * @param value
	 *          the value of the constant
	 * @return the LP variable representing the constant
	 * 
	 */
	public LPVariable createConstant(double value) {
		LPNamedVariable variable = createNewVariable();
		variable.solve(value);
		return variable;
	}

	/**
	 * 
	 * Create a new constant in the LP problem
	 * 
	 * @param name
	 *          name of the variable to create
	 * @param value
	 *          the value of the constant
	 * @return the LP variable representing the constant
	 * 
	 */
	public LPNamedVariable createNamedConstant(String name, double value) {
		LPNamedVariable variable = getNamedVariable(name);
		variable.solve(value);
		return variable;
	}

	/**
	 * 
	 * Set the objective function of the problem to solve
	 * 
	 * @param direction
	 *          optimization direction
	 * @param weights
	 *          the weights associated to each variable
	 * @param weightVariables
	 *          the set of variables to solve
	 * 
	 */
	public void setObjectiveFunction(Direction direction, List<Double> weights,
			List<LPNamedVariable> weightVariables) {
		this.direction = direction;
		this.weights = weights;
		this.weightVariables = weightVariables;
	}

	/**
	 * 
	 * Solve this LP problem
	 * 
	 */
	public void solve() {
		assert (weights.size() == weightVariables.size());
		File inputFile = writeProblemToFile();
		File outputFile = null;
		try {
			outputFile = File.createTempFile("spiti", ".sol");
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
			return;
		}
		runSolver(inputFile, outputFile);
		readSolution(outputFile, weightVariables);
		inputFile.delete();
		outputFile.delete();
	}

	/**
	 * 
	 * Run the glpsol solver
	 * 
	 * @param inputFile
	 *          input file in CPLEX LP format
	 * @param outputFile
	 *          file used to write the solution
	 * 
	 */
	protected void runSolver(File inputFile, File outputFile) {
		String timeoutOption = getTimeoutOption();
		String command = (GLPSOL + " --cpxlp --presol " + timeoutOption
				+ inputFile.getAbsolutePath() + " -o " + outputFile.getAbsolutePath());
		int exitCode = Exec.exec(command, log);
		assert (exitCode == 0);
	}

	/**
	 * 
	 * @return the option of the solver for the timeout parameter
	 * 
	 */
	protected String getTimeoutOption() {
		String timeoutOption = "";
		if (timeout != -1) {
			timeoutOption = "--tmlim " + timeout + " ";
		}
		return timeoutOption;
	}

	/**
	 * 
	 * Read the solution of the LP problem from the output file of glpsol
	 * 
	 * @param outputFile
	 *          the output file of glpsol
	 * @param weightVariables
	 *          the variable to find the solution for
	 * 
	 */
	private void readSolution(File outputFile,
			List<LPNamedVariable> weightVariables) {
		try {
			BufferedReader solution = new BufferedReader(new FileReader(outputFile));
			int index = 0;
			while (solution.ready() && (index < weightVariables.size())) {
				index = parse(solution.readLine(), weightVariables, index);
			}
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			System.exit(1);
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
		}
	}

	/**
	 * 
	 * Parse a line from the solution file
	 * 
	 * @param line
	 *          the line to parse
	 * @param weightVariables
	 *          the variables to find the solution
	 * @param index
	 *          the index of the current variable to look for
	 * @return the index of the variable to look for after this line
	 * 
	 */
	private int parse(String line, List<LPNamedVariable> weightVariables,
			int index) {
		LPNamedVariable variable = weightVariables.get(index);
		if (line.indexOf(variable.name) != -1) {
			String[] splited = line.split("\\s+");
			String value = splited[splited.length - 1];
			if (value.equals("eps")) {
				variable.setSolution(0.0);
			} else {
				variable.setSolution(Double.parseDouble(value));
			}
			return index + 1;
		}
		return index;
	}

	/**
	 * 
	 * Write the problem definition to a file
	 * 
	 * @return the file created to store the problem definition
	 * 
	 */
	private File writeProblemToFile() {
		File inputFile = null;
		try {
			inputFile = File.createTempFile("spiti", ".lp");
			FileWriter result = null;
			result = new FileWriter(inputFile);
			result.append("\\* spiti.lp *\\\n");
			result.append(objectiveToLPString());
			result.append("\n");
			constraintsToCPLEXLP(result);
			variablesToCPLEXLP(result);
			result.append("End\n");
			result.flush();
			result.close();
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
		}
		return inputFile;
	}

	/**
	 * 
	 * Convert the objectif function of this problem into a LP string
	 * 
	 * @return the LP string representation of the objective function
	 * 
	 */
	private String objectiveToLPString() {
		StringBuilder result = new StringBuilder();
		switch (direction) {
		case MINIMIZE:
			result.append("Minimize\n");
			break;
		case MAXIMIZE:
			result.append("Maximize\n");
			break;
		}
		result.append(" f : ");
		for (LPNamedVariable variable : variables) {
			int index = weightVariables.indexOf(variable);
			if (index != -1) {
				double weight = weights.get(index);
				if (weight >= 0) {
					result.append(" + ");
				} else {
					result.append(" - ");
				}
				result.append(String.valueOf(Math.abs(weight)));
				result.append(" ");
				result.append(variable.name);
			}
		}
		return result.toString();
	}

	/**
	 * 
	 * Convert the objectif function of this problem into a string
	 * 
	 * @return the string representation of the objective function
	 * 
	 */
	private String objectiveToString() {
		StringBuilder result = new StringBuilder();
		switch (direction) {
		case MINIMIZE:
			result.append("Minimize\n");
			break;
		case MAXIMIZE:
			result.append("Maximize\n");
			break;
		}
		result.append(" f : ");
		for (LPNamedVariable variable : variables) {
			int index = weightVariables.indexOf(variable);
			if (index != -1) {
				result.append(" + ");
				result.append(weights.get(index).toString());
				result.append(" ");
				result.append(variable.toString());
			}
		}
		result.append("\n");
		return result.toString();
	}

	/**
	 * 
	 * Sort a list of variables into constant and named variables
	 * 
	 * @param args
	 *          the list of variables to sort
	 * @param namedVariables
	 *          the named variables extracted
	 * 
	 * @return the sum of the constant
	 * 
	 */
	private double sortArgs(LPVariable[] args, Set<LPNamedVariable> namedVariables) {
		double result = 0.0;
		for (LPVariable variable : args) {
			if (variable.isSolved()) {
				result = result - variable.getSolution();
			} else {
				namedVariables.add((LPNamedVariable) variable);
			}
		}
		return result;
	}

	/**
	 * 
	 * Sum a set of variables
	 * 
	 * @param args
	 *          the set of variables to sum
	 * 
	 * @return the variable representing the sum
	 * 
	 */
	public LPVariable sum(LPVariable... args) {
		Set<LPNamedVariable> variables = new LinkedHashSet<LPNamedVariable>();
		double rightArg = sortArgs(args, variables);
		if (variables.size() == 0) {
			return createConstant(rightArg);
		}
		if (variables.size() == 1) {
			return variables.iterator().next();
		}
		LPNamedVariable definedVariable = createNewVariable();
		constraints
				.add(new LPConstraint(definedVariable, "=", rightArg, variables));
		return definedVariable;
	}

	/**
	 * 
	 * Get a variable that is the maximum of a list of variables
	 * 
	 * @param args
	 *          the list of variables to maximize
	 * @return the variable representing the maximization
	 * 
	 */
	public LPVariable max(LPVariable[] args) {
		LPNamedVariable definedVariable = createNewVariable();
		LPVariable[] buffer = new LPVariable[1];
		for (LPVariable arg : args) {
			buffer[0] = arg;
			max(definedVariable, buffer);
		}
		return definedVariable;
	}

	/**
	 * 
	 * Define a named variable as the maximum of a list of variables
	 * 
	 * @param namedVariable
	 *          the variable to defined as the maximum
	 * @param args
	 *          the list of variables to maximize
	 * 
	 */
	public void max(LPNamedVariable namedVariable, LPVariable[] args) {
		Set<LPNamedVariable> variables = new LinkedHashSet<LPNamedVariable>();
		double rightArg = sortArgs(args, variables);
		assert (variables.size() > 0);
		constraints.add(new LPConstraint(namedVariable, "<=", rightArg, variables));
	}

	/**
	 * 
	 * Create a variable representing a variable weighted
	 * 
	 * @param value
	 *          the weight
	 * @param variable
	 *          the variable weighted
	 * 
	 * @return the new variable
	 * 
	 */
	public LPVariable createWeightedVariable(Double value, LPVariable variable) {
		if (variable.isSolved()) {
			return createConstant(value * variable.getSolution());
		}
		LPWeightedVariable weightedVariable = new LPWeightedVariable(value,
				(LPNamedVariable) variable);
		LPNamedVariable definedVariable = createNewVariable();
		Set<LPNamedVariable> variableSet = new LinkedHashSet<LPNamedVariable>();
		variableSet.add(weightedVariable);
		constraints.add(new LPConstraint(definedVariable, "=", 0, variableSet));
		return definedVariable;
	}

	/**
	 * 
	 * Add a constraint to the problem
	 * 
	 * @param constraint
	 *          the constraint to add
	 * 
	 */
	public void addConstraint(LPConstraintLine constraint) {
		constraints.add(constraint);
	}

	/**
	 * @see animatlab.kodamat.core.io.ILog#setLog(animatlab.kodamat.core.io.Log)
	 */
	public void setLog(Log log) {
		this.log = log;
	}
}
