package animatlab.kodamat.reinforcementlearning.planning.spisvi;

import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.IModel;
import animatlab.kodamat.core.functions.IPartition;
import animatlab.kodamat.core.trees.AppendPolicy;
import animatlab.kodamat.core.trees.ClonePolicy;
import animatlab.kodamat.core.trees.ConvertPolicy;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.MergePolicy;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Parameters;

/**
 * 
 * Class used to build a policy by merging action value function
 * 
 * @author Thomas Degris
 * 
 */
public class ActionValueMergePolicy {
	/**
	 * 
	 * The result of the merge
	 * 
	 */
	private Tree<Set<String>> policy = null;

	/**
	 * 
	 * Model generating the tree
	 * 
	 */
	final IModel model;

	/**
	 * 
	 * Map of action value to remember the quality of each action
	 * 
	 */
	final Map<String, Tree<Double>> actionValues = new LinkedHashMap<String, Tree<Double>>();

	/**
	 * 
	 * Policy used to convert a value tree into an empty tree used in the merge
	 * 
	 */
	static final private ConvertPolicy<Double, Set<String>> convertPolicy = new ConvertPolicy<Double, Set<String>>() {
		/**
		 * 
		 * Convert a data of something into something else
		 * 
		 * @param leaf
		 *          the data to convert
		 * 
		 * @return the data converted
		 * 
		 */
		@Override
		public Set<String> convert(Leaf<Double> leaf) {
			return null;
		}
	};

	/**
	 * 
	 * Policy to use to merge policy tree together
	 * 
	 */
	final private class ActionAppendPolicy implements AppendPolicy<Set<String>> {
		/**
		 * 
		 * Action to merge with
		 * 
		 */
		private final String appAction;

		/**
		 * 
		 * Constructor
		 * 
		 * @param action
		 *          action to merge with
		 * 
		 */
		ActionAppendPolicy(String action) {
			appAction = action;
		}

		/**
		 * 
		 * Append two leaves together
		 * 
		 * @param originalLeaf
		 *          the original leaf. It MUST NOT be modified.
		 * @param appendedLeaf
		 *          the appended leaf that can be modified in the append operation
		 * 
		 */
		public void append(Leaf<Set<String>> originalLeaf,
				Leaf<Set<String>> appendedLeaf) {
			// Original data
			String origAction = originalLeaf.get().iterator().next();
			Leaf<Double> origValueLeaf = actionValues.get(origAction).getLeaf(
					originalLeaf.getContext());
			double origValue = Parameters.getVMaxFromReward(model.getRewardMax());
			if (origValueLeaf != null) {
				origValue = origValueLeaf.get();
			}
			// Appended data
			Context appContext = new Context(originalLeaf.getContext());
			appContext.putAll(appendedLeaf.getContext());
			Leaf<Double> appendedValueLeaf = actionValues.get(appAction).getLeaf(
					appContext);
			double appValue = Parameters.getVMaxFromReward(model.getRewardMax());
			if (appendedValueLeaf != null) {
				appValue = appendedValueLeaf.get();
			}
			// Merging data
			setAction(appendedLeaf, appAction);
			if (origValue > appValue + Parameters.EPSILON) {
				appendedLeaf.get().clear();
				appendedLeaf.get().addAll(originalLeaf.get());
			} else if (Math.abs(origValue - appValue) < Parameters.EPSILON) {
				appendedLeaf.get().addAll(originalLeaf.get());
			}
		}
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param model
	 *          model from which trees have been computed
	 * 
	 */
	public ActionValueMergePolicy(IModel model) {
		this.model = model;
	}

	/**
	 * 
	 * Merge a new action value tree
	 * 
	 * @param treeQaVs
	 *          the value tree to merge
	 * @param action
	 *          the action corresponding to the action value tree
	 * 
	 */
	public void merge(Tree<Double> treeQaVs, String action) {
		actionValues.put(action, treeQaVs);
		Tree<Set<String>> actionPolicy = new Tree<Set<String>>();
		convertPolicy.convert(treeQaVs, actionPolicy);
		if (policy == null) {
			for (IPartition<Set<String>> leaf : actionPolicy.partitions()) {
				setAction(leaf, action);
			}
		}
		MergePolicy<Set<String>> mergePolicy = new MergePolicy<Set<String>>(
				new ActionAppendPolicy(action), new ClonePolicy<Set<String>>() {
					public Set<String> cloneData(Set<String> data) {
						if (data == null) {
							return null;
						}
						return new LinkedHashSet<String>(data);
					}
				});
		policy = mergePolicy.merge(actionPolicy, policy);
	}

	/**
	 * 
	 * Set one action for a leaf
	 * 
	 * @param leaf
	 *          the leaf where to put the data
	 * @param action
	 *          the action to set
	 * 
	 */
	void setAction(IPartition<Set<String>> leaf, String action) {
		Set<String> actionSet = new LinkedHashSet<String>();
		actionSet.add(action);
		leaf.set(actionSet);
	}

	/**
	 * 
	 * @return the policy merged
	 * 
	 */
	public Tree<Set<String>> policy() {
		return policy;
	}

}
