package animatlab.kodamat.reinforcementlearning.planning.spisvi;

import animatlab.kodamat.core.functions.IPartition;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.reinforcementlearning.model.ILearningModel;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractForest;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractLearningForest;
import animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning;

/**
 * 
 * Learning version of the Factored IE algorithm
 * 
 * @author Thomas Degris
 * 
 */
public class LearningFactoredIE extends LearningSpiSvi {
	/**
	 * 
	 * Beta value to use to compute the exploration bonus
	 * 
	 */
	final private int beta;

	/**
	 * 
	 * Constructor
	 * 
	 * @param beta
	 *          beta value to use in the exploration bonus
	 * 
	 */
	public LearningFactoredIE(int beta) {
		this(beta, null);
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param beta
	 *          beta value to use in the exploration bonus
	 * @param forest
	 *          forest to use as model of the world
	 * 
	 */
	public LearningFactoredIE(int beta, AbstractForest forest) {
		super(forest);
		this.beta = beta;
		resetRewardFlag = (beta > 0);
	}

	// /**
	// * @see
	// animatlab.kodamat.reinforcementlearning.planning.spisvi.SpiSvi#getRegress(animatlab.kodamat.core.trees.Tree,
	// java.lang.String)
	// **/
	// @Override
	// public Tree<Double> getRegress(Tree<Double> treeVpi, String action) {
	// if (!getModel().getActionTestedSet().contains(action)) {
	// Tree<Double> result = new Tree<Double>(getModel().getLegend());
	// ((Leaf<Double>) result.getRoot()).set(
	// Parameters.getVMaxFromReward(Parameters.getRMaxFromReward(getModel().getRewardMax())));
	// return result;
	// }
	// log.out().print(".");
	// log.out().flush();
	// Tree<Map<Variable, IProbabilities<String>>> treePTreeQaV =
	// getPRegress(treeVpi.getRoot(), action, null);
	// if (treePTreeQaV == null) {
	// treePTreeQaV = new Tree<Map<Variable,
	// IProbabilities<String>>>(getModel().getLegend());
	// }
	// log.out().flush();
	// Tree<Double> treeFVTreeQaV = constructFVTree(action, treePTreeQaV,
	// treeVpi);
	// new UnaryOperatorPolicy.MultiplyOperator(gamma).applyPolicy(treeFVTreeQaV);
	// if ((beta > 0) && (treeFVTreeQaV.numberOfLeaves() > 1)) {
	// for (IPartition<Double> leaf : treeFVTreeQaV.partitions()) {
	// Leaf<Map<Variable, IProbabilities<String>>> probaLeaf =
	// treePTreeQaV.getLeaf(leaf.getContext());
	// if (probaLeaf.get() != null) {
	// leaf.set(leaf.get() + getParameterExplorationBonus(probaLeaf));
	// } else {
	// leaf.set(Parameters.getVMaxFromReward(getModel().getRewardMax()));
	// }
	// }
	// }
	// treeFVTreeQaV.simplify(EqualPolicy.valueLeafEqualPolicy);
	// Tree<Double> treeR = getModel().getRewardTree(action);
	// treeFVTreeQaV.append(treeR, appendRewardPolicy, null);
	// treeFVTreeQaV.simplify(EqualPolicy.valueLeafEqualPolicy);
	// return treeFVTreeQaV;
	// }
	//
	// /**
	// *
	// * Compute the exploration bonus for a leaf, given a distribution
	// *
	// * @param probaLeaf the leaf with the estimated proba
	// * @return the exploration bonus for this leaf
	// *
	// **/
	// private double getParameterExplorationBonus(
	// Leaf<Map<Variable, IProbabilities<String>>> probaLeaf) {
	// int minCi = Integer.MAX_VALUE;
	// for (IProbabilities<String> proba : probaLeaf.get().values()) {
	// minCi = Math.min(minCi, ((ILearningProbabilities<String>)
	// proba).getNumberOfExamples());
	// }
	// return beta / Math.sqrt(minCi);
	// }

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.spisvi.SpiSvi#getRegress(animatlab.kodamat.core.trees.Tree,
	 *      java.lang.String)
	 */
	@Override
	public Tree<Double> getRegress(Tree<Double> treeVpi, String action) {
		Tree<Double> treeQa = super.getRegress(treeVpi, action);
		if (beta > 0) {
			for (IPartition<Double> leaf : treeQa.partitions()) {
				leaf.set(leaf.get() + getExplorationBonus(action, leaf.getContext()));
			}
		}
		return treeQa;
	}

	/**
	 * 
	 * Compute the exploration bonus for a context
	 * 
	 * @param action
	 *          action of the action value tree
	 * @param context
	 *          context to get the number of examples with
	 * @return the exploration bonus for this context
	 * 
	 */
	private double getExplorationBonus(String action, Context context) {
		int minCi = Integer.MAX_VALUE;
		for (Variable variable : getModel().getVariableSet()) {
			minCi = Math.min(minCi, getModel().getNumberOfExamples(action, variable,
					context));
		}
		if (minCi == 0) {
			return Parameters.getVMaxFromReward(getModel().getRewardMax());
		}
		return beta / Math.sqrt(minCi);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getName()
	 */
	@Override
	public String getName() {
		String betaName = Parameters.format((double) beta);
		betaName = betaName.replaceAll("\\.", "");
		betaName = betaName.replaceAll(",", "");
		return super.getName() + betaName;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning#newInstance(animatlab.kodamat.reinforcementlearning.model.ILearningModel)
	 */
	@Override
	public IIncrementalPlanning newInstance(ILearningModel learningModel) {
		IIncrementalPlanning result = new LearningFactoredIE(beta,
				(AbstractLearningForest) learningModel);
		result.setLog(log);
		return result;
	}

}
