/*
 * 
 * Created on 8 janv. 2006
 *
 */
package animatlab.kodamat.reinforcementlearning.planning.spisvi;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.functions.FunctionsSum;
import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.io.DotUtils;
import animatlab.kodamat.core.io.HtmlPage;
import animatlab.kodamat.core.io.IToDot;
import animatlab.kodamat.core.trees.EqualPolicy;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.reinforcementlearning.model.ILearningModel;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractForest;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractLearningForest;
import animatlab.kodamat.reinforcementlearning.model.forest.RewardAdjustmentPolicy;
import animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning;
import animatlab.kodamat.supervisedlearning.core.Example;
import animatlab.kodamat.supervisedlearning.learningtree.ID3Tree;
import animatlab.kodamat.supervisedlearning.learningtree.LearningTree;
import animatlab.kodamat.supervisedlearning.regression.LeastSquareMeasure;

/**
 * 
 * Incremental SPI/SVI
 * 
 * @author Thomas Degris
 * 
 */
public class LearningSpiSvi extends SpiSvi implements IIncrementalPlanning {
	/**
	 * 
	 * The current map of quality functions
	 * 
	 */
	final Map<String, IFunction<Double>> qualityFunctions = new LinkedHashMap<String, IFunction<Double>>();

	/**
	 * 
	 * Current value function used to compute quality functions
	 * 
	 */
	Tree<Double> treeVk;

	/**
	 * 
	 * Flag to indicate weither or not to reset the value function when the reward
	 * changes
	 * 
	 */
	protected boolean resetRewardFlag = true;

	/**
	 * 
	 * Constructor
	 * 
	 * 
	 */
	public LearningSpiSvi() {
		this(null);
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param forest
	 *          forest to use as model of the world
	 * 
	 */
	public LearningSpiSvi(AbstractForest forest) {
		super(forest);
		setUnknownProbabilityPolicy(new OptimisticUnknownProbabilityPolicy());
		if (forest != null) {
			initializeCurrentValueFunction();
			if (forest.getParameters().hasKey(Parameters.PARAM_KEY_RESET_REWARD)) {
				resetRewardFlag = forest.getParameters().getKeyB(
						Parameters.PARAM_KEY_RESET_REWARD);
			}
		}
		if ((partitionsToTree != null) && (forest != null)) {
			if (forest.getParameters()
					.hasKey(Parameters.PARAM_KEY_ORDERING_THRESHOLD)) {
				partitionsToTree.getMeasure().setThreshold(
						forest.getParameters().getKeyD(
								Parameters.PARAM_KEY_ORDERING_THRESHOLD));
			} else {
				partitionsToTree.getMeasure().setThreshold(0.001);
			}
		}
	}

	/**
	 * 
	 * Initialize the current value function updated at each step
	 * 
	 */
	private void initializeCurrentValueFunction() {
		treeVk = new Tree<Double>();
		((Leaf<Double>) treeVk.getRoot()).set(0.0);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning#getIncrementalPlanning(Context,
	 *      java.lang.String, Context, List)
	 */
	public Map<String, IFunction<Double>> getIncrementalPlanning(
			Context lastState, String action, Context currentState,
			List<Double> reward) {
		return getIncrementalPlanning();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getModel()
	 */
	@Override
	public AbstractLearningForest getModel() {
		return (AbstractLearningForest) super.getModel();
	}

	/**
	 * 
	 * Set the quality function of an action
	 * 
	 * @param action
	 *          the action to set the quality
	 * @param treeQaVk
	 *          the corresponding quality function
	 * 
	 */
	private void setQualityFunction(String action, Tree<Double> treeQaVk) {
		qualityFunctions.put(action, treeQaVk);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getOptimalPolicy()
	 */
	@SuppressWarnings("unchecked")
	@Override
	public Tree<Set<String>> getOptimalPolicy() {
		RewardAdjustmentPolicy rewardAdjustmentPolicy = getModel()
				.getRewardAdjustmentPolicy();
		getModel().setRewardAdjustmentPolicy(null);
		List<LearningTree<Double, Double>> rewards = new ArrayList<LearningTree<Double, Double>>();
		FunctionsSum rewardSum = getModel().getRewardFunction(null);
		for (int funIndex = 0; funIndex < rewardSum.size(); funIndex++) {
			LearningTree<Double, Double> rewardLearningFun = (LearningTree<Double, Double>) rewardSum
					.functions().get(funIndex);
			rewards.add(rewardLearningFun);
			ID3Tree.Plain<Double> reward = new ID3Tree.Plain<Double>(
					new LeastSquareMeasure(null), getModel().getVariableOrdering());
			for (Map.Entry<Example<Double>, Integer> entry : rewardLearningFun
					.getExampleDatabase().entrySet()) {
				for (int counter = 0; counter < entry.getValue(); counter++) {
					reward.learn(entry.getKey());
				}
			}
			reward.flush();
			rewardSum.setFun(rewards.size() - 1, reward);
		}

		Tree<Set<String>> result = getSVI();

		getModel().setRewardAdjustmentPolicy(rewardAdjustmentPolicy);
		for (int funIndex = 0; funIndex < rewardSum.size(); funIndex++) {
			rewardSum.setFun(funIndex, rewards.get(funIndex));
		}

		return result;
	}

	/**
	 * 
	 * @return the current optimal policy used by the agent
	 * 
	 */
	public Tree<Set<String>> getCurrentOptimalPolicy() {
		ActionValueMergePolicy mergePolicy = new ActionValueMergePolicy(getModel());
		for (String action : getModel().getActionList()) {
			// 4. Compute Tree(QaV*) = Regress(Tree(V*),a) for each action a
			Tree<Double> treeQaVs = getRegress(getOptimalValueFunction(), action);
			// 5. Merge the trees Tree(QaVs) to obtain Tree(Pis)
			mergePolicy.merge(treeQaVs, action);
		}
		Tree<Set<String>> treePis = mergePolicy.policy();
		treePis.simplify(actionLeafEqualPolicy);
		return treePis;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getOptimalValueFunction()
	 */
	@Override
	public Tree<Double> getOptimalValueFunction() {
		// RewardAdjustmentPolicy rewardAdjustmentPolicy =
		// getModel().getRewardAdjustmentPolicy();
		// getModel().setRewardAdjustmentPolicy(null);
		// StaticSpiSvi spisvi = new StaticSpiSvi(getModel());
		// spisvi.getSVI();
		// getModel().setRewardAdjustmentPolicy(rewardAdjustmentPolicy);
		// return spisvi.getOptimalValueFunction();
		return treeVk;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.spisvi.SpiSvi#getRegress(animatlab.kodamat.core.trees.Tree,
	 *      java.lang.String)
	 */
	@Override
	public Tree<Double> getRegress(Tree<Double> treeVpi, String action) {
		if (!getModel().getActionTestedSet().contains(action)) {
			Tree<Double> result = new Tree<Double>();
			((Leaf<Double>) result.getRoot()).set(Parameters
					.getVMaxFromReward(getModel().getRewardMax()));
			return result;
		}
		return super.getRegress(treeVpi, action);
	}

	/**
	 * Compute one step of planning
	 * 
	 * @return the set of updated action value function
	 * 
	 */
	public Map<String, IFunction<Double>> getIncrementalPlanning() {
		// The reward has not been initialized, so we pass one step
		if (getModel().getRewardMax() == -Double.MAX_VALUE) {
			return null;
		}
		Tree<Double> treeVkp1 = null;

		if (resetRewardFlag && getModel().isRewardFunctionModifFlag()) {
			initializeCurrentValueFunction();
		}
		getModel().resetModifiedFlag();

		log.out().print("V[" + String.valueOf(treeVk.numberOfNodes()) + "]:");
		log.out().flush();
		for (String currentAction : getModel().getActionList()) {
			Tree<Double> treeQaVk = getRegress(treeVk, currentAction);
			setQualityFunction(currentAction, treeQaVk);
			treeVkp1 = maxPolicy.merge(treeQaVk, treeVkp1);
			treeVkp1.simplify(EqualPolicy.valueLeafEqualPolicy);
		}
		if (partitionsToTree != null) {
			treeVkp1 = partitionsToTree.build(treeVkp1);
		}
		treeVk = treeVkp1;
		setLastComputedValueFunction(treeVk);
		return qualityFunctions;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning#getActionValueFunctions()
	 */
	public Map<String, IFunction<Double>> getActionValueFunctions() {
		return qualityFunctions;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.spisvi.SpiSvi#spisviToDot(animatlab.kodamat.core.io.HtmlPage)
	 */
	@Override
	protected void spisviToDot(HtmlPage htmlPage) {
		super.spisviToDot(htmlPage);
		HtmlPage qualityPage = htmlPage.newChild("Quality (incremental)");
		for (String action : getModel().getActionList()) {
			Tree<Double> treeQaVk = getRegress(treeVk, action);
			String compactedAction = DotUtils.toNormalizeString(action);
			qualityPage.append(DotUtils.toHtmlDot(treeQaVk, IToDot.doublePolicy,
					qualityPage.getFolder(), "quality" + compactedAction));
		}
		HtmlPage rewardPage = htmlPage.newChild("Rewards (incremental)");
		for (String action : getModel().getActionList()) {
			Tree<Double> reward = getModel().getRewardTree(action);
			String compactedAction = DotUtils.toNormalizeString(action);
			rewardPage.append(DotUtils.toHtmlDot(reward, IToDot.doublePolicy,
					rewardPage.getFolder(), "reward" + compactedAction));
		}
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning#newInstance(animatlab.kodamat.reinforcementlearning.model.ILearningModel)
	 */
	public IIncrementalPlanning newInstance(ILearningModel learningModel) {
		IIncrementalPlanning result = new LearningSpiSvi(
				(AbstractLearningForest) learningModel);
		result.setLog(log);
		return result;
	}

	/**
	 * 
	 * Set the reset the value function when the reward changes flag
	 * 
	 * @param resetRewardFlag
	 *          new value of the flag
	 * 
	 */
	public void setResetRewardFlag(boolean resetRewardFlag) {
		this.resetRewardFlag = resetRewardFlag;
	}
}
