package animatlab.kodamat.reinforcementlearning.planning.spisvi;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.functions.UnaryOperatorPolicy;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.DNode;
import animatlab.kodamat.core.trees.EqualPolicy;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.trees.Vertex;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.reinforcementlearning.model.ILearningModel;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractForest;
import animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning;

/**
 * 
 * Incremental tree based planning
 * 
 * @author Thomas Degris
 * 
 */
public class LearningTreeBasedPlanning extends TreeBasedPlanning implements
		IIncrementalPlanning {
	/**
	 * 
	 * The current map of quality functions
	 * 
	 */
	final private Map<String, IFunction<Double>> qualityFunctions = new LinkedHashMap<String, IFunction<Double>>();

	/**
	 * 
	 * Current optimal value function
	 * 
	 */
	private Tree<Double> treeVk = null;

	/**
	 * 
	 * Constructor
	 * 
	 * @param forest
	 * 
	 */
	public LearningTreeBasedPlanning(AbstractForest forest) {
		super(forest);
		if (forest != null) {
			for (String action : forest.getActionList()) {
				Tree<Double> treeQ = new Tree<Double>();
				((Leaf<Double>) treeQ.getRoot()).set(0.0);
				qualityFunctions.put(action, treeQ);
			}
			setUnknownProbabilityPolicy(new OptimisticUnknownProbabilityPolicy());
		}
	}

	/**
	 * 
	 * Constructor
	 * 
	 * 
	 */
	public LearningTreeBasedPlanning() {
		this(null);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning#getIncrementalPlanning(Context,
	 *      java.lang.String, Context, java.util.List)
	 */
	public Map<String, IFunction<Double>> getIncrementalPlanning(
			Context lastState, String action, Context currentState,
			List<Double> reward) {
		if (lastState == null) {
			return qualityFunctions;
		}

		if (treeVk == null) {
			treeVk = new Tree<Double>();
			((Leaf<Double>) treeVk.getRoot()).set(0.0);
		}
		log.out().print("V[" + String.valueOf(treeVk.numberOfNodes()) + "]:");
		log.out().flush();

		Tree<Double> treeQa = getActionValueFunction(action);
		Leaf<Double> stateLeaf = treeQa.getLeaf(lastState);
		Context stateContext = stateLeaf.getContext();

		Tree<Map<Variable, IProbabilities<String>>> treePTreeQaV = getPRegress(
				treeVk.getRoot(), action, stateContext);

		Tree<Double> treeFVTreeQaV;
		if (treePTreeQaV != null) {
			treeFVTreeQaV = constructFVTree(action, treePTreeQaV, treeVk);
			new UnaryOperatorPolicy.MultiplyOperator(gamma)
					.applyPolicy(treeFVTreeQaV);
		} else {
			treeFVTreeQaV = treeVk.clone();
		}

		Tree<Double> treeR = getModel().getRewardTree(action).clone(stateContext);
		treeFVTreeQaV.append(treeR, appendRewardPolicy, null);

		stateLeaf.replace(treeFVTreeQaV);

		log.out().print("M");
		// treeVk = null;
		// for (String a : getModel().getActionList()) {
		// treeVk = maxPolicy.merge(getActionValueFunction(a), treeVk);
		// }
		merge(treeVk.getRoot(), treeQa, stateContext);

		if (partitionsToTree != null) {
			treeVk = partitionsToTree.build(treeVk);
		} else {
			treeVk.simplify(EqualPolicy.valueLeafEqualPolicy);
		}

		return qualityFunctions;
	}

	/**
	 * 
	 * Update a tree only for a specific context
	 * 
	 * @param vertex
	 *          the root of the tree where leaves are replaced
	 * @param treeQa
	 *          the tree updated in the context stateContext
	 * @param stateContext
	 *          context in which the subtree is merged
	 * 
	 */
	private void merge(Vertex<Double> vertex, Tree<Double> treeQa,
			Context stateContext) {
		if (!vertex.satisfy(stateContext)) {
			return;
		}
		if (vertex.isLeaf) {
			vertex.getTree().append((Leaf<Double>) vertex, treeQa,
					maxPolicy.getAppendPolicy(), null);
			return;
		}
		List<Vertex<Double>> children = new ArrayList<Vertex<Double>>(
				((DNode<Double>) vertex).children());
		for (Vertex<Double> child : children) {
			merge(child, treeQa, stateContext);
		}
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning#newInstance(animatlab.kodamat.reinforcementlearning.model.ILearningModel)
	 */
	public IIncrementalPlanning newInstance(ILearningModel learningModel) {
		return new LearningTreeBasedPlanning((AbstractForest) learningModel);
	}

	/**
	 * 
	 * Set a action-value function
	 * 
	 * @param action
	 *          the action to set
	 * @param treeQa
	 *          the new action-value function
	 * 
	 */
	public void setQualityFunction(String action, Tree<Double> treeQa) {
		qualityFunctions.put(action, treeQa);
	}

	/**
	 * 
	 * Get an action-value function
	 * 
	 * @param action
	 *          the action of the action-value function
	 * @return the action-value function
	 * 
	 */
	public Tree<Double> getActionValueFunction(String action) {
		return (Tree<Double>) qualityFunctions.get(action);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning#getActionValueFunctions()
	 */
	public Map<String, IFunction<Double>> getActionValueFunctions() {
		return qualityFunctions;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getOptimalValueFunction()
	 */
	@Override
	public Tree<Double> getOptimalValueFunction() {
		return treeVk;
	}
}
