/*
 * 
 * Created on Dec 29, 2005
 *
 */
package animatlab.kodamat.reinforcementlearning.planning.spisvi;

import static animatlab.kodamat.core.utils.Parameters.EPSILON;

import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import animatlab.kodamat.core.functions.IPartition;
import animatlab.kodamat.core.functions.UnaryOperatorPolicy;
import animatlab.kodamat.core.io.DotUtils;
import animatlab.kodamat.core.io.HtmlPage;
import animatlab.kodamat.core.io.IToDot;
import animatlab.kodamat.core.io.IToXml;
import animatlab.kodamat.core.io.Log;
import animatlab.kodamat.core.io.XmlUtils;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.AppendPolicy;
import animatlab.kodamat.core.trees.ClonePolicy;
import animatlab.kodamat.core.trees.ComparePolicy;
import animatlab.kodamat.core.trees.ConvertPolicy;
import animatlab.kodamat.core.trees.DNode;
import animatlab.kodamat.core.trees.EqualPolicy;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.MergePolicy;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.trees.Vertex;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractForest;
import animatlab.kodamat.reinforcementlearning.planning.IPlanning;
import animatlab.kodamat.supervisedlearning.entropy.Chi2Measure;
import animatlab.kodamat.supervisedlearning.regression.LeastSquareMeasure;
import animatlab.kodamat.supervisedlearning.utils.PartitionsToTree;

/**
 * 
 * Class containing the SPI and the SVI algorithm
 * 
 * @author Thomas Degris
 * 
 */
public class SpiSvi implements IPlanning, IToXml<Object> {
	/**
	 * 
	 * Prefix for the forest in a xml file
	 * 
	 */
	static public final String PREFIX = "spi:";

	/**
	 * 
	 * Name of the node in xml to represent an decision node
	 * 
	 */
	static public final String NODE_NAME_AGENT = PREFIX + "agent";

	/**
	 * 
	 * Flag to indicate if computed action-value are checked when assert are
	 * enabled
	 * 
	 */
	public boolean CHECK_QUALITY = true;

	/**
	 * 
	 * Maximum number of iteration
	 * 
	 */
	public int MAX_ITE = 1000;

	/**
	 * 
	 * Sum of all the probabilities computed
	 * 
	 */
	double probabilitySum = 0.0;

	/**
	 * 
	 * Gamma value used in planning computation
	 * 
	 */
	public double gamma = Parameters.GAMMA;

	/**
	 * 
	 * Name of the node in xml to represent an decision node
	 * 
	 */
	protected static final String NODE_NAME_VALUEFUNCTION = PREFIX
			+ "valuefunction";

	/**
	 * 
	 * Name of the node in xml to represent an decision node
	 * 
	 */
	protected static final String NODE_NAME_POLICY = PREFIX + "policy";

	/**
	 * 
	 * Name of the node in xml to represent an decision node
	 * 
	 */
	protected static final String NODE_NAME_ACTION = PREFIX + "action";

	/**
	 * 
	 * Name of the attribute to named a variable
	 * 
	 */
	protected static final String ATTRIBUTE_POLICY = "action";

	/**
	 * 
	 * Policy to decide the termination criterion of the SSA algorithm
	 * 
	 */
	public static interface SSATerminationCriterion {
		/**
		 * 
		 * Criterion to decide weither to value tree are considered as equivalent or
		 * not
		 * 
		 * @param treeVpik
		 *          the tree at the iteration k (WARNING: this instance may be
		 *          modified)
		 * @param treeVpikp1
		 *          the tree at the iteration k+1
		 * 
		 * @return true if treeVpik and treeVpikp1 are considered equivalent
		 * 
		 */
		boolean hasConverged(Tree<Double> treeVpik, Tree<Double> treeVpikp1);
	}

	/**
	 * 
	 * Policy used to convert a tree of probability distribution into a tree of
	 * sets of probability distributions
	 * 
	 * @author Thomas Degris
	 * 
	 */
	static protected class PdToPDSetConvertPolicy
			extends
			ConvertPolicy<IProbabilities<String>, Map<Variable, IProbabilities<String>>> {
		/**
		 * 
		 * Variable concerned by the sets
		 * 
		 */
		private Variable xPrime;

		/**
		 * 
		 * Constructor
		 * 
		 * @param xPrime
		 *          variable concerned by the sets
		 * 
		 */
		public PdToPDSetConvertPolicy(Variable xPrime) {
			this.xPrime = xPrime;
		}

		/**
		 * 
		 * Convert a data of something into something else
		 * 
		 * @param leaf
		 *          the data to convert
		 * 
		 * @return the data converted
		 * 
		 */
		@Override
		public Map<Variable, IProbabilities<String>> convert(
				Leaf<IProbabilities<String>> leaf) {
			if ((leaf == null) || (leaf.get() == null)) {
				return null;
			}
			Map<Variable, IProbabilities<String>> result = new LinkedHashMap<Variable, IProbabilities<String>>();
			result.put(xPrime, leaf.get());
			return result;
		}
	}

	/**
	 * 
	 * Policy used to compute the suprenum norm
	 * 
	 */
	static class MaxDiffComparePolicy implements ComparePolicy<Double> {
		/**
		 * 
		 * Maximum difference
		 * 
		 */
		double maxDiff = 0;

		/**
		 * 
		 * Compare a node with another node
		 * 
		 * @param node01
		 *          the first node to compare
		 * @param node02
		 *          the second node to compare
		 * 
		 */
		public void compare(Leaf<Double> node01, Leaf<Double> node02) {
			maxDiff = Math.max(maxDiff, Math.abs(node01.get() - node02.get()));
		}
	}

	/**
	 * 
	 * Policy to use when compute the value for an unknown policy
	 * 
	 */
	public class UnknownProbabilityPolicy {
		/**
		 * 
		 * @return the value associated to an unknown exception
		 * 
		 */
		protected Double probaToValue() {
			throw new Error("Unknown probabilities should not exist in this world");
		}
	}

	/**
	 * 
	 * Policy to use when compute the value for an unknown policy
	 * 
	 */
	public class OptimisticUnknownProbabilityPolicy extends
			UnknownProbabilityPolicy {
		/**
		 * 
		 * @return the value associated to an unknown exception
		 * 
		 */
		@Override
		protected Double probaToValue() {
			return Parameters.getVMaxFromReward(getModel().getRewardMax());
		}
	}

	/**
	 * 
	 * Policy to use when computing the value of an unknown policy
	 * 
	 */
	UnknownProbabilityPolicy unknownProbabilityPolicy = new UnknownProbabilityPolicy();

	/**
	 * 
	 * Criterion using the Span semi-norm defined in Stochastic Dynamic
	 * Programming with Factored Representations (Boutilier, Dearden, Goldszmidt)
	 * page 34
	 * 
	 */
	public final SSATerminationCriterion spanSeminormCriterion = new SSATerminationCriterion() {
		public boolean hasConverged(Tree<Double> treeVpik, Tree<Double> treeVpikp1) {
			double spanSemiNorm = getSpanSemiNorm(treeVpik, treeVpikp1);
			return (spanSemiNorm < epsilonThresholdConvergence);
		}
	};

	/**
	 * 
	 * Criterion using the Supremum norm defined in Stochastic Dynamic Programming
	 * with Factored Representations (Boutilier, Dearden, Goldszmidt) page 34
	 * 
	 */
	public final SSATerminationCriterion supremumNormCriterion = new SSATerminationCriterion() {
		public boolean hasConverged(Tree<Double> treeVpik, Tree<Double> treeVpikp1) {
			double supremum = getSupremumNorm(treeVpik, treeVpikp1);
			return (supremum < (epsilonThresholdConvergence * (1 - gamma) / (2 * gamma)));
		}
	};

	/**
	 * 
	 * Instance of a ValueLeafEqualPolicy
	 * 
	 */
	static final protected EqualPolicy<Set<String>> actionLeafEqualPolicy = new EqualPolicy<Set<String>>() {
		public boolean equals(Leaf<Set<String>> leaf01, Leaf<Set<String>> leaf02) {
			Set<String> set01 = leaf01.get();
			Set<String> set02 = leaf02.get();
			Set<String> intersection = new LinkedHashSet<String>(set01);
			intersection.retainAll(set02);
			if (intersection.isEmpty()) {
				return false;
			}
			set01.clear();
			set01.addAll(intersection);
			set02.clear();
			set02.addAll(intersection);
			return true;
		}
	};

	// new EqualPolicy<SpiSviInfo>() {
	// public boolean equals(SpiSviInfo data01, SpiSviInfo data02) {
	// assert(data01.action != null);
	// assert(data02.action != null);
	// return data01.action.equals(data02.action);
	// }
	// };

	/**
	 * 
	 * Instance of a ValueLeafEqualPolicy
	 * 
	 */
	static final protected EqualPolicy<Set<String>> exactActionLeafEqualPolicy = new EqualPolicy<Set<String>>() {
		public boolean equals(Leaf<Set<String>> leaf01, Leaf<Set<String>> leaf02) {
			return leaf01.get().equals(leaf02.get());
		}
	};

	/**
	 * 
	 * Policy used to clone the data into a SpiSviTree
	 * 
	 */
	static public final ClonePolicy<Map<Variable, IProbabilities<String>>> probaClonePolicy = new ClonePolicy<Map<Variable, IProbabilities<String>>>() {
		public Map<Variable, IProbabilities<String>> cloneData(
				Map<Variable, IProbabilities<String>> data) {
			if (data == null) {
				return null;
			}
			return new LinkedHashMap<Variable, IProbabilities<String>>(data);
		}
	};

	/**
	 * 
	 * Policy used to add two trees of double
	 * 
	 */
	public AppendPolicy<Double> appendRewardPolicy = new AppendPolicy<Double>() {
		public void append(Leaf<Double> originalLeaf, Leaf<Double> appendedLeaf) {
			if (appendedLeaf.get() == null) {
				appendedLeaf.set(Parameters
						.getVMaxFromReward(getModel().getRewardMax()));
			} else {
				appendedLeaf.set(appendedLeaf.get() + originalLeaf.get());
			}
		}
	};

	/**
	 * 
	 * Policy used to maximize two trees of double
	 * 
	 */
	final public MergePolicy<Double> maxPolicy = new MergePolicy<Double>(
			new AppendPolicy<Double>() {
				public void append(Leaf<Double> originalLeaf, Leaf<Double> appendedLeaf) {
					if (appendedLeaf.get() == null) {
						appendedLeaf.set(Parameters.getVMaxFromReward(getModel()
								.getRewardMax()));
					} else {
						appendedLeaf.set(Math.max(appendedLeaf.get(), originalLeaf.get()));
					}
				}
			}, null);

	/**
	 * 
	 * Forest used to compute a policy
	 * 
	 */
	private AbstractForest forest;

	/**
	 * 
	 * Last policy computed
	 * 
	 */
	private Tree<Set<String>> lastComputedPolicy = null;

	/**
	 * 
	 * Last value function computed
	 * 
	 */
	private Tree<Double> lastComputedValueFunction = null;

	/**
	 * 
	 * Last result of the SSA method
	 * 
	 */
	Tree<Double> lastSSAResult = null;

	/**
	 * 
	 * Where to log SPI/Svi
	 * 
	 */
	protected Log log = Log.sysout;

	/**
	 * 
	 * Reordering instance if reordering is enabled
	 * 
	 */
	protected PartitionsToTree<Double> partitionsToTree = null;

	/**
	 * 
	 * Threshold to use to compare two value functions and test if they have
	 * converged
	 * 
	 */
	double epsilonThresholdConvergence = Parameters.EPSILON_THRESHOLD_CONVERGENCE;

	/**
	 * 
	 * Policy used to convert a set of actions into a string
	 * 
	 */
	public ConvertPolicy<Set<String>, String> policyConversionPolicy = new ConvertPolicy<Set<String>, String>() {
		@Override
		public String convert(Leaf<Set<String>> leaf) {
			String actions = leaf.get().toString();
			actions = actions.replaceAll(",", "\\\\n");
			return actions.substring(1, actions.length() - 1);
		}
	};

	/**
	 * 
	 * Policy used to join two conditional probabilities distribution
	 * 
	 */
	static final AppendPolicy<Map<Variable, IProbabilities<String>>> UNION_APPEND_POLICY = new AppendPolicy<Map<Variable, IProbabilities<String>>>() {
		public void append(
				Leaf<Map<Variable, IProbabilities<String>>> originalLeaf,
				Leaf<Map<Variable, IProbabilities<String>>> appendedLeaf) {
			if (appendedLeaf.get() == null) {
				return;
			}
			if (originalLeaf.get() == null) {
				appendedLeaf.set(null);
				return;
			}
			appendedLeaf.get().putAll(originalLeaf.get());
			if (appendedLeaf.get().values().contains(null)) {
				appendedLeaf.set(null);
			}
		}
	};

	/**
	 * 
	 * Policy to merge PTree together
	 * 
	 */
	static protected final MergePolicy<Map<Variable, IProbabilities<String>>> probaMergePolicy = new MergePolicy<Map<Variable, IProbabilities<String>>>(
			UNION_APPEND_POLICY, probaClonePolicy);

	/**
	 * 
	 * Constructor
	 * 
	 * @param forest
	 *          forest to use as model of the world
	 * 
	 */
	public SpiSvi(AbstractForest forest) {
		setForest(forest);
	}

	/**
	 * 
	 * Set the forest of this instance
	 * 
	 * @param forest
	 *          the forest to set
	 * 
	 */
	final protected void setForest(AbstractForest forest) {
		this.forest = forest;
		if (forest != null) {
			if (!"false".equals(forest.getParameters().getKeyS(
					Parameters.PARAM_KEY_ORDERING))) {
				partitionsToTree = new PartitionsToTree<Double>(
						new LeastSquareMeasure());
			}
			if (forest.getParameters().hasKey(
					Parameters.PARAM_KEY_THRESHOLD_CONVERGENCE)) {
				epsilonThresholdConvergence = forest.getParameters().getKeyD(
						Parameters.PARAM_KEY_THRESHOLD_CONVERGENCE);
			}
			if (forest.getParameters().hasKey(Parameters.PARAM_KEY_GAMMA)) {
				gamma = forest.getParameters().getKeyD(Parameters.PARAM_KEY_GAMMA);
			}
		}
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getModel()
	 */
	public AbstractForest getModel() {
		return forest;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getOptimalPolicy()
	 */
	public Tree<Set<String>> getOptimalPolicy() {
		return lastComputedPolicy;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getOptimalValueFunction()
	 */
	public Tree<Double> getOptimalValueFunction() {
		return lastComputedValueFunction;
	}

	/**
	 * 
	 * Compute the value function of a policy until it converges. This is the
	 * algorithm SSA(Tree(R),Tree(pi)) defined in Stochastic Dynamic Programming
	 * with Factored Representations (Boutilier, Dearden, Goldszmidt) page 33
	 * figure 18 This method does not modify this instance.
	 * 
	 * @param policyTree
	 *          the policy to evaluate
	 * @param valueTree
	 *          value tree to start the evaluation on
	 * @param terminationCriterion
	 *          termination criterion
	 * @return the value function of the policy
	 * 
	 */
	public Tree<Double> ssa(Tree<Double> valueTree, Tree<Set<String>> policyTree,
			SpiSvi.SSATerminationCriterion terminationCriterion) {
		// 1. Set Tree(Vpi0) to be Tree(R) (immediate reward)
		Tree<Double> treeVpik = valueTree.clone();
		// 2. Until termination, compute Tree(Vpikp1) = Regress(Tree(Vkpi),
		// Tree(pi))
		Tree<Double> treeVpikp1 = null;
		boolean termination = false;
		int numberOfIteration = 0;
		while (!termination) {
			log.out().print(
					"Evaluating V (" + numberOfIteration + "[treeVpik:"
							+ treeVpik.numberOfLeaves() + "/" + treeVpik.numberOfNodes()
							+ "]:");
			log.out().flush();
			treeVpikp1 = getRegress(treeVpik, policyTree);
			termination = (terminationCriterion.hasConverged(treeVpik, treeVpikp1) || (numberOfIteration > MAX_ITE));
			treeVpik = treeVpikp1;
			numberOfIteration++;
			log.out().println();
		}
		// 3. Return the final tree Tree(Vpi) = Tree(Vpin)
		return treeVpikp1;
	}

	/**
	 * 
	 * Algorithm Regress(Tree(V), Tree(Pi)) defined in Stochastic Dynamic
	 * Programming with Factored Representations (Boutilier, Dearden, Goldszmidt)
	 * page 31 figure 14
	 * 
	 * @param treeVpik
	 *          value tree to compute the value function on
	 * @param policyTree
	 *          tree to build the quality tree from (warning, this tree is
	 *          modified and is transformed from a policy tree to the computed
	 *          value tree)
	 * @return a quality tree computed from a policy tree
	 * 
	 */
	public Tree<Double> getRegress(Tree<Double> treeVpik,
			Tree<Set<String>> policyTree) {
		// 1. For each action occuring in Tree(Pi), call Regress(Tree(V), a)
		// to produce Tree(QaV)
		final Map<String, Tree<Double>> qaVMap = new LinkedHashMap<String, Tree<Double>>();
		Set<String> actionSet = new LinkedHashSet<String>();
		for (IPartition<Set<String>> leaf : policyTree.partitions()) {
			actionSet.addAll(leaf.get());
		}
		for (String action : actionSet) {
			log.out().print(" " + action + "(");
			log.out().flush();
			Tree<Double> treeQaV = getRegress(treeVpik, action);
			qaVMap.put(action, treeQaV);
			log.out().print(")");
		}
		// 2. At each leaf of Tree(Pi) labeled by a, append Tree(QaV) to Tree(Pi)
		// retaining the leaf labels (values) from Tree(QaV)
		// (deleting the action labels from Tree(Pi))
		Tree<Double> treeQpiV = new Tree<Double>();
		(new ConvertPolicy<Set<String>, Double>() {
			@Override
			public Double convert(Leaf<Set<String>> leaf) {
				return null;
			}
		}).convert(policyTree, treeQpiV);
		for (Leaf<Double> leaf : treeQpiV.getClonedLeafSet()) {
			Set<String> actions = policyTree.getLeaf(leaf.getContext()).get();
			String action = actions.iterator().next();
			leaf.replace(qaVMap.get(action).clone(leaf.getContext()));
		}
		// 3. Return the (simplified) resulting tree, Tree(QpiV)
		treeQpiV.simplify(EqualPolicy.valueLeafEqualPolicy);
		if (partitionsToTree != null) {
			treeQpiV = partitionsToTree.build(treeQpiV);
		}
		return treeQpiV;
	}

	/**
	 * 
	 * Algorithm Regress(Tree(V),a) defined in Stochastic Dynamic Programming with
	 * Factored Representations (Boutilier, Dearden, Goldszmidt) page 28 figure 12
	 * 
	 * @param treeVpi
	 *          value tree to apply the regress operator on
	 * @param action
	 *          action used to compute this method
	 * @return the quality tree Tree(QaV)
	 * 
	 */
	public Tree<Double> getRegress(Tree<Double> treeVpi, String action) {
		assert (action != null);
		Tree<Double> treeFVTreeQaV = treeBackproj(treeVpi, action);
		treeFVTreeQaV.simplify(EqualPolicy.valueLeafEqualPolicy);

		// 3. Discount FVTree(QaV) with discount factor B (multiply each leaf label
		// by B)
		new UnaryOperatorPolicy.MultiplyOperator(gamma).applyPolicy(treeFVTreeQaV);
		// 4. Append (and simplify) the reward tree, Tree(R), to FVTree(QaV), using
		// addition
		// as a combination function. The resulting tree is Tree(QaV)
		log.out().flush();
		Tree<Double> treeR = forest.getRewardTree(action);
		treeFVTreeQaV.append(treeR, appendRewardPolicy, null);
		return treeFVTreeQaV;
	}

	/**
	 * 
	 * Compute the backprojection of a value tree
	 * 
	 * @param treeVpi
	 *          the value tree to compute the back projection for
	 * @param action
	 *          the action of the backprojection
	 * @return the backprojection value tree
	 * 
	 */
	public Tree<Double> treeBackproj(Tree<Double> treeVpi, String action) {
		// Let PTree(QaV) be the tree returned by PRegress(Tree(V),a),
		// a) we assume PRegress (Tree(V),a) is simplified (contains
		// no redundant nodes)
		log.out().print(".");
		log.out().flush();
		Tree<Map<Variable, IProbabilities<String>>> treePTreeQaV = getPRegress(
				treeVpi.getRoot(), action, null);
		if (treePTreeQaV == null) {
			treePTreeQaV = new Tree<Map<Variable, IProbabilities<String>>>();
		}
		// 2. Construct FVTree(QaV) as follows. For each branch b E
		// PRegress(Tree(V),a)
		// (with leaf node lb):
		// (a) Let Prb be the joint distribution obtained from the product of the
		// individual
		// variable distributions labeling lb
		// (b) Compute vb = EPrb(b')V(b')
		// (c) Re-label leaf lb with vb
		log.out().flush();
		Tree<Double> treeFVTreeQaV = constructFVTree(action, treePTreeQaV, treeVpi);
		return treeFVTreeQaV;
	}

	/**
	 * 
	 * Algorithm PRegress(Tree(V),a) defined in Stochastic Dynamic Programming
	 * with Factored Representations (Boutilier, Dearden, Goldszmidt) page 29
	 * figure 13
	 * 
	 * @param vertex
	 *          node where to build the tree from
	 * @param actionRegressed
	 *          actions used to compute this method
	 * @param context
	 *          context to restrict the computation
	 * @return the probability tree PTree(QaV)
	 * 
	 */
	public Tree<Map<Variable, IProbabilities<String>>> getPRegress(
			Vertex<Double> vertex, String actionRegressed, Context context) {
		// 1. If Tree(V) contains a single (leaf) node (necessary labeled with a
		// value),
		// return an empty tree PTree(QaV)
		if (vertex.isLeaf) {
			return null;
		}
		final DNode<Double> dnode = (DNode<Double>) vertex;

		// 2. Let X be the variable labeling the root of Tree(V).
		// Let TXP be the tree-structured CPT for X in the DBN for a
		// (with leaves labeled by distributions over val(X))
		Tree<IProbabilities<String>> treeTPX = getModel().getTransitionFunction(
				dnode.getTestedVariable(), actionRegressed);
		if (treeTPX == null) {
			return null;
		}
		if (context != null) {
			treeTPX = treeTPX.clone(context);
		}
		// 3. For each xi E val(X) that occurs with positive probability in the
		// distribution
		// at some leaf in TPX
		Set<String> valX = forest.getValueSet(dnode.getTestedVariable());
		if (valX == null) {
			return null;
		}
		Map<String, Tree<Map<Variable, IProbabilities<String>>>> treeTPxiSet = new LinkedHashMap<String, Tree<Map<Variable, IProbabilities<String>>>>();
		for (String xi : valX) {
			// (a) Let TxiV be the subtree in Tree(V) attached to the root X by arc xi
			Vertex<Double> subTreeTxiV = dnode.getChild(xi);
			// (b) Let TxiP be the tree produced by calling PRegress(TxiV, a)
			Tree<Map<Variable, IProbabilities<String>>> treeTxiV = null;
			if ((subTreeTxiV != null) && (!subTreeTxiV.isLeaf)) {
				treeTxiV = getPRegress(subTreeTxiV, actionRegressed, context);
				treeTPxiSet.put(xi, treeTxiV);
			}
		}
		// 4. For each l E TPX, labeled with distribution Prl:
		Tree<Map<Variable, IProbabilities<String>>> treeTPXClone = new Tree<Map<Variable, IProbabilities<String>>>();
		PdToPDSetConvertPolicy pdToPDSetConvertPolicy = new PdToPDSetConvertPolicy(
				dnode.getTestedVariable());
		if (getModel().isMultiValued()) {
			pdToPDSetConvertPolicy.convert(getModel().getValueSets(), treeTPX,
					treeTPXClone);
		} else {
			pdToPDSetConvertPolicy.convert(treeTPX, treeTPXClone);
		}

		// When we have nothing to merge, then we can directly return the CPTTree
		if (treeTPxiSet.size() == 0) {
			return treeTPXClone;
		}
		for (Leaf<Map<Variable, IProbabilities<String>>> cptNode : treeTPXClone
				.getClonedLeafSet()) {
			if ((cptNode.get() != null)
					&& (cptNode.get().get(dnode.getTestedVariable()) != null)) {
				// (a) Let vall(X) = {xi E val(X):Prl(xi) > 0}
				Set<String> vall = cptNode.get().get(dnode.getTestedVariable())
						.getValueSet();
				// (b) Let Tl = Merge({TxiP: xi E vall(X)}), using union to combine the
				// leaf labels
				// (which are distributions over sets of variables)
				Tree<Map<Variable, IProbabilities<String>>> treeTl = null;
				for (String xi : vall) {
					Tree<Map<Variable, IProbabilities<String>>> treeTPxi = treeTPxiSet
							.get(xi);
					if (treeTPxi != null) {
						if (treeTl == null) {
							treeTl = treeTPxi.clone(probaClonePolicy, cptNode.getContext());
						} else {
							treeTl = probaMergePolicy.merge(treeTPxi, treeTl);
						}
					}
				}
				// (c) Revise TXP by appending Tl to leaf l, using union to combine
				// the leaf labels (distributions) of Tl with the label (distribution
				// over X) of leaf l
				if (treeTl != null) {
					treeTl.cut(cptNode.getContext());
					treeTPXClone.append(cptNode, treeTl, SpiSvi.UNION_APPEND_POLICY,
							probaClonePolicy);
				}
			}
		}
		return treeTPXClone;
	}

	/**
	 * 
	 * The Structured Policy Iteration (SPI) algorithm SPI(Tree(R),Tree(Pi))
	 * defined in Stochastic Dynamic Programming with Factored Representations
	 * (Boutilier, Dearden, Goldszmidt) page 36 figure 20
	 * 
	 * @return the policy computed using the SPI algorithm Tree(Pi*)
	 * 
	 */
	public Tree<Set<String>> getSPI() {
		int numberOfIteration = 0;
		String initialRandomAction = forest.getActionList().iterator().next();
		Tree<Set<String>> treePi = getNewSpiSviPolicy(initialRandomAction);
		Tree<Double> treeVpi = forest.getRewardTree(initialRandomAction).clone();
		// 1. Set Tree(PiPrime) = Tree(Pi)
		Tree<Set<String>> treePiPrime = treePi;
		// 2. Repeat
		do {
			numberOfIteration++;
			log.out().println("Iterating policy (" + numberOfIteration + ")...");
			// (a) Set Tree(Pi) = Tree(PiPrime)
			treePi = treePiPrime;
			// (b) Compute Tree(Vpi) using SSA
			treeVpi = ssa(treeVpi, treePi, supremumNormCriterion);
			// (c) Compute Tree(QaVpi)=Regress(Tree(Vpi),a) for each action a
			ActionValueMergePolicy mergePolicy = new ActionValueMergePolicy(
					getModel());
			for (String action : forest.getActionList()) {
				// (d) Merge the trees Tree(QaVpi) to obtain Tree(PiPrime)
				// (where PiPrime is the greedy policy w.r.t. Vpi)
				Tree<Double> treeQaVpi = getRegress(treeVpi, action);
				mergePolicy.merge(treeQaVpi, action);
			}
			treePiPrime = mergePolicy.policy();
			treePiPrime.simplify(exactActionLeafEqualPolicy);
			log.out().println("Iteration " + numberOfIteration + "\n");
		} // until PiPrime = Pi
		while (!treePiPrime.equals(treePi, exactActionLeafEqualPolicy)
				&& (numberOfIteration < MAX_ITE));
		setLastComputedValueFunction(treeVpi);
		setLastComputedPolicyFunction(treePiPrime);
		return getOptimalPolicy();
	}

	/**
	 * The Structured Value Iteration (SVI) algorithm SVI(Tree(R),Tree(Pi))
	 * defined in Stochastic Dynamic Programming with Factored Representations
	 * (Boutilier, Dearden, Goldszmidt) page 37 figure 23
	 * 
	 * @return the policy that converges using the SVI algorithm Tree(Pi*)
	 * 
	 */
	public Tree<Set<String>> getSVI() {
		int numberOfIteration = 0;
		boolean termination = true;

		// 1. Set Tree(V0) = Tree(R)
		Tree<Double> treeVk = forest.getRewardTree(
				forest.getActionList().iterator().next()).clone();
		Tree<Double> treeVkp1;
		// 2. Repeat
		do {
			numberOfIteration++;
			log.out().print(
					"Evaluating V (" + numberOfIteration + "[treeVpik:"
							+ treeVk.numberOfNodes() + "]:");
			treeVkp1 = null;
			for (String action : forest.getActionList()) {
				// (a) Compute Tree(QaVk)=Regress(Tree(Vk),a) for each action a
				Tree<Double> treeQaVk = getRegress(treeVk, action);
				// (b) Merge (via maximization) the trees Tree(QaVk) to obtain
				// Tree(Vk+1)
				treeVkp1 = maxPolicy.merge(treeQaVk, treeVkp1);
			}
			treeVkp1.simplify(EqualPolicy.valueLeafEqualPolicy);
			if (partitionsToTree != null) {
				treeVkp1 = partitionsToTree.build(treeVkp1);
			}
			log.out().print("E=");
			double spanSemiNorm = getSpanSemiNorm(treeVk, treeVkp1);
			log.out().println(spanSemiNorm);
			termination = (spanSemiNorm < epsilonThresholdConvergence)
					|| (numberOfIteration > MAX_ITE);
			treeVk = treeVkp1;
		} while (!termination);
		// 3. Set Tree(V*) = Tree(Vk+1)
		Tree<Double> treeVs = treeVkp1;
		setLastComputedValueFunction(treeVs);
		ActionValueMergePolicy mergePolicy = new ActionValueMergePolicy(getModel());
		for (String action : forest.getActionList()) {
			// 4. Compute Tree(QaV*) = Regress(Tree(V*),a) for each action a
			Tree<Double> treeQaVs = getRegress(treeVs, action);
			// 5. Merge the trees Tree(QaVs) to obtain Tree(Pis)
			mergePolicy.merge(treeQaVs, action);
		}
		Tree<Set<String>> treePis = mergePolicy.policy();
		treePis.simplify(actionLeafEqualPolicy);
		// 6. Return Tree(Pis) and Tree(V*)
		setLastComputedPolicyFunction(treePis);
		return treePis;
	}

	/**
	 * 
	 * Set a SpiSviTree as the last computed policy function
	 * 
	 * @param policy
	 *          the last computed policy tree
	 * 
	 */
	protected void setLastComputedPolicyFunction(Tree<Set<String>> policy) {
		lastComputedPolicy = policy;
	}

	/**
	 * 
	 * Set a SpiSviTree as the last computed value function
	 * 
	 * @param valueTree
	 *          the last computed value tree
	 * 
	 */
	protected void setLastComputedValueFunction(Tree<Double> valueTree) {
		lastComputedValueFunction = valueTree;
	}

	/**
	 * 
	 * Create a new SpiSviTree to represent a policy proposing the same action for
	 * the states
	 * 
	 * @param action
	 *          the only action returned by the policy
	 * 
	 * @return a new SpiSviTree policy returning action as the only action for all
	 *         the states
	 * 
	 */
	public Tree<Set<String>> getNewSpiSviPolicy(String action) {
		Tree<Set<String>> policy = new Tree<Set<String>>();
		Set<String> data = new LinkedHashSet<String>();
		data.add(action);
		((Leaf<Set<String>>) policy.getRoot()).set(data);
		return policy;
	}

	/**
	 * 
	 * Evaluate a policy
	 * 
	 * @param policy
	 *          the policy to evaluate
	 * 
	 * @return the value function corresponding to the policy
	 * 
	 */
	public Tree<Double> getSSA(Tree<Set<String>> policy) {
		return getSSA(policy, supremumNormCriterion);
	}

	/**
	 * 
	 * Evaluate a policy
	 * 
	 * @param policy
	 *          the policy to evaluate
	 * @param terminationCriterion
	 *          termination criterion
	 * 
	 * @return the value function corresponding to the policy
	 * 
	 */
	protected Tree<Double> getSSA(Tree<Set<String>> policy,
			SSATerminationCriterion terminationCriterion) {
		String initialRandomAction = forest.getActionList().iterator().next();
		Tree<Set<String>> workingPolicy = policy.clone();
		if (lastSSAResult == null) {
			lastSSAResult = forest.getRewardTree(initialRandomAction);
		}
		lastSSAResult = ssa(lastSSAResult, workingPolicy, terminationCriterion);
		return lastSSAResult;
	}

	/**
	 * @see animatlab.kodamat.core.io.IToXml#toXml(org.w3c.dom.Document,
	 *      animatlab.kodamat.core.io.IToXml.IToXmlPolicy)
	 */
	public Element toXml(Document document, IToXmlPolicy<Object> policy) {
		Element agentNode = document.createElement(NODE_NAME_AGENT);
		agentNode.appendChild(forest.toXml(document, null));

		if (lastComputedValueFunction != null) {
			Element valueFunctionNode = document
					.createElement(NODE_NAME_VALUEFUNCTION);
			valueFunctionNode.appendChild(lastComputedValueFunction.toXml(document,
					new IToXmlPolicy<Double>() {
						public void toXml(Element parent, Double data) {
							parent.setAttribute(AbstractFactoredModel.ATTRIBUTE_QUALITY,
									String.valueOf(data));
						}
					}));
			agentNode.appendChild(valueFunctionNode);
		}

		if (lastComputedPolicy != null) {
			Element policyNode = document.createElement(NODE_NAME_POLICY);
			policyNode.appendChild(lastComputedPolicy.toXml(document,
					new IToXmlPolicy<Set<String>>() {
						public void toXml(Element parent, Set<String> data) {
							for (String action : data) {
								Element actionNode = parent.getOwnerDocument().createElement(
										NODE_NAME_ACTION);
								actionNode.setAttribute(ATTRIBUTE_POLICY, action);
								parent.appendChild(actionNode);
							}
						}
					}));
			agentNode.appendChild(policyNode);
		}
		return agentNode;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getName()
	 */
	public String getName() {
		return getClass().getSimpleName();
	}

	/**
	 * 
	 * Quickly check if the probability is null or not
	 * 
	 * @param probabilityDistribution
	 *          the probability distribution to check on
	 * @param tests
	 *          the list of variable/value
	 * 
	 * @return true if the probability is null (meaning it is not necessary to
	 *         compute the probability of sub-space)
	 * 
	 */
	public boolean isProbabilityNull(
			Map<Variable, IProbabilities<String>> probabilityDistribution,
			Context tests) {
		for (Map.Entry<Variable, String> entry : tests.entrySet()) {
			IProbabilities<String> variableDistribution = probabilityDistribution
					.get(entry.getKey());
			if (variableDistribution != null) {
				if (!variableDistribution.getValueSet().contains(entry.getValue())) {
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * 
	 * A probability that a list of variable/value occurs for this node
	 * 
	 * @param probabilityDistribution
	 *          the probability distribution to check on
	 * @param tests
	 *          the list of variable/value
	 * 
	 * @return the computed probablity
	 * 
	 */
	final public double getProbability(
			Map<Variable, IProbabilities<String>> probabilityDistribution,
			Context tests) {
		double probability = 1.0;
		for (Map.Entry<Variable, String> entry : tests.entrySet()) {
			IProbabilities<String> variableDistribution = probabilityDistribution
					.get(entry.getKey());
			if (variableDistribution != null) {
				probability *= variableDistribution.getProbability(entry.getValue());
			}
		}
		return probability;
	}

	/**
	 * 
	 * Compute a value for a node at time t from the value function at time t-1,
	 * from a probability distribution and from a knownledge measure of the
	 * probability distribution
	 * 
	 * @param probabilityDistribution
	 *          the probability distribution
	 * @param vertex
	 *          the node in the value function computed at the previous iteration
	 * @return the new value computed for this iteration
	 * 
	 */
	double computeValue(
			Map<Variable, IProbabilities<String>> probabilityDistribution,
			Vertex<Double> vertex) {
		Context tests = vertex.getContext();
		if (isProbabilityNull(probabilityDistribution, tests)) {
			return 0;
		}
		if (vertex.isLeaf) {
			Leaf<Double> valueNode = (Leaf<Double>) vertex;
			double probability = getProbability(probabilityDistribution, tests);
			// assert(probability > 0);
			probabilitySum += probability;
			assert (probabilitySum <= 1 + EPSILON);
			return valueNode.get() * probability;
		}
		DNode<Double> dnode = (DNode<Double>) vertex;
		double quality = 0;
		for (Vertex<Double> child : dnode.children()) {
			quality += computeValue(probabilityDistribution, child);
		}
		return quality;
	}

	/**
	 * 
	 * Construct a value tree from a probabilities tree
	 * 
	 * @param action
	 *          action associated with the PTree
	 * @param currentTree
	 *          the probabilities tree
	 * @param lastTree
	 *          the tree at time t-1
	 * 
	 * @return the value tree computed
	 * 
	 */
	public Tree<Double> constructFVTree(String action,
			Tree<Map<Variable, IProbabilities<String>>> currentTree,
			final Tree<Double> lastTree) {

		// If current tree is only one empty node
		// that means that we don't have anything to compute
		Tree<Double> result = new Tree<Double>();
		if ((currentTree.numberOfNodes() <= 1)
				&& (((Leaf<?>) currentTree.getRoot()).get() == null)) {
			((Leaf<Double>) result.getRoot()).set(Parameters.getVMaxFromReward(forest
					.getRewardMax()));
			return result;
		}

		new ConvertPolicy<Map<Variable, IProbabilities<String>>, Double>() {
			@Override
			public Double convert(Leaf<Map<Variable, IProbabilities<String>>> leaf) {
				final Tree<Double> valueTree = lastTree;
				if (!isProbaIsKnown(leaf.get())) {
					return unknownProbabilityPolicy.probaToValue();
				}
				// This a leaf node, we must compute the quality associated with this
				// node
				// from Tree(V) and the value distribution
				probabilitySum = 0;
				double quality = computeValue(leaf.get(), valueTree.getRoot());
				assert ((probabilitySum <= 1 + EPSILON));
				return quality;
			}

			private boolean isProbaIsKnown(Map<Variable, IProbabilities<String>> proba) {
				if (proba == null) {
					return false;
				}
				return true;
			}
		}.convert(currentTree, result);
		return result;
	}

	/**
	 * 
	 * Build a string to show the computation of a value in a value function
	 * 
	 * @param probabilityDistribution
	 *          the probability distribution
	 * @param vertex
	 *          the node in the value function computed at the previous iteration
	 * 
	 * @return a string showing the sum
	 * 
	 * 
	 */
	public String fvtreeToString(
			Map<Variable, IProbabilities<String>> probabilityDistribution,
			Vertex<Double> vertex) {
		Context tests = vertex.getContext();
		if (isProbabilityNull(probabilityDistribution, tests)) {
			return "";
		}
		if (vertex.isLeaf) {
			Leaf<Double> valueNode = (Leaf<Double>) vertex;
			double probability = getProbability(probabilityDistribution, tests);
			return Parameters.format(valueNode.get()) + "["
					+ Context.contextToString(valueNode.getContext()) + "]*"
					+ String.valueOf(probability);
		}
		DNode<Double> dnode = (DNode<Double>) vertex;
		StringBuilder result = new StringBuilder();
		for (Vertex<Double> child : dnode.children()) {
			String childResult = fvtreeToString(probabilityDistribution, child);
			if (childResult.length() > 0) {
				if (result.length() > 0) {
					result.append(" +\n");
				}
				result.append(childResult);
			}
		}
		return result.toString();
	}

	/**
	 * 
	 * Check the validity of the quality
	 * 
	 * @param quality
	 *          the quality to check
	 * 
	 * @return true if the value is valid
	 * 
	 */
	boolean checkQuality(double quality) {
		if (!CHECK_QUALITY) {
			return true;
		}
		return ((!Double.isNaN(quality) && !Double.isInfinite(quality)) && (gamma
				* quality + getModel().getRewardMax()) < (Parameters
				.getVMaxFromReward(Math.max(0, Parameters.getRMaxFromReward(getModel()
						.getRewardMax()))) + 10 * Parameters.EPSILON));
	}

	/**
	 * 
	 * Compute the Supremum norm defined in Stochastic Dynamic Programming with
	 * Factored Representations (Boutilier, Dearden, Goldszmidt) page 34 WARNING
	 * This tree is modified to compute the difference
	 * 
	 * @param refTree
	 *          reference tree (warning, this tree is modified)
	 * @param valueTree
	 *          the tree to compare with this
	 * 
	 * @return true if the trees are similar
	 * 
	 */
	public static double getSupremumNorm(Tree<Double> refTree,
			Tree<Double> valueTree) {
		SpiSvi.MaxDiffComparePolicy policy = new SpiSvi.MaxDiffComparePolicy();
		refTree.compare(valueTree, policy);
		return policy.maxDiff;
	}

	/**
	 * 
	 * Compute the Span semi-norm defined in Stochastic Dynamic Programming with
	 * Factored Representations (Boutilier, Dearden, Goldszmidt) page 34 WARNING
	 * This tree is modified to compute the difference
	 * 
	 * @param valueTree
	 *          the tree to compare with (warning, this tree is modified)
	 * @param refTree
	 *          reference tree
	 * 
	 * @return true if the trees are similar
	 * 
	 */
	public double getSpanSemiNorm(Tree<Double> valueTree, Tree<Double> refTree) {
		valueTree.append(refTree, new AppendPolicy<Double>() {
			public void append(Leaf<Double> originalLeaf, Leaf<Double> appendedLeaf) {
				Double data01 = originalLeaf.get();
				Double data02 = appendedLeaf.get();
				if ((data01 == null) || (data02 == null)) {
					appendedLeaf.set(Parameters.getVMaxFromReward(getModel()
							.getRewardMax()));
				} else {
					appendedLeaf.set(data02 - data01);
				}
			}
		}, null);
		double maximalElement = -Double.MAX_VALUE;
		double minimalElement = Double.MAX_VALUE;
		for (IPartition<Double> valueNode : valueTree.partitions()) {
			maximalElement = Math.max(valueNode.get(), maximalElement);
			minimalElement = Math.min(valueNode.get(), minimalElement);
		}
		return maximalElement - minimalElement;
	}

	/**
	 * 
	 * Dump this as a html page
	 * 
	 * @param folderPath
	 *          the folder where to output the html file and figures
	 * @param title
	 *          title of the results
	 * 
	 */
	public void dump(String folderPath, String title) {
		HtmlPage htmlDotPage = new HtmlPage(folderPath, "index.html");
		log.out().println("Dumping in " + htmlDotPage.getAbsolutePath());
		htmlDotPage.setHtmlPageSetTitle(title);
		htmlDotPage.setTitle("Index");
		spisviToDot(htmlDotPage);
		getModel().dump(htmlDotPage);
		htmlDotPage.write();
		XmlUtils.save(this, null, folderPath + "/spiti.agent");
	}

	/**
	 * 
	 * Dump this into a html page
	 * 
	 * @param htmlPage
	 *          the html page where to dump this
	 * 
	 */
	protected void spisviToDot(HtmlPage htmlPage) {
		HtmlPage valuePage = htmlPage.newChild("Value Function");
		valuePage.append(DotUtils.toHtmlDot(getOptimalValueFunction(), null,
				htmlPage.getFolder(), "value"));

		HtmlPage policyPage = htmlPage.newChild("Greedy policy");
		Tree<Set<String>> policy = getOptimalPolicy();
		policyPage.append(DotUtils.toHtmlDot(policy,
				new IToDot.IToDotPolicy<Set<String>>() {
					public String toDotLabel(Set<String> data) {
						String result = data.toString();
						return result.substring(1, result.length() - 1);
					}
				}, htmlPage.getFolder(), "policy"));
		Tree<String> convertedPolicy = new Tree<String>();
		policyConversionPolicy.convert(policy, convertedPolicy);
		convertedPolicy.simplify(null);
		policyPage.append("Converted policy: <br/>");
		policyPage.append(DotUtils.toHtmlDot(convertedPolicy, null, htmlPage
				.getFolder(), "policyconverted"));
		if (convertedPolicy.numberOfNodes() == 1) {
			return;
		}
		PartitionsToTree<String> functionToTree = new PartitionsToTree<String>(
				new Chi2Measure(null));
		Tree<String> reorgPolicy01 = functionToTree.build(convertedPolicy);
		policyPage.append("Reorganized policy 01: <br/>");
		policyPage.append(DotUtils.toHtmlDot(reorgPolicy01, null, htmlPage
				.getFolder(), "policyreorg01"));

		// ReorganizeTree<String> reorganizeTree = new ReorganizeTree<String>();
		// Tree<String> reorgPolicy02 =
		// reorganizeTree.reorganize(convertedPolicy, getModel().getValueSets(), new
		// Chi2Measure(null));
		// policyPage.append("Reorganized policy 02: <br/>");
		// policyPage.append(
		// DotUtils.toHtmlDot(reorgPolicy02, null, htmlPage.getFolder(),
		// "policyreorg02"));

	}

	/**
	 * @see animatlab.kodamat.core.io.ILog#setLog(animatlab.kodamat.core.io.Log)
	 */
	public void setLog(Log log) {
		this.log = log;
		if (partitionsToTree != null) {
			partitionsToTree.setLog(log);
		}
	}

	/**
	 * 
	 * Set the policy conversion policy
	 * 
	 * @param policyConversionPolicy
	 *          the new conversion policy
	 * 
	 */
	public void setPolicyConversionPolicy(
			ConvertPolicy<Set<String>, String> policyConversionPolicy) {
		this.policyConversionPolicy = policyConversionPolicy;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#solve()
	 */
	public void solve() {
		getSVI();
	}

	/**
	 * 
	 * Set the ordering instance
	 * 
	 * @param measure
	 *          the measure to use to reorder
	 * 
	 */
	public void setOrdering(LeastSquareMeasure measure) {
		if (measure == null) {
			partitionsToTree = null;
			return;
		}
		partitionsToTree = new PartitionsToTree<Double>(measure);
	}

	/**
	 * 
	 * Set the policy to use when proba are unknown
	 * 
	 * @param unknownProbabilityPolicy
	 *          The unknownProbabilityPolicy to set.
	 * 
	 */
	public void setUnknownProbabilityPolicy(
			UnknownProbabilityPolicy unknownProbabilityPolicy) {
		this.unknownProbabilityPolicy = unknownProbabilityPolicy;
	}

	/**
	 * 
	 * Set a new policy to add the reward to a value function
	 * 
	 * @param appendRewardPolicy
	 *          the policy used to append the reward to the value function
	 * 
	 */
	public void setAppendRewardPolicy(AppendPolicy<Double> appendRewardPolicy) {
		this.appendRewardPolicy = appendRewardPolicy;
	}

}
