/*
 * 
 * Created on Jan 12, 2006
 *
 */
package animatlab.kodamat.reinforcementlearning.planning.spisvi;

import static animatlab.kodamat.core.io.XmlUtils.getFirstChild;
import static animatlab.kodamat.core.io.XmlUtils.getRootFromFile;
import static animatlab.kodamat.core.io.XmlUtils.save;
import static animatlab.kodamat.reinforcementlearning.model.forest.StaticForest.PROBLEM_EXTENSION;

import java.util.LinkedHashSet;
import java.util.Set;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import animatlab.kodamat.core.io.DataLoadingPolicy;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Chrono;
import animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractForest;
import animatlab.kodamat.reinforcementlearning.model.forest.StaticForest;
import animatlab.kodamat.reinforcementlearning.planning.IPlanning;

/**
 * 
 * SpiSvi instance loaded from a xml file to get a pre-computed optimal value
 * function and optimal policy
 * 
 * @author Thomas Degris
 * 
 */
public class StaticSpiSvi extends SpiSvi {
	/**
	 * 
	 * File to load to run the test
	 * 
	 */
	static public String AGENT_EXTENSION = ".spisvi.agent";

	/**
	 * 
	 * Constructor
	 * 
	 * @param element
	 *          xml tree to load the agent from
	 * 
	 */
	public StaticSpiSvi(Element element) {
		super(null);
		assert (element.getNodeName().equals(NODE_NAME_AGENT) || element
				.getNodeName().equals(AbstractForest.NODE_NAME_FOREST));
		if (element.getNodeName().equals(AbstractForest.NODE_NAME_FOREST)) {
			setForest(new StaticForest(element));
		}

		NodeList nodeList = element.getChildNodes();
		for (int index = 0; index < nodeList.getLength(); index++) {
			Node childNode = nodeList.item(index);
			if (childNode.getNodeType() == Node.ELEMENT_NODE) {
				if (childNode.getNodeName().equals(AbstractForest.NODE_NAME_FOREST)) {
					setForest(new StaticForest((Element) childNode));
				}
				if (childNode.getNodeName().equals(NODE_NAME_VALUEFUNCTION)) {
					setLastComputedValueFunction(loadValueFunction(getFirstChild(childNode)));
				}
				if (childNode.getNodeName().equals(NODE_NAME_POLICY)) {
					setLastComputedPolicyFunction(loadPolicy(getFirstChild(childNode)));
				}
			}
		}
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param forest
	 *          the model used by the algorithm
	 * 
	 */
	public StaticSpiSvi(StaticForest forest) {
		super(forest);
	}

	/**
	 * 
	 * Load the optimal policy from a xml tree
	 * 
	 * @param xmlNode
	 *          the root the xml tree
	 * 
	 * @return the tree loaded
	 * 
	 */
	public Tree<Set<String>> loadPolicy(Element xmlNode) {
		return new Tree<Set<String>>(xmlNode, new DataLoadingPolicy<Set<String>>() {
			public Set<String> xmlToData(Element element) {
				Set<String> result = new LinkedHashSet<String>();
				if (element.hasAttribute(ATTRIBUTE_POLICY)) {
					result.add(element.getAttribute(ATTRIBUTE_POLICY));
					return result;
				}
				NodeList nodeList = element.getChildNodes();
				for (int index = 0; index < nodeList.getLength(); index++) {
					Node childNode = nodeList.item(index);
					if (childNode.getNodeType() == Node.ELEMENT_NODE) {
						assert (childNode.getNodeName().equals(NODE_NAME_ACTION));
						result.add(((Element) childNode).getAttribute(ATTRIBUTE_POLICY));
					}
				}
				return result;
			}
		});
	}

	/**
	 * 
	 * Load the optimal value function from a xml tree
	 * 
	 * @param xmlNode
	 *          the root the xml tree
	 * 
	 * @return the tree loaded
	 * 
	 */
	public Tree<Double> loadValueFunction(Element xmlNode) {
		return new Tree<Double>(xmlNode, new DataLoadingPolicy<Double>() {
			public Double xmlToData(Element element) {
				return Double.parseDouble(element
						.getAttribute(AbstractFactoredModel.ATTRIBUTE_QUALITY));
			}
		});
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getModel()
	 */
	@Override
	public StaticForest getModel() {
		return (StaticForest) super.getModel();
	}

	/**
	 * 
	 * Convert a forest xml file into an agent xml file
	 * 
	 * @param forest
	 *          the forest to convert
	 * @param outputFile
	 *          file where to save the forest
	 * @return the planning method used to compute the conversion
	 * 
	 */
	static public IPlanning convertProblemToAgent(AbstractForest forest,
			String outputFile) {
		SpiSvi spisvi = new SpiSvi(forest);
		spisvi.solve();
		spisvi.setLastComputedValueFunction(spisvi
				.getSSA(spisvi.getOptimalPolicy()));
		save(spisvi, null, outputFile + StaticSpiSvi.AGENT_EXTENSION);
		return spisvi;
	}

	/**
	 * 
	 * Main method
	 * 
	 * @param args
	 *          list of arguments passed to the program
	 * 
	 */
	public static void main(String[] args) {
		Chrono chrono = new Chrono();
		IPlanning planning = convertProblemToAgent(new StaticForest(
				getRootFromFile(args[0] + PROBLEM_EXTENSION)), args[0]);
		System.out.println("Time: " + chrono.toString());
		System.out.println("Size: "
				+ planning.getOptimalValueFunction().numberOfNodes());
	}

}
