package animatlab.kodamat.reinforcementlearning.planning.spisvi;

import java.util.Map;

import animatlab.kodamat.core.functions.IPartition;
import animatlab.kodamat.core.functions.UnaryOperatorPolicy.MultiplyOperator;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.AppendPolicy;
import animatlab.kodamat.core.trees.ConvertPolicy;
import animatlab.kodamat.core.trees.EqualPolicy;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.MergePolicy;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractForest;

/**
 * 
 * Incremental planning method based on tree representations
 * 
 * @author Thomas Degris
 * 
 */
public class TreeBasedPlanning extends SpiSvi {

	/**
	 * 
	 * Policy to convert a string into a double
	 * 
	 */
	private static class ProbaDistributionToProba extends
			ConvertPolicy<IProbabilities<String>, Double> {
		/**
		 * 
		 * Value to get the probability for
		 * 
		 */
		private String value;

		/**
		 * 
		 * Constructor
		 * 
		 */
		public ProbaDistributionToProba() {
			super();
		}

		/**
		 * @see animatlab.kodamat.core.trees.ConvertPolicy#convert(animatlab.kodamat.core.trees.Leaf)
		 */
		@Override
		public Double convert(Leaf<IProbabilities<String>> leaf) {
			if (leaf.get() == null) {
				return null;
			}
			return leaf.get().getProbability(value);
		}

		/**
		 * 
		 * Set value
		 * 
		 * @param value
		 *          the value to set.
		 * 
		 */
		public void setValue(String value) {
			this.value = value;
		}
	}

	/**
	 * 
	 * Policy used to convert a conditional probability distribution into a
	 * conditional probability function
	 * 
	 */
	final private ProbaDistributionToProba probaConvertPolicy = new ProbaDistributionToProba();

	/**
	 * 
	 * Operator used to multiply a tree by a constant
	 * 
	 */
	final protected MultiplyOperator multiplyOperator = new MultiplyOperator();

	/**
	 * 
	 * Policy used to multiply two trees of double
	 * 
	 */
	final private MergePolicy<Double> productPolicy = new MergePolicy<Double>(
			new AppendPolicy<Double>() {
				public void append(Leaf<Double> originalLeaf, Leaf<Double> appendedLeaf) {
					if ((appendedLeaf.get() == null) || (originalLeaf.get() == null)) {
						appendedLeaf.set(null);
					} else {
						appendedLeaf.set(appendedLeaf.get() * originalLeaf.get());
					}
				}
			}, null);

	/**
	 * 
	 * Constructor
	 * 
	 * @param forest
	 *          forest to use as model of the world
	 * 
	 */
	public TreeBasedPlanning(AbstractForest forest) {
		super(forest);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getName()
	 */
	@Override
	public String getName() {
		return "ltbp";
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.spisvi.SpiSvi#treeBackproj(animatlab.kodamat.core.trees.Tree,
	 *      java.lang.String)
	 */
	@Override
	public Tree<Double> treeBackproj(Tree<Double> treeVpi, String action) {
		assert (action != null);
		Tree<Double> result = new Tree<Double>();
		((Leaf<Double>) result.getRoot()).set(0.0);
		for (IPartition<Double> leaf : treeVpi.partitions()) {
			Tree<Double> probaTree = computeRelevantProba(leaf, action);
			multiplyOperator.setConstant(leaf.get());
			multiplyOperator.applyPolicy(probaTree);
			result.append(probaTree, appendRewardPolicy, null);
		}
		result = partitionsToTree.build(result);
		return result;
	}

	/**
	 * 
	 * Compute the regress operation on one leaf
	 * 
	 * @param leaf
	 *          the leaf to compute the regress operation for
	 * @param action
	 *          the action regressed
	 * @return the new tree computed from the leaf
	 * 
	 */
	protected Tree<Double> computeRelevantProba(IPartition<Double> leaf,
			String action) {
		Tree<Double> result = new Tree<Double>();
		((Leaf<Double>) result.getRoot()).set(1.0);
		for (Map.Entry<Variable, String> entry : leaf.getContext().entrySet()) {
			Tree<IProbabilities<String>> tree = getModel().getTransitionFunction(
					entry.getKey(), action);
			Tree<Double> probaTree = new Tree<Double>();
			probaConvertPolicy.setValue(entry.getValue());
			probaConvertPolicy.convert(tree, probaTree);
			result = productPolicy.merge(result, probaTree);
			result.simplify(EqualPolicy.valueLeafEqualPolicy);
		}
		return result;

	}

}
