package animatlab.kodamat.reinforcementlearning.planning.spudd;

import java.io.File;
import java.io.IOException;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.reinforcementlearning.model.ILearningModel;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractForest;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractLearningForest;
import animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning;

/**
 * 
 * Incremental SPUDD
 * 
 * @author Thomas Degris
 * 
 */
public class LearningSpudd extends Spudd implements IIncrementalPlanning {
	/**
	 * 
	 * Suffix used by Spudd to store ADD file
	 * 
	 */
	private static final String OPTDUAL_ADD_SUFFIX = "-OPTDual.ADD";

	/**
	 * 
	 * Number of iteration to execute by step
	 * 
	 */
	final static private int NUMBER_OF_ITERATION_BY_STEP = 1;

	/**
	 * 
	 * Flag to indicate weither or not to reset the value function when the reward
	 * changes
	 * 
	 */
	protected boolean resetRewardFlag = true;

	/**
	 * 
	 * Wrapper around the last policy computed
	 * 
	 */
	private class ActionValueWrapper implements IFunction<Double> {
		/**
		 * 
		 * Action represented by this wrapper
		 * 
		 */
		final private String action;

		/**
		 * 
		 * Constructor
		 * 
		 * @param action
		 *          action represented by this wrapper
		 * 
		 */
		public ActionValueWrapper(String action) {
			this.action = action;
		}

		/**
		 * @see animatlab.kodamat.core.functions.IFunction#classify(Context)
		 */
		public Double classify(Context state) {
			if ((lastComputedPolicy != null)
					&& lastComputedPolicy.classify(state).contains(action)) {
				return 1.0;
			}
			return 0.0;
		}

		/**
		 * @see animatlab.kodamat.core.functions.IFunction#numberOfNodes()
		 */
		public int numberOfNodes() {
			return lastComputedValueFunction.numberOfNodes();
		}

		/**
		 * @see animatlab.kodamat.core.functions.IFunction#clone()
		 */
		@Override
		public IFunction<Double> clone() {
			assert (false);
			return null;
		}
	}

	/**
	 * 
	 * File used to store the last ADDs representing the last computed value
	 * function
	 * 
	 */
	private File addVkFile = null;

	/**
	 * 
	 * File used to store the last ADDs representing the last computed value
	 * function and that we have to remove
	 * 
	 */
	private File oldVkFile = null;

	/**
	 * 
	 * Wrapper arround the last policy computed
	 * 
	 */
	final private Map<String, IFunction<Double>> actionValueFunctions = new LinkedHashMap<String, IFunction<Double>>();

	/**
	 * 
	 * Constructor
	 * 
	 */
	public LearningSpudd() {
		this(null);
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param forest
	 *          forest used to compute a policy
	 * 
	 */
	public LearningSpudd(AbstractForest forest) {
		super(forest);
		if (forest != null) {
			for (String action : forest.getActionList()) {
				actionValueFunctions.put(action, new ActionValueWrapper(action));
			}
			((AbstractLearningForest) forest).getInductionMeasure().noEmptyLeaf = true;
		}
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning#getIncrementalPlanning(Context,
	 *      java.lang.String, Context, java.util.List)
	 */
	public Map<String, IFunction<Double>> getIncrementalPlanning(
			Context lastState, String action, Context currentState,
			List<Double> reward) {
		if (getModel().getRewardTree(null) == null) {
			return null;
		}
		if (getModel().getActionTestedSet().size() < getModel().getActionList()
				.size()) {
			return null;
		}
		solve();
		cleanTemporaryFile();
		return actionValueFunctions;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning#getActionValueFunctions()
	 */
	public Map<String, IFunction<Double>> getActionValueFunctions() {
		return actionValueFunctions;
	}

	/**
	 * 
	 * Clean temporary file used for incremental learning
	 * 
	 * 
	 */
	private void cleanTemporaryFile() {
		if (oldVkFile != null) {
			oldVkFile.delete();
			new File(oldVkFile.getAbsolutePath() + OPTDUAL_ADD_SUFFIX).delete();
			new File(oldVkFile.getAbsolutePath() + "-stats.dat").delete();
		}
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getModel()
	 */
	@Override
	public AbstractLearningForest getModel() {
		return (AbstractLearningForest) super.getModel();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning#newInstance(animatlab.kodamat.reinforcementlearning.model.ILearningModel)
	 */
	public IIncrementalPlanning newInstance(ILearningModel learningModel) {
		return new LearningSpudd((AbstractForest) learningModel);
	}

	/**
	 * 
	 * Set the reset the value function when the reward changes flag
	 * 
	 * @param resetRewardFlag
	 *          new value of the flag
	 * 
	 */
	public void setResetRewardFlag(boolean resetRewardFlag) {
		this.resetRewardFlag = resetRewardFlag;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.spudd.Spudd#getAdditionalSpuddOptions()
	 */
	@Override
	protected String getAdditionalSpuddOptions() {
		StringBuilder result = new StringBuilder();
		if ((addVkFile != null)
				&& (!resetRewardFlag || !getModel().isRewardFunctionModifFlag())) {
			result
					.append("-dual " + addVkFile.getAbsolutePath() + OPTDUAL_ADD_SUFFIX);
		}
		getModel().resetModifiedFlag();
		oldVkFile = addVkFile;
		addVkFile = null;
		try {
			addVkFile = File.createTempFile("spudd", "");
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
		}
		result.append(" -o " + addVkFile.getAbsolutePath());
		result.append(" -i " + NUMBER_OF_ITERATION_BY_STEP);
		return result.toString();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getName()
	 */
	@Override
	public String getName() {
		return "learningspudd";
	}

	/**
	 * @see java.lang.Object#finalize()
	 */
	@Override
	public void finalize() {
		cleanTemporaryFile();
	}
}
