package animatlab.kodamat.reinforcementlearning.planning.spudd;

import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import animatlab.kodamat.core.adds.ADD;
import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.io.DataLoadingPolicy;
import animatlab.kodamat.core.io.Exec;
import animatlab.kodamat.core.io.IToXml;
import animatlab.kodamat.core.io.Log;
import animatlab.kodamat.core.io.XmlUtils;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.DNode;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Vertex;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractForest;
import animatlab.kodamat.reinforcementlearning.planning.IPlanning;

/**
 * 
 * Spudd wrapper on the spudd executable
 * 
 * @author Thomas Degris
 * 
 */
public class Spudd implements IPlanning, IToXml<Object> {
	/**
	 * 
	 * Name of the node in xml to represent an decision node
	 * 
	 */
	static public final String NODE_NAME_AGENT = ADD.NAMESPACE + "agent";

	/**
	 * 
	 * Name of the node in xml to represent an decision node
	 * 
	 */
	private static final String ROOT_NAME = ADD.NAMESPACE + ":" + "spuddspiti";

	/**
	 * 
	 * Name of the node in xml to represent an decision node
	 * 
	 */
	static protected final String NODE_NAME_VALUE_FUNCTION = ADD.NAMESPACE + ":"
			+ "valuefunction";

	/**
	 * 
	 * Name of the node in xml to represent an decision node
	 * 
	 */
	static protected final String NODE_NAME_VALUE_FUNCTION_SIZE = ADD.NAMESPACE
			+ ":" + "valfunsize";

	/**
	 * 
	 * Name of the node in xml to represent an decision node
	 * 
	 */
	static protected final String NODE_NAME_POLICY = ADD.NAMESPACE + ":"
			+ "policy";

	/**
	 * 
	 * Name of the node in xml to represent an decision node
	 * 
	 */
	protected static final String NODE_NAME_ACTION = ADD.NAMESPACE + ":"
			+ "action";

	/**
	 * 
	 * Name of the attribute to named a variable
	 * 
	 */
	protected static final String ATTRIBUTE_POLICY = "action";

	/**
	 * 
	 * Prefix for variable names
	 * 
	 */
	private static final String VARIABLES_PREFIX = "var";

	/**
	 * 
	 * Prefix for variable values
	 * 
	 */
	private static final String VALUES_PREFIX = "val";

	/**
	 * 
	 * Prefix for actions
	 * 
	 */
	private static final String ACTION_PREFIX = "act";

	/**
	 * 
	 * Prefix used for spuddinc installation
	 * 
	 */
	private static final String SPUDDINC_OPTIONS = "-r sift";

	/**
	 * 
	 * Path of spuddinc
	 * 
	 */
	private static final String SPUDDINC_PATH = "spuddinc";

	/**
	 * 
	 * Class used to convert a tree into a Spudd declaration
	 * 
	 * @param <T>
	 *          type of the tree to convert
	 * 
	 */
	static class TreeToDeclaration<T> {
		/**
		 * 
		 * Interface used to convert data to Spudd declaration
		 * 
		 * @param <T>
		 *          the type of data to convert
		 * 
		 */
		interface DataToDeclarationPolicy<T> {
			/**
			 * 
			 * Convert a data into a spudd declaration
			 * 
			 * @param data
			 *          the data to convert
			 * @return the string representing the string declaration
			 * 
			 */
			String dataToDeclaration(T data);
		}

		/**
		 * 
		 * Convert a tree into a Spudd declaration
		 * 
		 * @param node
		 *          the root of the tree to convert
		 * @param dataToDeclPolicy
		 *          the policy to use to convert the data
		 * 
		 * @return the string representing the declaration
		 * 
		 */
		String convert(Vertex<T> node, DataToDeclarationPolicy<T> dataToDeclPolicy) {
			StringBuilder result = new StringBuilder();
			String value = node.getTestedValue();
			if (value != null) {
				result.append("(");
				result.append(VALUES_PREFIX);
				result.append(value);
				result.append("\t");
			}
			result.append("(");
			if (node.isLeaf) {
				Leaf<T> leaf = (Leaf<T>) node;
				if (dataToDeclPolicy == null) {
					result.append(leaf.get().toString());
				} else {
					result.append(dataToDeclPolicy.dataToDeclaration(leaf.get()));
				}
			} else {
				result.append(convertDNode((DNode<T>) node, dataToDeclPolicy));
			}
			result.append(")");
			if (value != null) {
				result.append(")");
			}
			return result.toString();
		}

		/**
		 * 
		 * Convert a decision node into a Spudd declaration
		 * 
		 * @param dnode
		 *          the decision node to convert
		 * @param dataToDeclPolicy
		 *          the policy to use to convert the data
		 * @return the Spudd declaration corresponding to this decision
		 * 
		 */
		private String convertDNode(DNode<T> dnode,
				DataToDeclarationPolicy<T> dataToDeclPolicy) {
			StringBuilder result = new StringBuilder();
			result.append(VARIABLES_PREFIX);
			result.append(dnode.getTestedVariable().toString());
			result.append("\t");
			for (String childValue : getSortedValueSet(dnode.getChildrenValues())) {
				result.append(convert(dnode.getChild(childValue), dataToDeclPolicy));
				result.append(" ");
			}
			return result.toString();
		}
	}

	/**
	 * 
	 * Instance used to format value number
	 * 
	 */
	static final NumberFormat numberFormat = NumberFormat.getInstance();

	static {
		numberFormat.setGroupingUsed(false);
		numberFormat.setMaximumFractionDigits(6);
		numberFormat.setMinimumFractionDigits(6);
	}

	/**
	 * 
	 * Instance to convert a tree of double into a Spudd declaration
	 * 
	 */
	final static TreeToDeclaration<Double> doubleToDeclaration = new TreeToDeclaration<Double>();

	/**
	 * 
	 * Instance to convert a tree of double into a Spudd declaration
	 * 
	 */
	final static TreeToDeclaration<IProbabilities<String>> probaToDeclaration = new TreeToDeclaration<IProbabilities<String>>();

	/**
	 * 
	 * Forest used to compute a policy
	 * 
	 */
	protected AbstractForest forest;

	/**
	 * 
	 * Where to log SPI/Svi
	 * 
	 */
	public Log log = Log.sysout;

	/**
	 * 
	 * Gamma value used in planning computation
	 * 
	 */
	public double gamma = Parameters.GAMMA;

	/**
	 * 
	 * Last policy computed
	 * 
	 */
	protected ADD<Set<String>> lastComputedPolicy = null;

	/**
	 * 
	 * Last value function computed
	 * 
	 */
	protected ADD<Double> lastComputedValueFunction = null;

	/**
	 * 
	 * When we don't parse the value function, we still read its size from the xml
	 * file
	 * 
	 */
	protected Integer lastComputedValueFunctionSize = null;

	/**
	 * 
	 * Constructor
	 * 
	 * @param forest
	 *          forest used to compute a policy
	 * 
	 */
	public Spudd(AbstractForest forest) {
		this.forest = forest;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getOptimalPolicy()
	 */
	public ADD<Set<String>> getOptimalPolicy() {
		return lastComputedPolicy;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getOptimalValueFunction()
	 */
	public IFunction<Double> getOptimalValueFunction() {
		if (lastComputedValueFunction != null) {
			return lastComputedValueFunction;
		}
		if (lastComputedValueFunctionSize == null) {
			return null;
		}
		return new IFunction<Double>() {
			public Double classify(Context state) {
				assert (false);
				return null;
			}

			public int numberOfNodes() {
				return lastComputedValueFunctionSize;
			}

			@Override
			public IFunction<Double> clone() {
				assert (false);
				return null;
			}
		};
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getName()
	 */
	public String getName() {
		return "spudd";
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getModel()
	 */
	public AbstractForest getModel() {
		return forest;
	}

	/**
	 * @see animatlab.kodamat.core.io.ILog#setLog(animatlab.kodamat.core.io.Log)
	 */
	public void setLog(Log log) {
		this.log = log;
	}

	/**
	 * @see animatlab.kodamat.core.io.IToXml#toXml(org.w3c.dom.Document,
	 *      animatlab.kodamat.core.io.IToXml.IToXmlPolicy)
	 */
	public Element toXml(Document document, IToXmlPolicy<Object> policy) {
		Element agentNode = document.createElement(NODE_NAME_AGENT);
		agentNode.appendChild(forest.toXml(document, policy));

		if (lastComputedValueFunction != null) {
			Element valueFunctionNode = document
					.createElement(NODE_NAME_VALUE_FUNCTION);
			valueFunctionNode.appendChild(lastComputedValueFunction.toXml(document,
					new IToXmlPolicy<Double>() {
						public void toXml(Element parent, Double data) {
							parent.setAttribute(AbstractFactoredModel.ATTRIBUTE_QUALITY,
									String.valueOf(data));
						}
					}));
			agentNode.appendChild(valueFunctionNode);
		}

		if (lastComputedPolicy != null) {
			Element policyNode = document.createElement(NODE_NAME_POLICY);
			policyNode.appendChild(lastComputedPolicy.toXml(document,
					new IToXmlPolicy<Set<String>>() {
						public void toXml(Element parent, Set<String> data) {
							for (String action : data) {
								Element actionNode = parent.getOwnerDocument().createElement(
										NODE_NAME_ACTION);
								actionNode.setAttribute(ATTRIBUTE_POLICY, action);
								parent.appendChild(actionNode);
							}
						}
					}));
			agentNode.appendChild(policyNode);
		}
		return agentNode;
	}

	/**
	 * 
	 * @return the spudd representation of the forest
	 * 
	 */
	private String forestToProblem() {
		StringBuilder result = new StringBuilder("// Problem declaration\n");
		result.append(getVariablesDeclaration());
		result.append("\n");
		for (String action : forest.getActionList()) {
			result.append(getActionDeclaration(action));
			result.append("\n");
		}
		result.append("reward ");
		result.append(doubleToDeclaration.convert(forest.getRewardTree(null)
				.getRoot(), null));
		result.append("\ndiscount " + gamma + "\n");
		result.append("tolerance " + Parameters.EPSILON_THRESHOLD_CONVERGENCE
				+ "\n");
		result.append("\n");
		return result.toString();
	}

	/**
	 * 
	 * Build the declarations for an action
	 * 
	 * @param action
	 *          the action to get the declaration from
	 * @return a string representing the declaration of the transition model for
	 *         this action
	 * 
	 */
	private String getActionDeclaration(String action) {
		StringBuilder result = new StringBuilder("action ");
		result.append(ACTION_PREFIX);
		result.append(action);
		result.append("\n");
		for (Variable variable : forest.getVariableSet()) {
			if (forest.getValueSet(variable).size() > 1) {
				result.append("\t");
				result.append(VARIABLES_PREFIX);
				result.append(variable.toString());
				result.append("\t");
				final List<String> values = getSortedValueSet(forest
						.getValueSet(variable));
				result
						.append(probaToDeclaration
								.convert(
										forest.getTransitionFunction(variable, action).getRoot(),
										new TreeToDeclaration.DataToDeclarationPolicy<IProbabilities<String>>() {
											public String dataToDeclaration(
													IProbabilities<String> data) {
												StringBuilder result = new StringBuilder();
												for (String value : values) {
													String proba = numberFormat.format(data
															.getProbability(value));
													result.append(proba.replace(',', '.'));
													result.append(" ");
												}
												return result.toString();
											}
										}));
				result.append("\n");
			}
		}
		result.append("endaction");
		return result.toString();
	}

	/**
	 * 
	 * @return the declarations of the different variables of the problem
	 * 
	 */
	private String getVariablesDeclaration() {
		StringBuilder result = new StringBuilder("(variables ");
		for (Map.Entry<Variable, Set<String>> entry : forest.getValueSets()
				.entrySet()) {
			if (entry.getValue().size() > 1) {
				result.append("(" + VARIABLES_PREFIX);
				result.append(entry.getKey().toString());
				result.append(" ");
				for (String value : getSortedValueSet(entry.getValue())) {
					result.append(VALUES_PREFIX);
					result.append(value);
					result.append(" ");
				}
				result.append(") ");
			}
		}
		result.append(")");
		return result.toString();
	}

	/**
	 * 
	 * Sort a set of value
	 * 
	 * @param values
	 *          the set of value to sort
	 * 
	 * @return a list with the value sorted
	 * 
	 */
	static protected List<String> getSortedValueSet(Set<String> values) {
		ArrayList<String> result = new ArrayList<String>(values);
		Collections.sort(result);
		return result;
	}

	/**
	 * 
	 * Run Spudd Value Iteration method
	 * 
	 */
	public void solve() {
		if (!checkModel()) {
			return;
		}
		File inputFile = forestToSpuddFile();
		File outputFile = runSpudd(inputFile);
		Element rootNode = XmlUtils.getRootFromFile(outputFile.getAbsolutePath());
		assert (rootNode.getNodeName().equals(ROOT_NAME));
		NodeList nodeList = rootNode.getChildNodes();
		for (int index = 0; index < nodeList.getLength(); index++) {
			Node childNode = nodeList.item(index);
			if (childNode.getNodeType() == Node.ELEMENT_NODE) {
				if (childNode.getNodeName().equals(NODE_NAME_VALUE_FUNCTION)) {
					lastComputedValueFunction = loadValueFunction(XmlUtils
							.getFirstChild(childNode));
					lastComputedValueFunctionSize = null;
				}
				if (childNode.getNodeName().equals(NODE_NAME_VALUE_FUNCTION_SIZE)) {
					lastComputedValueFunction = null;
					lastComputedValueFunctionSize = Integer
							.parseInt(((Element) childNode).getAttribute("size"));
				}
				if (childNode.getNodeName().equals(NODE_NAME_POLICY)) {
					lastComputedPolicy = loadPolicy(XmlUtils.getFirstChild(childNode));
				}
			}
		}
		inputFile.delete();
		outputFile.delete();
	}

	/**
	 * 
	 * Load the policy from a xml tree
	 * 
	 * @param rootNode
	 *          the root node of the xml tree
	 * 
	 * @return the loaded policy
	 * 
	 */
	protected ADD<Set<String>> loadPolicy(Node rootNode) {
		return new ADD<Set<String>>((Element) rootNode,
				new DataLoadingPolicy<Set<String>>() {
					public Set<String> xmlToData(Element element) {
						final String dataAttr = element.getAttribute("data");
						if (dataAttr.length() > 0) {
							return loadFromAttr(dataAttr);
						}
						return loadChildNodes(element);
					}

					/**
					 * 
					 * Load the set of actions from a set of xml node
					 * 
					 * @param element
					 *          the parent node to load
					 * @return the set of actions loaded
					 * 
					 */
					private Set<String> loadChildNodes(Element element) {
						Set<String> result = new LinkedHashSet<String>();
						NodeList nodeList = element.getChildNodes();
						for (int index = 0; index < nodeList.getLength(); index++) {
							Node childNode = nodeList.item(index);
							if (childNode.getNodeType() == Node.ELEMENT_NODE) {
								assert (childNode.getNodeName().equals(NODE_NAME_ACTION));
								result
										.add(((Element) childNode).getAttribute(ATTRIBUTE_POLICY));
							}
						}
						return result;
					}

					/**
					 * 
					 * Load the set of actions from an attribute
					 * 
					 * @param dataAttr
					 *          the attribute to load the actions from
					 * @return the set of actions loaded
					 * 
					 */
					private Set<String> loadFromAttr(final String dataAttr) {
						String[] actions = dataAttr.split(":");
						Set<String> result = new LinkedHashSet<String>();
						for (String action : actions) {
							result.add(action);
						}
						return result;
					}
				});
	}

	/**
	 * 
	 * Load the value function from a xml tree
	 * 
	 * @param rootNode
	 *          the root node of the xml tree
	 * 
	 * @return the loaded value function
	 * 
	 */
	protected ADD<Double> loadValueFunction(Node rootNode) {
		return new ADD<Double>((Element) rootNode,
				DataLoadingPolicy.doubleLoadingPolicy);
	}

	/**
	 * 
	 * Run Spudd algorithm
	 * 
	 * @param inputFile
	 *          the file describing the problem to solve
	 * 
	 * @return the output file containing spiti data
	 * 
	 */
	private File runSpudd(File inputFile) {
		File outputFile = null;
		try {
			outputFile = File.createTempFile("output", ".spd");
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
			return null;
		}

		StringBuilder spuddCommand = new StringBuilder();
		spuddCommand.append(SPUDDINC_PATH);
		spuddCommand.append(" " + inputFile.getAbsolutePath());
		spuddCommand.append(" " + SPUDDINC_OPTIONS);
		spuddCommand.append(" -spiti " + outputFile.getAbsoluteFile());
		spuddCommand.append(" " + getAdditionalSpuddOptions());
		int exitCode = Exec.exec(spuddCommand.toString(), log);
		assert (exitCode == 0);
		return outputFile;
	}

	/**
	 * 
	 * @return additional Spudd options
	 * 
	 */
	protected String getAdditionalSpuddOptions() {
		return "";
	}

	/**
	 * 
	 * Save the forest into a file using spudd input format
	 * 
	 * @return the file used to save the forest
	 * 
	 */
	private File forestToSpuddFile() {
		File inputFile = null;
		try {
			inputFile = File.createTempFile("spudd", ".spd");
			PrintStream outputStream = new PrintStream(inputFile);
			outputStream.println(forestToProblem());
			outputStream.flush();
			outputStream.close();
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
		}
		return inputFile;
	}

	/**
	 * 
	 * @return check if the model is correct to run Spudd
	 * 
	 */
	private boolean checkModel() {
		for (Map.Entry<Variable, Set<String>> entry : forest.getValueSets()
				.entrySet()) {
			if (entry.getValue().size() <= 1) {
				return false;
			}
		}
		if (getModel().getRewardTree(null).numberOfNodes() <= 1) {
			return false;
		}
		return true;
	}

	/**
	 * 
	 * Set the forest of this instance
	 * 
	 * @param forest
	 *          the forest to set
	 * 
	 */
	final protected void setForest(AbstractForest forest) {
		this.forest = forest;
		if (forest != null) {
			if (forest.getParameters().hasKey(Parameters.PARAM_KEY_GAMMA)) {
				gamma = forest.getParameters().getKeyD(Parameters.PARAM_KEY_GAMMA);
			}
		}
	}
}
