package animatlab.kodamat.reinforcementlearning.planning.stats;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import animatlab.kodamat.core.IProblemDefinition;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.reinforcementlearning.agent.IAgent;
import animatlab.kodamat.reinforcementlearning.agent.IAgentFactory;
import animatlab.kodamat.reinforcementlearning.agent.ModelBasedAgentFactory;
import animatlab.kodamat.reinforcementlearning.core.stats.Measure;
import animatlab.kodamat.reinforcementlearning.core.stats.MeasureWriter;
import animatlab.kodamat.reinforcementlearning.core.trajectory.Trajectory;
import animatlab.kodamat.reinforcementlearning.planning.IPlanning;

/**
 * 
 * Measure used in batch protocole
 * 
 * @author Thomas Degris
 * 
 */
public abstract class BatchMeasure extends Measure {
	/**
	 * 
	 * Flag to decide weither or not an agent is loaded or the forest only
	 * 
	 */
	private boolean requireSolution = false;

	/**
	 * 
	 * List of algorithm to run
	 * 
	 */
	protected List<IAgentFactory> agentsFactory = new ArrayList<IAgentFactory>();

	/**
	 * 
	 * The measures associated to the experiment
	 * 
	 */
	protected List<Measure> measures = new ArrayList<Measure>();

	/**
	 * 
	 * Mapping from an agent to its measure writer
	 * 
	 */
	protected Map<IAgentFactory, MeasureWriter> agentWriters = null;

	/**
	 * 
	 * Trajectory used to learn
	 * 
	 */
	protected Trajectory trajectory = null;

	/**
	 * 
	 * Constructor
	 * 
	 * 
	 */
	public BatchMeasure() {
		super();
	}

	/**
	 * 
	 * Add new algorithm to run during the experiment
	 * 
	 * @param agentFactory
	 *          the algorithm to run for each pool
	 * 
	 */
	public void addAgentFactory(IAgentFactory agentFactory) {
		agentsFactory.add(agentFactory);
		requireSolution = requireSolution || agentFactory.requireSolution();
	}

	/**
	 * 
	 * Add a new measure to the experiment
	 * 
	 * @param measure
	 *          the measure to add
	 * 
	 */
	public void addMeasure(Measure measure) {
		measures.add(measure);
		requireSolution = requireSolution || measure.requireSolution();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.IStepListener#listenStep(int,
	 *      Context, java.lang.String, Context, java.util.List)
	 */
	@Override
	public void listenStep(int time, Context currentState, String actionChoosen,
			Context nextState, List<Double> lastReward) {
		trajectory.addObs(currentState, actionChoosen, lastReward, nextState);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.stats.Measure#setMeasureData(animatlab.kodamat.core.IProblemDefinition,
	 *      animatlab.kodamat.reinforcementlearning.planning.IPlanning,
	 *      animatlab.kodamat.reinforcementlearning.agent.IAgent,
	 *      java.lang.String)
	 */
	@Override
	public void setMeasureData(IProblemDefinition problem, IPlanning solution,
			IAgent agent, String resultsFolder) {
		super.setMeasureData(problem, solution, agent, resultsFolder);
		if (problem != null) {
			trajectory = new Trajectory(problem.getVariableOrdering(), problem
					.getActionList());
		}
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.stats.Measure#reset()
	 */
	@Override
	public void reset() {
		agentWriters = new LinkedHashMap<IAgentFactory, MeasureWriter>();
		for (IAgentFactory agentFactory : agentsFactory) {
			String agentName;
			if (agentFactory instanceof ModelBasedAgentFactory) {
				agentName = ((ModelBasedAgentFactory) agentFactory).getModel()
						.getName();
			} else {
				agentName = agentFactory.getAgentName();
			}
			agentWriters.put(agentFactory, new MeasureWriter(measures, resultsFolder
					+ "." + agentName));
		}
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.stats.Measure#requireSolution()
	 */
	@Override
	public boolean requireSolution() {
		return requireSolution;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.IStepListener#terminate()
	 */
	@Override
	public void terminate() {
		for (MeasureWriter writer : agentWriters.values()) {
			writer.terminate();
		}
		agentWriters = null;
	}
}
