package animatlab.kodamat.reinforcementlearning.planning.stats;

import static animatlab.kodamat.reinforcementlearning.agent.ModelBasedAgent.TIME_POINTS_LEARNING;
import static animatlab.kodamat.reinforcementlearning.agent.ModelBasedAgent.TIME_POINTS_PLANNING;

import java.util.Map;
import java.util.Random;

import animatlab.kodamat.core.IProblemDefinition;
import animatlab.kodamat.core.utils.Chrono;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.problems.ProblemRunner;
import animatlab.kodamat.reinforcementlearning.agent.IAgent;
import animatlab.kodamat.reinforcementlearning.agent.IAgentFactory;
import animatlab.kodamat.reinforcementlearning.agent.ModelBasedAgent;
import animatlab.kodamat.reinforcementlearning.core.acting.Greedy;
import animatlab.kodamat.reinforcementlearning.core.acting.IActing;
import animatlab.kodamat.reinforcementlearning.core.stats.Measure;
import animatlab.kodamat.reinforcementlearning.core.stats.MeasureWriter;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractLearningForest;
import animatlab.kodamat.reinforcementlearning.planning.IPlanning;
import animatlab.kodamat.reinforcementlearning.planning.PlanningUtils;

/**
 * 
 * Measure used to compute the quality of policy from a trajectory
 * 
 * @author Thomas Degris
 * 
 */
public class BatchPlanningMeasure extends BatchMeasure {
	/**
	 * 
	 * Time used to test the agent
	 * 
	 */
	private final int testingTime = 2000;

	/**
	 * 
	 * Random instance used to generate trajectory to evaluate agent
	 * 
	 */
	private Random random = new Random(0);

	/**
	 * 
	 * Constructor
	 * 
	 * 
	 */
	public BatchPlanningMeasure() {
		super();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.stats.Measure#computeMeasures(java.util.Map)
	 */
	@Override
	public void computeMeasures(Map<String, Double> data) {
		if ((agent instanceof ModelBasedAgent) && (agentsFactory.size() == 0)) {
			computeMeasureForRegisteredAgent(data);
			return;
		}
		for (IAgentFactory agentFactory : agentsFactory) {
			IAgent thisAgent = agentFactory.createAgent(random, problem, solution);
			MeasureWriter measureWriter = agentWriters.get(agentFactory);
			if (thisAgent instanceof ModelBasedAgent) {
				Chrono chrono = new Chrono();
				thisAgent = planFromTrajectory((ModelBasedAgent) thisAgent, chrono);
				if (chrono.getTimePoint(TIME_POINTS_LEARNING) != null) {
					measureWriter.addToNextStep("tl", chrono
							.getTimePoint(TIME_POINTS_LEARNING));
				}
				if (chrono.getTimePoint(TIME_POINTS_PLANNING) != null) {
					measureWriter.addToNextStep("tp", chrono
							.getTimePoint(TIME_POINTS_PLANNING));
				}
			}
			if (thisAgent != null) {
				runAgentInProblem(thisAgent);
				measureWriter.computeMeasures("time", trajectory.getTime());
			}
		}
	}

	/**
	 * 
	 * Compute the measures for the registered agent
	 * 
	 * @param data
	 *          the data where to put the results of the measure
	 * 
	 */
	private void computeMeasureForRegisteredAgent(Map<String, Double> data) {
		IAgent thisAgent = computePlanning((ModelBasedAgent) agent, null);
		runAgentInProblem(thisAgent);
		for (Measure measure : measures) {
			measure.computeMeasures(data);
		}
	}

	/**
	 * 
	 * Create a new agent executing the policy computed from a model learn from a
	 * trajectory
	 * 
	 * @param agent
	 *          the agent owning the learning method
	 * @param chrono
	 *          instance used to measure computation time
	 * 
	 * @return the new agent
	 * 
	 */
	private IAgent planFromTrajectory(ModelBasedAgent agent, Chrono chrono) {
		AbstractLearningForest learningModel = (AbstractLearningForest) agent
				.getModel();
		chrono.start();
		learningModel.learnFromTrajectory(trajectory, null);
		chrono.addTimePoint(TIME_POINTS_LEARNING);
		if (learningModel.getActionTestedSet().size() < problem.getActionList()
				.size()) {
			return null;
		}
		return computePlanning(agent, chrono);
	}

	/**
	 * 
	 * Compute an optimal policy from a model
	 * 
	 * @param agent
	 *          the agent for which to compute the optimal policy
	 * @param chrono
	 *          chrono used to measure times
	 * @return a new agent
	 * 
	 */
	private IAgent computePlanning(ModelBasedAgent agent, Chrono chrono) {
		AbstractLearningForest learningModel = (AbstractLearningForest) agent
				.getModel();
		IPlanning planning = agent.getPlanning();
		if (planning == null) {
			String planningMethod = learningModel.getParameters().getKeyS(
					Parameters.PARAM_KEY_PLANNING_METHOD);
			planning = PlanningUtils.getPlanningMethod(learningModel, planningMethod);
			if (planning == null) {
				return new ModelBasedAgent(learningModel, null);
			}
			planning.setLog(log);
			if (chrono != null) {
				chrono.start();
			}
			planning.solve();
			if (chrono != null) {
				chrono.addTimePoint(TIME_POINTS_PLANNING);
			}
		}
		return new ModelBasedAgent(learningModel, new Greedy(random, planning));
	}

	/**
	 * 
	 * Compute the set of measures for an agent
	 * 
	 * @param thisAgent
	 *          the agent to measure
	 * 
	 */
	private void runAgentInProblem(IAgent thisAgent) {
		IAgent agentInProblem = null;
		if (thisAgent instanceof ModelBasedAgent) {
			final IActing acting = ((ModelBasedAgent) thisAgent).getActing();
			if (acting != null) {
				agentInProblem = (IAgent) acting;
			}
		} else {
			agentInProblem = thisAgent;
		}
		for (Measure measure : measures) {
			measure.setMeasureData(problem, solution, thisAgent, resultsFolder);
		}
		if (agentInProblem == null) {
			return;
		}
		ProblemRunner problemRunner = new ProblemRunner(random, problem,
				agentInProblem, log);
		for (Measure measure : measures) {
			problemRunner.registerListener(measure);
		}
		problemRunner.runProblem(testingTime);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.stats.Measure#setMeasureData(animatlab.kodamat.core.IProblemDefinition,
	 *      animatlab.kodamat.reinforcementlearning.planning.IPlanning,
	 *      animatlab.kodamat.reinforcementlearning.agent.IAgent,
	 *      java.lang.String)
	 */
	@Override
	public void setMeasureData(IProblemDefinition problem, IPlanning solution,
			IAgent agent, String resultsFolder) {
		super.setMeasureData(problem, solution, agent, resultsFolder);
		reset();
	}
}
