/*
 * 
 * Created on 15 nov. 2005
 *
 */
package animatlab.kodamat.reinforcementlearning.planning.stats;

import static animatlab.kodamat.core.proba.Chi2Utils.chi2Utils;

import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.IProblemDefinition;
import animatlab.kodamat.core.proba.ILearningProbabilities;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.ComparePolicy;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.reinforcementlearning.agent.IAgent;
import animatlab.kodamat.reinforcementlearning.agent.ModelBasedAgent;
import animatlab.kodamat.reinforcementlearning.core.stats.Measure;
import animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractLearningForest;
import animatlab.kodamat.reinforcementlearning.planning.IPlanning;

/**
 * 
 * Measure logging the number of couple state/action tested by the agent
 * 
 * @author Thomas Degris
 * 
 */
public class Chi2ModelMeasure extends Measure {

	/**
	 * 
	 * Policy used to run all the Chi2 policies for each node in the tree
	 * 
	 */
	private static class Chi2ComparaisonPolicy implements
			ComparePolicy<IProbabilities<String>> {
		/**
		 * 
		 * List of Chi2 policy to run at each node
		 * 
		 */
		List<ChiStatisticPolicy> chiStatisticPolicyList;

		/**
		 * 
		 * Constructor
		 * 
		 * @param chiStatisticPolicyList
		 *          list of Chi2 policy to run at each node
		 * 
		 */
		Chi2ComparaisonPolicy(List<ChiStatisticPolicy> chiStatisticPolicyList) {
			this.chiStatisticPolicyList = chiStatisticPolicyList;
		}

		/**
		 * 
		 * Compare a node with another node
		 * 
		 * @param node01
		 *          the first node to compare
		 * @param node02
		 *          the second node to compare
		 * 
		 */
		public void compare(Leaf<IProbabilities<String>> node01,
				Leaf<IProbabilities<String>> node02) {
			for (ChiStatisticPolicy policy : chiStatisticPolicyList) {
				Leaf<IProbabilities<String>> referenceNode = node01;
				Leaf<IProbabilities<String>> learningNode = node02;
				policy.addNewStat(referenceNode, learningNode, chi2Utils
						.computeAlphaValue((ILearningProbabilities<String>) learningNode
								.get(), referenceNode.get()));
			}
		}

	}

	/**
	 * 
	 * Policy to measure the Chi2 between two nodes
	 * 
	 */
	private static abstract class ChiStatisticPolicy {
		/**
		 * 
		 * Current computed Chi2 value
		 * 
		 */
		protected double currentChiValue;

		/**
		 * 
		 * Model used as reference for the measure
		 * 
		 */
		final protected AbstractFactoredModel modelReference;

		/**
		 * 
		 * Constructor
		 * 
		 * @param modelReference
		 *          model used as reference for the measure
		 * 
		 */
		ChiStatisticPolicy(AbstractFactoredModel modelReference) {
			this.modelReference = modelReference;
		}

		/**
		 * 
		 * Reset the Chi value
		 * 
		 */
		protected void reset() {
			currentChiValue = 0.0;
		}

		/**
		 * 
		 * @return the computed Chi value
		 * 
		 */
		final protected double getChiValue() {
			return currentChiValue;
		}

		/**
		 * 
		 * Compute a new stat for the Chi2 measure
		 * 
		 * @param referenceNode
		 *          the probabilities used as reference
		 * @param learningNode
		 *          the probabilities learned until now
		 * @param alpha
		 *          alpha value compute for the two leaves
		 * 
		 */
		abstract protected void addNewStat(
				Leaf<IProbabilities<String>> referenceNode,
				Leaf<IProbabilities<String>> learningNode, double alpha);

		/**
		 * 
		 * @return the name of this Chi2 measure
		 * 
		 */
		abstract protected String getName();

		/**
		 * 
		 * @return the computed Chi2 value for this policy
		 * 
		 */
		public double getAverage() {
			assert (!Double.isNaN(currentChiValue));
			assert (Math.abs(currentChiValue) != Double.MAX_VALUE);
			return currentChiValue;
		}
	}

	/**
	 * 
	 * Measure the Chi2 min between two nodes
	 * 
	 */
	private static class MinChiStatisticPolicy extends ChiStatisticPolicy {
		/**
		 * 
		 * Constructor
		 * 
		 * @param modelReference
		 *          model used as reference for the measure
		 * 
		 */
		MinChiStatisticPolicy(AbstractFactoredModel modelReference) {
			super(modelReference);
		}

		/**
		 * @see animatlab.kodamat.reinforcementlearning.planning.stats.Chi2ModelMeasure.ChiStatisticPolicy#reset()
		 */
		@Override
		protected void reset() {
			currentChiValue = Double.MAX_VALUE;
		}

		/**
		 * @see animatlab.kodamat.reinforcementlearning.planning.stats.Chi2ModelMeasure.ChiStatisticPolicy#addNewStat(Leaf,
		 *      Leaf, double)
		 */
		@Override
		protected void addNewStat(Leaf<IProbabilities<String>> referenceNode,
				Leaf<IProbabilities<String>> learningNode, double alpha) {
			currentChiValue = Math.min(currentChiValue, alpha);
		}

		/**
		 * @see animatlab.kodamat.reinforcementlearning.planning.stats.Chi2ModelMeasure.ChiStatisticPolicy#getName()
		 */
		@Override
		protected String getName() {
			return "chimin";
		}
	}

	/**
	 * 
	 * Measure the Chi2 max between two nodes
	 * 
	 */
	private static class MaxChiStatisticPolicy extends ChiStatisticPolicy {
		/**
		 * 
		 * Constructor
		 * 
		 * @param modelReference
		 *          model used as reference for the measure
		 * 
		 */
		MaxChiStatisticPolicy(AbstractFactoredModel modelReference) {
			super(modelReference);
		}

		/**
		 * @see animatlab.kodamat.reinforcementlearning.planning.stats.Chi2ModelMeasure.ChiStatisticPolicy#reset()
		 */
		@Override
		protected void reset() {
			currentChiValue = -Double.MAX_VALUE;
		}

		/**
		 * @see animatlab.kodamat.reinforcementlearning.planning.stats.Chi2ModelMeasure.ChiStatisticPolicy#addNewStat(Leaf,
		 *      Leaf, double)
		 */
		@Override
		protected void addNewStat(Leaf<IProbabilities<String>> referenceNode,
				Leaf<IProbabilities<String>> learningNode, double alpha) {
			currentChiValue = Math.max(currentChiValue, alpha);
		}

		/**
		 * @see animatlab.kodamat.reinforcementlearning.planning.stats.Chi2ModelMeasure.ChiStatisticPolicy#getName()
		 */
		@Override
		protected String getName() {
			return "chimax";
		}
	}

	/**
	 * 
	 * Measure the Chi2 min between two nodes
	 * 
	 */
	private static class ExperimentedChiStatisticPolicy extends
			ChiStatisticPolicy {
		/**
		 * 
		 * Size of the state space
		 * 
		 */
		private int numberOfExamples;

		/**
		 * 
		 * Constructor
		 * 
		 * @param modelReference
		 *          model used as reference for the measure
		 * 
		 */
		ExperimentedChiStatisticPolicy(AbstractFactoredModel modelReference) {
			super(modelReference);
		}

		/**
		 * @see animatlab.kodamat.reinforcementlearning.planning.stats.Chi2ModelMeasure.ChiStatisticPolicy#reset()
		 */
		@Override
		protected void reset() {
			super.reset();
			numberOfExamples = 0;
		}

		/**
		 * @see animatlab.kodamat.reinforcementlearning.planning.stats.Chi2ModelMeasure.ChiStatisticPolicy#addNewStat(Leaf,
		 *      Leaf, double)
		 */
		@Override
		protected void addNewStat(Leaf<IProbabilities<String>> referenceNode,
				Leaf<IProbabilities<String>> learningNode, double alpha) {
			if (learningNode.get() != null) {
				int leafWeight = ((ILearningProbabilities<String>) learningNode.get())
						.getNumberOfExamples();
				currentChiValue = currentChiValue + leafWeight * alpha;
				numberOfExamples = numberOfExamples + leafWeight;
			}
		}

		/**
		 * @see animatlab.kodamat.reinforcementlearning.planning.stats.Chi2ModelMeasure.ChiStatisticPolicy#getName()
		 */
		@Override
		protected String getName() {
			return "chiexp";
		}

		/**
		 * @see animatlab.kodamat.reinforcementlearning.planning.stats.Chi2ModelMeasure.ChiStatisticPolicy#getAverage()
		 */
		@Override
		public double getAverage() {
			return currentChiValue / numberOfExamples;
		}

	}

	/**
	 * 
	 * Measure the Chi2 min between two nodes
	 * 
	 */
	private static class FullModelChiStatisticPolicy extends ChiStatisticPolicy {
		/**
		 * 
		 * Size of the state space
		 * 
		 */
		private long stateSpaceSize;

		/**
		 * 
		 * Constructor
		 * 
		 * @param modelReference
		 *          model used as reference for the measure
		 * 
		 */
		FullModelChiStatisticPolicy(AbstractFactoredModel modelReference) {
			super(modelReference);
		}

		/**
		 * @see animatlab.kodamat.reinforcementlearning.planning.stats.Chi2ModelMeasure.ChiStatisticPolicy#reset()
		 */
		@Override
		protected void reset() {
			super.reset();
			stateSpaceSize = 0;
		}

		/**
		 * @see animatlab.kodamat.reinforcementlearning.planning.stats.Chi2ModelMeasure.ChiStatisticPolicy#addNewStat(Leaf,
		 *      Leaf, double)
		 */
		@Override
		protected void addNewStat(Leaf<IProbabilities<String>> referenceNode,
				Leaf<IProbabilities<String>> learningNode, double alpha) {
			Set<Variable> testedVariableSet = new LinkedHashSet<Variable>();
			testedVariableSet.addAll(referenceNode.getContext().keySet());
			testedVariableSet.addAll(learningNode.getContext().keySet());

			long leafWeight = Context.getRestrictedStateSize(modelReference
					.getValueSets(), testedVariableSet);
			currentChiValue = currentChiValue + leafWeight * alpha;
			stateSpaceSize = stateSpaceSize + leafWeight;
		}

		/**
		 * @see animatlab.kodamat.reinforcementlearning.planning.stats.Chi2ModelMeasure.ChiStatisticPolicy#getName()
		 */
		@Override
		protected String getName() {
			return NAME_CHITHEO;
		}

		/**
		 * @see animatlab.kodamat.reinforcementlearning.planning.stats.Chi2ModelMeasure.ChiStatisticPolicy#getAverage()
		 */
		@Override
		public double getAverage() {
			// We disable this assert because we still want to compute
			// the value when the agent did not explore all the independant
			// partition of its environment
			// assert((stateSpaceSize == 0) || (stateSpaceSize ==
			// (modelReference.getNumberOfStates() *
			// modelReference.getActions().size() * modelReference.getStateSize())));
			return currentChiValue / stateSpaceSize;
		}
	}

	/**
	 * 
	 * Name for the theoric accuracy
	 * 
	 */
	public static final String NAME_CHITHEO = "chitheo";

	/**
	 * 
	 * The list of all the Chi2 policy used to compute the measure
	 * 
	 */
	private List<ChiStatisticPolicy> chiStatisticPolicyList = null;

	/**
	 * 
	 * Constructor
	 * 
	 * 
	 */
	public Chi2ModelMeasure() {
		super();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.stats.Measure#computeMeasures(Map)
	 */
	@Override
	public void computeMeasures(Map<String, Double> data) {
		if (!(agent instanceof ModelBasedAgent)) {
			return;
		}
		AbstractLearningForest model = (AbstractLearningForest) ((ModelBasedAgent) agent)
				.getModel();
		if ((model == null)
				|| (model.getActionTestedSet().size() < problem.getActionList().size())) {
			return;
		}
		computeAccuracy((AbstractFactoredModel) problem, model,
				chiStatisticPolicyList);
		for (ChiStatisticPolicy policy : chiStatisticPolicyList) {
			String key = policy.getName();
			Double value = policy.getAverage();
			data.put(key, value);
		}
	}

	/**
	 * Compute the accurancy of a learning forest compared to reference forest
	 * 
	 * @param referenceForest
	 *          the forest used as reference
	 * @param learningForest
	 *          the learning forest
	 * @param accuracyPolicyList
	 *          the list of accuracy policy to compute
	 * 
	 */
	static public void computeAccuracy(AbstractFactoredModel referenceForest,
			AbstractLearningForest learningForest,
			List<ChiStatisticPolicy> accuracyPolicyList) {
		for (ChiStatisticPolicy policy : accuracyPolicyList) {
			policy.reset();
		}
		Chi2ComparaisonPolicy comparePolicy = new Chi2ComparaisonPolicy(
				accuracyPolicyList);
		for (Variable variable : referenceForest.getVariableSet()) {
			for (String action : learningForest.getActionTestedSet()) {
				Tree<IProbabilities<String>> referenceTree = (Tree<IProbabilities<String>>) referenceForest
						.getTransitionFunction(variable, action);
				Tree<IProbabilities<String>> learningTree = learningForest
						.getTransitionFunction(variable, action);
				referenceTree.compare(learningTree, comparePolicy);
			}
		}
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.stats.Measure#setMeasureData(animatlab.kodamat.core.IProblemDefinition,
	 *      animatlab.kodamat.reinforcementlearning.planning.IPlanning,
	 *      animatlab.kodamat.reinforcementlearning.agent.IAgent,
	 *      java.lang.String)
	 */
	@Override
	public void setMeasureData(IProblemDefinition problem, IPlanning solution,
			IAgent agent, String resultsFolder) {
		super.setMeasureData(problem, solution, agent, resultsFolder);
		chiStatisticPolicyList = new ArrayList<ChiStatisticPolicy>();
		chiStatisticPolicyList.add(new MinChiStatisticPolicy(
				(AbstractFactoredModel) problem));
		chiStatisticPolicyList.add(new MaxChiStatisticPolicy(
				(AbstractFactoredModel) problem));
		chiStatisticPolicyList.add(new ExperimentedChiStatisticPolicy(
				(AbstractFactoredModel) problem));
		chiStatisticPolicyList.add(new FullModelChiStatisticPolicy(
				(AbstractFactoredModel) problem));
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.stats.Measure#reset()
	 */
	@Override
	public void reset() {
		// Nothing to do by default
	}

}
