package animatlab.kodamat.reinforcementlearning.planning.stats;

import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.util.LinkedHashMap;
import java.util.Map;

import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.reinforcementlearning.agent.IAgentFactory;
import animatlab.kodamat.reinforcementlearning.agent.ModelBasedAgent;
import animatlab.kodamat.reinforcementlearning.agent.ModelBasedAgentFactory;
import animatlab.kodamat.reinforcementlearning.core.stats.Measure;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractLearningForest;
import animatlab.kodamat.supervisedlearning.entropy.VariableMeasure;
import animatlab.kodamat.supervisedlearning.learningtree.LearningTree;

/**
 * 
 * Measure used for batch learning and different Chi2
 * 
 * @author Thomas Degris
 * 
 */
public class OverChi2Measure extends BatchMeasure {
	/**
	 * 
	 * Mapping from an agent to its measure writer
	 * 
	 */
	protected Map<IAgentFactory, PrintStream> resultsWriters = null;

	/**
	 * 
	 * Set of Chi2 value to measure
	 * 
	 */
	private double[] chi2MeasurePoints = null;

	/**
	 * 
	 * Constructor
	 * 
	 */
	public OverChi2Measure() {
		super();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.stats.Measure#computeMeasures(java.util.Map)
	 */
	@Override
	public void computeMeasures(Map<String, Double> data) {
		for (IAgentFactory agentFactory : agentsFactory) {
			ModelBasedAgentFactory learningAgentFactory = (ModelBasedAgentFactory) agentFactory;
			PrintStream measureWriter = resultsWriters.get(learningAgentFactory);
			for (double chiValue : chi2MeasurePoints) {
				measureWriter.print(trajectory.getTime() + " " + chiValue + " ");
				LearningTree<String, IProbabilities<String>> prototype = ((AbstractLearningForest) learningAgentFactory
						.getModel()).getInductionPrototype();
				VariableMeasure<String> variableMeasure = prototype.getMeasure();
				variableMeasure.setThreshold(chiValue);

				ModelBasedAgent learningAgent = learningAgentFactory.createAgent(null,
						problem, null);
				AbstractLearningForest model = (AbstractLearningForest) learningAgent
						.getModel();
				model.learnFromTrajectory(trajectory, null);
				for (Measure measure : measures) {
					measure.setMeasureData(problem, null, learningAgent, null);
					measure.reset();
				}
				LinkedHashMap<String, Double> localData = new LinkedHashMap<String, Double>();
				for (Measure measure : measures) {
					measure.computeMeasures(localData);
					for (Double value : localData.values()) {
						measureWriter.print(value + " ");
					}
					measureWriter.flush();
					measureWriter.print("\t");
				}
				measureWriter.println();
			}
			measureWriter.println();
		}
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.stats.Measure#reset()
	 */
	@Override
	public void reset() {
		resultsWriters = new LinkedHashMap<IAgentFactory, PrintStream>();
		for (IAgentFactory agentFactory : agentsFactory) {
			String agentName = ((ModelBasedAgentFactory) agentFactory).getModel()
					.getName();
			PrintStream resultsStream = null;
			try {
				resultsStream = new PrintStream(resultsFolder + "/" + agentName
						+ ".data");
			} catch (FileNotFoundException e) {
				e.printStackTrace();
				System.exit(1);
			}
			resultsWriters.put(agentFactory, resultsStream);
		}
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.IStepListener#terminate()
	 */
	@Override
	public void terminate() {
		for (PrintStream writer : resultsWriters.values()) {
			writer.flush();
			writer.close();
		}
		resultsWriters = null;
	}

	/**
	 * 
	 * Set the set of Chi2 value to measure
	 * 
	 * @param chi2MeasurePoints
	 *          the set of Chi2 value to measure
	 * 
	 */
	public void setChi2MeasurePoints(double[] chi2MeasurePoints) {
		this.chi2MeasurePoints = chi2MeasurePoints;
	}

}
