/*
 * 
 * Created on 15 nov. 2005
 *
 */
package animatlab.kodamat.reinforcementlearning.planning.stats;

import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.IProblemDefinition;
import animatlab.kodamat.core.trees.ComparePolicy;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.reinforcementlearning.agent.ModelBasedAgent;
import animatlab.kodamat.reinforcementlearning.core.stats.Measure;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractForest;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.SpiSvi;

/**
 * 
 * Measure comparing the value function of the agent with the optimal value
 * function
 * 
 * @author Thomas Degris
 * 
 */
public class RelativeErrorMeasure extends Measure {

	/**
	 * 
	 * Label used for the average
	 * 
	 */
	public static final String VALFUNAVE_LABEL = "valfunave";

	/**
	 * 
	 * Policy used to compare two value functions
	 * 
	 * @author Thomas Degris
	 * 
	 */
	public abstract static class ValueComparePolicy implements
			ComparePolicy<Double> {
		/**
		 * 
		 * The value computed by the measure
		 * 
		 */
		protected double diff = 0.0;

		/**
		 * 
		 * Compare two values together
		 * 
		 * @param referencedValue
		 *          the value of the value function used as reference
		 * @param learningValue
		 *          the value of the value function to measure
		 * 
		 * @return the result of the comparaison
		 * 
		 */
		protected double compare(double referencedValue, double learningValue) {
			if (referencedValue < learningValue) {
				referencedValue += Parameters.EPSILON_THRESHOLD_CONVERGENCE * 10;
			}
			// It is not possible that the value function of the evaluated
			// policy has a better value than the optimal policy
			assert (referencedValue >= learningValue);
			if (referencedValue == 0) {
				return 0.0;
			}

			return ((referencedValue - learningValue) / referencedValue);
		}

		/**
		 * 
		 * @return the value computed by the measure
		 * 
		 */
		public double getValue() {
			return diff;
		}
	}

	/**
	 * 
	 * Policy used to extract the max between two values functions
	 * 
	 */
	public static class MinDiffComparePolicy extends ValueComparePolicy {
		/**
		 * 
		 * Constructor
		 * 
		 */
		MinDiffComparePolicy() {
			diff = Double.MAX_VALUE;
		}

		/**
		 * 
		 * Compare a node with another node
		 * 
		 * @param node01
		 *          the first node to compare
		 * @param node02
		 *          the second node to compare
		 * 
		 */
		public void compare(Leaf<Double> node01, Leaf<Double> node02) {
			diff = Math.min(diff, compare(node01.get(), node02.get()));
		}
	}

	/**
	 * 
	 * Policy used to extract the min between two values functions
	 * 
	 */
	public static class MaxDiffComparePolicy extends ValueComparePolicy {
		/**
		 * 
		 * Constructor
		 * 
		 */
		public MaxDiffComparePolicy() {
			diff = -Double.MAX_VALUE;
		}

		/**
		 * 
		 * Compare a node with another node
		 * 
		 * @param node01
		 *          the first node to compare
		 * @param node02
		 *          the second node to compare
		 * 
		 */
		public void compare(Leaf<Double> node01, Leaf<Double> node02) {
			diff = Math.max(diff, compare(node01.get(), node02.get()));
			assert (!Double.isNaN(diff));
		}
	}

	/**
	 * 
	 * Comparaison policy used to compute the average deviation with the optimal
	 * policy
	 * 
	 */
	static class AverageDeviationComparaison extends ValueComparePolicy {
		/**
		 * 
		 * Sum of the weight applied on each deviation for each node
		 * 
		 */
		private long averageDeviationWeightSum = 0;

		/**
		 * 
		 * Agent used as reference
		 * 
		 */
		final private SpiSvi reference;

		/**
		 * 
		 * Constructor
		 * 
		 * @param reference
		 *          agent used as reference
		 * 
		 */
		AverageDeviationComparaison(SpiSvi reference) {
			this.reference = reference;
		}

		/**
		 * 
		 * Compare a node with another node
		 * 
		 * @param node01
		 *          the first node to compare
		 * @param node02
		 *          the second node to compare
		 * 
		 */
		public void compare(Leaf<Double> node01, Leaf<Double> node02) {
			Leaf<Double> optimalFunctionNode = node01;
			Leaf<Double> evaluatedFunctionNode = node02;

			Set<Variable> assignedVariableSet = new LinkedHashSet<Variable>();
			assignedVariableSet.addAll(optimalFunctionNode.getContext().keySet());
			assignedVariableSet.addAll(evaluatedFunctionNode.getContext().keySet());
			long weight = Context.getRestrictedStateSize(reference.getModel()
					.getValueSets(), assignedVariableSet);
			assert (weight > 0);
			averageDeviationWeightSum = averageDeviationWeightSum + weight;
			assert (averageDeviationWeightSum >= 0);

			double deviation = compare(optimalFunctionNode.get(),
					evaluatedFunctionNode.get());
			diff = diff + weight * deviation;

			assert (!Double.isNaN(diff));
		}

		/**
		 * @see animatlab.kodamat.reinforcementlearning.planning.stats.RelativeErrorMeasure.ValueComparePolicy#getValue()
		 */
		@Override
		public double getValue() {
			assert (averageDeviationWeightSum == ((IProblemDefinition) reference
					.getModel()).getNumberOfStates());
			return diff / averageDeviationWeightSum;
		}

	}

	/**
	 * 
	 * Data structure used to return the current evaluation of the greedy policy
	 * of an agent
	 * 
	 */
	static public class ValueResults {
		/**
		 * 
		 * Minimal distance bewteen the two value functions
		 * 
		 */
		public double min;

		/**
		 * 
		 * Maximum distance bewteen the two value functions
		 * 
		 */
		public double max;

		/**
		 * 
		 * Average distance bewteen the two value functions
		 * 
		 */
		public double average;

		/**
		 * @see java.lang.Object#toString()
		 */
		@Override
		public String toString() {
			return "(" + min + "," + average + "," + max + ")";
		}
	}

	/**
	 * 
	 * Constructor
	 * 
	 * 
	 */
	public RelativeErrorMeasure() {
		super();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.stats.Measure#computeMeasures(Map)
	 */
	@Override
	public void computeMeasures(Map<String, Double> data) {
		if (solution == null) {
			return;
		}
		SpiSvi spisvi = (SpiSvi) ((ModelBasedAgent) agent).getPlanning();
		Tree<Set<String>> currentAgentPolicy = null;
		if (spisvi != null) {
			currentAgentPolicy = spisvi.getOptimalPolicy();
		} else {
			spisvi = new SpiSvi((AbstractForest) ((ModelBasedAgent) agent).getModel());
			currentAgentPolicy = spisvi.getSVI();
		}
		ValueResults valueResults = evaluateSpiSviPolicy((SpiSvi) solution,
				currentAgentPolicy);
		data.put("valfunmax", valueResults.max);
		data.put("valfunmin", valueResults.min);
		data.put(VALFUNAVE_LABEL, valueResults.average);
	}

	/**
	 * 
	 * Evaluate a SPI/SVI policy compared to the optimal policy
	 * 
	 * @param reference
	 *          the reference to get the model and the optimal value function
	 * @param currentAgentPolicy
	 *          the policy to measure
	 * 
	 * @return the result of the evaluation
	 * 
	 */
	public static ValueResults evaluateSpiSviPolicy(SpiSvi reference,
			Tree<Set<String>> currentAgentPolicy) {
		if (currentAgentPolicy == null) {
			return null;
		}
		Tree<Double> currentPolicyValueFunction = reference
				.getSSA(currentAgentPolicy);
		return computeValues(reference, currentPolicyValueFunction);
	}

	/**
	 * 
	 * Compute the different values to estimate the greedy policy of a learning
	 * agent
	 * 
	 * @param reference
	 *          agent used as reference
	 * @param evaluatedValueFunction
	 *          the value function of the policy to evaluate
	 * @return the set of results
	 * 
	 */
	public static ValueResults computeValues(SpiSvi reference,
			Tree<Double> evaluatedValueFunction) {
		if (reference.getOptimalValueFunction() == null) {
			return null;
		}
		Tree<Double> optimalValueFunction = reference.getOptimalValueFunction();
		if (optimalValueFunction == null) {
			return null;
		}
		ValueResults valueResults = new ValueResults();
		valueResults.max = computeValueMeasure(evaluatedValueFunction,
				optimalValueFunction, new MaxDiffComparePolicy());
		valueResults.min = computeValueMeasure(evaluatedValueFunction,
				optimalValueFunction, new MinDiffComparePolicy());
		valueResults.average = computeValueMeasure(evaluatedValueFunction,
				optimalValueFunction, new AverageDeviationComparaison(reference));
		return valueResults;
	}

	/**
	 * 
	 * Compute the value measure for a given policy
	 * 
	 * @param evaluatedValueFunction
	 *          the value function modified to compute the difference
	 * @param optimalValueFunction
	 *          the value function used as reference (not modified)
	 * @param comparaisonPolicy
	 *          the policy to use to compute the measure
	 * 
	 * @return the measure computed
	 * 
	 */
	public static double computeValueMeasure(Tree<Double> evaluatedValueFunction,
			Tree<Double> optimalValueFunction, ValueComparePolicy comparaisonPolicy) {
		optimalValueFunction.compare(evaluatedValueFunction, comparaisonPolicy);
		return comparaisonPolicy.getValue();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.stats.Measure#reset()
	 */
	@Override
	public void reset() {
		// Nothing to reset
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.stats.Measure#requireSolution()
	 */
	@Override
	public boolean requireSolution() {
		return true;
	}

}
