/*
 * 
 * Created on Nov 1, 2005
 *
 */
package animatlab.kodamat.reinforcementlearning.planning.stats.tests;

import static animatlab.kodamat.core.io.XmlUtils.getRootFromFile;

import java.io.File;
import java.util.Set;

import junit.framework.TestCase;
import animatlab.kodamat.core.io.XmlUtils;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.problems.generated.syntheticmdps.LinearProblem;
import animatlab.kodamat.reinforcementlearning.model.forest.StaticForest;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.StaticSpiSvi;
import animatlab.kodamat.reinforcementlearning.planning.stats.RelativeErrorMeasure;
import animatlab.kodamat.reinforcementlearning.planning.stats.RelativeErrorMeasure.ValueResults;

/**
 * 
 * JUnit to test the calculation of the accuracy for the
 * LearningForestStatistics
 * 
 * @author Thomas Degris
 * 
 */
public class TestLearningForestStatistics extends TestCase {
	/**
	 * 
	 * World model used in the experiment
	 * 
	 */
	protected StaticSpiSvi staticSpiSvi = null;

	/**
	 * 
	 * File to load to run the test
	 * 
	 */
	final static String PROBLEM_PATH = "data/forests/coffeerobot";

	/**
	 * 
	 * Constant used to build dumbed value function
	 * 
	 */
	final static double DUMB_VALUE_FUNCTION = -1.0;

	/**
	 * @see junit.framework.TestCase#setUp()
	 */
	@Override
	public void setUp() {
		String agentPath = PROBLEM_PATH + StaticSpiSvi.AGENT_EXTENSION;
		if (!new File(agentPath).canRead()) {
			String forestPath = PROBLEM_PATH + StaticForest.PROBLEM_EXTENSION;
			StaticSpiSvi.convertProblemToAgent(new StaticForest(XmlUtils
					.getRootFromFile(forestPath)), PROBLEM_PATH);
		}
		staticSpiSvi = new StaticSpiSvi(XmlUtils.getRootFromFile(PROBLEM_PATH
				+ StaticSpiSvi.AGENT_EXTENSION));
		assert (staticSpiSvi.getOptimalValueFunction() != null);
	}

	/**
	 * 
	 * Test the value function distance of a LearningForestStatistics
	 * 
	 */
	public void testFunctionValueDistanceCalculationWithNullDistance() {
		Tree<Double> optimalValueFunctionTree = staticSpiSvi
				.getOptimalValueFunction();
		ValueResults valueResults = RelativeErrorMeasure.computeValues(
				staticSpiSvi, optimalValueFunctionTree);
		assertEquals(0.0, valueResults.min);
		assertEquals(0.0, valueResults.average);
		assertEquals(0.0, valueResults.max);
	}

	/**
	 * 
	 * Test the value function distance of a LearningForestStatistics
	 * 
	 */
	public void testFunctionValueDistanceCalculation() {
		Tree<Double> singleValue = new Tree<Double>();
		((Leaf<Double>) singleValue.getRoot()).set(DUMB_VALUE_FUNCTION);

		ValueResults valueResults = RelativeErrorMeasure.computeValues(
				staticSpiSvi, singleValue);
		assert (valueResults.min < valueResults.average);
		assert (valueResults.average < valueResults.max);
	}

	/**
	 * 
	 * Test the value function distance between two identical policies
	 * 
	 */
	public void testFunctionValueDistanceForSamePolicy() {
		String xmlString01 = "<dtr:dnode variable='hrc'>"
				+ "  <dtr:dnode value='1' variable='l'>"
				+ "    <dtr:leaf action='delc' value='1'/>"
				+ "    <dtr:leaf action='move' value='0'/>" + "  </dtr:dnode>"
				+ "  <dtr:dnode value='0' variable='l'>"
				+ "    <dtr:leaf action='delc' value='1'/>"
				+ "    <dtr:leaf action='move' value='0'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		String xmlString02 = "<dtr:dnode variable='l'>"
				+ "  <dtr:leaf action='delc' value='1'/>"
				+ "  <dtr:leaf action='move' value='0'/>" + "</dtr:dnode>";
		Tree<Set<String>> policy01 = staticSpiSvi.loadPolicy(XmlUtils
				.getRoot(xmlString01));
		Tree<Set<String>> policy02 = staticSpiSvi.loadPolicy(XmlUtils
				.getRoot(xmlString02));
		ValueResults policy01Results = RelativeErrorMeasure.evaluateSpiSviPolicy(
				staticSpiSvi, policy01);
		ValueResults policy02Results = RelativeErrorMeasure.evaluateSpiSviPolicy(
				staticSpiSvi, policy02);
		assert ((policy01Results.average - policy02Results.average) < Parameters.EPSILON_THRESHOLD_CONVERGENCE);
	}

	/**
	 * 
	 * Test the measure on a determinist best case problem
	 * 
	 */
	public void testOnDeterministBestCaseProblem() {
		LinearProblem bestCaseProblem = new LinearProblem();
		bestCaseProblem.setProblemData(8);
		StaticSpiSvi refSpiSvi = new StaticSpiSvi(bestCaseProblem.toForest());
		refSpiSvi.getSPI();

		StaticSpiSvi spisvi = new StaticSpiSvi(
				getRootFromFile("data/forests/bestcase08.static.junit"));
		ValueResults valueResults = RelativeErrorMeasure.evaluateSpiSviPolicy(
				refSpiSvi, spisvi.getOptimalPolicy());
		assert (valueResults.average < 1.0);
	}
}
