package animatlab.kodamat.supervisedlearning.core;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.io.DotUtils;
import animatlab.kodamat.core.io.HtmlPage;
import animatlab.kodamat.core.io.IContainerToString;
import animatlab.kodamat.core.io.IToDot.IToDotPolicy;
import animatlab.kodamat.core.proba.ILearningProbabilities;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.trees.Vertex;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.supervisedlearning.entropy.VariableMeasure;
import animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics;
import animatlab.kodamat.supervisedlearning.entropy.VariableMeasure.VariableMeasureData;
import animatlab.kodamat.supervisedlearning.learningtree.ILearningVertex;
import animatlab.kodamat.supervisedlearning.learningtree.LearningDNode;
import animatlab.kodamat.supervisedlearning.learningtree.LearningLeaf;

/**
 * 
 * Utils class to output a learning tree to a dot string
 * 
 * @author Thomas Degris
 * @param <U>
 *          type of the data in the trees
 * 
 */
public class LearningToDotUtils<U> {
	/**
	 * 
	 * Policy used to output this tree into dot
	 * 
	 * @param <T>
	 *          type contained in the tree
	 * 
	 */
	private static abstract class ToDotPolicy<T> implements IToDotPolicy<T>,
			IContainerToString<Vertex<?>> {
		/**
		 * 
		 * Constructor
		 * 
		 */
		protected ToDotPolicy() {
			super();
		}

		/**
		 * Convert the data into a dot label
		 * 
		 * @param data
		 *          the data to convert (is different of null)
		 * @return a string representing the dot label
		 */
		public String toDotLabel(T data) {
			return "";
		}

		/**
		 * 
		 * Print into a dot string the probabilities computed from the stats
		 * 
		 * @param stats
		 * 
		 * @return a short dot string representing this data
		 * 
		 */
		protected String shortProbaToDotString(VariableValueStatistics<T> stats) {
			return stats.toString();
		}

		/**
		 * 
		 * Print into a dot string the probabilities computed from the stats
		 * 
		 * @param stats
		 * 
		 * @return a short dot string representing this data
		 * 
		 */
		protected String longProbaToDotString(VariableValueStatistics<T> stats) {
			StringBuilder result = new StringBuilder();
			result.append(stats.getNumberOfExamples() + " example(s)\\n");
			result.append(stats.toString());
			return result.toString();
		}

		/**
		 * Convert the container from a default string
		 * 
		 * @param container
		 *          the container to convert
		 * @param defaultString
		 *          the default string obtained from the container
		 * @return the string corresponding from the policy
		 */
		@SuppressWarnings("unchecked")
		public String containerToString(Vertex<?> container, String defaultString) {
			if (container.isLeaf) {
				return leafToDot((LearningLeaf<T, ?>) container, defaultString);
			}
			return dnodeToDot((LearningDNode<T, ?>) container, defaultString);
		}

		/**
		 * Convert a regression decision node to a string
		 * 
		 * @param dnode
		 *          the regression decision node to convert
		 * @param defaultString
		 *          the default string obtained from the container
		 * @return the result of the conversion
		 */
		abstract protected String dnodeToDot(LearningDNode<T, ?> dnode,
				String defaultString);

		/**
		 * Convert a regression leaf to a string
		 * 
		 * @param leaf
		 *          the regression leaf to convert
		 * @param defaultString
		 *          the default string obtained from the container
		 * @return the result of the conversion
		 */
		@SuppressWarnings("unchecked")
		protected String leafToDot(LearningLeaf<T, ?> leaf, String defaultString) {
			if (!(leaf.get() instanceof IProbabilities)) {
				return String.valueOf(leaf.get());
			}
			ILearningProbabilities<String> proba = (ILearningProbabilities<String>) leaf
					.get();
			StringBuilder result = new StringBuilder();
			for (String value : proba.getValueSet()) {
				int percent = (int) (proba.getProbability(value) * 100);
				result.append(value + ": " + percent + "%");
				result.append("\\n");
			}
			return result.toString();
		}
	}

	/**
	 * 
	 * Policy to output a tree with the examples
	 * 
	 * @param <T>
	 *          type of the data stored in the tree
	 * 
	 */
	private static class IToDotLightPolicy<T> extends ToDotPolicy<T> {
		/**
		 * 
		 * Constructor
		 * 
		 */
		protected IToDotLightPolicy() {
			super();
		}

		/**
		 * @see animatlab.kodamat.supervisedlearning.core.LearningToDotUtils.ToDotPolicy#dnodeToDot(LearningDNode,
		 *      java.lang.String)
		 */
		@Override
		protected String dnodeToDot(LearningDNode<T, ?> dnode, String defaultString) {
			return defaultString;
		}
	}

	/**
	 * 
	 * Policy to output a tree with the examples
	 * 
	 * @param <T>
	 *          type of the data stored in the tree
	 * 
	 */
	public static class IToDotExamplesPolicy<T> extends IToDotLightPolicy<T> {
		/**
		 * 
		 * Constructor
		 * 
		 */
		public IToDotExamplesPolicy() {
			super();
		}

		/**
		 * @see animatlab.kodamat.supervisedlearning.core.LearningToDotUtils.ToDotPolicy#leafToDot(LearningLeaf,
		 *      java.lang.String)
		 */
		@Override
		protected String leafToDot(LearningLeaf<T, ?> leaf, String defaultString) {
			StringBuilder result = new StringBuilder(longProbaToDotString(leaf
					.getStats()));
			if (leaf.getExampleDatabase().size() > 0) {
				result.append("\\n");
				result.append(leaf.getExampleDatabase().toDotString());
			}
			return result.toString();
		}
	}

	/**
	 * 
	 * Policy to output a tree with the measures data
	 * 
	 * @param <T>
	 *          type of the data stored in the tree
	 * 
	 */
	private class IToDotFullPolicy<T> extends ToDotPolicy<T> {
		/**
		 * 
		 * Constructor
		 * 
		 */
		protected IToDotFullPolicy() {
			super();
		}

		/**
		 * @see animatlab.kodamat.supervisedlearning.core.LearningToDotUtils.ToDotPolicy#dnodeToDot(LearningDNode,
		 *      java.lang.String)
		 */
		@Override
		protected String dnodeToDot(LearningDNode<T, ?> dnode, String defaultString) {
			return vertexToDotString(dnode);
		}

		/**
		 * 
		 * Convert a vertex into a dot string
		 * 
		 * @param vertex
		 *          the vertex to convert
		 * @return the string representing the vertex
		 * 
		 */
		private String vertexToDotString(ILearningVertex<T> vertex) {
			VariableValueStatistics<T> stats = vertex.getStats();
			VariableMeasure<T> variableMeasure = vertex.getTree().getMeasure();
			Set<Variable> ignoredVariables = vertex.getContext().keySet();
			ExampleDatabase<T> exampleDatabase = vertex.getExampleDatabase();
			VariableMeasureData variableMeasureData = variableMeasure.getBestMeasure(
					exampleDatabase, stats, ignoredVariables);
			StringBuilder result = new StringBuilder();
			if (variableMeasureData != null) {
				result.append(variableMeasure.getBestMeasure(exampleDatabase, stats,
						ignoredVariables).toString());
			}
			if (exampleDatabase != null) {
				if (result.length() > 0) {
					result.append("\\n");
				}
				result.append(stats.getNumberOfExamples() + " examples");
			}
			if (result.length() > 0) {
				result.append("\\n");
			}
			result.append(Parameters.format(variableMeasure.getStatsError(stats,
					exampleDatabase, ignoredVariables))
					+ " (>" + Parameters.format(variableMeasure.getThreshold()) + ")");
			// if (variableMeasure instanceof CStrikeRegressionMeasure) {
			// if (result.length() > 0) {
			// result.append("\\n");
			// }
			// result.append(customizeDotInfo(vertex));
			// }
			return result.toString();
		}

		/**
		 * 
		 * Print some customized infos for the regression measure
		 * 
		 * @param vertex
		 *          the vertex for which the info is printed
		 * @return the string representing the customized info
		 * 
		 */
		@SuppressWarnings("unchecked")
		protected String customizeDotInfo(ILearningVertex<T> vertex) {
			VariableValueStatistics<Double> stats = (VariableValueStatistics<Double>) vertex
					.getStats();
			// RegressionMeasure variableMeasure = (RegressionMeasure)
			// vertex.getTree().getMeasure();
			// Set<Variable> ignoredVariables = vertex.getContext().keySet();
			// ExampleDatabase<Double> exampleDatabase = (ExampleDatabase<Double>)
			// vertex.getExampleDatabase();
			StringBuilder result = new StringBuilder();
			for (Double value : stats.getValueSet()) {
				result.append(Parameters.format(value) + ":"
						+ stats.getNumberOfExamples(value));
				result.append(" ");
			}
			return result.toString();
		}

		/**
		 * @see animatlab.kodamat.supervisedlearning.core.LearningToDotUtils.ToDotPolicy#leafToDot(LearningLeaf,
		 *      java.lang.String)
		 */
		@Override
		protected String leafToDot(LearningLeaf<T, ?> leaf, String defaultString) {
			StringBuilder result = new StringBuilder(vertexToDotString(leaf));
			if (result.length() > 0) {
				result.append("\\n");
			}
			result.append(super.leafToDot(leaf, defaultString));
			return result.toString();
		}
	}

	/**
	 * 
	 * List of policies to run to output a tree
	 * 
	 */
	public final Map<String, ToDotPolicy<U>> policyList = new LinkedHashMap<String, ToDotPolicy<U>>();

	/**
	 * 
	 * Constructor
	 * 
	 */
	public LearningToDotUtils() {
		policyList.put("light", new IToDotLightPolicy<U>());
		policyList.put("full", new IToDotFullPolicy<U>());
		// policyList.put("examples", new IToDotExamplesPolicy<U>());
	}

	/**
	 * 
	 * Dump a learning tree
	 * 
	 * @param tree
	 *          the tree to dump
	 * @param treesPage
	 *          page where to write the tree
	 * 
	 */
	public void dumpTree(Tree<U> tree, HtmlPage treesPage) {
		for (Map.Entry<String, ToDotPolicy<U>> entry : policyList.entrySet()) {
			treesPage.append(DotUtils.toHtmlDot(tree, entry.getValue(), treesPage
					.getFolder(), DotUtils.toNormalizeString(treesPage.getTitle())
					.toLowerCase()
					+ entry.getKey()));
		}
	}
}
