package animatlab.kodamat.supervisedlearning.entropy;

import java.util.Set;

import animatlab.kodamat.core.proba.ILearningProbabilities;
import animatlab.kodamat.core.proba.ValueStatistics;
import animatlab.kodamat.core.utils.Prototype;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.supervisedlearning.core.ExampleDatabase;

/**
 * 
 * Variable measure using Kolmogorov-Smirnoff metric to compute the measure
 * 
 * @author Thomas Degris
 * 
 */
public class GainRatioMeasure extends VariableMeasure<String> {
	/**
	 * 
	 * Current value of Ipn to evaluate the entropy
	 * 
	 */
	double ipn = Double.NaN;

	/**
	 * 
	 * Current value of currentG_ to evaluate the entropy
	 * 
	 */
	double g_ = Double.NaN;

	/**
	 * 
	 * Constructor
	 * 
	 */
	public GainRatioMeasure() {
		super((Double) null);
	}

	/**
	 * 
	 * Compute the I function (I(x,y))defined (page 3) in Utgoff, P.E. (1989).
	 * Incremental induction of decision trees. Machine Learning, 4, 161-186
	 * 
	 * @param valueStats
	 *          the value to partition
	 * @param numberOfTrainingInstance
	 *          total number of training instance
	 * 
	 * @return the I function defined as -x/(x+y)log(x/(x+y))-y/(x+y)log(y/(x+y))
	 * 
	 */
	protected double getIFun(ILearningProbabilities<String> valueStats,
			int numberOfTrainingInstance) {
		double result = 0;
		for (String className : valueStats.getValueSet()) {
			double division = valueStats.getNumberOfExamples(className)
					/ (double) numberOfTrainingInstance;
			result += -division * Math.log(division);
		}
		return result;
	}

	/**
	 * 
	 * Compute the I function (I(x,y))defined (page 3) in Utgoff, P.E. (1989).
	 * Incremental induction of decision trees. Machine Learning, 4, 161-186
	 * 
	 * @param mainStats
	 *          main stats to compare with
	 * @param valueStats
	 *          the value to partition
	 * 
	 * @return the I function defined as -x/(x+y)log(x/(x+y))-y/(x+y)log(y/(x+y))
	 * 
	 */
	protected double getIFun(ILearningProbabilities<String> mainStats,
			ILearningProbabilities<String> valueStats) {
		if (valueStats.getValueSet().size() < mainStats.getValueSet().size()) {
			return 0;
		}
		return getIFun(valueStats, valueStats.getNumberOfExamples());
	}

	/**
	 * 
	 * Compute the E-score (E(ai)) defined (page 3) in Utgoff, P.E. (1989).
	 * Incremental induction of decision trees. Machine Learning, 4, 161-186
	 * 
	 * @param mainStats
	 *          the stats to use to compute the measure
	 * @param variable
	 *          the variable for which the measure are computed
	 * 
	 * @return the E-score of the attribute over all the instances
	 * 
	 */
	protected double getEscore(VariableValueStatistics<String> mainStats,
			Variable variable) {
		double escore = 0;
		for (ValueStatistics<String> valueStats : mainStats.getValueStats(variable)
				.values()) {
			double sum = (double) valueStats.getNumberOfExamples()
					/ mainStats.getNumberOfExamples() * getIFun(mainStats, valueStats);
			escore += sum;
		}
		return escore;
	}

	/**
	 * 
	 * Compute the IV function (IV(ai)) defined (page 22) in Utgoff, P.E. (1989).
	 * Incremental induction of decision trees. Machine Learning, 4, 161-186
	 * 
	 * @param mainStats
	 *          the stats to use to compute the measure
	 * @param variable
	 *          the variable for which the measure are computed
	 * 
	 * @return the IV(ai)of the attribute
	 * 
	 */
	protected double getIV(VariableValueStatistics<String> mainStats,
			Variable variable) {
		double iv = 0;
		for (ValueStatistics<String> valueStats : mainStats.getValueStats(variable)
				.values()) {
			double pijPlusnijOnpPlusn = (double) valueStats.getNumberOfExamples()
					/ mainStats.getNumberOfExamples();
			iv += pijPlusnijOnpPlusn * Math.log(pijPlusnijOnpPlusn);
		}
		return -iv;
	}

	/**
	 * 
	 * Compute the g bar function (g_) defined (page 22) in Utgoff, P.E. (1989).
	 * Incremental induction of decision trees. Machine Learning, 4, 161-186
	 * 
	 * @param mainStats
	 *          the stats to use to compute the measure
	 * @param ipn
	 *          I(p,n) value cached (to avoid to recompute it)
	 * 
	 * @return the currentG_ of the attribute over all the instances
	 * 
	 */
	protected double getg_(VariableValueStatistics<String> mainStats, double ipn) {
		double g_Sum = 0;
		for (Variable variable : mainStats.getVariableSet()) {
			g_Sum += ipn - getEscore(mainStats, variable);
		}
		return ((1.0f / mainStats.getVariableSet().size()) * g_Sum);
	}

	/**
	 * 
	 * Compute some data used by the measure
	 * 
	 * @param variableValueStatistics
	 *          the stats to compute the measure with
	 * @param exampleDatabase
	 *          database of examples
	 * @param ignoredVariables
	 *          set of variable to ignore
	 * 
	 */
	@Override
	protected void computeCachedData(
			VariableValueStatistics<String> variableValueStatistics,
			ExampleDatabase<String> exampleDatabase, Set<Variable> ignoredVariables) {
		ipn = getIFun(variableValueStatistics, variableValueStatistics
				.getNumberOfExamples());
		g_ = getg_(variableValueStatistics, ipn);
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.entropy.VariableMeasure#computeMeasure(animatlab.kodamat.core.utils.Variable,
	 *      animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics,
	 *      ExampleDatabase, Set)
	 */
	@Override
	protected double computeMeasure(Variable variable,
			VariableValueStatistics<String> mainStats,
			ExampleDatabase<String> exampleDatabase, Set<Variable> ignoredVariables) {
		double result;
		double ipnMinusEscore = ipn - getEscore(mainStats, variable);
		if (ipnMinusEscore >= g_) {
			result = ipnMinusEscore / getIV(mainStats, variable);
		} else {
			result = Double.NEGATIVE_INFINITY;
		}
		return result;
	}

	/**
	 * @see animatlab.kodamat.core.utils.Prototype#newInstance()
	 */
	public Prototype newInstance() {
		GainRatioMeasure result = new GainRatioMeasure();
		result.set(this);
		return result;
	}
}
