package animatlab.kodamat.supervisedlearning.entropy;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.proba.ILearningProbabilities;
import animatlab.kodamat.core.proba.ValueStatistics;
import animatlab.kodamat.core.utils.Prototype;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.supervisedlearning.core.ExampleDatabase;

/**
 * 
 * Variable measure using Kolmogorov-Smirnoff metric to compute the measure
 * 
 * @author Thomas Degris
 * 
 */
public class KSMeasure extends VariableMeasure<String> {
	/**
	 * 
	 * Policy used to classified the value
	 * 
	 * @author Thomas Degris
	 * 
	 */
	static private interface PartitionPolicy {
		/**
		 * 
		 * Get the corresponding KS key for one value and one class
		 * 
		 * @param valueStats
		 *          the value for which we want to get the key
		 * @param className
		 *          the class for which we want to get the key
		 * @param mainStats
		 *          counter for all the values
		 * 
		 * @return the KS key depending on the policy implemented
		 * 
		 */
		double getKey(ILearningProbabilities<String> valueStats, String className,
				ILearningProbabilities<String> mainStats);
	}

	/**
	 * 
	 * Policy used to get the KS key for the left branch (meaning the test at the
	 * node is true)
	 * 
	 * @author Thomas Degris
	 * 
	 */
	static private class LeftPartitionPolicy implements PartitionPolicy {
		/**
		 * 
		 * Constructor
		 * 
		 */
		LeftPartitionPolicy() {
			super();
		}

		/**
		 * @see animatlab.kodamat.supervisedlearning.entropy.KSMeasure.PartitionPolicy#getKey(animatlab.kodamat.core.proba.ILearningProbabilities,
		 *      java.lang.String,
		 *      animatlab.kodamat.core.proba.ILearningProbabilities)
		 */
		public double getKey(ILearningProbabilities<String> valueStats,
				String className, ILearningProbabilities<String> mainStats) {
			double allValuesClassCounter = mainStats.getNumberOfExamples(className);
			return valueStats.getNumberOfExamples(className) / allValuesClassCounter;
		}
	}

	/**
	 * 
	 * Policy used to get the KS key for the right branch (meaning the test at the
	 * node is false)
	 * 
	 * @author Thomas Degris
	 * 
	 */
	static private class RightPartitionPolicy implements PartitionPolicy {
		/**
		 * 
		 * Constructor
		 * 
		 */
		RightPartitionPolicy() {
			super();
		}

		/**
		 * @see animatlab.kodamat.supervisedlearning.entropy.KSMeasure.PartitionPolicy#getKey(animatlab.kodamat.core.proba.ILearningProbabilities,
		 *      java.lang.String,
		 *      animatlab.kodamat.core.proba.ILearningProbabilities)
		 */
		public double getKey(ILearningProbabilities<String> valueStats,
				String className, ILearningProbabilities<String> mainStats) {
			double allValuesClassCounter = mainStats.getNumberOfExamples(className);
			return (allValuesClassCounter - valueStats.getNumberOfExamples(className))
					/ allValuesClassCounter;
		}
	}

	/**
	 * 
	 * Instance of the policy used to get the KS key for the left branch (meaning
	 * the test at the node is true)
	 * 
	 */
	private static PartitionPolicy leftPartitionPolicy = new LeftPartitionPolicy();

	/**
	 * 
	 * Instance of the policy used to get the KS key for the right branch (meaning
	 * the test at the node is false)
	 * 
	 */
	private static PartitionPolicy rightPartitionPolicy = new RightPartitionPolicy();

	/**
	 * 
	 * Constructor
	 * 
	 */
	public KSMeasure() {
		super((Double) null);
	}

	/**
	 * 
	 * Create a sorted list of the class counter for a specific value and for a
	 * partition policy
	 * 
	 * @param mainStats
	 *          main stats to compare with
	 * @param valueStats
	 *          the value to partition
	 * @param partitionPolicy
	 *          the policy of the partition
	 * 
	 * @return the sorted list
	 * 
	 */
	final private List<Double> getSortedMap(
			ILearningProbabilities<String> mainStats,
			ILearningProbabilities<String> valueStats, PartitionPolicy partitionPolicy) {
		List<Double> sortedList = new ArrayList<Double>();
		for (String className : mainStats.getValueSet()) {
			sortedList.add(partitionPolicy.getKey(valueStats, className, mainStats));
		}
		Collections.sort(sortedList);
		return sortedList;
	}

	/**
	 * 
	 * Compute the index to know the class subgrouping to compute the
	 * Kolmogorov-smirnoff metric
	 * 
	 * @param sortedList
	 *          sortedList to get the index for the subgrouping
	 * 
	 * @return the index of the first class on the upper group
	 * 
	 */
	final private int getSuperClassSplitIndex(List<Double> sortedList) {
		if (sortedList.size() < 2) {
			return 1;
		}
		Double lastValue = sortedList.get(1);
		double maxValue = lastValue.doubleValue()
				- sortedList.get(0).doubleValue();
		int splitIndex = 1;
		for (int index = 2; index < sortedList.size(); index++) {
			Double currentValue = sortedList.get(index);
			double value = currentValue.doubleValue() - lastValue.doubleValue();
			if (value > maxValue) {
				splitIndex = index;
				maxValue = value;
			}
		}
		return splitIndex;
	}

	/**
	 * 
	 * Compute the KS metric for a specific value and for a partition policy
	 * 
	 * @param mainStats
	 *          main stats to compare with
	 * @param valueStats
	 *          the value to partition
	 * @param partitionPolicy
	 *          the policy of the partition
	 * 
	 * @return the sorted list
	 * 
	 */
	final private double getKSDistance(ILearningProbabilities<String> mainStats,
			ILearningProbabilities<String> valueStats, PartitionPolicy partitionPolicy) {
		List<Double> sortedList = getSortedMap(mainStats, valueStats, partitionPolicy);
		int splitIndex = getSuperClassSplitIndex(sortedList);
		double lowerSum = 0;
		double upperSum = 0;
		double listSum = 0;
		for (int index = 0; index < sortedList.size(); index++) {
			double currentValue = sortedList.get(index);
			if (index < splitIndex) {
				lowerSum += currentValue;
			} else {
				upperSum += currentValue;
			}
			listSum += currentValue;
		}
		return ((upperSum / listSum) - (lowerSum / listSum));
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.entropy.VariableMeasure#computeMeasure(animatlab.kodamat.core.utils.Variable,
	 *      animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics,
	 *      ExampleDatabase, Set)
	 */
	@Override
	protected double computeMeasure(Variable variable,
			VariableValueStatistics<String> mainStats,
			ExampleDatabase<String> exampleDatabase, Set<Variable> ignoredVariables) {
		double result = -1;
		Map<String, ValueStatistics<String>> varStats = mainStats
				.getValueStats(variable);
		for (Map.Entry<String, ValueStatistics<String>> entry : varStats.entrySet()) {
			double metric = ((getKSDistance(mainStats, entry.getValue(),
					leftPartitionPolicy) + getKSDistance(mainStats, entry.getValue(),
					rightPartitionPolicy)) / 2.0f);
			result = Math.max(result, metric);
		}
		return result;
	}

	/**
	 * @see animatlab.kodamat.core.utils.Prototype#newInstance()
	 */
	public Prototype newInstance() {
		KSMeasure result = new KSMeasure();
		result.set(this);
		return result;
	}
}
