package animatlab.kodamat.supervisedlearning.entropy;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.proba.ValueStatistics;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Prototype;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.supervisedlearning.core.ExampleDatabase;

/**
 * 
 * Compute the best measure from a set of statistics given for each variables
 * 
 * @author Thomas Degris
 * @param <T>
 *          type of the data for which has been designed the test
 * 
 */
public abstract class VariableMeasure<T> implements Prototype {
	/**
	 * 
	 * Structure to store a test and its entropy value
	 * 
	 */
	static public class VariableMeasureData {
		/**
		 * 
		 * Variable concerned by this info
		 * 
		 */
		final public Variable variable;

		/**
		 * 
		 * Entropy measure corresponding to the test
		 * 
		 */
		public double measure = Double.NaN;

		/**
		 * 
		 * Constructor
		 * 
		 * @param variable
		 *          variable tested
		 * 
		 */
		public VariableMeasureData(Variable variable) {
			this.variable = variable;
		}

		/**
		 * @see java.lang.Object#toString()
		 */
		@Override
		public String toString() {
			return variable.toString() + ":" + Parameters.format(measure);
		}
	}

	/**
	 * 
	 * Measures computed for each variables
	 * 
	 */
	final private Map<Variable, VariableMeasureData> measures = new LinkedHashMap<Variable, VariableMeasureData>();

	/**
	 * 
	 * Threshold used when the problem is stochastic (set to null when the problem
	 * is not stochastic)
	 * 
	 */
	protected Double threshold = null;

	/**
	 * 
	 * Statistics used as reference
	 * 
	 */
	protected Map<Variable, Set<String>> allValueSets = null;

	/**
	 * 
	 * Flag to set to decide weither or not variables with missing examples
	 * (branches with no examples) may be installed or not. The variable refStats
	 * is used to check value sets
	 * 
	 */
	public boolean noEmptyLeaf = false;

	/**
	 * 
	 * Constructor
	 * 
	 * @param threshold
	 *          the threshold to use
	 * 
	 */
	public VariableMeasure(Double threshold) {
		this.threshold = threshold;
	}

	/**
	 * 
	 * The stats used to test if they are pure or not
	 * 
	 * @param variableValueStatistics
	 *          the stats to compute the measure with
	 * @param exampleDatabase
	 *          database of examples
	 * @param ignoredVariables
	 *          variables already set
	 * 
	 * @return true if all the examples at this vertex are from the same class
	 * 
	 */
	public boolean isPure(VariableValueStatistics<T> variableValueStatistics,
			ExampleDatabase<T> exampleDatabase, Set<Variable> ignoredVariables) {
		return (variableValueStatistics.getValueSet().size() <= 1);
	}

	/**
	 * 
	 * Compute the threshold value associated to a set of statistics
	 * 
	 * @param variableValueStatistics
	 *          the stats to compute the measure with
	 * @param exampleDatabase
	 *          database of examples
	 * @param ignoredVariables
	 *          variables already set
	 * @return the threshold used for the node
	 * 
	 */
	public double getStatsError(
			VariableValueStatistics<T> variableValueStatistics,
			ExampleDatabase<T> exampleDatabase, Set<Variable> ignoredVariables) {
		VariableMeasureData bestTest = getBestMeasure(exampleDatabase,
				variableValueStatistics, ignoredVariables);
		if (bestTest == null) {
			return 0;
		}
		return bestTest.measure;
	}

	/**
	 * 
	 * Build a string with the measure computed for every variables
	 * 
	 * @param exampleDatabase
	 *          the database of the examples
	 * @param variableValueStatistics
	 *          the stats to compute the measure with
	 * @param ignoredVariables
	 *          the variables to ignore for the computation
	 * 
	 * @return a string representing the measure computed for each variables
	 * 
	 * 
	 */
	protected String toStringDebug(ExampleDatabase<T> exampleDatabase,
			VariableValueStatistics<T> variableValueStatistics,
			Set<Variable> ignoredVariables) {
		computeCachedData(variableValueStatistics, exampleDatabase,
				ignoredVariables);
		StringBuilder result = new StringBuilder();
		for (Variable variable : variableValueStatistics.getVariableSet()) {
			result.append(variable.toString());
			result.append(": ");
			if ((variableValueStatistics.getValueSet(variable) != null)
					&& ((ignoredVariables == null) || !ignoredVariables
							.contains(variable))
					&& checkSampleSize(variableValueStatistics, variable)
					&& checkValueSet(variable, variableValueStatistics
							.getValueSet(variable))) {
				result.append(computeMeasure(variable, variableValueStatistics,
						exampleDatabase, ignoredVariables));
			} else {
				result.append("ign");
			}
			result.append("\n");
		}
		return result.toString();
	}

	/**
	 * 
	 * Compute the best measure for a stats
	 * 
	 * @param variableValueStatistics
	 *          the stats to compute the measuree for
	 * @param exampleDatabase
	 *          the example database to compute the measure with
	 * @param ignoredVariables
	 *          the set of variable to ignore
	 * @return the best measure
	 * 
	 */
	public VariableMeasureData getBestMeasure(ExampleDatabase<T> exampleDatabase,
			VariableValueStatistics<T> variableValueStatistics,
			Set<Variable> ignoredVariables) {
		if (isPure(variableValueStatistics, exampleDatabase, ignoredVariables)) {
			return null;
		}
		computeCachedData(variableValueStatistics, exampleDatabase,
				ignoredVariables);
		VariableMeasureData bestTest = null;
		for (Variable variable : variableValueStatistics.getVariableSet()) {
			if ((variableValueStatistics.getValueSet(variable) != null)
					&& ((ignoredVariables == null) || !ignoredVariables
							.contains(variable))
					&& checkSampleSize(variableValueStatistics, variable)
					&& checkValueSet(variable, variableValueStatistics
							.getValueSet(variable))) {
				VariableMeasureData variableMeasure = measures.get(variable);
				if (variableMeasure == null) {
					variableMeasure = new VariableMeasureData(variable);
					measures.put(variable, variableMeasure);
				}
				variableMeasure.measure = computeMeasure(variable,
						variableValueStatistics, exampleDatabase, ignoredVariables);
				if (!Double.isNaN(variableMeasure.measure)
						&& ((bestTest == null) || (variableMeasure.measure > (bestTest.measure + Parameters.EPSILON)))) {
					bestTest = variableMeasure;
				}
			}
		}
		return bestTest;
	}

	/**
	 * 
	 * Select the variable with the more info
	 * 
	 * @param exampleDatabase
	 *          the database of the examples
	 * @param variableValueStatistics
	 *          the stats to compute the measure with
	 * @param ignoredVariables
	 *          the variables to ignore for the computation
	 * @param currentTest
	 *          current tested variable
	 * 
	 * @return the index in the vector Example.values of the best variable
	 *         selected (may be null if no test has been found)
	 * 
	 */
	public VariableMeasureData getBestTest(ExampleDatabase<T> exampleDatabase,
			VariableValueStatistics<T> variableValueStatistics,
			Set<Variable> ignoredVariables, Variable currentTest) {
		if (isPure(variableValueStatistics, exampleDatabase, ignoredVariables)) {
			return null;
		}
		computeCachedData(variableValueStatistics, exampleDatabase,
				ignoredVariables);
		VariableMeasureData bestTest = getBestMeasure(exampleDatabase,
				variableValueStatistics, ignoredVariables);
		if ((threshold != null)
				&& (bestTest != null)
				&& !isSignificant(bestTest, variableValueStatistics, exampleDatabase,
						ignoredVariables, currentTest)) {
			return null;
		}
		return bestTest;
	}

	/**
	 * 
	 * Check if the value set of the variable contains all the value contained in
	 * the reference stats
	 * 
	 * @param variable
	 *          the variable to check
	 * @param valueSet
	 *          its value set to compare
	 * @return true if the set if ok
	 * 
	 */
	protected boolean checkValueSet(Variable variable, Set<String> valueSet) {
		if ((allValueSets == null) || !noEmptyLeaf) {
			return true;
		}
		if (valueSet.containsAll(allValueSets.get(variable))) {
			return true;
		}
		return false;
	}

	/**
	 * 
	 * Compute some data used by the measure
	 * 
	 * @param variableValueStatistics
	 *          the stats to compute the measure with
	 * @param exampleDatabase
	 *          database example
	 * @param ignoredVariables
	 *          the set of variable to ignore
	 * 
	 */
	protected void computeCachedData(
			VariableValueStatistics<T> variableValueStatistics,
			ExampleDatabase<T> exampleDatabase, Set<Variable> ignoredVariables) {
		// Nothing to do by default
	}

	/**
	 * 
	 * Check if the number of examples learned from a variable is equal to the
	 * number of examples learned
	 * 
	 * @param variableValueStatistics
	 *          the stats to check
	 * @param variable
	 *          the variable to check
	 * 
	 * @return true if the number of examples is the same
	 * 
	 */
	private boolean checkSampleSize(
			VariableValueStatistics<T> variableValueStatistics, Variable variable) {
		int numberOfExamples = 0;
		final Map<String, ValueStatistics<T>> varStats = variableValueStatistics
				.getValueStats(variable);
		// if (varStats.size() <= 1) {
		// return false;
		// }
		for (ValueStatistics<T> stat : varStats.values()) {
			numberOfExamples = numberOfExamples + stat.getNumberOfExamples();
		}
		return (numberOfExamples == variableValueStatistics.getNumberOfExamples());
	}

	/**
	 * 
	 * Test if a test on a variable is significant or not
	 * 
	 * @param bestTest
	 *          the test to check with
	 * @param variableValueStatistics
	 *          the stats to use to compute the measure
	 * @param exampleDatabase
	 *          database of examples
	 * @param ignoredVariables
	 *          the set of ignored variables
	 * @param currentTest
	 *          current test
	 * @return true is the test is significant
	 * 
	 */
	protected boolean isSignificant(VariableMeasureData bestTest,
			VariableValueStatistics<T> variableValueStatistics,
			ExampleDatabase<T> exampleDatabase, Set<Variable> ignoredVariables,
			Variable currentTest) {
		return true;
	}

	/**
	 * 
	 * Compute the measure for one variable
	 * 
	 * @param variable
	 *          the variable for which the measure are computed
	 * @param variableValueStatistics
	 *          the stats to use to compute the measure
	 * @param exampleDatabase
	 *          database of examples
	 * @param ignoredVariables
	 *          the set of ignored variables
	 * 
	 * @return the measure computed for the variable
	 * 
	 */
	abstract protected double computeMeasure(Variable variable,
			VariableValueStatistics<T> variableValueStatistics,
			ExampleDatabase<T> exampleDatabase, Set<Variable> ignoredVariables);

	/**
	 * 
	 * Aggregate a set of statistic to only one value
	 * 
	 * @param variableValueStatistics
	 *          the stats to aggregate
	 * @param exampleDatabase
	 *          database of examples
	 * @param ignoredVariables
	 *          set of variables to ignore
	 * @return the value representing the stats
	 * 
	 */
	public T aggregate(VariableValueStatistics<T> variableValueStatistics,
			ExampleDatabase<T> exampleDatabase, Set<Variable> ignoredVariables) {
		return variableValueStatistics.getBest();
	}

	/**
	 * 
	 * Build a string representing the stats
	 * 
	 * @param stats
	 *          the stats to represent in string
	 * @return a string representing
	 * 
	 */
	public String toString(VariableValueStatistics<T> stats) {
		return aggregate(stats, null, null).toString();
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return measures.toString();
	}

	/**
	 * 
	 * Print these stats into a dot string
	 * 
	 * @param stats
	 *          the stats to represent in string
	 * @param exampleDatabase
	 *          database of examples
	 * @param ignoredVariable
	 *          set of variables to ignored when computing the stats
	 * @param variableOrdering
	 *          the legend used to print out variable name
	 * @return a long dot string representing this data
	 * 
	 */
	public String infoToDotString(VariableValueStatistics<T> stats,
			ExampleDatabase<T> exampleDatabase, Set<Variable> ignoredVariable,
			VariableOrdering variableOrdering) {
		VariableMeasureData bestTestInfo = getBestTest(exampleDatabase, stats,
				ignoredVariable, null);
		if (bestTestInfo == null) {
			return "";
		}
		StringBuilder result = new StringBuilder();
		result.append("Best: ");
		result.append(bestTestInfo.variable.toString());
		result.append(" (" + Parameters.format(bestTestInfo.measure) + ")");
		return result.toString();
	}

	/**
	 * 
	 * Set the threshold used
	 * 
	 * @param threshold
	 *          the new threshold to use
	 * 
	 */
	public void setThreshold(Double threshold) {
		this.threshold = threshold;
	}

	/**
	 * 
	 * @return Returns the threshold.
	 * 
	 */
	public Double getThreshold() {
		return threshold;
	}

	/**
	 * 
	 * Set sets of values used as reference
	 * 
	 * @param allValueSets
	 *          map between variables and their value sets
	 * 
	 */
	public void setValuesSets(Map<Variable, Set<String>> allValueSets) {
		this.allValueSets = allValueSets;
	}

	/**
	 * 
	 * Set the parameters from a prototype
	 * 
	 * @param variableMeasure
	 *          the prototype to get the parameters from
	 * 
	 */
	protected void set(VariableMeasure<T> variableMeasure) {
		threshold = variableMeasure.threshold;
		noEmptyLeaf = variableMeasure.noEmptyLeaf;
	}

	/**
	 * 
	 * @return the name of the measure (only if necessary)
	 * 
	 */
	public String getName() {
		return "";
	}
}
