package animatlab.kodamat.supervisedlearning.entropy;

import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.proba.ILearningProbabilities;
import animatlab.kodamat.core.proba.ValueStatistics;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.supervisedlearning.core.Example;

/**
 * 
 * Statistics for a set of random variables
 * 
 * @author Thomas Degris
 * @param <T>
 *          type of value to build the stats for
 * 
 */
public class VariableValueStatistics<T> extends ValueStatistics<T> {
	/**
	 * 
	 * Statistics for all the values of a variable
	 * 
	 */
	private final Map<Variable, Map<String, ValueStatistics<T>>> stats = new LinkedHashMap<Variable, Map<String, ValueStatistics<T>>>();

	/**
	 * 
	 * Constructor
	 * 
	 */
	public VariableValueStatistics() {
		super();
	}

	/**
	 * 
	 * Add a new example to this stats
	 * 
	 * @param example
	 *          the example to add
	 * @param variableOrdering
	 *          the legend to use to get variable ordering
	 * 
	 */
	public void add(VariableOrdering variableOrdering, Example<T> example) {
		add(example.className);
		for (int index = 0; index < example.values.size(); index++) {
			Variable variable = variableOrdering.get(index);
			String value = example.values.get(index);
			if (value != null) {
				Map<String, ValueStatistics<T>> variableStats = stats.get(variable);
				if (variableStats == null) {
					variableStats = new LinkedHashMap<String, ValueStatistics<T>>();
					stats.put(variable, variableStats);
				}
				ValueStatistics<T> valueStats = variableStats.get(value);
				if (valueStats == null) {
					valueStats = new ValueStatistics<T>();
					variableStats.put(value, valueStats);
				}
				valueStats.add(example.className);
			}
		}
	}

	/**
	 * 
	 * @return the set of variable for which the stats are computed
	 * 
	 */
	public Set<Variable> getVariableSet() {
		return stats.keySet();
	}

	/**
	 * 
	 * Get a set of all the stats for all the values of one variable
	 * 
	 * @param variable
	 *          the variable to get the stats for
	 * @return a collection of the stats for one variable
	 * 
	 */
	public Set<ILearningProbabilities<T>> getStats(Variable variable) {
		Set<ILearningProbabilities<T>> result = new LinkedHashSet<ILearningProbabilities<T>>();
		result.addAll(stats.get(variable).values());
		return result;
	}

	/**
	 * 
	 * Get the domaine of a variable
	 * 
	 * @param variable
	 *          the variable to get the domaine from
	 * @return the domaine of the variable
	 * 
	 */
	public Set<String> getValueSet(Variable variable) {
		Map<String, ValueStatistics<T>> variableStats = stats.get(variable);
		if (variableStats == null) {
			return null;
		}
		return variableStats.keySet();
	}

	/**
	 * @see animatlab.kodamat.core.proba.ValueStatistics#merge(animatlab.kodamat.core.proba.ILearningProbabilities)
	 */
	@SuppressWarnings("unchecked")
	@Override
	public void merge(ILearningProbabilities<T> otherStats) {
		super.merge(otherStats);
		VariableValueStatistics<T> other = (VariableValueStatistics<T>) otherStats;
		for (Map.Entry<Variable, Map<String, ValueStatistics<T>>> entry : other.stats
				.entrySet()) {
			Variable variable = entry.getKey();
			Map<String, ValueStatistics<T>> variableStats = stats.get(variable);
			if (variableStats == null) {
				variableStats = new LinkedHashMap<String, ValueStatistics<T>>();
				stats.put(variable, variableStats);
			}
			for (Map.Entry<String, ValueStatistics<T>> varEntry : variableStats
					.entrySet()) {
				String value = varEntry.getKey();
				ValueStatistics<T> valueStats = variableStats.get(value);
				if (valueStats == null) {
					valueStats = new ValueStatistics<T>();
					variableStats.put(value, valueStats);
				}
				valueStats.merge(varEntry.getValue());
			}
		}
	}

	/**
	 * 
	 * Get the stats for a variable for each value
	 * 
	 * @param variable
	 *          the value to get the stats from
	 * @return the stats for each value of the variable
	 * 
	 */
	public Map<String, ValueStatistics<T>> getValueStats(Variable variable) {
		return stats.get(variable);
	}
}
