package animatlab.kodamat.supervisedlearning.learningtree;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.proba.ILearningProbabilities;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Vertex;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.supervisedlearning.core.Example;
import animatlab.kodamat.supervisedlearning.core.ExampleDatabase;
import animatlab.kodamat.supervisedlearning.entropy.VariableMeasure;
import animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics;

/**
 * 
 * ID3-like algorithm
 * 
 * @author Thomas Degris
 * @param <T>
 *          type of the examples to learn
 * @param <U>
 *          type of the data in the leaf
 * 
 */
public abstract class ID3Tree<T, U> extends LearningTree<T, U> {
	/**
	 * 
	 * Probability version of the ID3 algorithm
	 * 
	 * @author Thomas Degris
	 * @param <V>
	 *          type of the examples to learn
	 * 
	 */
	static public class Prob<V> extends ID3Tree<V, IProbabilities<V>> {
		/**
		 * 
		 * Constructor
		 * 
		 * @param variableMeasure
		 *          prototype of the measure used to select the test to install
		 * @param variableOrdering
		 *          variable ordering to use with the examples
		 * 
		 */
		public Prob(VariableMeasure<V> variableMeasure,
				VariableOrdering variableOrdering) {
			super(variableMeasure, variableOrdering);
		}

		/**
		 * 
		 * Set the data at a leaf
		 * 
		 * @param leaf
		 *          the leaf to set the data
		 * @param stats
		 *          the data to set
		 * @param exampleDatabase
		 *          database of examples
		 * 
		 */
		@Override
		protected void setDataToLeaf(Leaf<IProbabilities<V>> leaf,
				VariableValueStatistics<V> stats, ExampleDatabase<V> exampleDatabase) {
			leaf.set(stats);
		}

		/**
		 * @see animatlab.kodamat.core.utils.Prototype#newInstance()
		 */
		@Override
		public ID3Tree<V, IProbabilities<V>> newInstance() {
			return new Prob<V>(variableMeasure, variableOrdering);
		}
	}

	/**
	 * 
	 * Plain version of the ID3 algorithm
	 * 
	 * @author Thomas Degris
	 * @param <V>
	 *          type of the examples to learn
	 * 
	 */
	static public class Plain<V> extends ID3Tree<V, V> {
		/**
		 * 
		 * Constructor
		 * 
		 * @param variableMeasure
		 *          prototype of the measure used to select the test to install
		 * @param variableOrdering
		 *          variable ordering to use with the examples
		 * 
		 */
		public Plain(VariableMeasure<V> variableMeasure,
				VariableOrdering variableOrdering) {
			super(variableMeasure, variableOrdering);
		}

		/**
		 * 
		 * Set the data at a leaf
		 * 
		 * @param leaf
		 *          the leaf to set the data
		 * @param stats
		 *          the data to set
		 * @param exampleDatabase
		 *          database of examples
		 * 
		 */
		@Override
		protected void setDataToLeaf(Leaf<V> leaf,
				VariableValueStatistics<V> stats, ExampleDatabase<V> exampleDatabase) {
			leaf.set(variableMeasure.aggregate(stats, exampleDatabase, leaf
					.getContext().keySet()));
		}

		/**
		 * @see animatlab.kodamat.core.utils.Prototype#newInstance()
		 */
		@Override
		public ID3Tree<V, V> newInstance() {
			return new Plain<V>(variableMeasure, variableOrdering);
		}
	}

	/**
	 * 
	 * Prototype of the database used to store the example going though this node
	 * to be able to select the best test available at this node.
	 * 
	 */
	final private VariableValueStatistics<T> statsDatabase = new VariableValueStatistics<T>();

	/**
	 * 
	 * Database storing the number of time an example is added
	 * 
	 */
	final private ExampleDatabase<T> exampleDatabase = new ExampleDatabase<T>();

	/**
	 * 
	 * Constructor
	 * 
	 * @param variableMeasure
	 *          Prototype of the database used to store the example going though
	 *          this node to be able to select the best test available at this
	 *          node.
	 * @param variableOrdering
	 *          variable ordering to use with the examples
	 * 
	 */
	public ID3Tree(VariableMeasure<T> variableMeasure,
			VariableOrdering variableOrdering) {
		super(variableMeasure, variableOrdering);
		this.variableMeasure.noEmptyLeaf = true;
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.core.ISupervisedLearning#learn(animatlab.kodamat.supervisedlearning.core.Example)
	 */
	@Override
	public void learn(Example<T> example) {
		getRoot().remove();
		super.learn(example);
		statsDatabase.add(variableOrdering, example);
		exampleDatabase.add(example);
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.core.ISupervisedLearning#getStats()
	 */
	@Override
	public ILearningProbabilities<T> getStats() {
		return statsDatabase;
	}

	/**
	 * 
	 * Learn a new example (ID3-like algorithm)
	 * 
	 * @param leaf
	 *          the leaf from which we learn
	 * @param stats
	 *          the stats about the examples
	 * @param examples
	 *          the list of examples to learn
	 * 
	 */
	@SuppressWarnings("unchecked")
	private void learn(LearningLeaf<T, U> leaf, VariableValueStatistics<T> stats,
			ExampleDatabase<T> examples) {
		VariableMeasure.VariableMeasureData bestTest = variableMeasure.getBestTest(
				exampleDatabase, stats, leaf.getContext().keySet(), null);
		if (bestTest == null) {
			setDataToLeaf(leaf, stats, exampleDatabase);
			leaf.setStats(stats);
			return;
		}
		LearningDNode<T, U> dnode = (LearningDNode<T, U>) leaf
				.replace(bestTest.variable);
		dnode.setStats(stats);
		Map<String, VariableValueStatistics<T>> statsMap = new LinkedHashMap<String, VariableValueStatistics<T>>();
		Map<String, ExampleDatabase<T>> examplesMap = new LinkedHashMap<String, ExampleDatabase<T>>();
		final Set<String> bestTestValueSet = stats.getValueSet(bestTest.variable);
		for (String value : bestTestValueSet) {
			dnode.addLeaf(value);
			statsMap.put(value, new VariableValueStatistics<T>());
			examplesMap.put(value, new ExampleDatabase<T>());
		}
		for (Map.Entry<Example<T>, Integer> entry : examples.entrySet()) {
			String exampleValue = entry.getKey().values.get(variableOrdering
					.indexOf(bestTest.variable));
			if (exampleValue != null) {
				addExample(variableOrdering, statsMap.get(exampleValue), examplesMap
						.get(exampleValue), entry.getKey(), entry.getValue());
			} else {
				for (String value : bestTestValueSet) {
					addExample(variableOrdering, statsMap.get(value), examplesMap
							.get(value), entry.getKey(), entry.getValue());
				}
			}
		}
		for (String value : bestTestValueSet) {
			learn((LearningLeaf<T, U>) dnode.getChild(value), statsMap.get(value),
					examplesMap.get(value));
		}
	}

	/**
	 * 
	 * Add an example for a leaf
	 * 
	 * @param variableOrdering
	 *          variable ordering in the examples
	 * @param stats
	 *          the stats about the examples
	 * @param examples
	 *          the list of examples to learn
	 * @param example
	 *          the example to add
	 * @param times
	 *          number of times the example has to be added
	 * 
	 */
	private void addExample(VariableOrdering variableOrdering,
			VariableValueStatistics<T> stats, ExampleDatabase<T> examples,
			Example<T> example, int times) {
		for (int counter = 0; counter < times; counter++) {
			stats.add(variableOrdering, example);
			examples.add(example);
		}
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.core.ISupervisedLearning#flush()
	 */
	@SuppressWarnings("unchecked")
	@Override
	public void flush() {
		getRoot().remove();
		learn((LearningLeaf<T, U>) getRoot(), statsDatabase, exampleDatabase);
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.LearningTree#learn(animatlab.kodamat.core.trees.Vertex,
	 *      animatlab.kodamat.supervisedlearning.core.Example)
	 */
	@Override
	protected void learn(Vertex<U> node, Example<T> example) {
		// Nothing to do : ID3 is not incremental
	}
}
