/*
 * 
 * Created on 26 dec. 2005
 *
 */
package animatlab.kodamat.supervisedlearning.learningtree;

import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.DNode;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.trees.Vertex;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.supervisedlearning.core.Example;
import animatlab.kodamat.supervisedlearning.core.ExampleDatabase;
import animatlab.kodamat.supervisedlearning.entropy.VariableMeasure;
import animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics;
import animatlab.kodamat.supervisedlearning.entropy.VariableMeasure.VariableMeasureData;

/**
 * 
 * ID5 algorithm from Incremental Induction of Decision Trees (Utgoff, 1989)
 * 
 * @author Thomas Degris
 * 
 */
public class ID5RTree extends LearningTree<String, IProbabilities<String>> {
	/**
	 * 
	 * Constructor
	 * 
	 * @param variableMeasure
	 *          measure to use to select the best test
	 * @param variableOrdering
	 *          variable ordering to use with the examples
	 * 
	 */
	public ID5RTree(VariableMeasure<String> variableMeasure,
			VariableOrdering variableOrdering) {
		super(variableMeasure, variableOrdering);
	}

	/**
	 * @see animatlab.kodamat.core.trees.Tree#newInstance()
	 */
	@Override
	public ID5RTree newInstance() {
		return new ID5RTree(variableMeasure, variableOrdering);
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.LearningTree#newDNode(animatlab.kodamat.core.trees.DNode,
	 *      java.lang.String, animatlab.kodamat.core.utils.Variable)
	 */
	@Override
	protected ID5DNode newDNode(DNode<IProbabilities<String>> parent,
			String value, Variable testedVariable) {
		return new ID5DNode(this, parent, value, testedVariable);
	}

	/**
	 * 
	 * Learn a new example (ID5R tree-update algorithm described in Incremental
	 * Induction of Decision Trees, Utgoff, p6)
	 * 
	 * @param node
	 *          root of the subtree where to learn the example
	 * @param example
	 *          the example to learn
	 * 
	 */
	@SuppressWarnings("unchecked")
	@Override
	protected void learn(Vertex<IProbabilities<String>> node,
			Example<String> example) {
		if (node.isLeaf) {
			LearningLeaf<String, IProbabilities<String>> leaf = (LearningLeaf<String, IProbabilities<String>>) node;
			// 1. If the tree is empty, then define it as the unexpanded form, setting
			// the class name to the class of the instance, and the set of instances
			// to the singleton set containing the instance
			if (leaf.get() == null) {
				leaf.set(((ILearningVertex<String>) leaf).getStats());
				leaf.addExample(example);
				return;
			}
			// 2. Otherwise, if the tree is in unexpanded form and the instance is
			// from the same class, then add the instance to the set of instances
			// kept in the node
			if ((leaf.get().getBest().equals(example.className) || !leaf
					.getExampleDatabase().isConsistant(example))) {
				leaf.addExample(example);
				return;
			}
			// 3. Otherwise,
			// (a) if the tree is in unexpanded form, then expand it one level,
			// choosing
			// the test attribute for the root arbitrarily
			Variable variable = chooseTestAttributeArbitrarily(leaf);
			ExampleDatabase<String> exampleDatabase = leaf.getExampleDatabase();
			node = expand(leaf, variable, leaf.getStats().getValueSet(variable));
			sendExamplesToLeaf((ID5DNode) node, exampleDatabase);
		}
		ID5DNode dnode = (ID5DNode) node;
		// (b) for the test attribute and all nontest attributes at the current
		// node, update
		// the count of positive or negative instances for the value of that
		// attribute in the
		// training instance.
		dnode.addExample(example);
		// (c) If the current node contains an attribute test that does not have the
		// lowest E-score, then
		VariableMeasureData bestTest = getBestTest(dnode);
		if ((bestTest != null)
				&& !bestTest.variable.equals(dnode.getTestedVariable())) {
			// omegaIndex. Restructure the tree so that an attribute with the lowest
			// E-score is at the root
			pullUp(dnode, bestTest.variable);
			// ii. Recursively reestablish a best test attribute in each subtree
			// except the one that
			// will be updated in step 3d.
			for (Vertex<IProbabilities<String>> child : dnode.children()) {
				if (!testIsTrue(child, example.values)) {
					installBestTest(child);
				}
			}
		}
		// (d) Recursively update the decision tree below the current decision node
		// along the branch for the value of the test attribute that occurs in the
		// instance description. Grow the branch if necessary.
		Vertex<IProbabilities<String>> child = getChild(dnode, example.values);
		if (child == null) {
			child = dnode.addLeaf(example.values.get(variableOrdering.indexOf(dnode
					.getTestedVariable())));
		}
		learn(child, example);
	}

	/**
	 * 
	 * Restructure the decision tree to install a desired test attribute at this
	 * node (ID5R pull-up algorithm described in Incremental Induction of Decision
	 * Trees, Utgoff, p7)
	 * 
	 * @param vertex
	 *          vertex where to pullup the test
	 * @param anew
	 *          the test to install at the root of the subtree
	 * 
	 */
	@SuppressWarnings("unchecked")
	protected void pullUp(Vertex<IProbabilities<String>> vertex, Variable anew) {
		ID5DNode dnode = null;
		if (!vertex.isLeaf) {
			dnode = (ID5DNode) vertex;
			dnode.setFresh(false);
		}
		// 1. If the attributes anew to be pulled up is already at the root, then
		// stop.
		if ((!vertex.isLeaf) && anew.equals(dnode.getTestedVariable())) {
			return;
		}
		// 2. Otherwise,
		// (a) Recursively pull the attribute anew to the root of each immediate
		// subtree. Convert any unexpanded tree to expanded form as necessary,
		// choosing attribute anew as the test attribute.
		if (vertex.isLeaf) {
			LearningLeaf<String, IProbabilities<String>> leaf = (LearningLeaf<String, IProbabilities<String>>) vertex;
			ExampleDatabase<String> exampleDatabase = leaf.getExampleDatabase();
			dnode = expand(leaf, anew, leaf.getStats().getValueSet(anew));
			dnode.setFresh(false);
			sendExamplesToLeaf(dnode, exampleDatabase);
			return;
		}
		for (Vertex<IProbabilities<String>> child : dnode.getNewChildrenSet()) {
			pullUp(child, anew);

		}
		// (b) Transpose the tree, resulting in a new tree with anew at the root,
		// and the old root attribute aold at the root of each immediate subtree
		transpose(dnode);
	}

	/**
	 * 
	 * Expand the tree from this node expanding for a variable of the state
	 * 
	 * @param variable
	 *          variable to use to create the children
	 * @param vertex
	 *          the leaf to expand
	 * @param valueSet
	 *          set of value of the variable to create one leaf for each value
	 * @return the decision node replacing the leaf
	 * 
	 */
	@SuppressWarnings("unchecked")
	final protected ID5DNode expand(Vertex<IProbabilities<String>> vertex,
			Variable variable, Set<String> valueSet) {
		ID5DNode dnode = (ID5DNode) vertex.replace(variable);
		dnode.setData((ILearningVertex<String>) vertex);
		for (String value : valueSet) {
			dnode.addLeaf(value);
		}
		return dnode;
	}

	/**
	 * 
	 * Choose a new test attribute arbitrarily
	 * 
	 * @param leaf
	 *          the leaf where to choose the test
	 * 
	 * @return the attribute (variable) choosen
	 * 
	 */
	protected Variable chooseTestAttributeArbitrarily(
			LearningLeaf<String, IProbabilities<String>> leaf) {
		Map<Variable,String> testsList = leaf.getContext();
		for (Variable variable : leaf.getStats().getVariableSet()) {
			if (testsList.get(variable) == null) {
				return variable;
			}
		}
		return null;
	}

	/**
	 * 
	 * Check if the best test is installed at this node and its children
	 * 
	 * @param vertex
	 *          root of the subtree to check
	 * 
	 */
	protected void installBestTest(Vertex<IProbabilities<String>> vertex) {
		if (vertex.isLeaf) {
			return;
		}
		ID5DNode dnode = (ID5DNode) vertex;
		VariableMeasureData bestTest = getBestTest(dnode);
		if ((bestTest != null)
				&& !bestTest.variable.equals(dnode.getTestedVariable())) {
			pullUp(dnode, bestTest.variable);
		}
		for (Vertex<IProbabilities<String>> child : dnode.children()) {
			installBestTest(child);
		}
	}

	/**
	 * 
	 * Transpose this tree taking as root as the subtree this node
	 * 
	 * @param dnode
	 *          dnode to transpose
	 * 
	 */
	@SuppressWarnings("unchecked")
	private void transpose(LearningDNode<String, IProbabilities<String>> dnode) {
		// Register the actual tree structure
		Variable oldLevel01TestVariable = dnode.getTestedVariable();
		Variable oldLevel02TestVariable = ((LearningDNode<String, IProbabilities<String>>) dnode
				.children().iterator().next()).getTestedVariable();
		Map<String, Map<String, Tree<IProbabilities<String>>>> level01NodeMap = new LinkedHashMap<String, Map<String, Tree<IProbabilities<String>>>>();
		LinkedHashSet<String> level02ValueSet = new LinkedHashSet<String>();
		for (Vertex<IProbabilities<String>> level01Node : dnode.getNewChildrenSet()) {
			Map<String, Tree<IProbabilities<String>>> level02NodeMap = new LinkedHashMap<String, Tree<IProbabilities<String>>>();
			for (Vertex<IProbabilities<String>> level02Node : ((DNode<IProbabilities<String>>) level01Node)
					.getNewChildrenSet()) {
				level02ValueSet.add(level02Node.getTestedValue());
				level02NodeMap.put(level02Node.getTestedValue(), level02Node.extract());
			}
			level01NodeMap.put(level01Node.getTestedValue(), level02NodeMap);
			level01Node.remove();
		}
		// Build the new tree from the existing nodes
		dnode = expand(dnode, oldLevel02TestVariable, level02ValueSet);
		for (Vertex<IProbabilities<String>> level01Node : dnode.getNewChildrenSet()) {
			ID5DNode level01DNode = expand(level01Node, oldLevel01TestVariable,
					level01NodeMap.keySet());
			level01DNode.setFresh(false);
			for (Vertex<IProbabilities<String>> level02Node : level01DNode
					.getNewChildrenSet()) {
				Tree<IProbabilities<String>> subTree = level01NodeMap.get(
						level02Node.getTestedValue()).get(level01DNode.getTestedValue());
				if (subTree == null) {
					level02Node.remove();
				} else {
					level01DNode.getStats().merge(
							((ILearningVertex<String>) subTree.getRoot()).getStats());
					replace(level02Node, subTree);
				}
			}
		}
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.LearningTree#setDataToLeaf(animatlab.kodamat.core.trees.Leaf,
	 *      animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics,
	 *      ExampleDatabase)
	 */
	@Override
	protected void setDataToLeaf(Leaf<IProbabilities<String>> leaf,
			VariableValueStatistics<String> stats,
			ExampleDatabase<String> exampleDatabase) {
		assert (false);
	}
}
