/*
 * 
 * Created on 24 dec. 2005
 *
 */
package animatlab.kodamat.supervisedlearning.learningtree;

import java.util.List;
import java.util.Map;

import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.DNode;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Vertex;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.supervisedlearning.core.Example;
import animatlab.kodamat.supervisedlearning.core.ExampleDatabase;
import animatlab.kodamat.supervisedlearning.entropy.VariableMeasure;
import animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics;
import animatlab.kodamat.supervisedlearning.entropy.VariableMeasure.VariableMeasureData;

/**
 * A binary decision tree with methods to build it in incremental mode according
 * to Utgoff's ITI algorithm and to prune it virtually according to the minimum
 * description length principle.
 * <P>
 * <I>"The most useful aspect of the ITI algorithm is that it provides a
 * mechanism for incremental tree induction. If one has already constructed a
 * tree, and then obtains a new labeled example, it is possible to present it
 * the algorithm, and have the algorithm revise the tree as necessary. The
 * alternative would be to build a new tree from scratch, based on the now
 * augmented set of labeled examples, which is typically much more expensive.
 * ITI handles symbolic and numeric variables, and missing data values. It
 * includes a virtual pruning mechanism too."</I><BR>
 * Paul E. Utgoff,October 1997, Machine Learning Journal.
 * <P>
 * <B>Incremental mode</B><BR>
 * <I>"In incremental mode, each training example that is presented is
 * immediately incorporated into the tree, and the tree is then immediately
 * restructured as necessary so that every decision node has its most desired
 * test installed."</I><BR>
 * Paul E. Utgoff, October 1997, Machine Learning Journal.
 * <P>
 * <B>Virtual pruning mechanism</B><BR>
 * The vitual pruning mechanism aims to reduce the classification errors
 * (noise). <I>"An important component of decision tree induction is to avoid
 * overfitting the training data, especially when the data are known to contain
 * attribute or classification error (noise). A variety of methods have come
 * into existence, and the question is which of them is best suited to the
 * incremental induction problem. All of the approachesthat maintain a separate
 * pruning set are oxymoronic for incremental tree induction. For ITI, a
 * suitable approach is based on the minimum description length principle.[...]
 * For all practical purposes, such as classifying examples with the tree, or
 * inspecting the tree by printing it, a virtually pruned tree behaves and
 * appears as though it had been truly pruned."</I><BR>
 * Paul E. Utgoff, October 1997, Machine Learning Journal.
 * 
 * @author Julien Burlet and Thomas Degris
 * 
 * 
 * @version 1.0
 */
public class ItiTree extends LearningTree<String, IProbabilities<String>> {
	/**
	 * 
	 * Constructor
	 * 
	 * @param variableMeasure
	 *          measure to use to select the best test
	 * @param variableOrdering
	 *          variable ordering to use with the examples
	 * 
	 */
	public ItiTree(VariableMeasure<String> variableMeasure,
			VariableOrdering variableOrdering) {
		super(variableMeasure, variableOrdering);
	}

	/**
	 * @see animatlab.kodamat.core.trees.Tree#newInstance()
	 */
	@Override
	public ItiTree newInstance() {
		return new ItiTree(variableMeasure, variableOrdering);
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.LearningTree#newDNode(animatlab.kodamat.core.trees.DNode,
	 *      java.lang.String, animatlab.kodamat.core.utils.Variable)
	 */
	@Override
	protected ItiDNode newDNode(DNode<IProbabilities<String>> parent,
			String value, Variable testedVariable) {
		return new ItiDNode(this, parent, value, testedVariable);
	}

	/**
	 * 
	 * Incorporates a training example into this tree and restructures the tree as
	 * needed.
	 * <P>
	 * 
	 * <I>"When an example is to be incorporated into an empty tree, the tree is
	 * replaced by a leaf node that indicates the class of the leaf, and the
	 * example is saved at the leaf node. Whenever an example is to be
	 * incorporated, the branches of the tree are followed as far as possible
	 * according to the values in the example. If the example has the same class
	 * as the leaf, the example is simply added to the set of the examples saved
	 * at the leaf node. If the example has a different class label from the leaf,
	 * the algorithm attempts to turn the leaf into a decision node, picking the
	 * best test according to the test-selection metric. The examples saved at the
	 * node that was just converted from a leaf node to a decision node are then
	 * incorporated recursively by sending each one down its proper branch
	 * according to the new test."</I><BR>
	 * Paul Utgoff, October 1997, Machine Learning Journal.
	 * 
	 * @param example
	 *          Example to learn
	 * 
	 */
	@Override
	public void learn(Example<String> example) {
		learn(getRoot(), example);
		ensureBestTest(getRoot());
	}

	/**
	 * 
	 * Transpose the decision tree
	 * 
	 * @param dnode
	 *          the decision node where to transpose the tree
	 * @param bestTest
	 *          test to install at the node
	 * 
	 */
	private void transposeTree(ItiDNode dnode, VariableMeasureData bestTest) {
		ItiDNode newDNode = installDNode(dnode, bestTest);
		assert (newDNode == dnode);
		distributeExamplesToChildren(dnode);
	}

	/**
	 * 
	 * Recursively ensures that the best test is installed at each decision node
	 * for this tree and all subtrees.
	 * 
	 * @param node
	 *          the root of the subtree to ensure that best tests are installed
	 * 
	 */
	private void ensureBestTest(Vertex<IProbabilities<String>> node) {
		// if node is a decision node and node is stale
		if (!node.isLeaf) {
			ItiDNode dnode = (ItiDNode) node;
			if (!dnode.isFresh()) {
				// find best_test for this node
				VariableMeasureData bestTest = getBestTest(dnode);
				// if best_test is not already installed
				if ((bestTest != null)
						&& !dnode.getTestedVariable().equals(bestTest.variable)) {
					// transpose_tree(node, best_test)
					transposeTree(dnode, bestTest);
				}
				// mark node fresh
				dnode.setFresh(true);
				// ensure_best_test(node->left)
				ensureBestTest(dnode.left());
				// ensure_best_test(node->right)
				ensureBestTest(dnode.right());
			}
		}
	}

	/**
	 * Recursively ensures that the best test is installed at each decision node
	 * for this tree and all subtrees.
	 * 
	 * @param node
	 *          the root of the node in the tree to add the example
	 * @param example
	 *          example to learn in the tree
	 * 
	 */
	@SuppressWarnings("unchecked")
	@Override
	protected void learn(Vertex<IProbabilities<String>> node,
			Example<String> example) {
		ILearningVertex<String> learningVertex = (ILearningVertex<String>) node;
		// if node is NULL then convert node to an empty leaf
		if (node.isLeaf) {
			LearningLeaf<String, IProbabilities<String>> learningLeaf = (LearningLeaf<String, IProbabilities<String>>) node;
			if (learningLeaf.get() == null) {
				setDataToLeaf(learningLeaf, learningLeaf.getStats(), learningLeaf
						.getExampleDatabase());
			}
		}

		// We add the new example in the example database recording
		// all the examples and the number of times they have occured
		learningVertex.addExample(example);

		// if node is a leaf
		if (node.isLeaf) {
			LearningLeaf<String, IProbabilities<String>> leaf = (LearningLeaf<String, IProbabilities<String>>) node;
			// save example at node
			// (already done)
			// if node should be converted to a decision node
			VariableMeasureData bestTest = getBestTest(leaf);
			if (bestTest != null) {
				// construct information at node
				ItiDNode dnode = installDNode(node, bestTest);
				// mark node fresh
				dnode.setFresh(true);
				// for each example saved at node
				distributeExamplesToChildren(dnode);
			}
		} else {
			ItiDNode dnode = (ItiDNode) node;
			// update information at node
			// (already done)
			// mark node stale
			dnode.setFresh(false);
			// if test is true
			if (testIsTrue(dnode, example.values)) {
				learn(dnode.left(), example);
			} else {
				learn(dnode.right(), example);
			}
		}
	}

	/**
	 * 
	 * Distribute the examples of a database to a list of children
	 * 
	 * @param dnode
	 *          the node to get the examples from
	 * 
	 */
	private void distributeExamplesToChildren(ItiDNode dnode) {
		for (Map.Entry<Example<String>, Integer> entry : dnode.getExampleDatabase()
				.entrySet()) {
			for (int index = 0; index < entry.getValue(); index++) {
				// if test is true
				if (testIsTrue(dnode, entry.getKey().values)) {
					learn(dnode.left(), entry.getKey());
				} else {
					learn(dnode.right(), entry.getKey());
				}
			}
		}
	}

	/**
	 * 
	 * Install a test at a leaf
	 * 
	 * @param node
	 *          the leaf where to install the test
	 * @param test
	 *          test to install at the node
	 * 
	 * @return the new decision node installed instead at the leaf
	 * 
	 */
	@SuppressWarnings("unchecked")
	private ItiDNode installDNode(Vertex<IProbabilities<String>> node,
			VariableMeasureData test) {
		ItiDNode dnode = (ItiDNode) node.replace(test.variable);
		if (node.isLeaf) {
			dnode.setData((LearningLeaf<String, IProbabilities<String>>) node);
		}
		dnode.selectedValue = dnode.getStats().getValueSet(test.variable)
				.iterator().next();
		dnode.addLeaf(dnode.selectedValue);
		dnode.addLeaf(DNode.VALUE_WILDCARD);
		return dnode;
	}

	/**
	 * 
	 * Test if the test installed at this node returns true for an exemple
	 * 
	 * @param dnode
	 *          the node to test the test
	 * @param values
	 *          the example to test
	 * 
	 * @return true if the test is true against the example
	 * 
	 */
	private boolean testIsTrue(ItiDNode dnode, List<String> values) {
		return dnode.selectedValue.equals(values.get(variableOrdering.indexOf(dnode
				.getTestedVariable())));
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.LearningTree#setDataToLeaf(animatlab.kodamat.core.trees.Leaf,
	 *      animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics,
	 *      ExampleDatabase)
	 */
	@Override
	protected void setDataToLeaf(Leaf<IProbabilities<String>> leaf,
			VariableValueStatistics<String> stats,
			ExampleDatabase<String> exampleDatabase) {
		leaf.set(stats);
	}

	/**
	 * 
	 * Get the best test to install for a learning vertex
	 * 
	 * @param learningVertex
	 *          the vertex for which to select the best test
	 * @return the best test associated to the vertex
	 * 
	 */
	@Override
	protected VariableMeasureData getBestTest(
			ILearningVertex<String> learningVertex) {
		return variableMeasure.getBestTest(learningVertex.getExampleDatabase(),
				learningVertex.getStats(), null, null);
	}

}
