package animatlab.kodamat.supervisedlearning.learningtree;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import animatlab.kodamat.core.trees.DNode;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.supervisedlearning.core.Example;
import animatlab.kodamat.supervisedlearning.core.ExampleDatabase;
import animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics;
import animatlab.kodamat.supervisedlearning.entropy.VariableMeasure.VariableMeasureData;

/**
 * 
 * Learning leaf
 * 
 * @author Thomas Degris
 * @param <T>
 *          type of the examples to learn
 * @param <U>
 *          type contained at the leaves of the tree
 * 
 */
public class LearningLeaf<T, U> extends Leaf<U> implements ILearningVertex<T> {
	/**
	 * 
	 * Stats maintained at this node
	 * 
	 */
	private VariableValueStatistics<T> stats = null;

	/**
	 * 
	 * Database of the examples learned
	 * 
	 */
	private ExampleDatabase<T> exampleDatabase = null;

	/**
	 * 
	 * Constructor
	 * 
	 * @param tree
	 *          the tree owning this decision node
	 * @param value
	 *          value of the tested variable
	 * @param parent
	 *          Parent in the tree of this node
	 * 
	 */
	public LearningLeaf(Tree<U> tree, DNode<U> parent, String value) {
		super(tree, parent, value);
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.ILearningVertex#getExampleDatabase()
	 */
	final public ExampleDatabase<T> getExampleDatabase() {
		if (exampleDatabase == null) {
			exampleDatabase = new ExampleDatabase<T>();
		}
		return exampleDatabase;
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.ILearningVertex#getStats()
	 */
	final public VariableValueStatistics<T> getStats() {
		if (stats == null) {
			stats = new VariableValueStatistics<T>();
		}
		return stats;
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.ILearningVertex#addExample(animatlab.kodamat.supervisedlearning.core.Example)
	 */
	public void addExample(Example<T> example) {
		getExampleDatabase().add(example);
		getStats().add(getTree().getVariableOrdering(), example);
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.ILearningVertex#setData(animatlab.kodamat.supervisedlearning.learningtree.ILearningVertex)
	 */
	public void setData(ILearningVertex<T> learningVertex) {
		assert (stats == null);
		stats = learningVertex.getStats();
		assert (exampleDatabase == null);
		exampleDatabase = learningVertex.getExampleDatabase();
	}

	/**
	 * @see animatlab.kodamat.core.trees.Vertex#getTree()
	 */
	@SuppressWarnings("unchecked")
	@Override
	public LearningTree<T, U> getTree() {
		return (LearningTree<T, U>) super.getTree();
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.ILearningVertex#setStats(animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics)
	 */
	public void setStats(VariableValueStatistics<T> stats) {
		this.stats = stats;
	}

	/**
	 * @see animatlab.kodamat.core.io.IToXml#toXml(org.w3c.dom.Document,
	 *      animatlab.kodamat.core.io.IToXml.IToXmlPolicy)
	 */
	@Override
	public Element toXml(Document document, IToXmlPolicy<U> policy) {
		Element result = super.toXml(document, policy);
		VariableMeasureData measure = getTree().getMeasure().getBestMeasure(
				exampleDatabase, getStats(), getContext().keySet());
		if (measure != null) {
			result.setAttribute(ATTRIBUTE_MEASURE, String.valueOf(measure.measure));
		}
		result.setAttribute(NB_EXAMPLES_MEASURE, String.valueOf(getStats()
				.getNumberOfExamples()));
		return result;
	}
}
