package animatlab.kodamat.supervisedlearning.learningtree;

import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.proba.ILearningProbabilities;
import animatlab.kodamat.core.trees.ClonePolicy;
import animatlab.kodamat.core.trees.DNode;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.trees.Vertex;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.supervisedlearning.core.Example;
import animatlab.kodamat.supervisedlearning.core.ExampleDatabase;
import animatlab.kodamat.supervisedlearning.core.ISupervisedLearning;
import animatlab.kodamat.supervisedlearning.entropy.VariableMeasure;
import animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics;
import animatlab.kodamat.supervisedlearning.entropy.VariableMeasure.VariableMeasureData;

/**
 * 
 * Abstract class used for induction and regression tree
 * 
 * @author Thomas Degris
 * @param <T>
 *          type of the examples to learn
 * @param <U>
 *          type contained at the leaves of the tree
 * 
 */
public abstract class LearningTree<T, U> extends Tree<U> implements
		ISupervisedLearning<T> {
	/**
	 * 
	 * The measure to use to select the best test
	 * 
	 */
	final protected VariableMeasure<T> variableMeasure;

	/**
	 * 
	 * Set for all the variables
	 * 
	 */
	private final Map<Variable, Set<String>> allValuesSets = new LinkedHashMap<Variable, Set<String>>();

	/**
	 * 
	 * Variable ordering of the examples
	 * 
	 */
	protected VariableOrdering variableOrdering = null;

	/**
	 * 
	 * Constructor
	 * 
	 * @param variableMeasure
	 *          measure to use to select the best test
	 * @param variableOrdering
	 *          variable ordering to use with the examples
	 * 
	 */
	@SuppressWarnings("unchecked")
	public LearningTree(VariableMeasure<T> variableMeasure,
			VariableOrdering variableOrdering) {
		super();
		if (variableMeasure != null) {
			this.variableMeasure = (VariableMeasure<T>) variableMeasure.newInstance();
			this.variableMeasure.setValuesSets(allValuesSets);
		} else {
			this.variableMeasure = null;
		}
		this.variableOrdering = variableOrdering;
	}

	/**
	 * 
	 * Set the value set for all variable
	 * 
	 * @param variableValueSet
	 *          the set of values for all variables
	 * 
	 */
	public void setVariableValueSet(Map<Variable, Set<String>> variableValueSet) {
		allValuesSets.putAll(variableValueSet);
	}

	/**
	 * @see animatlab.kodamat.core.trees.Tree#newLeaf(animatlab.kodamat.core.trees.DNode,
	 *      java.lang.String)
	 */
	@Override
	protected Leaf<U> newLeaf(DNode<U> parent, String value) {
		return new LearningLeaf<T, U>(this, parent, value);
	}

	/**
	 * 
	 * Create a new decision node in this tree
	 * 
	 * @param parent
	 *          the parent of this decision node
	 * @param value
	 *          the value of the tested variable to reach this node
	 * @param testedVariable
	 *          the variable tested at the decision node
	 * 
	 * @return the new decision node
	 * 
	 */
	@Override
	protected LearningDNode<T, U> newDNode(DNode<U> parent, String value,
			Variable testedVariable) {
		return new LearningDNode<T, U>(this, parent, value, testedVariable);
	}

	/**
	 * 
	 * Send all the examples of the database to the children of this node
	 * 
	 * @param dnode
	 *          the root of the subtree to send the examples to
	 * @param exampleDatabase
	 *          the database storing the examples
	 * 
	 */
	protected void sendExamplesToLeaf(DNode<U> dnode,
			ExampleDatabase<T> exampleDatabase) {
		for (Map.Entry<Example<T>, Integer> entry : exampleDatabase.entrySet()) {
			for (int index = 0; index < entry.getValue(); index++) {
				Vertex<U> child = getChild(dnode, entry.getKey().values);
				if (child == null) {
					child = dnode.addLeaf(entry.getKey().values.get(variableOrdering
							.indexOf(dnode.getTestedVariable())));
				}
				learn(child, entry.getKey());
			}
		}
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.core.ISupervisedLearning#learn(animatlab.kodamat.supervisedlearning.core.Example)
	 */
	public void learn(Example<T> example) {
		boolean valueSetHasChanged = false;
		for (int variableIndex = 0; variableIndex < example.values.size(); variableIndex++) {
			Variable variable = variableOrdering.get(variableIndex);
			Set<String> valueSet = allValuesSets.get(variable);
			if (valueSet == null) {
				valueSet = new LinkedHashSet<String>();
				allValuesSets.put(variable, valueSet);
			}
			String value = example.values.get(variableIndex);
			if (value != null) {
				int sizeBackup = valueSet.size();
				valueSet.add(value);
				if (valueSet.size() > sizeBackup) {
					valueSetHasChanged = true;
				}
			}
		}

		if (variableMeasure.noEmptyLeaf && valueSetHasChanged && (root != null)) {
			ExampleDatabase<T> exampleDatabase = getExampleDatabase();
			getRoot().remove();
			for (Map.Entry<Example<T>, Integer> entry : exampleDatabase.entrySet()) {
				for (int index = 0; index < entry.getValue(); index++) {
					learn(getRoot(), entry.getKey());
				}
			}
		}

		learn(getRoot(), example);
	}

	/**
	 * 
	 * @return the complete example database of examples learned by this tree
	 * 
	 */
	@SuppressWarnings("unchecked")
	public ExampleDatabase<T> getExampleDatabase() {
		return ((ILearningVertex<T>) getRoot()).getExampleDatabase();
	}

	/**
	 * 
	 * Learn a new example (ID4 tree-update algorithm described in Incremental
	 * Induction of Decision Trees, Utgoff, p4)
	 * 
	 * @param node
	 *          root of the subtree where to learn the example
	 * @param example
	 *          the example to learn
	 * 
	 */
	abstract protected void learn(Vertex<U> node, Example<T> example);

	/**
	 * @see animatlab.kodamat.supervisedlearning.core.ISupervisedLearning#getStats()
	 */
	@SuppressWarnings("unchecked")
	public ILearningProbabilities<T> getStats() {
		return ((ILearningVertex<T>) getRoot()).getStats();
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.core.ISupervisedLearning#flush()
	 */
	public void flush() {
		// Nothing to do by default
	}

	/**
	 * 
	 * Add an exemple to a vertex
	 * 
	 * @param learningVertex
	 *          the vertex where to add the example
	 * @param example
	 *          the example to add
	 * 
	 */
	@SuppressWarnings("unchecked")
	protected void addExample(ILearningVertex<T> learningVertex,
			Example<T> example) {
		learningVertex.addExample(example);
		Vertex<U> vertex = (Vertex<U>) learningVertex;
		if (vertex.isLeaf) {
			setDataToLeaf((Leaf<U>) vertex, learningVertex.getStats(), learningVertex
					.getExampleDatabase());
		}
	}

	/**
	 * 
	 * Set the data at a leaf
	 * 
	 * @param leaf
	 *          the leaf to set the data
	 * @param stats
	 *          the data to set
	 * @param exampleDatabase
	 *          the database of examples
	 * 
	 */
	abstract protected void setDataToLeaf(Leaf<U> leaf,
			VariableValueStatistics<T> stats, ExampleDatabase<T> exampleDatabase);

	/**
	 * 
	 * Get the best test to install for a learning vertex
	 * 
	 * @param learningVertex
	 *          the vertex for which to select the best test
	 * @return the best test associated to the vertex
	 * 
	 */
	@SuppressWarnings("unchecked")
	protected VariableMeasureData getBestTest(ILearningVertex<T> learningVertex) {
		Variable currentTestedVariable = null;
		if (!((Vertex<U>) learningVertex).isLeaf) {
			currentTestedVariable = ((DNode<U>) learningVertex).getTestedVariable();
		}
		return variableMeasure.getBestTest(learningVertex.getExampleDatabase(),
				learningVertex.getStats(), learningVertex.getContext().keySet(),
				currentTestedVariable);
	}

	/**
	 * @see animatlab.kodamat.core.utils.Prototype#newInstance()
	 */
	@Override
	abstract public LearningTree<T, U> newInstance();

	/**
	 * 
	 * @return the measure to select the best test
	 * 
	 */
	public VariableMeasure<T> getMeasure() {
		return variableMeasure;
	}

	/**
	 * @see animatlab.kodamat.core.trees.Tree#newClonedTree()
	 */
	@Override
	protected Tree<U> newClonedTree() {
		return new ClonedTree<T, U>(getMeasure(), variableOrdering);
	}

	/**
	 * @see animatlab.kodamat.core.trees.Tree#clone(animatlab.kodamat.core.trees.ClonePolicy,
	 *      animatlab.kodamat.core.utils.Context)
	 */
	@SuppressWarnings("unchecked")
	@Override
	public Tree<U> clone(ClonePolicy<U> clonePolicy, Context testsList) {
		ClonedTree<T, U> result = (ClonedTree<T, U>) super.clone(clonePolicy,
				testsList);
		ExampleDatabase<T> clonedDatabase = getExampleDatabase();
		if ((testsList != null) && (clonedDatabase != null)) {
			clonedDatabase = new ExampleDatabase<T>();
			for (Map.Entry<Example<T>, Integer> entry : getExampleDatabase()
					.entrySet()) {
				if (Context.satisfy(variableOrdering, testsList, entry.getKey().values)) {
					clonedDatabase.put(entry.getKey(), entry.getValue());
				}
			}
		}
		result.setExampleDatabase(clonedDatabase);
		return result;
	}

	/**
	 * 
	 * Expand a decision node by creating a leaf for all the possible values of
	 * the tested variable
	 * 
	 * @param dnode
	 *          the decision node to expand
	 * 
	 */
	protected void expandDNode(DBLearningDNode<T, U> dnode) {
		for (String value : allValuesSets.get(dnode.getTestedVariable())) {
			setDataToLeaf(dnode.addLeaf(value), null, null);
		}
	}

	/**
	 * 
	 * @return returns the variableOrdering.
	 * 
	 */
	protected VariableOrdering getVariableOrdering() {
		return variableOrdering;
	}

	/**
	 * 
	 * Set variableOrdering
	 * 
	 * @param variableOrdering
	 *          the variableOrdering to set.
	 * 
	 */
	public void setVariableOrdering(VariableOrdering variableOrdering) {
		this.variableOrdering = variableOrdering;
	}

	/**
	 * 
	 * Get the child of a decision node
	 * 
	 * @param dnode
	 *          the decision node to get the child from
	 * @param values
	 *          the example used to select the child
	 * @return the child satisfying the example
	 * 
	 */
	final protected Vertex<U> getChild(DNode<U> dnode, List<String> values) {
		Variable testedVariable = dnode.getTestedVariable();
		int variableIndex = variableOrdering.indexOf(testedVariable);
		Vertex<U> child = dnode.getChild(values.get(variableIndex));
		return child;
	}

	/**
	 * 
	 * Check if an example is valid to reach this vertex
	 * 
	 * @param vertex
	 *          check if a vertex satisfy a list of values
	 * @param values
	 *          the example to test with
	 * 
	 * @return true if the example is valid
	 * 
	 */
	final public boolean testIsTrue(Vertex<U> vertex, List<String> values) {
		if (vertex.getParent() == null) {
			return true;
		}
		return vertex.getTestedValue().equals(
				values.get(variableOrdering.indexOf(vertex.getParent()
						.getTestedVariable())));
	}

}
