package animatlab.kodamat.supervisedlearning.learningtree;

import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Vertex;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.supervisedlearning.core.Example;
import animatlab.kodamat.supervisedlearning.core.ExampleDatabase;
import animatlab.kodamat.supervisedlearning.entropy.VariableMeasure;
import animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics;
import animatlab.kodamat.supervisedlearning.entropy.VariableMeasure.VariableMeasureData;

/**
 * 
 * Stochastic version of the ID4 algorithm
 * 
 * @author Thomas Degris
 * @param <T>
 *          type of the examples to learn
 * @param <U>
 *          type contained at the leaves of the tree
 * 
 */
public abstract class SID4Tree<T, U> extends ID4Tree<T, U> {
	/**
	 * 
	 * Probability version of the ID3 algorithm
	 * 
	 * @author Thomas Degris
	 * @param <V>
	 *          type of the examples to learn
	 * 
	 */
	static public class Prob<V> extends SID4Tree<V, IProbabilities<V>> {
		/**
		 * 
		 * Constructor
		 * 
		 * @param variableMeasure
		 *          prototype of the measure used to select the test to install
		 * @param variableOrdering
		 *          variable ordering to use with the examples
		 * 
		 */
		public Prob(VariableMeasure<V> variableMeasure,
				VariableOrdering variableOrdering) {
			super(variableMeasure, variableOrdering);
		}

		/**
		 * 
		 * Set the data at a leaf
		 * 
		 * @param leaf
		 *          the leaf to set the data
		 * @param stats
		 *          the data to set
		 * @param exampleDatabase
		 *          database of examples
		 * 
		 */
		@Override
		protected void setDataToLeaf(Leaf<IProbabilities<V>> leaf,
				VariableValueStatistics<V> stats, ExampleDatabase<V> exampleDatabase) {
			leaf.set(stats);
		}

		/**
		 * @see animatlab.kodamat.core.utils.Prototype#newInstance()
		 */
		@Override
		public SID4Tree<V, IProbabilities<V>> newInstance() {
			return new Prob<V>(variableMeasure, variableOrdering);
		}
	}

	/**
	 * 
	 * Plain version of the ID3 algorithm
	 * 
	 * @author Thomas Degris
	 * @param <V>
	 *          type of the examples to learn
	 * 
	 */
	static public class Plain<V> extends SID4Tree<V, V> {
		/**
		 * 
		 * Constructor
		 * 
		 * @param variableMeasure
		 *          prototype of the measure used to select the test to install
		 * @param variableOrdering
		 *          legend associated the variables used in this tree
		 * 
		 */
		public Plain(VariableMeasure<V> variableMeasure,
				VariableOrdering variableOrdering) {
			super(variableMeasure, variableOrdering);
		}

		/**
		 * 
		 * Set the data at a leaf
		 * 
		 * @param leaf
		 *          the leaf to set the data
		 * @param stats
		 *          the data to set
		 * @param exampleDatabase
		 *          database of examples
		 * 
		 */
		@Override
		protected void setDataToLeaf(Leaf<V> leaf,
				VariableValueStatistics<V> stats, ExampleDatabase<V> exampleDatabase) {
			leaf.set(variableMeasure.aggregate(stats, exampleDatabase, leaf
					.getContext().keySet()));
		}

		/**
		 * @see animatlab.kodamat.core.utils.Prototype#newInstance()
		 */
		@Override
		public SID4Tree<V, V> newInstance() {
			return new Plain<V>(variableMeasure, variableOrdering);
		}
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param variableMeasure
	 *          measure to use to select the best test
	 * @param variableOrdering
	 *          variable ordering to use with the examples
	 * 
	 */
	public SID4Tree(VariableMeasure<T> variableMeasure,
			VariableOrdering variableOrdering) {
		super(variableMeasure, variableOrdering);
	}

	/**
	 * 
	 * Learn a new example (ID4 tree-update algorithm described in Incremental
	 * Induction of Decision Trees, Utgoff, p4)
	 * 
	 * @param node
	 *          root of the subtree where to learn the example
	 * @param example
	 *          the example to learn
	 * 
	 */
	@SuppressWarnings("unchecked")
	@Override
	protected void learn(Vertex<U> node, Example<T> example) {
		ILearningVertex<T> learningVertex = (ILearningVertex<T>) node;
		// 1. For each possible test attribute at the current node, update the
		// count of positive or negative instances for the value of that attribute
		// in the training instance.
		addExample(learningVertex, example);
		VariableMeasureData bestTest = getBestTest(learningVertex);
		// 2. If all the instances observed at the current node are positive
		// (negative), then the decision tree at the current node is an answer
		// node containing a ‘‘+’’ (‘‘-’’) to indicate a positive (negative)
		// instance.
		if ((variableMeasure.isPure(learningVertex.getStats(), learningVertex
				.getExampleDatabase(), learningVertex.getContext().keySet()) && node.isLeaf)
				|| (node.isLeaf && (bestTest == null))) {
			return;
		}
		// 3. Otherwise,
		DBLearningDNode<T, U> dnode;
		// (a) If the current node is an answer node, then change it to a
		// decision node containing an attribute test with the lowest Escore.
		if ((node.isLeaf) && (bestTest != null)) {
			LearningLeaf<T, U> leaf = (LearningLeaf<T, U>) node;
			dnode = (DBLearningDNode<T, U>) leaf.replace(bestTest.variable);
			dnode.setData(leaf);
			expandDNode(dnode);
			sendExamplesToLeaf(dnode, dnode.getExampleDatabase());
			return;
		}
		// (pre-b) If the test is not valid anymore and this decision node is the
		// parent of leaves, then we transform the decision node into a leaf
		dnode = (DBLearningDNode<T, U>) node;
		if (bestTest == null) {
			if (canBeRemoved(dnode)) {
				LearningLeaf<T, U> leaf = (LearningLeaf<T, U>) dnode.replaceWithLeaf();
				leaf.setData(dnode);
				setDataToLeaf(leaf, leaf.getStats(), leaf.getExampleDatabase());
			}
			return;
		}
		// (b) Otherwise, if the current decision node contains an attribute
		// test that does not have the lowest E-score, then
		// i. Change the attribute test to one with the lowest E-score.
		// ii. Discard all existing subtrees below the decision node.
		if (!bestTest.variable.equals(dnode.getTestedVariable())) {
			dnode.replace(bestTest.variable);
			expandDNode(dnode);
			sendExamplesToLeaf(dnode, dnode.getExampleDatabase());
			return;
		}
		// (c) Recursively update the decision tree below the current decision
		// node along the branch of the value of the current test
		// attribute that occurs in the instance description. Grow the
		// branch if necessary.
		Vertex<U> child = getChild(dnode, example.values);
		if (child == null) {
			child = dnode.addLeaf(example.values.get(variableOrdering.indexOf(dnode
					.getTestedVariable())));
		}
		learn(child, example);
	}

	/**
	 * 
	 * Check if a decision node can be removed
	 * 
	 * @param node
	 *          the node to check
	 * 
	 * @return true if the decision can removed
	 * 
	 */
	private boolean canBeRemoved(DBLearningDNode<T, U> node) {
		// for (Vertex<U> child : node.children()) {
		// if (!child.isLeaf) {
		// return false;
		// }
		// }
		return true;
	}
}
