/*
 * 
 * Created on Dec 28, 2005
 *
 */
package animatlab.kodamat.supervisedlearning.learningtree;

import java.util.Collection;

import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.MergeLeavesPolicy;
import animatlab.kodamat.core.trees.Vertex;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.supervisedlearning.core.Example;
import animatlab.kodamat.supervisedlearning.core.ExampleDatabase;
import animatlab.kodamat.supervisedlearning.entropy.VariableMeasure;
import animatlab.kodamat.supervisedlearning.entropy.VariableMeasure.VariableMeasureData;

/**
 * 
 * Decision tree to classify different distributions of a value
 * 
 * @author Thomas Degris
 * 
 */
public class SID5RTree extends ID5RTree {
	/**
	 * 
	 * Default merge policy
	 * 
	 */
	static protected class DefaultMergeLeavesPolicy implements
			MergeLeavesPolicy<IProbabilities<String>> {
		/**
		 * 
		 * Merge policy
		 * 
		 * @param vertexSet
		 *          set of leaves to merge
		 * @return the selected leaves where all have been merged
		 * 
		 */
		@SuppressWarnings("unchecked")
		public Vertex<IProbabilities<String>> merge(
				Collection<Vertex<IProbabilities<String>>> vertexSet) {
			for (Vertex<IProbabilities<String>> vertex : vertexSet) {
				if (!vertex.isLeaf) {
					return null;
				}
			}
			LearningLeaf<String, IProbabilities<String>> selected = (LearningLeaf<String, IProbabilities<String>>) vertexSet
					.iterator().next();
			for (Vertex<IProbabilities<String>> vertex : vertexSet) {
				if (vertex != selected) {
					LearningLeaf<String, IProbabilities<String>> leaf = (LearningLeaf<String, IProbabilities<String>>) vertex;
					selected.getExampleDatabase().merge(leaf.getExampleDatabase());
					selected.getStats().merge(leaf.getStats());
				}
			}
			return selected;
		}
	}

	/**
	 * 
	 * Policy used to simplify learning trees
	 * 
	 */
	static final private DefaultMergeLeavesPolicy mergePolicy = new DefaultMergeLeavesPolicy();

	/**
	 * 
	 * Constructor
	 * 
	 * @param variableMeasure
	 *          measure to use to select the best test
	 * @param variableOrdering
	 *          legend associated the variables used in this tree
	 * 
	 */
	public SID5RTree(VariableMeasure<String> variableMeasure,
			VariableOrdering variableOrdering) {
		super(variableMeasure, variableOrdering);
	}

	/**
	 * @see animatlab.kodamat.core.trees.Tree#newInstance()
	 */
	@Override
	public SID5RTree newInstance() {
		return new SID5RTree(variableMeasure, variableOrdering);
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.core.ISupervisedLearning#learn(animatlab.kodamat.supervisedlearning.core.Example)
	 */
	@Override
	public void learn(Example<String> example) {
		learn(getRoot(), example);
		simplify();
		assert (checkLearning(getRoot()));
	}

	/**
	 * 
	 * Check of subtree of this tree
	 * 
	 * @param vertex
	 *          root of the subtree to check
	 * 
	 * @return check the current tree
	 * 
	 */
	private boolean checkLearning(Vertex<IProbabilities<String>> vertex) {
		if (vertex.isLeaf) {
			return true;
		}
		ID5DNode dnode = (ID5DNode) vertex;
		assert (dnode.children().size() > 1);
		for (Vertex<IProbabilities<String>> child : dnode.children()) {
			if (!checkLearning(child)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.ID5RTree#learn(animatlab.kodamat.core.trees.Vertex,
	 *      animatlab.kodamat.supervisedlearning.core.Example)
	 */
	@SuppressWarnings("unchecked")
	@Override
	protected void learn(Vertex<IProbabilities<String>> node,
			Example<String> example) {
		if (node.isLeaf) {
			LearningLeaf<String, IProbabilities<String>> leaf = (LearningLeaf<String, IProbabilities<String>>) node;
			if (leaf.get() == null) {
				leaf.set(((ILearningVertex<String>) leaf).getStats());
			}
			leaf.addExample(example);
			VariableMeasureData bestTest = null;
			if (leaf.get().getValueSet().size() > 1) {
				bestTest = getBestTest(leaf);
			}
			if (bestTest != null) {
				ExampleDatabase<String> exampleDatabase = leaf.getExampleDatabase();
				ID5DNode dnode = expand(leaf, bestTest.variable, leaf.getStats()
						.getValueSet(bestTest.variable));
				dnode.setFresh(false);
				sendExamplesToLeaf(dnode, exampleDatabase);
			}
		} else {
			ID5DNode dnode = (ID5DNode) node;
			dnode.addExample(example);
			Vertex<IProbabilities<String>> child = getChild(dnode, example.values);
			if (child == null) {
				child = dnode.addLeaf(example.values.get(variableOrdering.indexOf(dnode
						.getTestedVariable())));
			}
			learn(child, example);
			installBestTest(node);
		}
	}

	/**
	 * 
	 * Check if the best test is installed at this node and its children
	 * 
	 * @param vertex
	 *          root of the subtree to check
	 * 
	 */
	@Override
	protected void installBestTest(Vertex<IProbabilities<String>> vertex) {
		if (vertex.isLeaf) {
			return;
		}
		ID5DNode dnode = (ID5DNode) vertex;
		VariableMeasureData bestTest = getBestTest(dnode);
		if ((bestTest != null)
				&& !bestTest.variable.equals(dnode.getTestedVariable())) {
			pullUp(dnode, bestTest.variable);
		}
	}

	/**
	 * 
	 * Simplify this tree using a policy to compare the data at the leaves
	 * 
	 */
	public void simplify() {
		getRoot().simplify(null, mergePolicy);
	}
}
