package animatlab.kodamat.supervisedlearning.regression;

import java.util.Set;

import animatlab.kodamat.core.proba.ILearningProbabilities;
import animatlab.kodamat.core.utils.Prototype;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.supervisedlearning.core.ExampleDatabase;
import animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics;

/**
 * 
 * Variable measure using Chi2 tests to compute the measure
 * 
 * @author Thomas Degris
 * 
 */
public class LeastSquareMeasure extends RegressionMeasure {
	/**
	 * 
	 * Default threshold regression value
	 * 
	 */
	final static public double DEFAULT_THRESHOLD_VALUE = 8.0;

	/**
	 * 
	 * Constructor
	 * 
	 */
	public LeastSquareMeasure() {
		super(null);
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param threshold
	 * 
	 */
	public LeastSquareMeasure(Double threshold) {
		super(threshold);
	}

	/**
	 * @see animatlab.kodamat.core.utils.Prototype#newInstance()
	 */
	public Prototype newInstance() {
		LeastSquareMeasure result = new LeastSquareMeasure();
		result.set(this);
		return result;
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.regression.RegressionMeasure#computeValue(Variable,
	 *      String, ILearningProbabilities, ExampleDatabase)
	 */
	@Override
	protected double computeValue(Variable variable, String variableValue,
			ILearningProbabilities<Double> classNamesCounter,
			ExampleDatabase<Double> exampleDatabase) {
		double result = 0;
		int counter = 0;
		for (Double value : classNamesCounter.getValueSet()) {
			int weight = classNamesCounter.getNumberOfExamples(value);
			result = result + (weight * value);
			counter = counter + weight;
		}
		return result / counter;
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.regression.RegressionMeasure#computeError(Variable,
	 *      String, double, ILearningProbabilities, ExampleDatabase)
	 */
	@Override
	public double computeError(Variable variable, String variableValue,
			double average, ILearningProbabilities<Double> classNamesCounter,
			ExampleDatabase<Double> exampleDatabase) {
		double result = 0;
		int counter = 0;
		for (Double value : classNamesCounter.getValueSet()) {
			int weight = classNamesCounter.getNumberOfExamples(value);
			result = result + weight * ((value - average) * (value - average));
			counter = counter + weight;
		}
		return result / counter;
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.regression.RegressionMeasure#isPure(animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics,
	 *      animatlab.kodamat.supervisedlearning.core.ExampleDatabase,
	 *      java.util.Set)
	 */
	@Override
	public boolean isPure(
			VariableValueStatistics<Double> variableValueStatistics,
			ExampleDatabase<Double> exampleDatabase, Set<Variable> ignoredVariables) {
		if (threshold == null) {
			return super.isPure(variableValueStatistics, exampleDatabase,
					ignoredVariables);
		}
		double minValue = Double.MAX_VALUE;
		double maxValue = -Double.MAX_VALUE;
		double average = 0;
		int sampleSize = 0;
		for (Double value : variableValueStatistics.getValueSet()) {
			minValue = Math.min(minValue, value);
			maxValue = Math.max(maxValue, value);
			int counter = variableValueStatistics.getNumberOfExamples(value);
			sampleSize += counter;
			average += counter * value;
		}
		average = average / sampleSize;
		return ((maxValue - minValue) <= threshold);
	}

}
