package animatlab.kodamat.supervisedlearning.regression;

import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.proba.ILearningProbabilities;
import animatlab.kodamat.core.proba.ValueStatistics;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.supervisedlearning.core.ExampleDatabase;
import animatlab.kodamat.supervisedlearning.entropy.VariableMeasure;
import animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics;

/**
 * 
 * Measure used in regression tree
 * 
 * @author Thomas Degris
 * 
 */
public abstract class RegressionMeasure extends VariableMeasure<Double> {
	/**
	 * 
	 * Global error associated with the current stats
	 * 
	 */
	protected double globalError;

	/**
	 * 
	 * Global value associated with the current stats
	 * 
	 */
	private double globalValue;

	/**
	 * 
	 * The maximum value computed by this measure
	 * 
	 */
	private double max;

	/**
	 * 
	 * Constructor
	 * 
	 * @param threshold
	 *          threshold value
	 * 
	 */
	public RegressionMeasure(Double threshold) {
		super(threshold);
	}

	/**
	 * 
	 * Compute some data used by the measure
	 * 
	 * @param variableValueStatistics
	 *          the stats to compute the measure with
	 * @param exampleDatabase
	 *          database example
	 * @param ignoredVariables
	 *          the set of variable to ignore
	 */
	@Override
	protected void computeCachedData(
			VariableValueStatistics<Double> variableValueStatistics,
			ExampleDatabase<Double> exampleDatabase, Set<Variable> ignoredVariables) {
		computeCachedDataVariableValue(null, null, variableValueStatistics,
				exampleDatabase, ignoredVariables);
		globalValue = computeValue(null, null, variableValueStatistics,
				exampleDatabase);
		globalError = computeError(null, null, globalValue,
				variableValueStatistics, exampleDatabase);
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.entropy.VariableMeasure#computeMeasure(animatlab.kodamat.core.utils.Variable,
	 *      animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics,
	 *      ExampleDatabase, Set)
	 */
	@Override
	protected double computeMeasure(Variable variable,
			VariableValueStatistics<Double> stats,
			ExampleDatabase<Double> exampleDatabase, Set<Variable> ignoredVariables) {
		double result = 0;
		double minError = Double.MAX_VALUE;
		for (Map.Entry<String, ValueStatistics<Double>> entry : stats
				.getValueStats(variable).entrySet()) {
			final ValueStatistics<Double> valueStats = entry.getValue();
			computeCachedDataVariableValue(variable, entry.getKey(), stats,
					exampleDatabase, ignoredVariables);
			double average = computeValue(variable, entry.getKey(), valueStats,
					exampleDatabase);
			double error = computeError(variable, entry.getKey(), average,
					valueStats, exampleDatabase);
			minError = Math.min(minError, error);
			double weight = (double) valueStats.getNumberOfExamples()
					/ stats.getNumberOfExamples();
			result = result + weight * error;
		}
		return globalError - result;
	}

	/**
	 * 
	 * Compute cached data for a variable and a specific value for this variable
	 * 
	 * @param variable
	 *          the variable for which the cached data are computed
	 * @param value
	 *          the value of the variable for which the cached data are computed
	 * @param stats
	 *          the stats to compute the measure with
	 * @param exampleDatabase
	 *          the dabatase of examples
	 * @param ignoredVariables
	 *          the set of ignored variables
	 * 
	 */
	protected void computeCachedDataVariableValue(Variable variable,
			String value, VariableValueStatistics<Double> stats,
			ExampleDatabase<Double> exampleDatabase, Set<Variable> ignoredVariables) {
		// Nothing to compute by default
	}

	/**
	 * 
	 * Compute the current average
	 * 
	 * @param variable
	 *          variable for which to compute the measure
	 * @param variableValue
	 *          the value of the variable
	 * @param stats
	 *          counter to count the error on
	 * @param exampleDatabase
	 *          database of examples
	 * @return the average computed
	 * 
	 */
	abstract protected double computeValue(Variable variable,
			String variableValue, ILearningProbabilities<Double> stats,
			ExampleDatabase<Double> exampleDatabase);

	/**
	 * 
	 * Compute the current average
	 * 
	 * @param variable
	 *          variable for which to compute the measure
	 * @param variableValue
	 *          the value of the variable
	 * @param average
	 *          average to compute the fitting error with
	 * @param stats
	 *          counter to count the error on
	 * @param exampleDatabase
	 *          database of examples
	 * @return the least square fitting error
	 * 
	 */
	abstract public double computeError(Variable variable, String variableValue,
			double average, ILearningProbabilities<Double> stats,
			ExampleDatabase<Double> exampleDatabase);

	/**
	 * 
	 * The stats used to test if they are pure or not
	 * 
	 * @param variableValueStatistics
	 *          the stats to compute the measure with
	 * @param exampleDatabase
	 *          database of examples
	 * @param ignoredVariables
	 *          variables already set
	 * 
	 * @return true if all the examples at this vertex are from the same class
	 * 
	 */
	@Override
	public boolean isPure(
			VariableValueStatistics<Double> variableValueStatistics,
			ExampleDatabase<Double> exampleDatabase, Set<Variable> ignoredVariables) {
		if (threshold == null) {
			return super.isPure(variableValueStatistics, exampleDatabase,
					ignoredVariables);
		}
		computeCachedData(variableValueStatistics, exampleDatabase,
				ignoredVariables);
		return (globalError <= threshold);
		// double minValue = Double.MAX_VALUE;
		// double maxValue = -Double.MAX_VALUE;
		// double average = 0;
		// int sampleSize = 0;
		// for (Double value : variableValueStatistics.getValueSet()) {
		// minValue = Math.min(minValue, value);
		// maxValue = Math.max(maxValue, value);
		// int counter = variableValueStatistics.getNumberOfExamples(value);
		// sampleSize += counter;
		// average += counter * value;
		// }
		// average = average / sampleSize;
		// return ((maxValue - minValue) <= threshold);
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.entropy.VariableMeasure#getStatsError(animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics,
	 *      animatlab.kodamat.supervisedlearning.core.ExampleDatabase,
	 *      java.util.Set)
	 */
	@Override
	public double getStatsError(
			VariableValueStatistics<Double> variableValueStatistics,
			ExampleDatabase<Double> exampleDatabase, Set<Variable> ignoredVariables) {
		computeCachedData(variableValueStatistics, exampleDatabase,
				ignoredVariables);
		return globalError;
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.entropy.VariableMeasure#isSignificant(animatlab.kodamat.supervisedlearning.entropy.VariableMeasure.VariableMeasureData,
	 *      animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics,
	 *      animatlab.kodamat.supervisedlearning.core.ExampleDatabase,
	 *      java.util.Set, animatlab.kodamat.core.utils.Variable)
	 */
	@Override
	protected boolean isSignificant(VariableMeasureData bestTest,
			VariableValueStatistics<Double> variableValueStatistics,
			ExampleDatabase<Double> exampleDatabase, Set<Variable> ignoredVariables,
			Variable currentTest) {
		if (Double.isNaN(bestTest.measure)) {
			return false;
		}
		if (threshold == null) {
			return true;
		}
		return (bestTest.measure > threshold);
	}

	/**
	 * 
	 * Aggregate a set of statistic to only one value
	 * 
	 * @param variableValueStatistics
	 *          the stats to aggregate
	 * @param exampleDatabase
	 *          database of examples
	 * @param ignoredVariables
	 *          variables to ignore to compute the aggregation
	 * @return the value representing the stats
	 * 
	 */
	@Override
	final public Double aggregate(
			VariableValueStatistics<Double> variableValueStatistics,
			ExampleDatabase<Double> exampleDatabase, Set<Variable> ignoredVariables) {
		Double value = computeAggregate(variableValueStatistics, exampleDatabase,
				ignoredVariables);
		if ((value != null) && (!Double.isNaN(value))) {
			max = Math.max(max, value);
		}
		return value;
	}

	/**
	 * 
	 * Aggregate a set of statistic to only one value
	 * 
	 * @param variableValueStatistics
	 *          the stats to aggregate
	 * @param exampleDatabase
	 *          database of examples
	 * @param ignoredVariables
	 *          variables to ignore to compute the aggregation
	 * @return the value representing the stats
	 * 
	 */
	protected Double computeAggregate(
			VariableValueStatistics<Double> variableValueStatistics,
			ExampleDatabase<Double> exampleDatabase, Set<Variable> ignoredVariables) {
		if (variableValueStatistics == null) {
			return null;
		}
		return computeValue(null, null, variableValueStatistics, exampleDatabase);
	}

	/**
	 * 
	 * Build a string representing the stats
	 * 
	 * @param variableValueStatistics
	 *          the stats to represent in string
	 * @return a string representing
	 * 
	 */
	@Override
	public String toString(VariableValueStatistics<Double> variableValueStatistics) {
		double average = computeValue(null, null, variableValueStatistics, null);
		double error = computeError(null, null, average, variableValueStatistics,
				null);
		return (average + "±" + error + "x" + variableValueStatistics
				.getNumberOfExamples());
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.entropy.VariableMeasure#infoToDotString(animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics,
	 *      animatlab.kodamat.supervisedlearning.core.ExampleDatabase,
	 *      java.util.Set, animatlab.kodamat.core.utils.VariableOrdering)
	 */
	@Override
	public String infoToDotString(VariableValueStatistics<Double> stats,
			ExampleDatabase<Double> exampleDatabase, Set<Variable> ignoredVariable,
			VariableOrdering variableOrdering) {
		computeCachedData(stats, exampleDatabase, ignoredVariable);
		double average = computeValue(null, null, stats, exampleDatabase);
		double error = computeError(null, null, average, stats, exampleDatabase);
		return super.infoToDotString(stats, exampleDatabase, ignoredVariable,
				variableOrdering)
				+ (Parameters.format(average) + ">" + error + "<x" + stats
						.getNumberOfExamples());
	}

	/**
	 * 
	 * @return the maximum value computed by this measure
	 * 
	 */
	public double getMax() {
		return max;
	}
}
