package animatlab.kodamat.supervisedlearning.regression;

import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.proba.ILearningProbabilities;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Prototype;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.supervisedlearning.core.Example;
import animatlab.kodamat.supervisedlearning.core.ExampleDatabase;
import animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics;

/**
 * 
 * Measure used for to explore the structure of a problem
 * 
 * @author Thomas Degris
 * 
 */
public class StructureLearningMeasure extends LeastSquareMeasure {
	/**
	 * 
	 * Beta parameter to use when the example is unknown
	 * 
	 */
	private double betaU;

	/**
	 * 
	 * Weight for unknown states
	 * 
	 */
	final private double nU;

	/**
	 * 
	 * Size of the space represented by this leaf
	 * 
	 */
	private long restrictedSpaceSize;

	/**
	 * 
	 * Constructor
	 * 
	 * @param betau
	 *          beta parameter to use when the example is unknown
	 * @param nU
	 *          weight for unknown states
	 * 
	 * 
	 */
	public StructureLearningMeasure(double betau, double nU) {
		this(betau, nU, null);
	}

	/**
	 * 
	 * Constructor with a threshold
	 * 
	 * @param betau
	 *          beta parameter to use when the example is unknown
	 * @param nU
	 *          weight for unknown states
	 * @param epsilon
	 *          theshold to use
	 * 
	 */
	public StructureLearningMeasure(double betau, double nU, Double epsilon) {
		super(epsilon);
		this.betaU = betau;
		this.nU = nU;
		noEmptyLeaf = true;
	}

	/**
	 * @see animatlab.kodamat.core.utils.Prototype#newInstance()
	 */
	@Override
	public Prototype newInstance() {
		return new StructureLearningMeasure(betaU, nU, getThreshold());
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.entropy.VariableMeasure#checkValueSet(animatlab.kodamat.core.utils.Variable,
	 *      java.util.Set)
	 */
	@Override
	protected boolean checkValueSet(Variable variable, Set<String> valueSet) {
		if (allValueSets.get(variable).size() > 1) {
			return true;
		}
		return false;
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.regression.RegressionMeasure#computeCachedDataVariableValue(animatlab.kodamat.core.utils.Variable,
	 *      java.lang.String,
	 *      animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics,
	 *      animatlab.kodamat.supervisedlearning.core.ExampleDatabase, Set)
	 */
	@Override
	protected void computeCachedDataVariableValue(Variable variable,
			String value, VariableValueStatistics<Double> stats,
			ExampleDatabase<Double> exampleDatabase, Set<Variable> ignoredVariables) {
		if (variable != null) {
			return;
		}
		restrictedSpaceSize = Context.getRestrictedStateSize(allValueSets,
				ignoredVariables);
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.regression.RegressionMeasure#computeAggregate(animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics,
	 *      animatlab.kodamat.supervisedlearning.core.ExampleDatabase,
	 *      java.util.Set)
	 */
	@Override
	public Double computeAggregate(
			VariableValueStatistics<Double> variableValueStatistics,
			ExampleDatabase<Double> exampleDatabase, Set<Variable> ignoredVariables) {
		if (variableValueStatistics == null) {
			return betaU;
		}
		computeCachedDataVariableValue(null, null, variableValueStatistics,
				exampleDatabase, ignoredVariables);
		return computeValue(null, null, variableValueStatistics, exampleDatabase);
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.regression.RegressionMeasure#computeValue(Variable,
	 *      String, ILearningProbabilities, ExampleDatabase)
	 */
	@Override
	protected double computeValue(Variable variable, String variableValue,
			ILearningProbabilities<Double> classNamesCounter,
			ExampleDatabase<Double> exampleDatabase) {
		double result = 0;
		int counter = 0;
		for (Double value : classNamesCounter.getValueSet()) {
			int weight = classNamesCounter.getNumberOfExamples(value);
			result = result + weight * value;
			counter = counter + weight;
		}
		double unknownSpaceSize = (restrictedSpaceSize - exampleDatabase.size())
				* nU;
		return (result + (unknownSpaceSize * betaU)) / (counter + unknownSpaceSize);
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.regression.RegressionMeasure#computeError(Variable,
	 *      String, double, ILearningProbabilities, ExampleDatabase)
	 */
	@Override
	public double computeError(Variable variable, String variableValue,
			double average, ILearningProbabilities<Double> classNamesCounter,
			ExampleDatabase<Double> exampleDatabase) {
		double result = 0;
		int counter = 0;
		for (Double value : classNamesCounter.getValueSet()) {
			int weight = classNamesCounter.getNumberOfExamples(value);
			result = result + weight * ((value - average) * (value - average));
			counter = counter + weight;
		}
		double unknownSpaceSize = (restrictedSpaceSize - exampleDatabase.size())
				* nU;
		return (result + (((betaU - average) * (betaU - average)) * unknownSpaceSize))
				/ (counter + unknownSpaceSize);
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.regression.RegressionMeasure#isPure(animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics,
	 *      ExampleDatabase, Set)
	 */
	@Override
	public boolean isPure(
			VariableValueStatistics<Double> variableValueStatistics,
			ExampleDatabase<Double> exampleDatabase, Set<Variable> ignoredVariables) {
		computeCachedDataVariableValue(null, null, variableValueStatistics,
				exampleDatabase, ignoredVariables);
		if ((betaU > 0) && (restrictedSpaceSize > exampleDatabase.size())) {
			return false;
		}
		if (threshold == null) {
			double value = Double.NaN;
			int count = 0;
			for (Map.Entry<Example<Double>, Integer> entry : exampleDatabase
					.entrySet()) {
				if (Double.isNaN(value)) {
					value = entry.getKey().className;
					count = entry.getValue();
				}
				if ((value != entry.getKey().className) || (count != entry.getValue())) {
					return false;
				}
			}
			return true;
		}
		return super.isPure(variableValueStatistics, exampleDatabase,
				ignoredVariables);
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.regression.RegressionMeasure#isSignificant(animatlab.kodamat.supervisedlearning.entropy.VariableMeasure.VariableMeasureData,
	 *      animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics,
	 *      animatlab.kodamat.supervisedlearning.core.ExampleDatabase,
	 *      java.util.Set, animatlab.kodamat.core.utils.Variable)
	 */
	@Override
	protected boolean isSignificant(VariableMeasureData bestTest,
			VariableValueStatistics<Double> variableValueStatistics,
			ExampleDatabase<Double> exampleDatabase, Set<Variable> ignoredVariables,
			Variable currentTest) {
		return (globalError > threshold);
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.entropy.VariableMeasure#getName()
	 */
	@Override
	public String getName() {
		String name = Parameters.format(betaU);
		name = name.replaceAll("\\.", "");
		name = name.replaceAll(",", "");
		// if (threshold == null) {
		// name = name + "tn";
		// }
		return "strlrn" + name;
	}

	/**
	 * 
	 * @return returns the betaU parameter
	 * 
	 */
	public double getBetaU() {
		return betaU;
	}

	/**
	 * 
	 * Set betaU
	 * 
	 * @param betaU
	 *          set the betaU parameter
	 * 
	 */
	public void setBetaU(double betaU) {
		this.betaU = betaU;
	}

}
