/*
 * 
 * Created on Dec 28, 2005
 *
 */
package animatlab.kodamat.supervisedlearning.rmax;

import java.util.Collection;

import animatlab.kodamat.core.trees.DNode;
import animatlab.kodamat.core.trees.EqualPolicy;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.MergeLeavesPolicy;
import animatlab.kodamat.core.trees.Vertex;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.supervisedlearning.core.Example;
import animatlab.kodamat.supervisedlearning.core.ExampleDatabase;
import animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics;
import animatlab.kodamat.supervisedlearning.learningtree.LearningDNode;
import animatlab.kodamat.supervisedlearning.learningtree.LearningLeaf;
import animatlab.kodamat.supervisedlearning.learningtree.LearningTree;
import animatlab.kodamat.supervisedlearning.regression.LeastSquareMeasure;

/**
 * 
 * Tree learning specific examples and returning a default value for examples
 * that are not considered as known FIXME The tree is not simplified
 * 
 * @author Thomas Degris
 * 
 */
public class RMaxTree extends LearningTree<Double, Double> {
	/**
	 * 
	 * Default merge policy
	 * 
	 */
	static protected class DefaultMergeLeavesPolicy implements
			MergeLeavesPolicy<Double> {
		/**
		 * 
		 * Merge policy
		 * 
		 * @param vertexSet
		 *          set of leaves to merge
		 * @return the selected leaves where all have been merged
		 * 
		 */
		public Vertex<Double> merge(Collection<Vertex<Double>> vertexSet) {
			if (vertexSet.size() < 2) {
				return null;
			}
			for (Vertex<Double> vertex : vertexSet) {
				if (!vertex.isLeaf) {
					return null;
				}
			}
			RMaxLeaf selected = (RMaxLeaf) vertexSet.iterator().next();
			for (Vertex<Double> vertex : vertexSet) {
				if (vertex != selected) {
					RMaxLeaf leaf = (RMaxLeaf) vertex;
					selected.merge(leaf);
				}
			}
			return selected;
		}
	}

	/**
	 * 
	 * Reward maximum get by the agent
	 * 
	 */
	double rewardMaximum = -Double.MAX_VALUE;

	/**
	 * 
	 * RMax value used for unknown state/action
	 * 
	 */
	double rmax;

	/**
	 * 
	 * Statistics databases used for decision nodes
	 * 
	 */
	final VariableValueStatistics<Double> mainStatsDatabasePrototype = new VariableValueStatistics<Double>();

	/**
	 * 
	 * Number of try to consider that a node is known
	 * 
	 */
	final int numberOfTry;

	/**
	 * 
	 * Default value of this tree
	 * 
	 */
	Double defaultValue = 0.0;

	/**
	 * 
	 * Constructor
	 * 
	 * @param numberOfTry
	 *          number of try before to consider that the example is known
	 * @param variableOrdering
	 *          variable ordering to use to build example
	 * 
	 */
	public RMaxTree(int numberOfTry, VariableOrdering variableOrdering) {
		super(new LeastSquareMeasure(null), variableOrdering);
		this.numberOfTry = numberOfTry;
		this.variableOrdering = variableOrdering;
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.core.ISupervisedLearning#learn(animatlab.kodamat.supervisedlearning.core.Example)
	 */
	@Override
	public void learn(Example<Double> example) {
		mainStatsDatabasePrototype.add(variableOrdering, example);
		learn(getRoot(), example);
		simplify();
		double currentReward = example.className;
		if (currentReward > rewardMaximum) {
			rewardMaximum = currentReward;
			rmax = Parameters.getRMaxFromReward(rewardMaximum);
		}
	}

	/**
	 * @see animatlab.kodamat.core.trees.Tree#newInstance()
	 */
	@Override
	public RMaxTree newInstance() {
		return new RMaxTree(numberOfTry, variableOrdering);
	}

	/**
	 * 
	 * Create all the possible children for a node
	 * 
	 * @param vertex
	 *          the node to expand
	 * 
	 * @return the node expanded
	 * 
	 */
	@SuppressWarnings("unchecked")
	private LearningDNode<Double, Double> createChildren(Vertex<Double> vertex) {
		LearningDNode<Double, Double> result;
		// We find the variable to expand the value from
		if (vertex.isLeaf) {
			int variableIndex = 0;
			Variable selectedVariable = null;
			while (selectedVariable == null) {
				Variable variable = variableOrdering.get(variableIndex);
				if (!vertex.getContext().containsKey(variable)) {
					selectedVariable = variable;
				}
				variableIndex++;
			}
			result = (LearningDNode<Double, Double>) vertex.replace(selectedVariable);
		} else {
			result = (LearningDNode<Double, Double>) vertex;
		}
		// We expand the node with all the values we know
		for (String value : mainStatsDatabasePrototype.getValueSet(result
				.getTestedVariable())) {
			if (result.getChild(value) == null) {
				result.addLeaf(value);
			}
		}
		return result;
	}

	/**
	 * 
	 * Learn an example in a subtree
	 * 
	 * @param vertex
	 *          the root of the subtree
	 * @param example
	 *          the example to learn
	 * 
	 */
	@SuppressWarnings("unchecked")
	@Override
	protected void learn(Vertex<Double> vertex, Example<Double> example) {
		if (vertex.isLeaf) {
			// Test if the reward known is correct
			LearningLeaf<Double, Double> leaf = (LearningLeaf<Double, Double>) vertex;
			// Test if the tree is already expanded then the Rmax value needs to be
			// updated
			if (leaf.getContext().size() == example.values.size()
					|| (leaf.getStats().getNumberOfExamples() > numberOfTry)) {
				leaf.addExample(example);
				return;
			}
		}
		// Build the children if necessary
		Vertex<Double> child = null;
		if (!vertex.isLeaf) {
			((RMaxDNode) vertex).setFresh(false);
			child = getChild((DNode<Double>) vertex, example.values);
		}
		if (child == null) {
			child = createChildren(vertex);
		}
		// Continue to expand in the children
		learn(child, example);
	}

	/**
	 * 
	 * Create a new decision node in this tree
	 * 
	 * @param parent
	 *          the parent of this decision node
	 * @param value
	 *          the value of the tested variable to reach this node
	 * @param testedVariable
	 *          the variable tested at the decision node
	 * 
	 * @return the new decision node
	 * 
	 */
	@Override
	protected LearningDNode<Double, Double> newDNode(DNode<Double> parent,
			String value, Variable testedVariable) {
		return new RMaxDNode(this, parent, value, testedVariable);
	}

	/**
	 * 
	 * Create a new leaf in this tree
	 * 
	 * @param parent
	 *          the parent of this leaf
	 * @param value
	 *          the value of the tested variable to reach this node
	 * 
	 * @return the new leaf
	 * 
	 */
	@Override
	protected LearningLeaf<Double, Double> newLeaf(DNode<Double> parent,
			String value) {
		return new RMaxLeaf(this, parent, value);
	}

	/**
	 * 
	 * Simplify this tree
	 * 
	 */
	public void simplify() {
		getRoot().simplify(new EqualPolicy<Double>() {
			public boolean equals(Leaf<Double> leaf01, Leaf<Double> leaf02) {
				RMaxLeaf rmaxLeaf01 = (RMaxLeaf) leaf01;
				RMaxLeaf rmaxLeaf02 = (RMaxLeaf) leaf02;
				if ((rmaxLeaf01.getStats().getNumberOfExamples() < numberOfTry)
						|| (rmaxLeaf02.getStats().getNumberOfExamples() < numberOfTry)) {
					return false;
				}
				return leaf01.get().equals(leaf02.get());
			}
		}, new DefaultMergeLeavesPolicy());
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.LearningTree#setDataToLeaf(animatlab.kodamat.core.trees.Leaf,
	 *      animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics,
	 *      ExampleDatabase)
	 */
	@Override
	protected void setDataToLeaf(Leaf<Double> leaf,
			VariableValueStatistics<Double> stats,
			ExampleDatabase<Double> exampleDatabase) {
		// Nothing to do by default
	}
}
