/*
 * 
 * Created on Aug 9, 2005
 *
 */
package animatlab.kodamat.supervisedlearning.test;

import static animatlab.kodamat.core.io.XmlUtils.getRoot;
import static animatlab.kodamat.core.utils.Context.createStateSet;
import static animatlab.kodamat.core.utils.Context.extractList;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import junit.framework.TestCase;
import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.io.DataLoadingPolicy;
import animatlab.kodamat.core.io.Log;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.rules.PlainRules;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Chrono;
import animatlab.kodamat.core.utils.Prototype;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.supervisedlearning.core.Example;
import animatlab.kodamat.supervisedlearning.core.ISupervisedLearning;
import animatlab.kodamat.supervisedlearning.entropy.Chi2Measure;
import animatlab.kodamat.supervisedlearning.entropy.GainRatioMeasure;
import animatlab.kodamat.supervisedlearning.entropy.KSMeasure;
import animatlab.kodamat.supervisedlearning.entropy.VariableMeasure;
import animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics;
import animatlab.kodamat.supervisedlearning.entropy.VariableMeasure.VariableMeasureData;
import animatlab.kodamat.supervisedlearning.learningtree.ID3Tree;
import animatlab.kodamat.supervisedlearning.learningtree.ID4Tree;
import animatlab.kodamat.supervisedlearning.learningtree.ID5RTree;
import animatlab.kodamat.supervisedlearning.learningtree.ItiTree;
import animatlab.kodamat.supervisedlearning.learningtree.SID4Tree;
import animatlab.kodamat.supervisedlearning.learningtree.SID5RTree;
import animatlab.kodamat.supervisedlearning.regression.LeastSquareMeasure;
import animatlab.kodamat.supervisedlearning.utils.PartitionsToTree;

/**
 * 
 * JUnit test on the learning methods of incremental induction of decision tree
 * 
 * @author Thomas Degris
 * 
 */
public class TestDeterministicLearningTree extends TestCase {
	/**
	 * 
	 * Maximum number of variables to use to create examples set
	 * 
	 */
	private static final int EXAMPLES_SET_MAX_SIZE = 7;

	/**
	 * 
	 * Legend to use for the trees
	 * 
	 */
	private static final VariableOrdering variableOrdering = VariableOrdering
			.buildNumericVariableOrdering(EXAMPLES_SET_MAX_SIZE);

	/**
	 * 
	 * Randomizer for generating examples
	 * 
	 */
	private Random random = new Random(0);

	/**
	 * 
	 * Make the tree learn a set of examples and then classify the examples
	 * learned to check whether the tree has correctly learned or not
	 * 
	 * @param learningTree
	 *          the learningTree to check
	 * @param examples
	 *          a set of examples to test
	 * 
	 */
	@SuppressWarnings("unchecked")
	private void learnAndClassifyExamples(
			ISupervisedLearning<String> learningTree, Example<String>[] examples) {
		IFunction<String> fun = (IFunction<String>) learningTree;
		for (int numberOfExamplesLearned = 0; numberOfExamplesLearned < examples.length; numberOfExamplesLearned++) {
			learningTree.learn(examples[numberOfExamplesLearned]);
			learningTree.flush();
			for (int index = 0; index <= numberOfExamplesLearned; index++) {
				Object data = fun.classify(variableOrdering
						.listToState(examples[index].values));
				if (data instanceof IProbabilities) {
					assertEquals(examples[index].className, ((IProbabilities) data)
							.getBest());
				} else {
					assertEquals(examples[index].className, data);
				}
			}
		}
	}

	/**
	 * Test different measures on the same algorithme
	 * 
	 * @param treeToTest
	 *          the learning tree to test
	 * 
	 */
	@SuppressWarnings("unchecked")
	private void testMeasureOnTree(Class treeToTest) {
		Chrono chrono = new Chrono();
		Prototype learningTree;
		learningTree = createTree(treeToTest, new GainRatioMeasure());
		testExamplesCaseOnTree(learningTree);
		learningTree = createTree(treeToTest, new KSMeasure());
		testExamplesCaseOnTree(learningTree);
		learningTree = createTree(treeToTest, new Chi2Measure(null));
		testExamplesCaseOnTree(learningTree);
		Log.sysout.out().println(
				treeToTest.toString() + ": " + chrono.getCurrentChrono());
	}

	/**
	 * Create a full set of examples from a number of variables
	 * 
	 * @param random
	 *          random generator used to generate the class of the examples
	 * @param stateSize
	 *          the number of variables
	 * @param valueSet
	 *          value set of the states
	 * 
	 * @return the set of examples
	 */
	@SuppressWarnings("unchecked")
	static protected Example<String>[] createExamplesSet(Random random,
			int stateSize, Context.ValueSet valueSet) {
		Context[] states = createStateSet(stateSize, valueSet);
		Example<String>[] examples = new Example[states.length];
		for (int index = 0; index < states.length; index++) {
			String className;
			switch (random.nextInt(4)) {
			case 0:
				className = "-";
				break;
			case 1:
				className = "+";
				break;
			case 2:
				className = "/";
				break;
			default:
				className = "*";
				break;
			}
			examples[index] = new Example<String>(className, variableOrdering
					.stateToList(states[index]));
		}
		return examples;
	}

	/**
	 * Test different set of examples on a instance of a learning tree
	 * 
	 * @param learningTree
	 *          the learning tree to test
	 * 
	 */
	@SuppressWarnings("unchecked")
	private void testExamplesCaseOnTree(Prototype learningTree) {
		Example[] examples02 = {
				new Example<String>("-", extractList("short, dark, brown")),
				new Example<String>("+", extractList("short, blond, blue")),
				new Example<String>("-", extractList("short, blond, brown")),
				new Example<String>("+", extractList("short, dark, blue")) };
		learnAndClassifyExamples((ISupervisedLearning<String>) learningTree
				.newInstance(), examples02);

		Example[] examples03 = {
				new Example<String>("-", extractList("short, blond, brown")),
				new Example<String>("-", extractList("tall, dark, brown")),
				new Example<String>("+", extractList("tall, blond, blue")),
				new Example<String>("-", extractList("tall, dark, blue")),
				new Example<String>("-", extractList("short, dark, blue")),
				new Example<String>("+", extractList("tall, red, blue")),
				new Example<String>("-", extractList("tall, blond, brown")),
				new Example<String>("+", extractList("short, blond, blue")) };
		learnAndClassifyExamples((ISupervisedLearning<String>) learningTree
				.newInstance(), examples03);

		Example[] examples04 = { new Example<String>("b", extractList("0, 0, 0")),
				new Example<String>("a", extractList("0, 0, 1")),
				new Example<String>("a", extractList("0, 1, 0")),
				new Example<String>("b", extractList("0, 1, 1")),
				new Example<String>("a", extractList("1, 0, 0")),
				new Example<String>("b", extractList("1, 0, 1")),
				new Example<String>("b", extractList("1, 1, 0")),
				new Example<String>("a", extractList("1, 1, 1")) };
		learnAndClassifyExamples((ISupervisedLearning<String>) learningTree
				.newInstance(), examples04);

		for (int index = 1; index < EXAMPLES_SET_MAX_SIZE; index = index + 2) {
			learnAndClassifyExamples((ISupervisedLearning<String>) learningTree
					.newInstance(), createExamplesSet(random, index,
					Context.ValueSet.BINARY));
		}
		for (int index = 1; index < EXAMPLES_SET_MAX_SIZE / 2; index = index + 2) {
			learnAndClassifyExamples((ISupervisedLearning<String>) learningTree
					.newInstance(), createExamplesSet(random, index,
					Context.ValueSet.OCTAL));
		}
	}

	/**
	 * Create a learning tree with a specific measure
	 * 
	 * @param treeToTest
	 *          the tree to test
	 * @param measure
	 *          the measure to assign the tree
	 * 
	 * @return the new instance of the tree
	 * 
	 */
	@SuppressWarnings("unchecked")
	private Prototype createTree(Class treeToTest, VariableMeasure measure) {
		Class[] constructorSignature = { VariableMeasure.class,
				VariableOrdering.class };
		Object[] initargs = { measure, variableOrdering };
		try {
			Constructor constructor = treeToTest.getConstructor(constructorSignature);
			Prototype learningTree = (Prototype) constructor.newInstance(initargs);
			return learningTree;
		} catch (SecurityException e) {
			e.printStackTrace();
		} catch (NoSuchMethodException e) {
			e.printStackTrace();
		} catch (IllegalArgumentException e) {
			e.printStackTrace();
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		} catch (InvocationTargetException e) {
			e.printStackTrace();
		}
		return null;
	}

	/**
	 * 
	 * Test the Chi2 value to separate distributions that don't have the same
	 * values in it
	 * 
	 */
	public void testChiSquareValueWithMissingValueInDistribution() {
		VariableValueStatistics<String> stats = new VariableValueStatistics<String>();
		stats.add(variableOrdering, new Example<String>("b",
				extractList("false, false")));
		stats.add(variableOrdering, new Example<String>("b",
				extractList("false, true")));
		stats.add(variableOrdering, new Example<String>("a",
				extractList("true, false")));
		stats.add(variableOrdering, new Example<String>("a",
				extractList("true, true")));
		Chi2Measure measure = new Chi2Measure(0.0);
		VariableMeasureData bestTest = measure.getBestTest(null, stats, null, null);
		assertEquals(variableOrdering.get(0), bestTest.variable);
	}

	/**
	 * 
	 * Test the ID3 algorithm
	 * 
	 */
	public void testID3() {
		testMeasureOnTree(ID3Tree.Plain.class);
	}

	/**
	 * 
	 * Test the ID4 algorithm
	 * 
	 */
	public void testID4() {
		testMeasureOnTree(ID4Tree.Plain.class);
	}

	/**
	 * 
	 * Test the SID4 algorithm
	 * 
	 */
	public void testSID4() {
		testMeasureOnTree(SID4Tree.Plain.class);
	}

	/**
	 * 
	 * Test the ID5R algorithm
	 * 
	 */
	public void testID5R() {
		testMeasureOnTree(ID5RTree.class);
	}

	/**
	 * 
	 * Test the ID5 algorithm
	 * 
	 */
	public void testSID5R() {
		testMeasureOnTree(SID5RTree.class);
	}

	/**
	 * 
	 * Test the simplification in SID5
	 * 
	 */
	@SuppressWarnings("unchecked")
	public void testSID5RSimplification() {
		SID5RTree learningTree = new SID5RTree(new Chi2Measure(0.0),
				variableOrdering);
		Example[] examples = {
				new Example<String>("-", extractList("short, blond, brown")),
				new Example<String>("-", extractList("tall, dark, brown")),
				new Example<String>("+", extractList("tall, blond, blue")),
				new Example<String>("-", extractList("tall, dark, blue")),
				new Example<String>("-", extractList("short, dark, blue")),
				new Example<String>("+", extractList("tall, red, blue")),
				new Example<String>("-", extractList("tall, blond, brown")),
				new Example<String>("+", extractList("short, blond, blue")) };
		learnAndClassifyExamples(learningTree, examples);
		int numberOfNodesBeforeSimplification = learningTree.numberOfNodes();
		learningTree.simplify();
		int numberOfNodesAfterSimplification = learningTree.numberOfNodes();
		assertEquals(numberOfNodesAfterSimplification,
				numberOfNodesBeforeSimplification);
	}

	/**
	 * 
	 * Test the ITI algorithm
	 * 
	 */
	public void testIti() {
		testMeasureOnTree(ItiTree.class);
	}

	/**
	 * 
	 * Test the xor operator with decision tree
	 * 
	 */
	public void testXorOperator() {
		Chi2Measure measure = new Chi2Measure(7.0);
		ItiTree ititree = new ItiTree(measure, variableOrdering);
		for (int numberOfTrials = 0; numberOfTrials < 50; numberOfTrials++) {
			boolean value01 = false;
			boolean value02 = false;
			switch (numberOfTrials % 4) {
			case 0:
				break;
			case 1:
				value01 = false;
				value02 = true;
				break;
			case 2:
				value01 = true;
				value02 = false;
				break;
			case 3:
				value01 = true;
				value02 = true;
				break;
			default:
				assert (false);
			}
			boolean className = value02 ^ value01;
			ArrayList<String> state = new ArrayList<String>();
			state.add(String.valueOf(value01));
			state.add(String.valueOf(value02));
			state.add("");
			for (int numberOfExamples = 0; numberOfExamples < 6; numberOfExamples++) {
				state.set(state.size() - 1, String.valueOf(random.nextBoolean()));
				ititree.learn(new Example<String>(String.valueOf(className), state));
			}
		}
		assert (ititree.numberOfLeaves() == 4);
	}

	/**
	 * 
	 * Test tree reorganization
	 * 
	 */
	public void testReorganisationWhenNoReorganisation() {
		final String initial = "<dtr:dnode variable='4'>"
				+ "  <dtr:dnode value='0' variable='2'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:leaf value='1' data='and'/>"
				+ "    <dtr:leaf value='2' data='everybody'/>" + "  </dtr:dnode>"
				+ "  <dtr:dnode value='1' variable='3'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:leaf value='1' data='and'/>"
				+ "    <dtr:leaf value='2' data='everybody'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		reorganizeAndCheck(initial, initial);
	}

	/**
	 * 
	 * Test tree reorganization
	 * 
	 */
	public void testReorganisationWithFullReorganisation() {
		final String initial = "<dtr:dnode variable='4'>"
				+ "  <dtr:dnode value='0' variable='3'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:leaf value='1' data='and'/>"
				+ "    <dtr:leaf value='2' data='everybody'/>" + "  </dtr:dnode>"
				+ "  <dtr:dnode value='1' variable='3'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:leaf value='1' data='and'/>"
				+ "    <dtr:leaf value='2' data='everybody'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		final String expected = "<dtr:dnode value='0' variable='3'>"
				+ "  <dtr:leaf value='0' data='world'/>"
				+ "  <dtr:leaf value='1' data='and'/>"
				+ "  <dtr:leaf value='2' data='everybody'/>" + "</dtr:dnode>";
		reorganizeAndCheck(initial, expected);
	}

	/**
	 * 
	 * Test tree reorganization
	 * 
	 */
	public void testReorganisationWithUsefullReorganisation() {
		final String initial = "<dtr:dnode variable='4'>"
				+ "  <dtr:dnode value='0' variable='3'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:leaf value='1' data='and'/>" + "  </dtr:dnode>"
				+ "  <dtr:dnode value='1' variable='3'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:dnode value='1' variable='2'>"
				+ "      <dtr:leaf value='0' data='hello'/>"
				+ "      <dtr:leaf value='1' data='everybody'/>" + "    </dtr:dnode>"
				+ "  </dtr:dnode>" + "</dtr:dnode>";
		final String expected = "<dtr:dnode variable='3'>"
				+ "  <dtr:leaf value='0' data='world'/>"
				+ "  <dtr:dnode value='1' variable='4'>"
				+ "    <dtr:leaf value='0' data='and'/>"
				+ "    <dtr:dnode value='1' variable='2'>"
				+ "      <dtr:leaf value='0' data='hello'/>"
				+ "      <dtr:leaf value='1' data='everybody'/>" + "    </dtr:dnode>"
				+ "  </dtr:dnode>" + "</dtr:dnode>";
		reorganizeAndCheck(initial, expected);
	}

	/**
	 * 
	 * Test tree reorganization
	 * 
	 */
	public void testReorganisationWithPartialReorganisation() {
		final String initial = "<dtr:dnode variable='2'>"
				+ "  <dtr:dnode value='0' variable='4'>"
				+ "    <dtr:dnode value='0' variable='3'>"
				+ "      <dtr:leaf value='0' data='world'/>"
				+ "      <dtr:leaf value='1' data='and'/>"
				+ "      <dtr:leaf value='2' data='everybody'/>" + "    </dtr:dnode>"
				+ "    <dtr:dnode value='1' variable='3'>"
				+ "      <dtr:leaf value='0' data='world'/>"
				+ "      <dtr:leaf value='1' data='and'/>"
				+ "      <dtr:leaf value='2' data='everybody'/>" + "    </dtr:dnode>"
				+ "  </dtr:dnode>" + "  <dtr:leaf value='1' data='world'/>"
				+ "</dtr:dnode>";
		final String expected = "<dtr:dnode variable='2'>"
				+ "  <dtr:dnode value='0' variable='3'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:leaf value='1' data='and'/>"
				+ "    <dtr:leaf value='2' data='everybody'/>" + "  </dtr:dnode>"
				+ "  <dtr:leaf value='1' data='world'/>" + "</dtr:dnode>";
		reorganizeAndCheck(initial, expected);
	}

	/**
	 * 
	 * Reorganize a tree and check with an expected tree
	 * 
	 * @param initial
	 *          the tree to reorganize
	 * @param expected
	 *          the expected after the reorganization
	 * 
	 */
	private void reorganizeAndCheck(final String initial, final String expected) {
		Tree<String> tree = new Tree<String>(getRoot(initial),
				DataLoadingPolicy.stringLoadingPolicy);
		PartitionsToTree<String> functionToTree = new PartitionsToTree<String>(
				new Chi2Measure(null));
		Tree<String> newTree = functionToTree.build(tree);
		tree = new Tree<String>(getRoot(expected),
				DataLoadingPolicy.stringLoadingPolicy);
		assertEquals(tree, newTree);
		System.out.println();
	}

	/**
	 * 
	 * Compare the ID4 and ID3 algorithm
	 * 
	 * 
	 */
	public void testCompareID4AndID3() {
		Example<String>[] examples = createExamplesSet(new Random(0), 3,
				Context.ValueSet.OCTAL);
		ID4Tree.Plain<String> id4tree = new ID4Tree.Plain<String>(new Chi2Measure(
				null), variableOrdering);
		id4tree.setVariableOrdering(variableOrdering);
		ID3Tree.Plain<String> id3tree = new ID3Tree.Plain<String>(new Chi2Measure(
				null), variableOrdering);
		id3tree.setVariableOrdering(variableOrdering);
		for (Example<String> example : examples) {
			id4tree.learn(example);
			id3tree.learn(example);
		}
		id3tree.flush();
		assertEquals(id3tree, id4tree);
	}

	/**
	 * 
	 * Test when the no empty leaf flag is enable
	 * 
	 */
	@SuppressWarnings("unchecked")
	public void testNoEmptyLeaf() {
		Example[] examples = { new Example<String>("0", extractList("0,0")),
				new Example<String>("1", extractList("0,1")),
				new Example<String>("1", extractList("0,1")),
				new Example<String>("2", extractList("1,0")),
				new Example<String>("2", extractList("1,1")) };
		Chi2Measure measure = new Chi2Measure(null);
		measure.noEmptyLeaf = true;
		ID4Tree.Plain<String> id4tree = new ID4Tree.Plain<String>(measure,
				variableOrdering);
		id4tree.setVariableOrdering(variableOrdering);
		for (Example example : examples) {
			id4tree.learn(example);
		}
		id4tree.learn(new Example<String>("2", extractList("1,2")));
		assertEquals("1", id4tree.classify(Context.getState("0,2", null)));
	}

	/**
	 * 
	 * Test the number of examples required for a split for a given value of Chi2
	 * 
	 */
	public void testChi2Value() {
		final double PROBA_REF = 0.75;
		final int numberOfTests = 50;
		Chi2Measure measure = new Chi2Measure();
		int[] numberOfExamples = new int[numberOfTests];
		for (int index = 0; index < numberOfTests; index++) {
			numberOfExamples[index] = countExamples(PROBA_REF, index, measure);
		}
		double average = 0;
		double min = Double.MAX_VALUE;
		double max = 0;
		for (int number : numberOfExamples) {
			average += number;
			min = Math.min(min, number);
			max = Math.max(max, number);
		}
		System.out.println("The number of examples is " + min + "<"
				+ (average / numberOfTests) + "<" + max + " for Chi2="
				+ measure.getThreshold() + " and a probability of " + PROBA_REF);
	}

	/**
	 * 
	 * Count the number of examples required to split a node
	 * 
	 * @param probaRef
	 *          the probability to use as reference
	 * @param seed
	 *          the seed number to initialize the random instance
	 * @param measure
	 *          the measure used to split the node
	 * @return the number of examples required
	 * 
	 */
	private int countExamples(double probaRef, int seed, Chi2Measure measure) {
		Random random = new Random(seed);
		ID4Tree.Prob<String> sid4tree = new ID4Tree.Prob<String>(measure,
				variableOrdering);
		int numberOfExamples = 0;
		while (sid4tree.numberOfNodes() < 2) {
			numberOfExamples++;
			boolean attrValue = random.nextBoolean();
			List<String> values = new ArrayList<String>();
			values.add(String.valueOf(attrValue));
			double proba = probaRef;
			if (attrValue) {
				proba = 1 - proba;
			}
			String className;
			if (random.nextFloat() < proba) {
				className = "a";
			} else {
				className = "b";
			}
			sid4tree.learn(new Example<String>(className, values));
		}
		return numberOfExamples;
	}

	/**
	 * 
	 * Test RMax tree reorganisation
	 * 
	 * 
	 */
	public void testRMaxTree() {
		PlainRules<Double> rules = new PlainRules<Double>(getRoot("<rdr:rules>"
				+ "  <rdr:rule data='0.0'>" + "    <rdr:cond variable='0' value='0'/>"
				+ "    <rdr:cond variable='1' value='0'/>"
				+ "    <rdr:cond variable='2' value='0'/>"
				+ "    <rdr:cond variable='3' value='0'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='1.0'>" + "    <rdr:cond variable='0' value='0'/>"
				+ "    <rdr:cond variable='1' value='0'/>"
				+ "    <rdr:cond variable='2' value='0'/>"
				+ "    <rdr:cond variable='3' value='1'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='1.0'>" + "    <rdr:cond variable='0' value='0'/>"
				+ "    <rdr:cond variable='1' value='0'/>"
				+ "    <rdr:cond variable='2' value='1'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='0.0'>" + "    <rdr:cond variable='0' value='0'/>"
				+ "    <rdr:cond variable='1' value='1'/>"
				+ "    <rdr:cond variable='2' value='1'/>"
				+ "    <rdr:cond variable='3' value='1'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='1.0'>" + "    <rdr:cond variable='0' value='0'/>"
				+ "    <rdr:cond variable='1' value='1'/>"
				+ "    <rdr:cond variable='2' value='1'/>"
				+ "    <rdr:cond variable='3' value='0'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='1.0'>" + "    <rdr:cond variable='0' value='0'/>"
				+ "    <rdr:cond variable='1' value='1'/>"
				+ "    <rdr:cond variable='2' value='0'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='1.0'>" + "    <rdr:cond variable='0' value='1'/>"
				+ "  </rdr:rule>" + "</rdr:rules>"),
				DataLoadingPolicy.doubleLoadingPolicy);
		PartitionsToTree<Double> partitionsToTree = new PartitionsToTree<Double>(
				new LeastSquareMeasure());
		Tree<Double> reorganised = partitionsToTree.build(rules);
		System.out.println(reorganised.toString());
	}
}
