/*
 * 
 * Created on Aug 9, 2005
 *
 */
package animatlab.kodamat.supervisedlearning.test;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import junit.framework.TestCase;
import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.supervisedlearning.core.Example;
import animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics;
import animatlab.kodamat.supervisedlearning.learningtree.ID4Tree;
import animatlab.kodamat.supervisedlearning.learningtree.LearningTree;
import animatlab.kodamat.supervisedlearning.learningtree.SID4Tree;
import animatlab.kodamat.supervisedlearning.regression.LeastAbsoluteDeviationMeasure;
import animatlab.kodamat.supervisedlearning.regression.LeastSquareMeasure;
import animatlab.kodamat.supervisedlearning.regression.StructureLearningMeasure;
import animatlab.kodamat.supervisedlearning.rmax.RMaxTree;

/**
 * 
 * JUnit test on the learning methods of incremental regression of decision tree
 * 
 * @author Thomas Degris
 * 
 */
public class TestRegressionTree extends TestCase {
	/**
	 * 
	 * Maximum number of variables to use to create examples set
	 * 
	 */
	private static final int EXAMPLES_SET_MAX_SIZE = 3;

	/**
	 * 
	 * Sum used as reference function
	 * 
	 */
	private class Sum implements IFunction<Double> {
		/**
		 * 
		 * Number of accurate variables to use in the sum
		 * 
		 */
		final public int numberOfAccurateVariable;

		/**
		 * 
		 * Constructor
		 * 
		 * @param numberOfAccurateVariable
		 *          number of accurate variables to use in the sum
		 * 
		 */
		protected Sum(int numberOfAccurateVariable) {
			this.numberOfAccurateVariable = numberOfAccurateVariable;
		}

		/**
		 * @see animatlab.kodamat.core.functions.IFunction#classify(Context)
		 */
		public Double classify(Context state) {
			double result = 0;
			for (int variable = 0; variable < numberOfAccurateVariable; variable++) {
				result = result
						+ Double.parseDouble(state.get(Variable.getVariable(String
								.valueOf(variable))));
			}
			return result;
		}

		/**
		 * @see animatlab.kodamat.core.functions.IFunction#numberOfNodes()
		 */
		public int numberOfNodes() {
			assert (false);
			return 0;
		}

		/**
		 * @see animatlab.kodamat.core.functions.IFunction#clone()
		 */
		@Override
		public IFunction<Double> clone() {
			assert (false);
			return null;
		}
	}

	/**
	 * 
	 * The legend to use in the tree
	 * 
	 */
	public VariableOrdering variableOrdering = VariableOrdering
			.buildNumericVariableOrdering(3);

	/**
	 * 
	 * Randomizer to use to select examples
	 * 
	 */
	private Random random = new Random(0);

	/**
	 * 
	 * Test the regression measure
	 * 
	 */
	public void testLeastSquareMeasure() {
		VariableValueStatistics<Double> stats = new VariableValueStatistics<Double>();
		stats.add(variableOrdering, new Example<Double>(1.0, Context
				.extractList("0,0")));
		stats.add(variableOrdering, new Example<Double>(0.0, Context
				.extractList("1,0")));
		LeastSquareMeasure measure = new LeastSquareMeasure(null);
		assertEquals(0.5, measure.aggregate(stats, null, null));
		assertEquals(0.25, measure.getBestTest(null, stats, null, null).measure);
	}

	/**
	 * 
	 * Test the regression measure
	 * 
	 */
	public void testLeastAbsoluteDeviationMeasure() {
		VariableValueStatistics<Double> stats = new VariableValueStatistics<Double>();
		LeastAbsoluteDeviationMeasure measure = new LeastAbsoluteDeviationMeasure(
				null);
		stats.add(variableOrdering, new Example<Double>(1.0, Context
				.extractList("0,0")));

		stats.add(variableOrdering, new Example<Double>(0.0, Context
				.extractList("1,0")));
		assertEquals(0.5, measure.aggregate(stats, null, null));
		assertEquals(0.5, measure.computeError(null, null, measure.aggregate(stats,
				null, null), stats, null));

		stats.add(variableOrdering, new Example<Double>(0.0, Context
				.extractList("1,0")));
		assertEquals(0.0, measure.aggregate(stats, null, null));
		assertEquals(1.0 / 3.0, measure.computeError(null, null, measure.aggregate(
				stats, null, null), stats, null));

		stats.add(variableOrdering, new Example<Double>(1.0, Context
				.extractList("0,0")));
		stats.add(variableOrdering, new Example<Double>(1.0, Context
				.extractList("0,0")));
		assertEquals(1.0, measure.aggregate(stats, null, null));
		assertEquals(2.0 / 5.0, measure.computeError(null, null, measure.aggregate(
				stats, null, null), stats, null));

		stats.add(variableOrdering, new Example<Double>(0.0, Context
				.extractList("1,0")));
		assertEquals(0.5, measure.aggregate(stats, null, null));
		assertEquals(0.5, measure.computeError(null, null, measure.aggregate(stats,
				null, null), stats, null));

	}

	/**
	 * 
	 * Test a regression tree using the least square error measure
	 * 
	 */
	public void testLeastAbsoluteDeviationRegressionTree() {
		testRegressionTree(new ID4Tree.Plain<Double>(
				new LeastAbsoluteDeviationMeasure(null), variableOrdering));
	}

	/**
	 * 
	 * Test a regression tree using the least square error measure
	 * 
	 */
	public void testLeastSquareRegressionTree() {
		testRegressionTree(new ID4Tree.Plain<Double>(new LeastSquareMeasure(null),
				variableOrdering));
	}

	/**
	 * 
	 * Test the regression tree
	 * 
	 * @param tree
	 *          learning algorithm to test
	 * 
	 */
	private void testRegressionTree(LearningTree<Double, Double> tree) {
		Context[] states = Context.createStateSet(EXAMPLES_SET_MAX_SIZE,
				Context.ValueSet.OCTAL);
		for (int accurateVariable = 0; accurateVariable < EXAMPLES_SET_MAX_SIZE; accurateVariable++) {
			Sum reference = new Sum(accurateVariable);
			learnOnReference(states, reference, tree);
		}
	}

	/**
	 * 
	 * Build a regresssion tree using a function as reference
	 * 
	 * @param states
	 *          the set of states to build from
	 * @param reference
	 *          the reference to use to learn from
	 * @param prototype
	 *          learning algorithm to test
	 * 
	 */
	private void learnOnReference(Context[] states, Sum reference,
			LearningTree<Double, Double> prototype) {
		LearningTree<Double, Double> tree = prototype.newInstance();
		tree.setVariableOrdering(variableOrdering);
		List<Example<Double>> examples = new ArrayList<Example<Double>>();
		for (Context state : states) {
			examples.add(new Example<Double>(reference.classify(state),
					variableOrdering.stateToList(state)));
		}
		List<Example<Double>> learnedExamples = new LinkedList<Example<Double>>(
				examples);
		while (!learnedExamples.isEmpty()) {
			tree
					.learn(learnedExamples.remove(random.nextInt(learnedExamples.size())));
		}
		for (Example<Double> example : examples) {
			assertEquals(example.className, tree.classify(variableOrdering
					.listToState(example.values)));
		}
	}

	/**
	 * 
	 * Test the RMax algorithm with simple example
	 * 
	 */
	public void testRMaxWithSimpleExample() {
		testRMaxWithSimpleExample(new RMaxTree(1, variableOrdering));
	}

	/**
	 * 
	 * Test the RMax algorithm with simple example
	 * 
	 */
	public void testRMax2WithSimpleExample() {
		testRMaxWithSimpleExample(new SID4Tree.Plain<Double>(
				new StructureLearningMeasure(1.0, 1), variableOrdering));
	}

	/**
	 * 
	 * Test the RMax algorithms with simple example
	 * 
	 * @param rmaxTree
	 *          the tree to test
	 * 
	 */
	private void testRMaxWithSimpleExample(LearningTree<Double, Double> rmaxTree) {
		rmaxTree
				.learn(new Example<Double>(0.0, Context.extractList("false, false")));
		assertEquals(0.0, rmaxTree.classify(Context.getState("false, false", null)));
		rmaxTree
				.learn(new Example<Double>(1.0, Context.extractList("true, false")));
		assertEquals(1.0, rmaxTree.classify(Context.getState("true, false", null)));
	}

	/**
	 * 
	 * Test the RMax algorithm simplification procedure
	 * 
	 */
	public void testRMaxSimplification() {
		final RMaxTree rmaxTree = new RMaxTree(1, variableOrdering);
		testRMaxSimplification(rmaxTree);
		rmaxTree.learn(new Example<Double>(0.0, Context.extractList("true, true")));
		assertEquals(2, rmaxTree.numberOfLeaves());
	}

	/**
	 * 
	 * Test the RMax algorithm simplification procedure
	 * 
	 */
	public void testRMax2Simplification() {
		final SID4Tree.Plain<Double> rmaxTree = new SID4Tree.Plain<Double>(
				new StructureLearningMeasure(1.0, 1), variableOrdering);
		testRMaxSimplification(rmaxTree);
		Example<Double> example = new Example<Double>(0.0, Context
				.extractList("true, true"));
		rmaxTree.learn(example);
		assertEquals(0.5, rmaxTree.classify(variableOrdering
				.listToState(example.values)));
	}

	/**
	 * 
	 * Test the RMax algorithm simplification procedure
	 * 
	 * @param rmaxTree
	 *          the learning tree to test
	 * 
	 */
	@SuppressWarnings("unchecked")
	private void testRMaxSimplification(LearningTree<Double, Double> rmaxTree) {
		Example[] examples = {
				new Example<Double>(0.0, Context.extractList("false, false")),
				new Example<Double>(0.0, Context.extractList("false, true")),
				new Example<Double>(1.0, Context.extractList("true, false")),
				new Example<Double>(1.0, Context.extractList("true, true")) };
		for (Example<Double> example : examples) {
			rmaxTree.learn(example);
		}
		assertEquals(2, rmaxTree.numberOfLeaves());
		for (Example<Double> example : examples) {
			assertEquals(example.className, rmaxTree.classify(variableOrdering
					.listToState(example.values)));
		}
	}

	/**
	 * 
	 * Test all the RMax algorithm
	 * 
	 */
	public void testRMax() {
		final int numberOfTry = 3;
		final double knownValue = 1.0;
		RMaxTree rmaxTree = new RMaxTree(numberOfTry, variableOrdering);
		testRMax(numberOfTry, knownValue, rmaxTree);
	}

	/**
	 * 
	 * Test all the RMax algorithm
	 * 
	 * @param numberOfTry
	 *          the number of try before a value is known
	 * @param knownValue
	 *          the value used when it is known
	 * @param rmaxTree
	 *          the tree to test
	 * 
	 */
	private void testRMax(final int numberOfTry, final double knownValue,
			LearningTree<Double, Double> rmaxTree) {
		Context[] states = Context.createStateSet(EXAMPLES_SET_MAX_SIZE,
				Context.ValueSet.BINARY);

		rmaxTree.setVariableValueSet(createValueSet());

		LinkedHashMap<Context, Integer> knownLevel = new LinkedHashMap<Context, Integer>();

		for (Context state : states) {
			Example<Double> example = new Example<Double>(knownValue,
					variableOrdering.stateToList(state));
			int randomTryNumber = random.nextInt(numberOfTry) + 1;
			for (int tryIndex = 0; tryIndex < randomTryNumber; tryIndex++) {
				rmaxTree.learn(example);
				knownLevel.put(state, tryIndex + 1);

				checkRMaxTree(knownValue, numberOfTry, rmaxTree, states, knownLevel);
			}
		}
	}

	/**
	 * 
	 * Create a value set
	 * 
	 * @return the value set
	 * 
	 */
	private Map<Variable, Set<String>> createValueSet() {
		Map<Variable, Set<String>> variableValueSet = new LinkedHashMap<Variable, Set<String>>();
		for (int variableIndex = 0; variableIndex < EXAMPLES_SET_MAX_SIZE; variableIndex++) {
			Set<String> valueSet = new LinkedHashSet<String>();
			valueSet.add("0");
			valueSet.add("1");
			variableValueSet.put(Variable.getVariable(String.valueOf(variableIndex)),
					valueSet);
		}
		return variableValueSet;
	}

	/**
	 * 
	 * Check if the RMax tree correctly learn
	 * 
	 * @param knownValue
	 *          value learn for every branch
	 * @param numberOfTry
	 *          number of try to decide if a value is known
	 * @param rmaxTree
	 *          the tree to check
	 * @param states
	 *          the base of examples
	 * @param knownLevel
	 *          the base of examples tested
	 * 
	 */
	private void checkRMaxTree(double knownValue, int numberOfTry,
			LearningTree<Double, Double> rmaxTree, Context[] states,
			LinkedHashMap<Context, Integer> knownLevel) {
		double rmax = Parameters.getRMaxFromReward(knownValue);
		for (Context state : states) {
			Integer numberOfExamplesLearned = knownLevel.get(state);
			if ((numberOfExamplesLearned != null)
					&& (knownLevel.get(state) == numberOfTry)) {
				assertEquals(knownValue, rmaxTree.classify(state));
			} else {
				assertEquals(rmax, rmaxTree.classify(state));
			}
		}
	}

	/**
	 * 
	 * Test all the structure bonus learning algorithm
	 * 
	 */
	public void testStructureBonus() {
		final double rmaxValue = 10.0;
		final double betaK = 0.1;
		final double exampleValue = 1.0;
		final StructureLearningMeasure measure = new StructureLearningMeasure(
				rmaxValue, 1);
		final SID4Tree.Plain<Double> mbieebTree = new SID4Tree.Plain<Double>(
				measure, variableOrdering);
		mbieebTree.setVariableValueSet(createValueSet());

		Map<Context, Integer> countMap = new LinkedHashMap<Context, Integer>();
		Context[] states = Context.createStateSet(EXAMPLES_SET_MAX_SIZE,
				Context.ValueSet.BINARY);
		Random random = new Random(0);
		for (Context state : states) {
			if (random.nextBoolean()) {
				int numberOfTime = random.nextInt(5) + 1;
				for (int count = 1; count < numberOfTime; count++) {
					countMap.put(state, count);
					mbieebTree.learn(new Example<Double>(exampleValue, variableOrdering
							.stateToList(state)));
					checkStructureBonusTree(states, countMap, mbieebTree, exampleValue,
							betaK, rmaxValue);
				}
			}
		}
	}

	/**
	 * 
	 * Check if the tree learn correctly
	 * 
	 * @param states
	 *          set of states that may have learned
	 * @param countMap
	 *          number of times each example have been learned
	 * @param mbieebTree
	 *          tree learning the function
	 * @param exampleValue
	 *          value of an example when it is learn
	 * @param betaK
	 *          betaK parameter
	 * @param rmaxValue
	 *          the rmax value for unknown value
	 * 
	 */
	private void checkStructureBonusTree(Context[] states,
			Map<Context, Integer> countMap, SID4Tree<Double, Double> mbieebTree,
			double exampleValue, double betaK, double rmaxValue) {
		for (Context state : states) {
			Integer count = countMap.get(state);
			Leaf<Double> leaf = mbieebTree.getLeaf(state);
			if (count == null) {
				assertEquals(rmaxValue, leaf.get());
			} else {
				final double expectedValue = exampleValue;
				assertTrue(Math.abs(expectedValue - leaf.get()) < Parameters.EPSILON);
			}
		}
	}
}
